Type: Package
Package: flextable
Title: Functions for Tabular Reporting
Version: 0.9.12
Authors@R: c(
    person("David", "Gohel", , "david.gohel@ardata.fr", role = c("aut", "cre")),
    person("ArData", role = "cph"),
    person("Clementine", "Jager", role = "ctb"),
    person("Eli", "Daniels", role = "ctb"),
    person("Panagiotis", "Skintzos", , "panagiotis.skintzos@ardata.fr", role = "aut"),
    person("Quentin", "Fazilleau", role = "ctb"),
    person("Maxim", "Nazarov", role = "ctb"),
    person("Titouan", "Robert", role = "ctb"),
    person("Michael", "Barrowman", role = "ctb"),
    person("Atsushi", "Yasumoto", role = "ctb"),
    person("Paul", "Julian", role = "ctb"),
    person("Sean", "Browning", role = "ctb"),
    person("Rémi", "Thériault", role = "ctb",
           comment = c(ORCID = "0000-0003-4315-6788")),
    person("Samuel", "Jobert", role = "ctb"),
    person("Keith", "Newman", role = "ctb")
  )
Description: Use a grammar for creating and customizing pretty tables.
    The following formats are supported: 'HTML', 'PDF', 'RTF', 'Microsoft
    Word', 'Microsoft PowerPoint', R 'Grid Graphics' and 'patchwork'.
    'R Markdown', 'Quarto' and the package 'officer' can be used to
    produce the result files. The syntax is the same for the user
    regardless of the type of output to be produced. A set of functions
    allows the creation, definition of cell arrangement, addition of
    headers or footers, formatting and definition of cell content with
    text and or images. The package also offers a set of high-level
    functions that allow tabular reporting of statistical models and the
    creation of complex cross tabulations.
License: GPL-3
URL: https://ardata-fr.github.io/flextable-book/,
        https://davidgohel.github.io/flextable/
BugReports: https://github.com/davidgohel/flextable/issues
Imports: data.table (>= 1.13.0), gdtools (>= 0.5.0), graphics,
        grDevices, grid, htmltools, knitr, officer (>= 0.7.3), ragg,
        rlang, rmarkdown (>= 2.0), stats, utils, uuid (>= 0.1-4), xml2
Suggests: bookdown (>= 0.40), broom, broom.mixed, formatters, chromote,
        cluster, commonmark, doconv (>= 0.3.0), equatags, ggplot2,
        gtable, jsonlite, lme4, magick, mgcv, nlme, officedown,
        patchwork, pdftools, pkgdown (>= 2.0.0), rtables, scales,
        svglite, tables (>= 0.9.17), testthat (>= 3.0.0), webshot2,
        withr, xtable
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-02 05:37:13 UTC; davidgohel
Author: David Gohel [aut, cre],
  ArData [cph],
  Clementine Jager [ctb],
  Eli Daniels [ctb],
  Panagiotis Skintzos [aut],
  Quentin Fazilleau [ctb],
  Maxim Nazarov [ctb],
  Titouan Robert [ctb],
  Michael Barrowman [ctb],
  Atsushi Yasumoto [ctb],
  Paul Julian [ctb],
  Sean Browning [ctb],
  Rémi Thériault [ctb] (ORCID: <https://orcid.org/0000-0003-4315-6788>),
  Samuel Jobert [ctb],
  Keith Newman [ctb]
Maintainer: David Gohel <david.gohel@ardata.fr>
Repository: CRAN
Date/Publication: 2026-06-02 07:00:02 UTC
Built: R 4.6.0; ; 2026-06-02 11:51:08 UTC; unix

Type: Package
Package: flextable
Title: Functions for Tabular Reporting
Version: 0.10.0
Authors@R: c(
    person("David", "Gohel", , "david.gohel@ardata.fr", role = c("aut", "cre")),
    person("ArData", role = "cph"),
    person("Clementine", "Jager", role = "ctb"),
    person("Eli", "Daniels", role = "ctb"),
    person("Panagiotis", "Skintzos", , "panagiotis.skintzos@ardata.fr", role = "aut"),
    person("Quentin", "Fazilleau", role = "ctb"),
    person("Maxim", "Nazarov", role = "ctb"),
    person("Titouan", "Robert", role = "ctb"),
    person("Michael", "Barrowman", role = "ctb"),
    person("Atsushi", "Yasumoto", role = "ctb"),
    person("Paul", "Julian", role = "ctb"),
    person("Sean", "Browning", role = "ctb"),
    person("Rémi", "Thériault", role = "ctb",
           comment = c(ORCID = "0000-0003-4315-6788")),
    person("Samuel", "Jobert", role = "ctb"),
    person("Keith", "Newman", role = "ctb")
  )
Description: Use a grammar for creating and customizing pretty tables.
    The following formats are supported: 'HTML', 'PDF', 'Typst', 'RTF',
    'Microsoft Word', 'Microsoft PowerPoint', R 'Grid Graphics' and
    'patchwork'.  'R Markdown', 'Quarto' and the package 'officer' can be
    used to produce the result files. The syntax is the same for the user
    regardless of the type of output to be produced. A set of functions
    allows the creation, definition of cell arrangement, addition of
    headers or footers, formatting and definition of cell content with
    text and or images. The package also offers a set of high-level
    functions that allow tabular reporting of statistical models and the
    creation of complex cross tabulations.
License: GPL-3
URL: https://ardata-fr.github.io/flextable-book/,
        https://davidgohel.github.io/flextable/
BugReports: https://github.com/davidgohel/flextable/issues
Imports: data.table (>= 1.13.0), gdtools (>= 0.5.0), graphics,
        grDevices, grid, htmltools, knitr, officer (>= 0.7.5), ragg,
        rlang, rmarkdown (>= 2.0), stats, utils, uuid (>= 0.1-4), xml2
Suggests: bookdown (>= 0.40), broom, broom.mixed, chromote, cluster,
        commonmark, doconv (>= 0.3.0), equatags, formatters, ggplot2,
        gtable, jsonlite, lme4, magick, mgcv, nlme, officedown,
        patchwork, pdftools, pkgdown (>= 2.0.0), rtables, scales,
        svglite, tables (>= 0.9.17), testthat (>= 3.0.0), webshot2,
        withr, xtable
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-07-07 08:49:11 UTC; davidgohel
Author: David Gohel [aut, cre],
  ArData [cph],
  Clementine Jager [ctb],
  Eli Daniels [ctb],
  Panagiotis Skintzos [aut],
  Quentin Fazilleau [ctb],
  Maxim Nazarov [ctb],
  Titouan Robert [ctb],
  Michael Barrowman [ctb],
  Atsushi Yasumoto [ctb],
  Paul Julian [ctb],
  Sean Browning [ctb],
  Rémi Thériault [ctb] (ORCID: <https://orcid.org/0000-0003-4315-6788>),
  Samuel Jobert [ctb],
  Keith Newman [ctb]
Maintainer: David Gohel <david.gohel@ardata.fr>
Repository: CRAN
Date/Publication: 2026-07-07 11:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(as_flextable,ElementaryTable)
S3method(as_flextable,TableTree)
S3method(as_flextable,brmsfit)
S3method(as_flextable,compact_summary)
S3method(as_flextable,data.frame)
S3method(as_flextable,gam)
S3method(as_flextable,glm)
S3method(as_flextable,glmmTMB)
S3method(as_flextable,glmmadmb)
S3method(as_flextable,gls)
S3method(as_flextable,grouped_data)
S3method(as_flextable,htest)
S3method(as_flextable,kmeans)
S3method(as_flextable,lm)
S3method(as_flextable,lme)
S3method(as_flextable,merMod)
S3method(as_flextable,nlme)
S3method(as_flextable,pam)
S3method(as_flextable,summarizor)
S3method(as_flextable,table)
S3method(as_flextable,tabular)
S3method(as_flextable,tabulator)
S3method(as_flextable,xtable)
S3method(dim,flextable)
S3method(dim,flextableGrob)
S3method(drawDetails,flextableHighlightedGrob)
S3method(drawDetails,flextableRowPartGrob)
S3method(format_fun,Date)
S3method(format_fun,POSIXt)
S3method(format_fun,character)
S3method(format_fun,default)
S3method(format_fun,double)
S3method(format_fun,factor)
S3method(format_fun,integer)
S3method(format_fun,logical)
S3method(format_fun,pct)
S3method(knit_print,flextable)
S3method(knit_print,run_reference)
S3method(latex_container_str,latex_container_float)
S3method(latex_container_str,latex_container_none)
S3method(latex_container_str,latex_container_wrap)
S3method(makeContext,flextableGrob)
S3method(ph_with,flextable)
S3method(plot,flextable)
S3method(plot,flextableGrob)
S3method(print,flextable)
S3method(print,flextable_defaults)
S3method(print,tabulator)
S3method(rtf_add,flextable)
S3method(summary,tabulator)
S3method(to_html,flextable)
S3method(to_rtf,flextable)
S3method(to_wml,flextable)
export(add_body)
export(add_body_row)
export(add_footer)
export(add_footer_lines)
export(add_footer_row)
export(add_header)
export(add_header_lines)
export(add_header_row)
export(add_latex_dep)
export(align)
export(align_nottext_col)
export(align_text_col)
export(append_chunks)
export(as_b)
export(as_bracket)
export(as_chunk)
export(as_equation)
export(as_flextable)
export(as_grouped_data)
export(as_highlight)
export(as_i)
export(as_image)
export(as_paragraph)
export(as_qmd)
export(as_raster)
export(as_strike)
export(as_sub)
export(as_sup)
export(as_word_field)
export(autofit)
export(before)
export(bg)
export(body_add_flextable)
export(body_replace_flextable_at_bkm)
export(bold)
export(border)
export(border_inner)
export(border_inner_h)
export(border_inner_v)
export(border_outer)
export(border_remove)
export(chunk_dataframe)
export(colformat_char)
export(colformat_date)
export(colformat_datetime)
export(colformat_double)
export(colformat_image)
export(colformat_int)
export(colformat_lgl)
export(colformat_num)
export(color)
export(colorize)
export(compact_summary)
export(compose)
export(continuous_summary)
export(delete_columns)
export(delete_part)
export(delete_rows)
export(df_printer)
export(dim_pretty)
export(empty_blanks)
export(fit_columns)
export(fit_to_width)
export(fix_border_issues)
export(flextable)
export(flextable_dim)
export(flextable_html_dependency)
export(flextable_to_rmd)
export(fmt_2stats)
export(fmt_avg_dev)
export(fmt_dbl)
export(fmt_header_n)
export(fmt_int)
export(fmt_n_percent)
export(fmt_pct)
export(fmt_signif_after_zeros)
export(fmt_summarizor)
export(font)
export(fontsize)
export(footers_flextable_at_bkm)
export(footnote)
export(fp_border_default)
export(fp_text_default)
export(gen_grob)
export(get_flextable_defaults)
export(gg_chunk)
export(grid_chunk)
export(headers_flextable_at_bkm)
export(height)
export(height_all)
export(highlight)
export(hline)
export(hline_bottom)
export(hline_top)
export(hrule)
export(htmltools_value)
export(hyperlink_text)
export(information_data_cell)
export(information_data_chunk)
export(information_data_paragraph)
export(init_flextable_defaults)
export(italic)
export(keep_with_next)
export(labelizor)
export(line_spacing)
export(linerange)
export(lollipop)
export(merge_at)
export(merge_h)
export(merge_h_range)
export(merge_none)
export(merge_v)
export(minibar)
export(mk_par)
export(ncol_keys)
export(nrow_part)
export(padding)
export(paginate)
export(plot_chunk)
export(prepend_chunks)
export(proc_freq)
export(qflextable)
export(regulartable)
export(rotate)
export(save_as_docx)
export(save_as_html)
export(save_as_image)
export(save_as_pptx)
export(save_as_rtf)
export(separate_header)
export(set_caption)
export(set_flextable_defaults)
export(set_footer_df)
export(set_formatter)
export(set_formatter_type)
export(set_header_df)
export(set_header_labels)
export(set_table_properties)
export(shift_table)
export(split_columns)
export(split_rows)
export(split_to_pages)
export(style)
export(summarizor)
export(surround)
export(tab_settings)
export(tabulator)
export(tabulator_colnames)
export(theme_alafoli)
export(theme_apa)
export(theme_booktabs)
export(theme_borderless)
export(theme_box)
export(theme_tron)
export(theme_tron_legacy)
export(theme_vader)
export(theme_vanilla)
export(theme_zebra)
export(use_df_printer)
export(use_flextable_qmd)
export(use_model_printer)
export(valign)
export(vline)
export(vline_left)
export(vline_right)
export(void)
export(width)
export(wrap_flextable)
import(data.table)
importFrom(data.table,":=")
importFrom(data.table,.GRP)
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,dcast)
importFrom(data.table,fcase)
importFrom(data.table,fifelse)
importFrom(data.table,first)
importFrom(data.table,fsetdiff)
importFrom(data.table,groupingsets)
importFrom(data.table,is.data.table)
importFrom(data.table,rbindlist)
importFrom(data.table,rleid)
importFrom(data.table,rleidv)
importFrom(data.table,setDF)
importFrom(data.table,setDT)
importFrom(data.table,setcolorder)
importFrom(data.table,setkeyv)
importFrom(data.table,setnames)
importFrom(data.table,setorderv)
importFrom(data.table,shift)
importFrom(data.table,tstrsplit)
importFrom(gdtools,installed_gfonts)
importFrom(gdtools,strings_sizes)
importFrom(grDevices,as.raster)
importFrom(grDevices,col2rgb)
importFrom(grDevices,dev.list)
importFrom(grDevices,dev.off)
importFrom(grDevices,is.raster)
importFrom(grDevices,rgb)
importFrom(graphics,boxplot)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(grid,convertHeight)
importFrom(grid,convertWidth)
importFrom(grid,drawDetails)
importFrom(grid,gList)
importFrom(grid,gTree)
importFrom(grid,gpar)
importFrom(grid,grid.draw)
importFrom(grid,grid.layout)
importFrom(grid,grid.newpage)
importFrom(grid,grid.rect)
importFrom(grid,grobDescent)
importFrom(grid,grobHeight)
importFrom(grid,grobWidth)
importFrom(grid,is.grob)
importFrom(grid,makeContext)
importFrom(grid,popViewport)
importFrom(grid,pushViewport)
importFrom(grid,rasterGrob)
importFrom(grid,rectGrob)
importFrom(grid,segmentsGrob)
importFrom(grid,setChildren)
importFrom(grid,textGrob)
importFrom(grid,unit)
importFrom(grid,viewport)
importFrom(grid,vpStack)
importFrom(htmltools,HTML)
importFrom(htmltools,attachDependencies)
importFrom(htmltools,browsable)
importFrom(htmltools,div)
importFrom(htmltools,htmlDependency)
importFrom(htmltools,htmlEscape)
importFrom(htmltools,save_html)
importFrom(htmltools,tagList)
importFrom(htmltools,tags)
importFrom(knitr,asis_output)
importFrom(knitr,fig_path)
importFrom(knitr,include_graphics)
importFrom(knitr,is_html_output)
importFrom(knitr,is_latex_output)
importFrom(knitr,knit_child)
importFrom(knitr,knit_meta_add)
importFrom(knitr,knit_print)
importFrom(knitr,opts_current)
importFrom(knitr,opts_knit)
importFrom(knitr,pandoc_to)
importFrom(knitr,raw_block)
importFrom(knitr,raw_html)
importFrom(knitr,raw_latex)
importFrom(officer,add_slide)
importFrom(officer,body_add_par)
importFrom(officer,body_add_xml)
importFrom(officer,body_set_default_section)
importFrom(officer,cursor_bookmark)
importFrom(officer,fp_border)
importFrom(officer,fp_cell)
importFrom(officer,fp_par)
importFrom(officer,fp_par_lite)
importFrom(officer,fp_tab)
importFrom(officer,fp_tabs)
importFrom(officer,fp_text)
importFrom(officer,fp_text_lite)
importFrom(officer,fpar)
importFrom(officer,ftext)
importFrom(officer,image_to_base64)
importFrom(officer,officer_url_encode)
importFrom(officer,opts_current_table)
importFrom(officer,page_mar)
importFrom(officer,page_size)
importFrom(officer,ph_location_type)
importFrom(officer,ph_with)
importFrom(officer,prop_section)
importFrom(officer,prop_table)
importFrom(officer,read_docx)
importFrom(officer,read_pptx)
importFrom(officer,rtf_add)
importFrom(officer,rtf_doc)
importFrom(officer,run_autonum)
importFrom(officer,run_bookmark)
importFrom(officer,str_encode_to_rtf)
importFrom(officer,table_colwidths)
importFrom(officer,table_layout)
importFrom(officer,table_width)
importFrom(officer,to_html)
importFrom(officer,to_rtf)
importFrom(officer,to_wml)
importFrom(ragg,agg_capture)
importFrom(ragg,agg_png)
importFrom(rlang,"%||%")
importFrom(rlang,call_args)
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,eval_tidy)
importFrom(rlang,get_expr)
importFrom(rlang,new_function)
importFrom(rlang,pairlist2)
importFrom(rlang,quo)
importFrom(rlang,quo_name)
importFrom(rlang,quo_text)
importFrom(rlang,sym)
importFrom(rmarkdown,html_document)
importFrom(rmarkdown,latex_dependency)
importFrom(rmarkdown,pandoc_available)
importFrom(rmarkdown,pandoc_version)
importFrom(rmarkdown,pdf_document)
importFrom(rmarkdown,render)
importFrom(stats,AIC)
importFrom(stats,BIC)
importFrom(stats,IQR)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,density)
importFrom(stats,df.residual)
importFrom(stats,logLik)
importFrom(stats,mad)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,naprint)
importFrom(stats,nobs)
importFrom(stats,quantile)
importFrom(stats,reshape)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(utils,browseURL)
importFrom(utils,compareVersion)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,packageVersion)
importFrom(utils,str)
importFrom(utils,tail)
importFrom(uuid,UUIDgenerate)
importFrom(xml2,as_xml_document)

flextable R package
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

[![R build
status](https://github.com/davidgohel/flextable/workflows/R-CMD-check/badge.svg)](https://github.com/davidgohel/flextable/actions)
[![version](https://www.r-pkg.org/badges/version/flextable)](https://CRAN.R-project.org/package=flextable)
[![test
coverage](https://codecov.io/gh/davidgohel/flextable/branch/main/graph/badge.svg)](https://app.codecov.io/gh/davidgohel/flextable)
![Active](https://www.repostatus.org/badges/latest/active.svg)

The flextable package provides a framework for easily creating tables
for reporting and publications. Tables can be formatted with a set of
verbs such as `bold()`, `color()`, they can receive a header of more
than one line, cells can be merged or contain an image. The package
makes it possible to build any table for publication from a `data.frame`
and provides convenience function `as_flextable()` to convert several R
objects to a flextable, such as an object returned from `table()` or a
model.

Tables can be embedded within HTML, PDF, Word and PowerPoint documents
from R Markdown and Quarto documents and within RTF or Microsoft Word or
PowerPoint documents with package
[officer](https://davidgohel.github.io/officer/). Tables can also be
exported as image files (png, svg) or combined with ggplot2 plots in
patchwork layouts.

``` r
flextable(mtcars) |>
  theme_vanilla() |>
  save_as_docx(path = "mytable.docx")
```

A `flextable` object is a data.frame representation. An API is available
to let R users create tables for reporting and control their formatting
properties and their layout. The package provides functions that give
control over:

- header, body and footer content
- text, paragraphs, cells and border formatting of any element
- displayed values

The package also offers a set of high-level functions that allow tabular
reporting of statistical models and the creation of complex cross
tabulations.

## Examples

``` r
library(flextable)

flextable(head(airquality))
```

<img src="man/figures/README-simple-example-1.png" alt="" width="432" />

Formatting can be layered on with a set of functions:

``` r
flextable(head(mtcars)) |>
  highlight(i = ~ mpg < 22, j = "disp", color = "#ffe842") |>
  bg(
    j = c("hp", "drat", "wt"),
    bg = scales::col_quantile(palette = c("wheat", "red"), domain = NULL)
  ) |>
  add_footer_lines("The 'mtcars' dataset")
```

<img src="man/figures/README-mtcars-dataset-1.png" alt="" width="792" />

``` r
ggplot2::diamonds[, c("cut", "carat", "price", "clarity", "table")] |>
  summarizor(by = c("cut")) |>
  as_flextable(spread_first_col = TRUE)
```

<img src="man/figures/README-diamonds-dataset-1.png" alt="" width="740" />

## Installation

``` r
install.packages("flextable")
```

You can get the development version from GitHub:

``` r
devtools::install_github("davidgohel/flextable")
```

## Resources

- [User guide](https://ardata-fr.github.io/flextable-book/)
- [Reference
  manual](https://davidgohel.github.io/flextable/reference/index.html)
- [Gallery of examples](https://ardata.fr/en/flextable-gallery/)

## Bug reports

When you file a [bug
report](https://github.com/davidgohel/flextable/issues), please spend
some time making it easy for us to reproduce. If you take the time to
make the bug report consistent, it will be easier to fix.

flextable R package
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

[![R build
status](https://github.com/davidgohel/flextable/workflows/R-CMD-check/badge.svg)](https://github.com/davidgohel/flextable/actions)
[![version](https://www.r-pkg.org/badges/version/flextable)](https://CRAN.R-project.org/package=flextable)
[![test
coverage](https://codecov.io/gh/davidgohel/flextable/branch/main/graph/badge.svg)](https://app.codecov.io/gh/davidgohel/flextable)
![Active](https://www.repostatus.org/badges/latest/active.svg)

The flextable package provides a framework for easily creating tables
for reporting and publications. Tables can be formatted with a set of
verbs such as `bold()`, `color()`, they can receive a header of more
than one line, cells can be merged or contain an image. The package
makes it possible to build any table for publication from a `data.frame`
and provides convenience function `as_flextable()` to convert several R
objects to a flextable, such as an object returned from `table()` or a
model.

Tables can be embedded within HTML, PDF, Word and PowerPoint documents
from R Markdown and Quarto documents (plus Typst with Quarto) and within
RTF or Microsoft Word or PowerPoint documents with package
[officer](https://davidgohel.github.io/officer/). Tables can also be
exported as image files (png, svg) or combined with ggplot2 plots in
patchwork layouts.

``` r
flextable(mtcars) |>
  theme_vanilla() |>
  save_as_docx(path = "mytable.docx")
```

A `flextable` object is a data.frame representation. An API is available
to let R users create tables for reporting and control their formatting
properties and their layout. The package provides functions that give
control over:

- header, body and footer content
- text, paragraphs, cells and border formatting of any element
- displayed values

The package also offers a set of high-level functions that allow tabular
reporting of statistical models and the creation of complex cross
tabulations.

## Examples

``` r
library(flextable)

flextable(head(airquality))
```

<img src="man/figures/README-simple-example-1.png" alt="" width="432" />

Formatting can be layered on with a set of functions:

``` r
flextable(head(mtcars)) |>
  highlight(i = ~ mpg < 22, j = "disp", color = "#ffe842") |>
  bg(
    j = c("hp", "drat", "wt"),
    bg = scales::col_quantile(palette = c("wheat", "red"), domain = NULL)
  ) |>
  add_footer_lines("The 'mtcars' dataset")
```

<img src="man/figures/README-mtcars-dataset-1.png" alt="" width="792" />

``` r
ggplot2::diamonds[, c("cut", "carat", "price", "clarity", "table")] |>
  summarizor(by = c("cut")) |>
  as_flextable(spread_first_col = TRUE)
```

<img src="man/figures/README-diamonds-dataset-1.png" alt="" width="740" />

## Installation

``` r
install.packages("flextable")
```

You can get the development version from GitHub:

``` r
devtools::install_github("davidgohel/flextable")
```

## Resources

- [User guide](https://ardata-fr.github.io/flextable-book/)
- [Reference
  manual](https://davidgohel.github.io/flextable/reference/index.html)
- [Gallery of examples](https://ardata.fr/en/flextable-gallery/)

## Bug reports

When you file a [bug
report](https://github.com/davidgohel/flextable/issues), please spend
some time making it easy for us to reproduce. If you take the time to
make the bug report consistent, it will be easier to fix.

#' @title Merge flextable cells vertically
#'
#' @description Merge flextable cells vertically when consecutive cells have
#' identical values. Text of formatted values are used to compare
#' values if available.
#'
#' Two options are available, either a column-by-column algorithm or an
#' algorithm where the combinations of these columns are used once for all
#' target columns.
#'
#' @param x `flextable` object
#' @param j column to used to find consecutive values to be merged. Columns
#' from orignal dataset can also be used.
#' @param target columns names where cells have to be merged.
#' @param part partname of the table where merge has to be done.
#' @param combine If the value is TRUE, the columns defined by `j` will
#' be combined into a single column/value and the consecutive values of
#' this result will be used. Otherwise, the columns are inspected one
#' by one to perform cell merges.
#' @examples
#' ft_merge <- flextable(mtcars)
#' ft_merge <- merge_v(ft_merge, j = c("gear", "carb"))
#' ft_merge
#'
#' data_ex <- structure(list(srdr_id = c(
#'   "175124", "175124", "172525", "172525",
#'   "172545", "172545", "172609", "172609", "172609"
#' ), substances = c(
#'   "alcohol",
#'   "alcohol", "alcohol", "alcohol", "cannabis",
#'   "cannabis", "alcohol\n cannabis\n other drugs",
#'   "alcohol\n cannabis\n other drugs",
#'   "alcohol\n cannabis\n other drugs"
#' ), full_name = c(
#'   "TAU", "MI", "TAU", "MI (parent)", "TAU", "MI",
#'   "TAU", "MI", "MI"
#' ), article_arm_name = c(
#'   "Control", "WISEteens",
#'   "Treatment as usual", "Brief MI (b-MI)", "Assessed control",
#'   "Intervention", "Control", "Computer BI", "Therapist BI"
#' )), row.names = c(
#'   NA,
#'   -9L
#' ), class = c("tbl_df", "tbl", "data.frame"))
#'
#'
#' ft_1 <- flextable(data_ex)
#' ft_1 <- theme_box(ft_1)
#' ft_2 <- merge_v(ft_1,
#'   j = "srdr_id",
#'   target = c("srdr_id", "substances")
#' )
#' ft_2
#' @family flextable merging function
#' @export
merge_v <- function(x, j = NULL, target = NULL, part = "body", combine = FALSE) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "merge_v()"))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  j <- as_col_keys(x[[part]], j, blanks = x$blanks)

  if (!is.null(target)) {
    target <- as_col_keys(x[[part]], target, blanks = character())
  } else {
    target <- as_col_keys(x[[part]], j, blanks = character())
  }

  x[[part]] <- span_columns(x = x[[part]], columns = j, target = target, combine = combine)

  x
}



#' @title Merge flextable cells horizontally
#'
#' @description Merge flextable cells horizontally when consecutive cells have
#' identical values. Text of formatted values are used to compare
#' values.
#'
#' @param x `flextable` object
#' @param i rows where cells have to be merged.
#' @param part partname of the table where merge has to be done.
#' @family flextable merging function
#' @examples
#' dummy_df <- data.frame(
#'   col1 = letters,
#'   col2 = letters, stringsAsFactors = FALSE
#' )
#' ft_merge <- flextable(dummy_df)
#' ft_merge <- merge_h(x = ft_merge)
#' ft_merge
#' @export
merge_h <- function(x, i = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "merge_h()"))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  i <- get_rows_id(x[[part]], i)
  x[[part]] <- span_rows(x = x[[part]], rows = i)

  x
}


#' @title Delete flextable merging informations
#'
#' @description Delete all merging informations from a flextable.
#'
#' @param x `flextable` object
#' @param part partname of the table where merge has to be done.
#' @family flextable merging function
#' @export
#' @examples
#' typology <- data.frame(
#'   col_keys = c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Species"),
#'   what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
#'   measure = c("Length", "Width", "Length", "Width", "Species"),
#'   stringsAsFactors = FALSE
#' )
#'
#' ft <- flextable(head(iris))
#' ft <- set_header_df(ft, mapping = typology, key = "col_keys")
#' ft <- merge_v(ft, j = c("Species"))
#'
#' ft <- theme_tron_legacy(merge_none(ft))
#' ft
merge_none <- function(x, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "merge_none()"))
  }
  part <- match.arg(part, c("all", "body", "header", "footer"), several.ok = FALSE)

  if (part == "all") {
    args <- list()
    for (p in c("header", "footer", "body")) {
      x <- merge_none(x = x, part = p)
    }
  }

  x[[part]] <- span_free(x[[part]])

  x
}





#' @title Merge flextable cells into a single one
#'
#' @description Merge flextable cells into a single one. All
#' rows and columns must be consecutive.
#'
#' @param x `flextable` object
#' @param i,j columns and rows to merge
#' @param part partname of the table where merge has to be done.
#' @family flextable merging function
#' @examples
#' ft_merge <- flextable(head(mtcars), cwidth = .5)
#' ft_merge <- merge_at(ft_merge, i = 1:2, j = 1:2)
#' ft_merge
#' @export
merge_at <- function(x, i = NULL, j = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "merge_at()"))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  j <- get_columns_id(x[[part]], j = j)
  j <- x$col_keys[j]

  i <- get_rows_id(x[[part]], i)

  x[[part]] <- span_cells_at(x = x[[part]], columns = j, rows = i)

  x
}


#' @title Rowwise merge of a range of columns
#'
#' @description Merge flextable columns into a single one for each selected rows. All
#' columns must be consecutive.
#'
#' @param x `flextable` object
#' @param i selected rows
#' @param j1,j2 selected columns that will define the range of columns to merge.
#' @param part partname of the table where merge has to be done.
#' @family flextable merging function
#' @examples
#' ft <- flextable(head(mtcars), cwidth = .5)
#' ft <- theme_box(ft)
#' ft <- merge_h_range(ft, i = ~ cyl == 6, j1 = "am", j2 = "carb")
#' ft <- flextable::align(ft, i = ~ cyl == 6, align = "center")
#' ft
#' @export
merge_h_range <- function(x, i = NULL, j1 = NULL, j2 = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "merge_h_range()"))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  j1 <- get_columns_id(x[[part]], j = j1)
  j2 <- get_columns_id(x[[part]], j = j2)

  seq_cols <- j1:j2

  i <- get_rows_id(x[[part]], i)
  x[[part]]$spans$rows[i, seq_cols] <- 0
  x[[part]]$spans$rows[i, j1] <- length(seq_cols)
  check_merge(x[[part]])


  x
}

gdtools::register_liberationsans()

init_flextable_defaults()

set_flextable_defaults(
  font.family = "Liberation Sans",
  border.color = "#333333"
)


test_that("png is created", {
  ft <- as_flextable(iris)
  file <- tempfile(fileext = ".png")
  try(invisible(save_as_image(x = ft, path = file, res = 150)), silent = TRUE)
  expect_true(file.exists(file))
  expect_gt(file.info(file)$size, 20000)
})

test_that("merged borders", {
  local_edition(3)

  dat <- data.frame(a = c(1, 1, 2, 2, 5), b = 6:10)

  ft <- flextable(dat)
  ft <- merge_v(ft, ~a, part = "body")
  ft <- hline(
    x = ft,
    i = 2,
    part = "body",
    border = fp_border(color = "red")
  )

  gr <- gen_grob(ft)

  expect_length(gr$children, 10)

  expect_equal(gr$children[[3]]$children$borders$children[[1]]$gp$col, "red")

  expect_length(gr$children[[1]]$children$borders$children, 2)
  expect_equal(
    gr$children[[1]]$children$borders$children[[1]]$gp$col,
    "#333333"
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[1]]$x0,
    grid::unit(0, "npc")
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[1]]$x1,
    grid::unit(1, "npc")
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[1]]$y0,
    grid::unit(1, "npc")
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[1]]$y1,
    grid::unit(1, "npc")
  )

  expect_equal(
    gr$children[[1]]$children$borders$children[[2]]$gp$col,
    "#333333"
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[2]]$x0,
    grid::unit(0, "npc")
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[2]]$x1,
    grid::unit(1, "npc")
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[2]]$y0,
    grid::unit(0, "npc")
  )
  expect_equal(
    gr$children[[1]]$children$borders$children[[2]]$y1,
    grid::unit(0, "npc")
  )

  expect_length(gr$children[[10]]$children$borders$children, 1)

  expect_equal(
    gr$children[[10]]$children$borders$children[[1]]$gp$col,
    "#333333"
  )
  expect_equal(
    gr$children[[10]]$children$borders$children[[1]]$x0,
    grid::unit(0, "npc")
  )
  expect_equal(
    gr$children[[10]]$children$borders$children[[1]]$x1,
    grid::unit(1, "npc")
  )
  expect_equal(
    gr$children[[10]]$children$borders$children[[1]]$y0,
    grid::unit(0, "npc")
  )
  expect_equal(
    gr$children[[10]]$children$borders$children[[1]]$y1,
    grid::unit(0, "npc")
  )
})

test_that("text wrapping", {
  text <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."
  source1 <- "DATA_SOURCE_A.COURSE_TITLE\nDATA_SOURCE_A.SUBJECT_DESCR\nDATA_SOURCE_A.CATALOG_NUMBER"
  source2 <- "DATA_SOURCE_A.GRADING_BASIS\nDATA_SOURCE_A.OFFICIAL_GRADE\nDATA_SOURCE_B.STUDENT_GROUP"

  temp_dat <- data.frame(
    label = c("Sources", "", "Notes"),
    col1 = c(source1, "", text),
    col2 = c(source2, "", text)
  )

  # Create table
  ft <- flextable(temp_dat)
  ft <- merge_h(ft, part = "body")

  gr <- gen_grob(ft, fit = "fixed")

  expect_length(gr$children, 9)
  expect_equal(gr$children[[5]]$children$contents$ftgrobs[[1]]$label, source1)
  expect_equal(gr$children[[6]]$children$contents$ftgrobs[[1]]$label, source2)

  # check wrap on 3 rows
  expect_length(gr$children[[5]]$children$contents$children, 3)
  expect_length(gr$children[[6]]$children$contents$children, 3)
  expect_equal(gr$children[[8]]$children$contents$ftgrobs[[1]]$label, "Notes")
  expect_length(gr$children[[8]]$children$contents$children, 1)
  expect_equal(gr$children[[9]]$children$contents$ftgrobs[[1]]$label, text)
  # check wrap on 3 rows
  expect_length(gr$children[[9]]$children$contents$children, 3)

  # check that height and width are greater than those of smaller cells
  expect_gt(
    gr$children$cell_2_2$children$contents$ftpar$height,
    gr$children$cell_1_2$children$contents$ftpar$height
  )
  expect_gt(
    gr$children$cell_2_2$children$contents$ftpar$width,
    gr$children$cell_2_1$children$contents$ftpar$width
  )
})

test_that("grid with raster", {
  skip_if_not_installed("magick")

  img.file <- file.path(
    R.home("doc"),
    "html",
    "logo.jpg"
  )
  myft <- flextable(head(iris))
  myft <- prepend_chunks(
    x = myft,
    i = 1:2,
    j = 1,
    as_image(src = img.file),
    part = "body"
  )
  ft <- autofit(myft)

  gr <- gen_grob(ft)

  expect_s3_class(gr$children[[6]]$children$contents$ftgrobs[[1]], "rastergrob")
  expect_s3_class(gr$children[[6]]$children$contents$ftgrobs[[2]], "text")
  expect_s3_class(
    gr$children[[11]]$children$contents$ftgrobs[[1]],
    "rastergrob"
  )
  expect_s3_class(gr$children[[11]]$children$contents$ftgrobs[[2]], "text")
  expect_s3_class(gr$children[[12]]$children$contents$ftgrobs[[1]], "text")
})

ft <- flextable(iris)

test_that("print as log", {
  expect_output(print(ft, preview = "log"), "a flextable object")
  expect_output(print(ft, preview = "log"), "header has 1 row")
  expect_output(print(ft, preview = "log"), "body has 150 row")
})


test_that("data selectors", {
  ft <- flextable(
    data = iris,
    col_keys = c("ouch", "Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Species", "blop")
  )

  expect_equal(
    flextable:::as_col_keys(ft$body, 2, blanks = ft$blanks),
    "Sepal.Width"
  )

  expect_equal(
    flextable:::as_col_keys(ft$body, -5, blanks = ft$blanks),
    c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  )

  expect_equal(
    flextable:::as_col_keys(ft$body, c(1, 2), blanks = ft$blanks),
    c("Sepal.Length", "Sepal.Width")
  )

  expect_equal(
    flextable:::as_col_keys(ft$body, NULL, blanks = ft$blanks),
    colnames(iris)
  )

  expect_equal(
    flextable:::as_col_keys(ft$body, c(TRUE, FALSE, TRUE, FALSE, TRUE),
      blanks = ft$blanks
    ),
    c("Sepal.Length", "Petal.Length", "Species")
  )

  expect_warning(
    flextable:::as_col_keys(ft$body, "Julio-Iglesias", blanks = ft$blanks)
  )
})

test_that("selection and merge_v", {
  ft <- flextable(
    data = iris[98:103, ],
    col_keys = c("aaa", "Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  )
  ft <- theme_box(ft)
  ft <- merge_v(ft, target = "aaa", j = "Species")

  expect_equal(
    ft$body$spans$columns[, 1],
    c(3L, 0L, 0L, 3L, 0L, 0L)
  )

  expect_warning(merge_v(ft, target = "aaa", j = "zzz"))
  expect_error(merge_v(ft, target = "Species", j = "Sepal.Width"))
})

test_that("selection and colors", {
  colourer <- function(z) {
    x <- rep("pink", length(z))
    x[is.na(z)] <- "#999999"
    w_avg <- which(z < mean(z, na.rm = TRUE))
    x[w_avg] <- "cyan"
    x
  }

  dat <- iris[98:103, ]
  dat[1, 1] <- NA
  dat[2, 2] <- NA
  dat[3, 3] <- NA
  dat[4, 4] <- NA
  ft <- flextable(data = dat)
  ft <- theme_box(ft)
  ft <- bg(ft, j = ~ . - Species, bg = colourer)

  expected_values <- c(
    "#999999", "cyan", "cyan", "pink", "cyan", "pink",
    "cyan", "#999999", "cyan", "pink", "cyan", "pink", "cyan", "cyan",
    "#999999", "pink", "pink", "pink", "cyan", "cyan", "cyan", "#999999",
    "pink", "pink", "transparent", "transparent", "transparent",
    "transparent", "transparent", "transparent"
  )
  bg_values <- as.vector(ft$body$styles$cells$background.color$data)
  expect_equal(bg_values, expected_values)

  ft <- bg(ft,
    source = "Species", j = "Sepal.Length",
    bg = function(z) {
      x <- rep("red", length(z))
      x[is.na(z)] <- "#999999"
      w_ver <- which(z %in% "versicolor")
      x[w_ver] <- "blue"
      x
    }
  )
  bg_values <- as.vector(ft$body$styles$cells$background.color$data[, 1])
  expect_equal(bg_values, rep(c("blue", "red"), each = 3))
})

test_that("identical values within columns are merged", {
  dummy_df <- data.frame(
    values = rep(letters[1:3], each = 2),
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  ft <- merge_v(x = ft, j = "values")
  expect_equal(ft$body$spans$columns[, 1], rep(c(2, 0), 3), ignore_attr = TRUE)
})

test_that("identical values within rows are merged", {
  dummy_df <- data.frame(
    col1 = letters,
    col2 = letters,
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  ft <- merge_h(x = ft)
  ref <- matrix(c(rep(2, 26), rep(0, 26)), ncol = 2)
  expect_equal(ft$body$spans$rows, ref, ignore_attr = TRUE)
})


test_that("span at", {
  dummy_df <- data.frame(
    col1 = letters,
    col2 = letters,
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  ft <- merge_at(x = ft, i = 1:4, j = 1:2)
  ref <- matrix(c(rep(2, 4), rep(1, 22), rep(0, 4), rep(1, 22)), ncol = 2)
  expect_equal(ft$body$spans$rows, ref, ignore_attr = TRUE)
  ref <- matrix(c(4, rep(0, 3), rep(1, 22), 4, rep(0, 3), rep(1, 22)), ncol = 2)
  expect_equal(ft$body$spans$columns, ref, ignore_attr = TRUE)
})


test_that("merged cells can be un-merged", {
  dummy_df <- data.frame(
    col1 = rep("a", 5),
    col2 = rep("a", 5),
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  ft <- merge_h(x = ft)
  expect_true(all(ft$body$spans$rows[, 1] == 2))
  ft <- merge_none(ft)
  expect_true(all(ft$body$spans$rows == 1))
  ft <- merge_v(x = ft)
  expect_true(all(ft$body$spans$columns[1, ] == 5))
  expect_true(all(ft$body$spans$columns[-1, ] == 0))
  ft <- merge_none(ft)
  expect_true(all(ft$body$spans$columns == 1))
})


test_that("delete_rows preserves vertical spans when deletion does not break them", {
  ft <- flextable(data.frame(
    col1 = rep(letters[1:3], each = 2),
    col2 = 1:6,
    stringsAsFactors = FALSE
  ))
  ft <- merge_v(ft, j = "col1")
  expect_equal(
    ft$body$spans$columns[, 1],
    c(2, 0, 2, 0, 2, 0),
    ignore_attr = TRUE
  )

  # delete rows 5:6 (the "c" group) — no span is broken
  ft2 <- delete_rows(ft, i = 5:6)
  expect_equal(ft2$body$spans$columns[, 1], c(2, 0, 2, 0), ignore_attr = TRUE)
})

test_that("delete_rows resets spans when deletion breaks a vertical span", {
  ft <- flextable(data.frame(
    col1 = rep(letters[1:3], each = 2),
    col2 = 1:6,
    stringsAsFactors = FALSE
  ))
  ft <- merge_v(ft, j = "col1")

  # delete row 3 — breaks the "b" group (rows 3:4)
  ft2 <- delete_rows(ft, i = 3)
  expect_true(all(ft2$body$spans$columns == 1))
  expect_true(all(ft2$body$spans$rows == 1))
})

test_that("delete_columns preserves horizontal spans when deletion does not break them", {
  ft <- flextable(data.frame(
    col1 = c("x", "y"),
    col2 = c("x", "y"),
    col3 = c("a", "b"),
    stringsAsFactors = FALSE
  ))
  ft <- merge_h(ft)
  expect_equal(ft$body$spans$rows[1, ], c(2, 0, 1), ignore_attr = TRUE)

  # delete col3 — no horizontal span is broken
  ft2 <- delete_columns(ft, j = "col3")
  expect_equal(ft2$body$spans$rows[1, ], c(2, 0), ignore_attr = TRUE)
})

test_that("delete_columns resets spans when deletion breaks a horizontal span", {
  ft <- flextable(data.frame(
    col1 = c("x", "y"),
    col2 = c("x", "y"),
    col3 = c("a", "b"),
    stringsAsFactors = FALSE
  ))
  ft <- merge_h(ft)

  # delete col1 — breaks the horizontal span col1+col2
  ft2 <- delete_columns(ft, j = "col1")
  expect_true(all(ft2$body$spans$columns == 1))
  expect_true(all(ft2$body$spans$rows == 1))
})

c_if <- function(...) {
  by <- c(...)
  by <- Filter(function(z) !is.null(z), by)
  by <- na.omit(by)
  as.character(by)
}


full_shift_structure <- function(x, cn_is_baseline, baseline_identifier,
                                 cn_treatment, cn_lab_cat, cn_visit,
                                 cn_baseline, cn_grade,
                                 grade_levels) {
  DAT_UCAT <- as.data.table(x)
  DAT_UCAT <- DAT_UCAT[!DAT_UCAT[[cn_is_baseline]] %in% baseline_identifier, ]
  DAT_UCAT <- DAT_UCAT[, .SD, .SDcols = c_if(cn_treatment, cn_lab_cat, cn_visit)]
  DAT_UCAT <- unique(DAT_UCAT)
  DAT_UCAT$DUMMY_JOIN <- 1L

  RANGE_LEVELS_BASELINE <- c(grade_levels, "MISSING", "SUM")
  RANGE_LEVELS_SHIFT <- c(grade_levels, "MISSING")
  DAT_UBS <- expand.grid(
    X = RANGE_LEVELS_BASELINE,
    Y = RANGE_LEVELS_SHIFT,
    stringsAsFactors = FALSE
  )
  names(DAT_UBS) <- c(cn_baseline, cn_grade)
  DAT_UBS$DUMMY_JOIN <- 1L

  PIVOT_DATA <- merge(
    x = DAT_UCAT, y = DAT_UBS, by = "DUMMY_JOIN",
    allow.cartesian = TRUE,
    all.x = TRUE, all.y = TRUE
  )
  PIVOT_DATA$DUMMY_JOIN <- NULL
  setorderv(PIVOT_DATA, cols = c_if(cn_treatment, cn_lab_cat, cn_visit, cn_baseline, cn_grade))
  setDF(PIVOT_DATA)
  PIVOT_DATA
}

pivot_baseline <- function(x, cn_is_baseline, baseline_identifier, cn_visit, cn_usubjid, cn_grade, cn_lab_cat, cn_treatment, cn_baseline) {
  BASELINE <- as.data.table(x)
  BASELINE <- BASELINE[BASELINE[[cn_is_baseline]] %in% baseline_identifier, ]
  BASELINE <- BASELINE[, .SD, .SDcols = c_if(cn_usubjid, cn_grade, cn_lab_cat, cn_treatment)]
  setnames(BASELINE, old = cn_grade, new = cn_baseline)

  DATAMART <- as.data.table(x)
  DATAMART <- DATAMART[!DATAMART[[cn_is_baseline]] %in% baseline_identifier, ]
  DATAMART <- DATAMART[, .SD, .SDcols = c_if(cn_usubjid, cn_visit, cn_grade, cn_lab_cat, cn_treatment)]
  DATAMART <- merge(DATAMART, BASELINE, by = c_if(cn_lab_cat, cn_treatment, cn_usubjid))
  setDF(DATAMART)
  DATAMART
}

#' @importFrom rlang new_function pairlist2
facfun_visit <- function(x, cn_visit = "VISIT", cn_visit_num = "VISITNUM") {
  LEV_VISIT <- as.data.table(x)
  LEV_VISIT <- LEV_VISIT[, .SD, .SDcols = c(cn_visit, cn_visit_num)]
  LEV_VISIT <- unique(LEV_VISIT)
  setorderv(LEV_VISIT, cols = cn_visit_num)

  levels_visit <- LEV_VISIT$VISIT

  ## function creation
  visit_as_factor <- new_function(
    pairlist2(x = , levels = levels_visit),
    quote(factor(x, levels = levels))
  )
  visit_as_factor
}

facfun_grade <- function(x, grade_levels = c("LOW", "NORMAL", "HIGH"), grade_labels = grade_levels) {
  new_function(
    # categories "MISSING" and "SUM" are added
    pairlist2(
      x = ,
      levels = c(grade_levels, "MISSING", "SUM"),
      labels = c(grade_labels, "Missing", "Sum")
    ),
    quote(factor(x, levels = levels, labels = labels))
  )
}

#' @export
#' @title Create a shift table
#' @description Create a shift table ready to be used with `tabulator()`.
#'
#' The function is transforming a dataset representing some
#' 'Laboratory Tests Results' structured as *CDISC clinical trial
#' data sets* format to a dataset representing the shift table.
#'
#' Shift tables are tables used in clinical trial analysis.
#' They show the progression of change from the baseline, with the progression
#' often being along time; the number of subjects is displayed in different
#' range (e.g. low, normal, or high) at baseline and at selected time points
#' or intervals.
#' @param x Laboratory Tests Results data frame.
#' @param cn_visit column name containing visit names, default to "VISIT".
#' @param cn_visit_num column name containing visit numbers, default to "VISITNUM".
#' @param cn_grade column name containing reference range indicators, default to "LBNRIND".
#' @param cn_usubjid column name containing unique subject inditifiers, default to "USUBJID".
#' @param cn_lab_cat column name containing lab tests or examination names, default to "LBTEST".
#' @param cn_is_baseline column name containing baseline flags, default to "LBBLFL".
#' @param baseline_identifier baseline flag value to use for baseline
#' identification. Its default is "Y".
#' @param cn_treatment column name containing treatment names, default to `NA`.
#' @param grade_levels levels to use for reference range indicators
#' @param grade_labels labels to use for reference range indicators
#' @family tools for clinical reporting
#' @return the shift table as a data.frame. Additionnal elements are provided
#' in attributes:
#'
#' - "VISIT_N": count of unique subject id per visits, labs and eventually
#' treatments. This element is supposed to be used as value for argument
#' `hidden_data` of function `tabulator()`.
#' - "FUN_VISIT": a utility function to easily turn *visit* column as a factor
#' column. It should be applied after the shift table creation.
#' - "FUN_GRADE": a utility function to easily turn *grade* column as a factor
#' column. It adds "MISSING/Missing" and "SUM/Sum" at the end of the set of values specified
#' in arguments `grade_levels` and `grade_labels`. It should be applied after the shift
#' table creation.
#' @examples
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' library(data.table)
#' library(flextable)
#'
#' # data simulation ----
#' USUBJID <- sprintf("01-ABC-%04.0f", 1:200)
#' VISITS <- c("SCREENING 1", "WEEK 2", "MONTH 3")
#' LBTEST <- c("Albumin", "Sodium")
#'
#' VISITNUM <- seq_along(VISITS)
#' LBBLFL <- rep(NA_character_, length(VISITNUM))
#' LBBLFL[1] <- "Y"
#'
#' VISIT <- data.frame(
#'   VISIT = VISITS, VISITNUM = VISITNUM,
#'   LBBLFL = LBBLFL, stringsAsFactors = FALSE
#' )
#' labdata <- expand.grid(
#'   USUBJID = USUBJID, LBTEST = LBTEST,
#'   VISITNUM = VISITNUM,
#'   stringsAsFactors = FALSE
#' )
#' setDT(labdata)
#'
#' labdata <- merge(labdata, VISIT, by = "VISITNUM")
#'
#' subject_elts <- unique(labdata[, .SD, .SDcols = "USUBJID"])
#' subject_elts <- unique(subject_elts)
#' subject_elts[, c("TREAT") := list(
#'   sample(x = c("Treatment", "Placebo"), size = .N, replace = TRUE)
#' )]
#' subject_elts[, c("TREAT") := list(
#'   factor(.SD$TREAT, levels = c("Treatment", "Placebo"))
#' )]
#' setDF(subject_elts)
#' labdata <- merge(labdata, subject_elts,
#'   by = "USUBJID", all.x = TRUE, all.y = FALSE
#' )
#' labdata[, c("LBNRIND") := list(
#'   sample(
#'     x = c("LOW", "NORMAL", "HIGH"), size = .N,
#'     replace = TRUE, prob = c(.03, .9, .07)
#'   )
#' )]
#'
#' setDF(labdata)
#'
#'
#'
#'
#' # shift table calculation ----
#'
#' SHIFT_TABLE <- shift_table(
#'   x = labdata, cn_visit = "VISIT",
#'   cn_grade = "LBNRIND",
#'   cn_usubjid = "USUBJID",
#'   cn_lab_cat = "LBTEST",
#'   cn_treatment = "TREAT",
#'   cn_is_baseline = "LBBLFL",
#'   baseline_identifier = "Y",
#'   grade_levels = c("LOW", "NORMAL", "HIGH")
#' )
#'
#' # get attrs for post treatment ----
#' SHIFT_TABLE_VISIT <- attr(SHIFT_TABLE, "VISIT_N")
#' visit_as_factor <- attr(SHIFT_TABLE, "FUN_VISIT")
#' range_as_factor <- attr(SHIFT_TABLE, "FUN_GRADE")
#'
#' # post treatments ----
#' SHIFT_TABLE$VISIT <- visit_as_factor(SHIFT_TABLE$VISIT)
#' SHIFT_TABLE$BASELINE <- range_as_factor(SHIFT_TABLE$BASELINE)
#' SHIFT_TABLE$LBNRIND <- range_as_factor(SHIFT_TABLE$LBNRIND)
#'
#' SHIFT_TABLE_VISIT$VISIT <- visit_as_factor(SHIFT_TABLE_VISIT$VISIT)
#'
#' # tabulator ----
#'
#' my_format <- function(z) {
#'   formatC(z * 100,
#'     digits = 1, format = "f",
#'     flag = "0", width = 4
#'   )
#' }
#'
#' tab <- tabulator(
#'   x = SHIFT_TABLE,
#'   hidden_data = SHIFT_TABLE_VISIT,
#'   row_compose = list(
#'     VISIT = as_paragraph(VISIT, "\n(N=", N_VISIT, ")")
#'   ),
#'   rows = c("LBTEST", "VISIT", "BASELINE"),
#'   columns = c("TREAT", "LBNRIND"),
#'   `n` = as_paragraph(N),
#'   `%` = as_paragraph(as_chunk(PCT, formatter = my_format))
#' )
#'
#' # as_flextable ----
#'
#' ft_1 <- as_flextable(
#'   x = tab, separate_with = "VISIT",
#'   label_rows = c(
#'     LBTEST = "Lab Test", VISIT = "Visit",
#'     BASELINE = "Reference Range Indicator"
#'   )
#' )
#'
#' ft_1
shift_table <- function(
    x,
    cn_visit = "VISIT", cn_visit_num = "VISITNUM", cn_grade = "LBNRIND",
    cn_usubjid = "USUBJID", cn_lab_cat = NA_character_, cn_is_baseline = "LBBLFL",
    baseline_identifier = "Y", cn_treatment = NA_character_,
    grade_levels = c("LOW", "NORMAL", "HIGH"),
    grade_labels = c("Low", "Normal", "High")) {
  cn_baseline <- "BASELINE"

  PIVOT_DATA <- full_shift_structure(
    x = x, cn_is_baseline = cn_is_baseline, baseline_identifier = baseline_identifier,
    cn_treatment = cn_treatment, cn_lab_cat = cn_lab_cat,
    cn_visit = cn_visit,
    cn_baseline = cn_baseline,
    cn_grade = cn_grade,
    grade_levels = grade_levels
  )

  DATAMART <- pivot_baseline(
    x = x, cn_is_baseline = cn_is_baseline, baseline_identifier = baseline_identifier,
    cn_visit = cn_visit,
    cn_usubjid = cn_usubjid,
    cn_grade = cn_grade,
    cn_treatment = cn_treatment, cn_lab_cat = cn_lab_cat,
    cn_baseline = cn_baseline
  )

  SHIFT_TABLE_VISIT <- as.data.table(DATAMART)
  by <- c_if(cn_lab_cat, cn_treatment, cn_visit)
  SHIFT_TABLE_VISIT <- SHIFT_TABLE_VISIT[, list(N_VISIT = length(.SD[[cn_usubjid]])), by = by]
  setDF(SHIFT_TABLE_VISIT)

  SHIFT_TABLE_CT <- as.data.table(DATAMART)
  by <- c_if(cn_treatment, cn_lab_cat, cn_visit, cn_baseline, cn_grade)
  SHIFT_TABLE_CT <- SHIFT_TABLE_CT[, list(N = .N), by = by]
  setDF(SHIFT_TABLE_CT)

  SHIFT_TABLE_DETAIL <- as.data.table(SHIFT_TABLE_CT)
  by <- c_if(cn_treatment, cn_lab_cat, cn_visit)
  SHIFT_TABLE_DETAIL[, c("PCT") := list(.SD[["N"]] / sum(.SD[["N"]], na.rm = TRUE)), by = by]

  SHIFT_TABLE_SUM <- as.data.table(SHIFT_TABLE_CT)
  by <- c_if(cn_treatment, cn_lab_cat, cn_visit, cn_grade)
  SHIFT_TABLE_SUM <- SHIFT_TABLE_SUM[, list(N = sum(.SD[["N"]], na.rm = TRUE)), by = by]
  by <- c_if(cn_treatment, cn_lab_cat, cn_visit)
  SHIFT_TABLE_SUM[, c("PCT") := list(.SD[["N"]] / sum(.SD[["N"]], na.rm = TRUE)), by = by]
  SHIFT_TABLE_SUM[[cn_baseline]] <- "SUM"

  replct <- list("MISSING", "MISSING", N = 0L, PCT = 0.0)
  names(replct) <- c(cn_baseline, cn_grade, "N", "PCT")

  SHIFT_TABLE <- rbind(SHIFT_TABLE_DETAIL, SHIFT_TABLE_SUM)
  setDF(SHIFT_TABLE_DETAIL)
  setDF(SHIFT_TABLE_SUM)
  for (j in names(replct)) {
    set(SHIFT_TABLE, i = which(is.na(SHIFT_TABLE[[j]])), j = j, value = replct[[j]])
  }

  setDT(PIVOT_DATA)
  by <- c_if(cn_treatment, cn_lab_cat, cn_visit, cn_baseline, cn_grade)
  SHIFT_TABLE <- merge(PIVOT_DATA, SHIFT_TABLE,
    by = by, all.x = TRUE, all.y = TRUE
  )
  for (j in names(replct)) {
    set(SHIFT_TABLE, i = which(is.na(SHIFT_TABLE[[j]])), j = j, value = replct[[j]])
  }

  setDF(SHIFT_TABLE)

  attr(SHIFT_TABLE, "VISIT_N") <- SHIFT_TABLE_VISIT
  attr(SHIFT_TABLE, "FUN_VISIT") <- facfun_visit(x, cn_visit = cn_visit, cn_visit_num = cn_visit_num)
  attr(SHIFT_TABLE, "FUN_GRADE") <- facfun_grade(x, grade_levels = grade_levels, grade_labels = grade_labels)
  SHIFT_TABLE
}

init_flextable_defaults()

library(data.table)

set.seed(42)

USUBJID <- sprintf("01-ABC-%04.0f", 1:50)
VISITS <- c("SCREENING 1", "WEEK 2", "MONTH 3")
LBTEST <- c("Albumin", "Sodium")
VISITNUM <- seq_along(VISITS)
LBBLFL <- rep(NA_character_, length(VISITNUM))
LBBLFL[1] <- "Y"

VISIT <- data.frame(
  VISIT = VISITS,
  VISITNUM = VISITNUM,
  LBBLFL = LBBLFL,
  stringsAsFactors = FALSE
)
labdata <- expand.grid(
  USUBJID = USUBJID,
  LBTEST = LBTEST,
  VISITNUM = VISITNUM,
  stringsAsFactors = FALSE
)
setDT(labdata)
labdata <- merge(labdata, VISIT, by = "VISITNUM")
labdata[,
  c("LBNRIND") := list(
    sample(
      x = c("LOW", "NORMAL", "HIGH"),
      size = .N,
      replace = TRUE,
      prob = c(.03, .9, .07)
    )
  )
]
setDF(labdata)

test_that("shift_table without treatment", {
  st <- shift_table(
    x = labdata,
    cn_visit = "VISIT",
    cn_grade = "LBNRIND",
    cn_usubjid = "USUBJID",
    cn_lab_cat = "LBTEST",
    cn_is_baseline = "LBBLFL",
    baseline_identifier = "Y",
    grade_levels = c("LOW", "NORMAL", "HIGH")
  )

  expect_s3_class(st, "data.frame")
  expect_true("BASELINE" %in% names(st))
  expect_true("N" %in% names(st))
  expect_true("PCT" %in% names(st))

  # attributes
  visit_n <- attr(st, "VISIT_N")
  expect_s3_class(visit_n, "data.frame")
  expect_true("N_VISIT" %in% names(visit_n))

  fun_visit <- attr(st, "FUN_VISIT")
  fun_grade <- attr(st, "FUN_GRADE")
  expect_type(fun_visit, "closure")
  expect_type(fun_grade, "closure")

  # factor functions work
  st$VISIT <- fun_visit(st$VISIT)
  expect_s3_class(st$VISIT, "factor")
  expect_equal(levels(st$VISIT), VISITS)

  st$BASELINE <- fun_grade(st$BASELINE)
  expect_s3_class(st$BASELINE, "factor")
  expect_true("Missing" %in% levels(st$BASELINE))
  expect_true("Sum" %in% levels(st$BASELINE))

  # all PCT values between 0 and 1
  expect_true(all(st$PCT >= 0 & st$PCT <= 1))
})

test_that("shift_table with treatment", {
  subject_elts <- unique(labdata[, "USUBJID", drop = FALSE])
  subject_elts$TREAT <- sample(
    c("Treatment", "Placebo"),
    size = nrow(subject_elts),
    replace = TRUE
  )
  labdata_treat <- merge(labdata, subject_elts, by = "USUBJID")

  st <- shift_table(
    x = labdata_treat,
    cn_visit = "VISIT",
    cn_grade = "LBNRIND",
    cn_usubjid = "USUBJID",
    cn_lab_cat = "LBTEST",
    cn_treatment = "TREAT",
    cn_is_baseline = "LBBLFL",
    baseline_identifier = "Y",
    grade_levels = c("LOW", "NORMAL", "HIGH")
  )

  expect_s3_class(st, "data.frame")
  expect_true("TREAT" %in% names(st))

  visit_n <- attr(st, "VISIT_N")
  expect_true("TREAT" %in% names(visit_n))
})

test_that("shift_table works with tabulator and as_flextable", {
  st <- shift_table(
    x = labdata,
    cn_visit = "VISIT",
    cn_grade = "LBNRIND",
    cn_usubjid = "USUBJID",
    cn_lab_cat = "LBTEST",
    cn_is_baseline = "LBBLFL",
    baseline_identifier = "Y",
    grade_levels = c("LOW", "NORMAL", "HIGH")
  )

  visit_n <- attr(st, "VISIT_N")
  fun_visit <- attr(st, "FUN_VISIT")
  fun_grade <- attr(st, "FUN_GRADE")

  st$VISIT <- fun_visit(st$VISIT)
  st$BASELINE <- fun_grade(st$BASELINE)
  st$LBNRIND <- fun_grade(st$LBNRIND)
  visit_n$VISIT <- fun_visit(visit_n$VISIT)

  tab <- tabulator(
    x = st,
    hidden_data = visit_n,
    row_compose = list(
      VISIT = as_paragraph(VISIT, "\n(N=", N_VISIT, ")")
    ),
    rows = c("LBTEST", "VISIT", "BASELINE"),
    columns = "LBNRIND",
    `n` = as_paragraph(N)
  )

  ft <- as_flextable(tab, separate_with = "VISIT")
  expect_s3_class(ft, "flextable")
  expect_true(nrow_part(ft, "body") > 0)
})

test_that("shift_table with custom labels", {
  st <- shift_table(
    x = labdata,
    cn_visit = "VISIT",
    cn_grade = "LBNRIND",
    cn_usubjid = "USUBJID",
    cn_lab_cat = "LBTEST",
    cn_is_baseline = "LBBLFL",
    baseline_identifier = "Y",
    grade_levels = c("LOW", "NORMAL", "HIGH"),
    grade_labels = c("Bas", "Normal", "Haut")
  )

  fun_grade <- attr(st, "FUN_GRADE")
  levs <- levels(fun_grade(st$BASELINE))
  expect_equal(levs, c("Bas", "Normal", "Haut", "Missing", "Sum"))
})

init_flextable_defaults()

#' @importFrom data.table rbindlist setDF
rbind_match_columns <- function(list_df) {
  df <- rbindlist(list_df, use.names = TRUE, fill = TRUE)
  setDF(df)
  row.names(df) <- NULL
  df
}


check_formula_i_and_part <- function(i, part) {
  if (inherits(i, "formula") && part %in% c("header", "footer")) {
    stop("formulas are not supported in the 'header' and 'footer' parts.")
  }
  TRUE
}

absolute_path <- function(x) {
  if (length(x) != 1L) {
    stop("'x' must be a single character string")
  }
  epath <- path.expand(x)

  if (file.exists(epath)) {
    epath <- normalizePath(epath, "/", mustWork = TRUE)
  } else {
    if (!dir.exists(dirname(epath))) {
      stop(sprintf("directory of '%s' does not exist.", x), call. = FALSE)
    }
    cat("", file = epath)
    epath <- normalizePath(epath, "/", mustWork = TRUE)
    unlink(epath)
  }
  epath
}

#' @importFrom stats median median sd mad
#' @importFrom stats quantile
Q1 <- function(z) as.double(quantile(z, probs = .25, na.rm = TRUE, names = FALSE))
Q3 <- function(z) as.double(quantile(z, probs = .75, na.rm = TRUE, names = FALSE))
MEDIAN <- function(z) as.double(median(z, na.rm = TRUE))
MEAN <- function(z) as.double(mean(z, na.rm = TRUE))
SD <- function(z) as.double(sd(z, na.rm = TRUE))
MAD <- function(z) as.double(mad(z, na.rm = TRUE))
MIN <- function(z) as.double(min(z, na.rm = TRUE))
MAX <- function(z) as.double(max(z, na.rm = TRUE))
N <- function(z) length(z)
NAS <- function(z) sum(is.na(z))


as_bookmark <- function(id, str) {
  new_id <- UUIDgenerate()
  bm_start_str <- sprintf("<w:bookmarkStart w:id=\"%s\" w:name=\"%s\"/>", new_id, id)
  bm_start_end <- sprintf("<w:bookmarkEnd w:id=\"%s\"/>", new_id)
  paste0(bm_start_str, str, bm_start_end)
}

format_double <- function(x, digits = 2) {
  formatC(x, format = "f", digits = digits, decimal.mark = ".", drop0trailing = TRUE)
}

has_value <- function(x) {
  !is.null(x) && !is.na(x) && length(x) == 1
}

coalesce_options <- function(a = NULL, b = NULL) {
  if (is.null(a)) {
    return(b)
  }
  if (is.null(b)) {
    return(a)
  }
  if (length(b) == 1) {
    b <- rep(b, length(a))
  }
  out <- a
  out[!has_value(a)] <- b[!has_value(a)]
  out
}

mcoalesce_options <- function(...) {
  Reduce(coalesce_options, list(...))
}

safe_stat <- function(..., FUN = max, NA_value = NA_real_) {
  x <- na.omit(unlist(list(...)))
  if (length(x) > 0) {
    FUN(x)
  } else {
    NA_value
  }
}

safe_stat_ext <- function(..., FUN = max, NA_value = NA_real_, LENGTH = NULL) {
  x <- na.omit(unlist(list(...)))
  if (length(x) > 0 && (!is.numeric(LENGTH) || length(LENGTH) == 0 || is.na(LENGTH) || length(x) == LENGTH[1])) {
    FUN(x)
  } else {
    NA_value
  }
}

# metric units -----
cm_to_inches <- function(x) {
  x / 2.54
}
mm_to_inches <- function(x) {
  x / 25.4
}
convin <- function(unit, x) {
  unit <- match.arg(unit, choices = c("in", "cm", "mm"), several.ok = FALSE)
  if (!identical("in", unit)) {
    x <- do.call(paste0(unit, "_to_inches"), list(x = x))
  }
  x
}
convmeters <- function(unit, x) {
  unit <- match.arg(unit, choices = c("in", "cm", "mm"), several.ok = FALSE)
  if (identical("cm", unit)) {
    x <- x * 2.54
  } else if (identical("mm", unit)) {
    x <- x * 254
  }
  x
}

# check for gregexec -----
if (!"gregexec" %in% getNamespaceExports("base")) {
  # copied from R source, grep.R
  gregexec <- function(pattern, text, ignore.case = FALSE, perl = FALSE,
                       fixed = FALSE, useBytes = FALSE) {
    if (is.factor(text) && length(levels(text)) < length(text)) {
      out <- gregexec(
        pattern, c(levels(text), NA_character_),
        ignore.case, perl, fixed, useBytes
      )
      outna <- out[length(out)]
      out <- out[text]
      out[is.na(text)] <- outna
      return(out)
    }

    dat <- gregexpr(
      pattern = pattern, text = text, ignore.case = ignore.case,
      fixed = fixed, useBytes = useBytes, perl = perl
    )
    if (perl && !fixed) {
      ## Perl generates match data, so use that
      capt.attr <- c("capture.start", "capture.length", "capture.names")
      process <- function(x) {
        if (anyNA(x) || any(x < 0)) {
          y <- x
        } else {
          ## Interleave matches with captures
          y <- t(cbind(x, attr(x, "capture.start")))
          attributes(y)[names(attributes(x))] <- attributes(x)
          ml <- t(cbind(attr(x, "match.length"), attr(x, "capture.length")))
          nm <- attr(x, "capture.names")
          ## Remove empty names that `gregexpr` returns
          dimnames(ml) <- dimnames(y) <-
            if (any(nzchar(nm))) list(c("", nm), NULL)
          attr(y, "match.length") <- ml
          y
        }
        attributes(y)[capt.attr] <- NULL
        y
      }
      lapply(dat, process)
    } else {
      ## For TRE or fixed we must compute the match data ourselves
      m1 <- lapply(regmatches(text, dat),
        regexec,
        pattern = pattern, ignore.case = ignore.case,
        perl = perl, fixed = fixed, useBytes = useBytes
      )
      mlen <- lengths(m1)
      res <- vector("list", length(m1))
      im <- mlen > 0
      res[!im] <- dat[!im] # -1, NA
      res[im] <- Map(
        function(outer, inner) {
          tmp <- do.call(cbind, inner)
          attributes(tmp)[names(attributes(inner))] <- attributes(inner)
          attr(tmp, "match.length") <-
            do.call(cbind, lapply(inner, `attr`, "match.length"))
          # useBytes/index.type should be same for all so use outer vals
          attr(tmp, "useBytes") <- attr(outer, "useBytes")
          attr(tmp, "index.type") <- attr(outer, "index.type")
          tmp + rep(outer - 1L, each = nrow(tmp))
        },
        dat[im],
        m1[im]
      )
      res
    }
  }
}

collect_labels <- function(dataset, use_labels = TRUE) {
  lbls <- character()
  values_labels <- list()
  if (use_labels) {
    lbls <- lapply(dataset, function(x) attr(x, "label"))
    lbls <- Filter(function(x) !is.null(x), lbls)
    values_labels <- lapply(dataset, function(x) attr(x, "labels"))
    values_labels <- Filter(f = function(x) !is.null(x), values_labels)
    values_labels <- lapply(values_labels, function(x) {
      values_lbls <- names(x)
      names(values_lbls) <- as.character(unname(x))
      values_lbls
    })
  }
  list(variables_labels = lbls, values_labels = values_labels)
}

apply_labels <- function(ft, collected_labels) {
  if (length(collected_labels$values_labels) > 0) {
    values_labels <- collected_labels$values_labels
    for(i in seq_along(values_labels)) {
      colname <- names(values_labels)[i]
      values_lbls <- values_labels[[colname]]
      ft <- labelizor(x = ft, part = "body", j = colname, labels = values_lbls)
      ft <- align(x = ft, j = colname, align = "left", part = "header")
      ft <- align(x = ft, j = colname, align = "left", part = "body")
    }
  }
  ft
}

merge_rle <- function(values) {
  rle_ <- rle(x = values)

  vout <- lapply(rle_$lengths, function(l) {
    out <- rep(0L, l)
    out[1] <- l
    out
  })
  as.integer(unlist(x = vout))
}

check_merge <- function(x) {
  row_check <- all(rowSums(x$spans$rows) == ncol(x$spans$rows))
  col_check <- all(colSums(x$spans$columns) == nrow(x$spans$columns))

  if (!row_check || !col_check) {
    stop("invalid merging instructions", call. = FALSE)
  }
  x
}
span_columns <- function(x, columns = NULL, target = columns, combine = FALSE) {
  if (!all(columns %in% colnames(x$dataset))) {
    wrong_col <- columns[!columns %in% colnames(x$dataset)]
    stop(
      sprintf(
        "`%s` is using unknown variable(s): %s",
        "columns",
        paste0("`", wrong_col, "`", collapse = ",")
      ),
      call. = FALSE
    )
  }
  if (!all(target %in% x$col_keys)) {
    wrong_col <- target[!target %in% x$col_keys]
    stop(
      sprintf(
        "`%s` is using unknown variable(s): %s",
        "target",
        paste0("`", wrong_col, "`", collapse = ",")
      ),
      call. = FALSE
    )
  }

  if (length(target) == 1) {
    target <- rep(target, length(columns))
  }
  if (length(columns) == 1) {
    columns <- rep(columns, length(target))
  }

  if (combine) {
    temp <- rep(list(NULL), length(columns))
    for (k in seq_along(columns)) {
      column <- columns[k]
      if (column %in% x$col_keys) {
        values <- sapply(
          get_chunkset_struct_element(x$content, j = columns[k]),
          function(x) {
            paste(x$txt, collapse = "")
        })
      } else {
        values <- format(x$dataset[[column]], trim = TRUE, justify = "left")
      }
      temp[[k]] <- values
    }
    values <- do.call(cbind, temp)
    values <- apply(values, 1, paste0, collapse = "_")
    x$spans$columns[, match(target, x$col_keys)] <- merge_rle(values)
  } else {
    for (k in seq_along(columns)) {
      column <- columns[k]
      if (column %in% x$col_keys) {
        values <- sapply(
          get_chunkset_struct_element(x$content, j = columns[k]),
          function(x) {
            paste(x$txt, collapse = "")
          })
      } else {
        values <- format(x$dataset[[column]], trim = TRUE, justify = "left")
      }
      x$spans$columns[, match(target[k], x$col_keys)] <- merge_rle(values)
    }
  }

  check_merge(x)
}

span_cells_at <- function(x, columns = NULL, rows = NULL) {
  if (is.null(columns)) {
    columns <- x$col_keys
  }
  if (is.null(rows)) {
    rows <- get_rows_id(x, i = rows)
  }

  stopifnot(all(columns %in% x$col_keys))

  row_id <- match(rows, seq_len(nrow(x$dataset)))
  .col_id <- match(columns, x$col_keys)

  test_valid_r <- (length(row_id) > 1 && all(diff(row_id) == 1)) || length(row_id) == 1
  test_valid_c <- (length(.col_id) > 1 && all(diff(.col_id) == 1)) || length(.col_id) == 1

  if (!test_valid_r) {
    stop("selected rows should all be consecutive")
  }
  if (!test_valid_c) {
    stop("selected columns should all be consecutive")
  }

  x$spans$columns[row_id, .col_id] <- 0
  x$spans$rows[row_id, .col_id] <- 0
  x$spans$columns[row_id[1], .col_id] <- length(row_id)
  x$spans$rows[row_id, .col_id[1]] <- length(.col_id)

  check_merge(x)
}

span_rows <- function(x, rows = NULL) {
  row_id <- get_rows_id(x, i = rows)

  for (rowi in row_id) {
    values <- sapply(
      get_chunkset_struct_element(x$content, i = rowi),
      function(x) {
        paste(x$txt, collapse = "")
      })
    x$spans$rows[rowi, ] <- merge_rle(values)
  }

  check_merge(x)
}

span_free <- function(x) {
  x$spans$rows[] <- 1
  x$spans$columns[] <- 1
  x
}

#' @param x a complex_tabpart object
#' @noRd
as_col_keys <- function(x, j = NULL, blanks = character()) {
  candidates <- setdiff(colnames(x$dataset), blanks)

  if (is.null(j)) {
    j <- candidates
  } else if (inherits(j, "formula")) {
    tmp_dat <- as.list(candidates)
    names(tmp_dat) <- candidates
    tmp_dat <- as.data.frame(tmp_dat, check.names = FALSE)
    j <- get_j_from_formula(j, tmp_dat)
  } else if (is.logical(j)) {
    if (length(j) != length(candidates)) {
      stop("j (as logical) is expected to have the same length than the original 'data.frame' used in call to `flextable()`.")
    }
    j <- candidates[j]
  } else if (is.character(j)) {
    if (!all(j %in% candidates)) {
      wrong_col <- j[!j %in% candidates]
      wrong_col <- paste0(shQuote(wrong_col), collapse = ",")
      warning(sprintf("Some column(s) can not be found in the original data.frame: %s.", wrong_col),
        call. = FALSE
      )
    }
    j <- intersect(candidates, j)
  } else if (is.numeric(j) & all(j > 0)) {
    j <- candidates[intersect(seq_along(candidates), j)]
  } else if (is.numeric(j) & all(j < 0)) {
    j <- candidates[setdiff(seq_along(candidates), -j)]
  }

  j
}

get_columns_id <- function(x, j = NULL) {
  maxcol <- length(x$col_keys)

  if (is.null(j)) {
    j <- seq_along(x$col_keys)
  }

  if (inherits(j, "formula")) {
    tmp_dat <- as.list(x$col_keys)
    names(tmp_dat) <- x$col_keys
    tmp_dat <- as.data.frame(tmp_dat, check.names = FALSE)
    j <- get_j_from_formula(j, tmp_dat)
  }

  if (is.numeric(j)) {
    if (length(j) > 0 && all(j < 0)) {
      j <- setdiff(seq_along(x$col_keys), -j)
    }

    if (any(j < 1 | j > maxcol)) {
      stop(
        sprintf(
          "invalid columns selection\navailable range: [%s]\nissues: %s",
          paste0(range(seq_len(maxcol)), collapse = ", "),
          paste0(setdiff(j, seq_len(maxcol)), collapse = ", ")
        )
      )
    }
  } else if (is.logical(j)) {
    if (length(j) != maxcol) {
      stop(
        sprintf(
          "invalid columns selection\n`j` should have a length of %.0f.",
          maxcol
        )
      )
    } else {
      j <- which(j)
    }
  } else if (is.character(j)) {
    j <- gsub("(^`|`$)", "", j)
    if (any(is.na(j))) {
      stop("invalid columns selection: NA in selection")
    } else if (!all(is.element(j, x$col_keys))) {
      stop(
        sprintf(
          "`%s` is using unknown variable(s): %s",
          "i",
          paste0("`", j[!is.element(j, x$col_keys)], "`", collapse = ",")
        )
      )
    } else {
      j <- match(j, x$col_keys)
    }
  } else {
    stop("invalid columns selection: unknown selection type")
  }

  j
}

get_rows_id <- function(x, i = NULL) {
  maxrow <- nrow(x$dataset)

  if (is.null(i)) {
    i <- seq_len(maxrow)
  }
  if (inherits(i, "formula")) {
    i <- get_i_from_formula(i, x$dataset)
  }

  if (is.numeric(i)) {
    if (length(i) > 0 && all(i < 0)) {
      i <- setdiff(-i, seq_len(maxrow))
    }

    if (any(i < 1 | i > maxrow)) {
      stop("invalid row selection: out of range selection")
    }
  } else if (is.logical(i)) {
    if (length(i) != maxrow) {
      stop(
        sprintf(
          "invalid row selection. `length(i)` [%.0f] != `nrow(dataset)` [%.0f].",
          length(i), maxrow
        )
      )
    } else {
      i <- which(i)
    }
  } else if (is.character(i)) {
    any(is.na(i))
    rn <- row.names(x$dataset)
    if (any(is.na(i))) {
      stop("invalid row selection: NA in selection")
    } else if (!all(is.element(i, rn))) {
      stop("invalid row selection: unknown rownames")
    } else {
      i <- match(i, rn)
    }
  } else {
    stop("invalid row selection: unknown selection type")
  }

  i
}

# formula tools -----
get_i_from_formula <- function(f, data) {
  if (length(f) > 2) {
    stop("formula selection is not as expected ( ~ condition )", call. = FALSE)
  }
  i <- eval(as.call(f[[2]]), envir = data)
  if (!is.logical(i)) {
    stop("formula selection should return a logical vector", call. = FALSE)
  }
  i
}
get_j_from_formula <- function(f, data) {
  if (length(f) > 2) {
    stop("formula selection is not as expected ( ~ variables )", call. = FALSE)
  }
  j <- attr(terms(f, data = data), "term.labels")
  j <- gsub("(^`|`$)", "", j)
  names_ <- names(data)

  invalid_names <- (!j %in% names_)
  if (any(invalid_names)) {
    invalid_names <- paste0("`", j[invalid_names], "`", collapse = ", ")
    stop(
      sprintf(
        "`%s` is using unknown variable(s): %s",
        format(f),
        invalid_names
      ),
      call. = FALSE
    )
  }
  j
}

#' @title flextable: Functions for Tabular Reporting
#'
#' @description
#' The flextable package facilitates access to and manipulation of
#' tabular reporting elements from R.
#'
#' The documentation of functions can be opened with command `help(package = "flextable")`.
#'
#' To learn more about flextable, start with the vignettes:
#' `browseVignettes(package = "flextable")`.
#'
#' `flextable()` function is producing flexible tables where each cell
#' can contain several chunks of text with their own set of formatting
#' properties (bold, font color, etc.). Function [compose()] lets customise
#' text of cells.
#'
#' @seealso <https://davidgohel.github.io/flextable/>, [flextable()]
#' @docType package
#' @aliases flextable-package
#' @name flextable-package
NULL

get_docx_xml <- function(x) {
  if(inherits(x, "flextable")){
    docx_file <- tempfile(fileext = ".docx")
    doc <- read_docx()
    doc <- body_add_flextable(doc, value = x)
    print(doc, target = docx_file)
    x <- docx_file
  }
  redoc <- read_docx(x)
  xml2::xml_child(docx_body_xml(redoc))
}

get_pptx_xml <- function(x) {
  if(inherits(x, "flextable")){
    pptx_file <- tempfile(fileext = ".pptx")
    doc <- read_pptx()
    doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
    doc <- ph_with(doc, x, location = ph_location_type(type = "body"))
    print(doc, target = pptx_file)
    x <- pptx_file
  }

  redoc <- read_pptx(x)
  slide <- redoc$slide$get_slide(redoc$cursor)
  xml2::xml_child(slide$get())
}

get_html_xml <- function(x) {
  if(inherits(x, "flextable")){
    html_file <- tempfile(fileext = ".html")
    save_as_html(tab, path = html_file)
    x <- html_file
  }
  doc <- read_html(x)
  xml_child(doc, "body")
}
get_pdf_text <- function(x, extract_fun) {
  stopifnot(grepl("\\.pdf$", x))

  doc <- extract_fun(x)
  txtfile <- tempfile()
  cat(paste0(doc, collapse = "\n"), file = txtfile)
  readLines(txtfile)
}

render_rmd <- function(file, rmd_format) {
  unlink(file, force = TRUE)
  sucess <- FALSE
  tryCatch(
    {
      render(rmd_file,
             output_format = rmd_format,
             output_file = pdf_file,
             envir = new.env(),
             quiet = TRUE
      )
      sucess <- TRUE
    },
    warning = function(e) {
    },
    error = function(e) {
    }
  )
  sucess
}

init_flextable_defaults()

rmd_file_0 <- "rmd/captions.Rmd"
if (!file.exists(rmd_file_0)) { # just for dev purpose
  rmd_file_0 <- "tests/testthat/rmd/captions.Rmd"
}
rmd_file <- tempfile(fileext = ".Rmd")
file.copy(rmd_file_0, rmd_file, overwrite = TRUE)

html_file <- gsub("\\.Rmd$", ".html", rmd_file)
docx_file <- gsub("\\.Rmd$", ".docx", rmd_file)
pdf_file <- gsub("\\.Rmd$", ".pdf", rmd_file)

test_that("with html_document", {
  unlink(html_file, force = TRUE)
  render(rmd_file,
    output_format = rmarkdown::html_document(),
    output_file = html_file,
    envir = new.env(),
    quiet = TRUE
  )

  xml_doc <- get_html_xml(html_file)

  # id is there, caption is the one of set_caption, no numbering
  caption_id2 <- xml_find_first(xml_doc, "//table/caption[@id='id2']")
  expect_true(grepl("text-align:center;", xml_attr(caption_id2, "style")))
  expect_equal(xml_text(caption_id2), "azerty querty")

  # first_chunk has an defined class
  first_chunk_class <- xml_attr(xml_child(caption_id2, 1), "class")
  expect_true(grepl("^cl\\-", first_chunk_class))
  expect_true(any(grepl(first_chunk_class, xml_text(xml_find_all(xml_doc, "//style")))))

  crossref_chunk <- xml_find_first(xml_doc, "//a[@href='#tab:id1']")
  expect_true(inherits(crossref_chunk, "xml_missing"))

  body <- xml_find_first(xml_doc, "//*[id='tab:id1']")
  id_chunk <- xml_find_first(xml_doc, "//table/caption/p/span[@id='tab:id1']")
  expect_true(inherits(id_chunk, "xml_missing"))

  captions <- xml_find_all(xml_doc, "//table/caption/p")
  expect_true(all(!grepl("Table [0-9]+:", xml_text(captions))))
})

test_that("with html_document2", {
  skip_if_not_installed("bookdown")

  unlink(html_file, force = TRUE)
  render(rmd_file,
    output_format = bookdown::html_document2(keep_md = FALSE),
    output_file = html_file,
    envir = new.env(),
    quiet = TRUE
  )

  xml_doc <- get_html_xml(html_file)

  # id is there, caption is the one of set_caption, numbering
  caption_id2 <- xml_find_first(xml_doc, "//table/caption/span[@id='tab:id2']")
  expect_true(grepl("text-align:center;", xml_attr(xml_parent(caption_id2), "style")))
  expect_true(grepl("Table [0-9]+\\: azerty querty", xml_text(xml_parent(caption_id2))))

  # first_chunk has an defined class
  first_chunk_class <- xml_attr(xml_siblings(caption_id2), "class")
  expect_true(all(grepl("^cl\\-", first_chunk_class)))

  crossref_chunk <- xml_find_first(xml_doc, "//a[@href='#tab:id1']")
  expect_false(inherits(crossref_chunk, "xml_missing"))

  id_chunk <- xml_find_first(xml_doc, "//caption/span[@id='tab:id1']")
  expect_false(inherits(id_chunk, "xml_missing"))

  caption <- xml_find_first(xml_doc, "//caption[span/@id='tab:id1']")
  expect_true(grepl("Table 2:", xml_text(caption)))
})

test_that("with word_document", {
  skip_if(pandoc_version() == numeric_version("2.9.2.1"))

  unlink(docx_file, force = TRUE)
  render(rmd_file,
    output_format = rmarkdown::word_document(keep_md = FALSE),
    output_file = docx_file,
    envir = new.env(),
    quiet = TRUE
  )

  doc <- get_docx_xml(docx_file)

  caption_node <- xml_find_first(doc,
    xpath = "/w:document/w:body/w:tbl[4]/preceding-sibling::*[1]"
  )
  expect_false(grepl("Table [1-5]+", xml_text(caption_node), fixed = TRUE))

  style_nodes <- xml_find_all(doc, "//w:pStyle[@w:val='TableCaption']")
  expect_length(style_nodes, 4)
  expect_true(all(xml_attr(style_nodes, "val") %in% "TableCaption"))
  txt_nodes <- xml_parent(xml_parent(style_nodes))
  expect_true(
    all(!grepl("^Table [1-5]+\\:", xml_text(txt_nodes)))
  )

  bookmarks <- xml_find_all(doc, "//w:tbl/preceding-sibling::w:p[1]/w:bookmarkStart")
  expect_length(bookmarks, 0)
})

test_that("with word_document2", {
  skip_if(pandoc_version() <= numeric_version("2.7.3"))
  skip_if(pandoc_version() == numeric_version("2.9.2.1"))
  skip_if_not_installed("bookdown")

  unlink(docx_file, force = TRUE)
  render(rmd_file,
    output_format = bookdown::word_document2(keep_md = FALSE),
    output_file = docx_file,
    envir = new.env(),
    quiet = TRUE
  )

  doc <- get_docx_xml(docx_file)

  caption_node <- xml_find_first(doc,
    xpath = "/w:document/w:body/w:tbl[2]/preceding-sibling::*[1]"
  )
  expect_true(grepl("Table 2", xml_text(caption_node), fixed = TRUE))

  crossref_node <- xml_find_first(doc, "/w:document/w:body/w:p[2]")
  expect_equal("Cross-reference is there: 2", xml_text(crossref_node))

  style_nodes <- xml_find_all(doc, "//w:pStyle[@w:val='TableCaption']")
  expect_length(style_nodes, 4)
  expect_true(all(xml_attr(style_nodes, "val") %in% "TableCaption"))
  txt_nodes <- xml_parent(xml_parent(style_nodes))
  expect_true(
    all(
      grepl("^Table [1-4]+\\:", xml_text(txt_nodes))
    )
  )

  bookmarks <- xml_find_all(doc, "//w:tbl/preceding-sibling::w:p[1]/w:bookmarkStart")
  expect_length(bookmarks, 3)
})



test_that("word with officer", {
  unlink(docx_file, force = TRUE)
  ft <- flextable(head(cars))
  ft <- theme_vanilla(ft)
  ft <- autofit(ft)
  ft <- set_caption(
    x = ft,
    caption = as_paragraph(
      as_chunk("azerty ", props = fp_text_default(color = "cyan")),
      as_chunk("querty", props = fp_text_default(color = "orange"))
    ),
    autonum = run_autonum(seq_id = "tab", bkm = "id2"),
    fp_p = fp_par(
      padding = 10,
      border = fp_border_default(color = "red", width = 1)
    )
  )
  doc <- get_docx_xml(ft)

  caption_node <- xml_find_first(doc,
    xpath = "/w:document/w:body/w:tbl/preceding-sibling::*[1]"
  )
  expect_false(grepl("Table [1-5]+", xml_text(caption_node), fixed = TRUE))

  style_nodes <- xml_find_all(doc, "//w:pStyle[@w:val='TableCaption']")
  expect_length(style_nodes, 1)
  expect_true(all(xml_attr(style_nodes, "val") %in% "TableCaption"))
  bookmarks <- xml_find_all(doc, "//w:tbl/preceding-sibling::w:p[1]/w:bookmarkStart")
  expect_length(bookmarks, 1)
  expect_equal(xml_attr(bookmarks, "name"), "id2")
})


test_that("with pdf_document2", {
  skip_if(pandoc_version() <= numeric_version("2.7.3"))
  skip_if_not_installed("bookdown")
  skip_if_not_installed("pdftools")

  require("pdftools")
  sucess <- render_rmd(file = pdf_file, rmd_format = bookdown::pdf_document2(keep_md = FALSE))
  if (sucess) {
    doc <- get_pdf_text(pdf_file, extract_fun = pdftools::pdf_text)
    expect_true(any(grepl("Cross-reference is there: 2", doc, fixed = TRUE)))
  } else {
    expect_false(sucess) # only necessary to avoid a note
  }
})

init_flextable_defaults()

#' @importFrom utils head
add_rows.complex_tabpart <- function( x, rows, first = FALSE ){

  data <- x$dataset
  spans <- x$spans
  ncol <- length(x$col_keys)
  nrow <- nrow(rows)

  x$styles$cells <- add_rows(x$styles$cells, nrows = nrow, first = first)
  x$styles$pars <- add_rows(x$styles$pars, nrows = nrow, first = first)
  x$styles$text <- add_rows(x$styles$text, nrows = nrow, first = first)
  x$content <- add_rows(x$content, nrows = nrow, first = first, rows)

  span_new <- matrix( 1, ncol = ncol, nrow = nrow )
  rowheights <- x$rowheights
  hrule <- x$hrule

  if( !first ){
    data <- rbind(data, rows )
    spans$rows <- rbind( spans$rows, span_new )
    spans$columns <- rbind( spans$columns, span_new )
    rowheights <- c(rowheights, rep(rev(rowheights)[1], nrow(rows)))
    hrule <- c(hrule, rep(rev(hrule)[1], nrow(rows)))
  } else {
    data <- rbind(rows, data )
    spans$rows <- rbind( span_new, spans$rows )
    spans$columns <- rbind( span_new, spans$columns )
    rowheights <- c(rep(rowheights[1], nrow(rows)), rowheights)
    hrule <- c(rep(hrule[1], nrow(rows)), hrule)

  }
  x$rowheights <- rowheights
  x$dataset <- data
  x$spans <- spans
  x$hrule <- hrule
  x
}

pml_spans <- function(value) {
  span_data <- fortify_span(value)
  span_data$grid_span <- ifelse(
    span_data$rowspan == 1,
    "",
    ifelse(
      span_data$rowspan > 1,
      paste0(" gridSpan=\"", span_data$rowspan, "\""),
      " hMerge=\"true\""
    )
  )
  span_data$row_span <- ifelse(
    span_data$colspan == 1,
    "",
    ifelse(
      span_data$colspan > 1,
      paste0(" rowSpan=\"", span_data$colspan, "\""),
      " vMerge=\"true\""
    )
  )
  span_data
}
default_fp_text_pml <- function(value) {
  default_chunks_properties <- information_data_default_chunk(value)
  unique_text_props <- distinct_text_properties(default_chunks_properties)
  rpr <- sapply(
    split(
      unique_text_props[setdiff(colnames(unique_text_props), "classname")],
      unique_text_props$classname
    ),
    function(x) {
      z <- do.call(officer::fp_text_lite, x)
      val <- format(z, type = "pml")
      val <- gsub("<a:rPr", "<a:endParaRPr", val, fixed = TRUE)
      val <- gsub("</a:rPr>", "</a:endParaRPr>", val, fixed = TRUE)
      val
    }
  )

  unique_text_props$fp_txt_default <- unname(rpr[unique_text_props$classname])
  setDT(default_chunks_properties)
  by_cols <- c(
    "color",
    "font.size",
    "bold",
    "italic",
    "underlined",
    "strike",
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "vertical.align",
    "shading.color"
  )
  by_cols <- intersect(by_cols, colnames(default_chunks_properties))
  by_cols <- intersect(by_cols, colnames(unique_text_props))
  default_chunks_properties <- merge(
    default_chunks_properties,
    unique_text_props,
    by = by_cols
  )
  setDF(default_chunks_properties)
  default_chunks_properties <- default_chunks_properties[, c(
    ".part",
    ".row_id",
    ".col_id",
    "fp_txt_default"
  )]
  default_chunks_properties
}


#' @importFrom data.table shift
pml_cells <- function(value, cell_data) {
  cell_heights <- fortify_height(value)
  cell_widths <- fortify_width(value)
  cell_hrule <- fortify_hrule(value)

  cell_data$width <- NULL # need to get rid of originals that are empty, should probably rm them
  cell_data$height <- NULL
  cell_data$hrule <- NULL
  cell_data <- merge(cell_data, cell_widths, by = ".col_id")
  cell_data <- merge(cell_data, cell_heights, by = c(".part", ".row_id"))
  cell_data <- merge(cell_data, cell_hrule, by = c(".part", ".row_id"))

  setDT(cell_data)
  setorderv(cell_data, cols = c(".part", ".row_id", ".col_id"))

  data_ref_cells <- distinct_cells_properties(cell_data)

  ## cell style pml
  fp_cell_pml <- data_ref_cells
  classnames <- data_ref_cells$classname
  fp_cell_pml <- split(fp_cell_pml, classnames)
  fp_cell_pml <- lapply(fp_cell_pml, function(x) {
    zz <- as.list(x)
    zz$border.bottom <- fp_border(
      color = zz$border.color.bottom,
      width = zz$border.width.bottom,
      style = zz$border.style.bottom
    )
    zz$border.top <- fp_border(
      color = zz$border.color.top,
      width = zz$border.width.top,
      style = zz$border.style.top
    )
    zz$border.right <- fp_border(
      color = zz$border.color.right,
      width = zz$border.width.right,
      style = zz$border.style.right
    )
    zz$border.left <- fp_border(
      color = zz$border.color.left,
      width = zz$border.width.left,
      style = zz$border.style.left
    )

    zz[c(
      "border.width.bottom",
      "border.width.top",
      "border.width.left",
      "border.width.right",
      "border.color.bottom",
      "border.color.top",
      "border.color.left",
      "border.color.right",
      "border.style.bottom",
      "border.style.top",
      "border.style.left",
      "border.style.right",
      "width",
      "height",
      "hrule"
    )] <- NULL
    zz$classname <- NULL
    zz <- do.call(fp_cell, zz)
    format(zz, type = "pml")
  })
  style_dat <- data.frame(
    fp_cell_pml = as.character(fp_cell_pml),
    classname = classnames,
    stringsAsFactors = FALSE
  )

  # organise everything
  cell_data <- merge(
    cell_data,
    data_ref_cells,
    by = intersect(colnames(cell_data), colnames(data_ref_cells))
  )
  cell_data <- merge(cell_data, style_dat, by = "classname")
  cell_data <- cell_data[,
    .SD,
    .SDcols = c(".part", ".row_id", ".col_id", "fp_cell_pml")
  ]
  setDF(cell_data)
  cell_data
}

gen_raw_pml <- function(
  value,
  uid = 99999L,
  offx = 0,
  offy = 0,
  cx = 0,
  cy = 0
) {
  cell_attributes <- information_data_cell(value)
  par_attributes <- information_data_paragraph(value)

  # cell_attributes and par_attributes must be ordered identically
  new_pos <- ooxml_rotation_alignments(
    rotation = cell_attributes$text.direction,
    valign = cell_attributes$vertical.align,
    align = par_attributes$text.align
  )

  par_attributes$text.align <- new_pos$align
  cell_attributes$vertical.align <- new_pos$valign

  setDT(cell_attributes)
  cell_attributes <- merge(
    cell_attributes,
    par_attributes[, c(
      ".part",
      ".row_id",
      ".col_id",
      "padding.bottom",
      "padding.top"
    )],
    by = c(".part", ".row_id", ".col_id")
  )
  cell_attributes[,
    c("margin.bottom", "margin.top") := list(
      .SD$padding.bottom,
      .SD$padding.top
    )
  ]
  cell_attributes[, c("padding.bottom", "padding.top") := NULL]
  setDF(cell_attributes)

  txt_data <- runs_as_pml(value)
  par_data <- ooxml_ppr(par_attributes, type = "pml")
  span_data <- pml_spans(value)
  cell_data <- pml_cells(value, cell_attributes)
  cell_heights <- fortify_height(value)

  default_chunks_properties <- default_fp_text_pml(value)

  setDT(cell_data)

  tab_data <- merge(cell_data, par_data, by = c(".part", ".row_id", ".col_id"))
  tab_data <- merge(
    tab_data,
    default_chunks_properties,
    by = c(".part", ".row_id", ".col_id")
  )
  tab_data <- merge(tab_data, txt_data, by = c(".part", ".row_id", ".col_id"))

  tab_data[
    tab_data$is_empty %in% TRUE,
    c("fp_par_xml") := list(
      paste0(.SD$fp_par_xml, .SD$fp_txt_default)
    )
  ]
  tab_data[, c("fp_txt_default", "is_empty") := list(NULL, NULL)]

  tab_data <- merge(tab_data, span_data, by = c(".part", ".row_id", ".col_id"))
  tab_data$.col_id <- factor(tab_data$.col_id, levels = value$col_keys)
  setorderv(tab_data, cols = c(".part", ".row_id", ".col_id"))

  tab_data[,
    c("pml") := list(
      paste0(
        "<a:txBody><a:bodyPr/><a:lstStyle/>",
        "<a:p>",
        .SD$fp_par_xml,
        .SD$par_nodes_str,
        "</a:p>",
        "</a:txBody>"
      )
    )
  ]
  tab_data[, c("fp_par_xml", "par_nodes_str") := list(NULL, NULL)]

  tab_data[,
    c("pml") := list(
      paste0(
        "<a:tc",
        .SD$grid_span,
        .SD$row_span,
        ">",
        .SD$pml,
        .SD$fp_cell_pml,
        "</a:tc>"
      )
    )
  ]
  tab_data[,
    c("fp_cell_pml", "grid_span", "row_span") := list(NULL, NULL, NULL)
  ]

  cells <- dcast(
    tab_data,
    .part + .row_id ~ .col_id,
    drop = TRUE,
    fill = "",
    value.var = "pml",
    fun.aggregate = I
  )
  cells <- merge(cells, cell_heights, by = c(".part", ".row_id"))

  rowheights <- cells$height
  cells[, c(".row_id", "height", ".part") := list(NULL, NULL, NULL)]
  rows <- apply(as.matrix(cells), 1, paste0, collapse = "")
  rows <- sprintf("<a:tr h=\"%.0f\">%s</a:tr>", rowheights * 914400, rows)
  rows <- paste0(rows, collapse = "")

  out <- "<a:tbl>"
  dims <- dim(value)
  widths <- dims$widths
  colswidths <- paste0(
    sprintf("<a:gridCol w=\"%.0f\"/>", widths * 914400),
    collapse = ""
  )

  out <- paste0(out, "<a:tblPr/><a:tblGrid>")
  out <- paste0(out, colswidths)
  out <- paste0(out, "</a:tblGrid>")
  out <- paste0(out, rows)

  out <- paste0(out, "</a:tbl>")

  graphic_frame <- paste0(
    "<p:graphicFrame ",
    "xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" ",
    "xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" ",
    "xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
    "<p:nvGraphicFramePr>",
    sprintf("<p:cNvPr id=\"%.0f\" name=\"\"/>", uid),
    "<p:cNvGraphicFramePr><a:graphicFrameLocks noGrp=\"true\"/></p:cNvGraphicFramePr>",
    "<p:nvPr/>",
    "</p:nvGraphicFramePr>",
    "<p:xfrm rot=\"0\">",
    sprintf("<a:off x=\"%.0f\" y=\"%.0f\"/>", offx * 914400, offy * 914400),
    sprintf("<a:ext cx=\"%.0f\" cy=\"%.0f\"/>", cx * 914400, cy * 914400),
    "</p:xfrm>",
    "<a:graphic>",
    "<a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/table\">",
    out,
    "</a:graphicData>",
    "</a:graphic>",
    "</p:graphicFrame>"
  )
  graphic_frame
}

#' @importFrom officer fp_text fp_par fp_border
#' @title Transform a 'xtable' object into a flextable
#'
#' @description Get a `flextable` object from
#' a `xtable` object.
#'
#' @param x `xtable` object
#' @param text.properties default text formatting properties
#' @param format.args List of arguments for the formatC function.
#' See argument `format.args` of `print.xtable`. Not yet
#' implemented.
#' @param rowname_col colname used for row names column
#' @param hline.after see `?print.xtable`.
#' @param NA.string see `?print.xtable`.
#' @param include.rownames see `?print.xtable`.
#' @param rotate.colnames see `?print.xtable`.
#' @param ... unused arguments
#' @examples
#' library(officer)
#' if( require("xtable") ){
#'
#'   data(tli)
#'   tli.table <- xtable(tli[1:10, ])
#'   align(tli.table) <- rep("r", 6)
#'   align(tli.table) <- "|r|r|clr|r|"
#'   ft_1 <- as_flextable(
#'     tli.table,
#'     rotate.colnames = TRUE,
#'     include.rownames = FALSE)
#'   ft_1 <- height(ft_1, i = 1, part = "header", height = 1)
#'   ft_1
#'
#'   \donttest{
#'   Grade3 <- c("A","B","B","A","B","C","C","D","A","B",
#'     "C","C","C","D","B","B","D","C","C","D")
#'   Grade6 <- c("A","A","A","B","B","B","B","B","C","C",
#'     "A","C","C","C","D","D","D","D","D","D")
#'   Cohort <- table(Grade3, Grade6)
#'   ft_2 <- as_flextable(xtable(Cohort))
#'   ft_2 <- set_header_labels(ft_2, rowname = "Grade 3")
#'   ft_2 <- autofit(ft_2)
#'   ft_2 <- add_header(ft_2, A = "Grade 6")
#'   ft_2 <- merge_at(ft_2, i = 1, j = seq_len( ncol(Cohort) ) + 1,
#'     part = "header" )
#'   ft_2 <- bold(ft_2, j = 1, bold = TRUE, part = "body")
#'   ft_2 <- height_all(ft_2, part = "header", height = .4)
#'   ft_2
#'
#'   temp.ts <- ts(cumsum(1 + round(rnorm(100), 0)),
#'     start = c(1954, 7), frequency = 12)
#'   ft_3 <- as_flextable(x = xtable(temp.ts, digits = 0),
#'     NA.string = "-")
#'   ft_3
#'   }
#'   detach("package:xtable", unload = TRUE)
#' }
#' @export
#' @family as_flextable methods
as_flextable.xtable <- function(
  x,
  text.properties = fp_text_default(),
  format.args = getOption("xtable.format.args", NULL),
  rowname_col = "rowname",
  hline.after = getOption("xtable.hline.after", c(-1, 0, nrow(x))),
  NA.string = getOption("xtable.NA.string", ""),
  include.rownames = TRUE,
  rotate.colnames = getOption("xtable.rotate.colnames", FALSE),
  ...
) {
  padding.left <- 4
  padding.right <- 4

  stopifnot(inherits(x, "xtable"))

  if (!is.null(hline.after)) {
    if (any(hline.after < -1) | any(hline.after > nrow(x))) {
      stop("'hline.after' must be inside [-1, nrow(x)]")
    }
  }

  if (!include.rownames) {
    data <- as.data.frame(x, stringsAsFactors = FALSE)
    col_labels <- names(x)
    .col_id <- make.names(col_labels)
  } else {
    rn_x <- row.names(x)
    data <- cbind(
      structure(
        list(rn_x),
        .Names = rowname_col,
        row.names = seq_along(rn_x),
        class = "data.frame"
      ),
      as.data.frame(x, stringsAsFactors = FALSE)
    )
    col_labels <- c("", names(x))
    .col_id <- make.names(col_labels)
    .col_id[1] <- rowname_col
  }
  names(data) <- .col_id

  nrow_ <- nrow(data)
  ncol_ <- ncol(data)

  ina <- matrix(FALSE, nrow = nrow(data), ncol = ncol(data))

  for (j in seq_along(.col_id)) {
    if (is.factor(data[[j]])) {
      data[[j]] <- as.character(data[[j]])
    }
    if (is.list(data[[j]])) {
      data[[j]] <- sapply(data[[j]], unlist)
    }
    ina[, j] <- is.na(data[[j]])
  }

  if (is.null(format.args)) {
    format.args <- list()
  }
  if (is.null(format.args$decimal.mark)) {
    format.args$decimal.mark <- options()$OutDec
  }

  digits_val <- attr(x, "digits", exact = TRUE)
  display_val <- attr(x, "display", exact = TRUE)
  align <- attr(x, "align")

  if (!include.rownames) {
    digits_val <- digits_val[-1]
    display_val <- display_val[-1]
    which_grep <- grep("^[a-zA-Z]", align)[1]
    align <- align[-seq_len(which_grep)]
  }

  if (!is.matrix(digits_val)) {
    digits_val <- rep(digits_val, each = nrow_)
  }
  display_val <- ifelse(
    digits_val < 0,
    "E",
    rep(display_val, each = nrow_)
  )

  col_names_ <- rep(.col_id, each = nrow_)
  rows_index <- rep(seq_len(nrow_), ncol_)

  ft <- flextable(data)

  ft <- set_header_df(
    ft,
    mapping = data.frame(
      col_keys = .col_id,
      label = col_labels,
      stringsAsFactors = FALSE
    )
  )
  for (iter in seq_along(rows_index)) {
    # val <- sprintf("value ~ formatC_with_na(%s, digits = %.0f, format = '%s', na_string = '%s')", col_names_[iter], digits_val[iter], display_val[iter], NA.string )
    ft <- compose(
      ft,
      j = col_names_[iter],
      i = rows_index[iter],
      value = as_paragraph(as_chunk(
        get(col_names_[iter]),
        formatter = format_fun
      ))
    )
  }
  ft <- border(x = ft, border = fp_border(width = 0), part = "all")
  ft <- style(x = ft, pr_t = text.properties, part = "all")
  ft <- bg(x = ft, bg = "transparent", part = "all")
  ft <- bold(x = ft, bold = TRUE, part = "header")
  if (include.rownames) {
    ft <- bold(x = ft, j = 1, bold = TRUE, part = "body")
  }

  parProp <- fp_par(padding.left = padding.left, padding.right = padding.right)

  if (any(align == "|")) {
    new_align <- character(0)
    border_right_pos <- integer(0)
    do_left_table <- FALSE
    do_right_table <- FALSE
    for (i in seq_along(align)) {
      if (i == 1 && align[i] == "|") {
        do_left_table <- TRUE
      } else if (align[i] == "|" && i < length(align)) {
        border_right_pos <- append(border_right_pos, length(new_align))
      } else if (align[i] == "|" && i == length(align)) {
        do_right_table <- TRUE
      } else {
        new_align <- append(new_align, align[i])
      }
    }
    align <- new_align

    if (do_left_table) {
      ft <- border(ft, j = 1, border.left = fp_border(), part = "all")
    }
    if (length(border_right_pos) > 0) {
      ft <- border(
        ft,
        j = border_right_pos,
        border.right = fp_border(),
        part = "all"
      )
    }
    if (do_right_table) {
      ft <- border(
        ft,
        j = length(align),
        border.right = fp_border(),
        part = "all"
      )
    }
  }

  if (rotate.colnames) {
    ft <- rotate(x = ft, rotation = "btlr", part = "header")
    header_dims <- dim_pretty(ft, part = "header")
    body_dims <- dim_pretty(ft, part = "body")
    footer_dims <- dim_pretty(ft, part = "footer")

    widths_header <- rep(header_dims$heights, each = length(ft$col_keys))
    widths_header <- matrix(widths_header, ncol = length(ft$col_keys))
    widths_header <- as.numeric(apply(widths_header, 2, max, na.rm = TRUE))
    heights_header <- rep(max(header_dims$widths), nrow_part(ft, "header"))
    header_dims$widths <- widths_header
    header_dims$heights <- heights_header

    widths_ <- do.call(
      rbind,
      list(header_dims$widths, body_dims$widths, footer_dims$widths)
    )
    widths_ <- as.numeric(apply(widths_, 2, max, na.rm = TRUE))
    heights_ <- do.call(
      c,
      list(header_dims$heights, body_dims$heights, footer_dims$heights)
    )
    ft <- width(ft, width = widths_)
    ft <- height(ft, height = header_dims$heights, part = "header")
    ft <- height(ft, height = body_dims$heights, part = "body")
    ft <- height(ft, height = footer_dims$heights, part = "footer")
    ft <- align(ft, align = "left", part = "header")
  } else {
    ft <- autofit(ft)
  }

  widths <- get_xtable_widths(align)
  align[!is.na(widths)] <- "j"
  if (!all(is.na(widths))) {
    j <- which(!is.na(widths))
    width <- widths[j]
    ft <- width(ft, j = j, width = width)
  }

  ft <- align(ft, j = align %in% "r", part = "all", align = "right")
  ft <- align(ft, j = align %in% "l", part = "all", align = "left")
  ft <- align(ft, j = align %in% "c", part = "all", align = "center")
  ft <- align(ft, j = align %in% "j", part = "all", align = "justify")

  if (!is.null(hline.after)) {
    if (-1 %in% hline.after) {
      hline.after <- setdiff(hline.after, -1)
      ft <- border(ft, i = 1, border.top = fp_border(), part = "header")
    }
    if (0 %in% hline.after) {
      hline.after <- setdiff(hline.after, 0)
      ft <- border(ft, border.bottom = fp_border(), part = "header")
    }
    if (length(hline.after) > 0) {
      ft <- border(
        ft,
        i = hline.after,
        border.bottom = fp_border(),
        part = "body"
      )
    }
  }

  ft
}


get_xtable_widths <- function(align, default_width = .3) {
  rex <- "^(p\\{)([0-9\\.]+)(cm|in|px)(\\}$)"
  width <- rep(default_width, length(align))
  w_matches <- grepl(rex, align)
  width[w_matches]
  newwidths <- gsub(rex, "\\2", align)
  newwidths[!w_matches] <- ""
  newwidths <- as.numeric(newwidths)

  units <- gsub(rex, "\\3", align)
  units[!w_matches] <- "in"
  if (any(!units %in% c("in", "cm", "px"))) {
    stop("unknown unit, supported units for column width are in, cm and px")
  }
  newwidths[units %in% "cm"] <- newwidths[units %in% "cm"] / 2.54
  newwidths[units %in% "px"] <- newwidths[units %in% "cm"] / 72
  newwidths
}

# helpers ----

.levels_summary <- function(x, max_levels) {
  if (length(x) > max_levels) {
    paste0(
      paste0("'", head(x, n = max_levels), "'", collapse = ", "),
      ", ..."
    )
  } else {
    paste0("'", x, "'", collapse = ", ")
  }
}

.col_summary <- function(col, nm, max_levels) {
  na_count <- sum(is.na(col))
  non_na <- sum(!is.na(col))

  out <- data.frame(
    col_name = nm,
    col_type = NA_character_,
    n = NA_integer_,
    na_count = na_count,
    min_val = NA_real_,
    max_val = NA_real_,
    str_val = NA_character_,
    data_kind = NA_character_,
    stringsAsFactors = FALSE
  )

  if (is.logical(col)) {
    out$col_type <- "logical"
    out$n <- non_na
    out$data_kind <- "logical"
    if (non_na > 0L) {
      out$str_val <- paste0(
        "TRUE: ",
        sum(col, na.rm = TRUE),
        ", FALSE: ",
        sum(!col, na.rm = TRUE)
      )
    }
  } else if (is.numeric(col)) {
    out$col_type <- if (is.integer(col)) "integer" else "numeric"
    out$n <- non_na
    out$data_kind <- "numeric"
    if (non_na > 0L) {
      out$min_val <- min(col, na.rm = TRUE)
      out$max_val <- max(col, na.rm = TRUE)
    }
  } else if (is.factor(col)) {
    levs <- levels(droplevels(col))
    out$col_type <- "factor"
    out$n <- length(levs)
    out$data_kind <- "discrete"
    if (length(levs) > 0L) {
      out$str_val <- .levels_summary(levs, max_levels = max_levels)
    }
  } else if (is.character(col)) {
    vals <- sort(unique(na.omit(col)))
    out$col_type <- "character"
    out$n <- length(vals)
    out$data_kind <- "discrete"
    if (length(vals) > 0L) {
      out$str_val <- .levels_summary(vals, max_levels = max_levels)
    }
  } else if (
    inherits(col, "Date") ||
      inherits(col, "POSIXt") ||
      inherits(col, "hms") ||
      inherits(col, "difftime")
  ) {
    out$col_type <- if (inherits(col, "Date")) {
      "Date"
    } else if (inherits(col, "POSIXt")) {
      "POSIXct"
    } else {
      "hms"
    }
    out$n <- non_na
    out$data_kind <- "range"
    if (non_na > 0L) {
      out$str_val <- paste0(
        format(min(col, na.rm = TRUE)),
        " - ",
        format(max(col, na.rm = TRUE))
      )
    }
  } else {
    return(NULL)
  }
  out
}

# main ----

#' @export
#' @title Compact Summary of a Dataset
#' @description
#' Produces a compact summary of a data.frame. Each row of the
#' result describes one column of the input with a type-specific
#' synopsis.
#'
#' Supported types and their summaries:
#'
#' - **numeric / integer**: count of non-`NA` values, min and max.
#' - **character**: number of unique values, first values listed.
#' - **factor**: number of levels, levels listed.
#' - **logical**: count of non-`NA` values, counts of `TRUE` and
#'   `FALSE`.
#' - **Date**: count of non-`NA` values, date range.
#' - **POSIXct / POSIXlt**: count of non-`NA` values, datetime
#'   range.
#' - **hms / difftime**: count of non-`NA` values, time range.
#'
#' Character and factor columns share the same summary layout
#' but report a different type label.
#'
#' The result has class `"compact_summary"` and can be converted
#' into a flextable with [as_flextable()].
#' @param x A data.frame.
#' @param show_type If `TRUE`, a *Type* column is added when the
#'   object is rendered as a flextable.
#' @param show_na If `TRUE`, a *NA* column showing the count of
#'   missing values per column is added when rendered as a
#'   flextable.
#' @param max_levels Maximum number of levels or unique values
#'   displayed for factor and character columns. Additional
#'   values are replaced by `", ..."`.
#' @return A data.frame with additional class `"compact_summary"`.
#' @examples
#' z <- compact_summary(iris)
#' as_flextable(z)
#'
#' z <- compact_summary(iris, show_type = TRUE, show_na = TRUE)
#' as_flextable(z)
#' @family as_flextable methods
#' @seealso [as_flextable.compact_summary()]
compact_summary <- function(
  x,
  show_type = FALSE,
  show_na = FALSE,
  max_levels = 10L
) {
  if (!is.data.frame(x)) {
    stop("compact_summary() expects a data.frame.", call. = FALSE)
  }

  summaries <- mapply(
    .col_summary,
    col = x,
    nm = names(x),
    MoreArgs = list(max_levels = max_levels),
    SIMPLIFY = FALSE
  )
  result <- rbindlist(Filter(Negate(is.null), summaries))
  setDF(result)

  if (is.null(result)) {
    result <- data.frame(
      col_name = character(0),
      col_type = character(0),
      n = integer(0),
      na_count = integer(0),
      min_val = numeric(0),
      max_val = numeric(0),
      str_val = character(0),
      data_kind = character(0),
      stringsAsFactors = FALSE
    )
  }
  rownames(result) <- NULL

  attr(result, "show_type") <- show_type
  attr(result, "show_na") <- show_na
  class(result) <- c("compact_summary", class(result))
  result
}

#' @export
#' @title Transform a 'compact_summary' object into a flextable
#' @description `compact_summary` objects can be transformed into
#' a flextable with method [as_flextable()].
#'
#' Numeric columns are formatted with [formatC()] using the
#' `digits` value stored in the object and the current
#' flextable defaults for `big.mark` and `decimal.mark`.
#' @param x A `compact_summary` object produced by
#'   [compact_summary()].
#' @param ... unused arguments.
#' @return A [flextable()] object.
#' @examples
#' z <- compact_summary(iris, show_type = TRUE, show_na = TRUE)
#' as_flextable(z)
#' @family as_flextable methods
as_flextable.compact_summary <- function(x, ...) {
  show_type <- attr(x, "show_type")
  show_na <- attr(x, "show_na")

  z <- get_flextable_defaults()
  format_num <- function(val) {
    formatC(
      val,
      format = "f",
      big.mark = z$big.mark,
      digits = z$digits,
      decimal.mark = z$decimal.mark
    )
  }

  col_keys <- "col_name"
  if (isTRUE(show_type)) {
    col_keys <- c(col_keys, "col_type")
  }
  col_keys <- c(col_keys, "n")
  if (isTRUE(show_na)) {
    col_keys <- c(col_keys, "na_count")
  }
  col_keys <- c(col_keys, "Values")

  ft <- flextable(x, col_keys = col_keys)
  ft <- set_header_labels(
    ft,
    col_name = "Column",
    col_type = "Type",
    n = "N",
    na_count = "NA",
    Values = "Values"
  )

  is_num <- which(x$data_kind %in% "numeric")
  if (length(is_num) > 0L) {
    ft <- mk_par(
      ft,
      i = is_num,
      j = "Values",
      value = as_paragraph(
        "Min: ",
        as_chunk(min_val, formatter = format_num),
        ", Max: ",
        as_chunk(max_val, formatter = format_num)
      )
    )
  }

  is_str <- which(!x$data_kind %in% "numeric")
  if (length(is_str) > 0L) {
    ft <- mk_par(
      ft,
      i = is_str,
      j = "Values",
      value = as_paragraph(as_chunk(str_val))
    )
  }

  ft <- colformat_int(ft, j = "n")
  if (isTRUE(show_na)) {
    ft <- colformat_int(ft, j = "na_count")
  }

  autofit(ft)
}

# fpstruct ------

fpstruct <- function(nrow, keys, default) {
  ncol <- length(keys)
  data <- rep(default, length.out = nrow * ncol)
  map_data <- matrix(
    data = data,
    nrow = nrow,
    ncol = ncol,
    dimnames = list(NULL, keys)
  )

  x <- list(
    data = map_data,
    keys = keys,
    nrow = nrow,
    ncol = ncol,
    default = default
  )
  class(x) <- "fpstruct"
  x
}

delete_row_from_fpstruct <- function(x, i) {
  x$data <- x$data[-i, , drop = FALSE]
  x$nrow <- x$nrow - length(i)
  x
}
delete_col_from_fpstruct <- function(x, j) {
  if (is.null(x$data)) {
    stop("unexpected error, could not find any data to drop")
  }
  x$data <- x$data[, !colnames(x$data) %in% j, drop = FALSE]
  x$ncol <- x$ncol - length(j)
  x$keys <- setdiff(x$keys, j)

  x
}

get_fpstruct_elements <- function(x, i, j) {
  if (is.null(x$data)) {
    stop("data coumpound does not exits.")
  }
  x$data[i, j, drop = FALSE]
}


#' @importFrom utils head tail
add_rows_fpstruct <- function(x, nrows, first, default = x$default, ...) {
  if (nrow(x$data) < 1) {
    new <- matrix(rep(default, x$ncol * nrows), ncol = x$ncol)
  } else if (first) {
    default <- as.vector(head(x$data, n = 1))
    new <- matrix(rep(default, each = nrows), ncol = x$ncol)
  } else {
    default <- as.vector(tail(x$data, n = 1))
    new <- matrix(rep(default, each = nrows), ncol = x$ncol)
  }
  if (first) {
    x$data <- rbind(new, x$data)
  } else {
    x$data <- rbind(x$data, new)
  }
  x$nrow <- nrow(x$data)
  x
}

# text_struct ------
text_struct <- function(
  nrow,
  keys,
  color = "black",
  font.size = 10,
  bold = FALSE,
  italic = FALSE,
  underlined = FALSE,
  strike = FALSE,
  font.family = "Arial",
  hansi.family = "Arial",
  eastasia.family = "Arial",
  cs.family = "Arial",
  vertical.align = "baseline",
  shading.color = "transparent",
  ...
) {
  x <- list(
    color = fpstruct(nrow = nrow, keys = keys, default = color),
    font.size = fpstruct(nrow = nrow, keys = keys, default = font.size),
    bold = fpstruct(nrow = nrow, keys = keys, default = bold),
    italic = fpstruct(nrow = nrow, keys = keys, default = italic),
    underlined = fpstruct(nrow = nrow, keys = keys, default = underlined),
    strike = fpstruct(nrow = nrow, keys = keys, default = strike),
    font.family = fpstruct(nrow = nrow, keys = keys, default = font.family),
    hansi.family = fpstruct(nrow = nrow, keys = keys, default = hansi.family),
    eastasia.family = fpstruct(
      nrow = nrow,
      keys = keys,
      default = eastasia.family
    ),
    cs.family = fpstruct(nrow = nrow, keys = keys, default = cs.family),
    vertical.align = fpstruct(
      nrow = nrow,
      keys = keys,
      default = vertical.align
    ),
    shading.color = fpstruct(nrow = nrow, keys = keys, default = shading.color)
  )
  class(x) <- "text_struct"
  x
}

set_text_struct_values <- function(x, i, j, property, value) {
  if (is.null(j)) {
    j <- x$color$keys
  }
  if (is.null(i)) {
    i <- seq_len(x$color$nrow)
  }

  if (inherits(value, "fp_text")) {
    for (property in intersect(names(value), names(x))) {
      x[[property]]$data[i, j] <- value[[property]]
    }
  } else if (property %in% names(x)) {
    x[[property]]$data[i, j] <- value
  }

  x
}
# `[.text_struct` <- function(x, i, j, property, value) {
#   x[[property]][i, j]
# }

delete_style_row <- function(x, i) {
  for (property in names(x)) {
    x[[property]] <- delete_row_from_fpstruct(x[[property]], i)
  }
  x
}
delete_style_col <- function(x, j) {
  for (property in names(x)) {
    x[[property]] <- delete_col_from_fpstruct(x[[property]], j)
  }
  x
}

add_rows_to_struct <- function(x, nrows, first, ...) {
  for (i in seq_len(length(x))) {
    x[[i]] <- add_rows_fpstruct(x[[i]], nrows, first = first)
  }
  x
}

text_struct_to_df <- function(object, ...) {
  data <- lapply(object, function(x) {
    as.vector(x$data)
  })
  data$.row_id <- rep(seq_len(nrow(object$color$data)), ncol(object$color$data))
  data$.col_id <- rep(object$color$keys, each = nrow(object$color$data))
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  data$.col_id <- factor(data$.col_id, levels = object$color$keys)
  data
}


# par_struct -----
par_struct <- function(
  nrow,
  keys,
  text.align = "left",
  line_spacing = 1,
  padding.bottom = 0,
  padding.top = 0,
  padding.left = 0,
  padding.right = 0,
  border.width.bottom = 0,
  border.width.top = 0,
  border.width.left = 0,
  border.width.right = 0,
  border.color.bottom = "transparent",
  border.color.top = "transparent",
  border.color.left = "transparent",
  border.color.right = "transparent",
  border.style.bottom = "solid",
  border.style.top = "solid",
  border.style.left = "solid",
  border.style.right = "solid",
  keep_with_next = FALSE,
  word_style = NA_character_,
  tabs = NA_character_,
  shading.color = "transparent",
  first_line = NA_real_,
  hanging = NA_real_,
  ...
) {
  x <- list(
    text.align = fpstruct(nrow = nrow, keys = keys, default = text.align),
    padding.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = padding.bottom
    ),
    padding.top = fpstruct(nrow = nrow, keys = keys, default = padding.top),
    padding.left = fpstruct(nrow = nrow, keys = keys, default = padding.left),
    padding.right = fpstruct(nrow = nrow, keys = keys, default = padding.right),
    line_spacing = fpstruct(nrow = nrow, keys = keys, default = line_spacing),
    border.width.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.bottom
    ),
    border.width.top = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.top
    ),
    border.width.left = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.left
    ),
    border.width.right = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.right
    ),
    border.color.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.bottom
    ),
    border.color.top = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.top
    ),
    border.color.left = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.left
    ),
    border.color.right = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.right
    ),
    border.style.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.bottom
    ),
    border.style.top = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.top
    ),
    border.style.left = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.left
    ),
    border.style.right = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.right
    ),
    shading.color = fpstruct(nrow = nrow, keys = keys, default = shading.color),
    keep_with_next = fpstruct(
      nrow = nrow,
      keys = keys,
      default = keep_with_next
    ),
    word_style = fpstruct(nrow = nrow, keys = keys, default = word_style),
    tabs = fpstruct(nrow = nrow, keys = keys, default = tabs),
    first_line = fpstruct(nrow = nrow, keys = keys, default = first_line),
    hanging = fpstruct(nrow = nrow, keys = keys, default = hanging)
  )
  class(x) <- "par_struct"
  x
}

set_par_struct_values <- function(x, i, j, property, value) {
  if (is.null(j)) {
    j <- x$text.align$keys
  }
  if (is.null(i)) {
    i <- seq_len(x$text.align$nrow)
  }

  if (inherits(value, "fp_par")) {
    if (!is.null(value$tabs) && !isFALSE(value$tabs)) {
      value$tabs <- as.character(value$tabs)
    } else {
      value$tabs <- as.character(fp_tabs())
    }
    value <- cast_borders(value)
    for (property in intersect(names(value), names(x))) {
      x[[property]]$data[i, j] <- value[[property]]
    }
  } else if (property %in% names(x)) {
    x[[property]]$data[i, j] <- value
  }

  x
}

par_struct_to_df <- function(object, ...) {
  data <- lapply(object, function(x) {
    as.vector(x$data)
  })
  data$.row_id <- rep(
    seq_len(nrow(object$text.align$data)),
    ncol(object$text.align$data)
  )
  data$.col_id <- rep(
    object$text.align$keys,
    each = nrow(object$text.align$data)
  )
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  data$.col_id <- factor(data$.col_id, levels = object$text.align$keys)
  data
}


# cell_struct -----
cell_struct <- function(
  nrow,
  keys,
  vertical.align = "top",
  text.direction = "lrtb",
  margin.bottom = 0,
  margin.top = 0,
  margin.left = 0,
  margin.right = 0,
  border.width.bottom = 1,
  border.width.top = 1,
  border.width.left = 1,
  border.width.right = 1,
  border.color.bottom = "transparent",
  border.color.top = "transparent",
  border.color.left = "transparent",
  border.color.right = "transparent",
  border.style.bottom = "solid",
  border.style.top = "solid",
  border.style.left = "solid",
  border.style.right = "solid",
  background.color = "#34CC27",
  width = NA_real_,
  height = NA_real_,
  hrule = "auto",
  ...
) {
  check_choice(value = vertical.align, choices = c("top", "center", "bottom"))
  check_choice(value = text.direction, choices = c("lrtb", "tbrl", "btlr"))

  x <- list(
    vertical.align = fpstruct(
      nrow = nrow,
      keys = keys,
      default = vertical.align
    ),
    width = fpstruct(nrow = nrow, keys = keys, default = width),
    height = fpstruct(nrow = nrow, keys = keys, default = height),
    margin.bottom = fpstruct(nrow = nrow, keys = keys, default = margin.bottom),
    margin.top = fpstruct(nrow = nrow, keys = keys, default = margin.top),
    margin.left = fpstruct(nrow = nrow, keys = keys, default = margin.left),
    margin.right = fpstruct(nrow = nrow, keys = keys, default = margin.right),
    border.width.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.bottom
    ),
    border.width.top = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.top
    ),
    border.width.left = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.left
    ),
    border.width.right = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.width.right
    ),
    border.color.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.bottom
    ),
    border.color.top = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.top
    ),
    border.color.left = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.left
    ),
    border.color.right = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.color.right
    ),
    border.style.bottom = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.bottom
    ),
    border.style.top = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.top
    ),
    border.style.left = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.left
    ),
    border.style.right = fpstruct(
      nrow = nrow,
      keys = keys,
      default = border.style.right
    ),
    text.direction = fpstruct(
      nrow = nrow,
      keys = keys,
      default = text.direction
    ),
    background.color = fpstruct(
      nrow = nrow,
      keys = keys,
      default = background.color
    ),
    hrule = fpstruct(nrow = nrow, keys = keys, default = hrule)
  )
  class(x) <- "cell_struct"
  x
}

cell_struct_to_df <- function(object, ...) {
  data <- lapply(object, function(x) {
    as.vector(x$data)
  })

  data$.row_id <- rep(
    seq_len(nrow(object$background.color$data)),
    ncol(object$background.color$data)
  )
  data$.col_id <- rep(
    object$background.color$keys,
    each = nrow(object$background.color$data)
  )
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  data$.col_id <- factor(data$.col_id, levels = object$background.color$keys)
  data
}


# chunkset_struct ---------------------------------------------------------

# This object is used to capture paragraphs of a part of a flextable
# It is a matrix, each column is a colkey, each row is a row
# It contains paragraphs, paragraphs are made of chunks
new_chunkset_struct <- function(col_keys, data) {
  chunkdata <- fpstruct(
    nrow = nrow(data),
    keys = col_keys,
    default = as_paragraph(as_chunk(""))
  )
  class(chunkdata) <- c("chunkset_struct")

  if (nrow(data) > 0) {
    newchunkdata <- lapply(
      data[col_keys],
      function(x) {
        as_paragraph(as_chunk(x, formatter = format_fun.default))
      }
    )
    newchunkdata <- matrix(
      do.call(c, newchunkdata),
      ncol = length(col_keys),
      dimnames = list(NULL, col_keys)
    )
    chunkdata <- set_chunkset_struct_element(
      x = chunkdata,
      i = seq_len(nrow(data)),
      j = col_keys,
      value = newchunkdata
    )
  }
  chunkdata
}

add_rows_to_chunkset_struct <- function(x, nrows, first, data, ...) {
  names_ <- names(data)
  stopifnot(!is.null(names_))

  x <- add_rows_fpstruct(
    x,
    nrows,
    first = first,
    default = as_paragraph(as_chunk(""))
  )
  if (first) {
    id <- seq_len(nrows)
  } else {
    id <- rev(rev(seq_len(x$nrow))[seq_len(nrows)])
  }

  newchunkdata <- lapply(data[x$keys], function(x) {
    as_paragraph(as_chunk(x, formatter = format_fun.default))
  })
  newchunkdata <- matrix(
    do.call(c, newchunkdata),
    ncol = length(x$keys),
    dimnames = list(NULL, x$keys)
  )

  x <- set_chunkset_struct_element(
    x = x,
    i = id,
    j = x$keys,
    value = newchunkdata
  )
  x
}

as_chunkset_struct <- function(l_paragraph, keys, i = NULL) {
  if (
    !is.null(i) &&
      length(l_paragraph) == length(i) &&
      length(keys) > 1
  ) {
    l_paragraph <- rep(l_paragraph, length(keys))
  }
  # temp fix for ftExtra
  l_paragraph[] <- lapply(l_paragraph, function(x) {
    .names <- colnames(x)
    .names[.names %in% "seq_index"] <- ".chunk_index"
    colnames(x) <- .names
    x
  })

  matrix(
    data = l_paragraph,
    ncol = length(keys),
    dimnames = list(NULL, keys)
  )
}

is_paragraph <- function(x) {
  chunk_str_names <- c(
    "txt",
    "font.size",
    "italic",
    "bold",
    "underlined",
    "strike",
    "color",
    "shading.color",
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "vertical.align",
    "width",
    "height",
    "url",
    "eq_data",
    "word_field_data",
    "qmd_data",
    "img_data",
    "alt",
    ".chunk_index"
  )
  is.data.frame(x) &&
    all(colnames(x) %in% chunk_str_names)
}
set_chunkset_struct_element <- function(x, i, j, value) {
  names_ <- colnames(value)
  stopifnot(
    is.matrix(value),
    !is.null(names_),
    mode(value) == "list",
    all(sapply(value, is_paragraph)),
    all(names_ %in% x$keys)
  )

  x$data[i, j] <- value
  x
}

append_chunkset_struct_element <- function(x, i, j, chunk_data, last = TRUE) {
  chunk_str_names <- c(
    "txt",
    "font.size",
    "italic",
    "bold",
    "underlined",
    "strike",
    "color",
    "shading.color",
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "vertical.align",
    "width",
    "height",
    "url",
    "eq_data",
    "word_field_data",
    "qmd_data",
    "img_data"
  )
  stopifnot(
    is.data.frame(chunk_data),
    all(chunk_str_names %in% colnames(chunk_data))
  )
  chunk_data <- chunk_data[, chunk_str_names, drop = FALSE]

  chunk_data_length <- nrow(chunk_data)
  i_length <- length(i)
  j_length <- length(j)
  expected_length <- j_length * i_length

  if (chunk_data_length == 1L && i_length != chunk_data_length) {
    chunk_data <- rep(list(chunk_data), i_length)
    chunk_data <- rbind_match_columns(chunk_data)
  }

  if (expected_length / nrow(chunk_data) == j_length) {
    chunk_data <- rep(list(chunk_data), j_length)
    chunk_data <- rbind_match_columns(chunk_data)
  }

  stopifnot(nrow(chunk_data) == expected_length)

  if (nrow(chunk_data) == 1) {
    chunk_data <- list(chunk_data)
  } else {
    chunk_data <- split(chunk_data, seq_len(expected_length))
    names(chunk_data) <- NULL
  }

  values <- get_chunkset_struct_element(x, i = i, j = j)
  values <- do.call(c, apply(values, 2, function(x) x))
  names(values) <- NULL

  values <- mapply(
    function(x, y, last = TRUE) {
      if (last) {
        y$.chunk_index <- max(x$.chunk_index, na.rm = TRUE) + 1
        x <- rbind_match_columns(list(x, y))
      } else {
        y$.chunk_index <- min(x$.chunk_index, na.rm = TRUE) - 1
        x <- rbind_match_columns(list(y, x))
      }
      x$.chunk_index <- rleid(x$.chunk_index)
      x
    },
    x = values,
    y = chunk_data,
    SIMPLIFY = FALSE,
    MoreArgs = list(last = last)
  )

  x$data[i, j] <- values
  x
}

get_chunkset_struct_element <- function(x, i, j) {
  x$data[i, j, drop = FALSE]
}

replace_missing_fptext_by_default <- function(x, default) {
  by_columns <- c(
    "font.size",
    "italic",
    "bold",
    "underlined",
    "strike",
    "color",
    "shading.color",
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "vertical.align"
  )
  by_columns <- intersect(by_columns, names(default))

  keys <- default[, setdiff(names(default), by_columns), drop = FALSE]
  values <- default[, by_columns, drop = FALSE]
  names(values) <- paste0(by_columns, "_default")
  defdata <- cbind(keys, values)

  newx <- x
  setDT(newx)
  setDT(defdata)
  newx <- newx[defdata, on = names(keys)]
  setDF(newx)
  for (j in by_columns) {
    if (!is.null(newx[[j]])) {
      newx[[j]] <- ifelse(
        is.na(newx[[j]]),
        newx[[paste0(j, "_default")]],
        newx[[j]]
      )
    } else {
      newx[[j]] <- newx[[paste0(j, "_default")]]
    }
    newx[[paste0(j, "_default")]] <- NULL
  }
  newx
}

#' @importFrom data.table shift
rtf_cells <- function(value, cell_data, layout = "fixed") {
  cell_heights <- fortify_height(value)
  cell_widths <- fortify_width(value)

  cell_data$width <- NULL # need to get rid of originals that are empty, should probably rm them
  cell_data$height <- NULL
  cell_data <- merge(cell_data, cell_widths, by = ".col_id")
  cell_data <- merge(cell_data, cell_heights, by = c(".part", ".row_id"))

  setDT(cell_data)
  setorderv(cell_data, cols = c(".part", ".row_id", ".col_id"))

  # fix for word horiz. borders, copying the bottom props to top props of the next cell
  cell_data <- copy_border_bottom_to_next_border_top(cell_data, value = value)

  data_ref_cells <- distinct_cells_properties(cell_data)

  ## cell style wml
  fp_cell_rtf <- data_ref_cells
  classnames <- data_ref_cells$classname
  fp_cell_rtf$classname <- NULL

  fp_cell_rtf <- split(fp_cell_rtf, classnames)
  fp_cell_rtf <- vapply(
    fp_cell_rtf,
    function(x) {
      zz <- as.list(x)
      zz$border.bottom <- fp_border(
        color = zz$border.color.bottom,
        width = zz$border.width.bottom,
        style = zz$border.style.bottom
      )
      zz$border.top <- fp_border(
        color = zz$border.color.top,
        width = zz$border.width.top,
        style = zz$border.style.top
      )
      zz$border.right <- fp_border(
        color = zz$border.color.right,
        width = zz$border.width.right,
        style = zz$border.style.right
      )
      zz$border.left <- fp_border(
        color = zz$border.color.left,
        width = zz$border.width.left,
        style = zz$border.style.left
      )

      zz[c(
        "border.width.bottom",
        "border.width.top",
        "border.width.left",
        "border.width.right",
        "border.color.bottom",
        "border.color.top",
        "border.color.left",
        "border.color.right",
        "border.style.bottom",
        "border.style.top",
        "border.style.left",
        "border.style.right",
        "width",
        "height",
        "hrule"
      )] <- NULL
      zz$classname <- NULL
      zz <- do.call(fp_cell, zz)
      zz <- format(zz, type = "rtf")
      zz
    },
    FUN.VALUE = "",
    USE.NAMES = FALSE
  )

  style_dat <- data.frame(
    fp_cell_rtf = fp_cell_rtf,
    classname = classnames,
    stringsAsFactors = FALSE
  )

  # organise everything
  cell_data <- merge(
    cell_data,
    data_ref_cells,
    by = intersect(colnames(cell_data), colnames(data_ref_cells))
  )
  cell_data <- merge(cell_data, style_dat, by = "classname")
  setorderv(cell_data, cols = c(".part", ".row_id", ".col_id"))
  if (layout %in% "fixed") {
    cell_data[,
      c("fp_cell_rtf") := list(
        sprintf("%s\\cellx%.0f", .SD$fp_cell_rtf, cumsum(.SD$width) * 1440)
      ),
      by = c(".part", ".row_id")
    ]
  } else {
    cell_data[,
      c("fp_cell_rtf") := list(
        sprintf("%s\\cellx", .SD$fp_cell_rtf)
      ),
      by = c(".part", ".row_id")
    ]
  }
  cell_data <- cell_data[,
    .SD,
    .SDcols = c(".part", ".row_id", ".col_id", "fp_cell_rtf")
  ]
  setDF(cell_data)
  cell_data
}


rtf_rows <- function(value) {
  # prepare cells formatting properties and add span information
  cell_attributes <- information_data_cell(value)
  span_data <- fortify_span(value)
  setDT(cell_attributes)
  cell_attributes <- merge(
    cell_attributes,
    span_data,
    by = c(".part", ".row_id", ".col_id")
  )
  setDF(cell_attributes)

  # prepare paragraphs formatting properties
  paragraphs_properties <- information_data_paragraph(value)

  # transform alignments for rotated text
  # and add them back to paragraphs_properties and cell_attributes
  new_pos <- ooxml_rotation_alignments(
    rotation = cell_attributes$text.direction,
    valign = cell_attributes$vertical.align,
    align = paragraphs_properties$text.align
  )
  paragraphs_properties$text.align <- new_pos$align
  cell_attributes$vertical.align <- new_pos$valign

  # get runs in wml format and get hyperlinks and images information
  run_data <- runs_as_rtf(value, chunk_data = information_data_chunk(value))
  cell_data <- rtf_cells(
    value,
    cell_attributes,
    layout = value$properties$layout
  )
  cell_heights <- fortify_height(value)
  cell_hrule <- fortify_hrule(value)

  setDT(cell_data)
  tab_data <- merge(
    cell_data,
    ooxml_ppr(paragraphs_properties, type = "rtf"),
    by = c(".part", ".row_id", ".col_id")
  )
  tab_data <- merge(tab_data, run_data, by = c(".part", ".row_id", ".col_id"))
  tab_data <- merge(tab_data, span_data, by = c(".part", ".row_id", ".col_id"))
  setorderv(tab_data, cols = c(".part", ".row_id", ".col_id"))

  tab_data[tab_data$colspan < 1, c("txt") := list("")]
  tab_data[tab_data$rowspan < 1, c("txt") := list("")]

  tab_data[,
    c("txt", "fp_par_xml") := list(
      paste0(.SD$fp_par_xml, .SD$txt),
      NULL
    )
  ]

  cells <- tab_data[,
    list(
      fp_cell_rtf = paste0(.SD$fp_cell_rtf, collapse = ""),
      content_rtf = paste0(.SD$txt, "\\cell", collapse = "")
    ),
    by = c(".part", ".row_id")
  ]

  split_str <- ""
  if (!value$properties$opts_word$split) {
    split_str <- "\\trkeep"
  }

  align <- match.arg(
    value$properties$align,
    c("center", "left", "right"),
    several.ok = FALSE
  )
  align <- c("center" = "\\trqc", "left" = "\\trql", "right" = "\\trqr")[align]

  hrule <- cell_hrule$hrule
  heights <- round(cell_heights$height * 1440, 0)
  heights[hrule %in% "auto"] <- 0
  heights[hrule %in% "exact"] <- -heights[hrule %in% "exact"]
  heights <- paste0("\\trrh", heights)

  header_str <- rep("", nrow(cell_hrule))
  header_str[cell_hrule$.part %in% "header"] <- "\\trhdr"

  if (value$properties$layout %in% "fixed") {
    autofit <- "\\trautofit0"
  } else {
    autofit <- "\\trautofit1"
  }

  header_str <- rep("", nrow(cell_hrule))
  header_str[cell_hrule$.part %in% "header"] <- "\\trhdr"

  rows <- paste0(
    "\\trowd",
    autofit,
    "\\trgaph0\\trleft0\\tapltr",
    align,
    split_str,
    heights,
    header_str,
    cells$fp_cell_rtf,
    cells$content_rtf,
    "\\row\\pard\n"
  )

  paste0(paste0(rows, collapse = ""), "\\vertalt")
}

#' @export
to_rtf.flextable <- function(x, topcaption = TRUE, ...) {
  tab_str <- rtf_rows(x)
  if (has_caption(x)) {
    if (topcaption) {
      tab_str <- paste0(caption_default_rtf(x), "\n\n", tab_str)
    } else {
      tab_str <- paste0(tab_str, "\n\n", caption_default_rtf(x))
    }
  }
  tab_str
}

#' @export
#' @importFrom officer rtf_add
#' @title Add a 'flextable' into an RTF document
#' @family officer_integration
#' @description [officer::rtf_add()] method for adding
#' flextable objects into 'RTF' documents.
#' @param x rtf object, created by [officer::rtf_doc()].
#' @param value a flextable object
#' @param ... unused arguments
#' @examples
#' library(flextable)
#' library(officer)
#'
#' ft <- flextable(head(iris))
#' ft <- autofit(ft)
#'
#' z <- rtf_doc()
#' z <- rtf_add(z, ft)
#'
#' print(z, target = tempfile(fileext = ".rtf"))
rtf_add.flextable <- function(x, value, ...) {
  x$content[[length(x$content) + 1]] <- value
  x
}

ooxml_rotation_alignments <- function(rotation, align, valign) {
  halign_out <- align
  valign_out <- valign

  left_top <- rotation %in% "btlr" & valign %in% "top" & align %in% "left"
  center_top <- rotation %in% "btlr" & valign %in% "top" & align %in% "center"
  right_top <- rotation %in% "btlr" & valign %in% "top" & align %in% "right"
  left_middle <- rotation %in% "btlr" & valign %in% "center" & align %in% "left"
  center_middle <- rotation %in% "btlr" & valign %in% "center" & align %in% "center"
  right_middle <- rotation %in% "btlr" & valign %in% "center" & align %in% "right"
  left_bottom <- rotation %in% "btlr" & valign %in% "bottom" & align %in% "left"
  center_bottom <- rotation %in% "btlr" & valign %in% "bottom" & align %in% "center"
  right_bottom <- rotation %in% "btlr" & valign %in% "bottom" & align %in% "right"

  # left-top to right-top
  halign_out[left_top] <- "right"
  valign_out[left_top] <- "top"
  # center-top to right-center
  halign_out[center_top] <- "right"
  valign_out[center_top] <- "center"
  # right-top to right-bottom
  halign_out[right_top] <- "right"
  valign_out[right_top] <- "bottom"
  # left_middle to center-top
  halign_out[left_middle] <- "center"
  valign_out[left_middle] <- "top"
  # center_middle to center-center
  halign_out[center_middle] <- "center"
  valign_out[center_middle] <- "center"
  # right_middle to center-bottom
  halign_out[right_middle] <- "center"
  valign_out[right_middle] <- "bottom"
  # left_bottom to left-top
  halign_out[left_bottom] <- "left"
  valign_out[left_bottom] <- "top"
  # center_bottom to left-center
  halign_out[center_bottom] <- "left"
  valign_out[center_bottom] <- "center"
  # right_bottom to left-bottom
  halign_out[right_bottom] <- "left"
  valign_out[right_bottom] <- "bottom"

  left_top <- rotation %in% "tbrl" & valign %in% "top" & align %in% "left"
  center_top <- rotation %in% "tbrl" & valign %in% "top" & align %in% "center"
  right_top <- rotation %in% "tbrl" & valign %in% "top" & align %in% "right"
  left_middle <- rotation %in% "tbrl" & valign %in% "center" & align %in% "left"
  center_middle <- rotation %in% "tbrl" & valign %in% "center" & align %in% "center"
  right_middle <- rotation %in% "tbrl" & valign %in% "center" & align %in% "right"
  left_bottom <- rotation %in% "tbrl" & valign %in% "bottom" & align %in% "left"
  center_bottom <- rotation %in% "tbrl" & valign %in% "bottom" & align %in% "center"
  right_bottom <- rotation %in% "tbrl" & valign %in% "bottom" & align %in% "right"

  # left-top to left-bottom
  halign_out[left_top] <- "left"
  valign_out[left_top] <- "bottom"
  # center-top to left-center
  halign_out[center_top] <- "left"
  valign_out[center_top] <- "center"
  # right-top to left-top
  halign_out[right_top] <- "left"
  valign_out[right_top] <- "top"

  # left_middle
  halign_out[left_middle] <- "center"
  valign_out[left_middle] <- "bottom"
  # center_middle
  halign_out[center_middle] <- "center"
  valign_out[center_middle] <- "center"
  # right_middle
  halign_out[right_middle] <- "center"
  valign_out[right_middle] <- "top"

  # left_bottom
  halign_out[left_bottom] <- "right"
  valign_out[left_bottom] <- "bottom"
  # center_bottom
  halign_out[center_bottom] <- "right"
  valign_out[center_bottom] <- "center"
  # right_bottom
  halign_out[right_bottom] <- "right"
  valign_out[right_bottom] <- "top"
  list(align = halign_out, valign = valign_out)
}

#' @title Prepare descriptive statistics for flextable
#' @description It performs a univariate statistical analysis of a dataset
#' by group and formats the results so that they can be used with
#' the [tabulator()] function or directly with [as_flextable][as_flextable.summarizor].
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/flextable-018-square.png" alt="summarizor illustration" style="width:100\%;">
#' }}
#' @note
#' This is very first version of the function; be aware it
#' can evolve or change.
#' @param x dataset
#' @param by columns names to be used as grouping columns
#' @param overall_label label to use as overall label
#' @param num_stats available statistics for numerical columns
#' to show, available options are "mean_sd", "median_iqr" and "range".
#' @param hide_null_na if TRUE (default), NA counts will not be shown when 0.
#' @param use_labels Logical; if TRUE, any column labels or value labels
#' present in the dataset will be used for display purposes. Defaults
#' to TRUE.
#' @seealso [fmt_summarizor()], [labelizor()]
#' @examples
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' z <- summarizor(CO2[-c(1, 4)],
#'   by = "Treatment",
#'   overall_label = "Overall"
#' )
#' ft_1 <- as_flextable(z)
#' ft_1
#'
#' ft_2 <- as_flextable(z, sep_w = 0, spread_first_col = TRUE)
#' ft_2
#'
#' z <- summarizor(CO2[-c(1, 4)])
#' ft_3 <- as_flextable(z, sep_w = 0, spread_first_col = TRUE)
#' ft_3
#' @export
summarizor <- function(
  x,
  by = character(),
  overall_label = NULL,
  num_stats = c("mean_sd", "median_iqr", "range"),
  hide_null_na = TRUE,
  use_labels = TRUE
) {
  if (length(by) < 1) {
    by <- ".overall"
    x[[by]] <- by
    last_by <- character()
  } else {
    last_by <- by[length(by)]
  }

  list_lbls <- collect_labels(dataset = x, use_labels = use_labels)
  for (colname in names(list_lbls$values_labels)) {
    x[[colname]] <- factor(
      x = x[[colname]],
      levels = names(list_lbls$values_labels[[colname]]),
      labels = unname(list_lbls$values_labels[[colname]])
    )
  }

  x <- as.data.frame(x)

  if (length(last_by) > 0 && !is.null(overall_label)) {
    for (i in rev(seq_along(by))) {
      col_i <- by[i]
      xoverall <- x
      z <- xoverall[[col_i]]
      levels_ <- if (is.factor(z)) {
        c(levels(z), overall_label)
      } else {
        c(sort(unique(z)), overall_label)
      }
      z <- rep(overall_label, nrow(xoverall))
      xoverall[[col_i]] <- z
      x <- rbind(x, xoverall)
      x[[col_i]] <- factor(x[[col_i]], levels = levels_)
    }
  }

  cols <- setdiff(colnames(x), by)
  dtx <- as.data.table(x)
  datn <- dtx[, list(n = .N), by = by]
  setDF(datn)
  dat <- dtx[, list(data = list(.SD)), by = by, .SDcols = cols]
  dat$data <- lapply(dat$data, dataset_describe)

  dat$data <- lapply(
    dat$data,
    function(dat, drop_stats) {
      dat <- dat[!dat$stat %in% drop_stats, ]
      if (hide_null_na) {
        dat <- dat[!(dat$stat %in% "missing" & dat$cts < 1), ]
      }
      dat
    },
    drop_stats = setdiff(c("mean_sd", "median_iqr", "range"), num_stats)
  )

  for (i in seq_len(nrow(dat))) {
    w <- as.data.frame(dat$data[[i]])
    w[by] <- lapply(dat[i, .SD, .SDcols = by], function(z) {
      rep(z, length.out = nrow(w))
    })
    dat$data[[i]] <- w
  }
  dat <- rbindlist(dat$data)

  first_levels <- c("n", "mean_sd", "median_iqr", "range")
  last_levels <- c("missing")
  levs <- c(
    first_levels,
    setdiff(unique(dat$stat), c(first_levels, last_levels)),
    last_levels
  )
  labs <- levs
  dat$stat <- factor(dat$stat, levels = levs, labels = labs)

  dat$variable <- factor(dat$variable, levels = cols)
  setDF(dat)
  attr(dat, "use_labels") <- list(
    stat = c(
      stat = "",
      mean_sd = "Mean (SD)",
      median_iqr = "Median (IQR)",
      range = "Range",
      missing = "Missing"
    ),
    variable = c(variable = "", unlist(list_lbls$variables_labels))
  )
  class(dat) <- c("summarizor", class(dat))
  attr(dat, "n_by") <- datn
  attr(dat, "by") <- by
  dat
}

#' @export
#' @title Transform a 'summarizor' object into a flextable
#' @description `summarizor` object should be transformed into a flextable
#' with method [as_flextable()].
#' @param x result from [summarizor()]
#' @param ... arguments for [as_flextable.tabulator()]
#' @family as_flextable methods
#' @examples
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' z <- summarizor(CO2[-c(1, 4)],
#'   by = "Treatment",
#'   overall_label = "Overall"
#' )
#' ft_1 <- as_flextable(z, spread_first_col = TRUE)
#' ft_1 <- prepend_chunks(ft_1,
#'   i = ~ is.na(variable), j = 1,
#'   as_chunk("\t")
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
as_flextable.summarizor <- function(x, ...) {
  tab <- tabulator(
    x = x,
    rows = c("variable", "stat"),
    columns = attr(x, "by"),
    blah = as_paragraph(
      as_chunk(
        fmt_summarizor(
          stat = !!sym("stat"),
          num1 = !!sym("value1"),
          num2 = !!sym("value2"),
          cts = !!sym("cts"),
          pcts = !!sym("percent")
        )
      )
    )
  )
  ft <- as_flextable(tab, ...)
  ft <- labelizor(ft, labels = c(".overall" = "Statistic"), part = "header")

  ft
}

#' @importFrom stats sd IQR median
dataset_describe <- function(dataset) {
  w <- lapply(dataset, function(x) {
    if (is.factor(x) || is.character(x)) {
      z <- table(x, useNA = "always")
      names(z)[is.na(names(z))] <- "missing"
      z <- as.data.frame(z)
      colnames(z) <- c("stat", "cts")
      z$percent <- z$cts / sum(z$cts)
      z$data_type <- "discrete"
      z$value1 <- rep(NA_real_, nrow(z))
      z$value2 <- rep(NA_real_, nrow(z))
      z
    } else if (is.numeric(x)) {
      z <- data.frame(
        stat = c("mean_sd", "median_iqr", "range", "missing"),
        value1 = c(
          mean(x, na.rm = TRUE),
          as.double(median(x, na.rm = TRUE)),
          min(x, na.rm = TRUE),
          NA_real_
        ),
        value2 = c(
          sd(x, na.rm = TRUE),
          as.double(IQR(x, na.rm = TRUE)),
          max(x, na.rm = TRUE),
          NA_real_
        ),
        cts = NA_real_,
        percent = NA_real_
      )

      z$cts[z$stat %in% "missing"] <- sum(is.na(x))
      z$percent[z$stat %in% "missing"] <- sum(is.na(x)) / length(x)
      z$data_type <- "continuous"
      z
    }
  })
  w <- rbindlist(w, use.names = TRUE, fill = TRUE, idcol = "variable")
  setDF(w)
  w
}

#' @export
#' @title Format summarizor statistics as text
#' @description Format the output of [summarizor()] into display strings:
#' quantitative variables are shown as mean (sd), median (IQR), or range;
#' qualitative variables are shown as count (percentage).
#' @param stat a character column containing the name of statictics
#' @param num1 a numeric statistic to display such as a mean or a median
#' @param num2 a numeric statistic to display such as a standard
#' deviation or a median absolute deviation.
#' @param cts a count to display
#' @param pcts a percentage to display
#' @param ... unused arguments
#' @seealso [summarizor()], [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#' z <- summarizor(iris, by = "Species")
#'
#' tab_1 <- tabulator(
#'   x = z,
#'   rows = c("variable", "stat"),
#'   columns = "Species",
#'   blah = as_paragraph(
#'     as_chunk(
#'       fmt_summarizor(
#'         stat = stat,
#'         num1 = value1, num2 = value2,
#'         cts = cts, pcts = percent
#'       )
#'     )
#'   )
#' )
#'
#' ft_1 <- as_flextable(x = tab_1, separate_with = "variable")
#' ft_1 <- labelizor(
#'   x = ft_1, j = "stat",
#'   labels = c(
#'     mean_sd = "Moyenne (ecart-type)",
#'     median_iqr = "Mediane (IQR)",
#'     range = "Etendue",
#'     missing = "Valeurs manquantes"
#'   )
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_2stats <- function(stat, num1, num2, cts, pcts, ...) {
  z_num <- character(length = length(num1))
  z_cts <- character(length = length(num1))

  is_mean_sd <- !is.na(num1) & !is.na(num2) & stat %in% "mean_sd"
  is_median_iqr <- !is.na(num1) & !is.na(num2) & stat %in% "median_iqr"
  is_range <- !is.na(num1) & !is.na(num2) & stat %in% "range"
  is_num_1 <- !is.na(num1) & is.na(num2)
  is_cts_2 <- !is.na(cts) & !is.na(pcts)
  is_cts_1 <- !is.na(cts) & is.na(pcts)

  z_num[is_num_1] <- fmt_dbl(num1[is_num_1])

  z_num[is_mean_sd] <- paste0(
    fmt_dbl(num1[is_mean_sd]),
    " (",
    fmt_dbl(num2[is_mean_sd]),
    ")"
  )
  z_num[is_median_iqr] <- paste0(
    fmt_dbl(num1[is_median_iqr]),
    " (",
    fmt_dbl(num2[is_median_iqr]),
    ")"
  )
  z_num[is_range] <- paste0(
    fmt_dbl(num1[is_range]),
    " - ",
    fmt_dbl(num2[is_range])
  )

  z_cts[is_cts_2] <- paste0(
    fmt_int(cts[is_cts_2]),
    " (",
    fmt_pct(pcts[is_cts_2]),
    ")"
  )
  z_cts[is_cts_1] <- fmt_int(cts[is_cts_1])

  paste0(z_num, z_cts)
}
#' @export
#' @rdname fmt_2stats
fmt_summarizor <- fmt_2stats


#' @export
#' @title Format count and percentage as text
#' @description Format counts and percentages as `n (xx.x%)`.
#' If percentages are missing, only the count is shown.
#' @param n count values
#' @param pct percent values
#' @param digit number of digits for the percentages
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#'
#' df <- structure(
#'   list(
#'     cut = structure(
#'       .Data = 1:5, levels = c(
#'         "Fair", "Good", "Very Good", "Premium", "Ideal"
#'       ),
#'       class = c("ordered", "factor")
#'     ),
#'     n = c(1610L, 4906L, 12082L, 13791L, 21551L),
#'     pct = c(0.0299, 0.0909, 0.2239, 0.2557, 0.3995)
#'   ),
#'   row.names = c(NA, -5L),
#'   class = "data.frame"
#' )
#'
#' ft_1 <- flextable(df, col_keys = c("cut", "txt"))
#' ft_1 <- mk_par(
#'   x = ft_1, j = "txt",
#'   value = as_paragraph(fmt_n_percent(n, pct))
#' )
#' ft_1 <- align(ft_1, j = "txt", part = "all", align = "right")
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_n_percent <- function(n, pct, digit = 1) {
  z1 <- character(length(n))
  z2 <- character(length(n))
  z1[!is.na(n)] <- sprintf("%s", fmt_int(n[!is.na(n)]))
  z2[!is.na(pct)] <- paste0(" (", fmt_pct(pct[!is.na(pct)]), ")")
  paste0(z1, z2)
}

#' @export
#' @title Format count as '(N=XX)' for column headers
#' @description Format counts as `\n(N=XX)` for appending
#' sample sizes to column titles.
#' @param n count values
#' @param newline indicates to prefix the text with a new line
#' (sof return).
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#'
#' df <- data.frame(zz = 1)
#'
#' ft_1 <- flextable(df)
#' ft_1 <- append_chunks(
#'   x = ft_1, j = 1, part = "header",
#'   value = as_chunk(fmt_header_n(200))
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_header_n <- function(n, newline = TRUE) {
  z1 <- character(length(n))

  mask <- "\n(N=%s)"
  if (!newline) {
    mask <- "(N=%s)"
  }

  z1[!is.na(n)] <- sprintf(mask, fmt_int(n[!is.na(n)]))
  z1
}

#' @export
#' @title Format numbers as integers
#' @description Format numeric values as integers (no decimals).
#' @param x numeric values
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#'
#' df <- data.frame(zz = 1.23)
#'
#' ft_1 <- flextable(df)
#' ft_1 <- mk_par(
#'   x = ft_1, j = 1, part = "body",
#'   value = as_paragraph(as_chunk(zz, formatter = fmt_int))
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_int <- function(x) {
  big.mark <- get_flextable_defaults()$big.mark
  na_str <- get_flextable_defaults()$na_str
  nan_str <- get_flextable_defaults()$nan_str
  format_fun.integer(x, big.mark = big.mark, na_str = na_str, nan_str = nan_str)
}

#' @export
#' @title Format numbers as percentages
#' @description Format numeric values as percentages (e.g. `"45.0%"`).
#' @param x numeric values
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#'
#' df <- data.frame(zz = .45)
#'
#' ft_1 <- flextable(df)
#' ft_1 <- mk_par(
#'   x = ft_1, j = 1, part = "body",
#'   value = as_paragraph(as_chunk(zz, formatter = fmt_pct))
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_pct <- function(x) {
  big.mark <- get_flextable_defaults()$big.mark
  decimal.mark <- get_flextable_defaults()$decimal.mark
  pct_digits <- get_flextable_defaults()$pct_digits
  na_str <- get_flextable_defaults()$na_str
  nan_str <- get_flextable_defaults()$nan_str
  format_fun.pct(
    x,
    big.mark = big.mark,
    decimal.mark = decimal.mark,
    pct_digits = pct_digits,
    na_str = na_str,
    nan_str = nan_str
  )
}

#' @export
#' @title Format numbers as doubles
#' @description Format numeric values with decimal digits, using
#' the flextable default settings for separators and precision.
#' @param x numeric values
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#'
#' df <- data.frame(zz = .45)
#'
#' ft_1 <- flextable(df)
#' ft_1 <- mk_par(
#'   x = ft_1, j = 1, part = "body",
#'   value = as_paragraph(as_chunk(zz, formatter = fmt_dbl))
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_dbl <- function(x) {
  big.mark <- get_flextable_defaults()$big.mark
  decimal.mark <- get_flextable_defaults()$decimal.mark
  digits <- get_flextable_defaults()$digits
  na_str <- get_flextable_defaults()$na_str
  nan_str <- get_flextable_defaults()$nan_str

  format_fun.double(
    x,
    big.mark = big.mark,
    decimal.mark = decimal.mark,
    digits = digits,
    na_str = na_str,
    nan_str = nan_str
  )
}


#' @export
#' @title Format mean and standard deviation as text
#' @description Format means and standard deviations as `mean (sd)`.
#' @param avg,dev mean and sd values
#' @param digit1,digit2 number of digits to show when printing 'mean' and 'sd'.
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' library(flextable)
#'
#' df <- data.frame(avg = 1:3 * 3, sd = 1:3)
#'
#' ft_1 <- flextable(df, col_keys = "avg")
#' ft_1 <- mk_par(
#'   x = ft_1, j = 1, part = "body",
#'   value = as_paragraph(fmt_avg_dev(avg = avg, dev = sd))
#' )
#' ft_1 <- autofit(ft_1)
#' ft_1
fmt_avg_dev <- function(avg, dev, digit1 = 1, digit2 = 1) {
  z1 <- character(length(avg))
  z2 <- character(length(avg))
  z1[!is.na(avg)] <- sprintf(paste0("%.", digit1, "f"), avg[!is.na(avg)])
  z2[!is.na(dev)] <- sprintf(paste0(" (%.", digit2, "f)"), dev[!is.na(dev)])
  paste0(z1, z2)
}

#' @export
#' @title Format with significant figures after zeros
#' @description Rounds significant figures after zeros in numeric vectors.
#' The number of digits displayed after the leading zeros is
#' customizable using the `digits` parameter.
#' @param x numeric values
#' @param digits number of digits displayed after the leading zeros
#' @seealso [tabulator()], [mk_par()]
#' @family value formatters
#' @examples
#' x <- data.frame(
#'   x = c(0.00000004567, 2.000003456, 3, pi)
#' )
#' ft_1 <- flextable(x)
#' ft_1 <- align(x = ft_1, j = 1, align = "left")
#' mk_par(ft_1, value = as_paragraph(
#'   fmt_signif_after_zeros(x)))
fmt_signif_after_zeros <- function(x, digits = 3) {
  na_str = flextable_global$defaults$na_str
  nan_str = flextable_global$defaults$nan_str
  decimal.mark = flextable_global$defaults$decimal.mark

  wisna <- is.na(x)
  wisnan <- is.nan(x)

  # as character
  str_rounded <- formatC(x, format = "f", digits = 64)

  # decimal point index
  decimal_index <- regexpr("\\.", str_rounded)

  # decimal part extraction
  decimal_part <- substr(str_rounded, decimal_index + 1, nchar(str_rounded))

  # first non zero index
  regex_non0 <- regexpr("^[0]+", decimal_part)
  pos_non0 <- attr(regex_non0, "match.length")
  pos_non0[pos_non0 < 0] <- nchar(decimal_part)[pos_non0 < 0]

  dec_str <- substr(decimal_part, start = regex_non0, stop = pos_non0 + digits)
  dec_str[regex_non0 < 0] <- substr(
    decimal_part[regex_non0 < 0],
    start = 1,
    stop = digits
  )
  which_have_dec <- grepl("[^0]", dec_str)
  dec_str[which_have_dec] <- paste0(decimal.mark, dec_str[which_have_dec])
  dec_str[!which_have_dec] <- ""

  int_part <- substr(str_rounded, 1, decimal_index - 1)
  out <- paste0(int_part, dec_str)

  out[wisna] <- na_str
  out[wisnan] <- nan_str
  out
}

#' @title flextable: Functions for Tabular Reporting
#'
#' @description
#' The flextable package facilitates access to and manipulation of
#' tabular reporting elements from R.
#'
#' The documentation of functions can be opened with command `help(package = "flextable")`.
#'
#' `flextable()` function is producing flexible tables where each cell
#' can contain several chunks of text with their own set of formatting
#' properties (bold, font color, etc.). Function [mk_par()] lets customise
#' text of cells.
#'
#' Each cell holds a single paragraph composed of inline chunks
#' (see [as_paragraph()]). This means cell content is strictly
#' inline: bold, italic, links, images, equations, inline code, etc.
#' Block-level structures (multiple paragraphs, bullet lists,
#' headings or fenced code blocks) cannot be placed inside a cell.
#' Soft line breaks (`\n`) are however supported.
#'
#' The [as_flextable()] function is used to transform specific objects into
#' flextable objects. For example, you can transform a crosstab produced with
#' the 'tables' package into a flextable which can then be formatted,
#' annotated or augmented with footnotes.
#'
#' In order to reduce the homogenization efforts and the number of functions
#' to be called, it is recommended to define formatting properties such as
#' font, border color, number of decimals displayed which will then be applied
#' by default. See [set_flextable_defaults()] for more details.
#'
#' @section Table Structure:
#'
#' A flextable is composed of three distinct parts:
#'
#' - `header`: By default, contains one row with the column names from the data.frame
#' - `body`: Contains the actual data from the data.frame
#' - `footer`: Empty by default, but can contain content (commonly used for footnotes or #' summary rows)
#'
#'
#' ```
#'        HEADER         <- Column names, labels, spanning headers
#' --------------------
#'         BODY          <- Data rows
#' --------------------
#'        FOOTER         <- Summary rows, notes, footnotes (optional)
#' ```
#'
#' A basic flextable has:
#'
#' - in the part 'header': 1 row with column names
#' - in the part 'body': as many rows as there are in the input data.frame
#' - no footer
#'
#' Rows and columns can be added or removed to the basic flextable:
#'
#' - Add new rows in header with [add_header()], [add_header_row()],
#' [add_header_lines()] and [set_header_labels()].
#' - Add new rows in footer with [add_footer()], [add_footer_lines()],
#' [set_header_footer_df()] and [add_footer_row()].
#' - Add new rows in body with [add_body()] and [add_body_row()].
#' - Delete columns with [delete_columns()].
#' - Delete a part with [delete_part()].
#' - Use column names to separate a simple header row into multiple nested
#' rows with [separate_header()].
#'
#' @section Selectors in flextable:
#' Selectors are a core feature of flextable that allow you to specify which
#' parts (`part`), rows (`i`) and columns (`j`) should be affected by formatting,
#' styling, or content operations. See the corresponding manual:
#' <[`Selectors in flextable`][flextable_selectors]>.
#'
#' @seealso <https://davidgohel.github.io/flextable/>,
#' <https://ardata-fr.github.io/flextable-book/>, [flextable()]
#' @docType package
#' @aliases flextable-package
#' @name flextable-package
"_PACKAGE"

sanitize_latex_str <- function(str) {
  z <- gsub("[\\\\]", "\\\\textbackslash", str)
  z <- gsub("([&%$#_{} ]{1})", "\\\\\\1", z)
  z <- gsub("[~]", "\\\\\\textasciitilde ", z)
  z <- gsub("^", "\\\\\\textasciicircum ", z, fixed = TRUE)

  z
}

as_table_latexstyle_lr <- function(x) {
  left <- character(nrow(x))
  right <- character(nrow(x))

  fontcolor_left <- paste0(latex_fontcolor(x$color), "{")
  fontcolor_right <- rep("}", nrow(x))
  fontcolor_left[is.na(x$color)] <- ""
  fontcolor_right[is.na(x$color)] <- ""
  left <- paste0(fontcolor_left, left)
  right <- paste0(right, fontcolor_right)

  highlight_left <- paste0(latex_highlightcolor(x$shading.color), "{")
  highlight_right <- rep("}", length(x$shading.color))
  highlight_left[is.na(x$shading.color) | colalpha(x$shading.color) < 1] <- ""
  highlight_right[is.na(x$shading.color) | colalpha(x$shading.color) < 1] <- ""
  left <- paste0(left, highlight_left)
  right <- paste0(highlight_right, right)

  fontsize_left <- paste0(latex_fontsize(x$font.size, x$line_spacing), "{")
  fontsize_right <- rep("}", nrow(x))
  fontsize_left[is.na(x$font.size) | is.na(x$line_spacing)] <- ""
  fontsize_right[is.na(x$font.size) | is.na(x$line_spacing)] <- ""
  left <- paste0(left, fontsize_left)
  right <- paste0(fontsize_right, right)

  fonts_ok <- get_pdf_engine() %in% c("xelatex", "lualatex")
  if (fonts_ok && !flextable_global$defaults$fonts_ignore) {
    font_family_left <- sprintf("{\\fontspec{%s} ", x$font.family)
    font_family_right <- rep("}", nrow(x))
    font_family_left[is.na(x$font.family)] <- ""
    font_family_right[is.na(x$font.family)] <- ""
    left <- paste0(left, font_family_left)
    right <- paste0(font_family_right, right)
  }

  bold_left <- rep("", nrow(x))
  bold_right <- rep("", nrow(x))
  bold_left[x$bold %in% TRUE] <- "\\textbf{"
  bold_right[x$bold %in% TRUE] <- "}"
  left <- paste0(left, bold_left)
  right <- paste0(bold_right, right)

  italic_left <- rep("", nrow(x))
  italic_right <- rep("", nrow(x))
  italic_left[x$italic %in% TRUE] <- "\\textit{"
  italic_right[x$italic %in% TRUE] <- "}"
  left <- paste0(left, italic_left)
  right <- paste0(italic_right, right)

  underlined_left <- rep("", nrow(x))
  underlined_right <- rep("", nrow(x))
  underlined_left[x$underlined %in% TRUE] <- "\\underline{"
  underlined_right[x$underlined %in% TRUE] <- "}"
  underlined_left[x$strike %in% TRUE] <- "\\sout{"
  underlined_right[x$strike %in% TRUE] <- "}"
  left <- paste0(left, underlined_left)
  right <- paste0(underlined_right, right)

  valign_left <- character(nrow(x))
  valign_right <- character(nrow(x))
  valign_left[x$vertical.align %in% "subscript"] <- "\\textsubscript{"
  valign_right[x$vertical.align %in% "subscript"] <- "}"
  valign_left[x$vertical.align %in% "superscript"] <- "\\textsuperscript{"
  valign_right[x$vertical.align %in% "superscript"] <- "}"
  left <- paste0(left, valign_left)
  right <- paste0(valign_right, right)

  data.frame(
    left = left,
    right = right,
    classname = x$classname,
    stringsAsFactors = TRUE
  )
}

latex_fontsize <- function(x, line_spacing = 1, digits = 0) {
  size <- format_double(x, digits = 0)
  baselineskip <- format_double(x * line_spacing, digits = 0)
  z <- sprintf("\\fontsize{%s}{%s}\\selectfont", size, baselineskip)
  z[is.na(x)] <- ""
  z
}

latex_fontcolor <- function(x) {
  col <- colcode0(x)
  z <- sprintf("\\textcolor[HTML]{%s}", col)
  z
}

latex_cell_bgcolor <- function(x) {
  is_transparent <- colalpha(x) < 1
  z <- sprintf("\\cellcolor[HTML]{%s}", colcode0(x))
  z[is_transparent] <- ""
  z
}

latex_text_direction <- function(x, left = TRUE) {
  textdir <- rep("", length(x))
  if (left) {
    textdir[x %in% "tbrl"] <- "\\rotatebox[origin=c]{270}{"
    textdir[x %in% "btlr"] <- "\\rotatebox[origin=c]{90}{"
  } else {
    textdir[x %in% c("tbrl", "btlr")] <- "}"
  }
  textdir
}


latex_highlightcolor <- function(x) {
  col <- colcode0(x)
  z <- sprintf("\\colorbox[HTML]{%s}", col)
  z
}

#' @importFrom knitr fig_path
img_to_latex <- function(img_data, width, height) {
  new_files <- fig_path(suffix = ".png", number = seq_along(img_data))
  for (d in unique(dirname(new_files))) {
    dir.create(d, showWarnings = FALSE, recursive = TRUE)
  }

  str_raster <- mapply(
    function(img_raster, new_file, width, height) {
      if (inherits(img_raster, "raster")) {
        plot_in_png(
          code = {
            op <- par(mar = rep(0, 4))
            plot(img_raster, interpolate = FALSE, asp = NA)
            par(op)
          },
          width = width,
          height = height,
          res = 300,
          units = "in",
          path = new_file
        )
      } else if (is.character(img_raster)) {
        if (!file.exists(img_raster)) {
          stop(sprintf("File '%s' could not be found.", img_raster))
        }
        file.copy(from = img_raster, to = new_file, overwrite = TRUE)
      } else {
        stop("`image_data` can only be a raster or a filename.")
      }
      sprintf(
        "\\includegraphics[width=%sin, height=%sin]{%s}",
        width,
        height,
        new_file
      )
    },
    img_data,
    new_files,
    width,
    height,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  str_raster <- as.character(unlist(str_raster))
  str_raster
}

#' @importFrom officer table_layout table_width table_colwidths prop_table
#' @export
#' @title Set table layout and width properties
#' @description Set table layout and table width. Default to fixed
#' algorithm.
#'
#' If layout is fixed, column widths will be used to display the table;
#' `width` is ignored.
#'
#' If layout is autofit, column widths will not be used;
#' table width is used (as a percentage).
#' @note
#' PowerPoint output ignore 'autofit layout'.
#' @inheritParams args_x_only
#' @param layout 'autofit' or 'fixed' algorithm. Default to 'fixed'.
#' @param width The parameter has a different effect depending on the
#' output format. Users should consider it as a minimum width.
#' In HTML, it is the minimum width of the space that the
#' table should occupy. In Word, it is a preferred size and Word
#' may decide not to strictly stick to it. It has no effect on
#' PowerPoint and PDF output. Its default value is 0, as an effect, it
#' only use necessary width to display all content. It is not used by the
#' PDF output.
#' @param align alignment in document (only Word, HTML and PDF),
#' supported values are 'left', 'center' and 'right'.
#' @param opts_html html options as a list. Supported elements are:
#' - 'extra_css': extra css instructions to be integrated with the HTML
#' code of the table.
#' - 'scroll': NULL or a list if you want to add a scroll-box.
#'     - Use an empty list to add an horizontal scroll.  The with
#'     is fixed, corresponding to the container's width.
#'     - If the list has a value named `height` it will be used as
#'     height and the scroll will happen also vertically. The height
#'     will be in pixel if numeric, if a string it should be a valid css
#'     measure.
#'     - If the list has a value named `freeze_first_column` set to `TRUE`,
#'     the first column is set as a *sticky* column.
#'     - If the list has a value named `add_css` it will be used as extra
#'     css to add, .i.e: `border:1px solid red;`.
#' - 'extra_class': extra classes to add in the table tag
#' @param opts_word Word options as a list. Supported elements are:
#' - 'split':  Word option 'Allow row to break across pages' can be
#' activated when TRUE.
#' - 'keep_with_next': Word option 'keep rows together' is
#' activated when TRUE. It avoids page break within tables. This
#' is handy for small tables, i.e. less than a page height.
#' - 'repeat_headers': Word option 'Repeat as header row' is
#' activated and associated to header rows when TRUE.
#' @param opts_pdf PDF options as a list. Supported elements are:
#' - 'tabcolsep': space between the text and the left/right border of its containing
#' cell.
#' - 'arraystretch': height of each row relative to its default
#' height, the default value is 1.5.
#' - 'float': type of floating placement in the PDF document, one of:
#'     * 'none' (the default value), table is placed after the preceding
#' paragraph.
#'     * 'float', table can float to a place in the text where it fits best
#'     * 'wrap-r', wrap text around the table positioned to the right side of the text
#'     * 'wrap-l', wrap text around the table positioned to the left side of the text
#'     * 'wrap-i', wrap text around the table positioned inside edge-near the binding
#'     * 'wrap-o', wrap text around the table positioned outside edge-far from the binding
#' - 'fonts_ignore': if TRUE, pdf-engine 'pdflatex' can be used instead of
#' 'xelatex' or 'lualatex.' If pdflatex is used, fonts will be ignored because they are
#' not supported by pdflatex, whereas with the xelatex and lualatex engines they are.
#' - 'caption_repeat': a boolean that indicates if the caption should be
#' repeated along pages. Its default value is `TRUE`.
#' - 'footer_repeat': a boolean that indicates if the footer should be
#' repeated along pages. Its default value is `TRUE`.
#' - 'default_line_color': default line color, restored globally after
#' the flextable is produced.
#' @param word_title alternative text for Word table (used as title of the table)
#' @param word_description alternative text for Word table (used as description of the table)
#' @examples
#' library(flextable)
#' ft_1 <- flextable(head(cars))
#' ft_1 <- autofit(ft_1)
#' ft_2 <- set_table_properties(ft_1, width = .5, layout = "autofit")
#' ft_2
#' ft_3 <- set_table_properties(ft_1, width = 1, layout = "autofit")
#'
#' # add scroll for HTML ----
#' set.seed(2)
#' dat <- lapply(1:14, function(x) rnorm(n = 20))
#' dat <- setNames(dat, paste0("colname", 1:14))
#' dat <- as.data.frame(dat)
#'
#' ft_4 <- flextable(dat)
#' ft_4 <- colformat_double(ft_4)
#' ft_4 <- bg(ft_4, j = 1, bg = "#DDDDDD", part = "all")
#' ft_4 <- bg(ft_4, i = 1, bg = "#DDDDDD", part = "header")
#' ft_4 <- autofit(ft_4)
#' ft_4 <- set_table_properties(
#'   x = ft_4,
#'   opts_html = list(
#'     scroll = list(
#'       height = "500px",
#'       freeze_first_column = TRUE
#'     )
#'   )
#' )
#' ft_4
#' @seealso [flextable()], [as_flextable()], [autofit()],
#' [knit_print.flextable()]
#' @family functions for flextable size management
set_table_properties <- function(
  x,
  layout = "fixed",
  width = 0,
  align = NULL,
  opts_html = list(),
  opts_word = list(),
  opts_pdf = list(),
  word_title = NULL,
  word_description = NULL
) {
  stopifnot(
    `wrong layout value` = layout %in% c("fixed", "autofit"),
    `width is not numeric` = is.numeric(width),
    `width is > 1` = width <= 1
  )

  if (!is.null(word_title)) {
    stopifnot(
      is.character(word_title),
      length(word_title) == 1
    )
    stopifnot(
      is.character(word_description),
      length(word_description) == 1
    )
  }

  x$properties <- list(
    layout = layout,
    width = width,
    align = if (is.null(align)) get_flextable_defaults()$table_align else align,
    opts_html = do.call(opts_ft_html, opts_html),
    opts_word = do.call(opts_ft_word, opts_word),
    opts_pdf = do.call(opts_ft_pdf, opts_pdf),
    word_title = word_title,
    word_description = word_description
  )
  x
}

opts_ft_html <- function(
  extra_css = get_flextable_defaults()$extra_css,
  scroll = get_flextable_defaults()$scroll,
  extra_class = NULL,
  ...
) {
  if (
    !is.character(extra_css) || length(extra_css) != 1 || any(is.na(extra_css))
  ) {
    stop(
      sprintf("'%s' is expected to be a single %s.", "extra_css", "character"),
      call. = FALSE
    )
  }
  if (!is.null(scroll) && !is.list(scroll)) {
    stop(
      sprintf("'%s' is expected to be %s.", "scroll", "NULL or a list"),
      call. = FALSE
    )
  }

  z <- list(extra_css = extra_css, scroll = scroll, extra_class = extra_class)
  class(z) <- "opts_ft_html"
  z
}

opts_ft_word <- function(
  split = get_flextable_defaults()$split,
  keep_with_next = get_flextable_defaults()$keep_with_next,
  repeat_headers = TRUE
) {
  if (!is.logical(split) || length(split) != 1) {
    stop(
      sprintf("'%s' is expected to be a single %s.", "split", "logical"),
      call. = FALSE
    )
  }
  if (!is.logical(keep_with_next) || length(keep_with_next) != 1) {
    stop(
      sprintf(
        "'%s' is expected to be a single %s.",
        "keep_with_next",
        "logical"
      ),
      call. = FALSE
    )
  }
  if (!is.logical(repeat_headers) || length(repeat_headers) != 1) {
    stop(
      sprintf(
        "'%s' is expected to be a single %s.",
        "repeat_headers",
        "logical"
      ),
      call. = FALSE
    )
  }

  z <- list(
    split = split,
    keep_with_next = keep_with_next,
    repeat_headers = repeat_headers
  )
  class(z) <- "opts_ft_word"
  z
}

opts_ft_pdf <- function(
  tabcolsep = get_flextable_defaults()$tabcolsep,
  arraystretch = get_flextable_defaults()$arraystretch,
  float = get_flextable_defaults()$float,
  fonts_ignore = get_flextable_defaults()$fonts_ignore,
  caption_repeat = TRUE,
  footer_repeat = FALSE,
  default_line_color = "black"
) {
  if (!is.logical(fonts_ignore) || length(fonts_ignore) != 1) {
    stop(
      sprintf("'%s' is expected to be a single %s.", "fonts_ignore", "logical"),
      call. = FALSE
    )
  }
  if (
    !is.numeric(tabcolsep) || length(tabcolsep) != 1 || any(sign(tabcolsep) < 0)
  ) {
    stop(
      sprintf(
        "'%s' is expected to be a single %s.",
        "tabcolsep",
        "positive number"
      ),
      call. = FALSE
    )
  }
  if (
    !is.numeric(arraystretch) ||
      length(arraystretch) != 1 ||
      any(sign(arraystretch) < 0)
  ) {
    stop(
      sprintf(
        "'%s' is expected to be a single %s.",
        "arraystretch",
        "positive number"
      ),
      call. = FALSE
    )
  }
  if (
    !is.character(float) ||
      length(float) != 1 ||
      !all(
        float %in% c("none", "float", "wrap-r", "wrap-l", "wrap-i", "wrap-o")
      )
  ) {
    stop(
      sprintf(
        "'%s' is expected to be a single %s.",
        "float",
        "character (one of 'none', 'float', 'wrap-r', 'wrap-l', 'wrap-i', 'wrap-o')"
      ),
      call. = FALSE
    )
  }
  if (!is.logical(caption_repeat) || length(caption_repeat) != 1) {
    stop(
      sprintf(
        "'%s' is expected to be a single %s.",
        "caption_repeat",
        "logical"
      ),
      call. = FALSE
    )
  }

  z <- list(
    tabcolsep = tabcolsep,
    arraystretch = arraystretch,
    float = float,
    default_line_color = default_line_color,
    caption_repeat = caption_repeat,
    footer_repeat = footer_repeat,
    fonts_ignore = fonts_ignore
  )
  class(z) <- "opts_ft_pdf"
  z
}

init_flextable_defaults()

ft <- flextable(head(iris, 2))

# --- set_table_properties validation ---

test_that("set_table_properties rejects invalid layout", {
  expect_error(set_table_properties(ft, layout = "bogus"), "layout")
})

test_that("set_table_properties rejects invalid width", {
  expect_error(set_table_properties(ft, width = "a"), "width")
  expect_error(set_table_properties(ft, width = 1.5), "width")
})

test_that("set_table_properties validates word_title and word_description", {
  expect_error(
    set_table_properties(ft, word_title = 123),
    "is.character"
  )
  expect_error(
    set_table_properties(ft, word_title = "Title"),
    "is.character"
  )
  expect_no_error(
    set_table_properties(ft, word_title = "Title", word_description = "Desc")
  )
})

# --- opts_ft_html validation ---

test_that("opts_ft_html rejects invalid extra_css", {
  expect_error(
    set_table_properties(ft, opts_html = list(extra_css = 123)),
    "extra_css"
  )
  expect_error(
    set_table_properties(ft, opts_html = list(extra_css = c("a", "b"))),
    "extra_css"
  )
})

test_that("opts_ft_html rejects invalid scroll", {
  expect_error(
    set_table_properties(ft, opts_html = list(scroll = "yes")),
    "scroll"
  )
  expect_no_error(
    set_table_properties(ft, opts_html = list(scroll = NULL))
  )
  expect_no_error(
    set_table_properties(ft, opts_html = list(scroll = list(height = "300px")))
  )
})

# --- opts_ft_word validation ---

test_that("opts_ft_word rejects invalid split", {
  expect_error(
    set_table_properties(ft, opts_word = list(split = "yes")),
    "split"
  )
  expect_error(
    set_table_properties(ft, opts_word = list(split = c(TRUE, FALSE))),
    "split"
  )
})

test_that("opts_ft_word rejects invalid keep_with_next", {
  expect_error(
    set_table_properties(ft, opts_word = list(keep_with_next = 1)),
    "keep_with_next"
  )
})

test_that("opts_ft_word rejects invalid repeat_headers", {
  expect_error(
    set_table_properties(ft, opts_word = list(repeat_headers = "yes")),
    "repeat_headers"
  )
})

test_that("opts_ft_word accepts valid values", {
  ft2 <- set_table_properties(
    ft,
    opts_word = list(
      split = TRUE,
      keep_with_next = FALSE,
      repeat_headers = TRUE
    )
  )
  expect_true(ft2$properties$opts_word$split)
  expect_false(ft2$properties$opts_word$keep_with_next)
  expect_true(ft2$properties$opts_word$repeat_headers)
})

# --- opts_ft_pdf validation ---

test_that("opts_ft_pdf rejects invalid fonts_ignore", {
  expect_error(
    set_table_properties(ft, opts_pdf = list(fonts_ignore = "yes")),
    "fonts_ignore"
  )
  expect_error(
    set_table_properties(ft, opts_pdf = list(fonts_ignore = c(TRUE, FALSE))),
    "fonts_ignore"
  )
})

test_that("opts_ft_pdf rejects invalid tabcolsep", {
  expect_error(
    set_table_properties(ft, opts_pdf = list(tabcolsep = "big")),
    "tabcolsep"
  )
  expect_error(
    set_table_properties(ft, opts_pdf = list(tabcolsep = -1)),
    "tabcolsep"
  )
})

test_that("opts_ft_pdf rejects invalid arraystretch", {
  expect_error(
    set_table_properties(ft, opts_pdf = list(arraystretch = "big")),
    "arraystretch"
  )
  expect_error(
    set_table_properties(ft, opts_pdf = list(arraystretch = -0.5)),
    "arraystretch"
  )
})

test_that("opts_ft_pdf rejects invalid float", {
  expect_error(
    set_table_properties(ft, opts_pdf = list(float = "top")),
    "float"
  )
  expect_error(
    set_table_properties(ft, opts_pdf = list(float = 1)),
    "float"
  )
})

test_that("opts_ft_pdf rejects invalid caption_repeat", {
  expect_error(
    set_table_properties(ft, opts_pdf = list(caption_repeat = "yes")),
    "logical"
  )
})

test_that("opts_ft_pdf accepts all valid float values", {
  for (fval in c("none", "float", "wrap-r", "wrap-l", "wrap-i", "wrap-o")) {
    ft2 <- set_table_properties(ft, opts_pdf = list(float = fval))
    expect_equal(ft2$properties$opts_pdf$float, fval)
  }
})

test_that("opts_ft_pdf accepts valid combinations", {
  ft2 <- set_table_properties(
    ft,
    opts_pdf = list(
      tabcolsep = 4,
      arraystretch = 2,
      float = "float",
      fonts_ignore = TRUE,
      caption_repeat = FALSE,
      footer_repeat = TRUE
    )
  )
  expect_equal(ft2$properties$opts_pdf$tabcolsep, 4)
  expect_equal(ft2$properties$opts_pdf$arraystretch, 2)
  expect_equal(ft2$properties$opts_pdf$float, "float")
  expect_true(ft2$properties$opts_pdf$fonts_ignore)
  expect_false(ft2$properties$opts_pdf$caption_repeat)
  expect_true(ft2$properties$opts_pdf$footer_repeat)
})

# --- tab_settings ---

test_that("tab_settings works", {
  z <- data.frame(
    Statistic = c("Median (Q1 ; Q3)", "Min ; Max"),
    Value = c(
      "\t999.99\t(99.9 ; 99.9)",
      "\t9.99\t(9999.9 ; 99.9)"
    )
  )

  ts <- fp_tabs(
    fp_tab(pos = 0.4, style = "decimal"),
    fp_tab(pos = 1.4, style = "decimal")
  )

  ft <- flextable(z)
  ft <- tab_settings(ft, i = 1, j = 2, value = ts)
  ft <- width(ft, width = c(1.5, 2))

  docx_file <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- body_add_flextable(doc, value = ft)
  doc <- print(doc, target = docx_file)

  main_folder <- file.path(getwd(), "docx_folder")
  unpack_folder(file = docx_file, folder = main_folder)

  doc_file <- file.path(main_folder, "/word/document.xml")
  doc <- read_xml(doc_file)

  tabnode <- xml_find_first(
    doc,
    "w:body/w:tbl/w:tr[3]/w:tc[2]/w:p/w:pPr/w:tabs"
  )
  expect_true(inherits(tabnode, "xml_missing"))

  tabnode <- xml_find_first(
    doc,
    "w:body/w:tbl/w:tr[2]/w:tc[2]/w:p/w:pPr/w:tabs"
  )

  expect_false(inherits(tabnode, "xml_missing"))
  tab1 <- xml_child(tabnode, "w:tab[1]")
  tab2 <- xml_child(tabnode, "w:tab[2]")

  expect_equal(
    xml_attr(tab1, "val"),
    "decimal"
  )
  expect_equal(
    xml_attr(tab1, "pos"),
    "576"
  )
  expect_equal(
    xml_attr(tab2, "val"),
    "decimal"
  )
  expect_equal(
    xml_attr(tab2, "pos"),
    "2016"
  )

  unlink(main_folder, recursive = TRUE, force = TRUE)
})

# main -----

#' @export
#' @title Create pivot-style summary tables
#' @description It tabulates a data.frame representing an aggregation
#' which is then transformed as a flextable with
#' [as_flextable][as_flextable.tabulator]. The function
#' allows to define any display with the syntax of flextable in
#' a table whose layout is showing dimensions of the aggregation
#' across rows and columns.
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/ft-0-7-2-001-square.png" alt="tabulator illustration" style="width:100\%;">
#' }}
#' @note
#' This is very first version of the function; be aware it
#' can evolve or change.
#' @param x an aggregated data.frame
#' @param rows column names to use in rows dimensions
#' @param columns column names to use in columns dimensions
#' @param datasup_first additional data that will be merged with
#' table and placed after the columns presenting
#' the row dimensions.
#' @param datasup_last additional data that will be merged with
#' table and placed at the end of the table.
#' @param hidden_data additional data that will be merged with
#' table, the columns are not presented but can be used
#' with [compose()] or [mk_par()] function.
#' @param row_compose a list of call to [as_paragraph()] - these
#' calls will be applied to the row dimensions (the name is
#' used to target the displayed column).
#' @param ... named arguments calling function [as_paragraph()].
#' The names are used as labels and the values are evaluated
#' when the flextable is created.
#' @return an object of class `tabulator`.
#' @examples
#' \dontrun{
#' set_flextable_defaults(digits = 2, border.color = "gray")
#'
#' library(data.table)
#' # example 1 ----
#' if (require("stats")) {
#'   dat <- aggregate(breaks ~ wool + tension,
#'     data = warpbreaks, mean
#'   )
#'
#'   cft_1 <- tabulator(
#'     x = dat, rows = "wool",
#'     columns = "tension",
#'     `mean` = as_paragraph(as_chunk(breaks)),
#'     `(N)` = as_paragraph(as_chunk(length(breaks), formatter = fmt_int))
#'   )
#'
#'   ft_1 <- as_flextable(cft_1)
#'   ft_1
#' }
#'
#' # example 2 ----
#' if (require("ggplot2")) {
#'   multi_fun <- function(x) {
#'     list(mean = mean(x), sd = sd(x))
#'   }
#'
#'   dat <- as.data.table(ggplot2::diamonds)
#'   dat <- dat[cut %in% c("Fair", "Good", "Very Good")]
#'
#'   dat <- dat[, unlist(lapply(.SD, multi_fun),
#'     recursive = FALSE
#'   ),
#'   .SDcols = c("z", "y"),
#'   by = c("cut", "color")
#'   ]
#'
#'   tab_2 <- tabulator(
#'     x = dat, rows = "color",
#'     columns = "cut",
#'     `z stats` = as_paragraph(as_chunk(fmt_avg_dev(z.mean, z.sd, digit2 = 2))),
#'     `y stats` = as_paragraph(as_chunk(fmt_avg_dev(y.mean, y.sd, digit2 = 2)))
#'   )
#'   ft_2 <- as_flextable(tab_2)
#'   ft_2 <- autofit(x = ft_2, add_w = .05)
#'   ft_2
#' }
#'
#' # example 3 ----
#' # data.table version
#' dat <- melt(as.data.table(iris),
#'   id.vars = "Species",
#'   variable.name = "name", value.name = "value"
#' )
#' dat <- dat[,
#'   list(
#'     avg = mean(value, na.rm = TRUE),
#'     sd = sd(value, na.rm = TRUE)
#'   ),
#'   by = c("Species", "name")
#' ]
#' # dplyr version
#' # library(dplyr)
#' # dat <- iris %>%
#' #   pivot_longer(cols = -c(Species)) %>%
#' #   group_by(Species, name) %>%
#' #   summarise(avg = mean(value, na.rm = TRUE),
#' #   sd = sd(value, na.rm = TRUE),
#' #   .groups = "drop")
#'
#' tab_3 <- tabulator(
#'   x = dat, rows = c("Species"),
#'   columns = "name",
#'   `mean (sd)` = as_paragraph(
#'     as_chunk(avg),
#'     " (", as_chunk(sd), ")"
#'   )
#' )
#' ft_3 <- as_flextable(tab_3)
#' ft_3
#'
#' init_flextable_defaults()
#' }
#' @importFrom rlang enquos enquo call_args
#' @importFrom data.table rleidv as.data.table
#' @seealso [as_flextable.tabulator()], [summarizor()],
#' [as_grouped_data()], [tabulator_colnames()]
tabulator <- function(
  x,
  rows,
  columns,
  datasup_first = NULL,
  datasup_last = NULL,
  hidden_data = NULL,
  row_compose = list(),
  ...
) {
  stopifnot(
    `rows can not be empty` = length(rows) > 0,
    `columns can not be empty` = length(columns) > 0
  )
  use_labels <- attr(x, "use_labels")
  n_by <- attr(x, "n_by")

  x <- as.data.frame(x)

  row_compose <- enquo(row_compose)
  row_compose <- call_args(row_compose)

  col_exprs <- enquos(...)
  data_colnames <- setdiff(colnames(x), c(rows, columns))
  col_expr_names <- names(col_exprs)

  hidden_columns <- map_hidden_columns(
    dat = x,
    columns = columns,
    rows = rows
  )

  x <- add_fake_columns(x, col_expr_names)
  supp_colnames <- setdiff(colnames(datasup_first), rows)
  supp_colnames_last <- setdiff(colnames(datasup_last), rows)
  visible_columns <- map_visible_columns(
    dat = x,
    columns = columns,
    rows = rows,
    supp_colnames = supp_colnames,
    supp_colnames_last = supp_colnames_last,
    value_names = col_expr_names
  )

  # check dimensions
  cts <- as.data.table(x)[, c("cts") := .N, by = c(columns, rows)]
  setDF(cts)
  cts <- cts[cts$cts > 1, ]
  if (nrow(cts) > 0) {
    all_dims <- paste0("`", c(columns, rows), "`", collapse = ", ")
    stop(sprintf(
      "number of rows is not unique for some combinations of rows and columns: %s.",
      all_dims
    ))
  }

  .formula <- paste(
    paste0("`", rows, "`", collapse = "+"),
    "~",
    paste0("`", columns, "`", collapse = "+")
  )
  value_vars <- c(data_colnames, col_expr_names)

  dat <- dcast(
    data = as.data.table(x),
    formula = .formula,
    value.var = value_vars,
    sep = "@"
  )
  setDF(dat)

  dat <- merge_additional_dataset(dat, datasup_first, rows = rows)
  dat <- merge_additional_dataset(dat, hidden_data, rows = rows)
  dat <- merge_additional_dataset(dat, datasup_last, rows = rows)
  setDF(dat)

  z <- list(
    data = dat,
    rows = rows,
    columns = columns,
    visible_columns = visible_columns,
    hidden_columns = hidden_columns,
    col_exprs = col_exprs,
    row_exprs = row_compose,
    use_labels = use_labels,
    n_by = n_by
  )

  class(z) <- "tabulator"

  z
}


#' @export
#' @title Transform a 'tabulator' object into a flextable
#' @description [tabulator()] object can be transformed as a flextable
#' with method [as_flextable()].
#' @param x result from [tabulator()]
#' @param separate_with columns used to sepatate the groups
#' with an horizontal line.
#' @param big_border,small_border big and small border properties defined
#' by a call to [fp_border_default()] or [officer::fp_border()].
#' @param rows_alignment,columns_alignment alignments to apply to
#' columns corresponding to `rows` and `columns`; see arguments
#' `rows` and `columns` in [tabulator()].
#' @param label_rows labels to use for the first column names, i.e.
#' the *row* column names. It must be a named vector, the values will
#' be matched based on the names.
#' @param spread_first_col if TRUE, first row is spread as a new line separator
#' instead of being a column. This helps to reduce the width and allows for
#' clear divisions.
#' @param expand_single if FALSE (the default), groups with only one
#' row will not be expanded with a title row. If TRUE,
#' single row groups and multi-row groups are all
#' restructured.
#' @param sep_w blank column separators'width to be used. If 0,
#' blank column separators will not be used.
#' @param unit unit of argument `sep_w`, one of "in", "cm", "mm".
#' @param ... unused argument
#' @family as_flextable methods
#' @seealso [summarizor()], [as_grouped_data()]
#' @examples
#' \dontrun{
#' library(flextable)
#'
#' set_flextable_defaults(digits = 2, border.color = "gray")
#'
#' if (require("stats")) {
#'   dat <- aggregate(breaks ~ wool + tension,
#'     data = warpbreaks, mean
#'   )
#'
#'   cft_1 <- tabulator(
#'     x = dat,
#'     rows = "wool",
#'     columns = "tension",
#'     `mean` = as_paragraph(as_chunk(breaks)),
#'     `(N)` = as_paragraph(
#'       as_chunk(length(breaks))
#'     )
#'   )
#'
#'   ft_1 <- as_flextable(cft_1, sep_w = .1)
#'   ft_1
#' }
#'
#' if (require("stats")) {
#'   set_flextable_defaults(
#'     padding = 1, font.size = 9,
#'     border.color = "orange"
#'   )
#'
#'   ft_2 <- as_flextable(cft_1, sep_w = 0)
#'   ft_2
#' }
#'
#' if (require("stats")) {
#'   set_flextable_defaults(
#'     padding = 6, font.size = 11,
#'     border.color = "white",
#'     font.color = "white",
#'     background.color = "#333333"
#'   )
#'
#'   ft_3 <- as_flextable(
#'     x = cft_1, sep_w = 0,
#'     rows_alignment = "center",
#'     columns_alignment = "right"
#'   )
#'   ft_3
#' }
#'
#' init_flextable_defaults()
#' }
as_flextable.tabulator <- function(
  x,
  separate_with = character(0),
  big_border = fp_border_default(width = 1.5),
  small_border = fp_border_default(width = .75),
  rows_alignment = "left",
  columns_alignment = "center",
  label_rows = x$rows,
  spread_first_col = FALSE,
  expand_single = FALSE,
  sep_w = .05,
  unit = "in",
  ...
) {
  # get necessary element
  dat <- x$data

  rows <- x$rows
  columns <- x$columns

  if (spread_first_col) {
    dat <- as_grouped_data(dat, groups = rows[1], expand_single = expand_single)
  }

  visible_columns <- x$visible_columns
  hidden_columns <- x$hidden_columns

  col_exprs <- x$col_exprs

  if (sep_w < 0.001) {
    visible_columns <- visible_columns[
      !visible_columns[[".tab_columns"]] %in% "dummy",
    ]
  }

  visible_columns_keys <- visible_columns[
    visible_columns$.type. %in%
      "columns" &
      !visible_columns[[".tab_columns"]] %in% "dummy",
    "col_keys"
  ]
  blank_columns <- visible_columns[
    visible_columns[[".tab_columns"]] %in% "dummy",
    "col_keys"
  ]

  # create border_h_major from separate_with
  stopifnot(
    `separate_with is not part of rows` = length(separate_with) < 1 ||
      all(separate_with %in% rows)
  )

  border_h_major <- integer()

  if (length(separate_with) > 0) {
    rle <- rleidv(dat[separate_with])
    border_h_major <- which(rle != c(-1, rle[-length(rle)])) - 1
    border_h_major <- setdiff(border_h_major, 0)
  }
  if (length(separate_with) > 0 && spread_first_col) {
    border_h_major <- setdiff(border_h_major, which(!is.na(dat[[rows[1]]])))
  }

  if (spread_first_col) {
    visible_columns <- visible_columns[-1, ]
  }

  # for later iteration, a list of visible columns
  # to use when filling the table
  visible_columns_mapping <- visible_columns[
    visible_columns$.type. %in%
      "columns" &
      !visible_columns[[".tab_columns"]] %in% "dummy",
  ]
  visible_columns_mapping$.type. <- NULL
  visible_columns_mapping <- split(
    visible_columns_mapping,
    visible_columns_mapping[columns],
    drop = TRUE
  )

  # for later iteration, a list of hidden columns that can
  # be used by expressions defined by the user
  hidden_columns_mapping <- split(
    hidden_columns,
    hidden_columns[columns],
    drop = TRUE
  )

  ft <- flextable(dat, col_keys = visible_columns$col_keys)
  ft <- border_remove(ft)

  labels_tab <- visible_columns
  labels_tab <- labels_tab[!labels_tab$.tab_columns %in% "dummy", ]
  if (length(col_exprs) > 1) {
    labels <- labels_tab$.tab_columns
    names(labels) <- labels_tab$col_keys
    ft <- set_header_labels(x = ft, values = labels)
  } else {
    labels <- labels_tab[[columns[length(columns)]]]
    names(labels) <- labels_tab$col_keys
    ft <- set_header_labels(x = ft, values = labels)
    columns <- columns[-length(columns)]
  }

  for (j in names(visible_columns_mapping)) {
    visible_columns_mapping_j <- visible_columns_mapping[[j]]
    replication_info <- hidden_columns_mapping[[j]]
    ft$body$dataset[as.character(
      replication_info$.user_columns
    )] <- ft$body$dataset[replication_info$col_keys]

    for (i in seq_len(nrow(visible_columns_mapping_j))) {
      colname <- visible_columns_mapping_j[i, "col_keys"]
      exp_name <- visible_columns_mapping_j[i, ".tab_columns"]
      ft <- mk_par(ft, j = colname, value = !!col_exprs[[exp_name]])
    }
    ft$body$dataset[as.character(replication_info$.user_columns)] <- rep(
      list(NULL),
      nrow(replication_info)
    )
  }

  row_spanner <- character(length = 0L)
  if (spread_first_col) {
    row_spanner <- rows[1]
    rows <- rows[-1]

    row_spanner_map <- x$use_labels[[row_spanner]]
    apply_row_spanner_map <- function(values) {
      if (is.null(row_spanner_map)) {
        return(values)
      }
      # coerce to character first: dat[[row_spanner]] is typically a
      # factor and assigning a replacement not in its levels would
      # silently turn the cell into NA.
      values <- as.character(values)
      idx <- match(values, names(row_spanner_map))
      values[!is.na(idx)] <- row_spanner_map[idx[!is.na(idx)]]
      values
    }

    # treatment of groups
    rleid_ <- do.call(rleid, dat[row_spanner])
    table_rleid <- table(rleid_[is.na(dat[[row_spanner]])])
    table_uid <- as.integer(names(table_rleid[table_rleid > 1])) - 1 # considered as title row for non single group

    # write title rows for non single groups
    sel <- rleid_ %in% table_uid
    row_spanner_labels <- apply_row_spanner_map(dat[[row_spanner]][sel])
    ft <- mk_par(ft, i = sel, value = as_paragraph(row_spanner_labels))
    ft <- merge_h(ft, i = sel)

    # write title rows for single groups
    sna <- !is.na(dat[[row_spanner]])
    sna <- c(sna[-length(sna)] == sna[-1], FALSE) & sna
    row_spanner_labels <- apply_row_spanner_map(dat[[row_spanner]][sna])
    ft <- mk_par(ft, i = sna, j = 1, value = as_paragraph(row_spanner_labels))
  }

  ft <- merge_v(ft, j = rows, part = "body")
  ft <- valign(ft, valign = "top", j = rows)

  for (j in names(x$row_exprs)) {
    if (!j %in% row_spanner) {
      ft <- mk_par(ft, i = !is.na(dat[[j]]), j = j, value = !!x$row_exprs[[j]])
    } else {
      ft <- mk_par(ft, i = !is.na(dat[[j]]), j = 1, value = !!x$row_exprs[[j]])
    }
  }

  for (column in rev(columns)) {
    rel_ <- rle(visible_columns[[column]])
    rel_$values[rel_$values %in% "dummy"] <- ""
    ft <- add_header_row(
      x = ft,
      values = rel_$values,
      colwidths = rel_$lengths,
      top = TRUE
    )
    ft <- hline(
      ft,
      i = 1,
      j = which(!visible_columns[[column]] %in% c("dummy", rows)),
      border = small_border,
      part = "header"
    )
    ft <- align(x = ft, i = 1, align = "center", part = "header")
  }

  rows_supp <- visible_columns$col_keys[
    visible_columns$.type. %in%
      c("rows_supp", "rows_supp_last") &
      !visible_columns$.tab_columns %in% "dummy"
  ]
  ft <- merge_v(
    x = ft,
    j = c(rows, rows_supp),
    part = "header"
  )

  ft <- valign(ft, valign = "bottom", j = c(rows, rows_supp), part = "header")
  ft <- valign(ft, valign = "top", part = "body")

  ft <- align(
    x = ft,
    j = visible_columns_keys,
    align = columns_alignment,
    part = "all"
  )
  ft <- align(
    x = ft,
    j = c(rows, rows_supp),
    align = rows_alignment,
    part = "all"
  )

  if (sep_w > 0) {
    ft <- padding(ft, j = blank_columns, padding = 0, part = "all")
  }

  # dummy as blank columns

  ft <- hline(ft, i = border_h_major, border = small_border)

  ft <- hline(
    ft,
    i = nrow_part(ft, part = "header"),
    j = setdiff(visible_columns$col_keys, blank_columns),
    border = big_border,
    part = "header"
  )
  ft <- hline_top(x = ft, border = big_border, part = "header")
  ft <- hline_bottom(x = ft, border = big_border, part = "body")

  if (sep_w > 0) {
    ft <- border(
      ft,
      j = blank_columns,
      border.top = fp_border_default(width = 0),
      border.bottom = fp_border_default(width = 0),
      part = "all"
    )
    ft <- bg(ft, j = blank_columns, bg = "transparent", part = "all")
    ft <- void(ft, j = blank_columns, part = "all")
    ft <- width(ft, j = blank_columns, width = sep_w, unit = unit)
  }

  if (!is.null(names(label_rows))) {
    j_labs <- names(label_rows)
    if (spread_first_col) {
      j_labs <- j_labs[!names(label_rows) %in% row_spanner]
      label_rows <- label_rows[!names(label_rows) %in% row_spanner]
    }
    ft <- mk_par(
      ft,
      i = 1,
      j = j_labs,
      value = as_paragraph(as.character(label_rows)),
      part = "header"
    )
  }

  if (spread_first_col) {
    ft <- align(
      x = ft,
      i = !is.na(dat[[row_spanner]]),
      align = columns_alignment
    )
  }

  if (!is.null(x$use_labels)) {
    for (labj in names(x$use_labels)) {
      if (labj %in% ft$col_keys) {
        ft <- labelizor(
          x = ft,
          j = labj,
          labels = x$use_labels[[labj]],
          part = "all"
        )
      }
    }
  }
  if (!is.null(x$n_by)) {
    sum_x <- visible_columns[
      visible_columns$.tab_columns %in%
        names(x$col_exprs)[1] &
        visible_columns$.type. %in% "columns",
      ,
      drop = FALSE
    ]
    header_row <- length(x$columns)

    for (k in seq_len(nrow(x$n_by))) {
      mask <- rep(TRUE, nrow(sum_x))
      for (col in x$columns) {
        mask <- mask &
          as.character(sum_x[[col]]) == as.character(x$n_by[[col]][k])
      }
      if (any(mask)) {
        ft <- append_chunks(
          x = ft,
          j = sum_x[mask, "col_keys"][1],
          i = header_row,
          part = "header",
          as_chunk(x$n_by$n[k], formatter = fmt_header_n)
        )
      }
    }
  }

  ft <- autofit(ft, part = "all", add_w = .2, add_h = .0, unit = "cm")
  ft <- fix_border_issues(ft, part = "all")
  ft
}

# util methods -----

#' @export
#' @describeIn tabulator call `summary()` to get
#' a data.frame describing mappings between variables
#' and their names in the flextable. This data.frame contains
#' a column named `col_keys` where are stored the names that
#' can be used for further selections.
#' @param object an object returned by function
#' `tabulator()`.
summary.tabulator <- function(object, ...) {
  hidden_columns <- object$hidden_columns
  names(hidden_columns)[names(hidden_columns) %in% ".user_columns"] <- "column"
  hidden_columns$.type. <- "hidden"

  visible_columns <- object$visible_columns
  names(visible_columns)[names(visible_columns) %in% ".tab_columns"] <- "column"
  dat <- rbind(visible_columns, hidden_columns)
  dat
}

#' @export
#' @importFrom rlang quo_text
#' @title Column keys of tabulator objects
#' @description The function provides a way to get column keys
#' associated with the flextable corresponding to a [tabulator()]
#' object. It helps in customizing or programing with `tabulator`.
#'
#' The function is using column names from the original
#' dataset, eventually filters and returns the names
#' corresponding to the selection.
#' @param x a [tabulator()] object
#' @param columns column names to look for
#' @param type the type of column to look for, it can be:
#'
#' * 'columns': visible columns, corresponding to names provided
#' in the '...' arguments of your call to 'tabulator()'.
#' * 'hidden': unvisible columns, corresponding to names of
#' the original dataset columns.
#' * 'rows': visible columns used as 'row' content
#' * 'rows_supp': visible columns used as 'rows_supp' content
#' * NULL: any type of column
#' @param ... any filter conditions that use variables
#' names, the same than the argument `columns` of function [tabulator()]
#' (`tabulator(columns = c("col1", "col2"))`).
#' @seealso [tabulator()], [as_flextable.tabulator()]
#' @examples
#' library(flextable)
#'
#' cancer_dat <- data.frame(
#'   count = c(
#'     9L, 5L, 1L, 2L, 2L, 1L, 9L, 3L, 1L, 10L, 2L, 1L, 1L, 2L, 0L, 3L,
#'     2L, 1L, 1L, 2L, 0L, 12L, 4L, 1L, 7L, 3L, 1L, 5L, 5L, 3L, 10L,
#'     4L, 1L, 4L, 2L, 0L, 3L, 1L, 0L, 4L, 4L, 2L, 42L, 28L, 19L, 26L,
#'     19L, 11L, 12L, 10L, 7L, 10L, 5L, 6L, 5L, 0L, 3L, 4L, 3L, 3L,
#'     1L, 2L, 3L
#'   ),
#'   risktime = c(
#'     157L, 77L, 21L, 139L, 68L, 17L, 126L, 63L, 14L, 102L, 55L,
#'     12L, 88L, 50L, 10L, 82L, 45L, 8L, 76L, 42L, 6L, 134L, 71L,
#'     22L, 110L, 63L, 18L, 96L, 58L, 14L, 86L, 42L, 10L, 66L,
#'     35L, 8L, 59L, 32L, 8L, 51L, 28L, 6L, 212L, 130L, 101L,
#'     136L, 72L, 63L, 90L, 42L, 43L, 64L, 21L, 32L, 47L, 14L,
#'     21L, 39L, 13L, 14L, 29L, 7L, 10L
#'   ),
#'   time = rep(as.character(1:7), 3),
#'   histology = rep(as.character(1:3), 21),
#'   stage = rep(as.character(1:3), each = 21)
#' )
#'
#' datasup_first <- data.frame(
#'   time = factor(1:7, levels = 1:7),
#'   zzz = runif(7)
#' )
#'
#' z <- tabulator(cancer_dat,
#'   rows = "time",
#'   columns = c("histology", "stage"),
#'   datasup_first = datasup_first,
#'   n = as_paragraph(as_chunk(count))
#' )
#'
#' j <- tabulator_colnames(
#'   x = z, type = "columns",
#'   columns = c("n"),
#'   stage %in% 1
#' )
#'
#' src <- tabulator_colnames(
#'   x = z, type = "hidden",
#'   columns = c("count"),
#'   stage %in% 1
#' )
#'
#' if (require("scales")) {
#'   colourer <- col_numeric(
#'     palette = c("wheat", "red"),
#'     domain = c(0, 45)
#'   )
#'   ft_1 <- as_flextable(z)
#'   ft_1 <- bg(
#'     ft_1,
#'     bg = colourer, part = "body",
#'     j = j, source = src
#'   )
#'   ft_1
#' }
tabulator_colnames <- function(x, columns, ..., type = NULL) {
  dat <- summary(x)

  if (!is.null(type)) {
    dat <- dat[dat$.type. %in% type, ]
  }

  exprs <- enquos(...)
  exprs_evals <- lapply(
    exprs,
    function(expr_filter, dat) {
      check_filter_expr(expr_filter, dat)
      eval_tidy({{ expr_filter }}, data = dat)
    },
    dat = dat
  )
  exprs_evals <- append(exprs_evals, list(dat$column %in% columns))
  keep <- Reduce(`&`, exprs_evals)

  dat["col_keys"][keep, ]
}


#' @export
print.tabulator <- function(x, ...) {
  cat("layout:\n")
  cat(
    "* row(s): ",
    paste0("`", x$rows, "`", collapse = ", "),
    "\n"
  )
  cat(
    "* column(s): ",
    paste0("`", x$columns, "`", collapse = ", "),
    "\n"
  )
  cat(
    "* content(s): ",
    paste0("`", names(x$col_exprs), "`", collapse = ", "),
    "\n"
  )
  visible_columns <- x$visible_columns
  columns_keys <- visible_columns[
    visible_columns$.type. %in%
      "columns" &
      !visible_columns[[".tab_columns"]] %in% "dummy",
    "col_keys"
  ]

  cat(
    "\ncol_keys: c(",
    paste0(shQuote(columns_keys, type = "cmd"), collapse = ", "),
    ")\n",
    sep = ""
  )

  print(as.data.table(x$data))
  invisible()
}


# utils -----
#' @importFrom rlang quo_text
check_filter_expr <- function(filter_expr, x) {
  filter_varnames <- all.vars(filter_expr)
  missing_varnames <- setdiff(filter_varnames, colnames(x))

  if (length(missing_varnames) > 0) {
    stop(
      sprintf(
        "`%s` is using unknown variable(s): %s",
        quo_text(filter_expr),
        paste0("`", missing_varnames, "`", collapse = ",")
      ),
      call. = FALSE
    )
  }
}

add_fake_columns <- function(x, fake_columns) {
  x[fake_columns] <- rep(list(character(nrow(x))), length(fake_columns))
  x
}

merge_additional_dataset <- function(a, b, rows) {
  if (!is.null(b)) {
    by <- intersect(rows, colnames(b))
    a$.keep_order_a <- seq_len(nrow(a))
    b$.keep_order_b <- seq_len(nrow(b))
    z <- merge(a, b, by = by, all.x = TRUE, all.y = FALSE)
    z <- z[order(z$.keep_order_a, z$.keep_order_b), ]
    z[c(".keep_order_a", ".keep_order_b")] <- list(NULL, NULL)
    a <- z
  }
  a
}

#' @importFrom data.table setorderv
map_visible_columns <- function(
  dat,
  columns,
  rows,
  value_names = character(0),
  supp_colnames = character(0),
  supp_colnames_last = character(0)
) {
  dat <- dat[c(columns, rows)]
  dat[value_names] <- lapply(
    value_names,
    function(x, n) character(n),
    n = nrow(dat)
  )
  dat <- as.data.table(dat)
  dat <- melt(
    as.data.table(dat),
    id.vars = c(rows, columns),
    measure.vars = value_names,
    variable.name = ".tab_columns"
  )
  setorderv(dat, cols = c(rows, columns))

  columns <- c(columns, ".tab_columns")
  ldims <- dat[, .SD, .SDcols = columns]
  ldims <- unique(ldims)
  setorderv(ldims, cols = columns[-length(columns)]) # important - order matters
  uid_cols <- columns

  last_m1_column <- length(columns) - 1
  ldims <- split(ldims, rleid(ldims[[last_m1_column]]))
  ldims <- lapply(
    ldims,
    function(x, j) {
      x1 <- x[1, ]
      x1[, j] <- "dummy"
      rbind(x, x1)
    },
    j = seq(last_m1_column, length(columns), by = 1L)
  )
  ldims <- do.call(rbind, ldims)

  sel_columns <- columns[seq_len(length(columns) - 2)]

  for (j in rev(seq_along(sel_columns))) {
    # Check if any group has more than 1 row (only then do we need separators)
    group_sizes <- table(rleid(ldims[[j]]))
    if (all(group_sizes <= 1)) {
      next
    }

    ldims <- split(ldims, rleid(ldims[[j]]))
    ldims <- lapply(
      ldims,
      function(x, j) {
        x[nrow(x), j] <- "dummy"
        x
      },
      j = j
    )
    ldims <- do.call(rbind, ldims)
  }
  ldims <- ldims[-nrow(ldims), ]

  rdims <- lapply(rows, function(x, n) rep(x, n), n = ncol(ldims))
  rdims <- do.call(rbind, rdims)
  x1 <- rdims[1, ]
  x1[] <- "dummy"
  rdims <- rbind(rdims, x1)
  colnames(rdims) <- names(ldims)
  rdims <- as.data.frame(rdims, row.names = FALSE)

  rdims_supp <- NULL
  if (length(supp_colnames) > 0) {
    rdims_supp <- lapply(
      supp_colnames,
      function(x, n) rep(x, n),
      n = ncol(ldims)
    )
    rdims_supp <- do.call(rbind, rdims_supp)
    x1 <- rdims_supp[1, ]
    x1[] <- "dummy"
    rdims_supp <- rbind(rdims_supp, x1)
    colnames(rdims_supp) <- names(ldims)
    rdims_supp <- as.data.frame(rdims_supp, row.names = FALSE)
    rdims_supp$.type. <- "rows_supp"
  }

  rdims_supp_last <- NULL
  if (length(supp_colnames_last) > 0) {
    rdims_supp_last <- lapply(
      supp_colnames_last,
      function(x, n) rep(x, n),
      n = ncol(ldims)
    )
    rdims_supp_last <- do.call(rbind, rdims_supp_last)
    x1 <- rdims_supp_last[1, ]
    x1[] <- "dummy"
    rdims_supp_last <- rbind(x1, rdims_supp_last)
    colnames(rdims_supp_last) <- names(ldims)
    rdims_supp_last <- as.data.frame(rdims_supp_last, row.names = FALSE)
    rdims_supp_last$.type. <- "rows_supp_last"
  }

  ldims$.type. <- "columns"
  rdims$.type. <- "rows"
  last_column <- columns[length(columns)]
  dims <- rbind(rdims, rdims_supp, ldims, rdims_supp_last)

  is_dummy <- dims[[last_column]] %in% "dummy"
  dims$col_keys <- do.call(
    paste,
    append(as.list(dims[uid_cols]), list(sep = "@"))
  )
  if (length(value_names) > 1) {
    dims$col_keys <- paste0(dims$.stat_name, "@", dims$col_keys)
  }
  dims$col_keys[is_dummy] <- paste0("dummy", seq_len(sum(is_dummy)))

  dims$col_keys[dims$.type. %in% "rows" & !is_dummy] <- dims[[last_column]][
    dims$.type. %in% "rows" & !is_dummy
  ]
  dims$col_keys[dims$.type. %in% "rows_supp" & !is_dummy] <- dims[[
    last_column
  ]][dims$.type. %in% "rows_supp" & !is_dummy]
  dims$col_keys[dims$.type. %in% "rows_supp_last" & !is_dummy] <- dims[[
    last_column
  ]][dims$.type. %in% "rows_supp_last" & !is_dummy]
  setDF(dims)
  dims
}

map_hidden_columns <- function(dat, columns, rows) {
  user_columns <- setdiff(colnames(dat), c(columns, rows))
  dat <- as.data.table(dat)
  dat[,
    c(user_columns) := lapply(.SD, function(x) character(length(x))),
    .SDcols = user_columns
  ]
  dat <- melt(
    as.data.table(dat),
    id.vars = c(columns, rows),
    measure.vars = user_columns,
    variable.name = ".user_columns"
  )
  columns <- c(columns, ".user_columns")
  dims <- dat[, .SD, .SDcols = columns]
  setDF(dat)
  dims <- unique(dims)
  dims$col_keys <- do.call(
    paste,
    append(
      as.list(dims[, .SD, .SDcols = columns[-length(columns)]]),
      list(sep = "@")
    )
  )
  dims$col_keys <- paste0(dims$.user_columns, "@", dims$col_keys)
  setDF(dims)
  dims
}

# grid_data ---------------------------------------------------------------

#' @importFrom data.table data.table rbindlist rleid .SD .N first fcase fifelse
get_grid_data <- function(x, autowidths, wrapping) {
  parts <- c("header", "body", "footer")
  names(parts) <- parts
  grid_data <- rbindlist(
    lapply(parts, function(part) {
      nr <- nrow_part(x, part)
      if (nr > 0) {
        data.frame(
          .col_id = rep(x$col_keys, each = nr),
          .row_id = rep(seq_len(nr), length(x$col_keys)),
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      }
    }),
    use.names = TRUE,
    idcol = ".part"
  )

  grid_data$.part <- factor(
    grid_data$.part,
    levels = intersect(parts, grid_data$.part)
  )
  grid_data$.col_id <- factor(grid_data$.col_id, levels = x$col_keys)
  keycols <- c(".part", ".row_id", ".col_id")
  setorderv(grid_data, cols = keycols)
  grid_data[,
    c("row_index", "col_index") := list(
      rleid(.SD$.part, .SD$.row_id),
      as.integer(.SD$.col_id)
    )
  ]
  setorderv(grid_data, cols = c("row_index", "col_index"))

  # add column widths
  col_widths <- fortify_width(x)
  colnames(col_widths)[colnames(col_widths) == "width"] <- "col_width"
  grid_data <- merge(grid_data, col_widths, by = c(".col_id"))

  # add row heights
  row_heights <- fortify_height(x)
  colnames(row_heights)[colnames(row_heights) == "height"] <- "row_height"
  grid_data <- merge(grid_data, row_heights, by = c(".part", ".row_id"))

  # add hrule
  row_hrules <- fortify_hrule(x)
  grid_data <- merge(grid_data, row_hrules, by = c(".part", ".row_id"))

  # calculate minimum and maximum cell height according to hrule
  grid_data[,
    "row_min_height" := fcase(
      .SD$hrule %in% "exact"   , .SD$row_height ,
      .SD$hrule %in% "atleast" , .SD$row_height ,
      default = NA_real_
    )
  ]
  grid_data[,
    "row_max_height" := fcase(
      .SD$hrule %in% "exact" , .SD$row_height ,
      default = NA_real_
    )
  ]

  # add spans
  spans <- fortify_span(x)
  grid_data <- merge(grid_data, spans, by = keycols)
  # create span groups
  setorderv(grid_data, cols = c("row_index", "col_index"))
  grid_data[, "rowspan_group" := calc_grid_span_group(.SD$rowspan)]
  setorderv(grid_data, cols = c("col_index", "row_index"))
  grid_data[, "colspan_group" := calc_grid_span_group(.SD$colspan)]

  # set cell width and height
  grid_data[,
    "cell_width" := safe_stat(.SD$col_width, FUN = sum, NA_value = NA_real_),
    by = "rowspan_group"
  ]
  grid_data[,
    "cell_height" := safe_stat(.SD$row_height, FUN = sum, NA_value = NA_real_),
    by = "colspan_group"
  ]
  grid_data[,
    "cell_min_height" := safe_stat(
      .SD$row_min_height,
      FUN = sum,
      NA_value = NA_real_
    ),
    by = "colspan_group"
  ]
  grid_data[,
    "cell_max_height" := safe_stat_ext(
      .SD$row_max_height,
      FUN = sum,
      LENGTH = max(.SD$colspan),
      NA_value = NA_real_
    ),
    by = "colspan_group"
  ]
  grid_data[,
    "cell_height" := safe_stat(
      safe_stat(
        .SD$cell_height,
        .SD$cell_min_height,
        FUN = max,
        NA_value = NA_real_
      ),
      .SD$cell_max_height,
      FUN = min,
      NA_value = NA_real_
    ),
    by = c("col_index", "row_index")
  ]

  # keep only active cells
  grid_data <- grid_data[
    grid_data$rowspan > 0 & grid_data$colspan > 0,
    ,
    drop = FALSE
  ]

  # generate argument list for creating the cell viewports
  setorderv(grid_data, cols = c("row_index", "col_index"))
  grid_data[,
    "cell_vp" := mapply(
      function(row, colspan, column, rowspan) {
        list(
          layout.pos.row = seq(row, length.out = colspan),
          layout.pos.col = seq(column, length.out = rowspan),
          clip = "on"
        )
      },
      row = .SD$row_index,
      colspan = .SD$colspan,
      column = .SD$col_index,
      rowspan = .SD$rowspan,
      SIMPLIFY = FALSE
    )
  ]

  # drop unneeded columns
  grid_data[,
    c(
      "col_width",
      "row_height",
      "row_min_height",
      "row_max_height",
      "hrule",
      "rowspan_group",
      "colspan_group"
    ) := NULL
  ]

  grid_data
}


grid_data_add_cell_info <- function(grid_data, x) {
  keycols <- c(".part", ".row_id", ".col_id")
  spans <- fortify_span(x) # redondant avec get_grid_data, pas bien
  cell_data <- merge(
    as.data.table(information_data_cell(x)),
    spans,
    by = keycols
  )

  fortify_borders_data <- fortify_latex_borders(cell_data)
  # apply a correction to borders for vert. merged cells
  last_bottom_borders_data <- fortify_borders_data[,
    list(
      border.width.bottom = last(.SD$border.width.bottom),
      border.color.bottom = last(.SD$border.color.bottom)
    ),
    by = c(".part", ".col_id", "vspan_id")
  ]
  fortify_borders_data$border.width.bottom <- NULL
  fortify_borders_data$border.color.bottom <- NULL
  fortify_borders_data <- merge(
    fortify_borders_data,
    last_bottom_borders_data,
    by = c(".part", ".col_id", "vspan_id")
  )

  fortify_borders_data <- fortify_borders_data[,
    .SD,
    .SDcols = c(
      ".part",
      ".row_id",
      ".col_id",
      "border.color.top",
      "border.color.bottom",
      "border.width.left",
      "border.color.left",
      "border.width.right",
      "border.color.right",
      "border.width.top",
      "border.width.bottom"
    )
  ]
  # apply a correction to overlapping vert. borders
  setorderv(fortify_borders_data, cols = c(".part", ".row_id", ".col_id"))
  fortify_borders_data[
    fortify_borders_data$border.width.right ==
      shift(fortify_borders_data$border.width.left, type = "lead", fill = -1) &
      fortify_borders_data$.col_id != tail(x$col_keys, 1),
    c("border.width.right") := 0,
    by = c(".part", ".row_id")
  ]

  cell_data <- cell_data[,
    .SD,
    .SDcols = setdiff(
      colnames(cell_data),
      c(
        "border.color.top",
        "border.color.bottom",
        "border.width.left",
        "border.color.left",
        "border.width.right",
        "border.color.right",
        "border.width.top",
        "border.width.bottom"
      )
    )
  ]

  cell_data <- merge(cell_data, fortify_borders_data, by = keycols)

  # merge with grid_data to keep only active cells
  cell_data <- merge(
    grid_data[, keycols, with = FALSE],
    cell_data,
    by = keycols
  )

  # generate argument list for creating the cell background rect
  cell_data[,
    "background" := mapply(
      function(fill) {
        gp <- gpar_background(fill)
        if (length(gp) > 0) {
          list(gp = gp)
        }
      },
      fill = .SD$background.color,
      SIMPLIFY = FALSE
    )
  ]

  # generate argument list for creating the border segments
  cell_data[,
    "borders" := mapply(
      function(t, l, b, r) {
        list(
          if (length(t) > 0) {
            list(x0 = 0, x1 = 1, y0 = 1, y1 = 1, gp = t)
          },
          if (length(l) > 0) {
            list(x0 = 0, x1 = 0, y0 = 0, y1 = 1, gp = l)
          },
          if (length(b) > 0) {
            list(x0 = 0, x1 = 1, y0 = 0, y1 = 0, gp = b)
          },
          if (length(r) > 0) {
            list(x0 = 1, x1 = 1, y0 = 0, y1 = 1, gp = r)
          }
        )
      },
      t = mapply(
        gpar_border,
        width = cell_data$border.width.top,
        color = cell_data$border.color.top,
        style = cell_data$border.style.top,
        SIMPLIFY = FALSE
      ),
      l = mapply(
        gpar_border,
        width = cell_data$border.width.left,
        color = cell_data$border.color.left,
        style = cell_data$border.style.left,
        SIMPLIFY = FALSE
      ),
      b = mapply(
        gpar_border,
        width = cell_data$border.width.bottom,
        color = cell_data$border.color.bottom,
        style = cell_data$border.style.bottom,
        SIMPLIFY = FALSE
      ),
      r = mapply(
        gpar_border,
        width = cell_data$border.width.right,
        color = cell_data$border.color.right,
        style = cell_data$border.style.right,
        SIMPLIFY = FALSE
      ),
      SIMPLIFY = FALSE
    )
  ]

  # set rotation angle
  cell_data[,
    "angle" := fcase(
      .SD$text.direction == "tbrl" , 270 ,
      .SD$text.direction == "btlr" ,  90 ,
      default = 0
    )
  ]

  # merge needed columns
  cell_data <- cell_data[,
    c(
      keycols,
      "background",
      "borders",
      "angle",
      "vertical.align"
    ),
    with = FALSE
  ]
  grid_data <- merge(grid_data, cell_data, by = keycols, all.x = TRUE)
  setorderv(grid_data, cols = c("row_index", "col_index"))
  grid_data
}

#' @importFrom grid unit
grid_data_add_par_info <- function(grid_data, x) {
  par_data <- information_data_paragraph(x)

  # merge with grid_data to keep only active cells
  keycols <- c(".part", ".row_id", ".col_id")
  par_data <- merge(
    grid_data[,
      c(
        keycols,
        "vertical.align",
        "angle"
      ),
      with = FALSE
    ],
    par_data,
    by = keycols
  )

  # calculate justification
  par_data[,
    "just" := mapply(
      calc_grid_just,
      halign = .SD$text.align,
      valign = .SD$vertical.align,
      SIMPLIFY = FALSE
    )
  ]
  # calculate rotated justification
  par_data[,
    "justr" := mapply(
      calc_grid_rotated_just,
      just = .SD$just,
      angle = .SD$angle,
      SIMPLIFY = FALSE
    )
  ]

  # set padding as units
  par_data[, "padding.top" := .SD$padding.top / 72]
  par_data[, "padding.left" := .SD$padding.left / 72]
  par_data[, "padding.bottom" := .SD$padding.bottom / 72]
  par_data[, "padding.right" := .SD$padding.right / 72]

  # first-line/hanging indents in inches (hanging wins, as in officer).
  # Word semantics: all lines at padding.left, first line outdented by
  # `hanging` -> place the contents viewport at the first-line position
  # and indent continuation lines (see arrange_contents_grob)
  par_data[,
    "hanging" := fifelse(
      !is.na(.SD$hanging) & .SD$hanging > 0,
      .SD$hanging / 72,
      0
    )
  ]
  par_data[,
    "first_line" := fifelse(
      .SD$hanging <= 0 & !is.na(.SD$first_line) & .SD$first_line > 0,
      .SD$first_line / 72,
      0
    )
  ]
  par_data[, "padding.left" := pmax(0, .SD$padding.left - .SD$hanging)]

  par_data[, "paddingx" := .SD$padding.left + .SD$padding.right]
  par_data[, "paddingy" := .SD$padding.top + .SD$padding.bottom]

  # generate argument list for creating the contents viewport
  par_data[,
    "contents_vp" := mapply(
      function(just, justr, t, l, b, r, angle) {
        x <- unit(just[[1]], "npc")
        if (just[[1]] == 0) {
          x <- x + unit(l, "in")
        } else if (just[[1]] == 1) {
          x <- x - unit(r, "in")
        }
        y <- unit(just[[2]], "npc")
        if (just[[2]] == 0) {
          y <- y + unit(b, "in")
        } else if (just[[2]] == 1) {
          y <- y - unit(t, "in")
        }
        list(
          x = x,
          y = y,
          just = justr,
          angle = angle,
          width = unit(1, "npc") - unit(l + r, "in"),
          height = unit(1, "npc") - unit(t + b, "in"),
          clip = "inherit"
        )
      },
      just = .SD$just,
      justr = .SD$justr,
      t = .SD$padding.top,
      l = .SD$padding.left,
      b = .SD$padding.bottom,
      r = .SD$padding.right,
      angle = .SD$angle,
      SIMPLIFY = FALSE
    )
  ]

  # merge needed columns
  par_data <- par_data[,
    c(
      keycols,
      "line_spacing",
      "just",
      "justr",
      "paddingx",
      "paddingy",
      "first_line",
      "hanging",
      "contents_vp"
    ),
    with = FALSE
  ]
  grid_data <- merge(grid_data, par_data, by = keycols, all.x = TRUE)
  setorderv(grid_data, cols = c("row_index", "col_index"))

  # merge cell parameters
  grid_data[,
    "cell_params" := mapply(
      list,
      min_height = .SD$cell_min_height,
      max_height = .SD$cell_max_height,
      just = .SD$just,
      justr = .SD$justr,
      paddingx = .SD$paddingx,
      paddingy = .SD$paddingy,
      first_line = .SD$first_line,
      hanging = .SD$hanging,
      angle = .SD$angle,
      lineheight = .SD$line_spacing,
      SIMPLIFY = FALSE
    )
  ]

  grid_data
}

#' @importFrom grDevices is.raster
grid_data_add_chunk_info <- function(grid_data, x, autowidths, wrapping) {
  chunk_data <- information_data_chunk(x, expand_special_chars = FALSE)

  # merge with grid_data to keep only active cells
  keycols <- c(".part", ".row_id", ".col_id")
  chunk_data <- merge(
    grid_data[, c(keycols, "line_spacing", "angle"), with = FALSE],
    chunk_data,
    by = keycols
  )

  # chunk types
  chunk_data[,
    "is_raster" := sapply(.SD$img_data, is.raster) |
      sapply(.SD$img_data, is.character)
  ]
  chunk_data[, "is_equation" := !is.na(.SD$eq_data)]
  chunk_data[, "is_text" := !.SD$is_raster & !.SD$is_equation]

  # avoid CMD check warnings
  is_raster <- is_text <- txt <- is_newline <- is_space <-
    is_superscript <- is_subscript <- angle <- NULL

  # remove empty text
  chunk_data <- chunk_data[!((is_text) & txt %in% ""), , drop = FALSE]

  # remove breaks and tabs from superscript/subscript
  chunk_data[,
    "is_superscript" := .SD$is_text & .SD$vertical.align %in% "superscript"
  ]
  chunk_data[,
    "is_subscript" := .SD$is_text & .SD$vertical.align %in% "subscript"
  ]
  chunk_data[
    (is_superscript | is_subscript),
    "txt" := gsub("[\n\t]", "", .SD$txt)
  ]

  # set chunk index
  chunk_data[, "chunk_index" := 1:.N, by = keycols]
  setorderv(chunk_data, cols = c(keycols, "chunk_index"))

  # calculate fontface
  chunk_data[
    (is_text),
    "font.face" := fcase(
      .SD$bold & !.SD$italic , 2L ,
      !.SD$bold & .SD$italic , 3L ,
      .SD$bold & .SD$italic  , 4L ,
      default = 1L
    )
  ]

  # create gpar
  chunk_data[,
    "gp" := mapply(
      function(is_text, ...) {
        if (is_text) {
          gpar_text(...)
        } else {
          gpar()
        }
      },
      is_text = .SD$is_text,
      color = .SD$color,
      fontfamily = .SD$font.family,
      fontsize = .SD$font.size,
      fontface = .SD$font.face,
      lineheight = .SD$line_spacing,
      SIMPLIFY = FALSE
    )
  ]

  # resolve image data
  if (is.character(chunk_data$img_data)) {
    # fix to avoid data.table complaining about casting from char to list
    chunk_data$img_data <- as.list(chunk_data$img_data)
  }
  chunk_data[
    (is_raster),
    "img_data" := mapply(
      calc_grid_image,
      .SD$img_data,
      width = .SD$width,
      height = .SD$height,
      SIMPLIFY = FALSE
    )
  ]

  # create chunk .part data (parts split by newline)
  part_data <- chunk_data[,
    c(
      ".part",
      ".row_id",
      ".col_id",
      "line_spacing",
      "angle",
      "font.family",
      "font.size",
      "italic",
      "bold",
      "txt",
      "is_text",
      "is_superscript",
      "is_subscript",
      "width",
      "height",
      ".chunk_index",
      "chunk_index"
    ),
    with = FALSE
  ]
  # expand by newline
  part_data <- expand_special_char(part_data, what = "\n", with = NA)
  # set part_index per chunk
  setorderv(part_data, cols = c(keycols, "chunk_index", ".chunk_index"))
  part_data[, "part_index" := 1:.N, by = c(keycols, "chunk_index")]
  setorderv(part_data, cols = c(keycols, "chunk_index", "part_index"))
  # calculate metrics
  part_data <- calc_grid_text_metrics(part_data)

  # calculate now content dimensions if autowidths is TRUE and wrapping is FALSE
  if (autowidths && !wrapping) {
    # create line index per chunk
    part_data[, c("line_index") := fifelse(.SD$is_newline, 1L, 0L)]
    part_data[, c("line_index") := cumsum(.SD$line_index) + 1L, by = keycols]
    # calculate line metrics
    line_data <- part_data[,
      list(
        line_width = sum(.SD$width, na.rm = TRUE),
        line_height = max(.SD$ascent, na.rm = TRUE) +
          max(.SD$descent, na.rm = TRUE)
      ),
      by = c(keycols, "line_index")
    ]
    line_data <- line_data[,
      list(
        content_width = max(.SD$line_width, na.rm = TRUE),
        content_height = sum(.SD$line_height, na.rm = TRUE)
      ),
      by = keycols
    ]
    part_data <- merge(part_data, line_data, by = keycols)
    part_data[
      angle != 0,
      c("content_width", "content_height") := list(
        .SD$content_height,
        .SD$content_width
      )
    ]

    # merge content_width/content_height to grid data
    content_data <- part_data[,
      list(
        content_width = first(.SD$content_width),
        content_height = first(.SD$content_height)
      ),
      by = keycols
    ]
    grid_data <- merge(grid_data, content_data, by = keycols, all.x = TRUE)

    # adjust column widths & row heights
    grid_data <- grid_data_adjust_widths(grid_data)
    grid_data <- grid_data_adjust_heights(grid_data)

    # drop unneeded columns
    part_data[, c("line_index", "content_width", "content_height") := NULL]
    line_data <- NULL
  }

  # handle wrapping
  if (wrapping) {
    # create chunk word data (parts split by whitespace)
    word_data <- part_data
    word_data[,
      "wrapping" := fifelse(
        .SD$is_text &
          !.SD$is_space &
          !.SD$is_newline &
          !.SD$is_superscript &
          !.SD$is_subscript,
        "whitespace",
        "none"
      )
    ]
    # split by wrapping method
    split_list <- split(word_data, word_data$wrapping, drop = FALSE)
    if (is.data.table(split_list$whitespace)) {
      # expand by whitespace
      split_list$whitespace <- expand_special_char(
        split_list$whitespace,
        what = "[ \t]",
        with = NA
      )
    }
    # bind them together
    word_data <- rbindlist(split_list, use.names = TRUE, fill = TRUE)
    split_list <- NULL

    # set word_index and word_count per chunk part
    setorderv(
      word_data,
      cols = c(keycols, "chunk_index", "part_index", ".chunk_index")
    )
    word_data[,
      c("word_index", "word_count") := list(1:.N, .N),
      by = c(keycols, "chunk_index", "part_index")
    ]
    setorderv(
      word_data,
      cols = c(keycols, "chunk_index", "part_index", "word_index")
    )

    # calculate metrics
    word_data <- calc_grid_text_metrics(word_data)

    # calculate now content dimensions if autowidths is TRUE and wrapping is TRUE
    if (autowidths) {
      # don't take into account whitespace dimensions, except newline char in the beggining
      word_data[,
        "part_width" := fifelse(
          .SD$is_newline & .SD$part_index > 1 | .SD$is_space,
          0,
          .SD$width
        )
      ]
      word_data[,
        "part_ascent" := fifelse(
          .SD$is_newline & .SD$part_index > 1 | .SD$is_space,
          0,
          .SD$ascent
        )
      ]
      word_data[,
        "part_descent" := fifelse(
          .SD$is_newline & .SD$part_index > 1 | .SD$is_space,
          0,
          .SD$descent
        )
      ]

      # calculate content metrics
      word_data[,
        "content_width" := max(.SD$part_width, na.rm = TRUE),
        by = keycols
      ]
      word_data[,
        "content_height" := (sum(.SD$part_width, na.rm = TRUE) /
          .SD$content_width) *
          (max(.SD$part_ascent, na.rm = TRUE) +
            max(.SD$part_descent, na.rm = TRUE)),
        by = keycols
      ]
      word_data[
        angle != 0,
        c("content_width", "content_height") := list(
          .SD$content_height,
          .SD$content_width
        )
      ]

      # merge content_width to grid data
      content_data <- word_data[,
        list(
          content_width = first(.SD$content_width)
        ),
        by = keycols
      ]
      grid_data <- merge(grid_data, content_data, by = keycols, all.x = TRUE)
      # adjust column widths
      grid_data <- grid_data_adjust_widths(grid_data)
      # get back cell width
      content_data <- grid_data[,
        list(
          content_width = first(.SD$cell_width)
        ),
        by = keycols
      ]
      word_data[, "content_width" := NULL]
      word_data <- merge(word_data, content_data, by = keycols, all.x = TRUE)
      # now recalculate the content height
      word_data[,
        "content_height" := (sum(.SD$part_width, na.rm = TRUE) /
          .SD$content_width) *
          (max(.SD$part_ascent, na.rm = TRUE) +
            max(.SD$part_descent, na.rm = TRUE)),
        by = keycols
      ]
      word_data[
        angle != 0,
        c("content_width", "content_height") := list(
          .SD$content_height,
          .SD$content_width
        )
      ]
      # merge content_height to grid data
      content_data <- word_data[,
        list(
          content_height = first(.SD$content_height)
        ),
        by = keycols
      ]
      grid_data <- merge(grid_data, content_data, by = keycols, all.x = TRUE)
      # adjust row heights
      grid_data <- grid_data_adjust_heights(grid_data)

      # drop not needed columns
      word_data[,
        c(
          "content_width",
          "content_height",
          "wrapping",
          "part_width",
          "part_ascent",
          "part_descent"
        ) := NULL
      ]
    }

    # create chunk word data (parts split by any character)
    char_data <- word_data[
      (is_text & !is_space & !is_newline & !is_superscript & !is_subscript),
      ,
      drop = FALSE
    ]
    # expand by character
    char_data <- expand_special_char(char_data, what = ".", with = NA)

    # set word_index and word_count per chunk part
    setorderv(
      char_data,
      cols = c(
        keycols,
        "chunk_index",
        "part_index",
        "word_index",
        ".chunk_index"
      )
    )
    char_data[,
      c("char_index", "char_count") := list(seq_len(.N), .N),
      by = c(keycols, "chunk_index", "part_index", "word_index")
    ]

    # calculate metrics
    if (nrow(char_data) > 0) {
      char_data <- calc_grid_text_metrics(char_data)
    }

    # merge char_count to word data
    word_char_data <- char_data[,
      list(
        char_count = first(.SD$char_count)
      ),
      by = c(keycols, "chunk_index", "part_index", "word_index")
    ]
    word_data <- merge(
      word_data,
      word_char_data,
      by = c(keycols, "chunk_index", "part_index", "word_index"),
      all.x = TRUE
    )

    # merge word_count to part data
    part_word_data <- word_data[,
      list(
        word_count = first(.SD$word_count),
        char_count = sum(.SD$char_count)
      ),
      by = c(keycols, "chunk_index", "part_index")
    ]
    part_data <- merge(
      part_data,
      part_word_data,
      by = c(keycols, "chunk_index", "part_index"),
      all.x = TRUE
    )

    # merge part/word/char data together
    word_index <- word_count <- char_index <- char_count <- NULL
    part_data <- rbindlist(
      list(
        part_data,
        word_data[word_count > 1], # only multiple words
        char_data[char_count > 1] # only multiple chars
      ),
      use.names = TRUE,
      fill = TRUE,
      idcol = "wrap_level" # set idcol to know where data came from
    )
    part_data[is.na(word_index), "word_index" := 0L]
    part_data[is.na(word_count), "word_count" := 0L]
    part_data[is.na(char_index), "char_index" := 0L]
    part_data[is.na(char_count), "char_count" := 0L]
    setorderv(
      part_data,
      cols = c(
        keycols,
        "chunk_index",
        "part_index",
        "word_index",
        "char_index"
      )
    )
    # set .chunk_index unique for each cell
    part_data[, ".chunk_index" := 1:.N, by = keycols]

    # set wrap child count
    part_data[,
      "child_count" := fcase(
        .SD$wrap_level == 1 & .SD$word_count > 1 , .SD$word_count ,
        .SD$wrap_level == 1 & .SD$char_count > 1 , .SD$char_count ,
        .SD$wrap_level == 2                      , .SD$char_count ,
        default = 0L
      )
    ]
    # adjust wrap level
    part_data[,
      "wrap_level" := fcase(
        .SD$wrap_level == 1                       , 1L ,
        .SD$wrap_level == 2                       , 2L ,
        .SD$wrap_level == 3 & .SD$word_count == 1 , 2L ,
        default = 3L
      )
    ]

    # calculate wrap children indices
    # for any row which has wrapped children
    part_data[,
      "children_index" := list(calc_grid_wrap_children(.SD)),
      by = c(keycols, "chunk_index")
    ]
  } else {
    # no wrapping
    part_data[, ".chunk_index" := 1:.N, by = keycols]
    part_data[, "child_count" := 0L]
    part_data[, "children_index" := integer(0)]
  }
  setorderv(part_data, cols = c(keycols, ".chunk_index"))

  # replace special chars
  part_data[(is_newline), "txt" := ""]
  part_data[(is_space), "txt" := " "]

  # merge chunk data
  grid_data <- merge(
    grid_data,
    chunk_data[,
      list(
        chunk_data = list(
          data.table(
            chunk_index = .SD$chunk_index,
            is_text = .SD$is_text,
            is_raster = .SD$is_raster,
            is_equation = .SD$is_equation,
            is_underlined = .SD$underlined,
            is_bold = .SD$bold,
            is_italic = .SD$italic,
            valign = .SD$vertical.align,
            width = .SD$width,
            height = .SD$height,
            gp = .SD$gp,
            txt_data = .SD$txt,
            img_data = .SD$img_data,
            eq_data = .SD$eq_data,
            shading_color = .SD$shading.color
          )
        )
      ),
      by = keycols
    ],
    by = keycols,
    all.x = TRUE
  )

  # merge part data
  grid_data <- merge(
    grid_data,
    part_data[,
      list(
        chunk_part_data = list(
          data.table(
            .chunk_index = .SD$.chunk_index,
            chunk_index = .SD$chunk_index,
            child_count = .SD$child_count,
            children_index = .SD$children_index,
            row_index = 0L,
            col_index = 0L,
            is_newline = .SD$is_newline,
            is_whitespace = .SD$is_space,
            width = .SD$width,
            height = .SD$height,
            ascent = .SD$ascent,
            descent = .SD$descent,
            txt_data = .SD$txt
          )
        )
      ),
      by = keycols
    ],
    by = keycols,
    all.x = TRUE
  )
  setorderv(grid_data, cols = c("row_index", "col_index"))

  # drop unneeded columns
  grid_data[,
    c(
      "cell_min_height",
      "cell_max_height",
      "just",
      "justr",
      "paddingx",
      "paddingy",
      "angle",
      "line_spacing",
      "vertical.align"
    ) := NULL
  ]

  grid_data
}

#' @importFrom data.table fsetdiff
grid_data_adjust_widths <- function(grid_data) {
  # collect the minimum width of each cell
  dat <- grid_data[,
    list(
      content_min_width = max(
        .SD$cell_width,
        safe_stat(
          .SD$content_width + .SD$paddingx,
          FUN = max,
          NA_value = NA_real_
        ),
        na.rm = TRUE
      )
    ),
    by = c("col_index", "rowspan")
  ]
  # separate columns with minimum span
  dat_minspan <- dat[, .SD[which.min(.SD$rowspan), ], by = "col_index"]
  # separate columns with maximum span
  dat_maxspan <- fsetdiff(dat, dat_minspan)
  if (nrow(dat_maxspan) > 0) {
    col_index <- NULL
    # recalculate minimum width for spanned cells
    dat_maxspan[,
      c("prev_width", "content_min_width") := list(
        .SD$content_min_width,
        mapply(
          function(index, span, width) {
            sum(
              dat_minspan[
                col_index >= index & col_index < index + span
              ]$content_min_width
            )
          },
          .SD$col_index,
          .SD$rowspan,
          .SD$content_min_width
        )
      )
    ]
    # check the difference of new width vs the previous
    dat_maxspan[, "diff" := .SD$prev_width - .SD$content_min_width]
    # if there is a positive difference, it means that some spanned cells are wider than
    # the cummulative widths of their columns
    # in that case we need to re-adjust the column widths
    if (any(dat_maxspan$diff > 0)) {
      dat_overflow <- dat_maxspan[
        diff > 0,
        .SD[which.max(.SD$diff), ],
        by = "col_index"
      ]
      dat_overflow[, "group" := .SD$col_index]
      dat_overflow <- dat_overflow[,
        list(
          col_index = seq.int(
            from = .SD$col_index,
            length.out = .SD$rowspan,
            by = 1L
          ),
          diff = .SD$diff / .SD$rowspan
        ),
        by = "group"
      ]
      dat_overflow <- dat_overflow[,
        .SD[which.max(.SD$diff), ],
        by = "col_index"
      ]
      dat_overflow[, "group" := NULL]
      dat <- merge(
        dat,
        dat_overflow,
        by = "col_index",
        all.x = TRUE
      )
      dat[, "diff" := .SD$diff * .SD$rowspan]
      dat[,
        "content_min_width" := sum(
          .SD$content_min_width,
          .SD$diff,
          na.rm = TRUE
        ),
        by = c("col_index", "rowspan")
      ]
      dat[, "diff" := NULL]
      dat_minspan <- dat[, .SD[which.min(.SD$rowspan), ], by = "col_index"]
      dat_maxspan <- fsetdiff(dat, dat_minspan)
      dat_maxspan[,
        c("prev_width", "content_min_width") := list(
          .SD$content_min_width,
          mapply(
            function(index, span, width) {
              sum(
                dat_minspan[
                  col_index >= index & col_index < index + span
                ]$content_min_width
              )
            },
            .SD$col_index,
            .SD$rowspan,
            .SD$content_min_width
          )
        )
      ]
    }
    dat <- rbindlist(list(
      dat_minspan[, c("col_index", "rowspan", "content_min_width")],
      dat_maxspan[, c("col_index", "rowspan", "content_min_width")]
    ))
  }
  grid_data <- merge(
    grid_data,
    dat,
    by = c("col_index", "rowspan"),
    all.x = TRUE
  )
  grid_data[, "cell_width" := .SD$content_min_width]
  grid_data
}

grid_data_adjust_heights <- function(grid_data) {
  # collect the minimum height of each cell
  dat <- grid_data[,
    list(
      content_min_height = safe_stat(
        safe_stat(
          .SD$cell_min_height,
          safe_stat(.SD$cell_height, FUN = max, NA_value = NA_real_),
          safe_stat(
            .SD$content_height + .SD$paddingy,
            FUN = max,
            NA_value = NA_real_
          ),
          FUN = max,
          NA_value = NA_real_
        ),
        .SD$cell_max_height,
        FUN = min,
        NA_value = NA_real_
      )
    ),
    by = c("row_index", "colspan")
  ]
  # separate rows with minimum span
  dat_minspan <- dat[, .SD[which.min(.SD$colspan), ], by = "row_index"]
  # separate rows with maximum span
  dat_maxspan <- fsetdiff(dat, dat_minspan)
  if (nrow(dat_maxspan) > 0) {
    row_index <- NULL
    # recalculate minimum height for spanned cells
    dat_maxspan[,
      c("prev_height", "content_min_height") := list(
        .SD$content_min_height,
        mapply(
          function(index, span, height) {
            sum(
              dat_minspan[
                row_index >= index & row_index < index + span
              ]$content_min_height
            )
          },
          .SD$row_index,
          .SD$colspan,
          .SD$content_min_height
        )
      )
    ]
    # check the difference of new height vs the previous
    dat_maxspan[, "diff" := .SD$prev_height - .SD$content_min_height]
    # if there is a positive difference, it means that some spanned cells are taller than
    # the cummulative heights of their rows
    # in that case we need to re-adjust the row heights
    if (any(dat_maxspan$diff > 0)) {
      dat_overflow <- dat_maxspan[
        diff > 0,
        .SD[which.max(.SD$diff), ],
        by = "row_index"
      ]
      dat_overflow[, "group" := .SD$row_index]
      dat_overflow <- dat_overflow[,
        list(
          row_index = seq.int(
            from = .SD$row_index,
            length.out = .SD$colspan,
            by = 1L
          ),
          diff = diff / .SD$colspan
        ),
        by = "group"
      ]
      dat_overflow <- dat_overflow[,
        .SD[which.max(.SD$diff), ],
        by = "row_index"
      ]
      dat_overflow[, "group" := NULL]
      dat <- merge(
        dat,
        dat_overflow,
        by = "row_index",
        all.x = TRUE
      )
      dat[, "diff" := .SD$diff * .SD$colspan]
      dat[,
        "content_min_height" := sum(
          .SD$content_min_height,
          .SD$diff,
          na.rm = TRUE
        ),
        by = c("row_index", "colspan")
      ]
      dat[, "diff" := NULL]
      dat_minspan <- dat[, .SD[which.min(.SD$colspan), ], by = "row_index"]
      dat_maxspan <- fsetdiff(dat, dat_minspan)
      dat_maxspan[,
        c("prev_height", "content_min_height") := list(
          .SD$content_min_height,
          mapply(
            function(index, span, height) {
              sum(
                dat_minspan[
                  row_index >= index & row_index < index + span
                ]$content_min_height
              )
            },
            .SD$row_index,
            .SD$colspan,
            .SD$content_min_height
          )
        )
      ]
    }
    dat <- rbindlist(list(
      dat_minspan[, c("row_index", "colspan", "content_min_height")],
      dat_maxspan[, c("row_index", "colspan", "content_min_height")]
    ))
  }
  grid_data <- merge(
    grid_data,
    dat,
    by = c("row_index", "colspan"),
    all.x = TRUE
  )
  grid_data[, "cell_height" := .SD$content_min_height]
  grid_data
}

# gpar --------------------------------------------------------------------

#' @importFrom grid gpar
gpar_background <- function(fill) {
  if (isTRUE(!is.na(fill) && fill != "transparent")) {
    gpar(
      lwd = 0.5,
      fill = fill,
      col = fill
    )
  }
}

gpar_ltys <- c(
  solid = 1,
  dashed = 2,
  dotted = 3,
  dotdash = 4,
  longdash = 5,
  twodash = 6
)

gpar_border <- function(width, color, style) {
  if (
    isTRUE(
      width > 0 &&
        !is.na(color) &&
        !color %in% "transparent" &&
        (style %in% gpar_ltys || style %in% names(gpar_ltys))
    )
  ) {
    gpar(
      lwd = width * 72.27 / 25.4,
      col = color,
      lty = style
    )
  }
}

gpar_text <- function(color, fontfamily, fontsize, fontface, lineheight) {
  gpar(
    col = color,
    fontfamily = fontfamily,
    fontsize = fontsize,
    fontface = fontface,
    lineheight = lineheight
  )
}

# utils -------------------------------------------------------------------

# helper to create grouping by rowspan or colspan
calc_grid_span_group <- function(x) {
  x <- x[x > 0]
  inverse.rle(structure(
    list(lengths = x, values = seq_along(x)),
    class = "rle"
  ))
}

calc_grid_just <- function(halign, valign) {
  c(
    # hjust
    fcase(
      halign == "left"  , 0 ,
      halign == "right" , 1 ,
      default = 0.5
    ),
    # vjust
    fcase(
      valign == "bottom" , 0 ,
      valign == "top"    , 1 ,
      default = 0.5
    )
  )
}

calc_grid_rotated_just <- function(just, angle) {
  hjust <- just[[1]]
  vjust <- just[[2]]
  c(
    # hjust
    fcase(
      0 <= angle & angle < 90                                   , hjust     ,
      90 <= angle & angle < 180                                 ,
      fifelse(vjust == 0, 0, fifelse(vjust == 1, 1, 1 - vjust)) ,
      180 <= angle & angle < 270                                , 1 - hjust ,
      270 <= angle & angle < 360                                ,
      fifelse(vjust == 0, 1, fifelse(vjust == 1, 0, vjust))
    ),
    # vjust
    fcase(
      0 <= angle & angle < 90                                   , vjust     ,
      90 <= angle & angle < 180                                 ,
      fifelse(hjust == 0, 1, fifelse(hjust == 1, 0, hjust))     ,
      180 <= angle & angle < 270                                , 1 - vjust ,
      270 <= angle & angle < 360                                ,
      fifelse(hjust == 0, 0, fifelse(hjust == 1, 1, 1 - hjust))
    )
  )
}

#' @importFrom gdtools strings_sizes
calc_grid_text_metrics <- function(dat) {
  # avoid CMD check warnings
  is_text <- is_newline <- is_space <- is_superscript <- is_subscript <- NULL

  # handle special chars
  dat[, "is_newline" := .SD$is_text & .SD$txt %in% "\n"]
  dat[, "txt" := gsub("\t", "\u020\u020\u020\u020", .SD$txt)]
  dat[, "is_space" := .SD$is_text & .SD$txt %in% " "]
  # temporarily replace these, so that they get some height
  dat[(is_newline), "txt" := "."]

  # calculate string metrics
  txt_metrics <- gdtools::strings_sizes(
    dat$txt,
    fontname = dat$font.family,
    fontsize = dat$font.size,
    bold = dat$bold,
    italic = dat$italic
  )
  colnames(txt_metrics) <- c("part_width", "part_ascent", "part_descent")
  dat <- cbind(dat, txt_metrics)

  # set width/ascent/descent/height
  dat[(is_text), "width" := .SD$part_width]
  dat[(is_superscript | is_subscript), "width" := .SD$width * 0.75]
  dat[(is_text), "ascent" := .SD$part_ascent * .SD$line_spacing]
  dat[(is_text), "descent" := .SD$part_descent * .SD$line_spacing]
  dat[(is_text), "height" := .SD$ascent + .SD$descent]
  dat[!(is_text), c("ascent", "descent") := list(.SD$height, 0)]
  dat[(is_newline), "txt" := "\n"]

  # drop not needed columns
  dat[, c("part_width", "part_ascent", "part_descent") := NULL]

  # done
  dat
}

calc_grid_wrap_children <- function(dat) {
  wrap_level <- .chunk_index <- NULL
  mapply(
    function(level, index, child_count) {
      if (child_count > 0) {
        head(
          dat[wrap_level == (level + 1) & .chunk_index > index, , drop = FALSE],
          n = child_count
        )$.chunk_index
      } else {
        integer(0)
      }
    },
    dat$wrap_level,
    dat$.chunk_index,
    dat$child_count,
    SIMPLIFY = FALSE
  )
}

calc_grid_label <- function(
  x,
  valign = "baseline",
  is_underlined = FALSE,
  is_bold = FALSE,
  is_italic = FALSE
) {
  if (valign %in% c("superscript", "subscript") || isTRUE(is_underlined)) {
    x <- paste0("\"", x, "\"")
    if (valign %in% "superscript") {
      x <- paste0("\"\"^", x)
    } else if (valign %in% "subscript") {
      x <- paste0("\"\"[", x, "]")
    }
    if (isTRUE(is_underlined)) {
      x <- paste0("underline(", x, ")")
    }
    if (isTRUE(is_bold) && isTRUE(is_italic)) {
      x <- paste0("bolditalic(", x, ")")
    } else if (isTRUE(is_bold)) {
      x <- paste0("bold(", x, ")")
    } else if (isTRUE(is_italic)) {
      x <- paste0("italic(", x, ")")
    }
    parse(text = x)
  } else {
    x
  }
}

blank_raster_image <- as.raster(matrix("transparent", nrow = 1, ncol = 1))

calc_grid_image <- function(img_data, width = NULL, height = NULL) {
  image <- blank_raster_image
  if (is.raster(img_data)) {
    image <- img_data
    class(image) <- c("flextableRasterChunk", class(image))
  } else if (is.character(img_data)) {
    if (!file.exists(img_data)) {
      warning(sprintf("file '%s' can not be found.", img_data))
    } else if (requireNamespace("magick", quietly = TRUE)) {
      tryCatch(
        {
          if (grepl("\\.svg$", ignore.case = TRUE, x = img_data)) {
            image <- magick::image_read_svg(img_data, width * 72, height * 72)
          } else {
            image <- magick::image_read(img_data)
          }
          image <- as.raster(image)
        },
        error = function(e) {
          warning("error reading image: ", e)
        }
      )
    } else {
      warning(sprintf("package 'magick' is required to read image files"))
    }
  }
  image
}

# Word ----
#' @importFrom officer run_bookmark ftext
caption_bookdown_docx_md <- function(x, tab_props = opts_current_table()) {
  # for bookdown::word_document2.
  # 'bookdown' wants a table reference as (#tab:bookmark) to enable cross-references
  # in a non raw block. It is then only possible to format chunk of text but
  # it is not possible to add paragraph settings as keep_with_next or ft.align.
  #
  # So, this caption use Word style name ('Table Caption' by default) as only
  # solution to control paragraph formattings. The chunks of text can be made
  # with `set_caption(caption=as_paragraph(...))`. It is defined with a custom
  # raw bloc (surrounded with ::::).
  #
  # Table captions formattings should be defined in the reference_docx file
  # and ft.align should be defined with the same value than the alignement
  # of the 'Table caption' paragraph properties
  if (!has_caption(x)) {
    return("")
  }
  style_start <- "::: {custom-style=\"Table Caption\"}\n"
  style_end <- "\n:::\n"

  caption_chunks_str <- caption_chunks_wml(x)
  caption_chunks_str <- paste("`", caption_chunks_str, "`{=openxml}", sep = "")

  caption_id <- if (!is.null(x$caption$autonum$bookmark)) {
    paste0(
      "(\\#",
      x$caption$autonum$seq_id,
      ":",
      x$caption$autonum$bookmark,
      ")"
    )
  } else {
    ""
  }

  caption <- paste(
    style_start,
    paste0("<caption>", caption_id, caption_chunks_str, "</caption>"),
    style_end,
    "",
    sep = "\n"
  )
  caption
}

caption_default_docx_openxml <- function(
  x,
  keep_with_next = FALSE,
  allow_autonum = TRUE
) {
  # for rmarkdown::word_document
  # There is no table reference and no cross-references.
  # It is possible to format chunk of text and also define
  # paragraph settings as keep_with_next or ft.align.
  #
  # So, this caption use Word style name ('Table Caption' by default)
  # The chunks of text can be made
  # with `set_caption(caption=as_paragraph(...))`. The paragraph properties
  # can be defined with `set_caption(fp_p=fp_par(...))`.
  if (!has_caption(x)) {
    return("")
  }

  p_pr <- ""
  x <- process_caption_fp_par(x, keep_with_next = keep_with_next)
  if (!is.null(x$caption$fp_p)) {
    p_pr <- format(x$caption$fp_p, type = "wml")
    if (packageVersion("officer") > numeric_version("0.6.8")) {
      m <- regexpr("w:pstlname=\"[^\"]+\"", p_pr)
      pstlname <- regmatches(x = p_pr, m)
      style_name <- gsub("(w:pstlname=)(\")([^\"]+)(\")", "\\3", pstlname)
      style_id <- gsub("[^a-zA-Z0-9]", "", style_name)
      p_pr <- gsub(
        pstlname,
        sprintf("w:val=\"%s\"", style_id),
        p_pr,
        fixed = TRUE
      )
    }
  } else {
    cap_style_id <- gsub("[^a-zA-Z0-9]", "", x$caption$word_stylename)
    p_pr <- paste0(
      "<w:pPr>",
      sprintf("<w:pStyle w:val=\"%s\"/>", cap_style_id),
      sprintf("<w:jc w:val=\"%s\"/>", x$properties$align),
      if (keep_with_next) "<w:keepNext/>",
      "</w:pPr>"
    )
  }

  caption_chunks_str <- caption_chunks_wml(x)

  autonum <- ""
  if (has_autonum(x) && allow_autonum) {
    autonum <- to_wml(x$caption$autonum)
  }

  caption_str <- paste0(
    c(
      wp_ns_yes,
      p_pr,
      autonum,
      caption_chunks_str,
      "</w:p>"
    ),
    collapse = ""
  )
  caption_str
}
caption_default_rtf <- function(x, keep_with_next = FALSE) {
  if (!has_caption(x)) {
    return("")
  }

  p_pr <- ""
  x <- process_caption_fp_par(x, keep_with_next = keep_with_next)
  if (!is.null(x$caption$fp_p)) {
    p_pr <- format(x$caption$fp_p, type = "rtf")
  }

  caption_chunks_str <- caption_chunks_rtf(x)

  autonum <- ""
  if (has_autonum(x)) {
    autonum <- to_rtf(x$caption$autonum)
  }

  caption_str <- paste0(
    c(
      "{\\pard",
      p_pr,
      autonum,
      caption_chunks_str,
      "\\par}"
    ),
    collapse = ""
  )
  caption_str
}

# HTML ----

caption_bookdown_html <- function(x) {
  # for bookdown::html_document2.
  # 'bookdown' wants a table reference as (#tab:bookmark) to enable cross-references
  # in a non raw block. It is then only possible to format chunk of text but
  # it is not possible to add paragraph settings as keep_with_next or ft.align.
  #
  # So, this caption use Word style name ('Table Caption' by default) as only
  # solution to control paragraph formattings. The chunks of text can be made
  # with `set_caption(caption=as_paragraph(...))`. It is defined with a custom
  # raw bloc (surrounded with ::::).
  #
  # Table captions formattings should be defined in the reference_docx file
  # and ft.align should be defined with the same value than the alignement
  # of the 'Table caption' paragraph properties

  if (!has_caption(x)) {
    return("")
  }
  x <- process_caption_fp_par(x, keep_with_next = FALSE)

  reference_label <- ""
  if (!is.null(x$caption$autonum$bookmark)) {
    reference_label <- paste0(
      "(#",
      x$caption$autonum$seq_id,
      ":",
      x$caption$autonum$bookmark,
      ")"
    )
  }

  if (!is.null(x$caption$fp_p)) {
    inline_css <- sprintf(
      " style=\"display:table-caption;%s\"",
      format(x$caption$fp_p, type = "html")
    )
  } else {
    inline_css <- " style=\"display:table-caption;\""
  }

  if (is.null(x$caption$html_classes)) {
    caption_class <- ""
  } else {
    caption_class <- sprintf(" class=\"%s\"", x$caption$html_classes)
  }

  html_caption <- caption_chunks_html(x)
  caption_chunks_str <- html_caption$html
  css <- html_caption$css
  caption_str <- paste0(
    sprintf("\n<caption%s%s>", inline_css, caption_class),
    reference_label,
    caption_chunks_str,
    "</caption>\n"
  )

  attr(caption_str, "css") <- css
  caption_str
}

caption_default_html <- function(x) {
  if (!has_caption(x)) {
    caption_str <- ""
    attr(caption_str, "css") <- ""
    return(caption_str)
  }

  x <- process_caption_fp_par(x, keep_with_next = FALSE)
  reference_label <- ""

  if (!is.null(x$caption$autonum$bookmark)) {
    reference_label <- sprintf(" id=\"%s\"", x$caption$autonum$bookmark)
  }

  if (!is.null(x$caption$fp_p)) {
    inline_css <- sprintf(
      " style=\"display:table-caption;%s\"",
      format(x$caption$fp_p, type = "html")
    )
  } else {
    inline_css <- " style=\"display:table-caption;\""
  }

  if (is.null(x$caption$html_classes)) {
    caption_class <- ""
  } else {
    caption_class <- sprintf(" class=\"%s\"", x$caption$html_classes)
  }

  html_caption <- caption_chunks_html(x)
  caption_chunks_str <- html_caption$html
  has_caption_label <- !is.null(caption_chunks_str)
  css <- html_caption$css

  caption_str <- paste0(
    sprintf("<caption%s%s%s>", inline_css, caption_class, reference_label),
    caption_chunks_str,
    "</caption>"
  )
  attr(caption_str, "css") <- css

  caption_str
}

# PDF ----
caption_default_latex <- function(x) {
  if (!has_caption(x)) {
    return("")
  }
  caption_chunks_str <- caption_chunks_latex(x)

  caption <- paste0(
    "\\caption{",
    caption_chunks_str,
    "}",
    "\\\\"
  )
  caption
}

caption_bookdown_latex <- function(x) {
  if (!has_caption(x)) {
    return("")
  }

  reference_label <- ""
  if (!is.null(x$caption$autonum$bookmark)) {
    reference_label <- paste0(
      "(\\#",
      x$caption$autonum$seq_id,
      ":",
      x$caption$autonum$bookmark,
      ")"
    )
  }

  caption_chunks_str <- caption_chunks_latex(x)

  caption <- paste0(
    "\\caption{",
    caption_chunks_str,
    "}",
    reference_label,
    "\\\\"
  )
  caption
}

# utils -----
wp_ns_yes <- "<w:p xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">"
wp_ns_no <- "<w:p>"
has_caption <- function(x) {
  has_it <- FALSE
  if (!is.null(x$caption$value)) {
    has_it <- TRUE
  }
  has_it
}

has_autonum <- function(x) {
  has_it <- FALSE
  if (!is.null(x$caption$autonum)) {
    has_it <- TRUE
  }
  has_it
}
process_caption_fp_par <- function(x, keep_with_next) {
  if (!is.null(x$caption$fp_p)) {
    fp_p <- x$caption$fp_p
    if (x$caption$align_with_table) {
      fp_p <- update(fp_p, text.align = x$properties$align)
    }
    fp_p <- update(fp_p, word_style = x$caption$word_stylename)
    if (keep_with_next) {
      fp_p <- update(fp_p, keep_with_next = TRUE)
    }
    x$caption$fp_p <- fp_p
  }
  x
}

caption_chunks_wml <- function(x) {
  if (!x$caption$simple_caption) {
    spans <- runs_as_wml(x, txt_data = x$caption$value)
    caption_chunks_str <- paste(spans$run_openxml, collapse = "")
  } else {
    caption_chunks_str <- paste0(
      "<w:r><w:t xml:space=\"preserve\">",
      htmlEscape(x$caption$value),
      "</w:t></w:r>"
    )
  }
  caption_chunks_str
}
caption_chunks_rtf <- function(x) {
  if (!x$caption$simple_caption) {
    spans <- runs_as_rtf(x, chunk_data = x$caption$value)
    caption_chunks_str <- paste(spans$txt, collapse = "")
  } else {
    caption_chunks_str <- x$caption$value
  }
  caption_chunks_str
}

caption_chunks_html <- function(x) {
  if (!x$caption$simple_caption) {
    caption_df <- x$caption$value
    caption_spans <- runs_as_html(x, chunk_data = caption_df)
    css <- attr(caption_spans, "css")
    caption_chunks_str <- paste0(caption_spans$span_tag, collapse = "")
  } else {
    css <- ""
    caption_chunks_str <- paste0(
      "<span>",
      htmlize(x$caption$value),
      "</span>"
    )
  }
  list(html = caption_chunks_str, css = css)
}

caption_chunks_latex <- function(x) {
  if (!x$caption$simple_caption) {
    caption_str <- runs_as_latex(x = x, chunk_data = x$caption$value)
    caption_chunks_str <- paste(caption_str$txt, collapse = "")
    caption_chunks_str <- gsub(
      "\\linebreak ",
      "\\newline ",
      caption_chunks_str,
      fixed = TRUE
    )
  } else {
    caption_chunks_str <- sanitize_latex_str(x$caption$value)
    caption_chunks_str <- gsub(
      "\n",
      "\\newline ",
      caption_chunks_str,
      fixed = TRUE
    )
  }
  caption_chunks_str
}

#' @importFrom data.table is.data.table .N
expand_special_char <- function(x, what, with = NA, ...) {
  m <- gregexec(pattern = what, x$txt, ...)
  if (isTRUE(any(unlist(m) > -1))) {
    txt <- regmatches(x$txt, m, invert = NA)
    txt <- lapply(txt, function(z) z[nzchar(z)])
    if (is.character(with) && !is.na(with)) {
      txt <- lapply(txt, gsub, pattern = what, replacement = with, ...)
    }
    len <- lapply(txt, length)

    was_dt <- is.data.table(x)
    setDT(x)
    x <- x[rep(seq_len(.N), len)][, ".chunk_index" := seq_len(.N)]
    x$txt <- unlist(txt)
    if (!was_dt) setDF(x)
  }
  x
}


#' @noRd
#' @title fortify width
#' @description create a data.frame with width information.
fortify_width <- function(x) {
  dat <- list()
  for (part in c("header", "body", "footer")) {
    nr <- nrow_part(x, part)
    if (nr > 0) {
      dat[[part]] <- data.frame(
        .col_id = x$col_keys,
        width = x[[part]]$colwidths,
        stringsAsFactors = FALSE
      )
    }
  }
  dat <- data.table::rbindlist(dat)
  dat <- dat[, list(width = safe_stat(.SD$width, FUN = max)), by = ".col_id"]
  setDF(dat)
  dat$.col_id <- factor(dat$.col_id, levels = x$col_keys)
  setorderv(dat, cols = c(".col_id"))

  dat
}

#' @noRd
#' @title fortify width
#' @description create a data.frame with height information.
fortify_height <- function(x) {
  rows <- list()
  for (part in c("header", "body", "footer")) {
    nr <- nrow_part(x, part)
    if (nr > 0) {
      rows[[part]] <- data.frame(
        .row_id = seq_len(nr),
        height = x[[part]]$rowheights,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  dat <- rbindlist(rows, use.names = TRUE, idcol = ".part")
  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  setorderv(dat, cols = c(".part", ".row_id"))

  setDF(dat)
  dat
}

#' @noRd
#' @title fortify hrule
#' @description create a data.frame with hrule information.
fortify_hrule <- function(x) {
  rows <- list()
  for (part in c("header", "body", "footer")) {
    nr <- nrow_part(x, part)
    if (nr > 0) {
      rows[[part]] <- data.frame(
        .row_id = seq_len(nr),
        hrule = x[[part]]$hrule,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  dat <- rbindlist(rows, use.names = TRUE, idcol = ".part")
  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  setorderv(dat, cols = c(".part", ".row_id"))
  setDF(dat)
  dat
}

#' @noRd
#' @title fortify rows and columns spans
#' @description create a data.frame with span information.
fortify_span <- function(x, parts = c("header", "body", "footer")) {
  rows <- list()
  for (part in parts) {
    if (nrow_part(x, part) > 0) {
      nr <- nrow(x[[part]]$spans$rows)
      rows[[part]] <- data.frame(
        .col_id = rep(x$col_keys, each = nr),
        .row_id = rep(seq_len(nr), length(x$col_keys)),
        rowspan = as.vector(x[[part]]$spans$rows),
        colspan = as.vector(x[[part]]$spans$columns),
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }
  dat <- rbindlist(rows, use.names = TRUE, idcol = ".part")
  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  dat$.col_id <- factor(dat$.col_id, levels = x$col_keys)
  setorderv(dat, cols = c(".part", ".row_id", ".col_id"))

  setDF(dat)

  dat
}


# distinct_properties ----
#' @importFrom data.table setDT
#' @importFrom uuid UUIDgenerate
#' @noRd
distinct_text_properties <- function(x, add_columns = character(length = 0L)) {
  columns <- c(
    "color",
    "font.size",
    "bold",
    "italic",
    "underlined",
    "strike",
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "vertical.align",
    "shading.color",
    add_columns
  )
  columns <- intersect(columns, colnames(x))
  dat <- as.data.table(x[columns])
  uid <- unique(dat)
  setDF(dat)

  classname <- UUIDgenerate(n = nrow(uid), use.time = TRUE)
  classname <- gsub("(^[[:alnum:]]+)(.*)$", "cl-\\1", classname)
  uid$classname <- classname

  setDF(uid)

  uid
}
distinct_paragraphs_properties <- function(x) {
  # fp_columns <- intersect(names(formals(officer::fp_par)), colnames(x))
  columns <- c(
    "text.align",
    "line_spacing",
    "padding.bottom",
    "padding.top",
    "padding.left",
    "padding.right",
    "shading.color",
    "keep_with_next",
    "border.width.bottom",
    "border.width.top",
    "border.width.left",
    "border.width.right",
    "border.color.bottom",
    "border.color.top",
    "border.color.left",
    "border.color.right",
    "border.style.bottom",
    "border.style.top",
    "border.style.left",
    "border.style.right",
    "text.direction",
    "vertical.align",
    "tabs",
    "word_style",
    "first_line",
    "hanging"
  )
  columns <- intersect(columns, colnames(x))
  dat <- as.data.frame(x)[columns]
  setDT(dat)

  uid <- unique(dat)

  classname <- UUIDgenerate(n = nrow(uid), use.time = TRUE)
  classname <- gsub("(^[[:alnum:]]+)(.*)$", "cl-\\1", classname)
  uid$classname <- classname

  setDF(uid)

  uid
}

distinct_cells_properties <- function(x) {
  # fp_columns <- intersect(names(formals(officer::fp_cell)), colnames(x))
  columns <- c(
    "vertical.align",
    "margin.bottom",
    "margin.top",
    "margin.left",
    "margin.right",
    "background.color",
    "text.direction",
    "text.align",
    "width",
    "height",
    "hrule", # workaround for some formats
    "border.width.bottom",
    "border.width.top",
    "border.width.left",
    "border.width.right",
    "border.color.bottom",
    "border.color.top",
    "border.color.left",
    "border.color.right",
    "border.style.bottom",
    "border.style.top",
    "border.style.left",
    "border.style.right",
    "rowspan",
    "colspan"
  )
  columns <- intersect(columns, colnames(x))

  dat <- as.data.frame(x)[columns]
  setDT(dat)

  uid <- unique(dat)
  classname <- UUIDgenerate(n = nrow(uid), use.time = TRUE)
  classname <- gsub("(^[[:alnum:]]+)(.*)$", "cl-\\1", classname)
  uid$classname <- classname

  setDF(uid)

  uid
}

# information data -----
fortify_content <- function(
  x,
  default_chunk_fmt,
  ...,
  expand_special_chars = TRUE
) {
  if (isTRUE(expand_special_chars)) {
    x$data[] <- lapply(x$data, expand_special_char, what = "\n", with = "<br>")
    x$data[] <- lapply(x$data, expand_special_char, what = "\t", with = "<tab>")
  }

  row_id <- unlist(mapply(
    function(rows, data) {
      rep(rows, nrow(data))
    },
    rows = rep(seq_len(nrow(x$data)), ncol(x$data)),
    x$data,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  ))

  .col_id <- unlist(mapply(
    function(columns, data) {
      rep(columns, nrow(data))
    },
    columns = rep(x$keys, each = nrow(x$data)),
    x$data,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  ))

  out <- rbindlist(
    apply(x$data, 2, function(col) {
      rbindlist(col, use.names = TRUE, fill = TRUE)
    }),
    use.names = TRUE,
    fill = TRUE
  )
  out$.row_id <- row_id
  out$.col_id <- .col_id
  setDF(out)

  default_props <- text_struct_to_df(
    default_chunk_fmt,
    stringsAsFactors = FALSE
  )
  out <- replace_missing_fptext_by_default(out, default_props)

  out$.col_id <- factor(out$.col_id, levels = default_chunk_fmt$color$keys)
  out <- out[order(out$.col_id, out$.row_id, out$.chunk_index), ]
  out
}

information_data_default_chunk <- function(x) {
  dat <- list()
  if (nrow_part(x, "header") > 0) {
    dat$header <- text_struct_to_df(x$header$styles[["text"]])
  }
  if (nrow_part(x, "body") > 0) {
    dat$body <- text_struct_to_df(x$body$styles[["text"]])
  }
  if (nrow_part(x, "footer") > 0) {
    dat$footer <- text_struct_to_df(x$footer$styles[["text"]])
  }
  dat <- rbindlist(dat, use.names = TRUE, idcol = ".part")

  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  dat$.col_id <- factor(dat$.col_id, levels = x$col_keys)
  setorderv(dat, cols = c(".part", ".row_id", ".col_id"))
  setcolorder(dat, neworder = c(".part", ".row_id", ".col_id"))

  setDF(dat)

  dat
}

#' @importFrom data.table rbindlist setDF setcolorder
#' @export
#' @title Get chunk-level content information from a flextable
#' @description
#' This function takes a flextable object and returns a data.frame containing
#' information about each text chunk within the flextable. The data.frame includes
#' details such as the text content, formatting properties, position within the
#' paragraph, paragraph row, and column.
#' @inheritParams args_x_only
#' @section don't use this:
#'
#' These data structures should not be used, as they
#' represent an interpretation of the underlying data
#' structures, which may evolve over time.
#'
#' **They are exported to enable two packages that exploit
#' these structures to make a transition, and should not
#' remain available for long.**
#'
#' @return a data.frame containing information about chunks:
#'
#' - text chunk (column `txt`) and other content (`url`
#' for the linked url, `eq_data` for content of type 'equation',
#' `word_field_data` for content of type 'word_field' and
#' `img_data` for content of type 'image'),
#' - formatting properties,
#' - part (`.part`), position within the paragraph (`.chunk_index`),
#' row (`.row_id`) and column (`.col_id`).
#' @keywords internal
#' @examples
#' ft <- as_flextable(iris)
#' x <- information_data_chunk(ft)
#' head(x)
#' @family table data extraction
information_data_chunk <- function(x, expand_special_chars = TRUE) {
  dat <- list()
  if (nrow_part(x, "header") > 0) {
    dat$header <- fortify_content(
      x$header$content,
      default_chunk_fmt = x$header$styles$text,
      expand_special_chars = expand_special_chars
    )
  }
  if (nrow_part(x, "body") > 0) {
    dat$body <- fortify_content(
      x$body$content,
      default_chunk_fmt = x$body$styles$text,
      expand_special_chars = expand_special_chars
    )
  }
  if (nrow_part(x, "footer") > 0) {
    dat$footer <- fortify_content(
      x$footer$content,
      default_chunk_fmt = x$footer$styles$text,
      expand_special_chars = expand_special_chars
    )
  }
  dat <- rbindlist(dat, use.names = TRUE, fill = TRUE, idcol = ".part")

  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  dat$.col_id <- factor(dat$.col_id, levels = x$col_keys)
  setorderv(dat, cols = c(".part", ".row_id", ".col_id"))
  setcolorder(dat, neworder = c(".part", ".row_id", ".col_id", ".chunk_index"))

  setDF(dat)
  dat
}

# Cheap detection of equation content.
#
# `information_data_chunk()` is expensive (it runs `fortify_content()` over
# every part), so it should not be called just to find out whether a table
# contains any equation - equations are rare and the answer is almost always
# FALSE. Equations are stored by `as_equation()` as an `eq_data` column in the
# per-cell chunk data.frames; a cell without an equation does not even carry
# that column. This probe scans the raw content structures directly and
# short-circuits on the first equation found, avoiding the full pipeline.
has_equation <- function(x) {
  for (part in c("header", "body", "footer")) {
    cnt <- x[[part]]$content
    if (is.null(cnt) || nrow(cnt$data) < 1) {
      next
    }
    for (cell in cnt$data) {
      eq <- cell[["eq_data"]]
      if (!is.null(eq) && !all(is.na(eq))) {
        return(TRUE)
      }
    }
  }
  FALSE
}

# Cheap detection of image content, same rationale as `has_equation()`.
#
# In the raw chunk data.frames `img_data` is either a list column (raster
# objects, file paths - `gg_chunk()`/`plot_chunk()` render to PNG at chunk
# creation) or a plain character column of file paths; an empty slot is NULL
# or a scalar NA. The predicate matches `runs_types()$is_raster`.
has_raster <- function(x) {
  for (part in c("header", "body", "footer")) {
    cnt <- x[[part]]$content
    if (is.null(cnt) || nrow(cnt$data) < 1) {
      next
    }
    for (cell in cnt$data) {
      img <- cell[["img_data"]]
      if (is.null(img)) {
        next
      }
      if (!is.list(img)) {
        img <- as.list(img)
      }
      for (z in img) {
        if (inherits(z, "raster") || (is.character(z) && any(!is.na(z)))) {
          return(TRUE)
        }
      }
    }
  }
  FALSE
}

# Cheap detection of hyperlink content, same rationale as `has_equation()`.
has_hlink <- function(x) {
  for (part in c("header", "body", "footer")) {
    cnt <- x[[part]]$content
    if (is.null(cnt) || nrow(cnt$data) < 1) {
      next
    }
    for (cell in cnt$data) {
      url <- cell[["url"]]
      if (!is.null(url) && !all(is.na(url))) {
        return(TRUE)
      }
    }
  }
  FALSE
}


#' @importFrom data.table rbindlist setDF
#' @title Get paragraph-level information from a flextable
#' @description
#' This function takes a flextable object and returns a data.frame containing
#' information about each paragraph within the flextable. The data.frame includes
#' details about formatting properties and position within the
#' row and column.
#' @inheritParams args_x_only
#' @section don't use this:
#'
#' These data structures should not be used, as they
#' represent an interpretation of the underlying data
#' structures, which may evolve over time.
#'
#' **They are exported to enable two packages that exploit
#' these structures to make a transition, and should not
#' remain available for long.**
#'
#' @return a data.frame containing information about paragraphs:
#'
#' - formatting properties,
#' - part (`.part`), row (`.row_id`) and column (`.col_id`).
#' @keywords internal
#' @examples
#' ft <- as_flextable(iris)
#' x <- information_data_paragraph(ft)
#' head(x)
#' @export
#' @family table data extraction
information_data_paragraph <- function(x) {
  dat <- list()
  if (nrow_part(x, "header") > 0) {
    dat$header <- par_struct_to_df(x$header$styles[["pars"]])
  }
  if (nrow_part(x, "body") > 0) {
    dat$body <- par_struct_to_df(x$body$styles[["pars"]])
  }
  if (nrow_part(x, "footer") > 0) {
    dat$footer <- par_struct_to_df(x$footer$styles[["pars"]])
  }
  dat <- rbindlist(dat, use.names = TRUE, idcol = ".part")

  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  dat$.col_id <- factor(dat$.col_id, levels = x$col_keys)
  setorderv(dat, cols = c(".part", ".row_id", ".col_id"))
  setcolorder(dat, neworder = c(".part", ".row_id", ".col_id"))

  setDF(dat)

  dat
}

#' @title Get cell-level information from a flextable
#' @description
#' This function takes a flextable object and returns a data.frame containing
#' information about each cell within the flextable. The data.frame includes
#' details about formatting properties and position within the
#' row and column.
#' @inheritParams args_x_only
#' @section don't use this:
#'
#' These data structures should not be used, as they
#' represent an interpretation of the underlying data
#' structures, which may evolve over time.
#'
#' **They are exported to enable two packages that exploit
#' these structures to make a transition, and should not
#' remain available for long.**
#'
#' @return a data.frame containing information about cells:
#'
#' - formatting properties,
#' - part (`.part`), row (`.row_id`) and column (`.col_id`).
#' @keywords internal
#' @examples
#' ft <- as_flextable(iris)
#' x <- information_data_cell(ft)
#' head(x)
#' @export
#' @family table data extraction
information_data_cell <- function(x) {
  dat <- list()
  if (nrow_part(x, "header") > 0) {
    dat$header <- cell_struct_to_df(x$header$styles[["cells"]])
  }
  if (nrow_part(x, "body") > 0) {
    dat$body <- cell_struct_to_df(x$body$styles[["cells"]])
  }
  if (nrow_part(x, "footer") > 0) {
    dat$footer <- cell_struct_to_df(x$footer$styles[["cells"]])
  }
  dat <- rbindlist(dat, use.names = TRUE, idcol = ".part")

  dat$.part <- factor(dat$.part, levels = c("header", "body", "footer"))
  dat$.col_id <- factor(dat$.col_id, levels = x$col_keys)
  setorderv(dat, cols = c(".part", ".row_id", ".col_id"))
  setcolorder(dat, neworder = c(".part", ".row_id", ".col_id"))

  setDF(dat)

  dat
}

test_that("data.frame", {
  dummy_df <- data.frame(
    A = rep(letters[1:3], each = 2),
    B = seq(0, 1, length = 6)
  )
  ft <- as_flextable(dummy_df)
  expect_equal(
    information_data_chunk(ft)$txt,
    c(
      "A",
      "B",
      "character",
      "numeric",
      "a",
      "0.0",
      "a",
      "0.2",
      "b",
      "0.4",
      "b",
      "0.6",
      "c",
      "0.8",
      "c",
      "1.0",
      "n: 6",
      "n: 6"
    )
  )
  ft <- as_flextable(dummy_df[1, ])
  expect_equal(
    information_data_chunk(ft)$txt,
    c("A", "<br>", "character", "a", "B", "<br>", "numeric", "0")
  )
})

test_that("grouped_data", {
  my_CO2 <- CO2
  setDT(my_CO2)
  my_CO2$conc <- as.integer(my_CO2$conc)
  data_co2 <- dcast(
    my_CO2,
    Treatment + conc ~ Type,
    value.var = "uptake",
    fun.aggregate = mean
  )
  expect_silent(
    data_co2 <- as_grouped_data(x = data_co2, groups = c("Treatment"))
  )
  expect_equal(
    data_co2$Treatment[seq_len(2)],
    factor(c("nonchilled", NA), levels = c("nonchilled", "chilled"))
  )
  expect_equal(
    data_co2$Treatment[c(8, 9, 10)],
    factor(c(NA, "chilled", NA), levels = c("nonchilled", "chilled"))
  )

  out_tmp <- data_co2[1, , drop = TRUE]
  expect_equal(attr(out_tmp, "groups"), "Treatment")
  expect_equal(attr(out_tmp, "columns"), c("conc", "Quebec", "Mississippi"))
  expect_equal(unlist(out_tmp, use.names = FALSE), c(1, NA, NA, NA))

  expect_s3_class(data_co2, "grouped_data")

  expect_silent(
    data_co2 <- as_grouped_data(
      x = data_co2,
      groups = c("Treatment"),
      expand_single = TRUE
    )
  )
  expect_true(all(is.na(unlist(
    data_co2[c(12, 13), , drop = TRUE],
    use.names = FALSE
  ))))

  ft <- as_flextable(data_co2)
  expect_equal(
    information_data_chunk(ft)$txt[seq_len(9)],
    c(
      "conc",
      "Quebec",
      "Mississippi",
      "Treatment",
      ": ",
      "nonchilled",
      "",
      "",
      ""
    )
  )
  expect_equal(information_data_chunk(ft)$txt[15], "95")

  ft <- as_flextable(data_co2, hide_grouplabel = TRUE)
  expect_equal(
    information_data_chunk(ft)$txt[seq_len(9)],
    c("conc", "Quebec", "Mississippi", "nonchilled", "", "", "", "", "")
  )
})

test_that("glm and lm", {
  skip_if_not_installed("broom")
  options("show.signif.stars" = TRUE)
  dat <- attitude
  dat$high.rating <- (dat$rating > 70)
  probit.model <- glm(
    high.rating ~ learning + critical + advance,
    data = dat,
    family = binomial(link = "probit")
  )
  expect_silent(ft <- as_flextable(probit.model))

  expect_equal(
    information_data_chunk(ft)$txt[5],
    "Pr(>|z|)"
  )
  expect_equal(
    information_data_chunk(ft)$txt[31],
    "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05"
  )

  lmod <- lm(
    rating ~ complaints + privileges + learning + raises + critical,
    data = attitude
  )
  ft <- as_flextable(lmod)
  expect_equal(
    information_data_chunk(ft)$txt[5],
    "Pr(>|t|)"
  )
  expect_equal(
    information_data_chunk(ft)$txt[44],
    "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05"
  )
  expect_equal(
    information_data_chunk(ft)$txt[72],
    "F-statistic: 12.06 on 24 and 5 DF, p-value: 0.0000"
  )
})

test_that("htest", {
  set.seed(16)
  M <- as.table(rbind(c(762, 327, 468), c(484, 239, 477)))
  dimnames(M) <- list(
    gender = c("F", "M"),
    party = c("Democrat", "Independent", "Republican")
  )
  ft <- as_flextable(stats::chisq.test(M))
  expect_equal(
    information_data_chunk(ft)$txt[6],
    "0.0000"
  )
})

test_that("continuous_summary works", {
  ft_1 <- continuous_summary(
    iris,
    names(iris)[1:4],
    by = "Species",
    hide_grouplabel = FALSE
  )
  expect_identical(
    information_data_chunk(ft_1)$txt[c(1, 11, 14, 71)],
    c("Species", "# na", "Sepal.Length", "setosa")
  )
})

test_that("transformation of mixed models works", {
  skip_if_not_installed("broom.mixed")
  skip_if_not_installed("nlme")
  m1 <- nlme::lme(distance ~ age, data = nlme::Orthodont)
  ft <- as_flextable(m1)
  expect_equal(
    information_data_chunk(ft)$txt[c(18, 108)],
    c("(Intercept)", "Akaike Information Criterion: 454.6")
  )
})

test_that("kmeans works", {
  set.seed(11)
  cl <- kmeans(scale(mtcars[1:7]), 5)
  ft <- as_flextable(cl)
  expect_equal(
    information_data_chunk(ft)$txt[c(37, 163)],
    c("1.0906", "BSS/TSS ratio: 80.1%")
  )
})

test_that("partitioning around medoids works", {
  skip_if_not_installed("cluster")
  set.seed(11)
  dat <- as.data.frame(scale(mtcars[1:7]))
  cl <- cluster::pam(dat, 3)
  ft <- as_flextable(cl)
  expect_equal(
    information_data_chunk(ft)$txt[c(37, 163, 17)],
    c("", NA, "2.2")
  )
})

test_that("grouped data structure", {
  init_flextable_defaults()
  set_flextable_defaults(
    post_process_pptx = function(x) {
      autofit(set_table_properties(x, layout = "fixed"))
    }
  )

  data_co2 <-
    structure(
      list(
        Treatment = structure(
          c(3L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 4L, 4L),
          levels = c("nonchilled", "chilled", "zoubi", "bisou"),
          class = "factor"
        ),
        conc = c(
          85L,
          95L,
          175L,
          250L,
          350L,
          500L,
          675L,
          1000L,
          95L,
          175L,
          250L,
          350L,
          500L,
          675L,
          1000L,
          NA,
          1000L
        ),
        Quebec = c(
          12,
          15.2666666666667,
          30.0333333333333,
          37.4,
          40.3666666666667,
          39.6,
          41.5,
          43.1666666666667,
          12.8666666666667,
          24.1333333333333,
          34.4666666666667,
          35.8,
          36.6666666666667,
          37.5,
          40.8333333333333,
          43,
          43
        ),
        Mississippi = c(
          10,
          11.3,
          20.2,
          27.5333333333333,
          29.9,
          30.6,
          30.5333333333333,
          31.6,
          9.6,
          14.7666666666667,
          16.1,
          16.6,
          16.6333333333333,
          18.2666666666667,
          18.7333333333333,
          19,
          19
        )
      ),
      row.names = c(NA, -17L),
      class = "data.frame"
    )
  gdata <- as_grouped_data(x = data_co2, groups = c("Treatment"))

  ft_1 <- as_flextable(gdata)
  ft_1 <- colformat_double(ft_1, digits = 2)
  ft_1 <- set_table_properties(ft_1, layout = "autofit")

  # pptx testing
  pptx_file <- tempfile(fileext = ".pptx")
  save_as_pptx(ft_1, path = pptx_file)
  doc <- read_pptx(pptx_file)

  xml_body <- doc$slide$get_slide(1)$get()
  xml_tbl <- xml_find_first(
    xml_body,
    "/p:sld/p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl"
  )

  xml_cell_2_1 <- xml_child(xml_tbl, "a:tr[2]/a:tc[1]")
  expect_equal(xml_text(xml_cell_2_1), "Treatment: zoubi")
  expect_equal(xml_attr(xml_cell_2_1, "gridSpan"), "3")
  xml_cell_2_2 <- xml_child(xml_tbl, "a:tr[2]/a:tc[2]")
  expect_equal(xml_text(xml_cell_2_2), "")
  expect_equal(xml_attr(xml_cell_2_2, "hMerge"), "true")
  xml_cell_2_3 <- xml_child(xml_tbl, "a:tr[2]/a:tc[3]")
  expect_equal(xml_text(xml_cell_2_3), "")
  expect_equal(xml_attr(xml_cell_2_3, "hMerge"), "true")

  xml_cell_3_1 <- xml_child(xml_tbl, "a:tr[3]/a:tc[1]")
  expect_equal(xml_text(xml_cell_3_1), "85")
  xml_cell_3_2 <- xml_child(xml_tbl, "a:tr[3]/a:tc[2]")
  expect_equal(xml_text(xml_cell_3_2), "12.00")
  xml_cell_3_3 <- xml_child(xml_tbl, "a:tr[3]/a:tc[3]")
  expect_equal(xml_text(xml_cell_3_3), "10.00")

  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[1]/a:tcPr/a:lnL"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[1]/a:tcPr/a:lnR"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[1]/a:tcPr/a:lnB"), "w"),
    "19050"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[1]/a:tcPr/a:lnT"), "w"),
    "19050"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[2]/a:tcPr/a:lnL"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[2]/a:tcPr/a:lnR"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[2]/a:tcPr/a:lnB"), "w"),
    "19050"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[2]/a:tcPr/a:lnT"), "w"),
    "19050"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[3]/a:tcPr/a:lnL"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[3]/a:tcPr/a:lnR"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[3]/a:tcPr/a:lnB"), "w"),
    "19050"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[1]/a:tc[3]/a:tcPr/a:lnT"), "w"),
    "19050"
  )

  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[2]/a:tc[1]/a:tcPr/a:lnL"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[2]/a:tc[1]/a:tcPr/a:lnR"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[2]/a:tc[1]/a:tcPr/a:lnB"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[2]/a:tc[1]/a:tcPr/a:lnT"), "w"),
    "0"
  )

  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[22]/a:tc[1]/a:tcPr/a:lnL"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[22]/a:tc[1]/a:tcPr/a:lnR"), "w"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[22]/a:tc[1]/a:tcPr/a:lnB"), "w"),
    "19050"
  )
  expect_equal(
    xml_attr(xml_child(xml_tbl, "a:tr[22]/a:tc[1]/a:tcPr/a:lnT"), "w"),
    "0"
  )

  # docx testing
  docx_file <- tempfile(fileext = ".docx")
  save_as_docx(ft_1, path = docx_file)
  doc <- read_docx(docx_file)
  xml_doc <- docx_body_xml(doc)
  xml_tbl <- xml_find_first(xml_doc, "/w:document/w:body/w:tbl")
  xml_cell_2_1 <- xml_child(xml_tbl, "w:tr[2]/w:tc[1]")
  expect_equal(xml_text(xml_cell_2_1), "Treatment: zoubi")
  expect_equal(
    xml_attr(xml_child(xml_cell_2_1, "w:tcPr/w:gridSpan"), "val"),
    "3"
  )
  xml_cell_2_2 <- xml_child(xml_tbl, "w:tr[2]/w:tc[2]")
  expect_s3_class(xml_cell_2_2, "xml_missing")

  xml_cell_1_1 <- xml_child(xml_tbl, "w:tr[1]/w:tc[1]")
  expect_equal(xml_text(xml_cell_1_1), "conc")
  xml_cell_1_2 <- xml_child(xml_tbl, "w:tr[1]/w:tc[2]")
  expect_equal(xml_text(xml_cell_1_2), "Quebec")
  xml_cell_1_3 <- xml_child(xml_tbl, "w:tr[1]/w:tc[3]")
  expect_equal(xml_text(xml_cell_1_3), "Mississippi")

  expect_equal(
    xml_attr(xml_child(xml_cell_1_3, "w:tcPr/w:tcBorders/w:bottom"), "sz"),
    "12"
  )
  expect_equal(
    xml_attr(xml_child(xml_cell_1_3, "w:tcPr/w:tcBorders/w:top"), "sz"),
    "12"
  )
  expect_equal(
    xml_attr(xml_child(xml_cell_1_3, "w:tcPr/w:tcBorders/w:left"), "sz"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_cell_1_3, "w:tcPr/w:tcBorders/w:right"), "sz"),
    "0"
  )

  expect_equal(
    xml_attr(xml_child(xml_cell_2_1, "w:tcPr/w:tcBorders/w:bottom"), "sz"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_cell_2_1, "w:tcPr/w:tcBorders/w:top"), "sz"),
    "12"
  )
  expect_equal(
    xml_attr(xml_child(xml_cell_2_1, "w:tcPr/w:tcBorders/w:left"), "sz"),
    "0"
  )
  expect_equal(
    xml_attr(xml_child(xml_cell_2_1, "w:tcPr/w:tcBorders/w:right"), "sz"),
    "0"
  )

  # html testing
  html_file <- tempfile(fileext = ".html")
  save_as_html(ft_1, path = html_file)
  xml_doc <- read_html(html_file)
  xml_tbl <- xml_find_first(xml_doc, "//table")

  xml_cell_2_1 <- xml_child(xml_tbl, "tbody/tr[1]/td[1]")
  expect_equal(xml_text(xml_cell_2_1), "Treatment: zoubi")
  expect_equal(xml_attr(xml_cell_2_1, "colspan"), "3")
  xml_cell_2_2 <- xml_child(xml_tbl, "tbody/tr[1]/td[2]")
  expect_s3_class(xml_cell_2_2, "xml_missing")

  xml_cell_1_1 <- xml_child(xml_tbl, "thead/tr[1]/th[1]")
  expect_equal(xml_text(xml_cell_1_1), "conc")
  xml_cell_1_2 <- xml_child(xml_tbl, "thead/tr[1]/th[2]")
  expect_equal(xml_text(xml_cell_1_2), "Quebec")
  xml_cell_1_3 <- xml_child(xml_tbl, "thead/tr[1]/th[3]")
  expect_equal(xml_text(xml_cell_1_3), "Mississippi")

  init_flextable_defaults()
})

labelled_df <- data.frame(
  region = structure(
    c(1, 2, 1, 9, 2, 3),
    labels = c(north = 1, south = 2, center = 3, missing = 9),
    label = "Region of the respondent"
  ),
  sex = structure(
    c("f", "f", "m", "m", "m", "f"),
    labels = c(female = "f", male = "m"),
    label = "Sex of the respondent"
  ),
  value = 1:6
)

test_that("labelled data", {
  ft <- flextable(labelled_df, use_labels = TRUE)
  expected_txt <- c(
    "Region of the respondent",
    "Sex of the respondent",
    "value",
    "north",
    "female",
    "1",
    "south",
    "female",
    "2",
    "north",
    "male",
    "3",
    "missing",
    "male",
    "4",
    "south",
    "male",
    "5",
    "center",
    "female",
    "6"
  )
  expect_equal(
    information_data_chunk(ft)$txt,
    expected_txt
  )

  ft <- flextable(labelled_df, use_labels = FALSE)
  expected_txt <- c(
    "region",
    "sex",
    "value",
    "1",
    "f",
    "1",
    "2",
    "f",
    "2",
    "1",
    "m",
    "3",
    "9",
    "m",
    "4",
    "2",
    "m",
    "5",
    "3",
    "f",
    "6"
  )
  expect_equal(
    information_data_chunk(ft)$txt,
    expected_txt
  )

  expected_txt <- c(
    "region",
    "sex",
    "sex",
    "sex",
    "region",
    "female",
    "male",
    "Total",
    "north",
    "1",
    "",
    "1",
    "",
    "2",
    "",
    "south",
    "1",
    "",
    "1",
    "",
    "2",
    "",
    "center",
    "1",
    "",
    "",
    "",
    "1",
    "",
    "missing",
    "",
    "",
    "1",
    "",
    "1",
    "",
    "Total",
    "3",
    "",
    "3",
    "",
    "6",
    ""
  )
  ft <- proc_freq(
    labelled_df,
    row = "region",
    col = "sex",
    include.row_percent = FALSE,
    include.column_percent = FALSE,
    include.table_percent = FALSE
  )
  expect_equal(
    information_data_chunk(ft)$txt,
    expected_txt
  )

  expected_txt <- c(
    "",
    "",
    "",
    "Statistic",
    "<br>",
    "(N=6)",
    "Region of the respondent",
    "north",
    "",
    "2 (33.3%)",
    "Region of the respondent",
    "south",
    "",
    "2 (33.3%)",
    "Region of the respondent",
    "center",
    "",
    "1 (16.7%)",
    "Region of the respondent",
    "Missing",
    "",
    "1 (16.7%)",
    "Sex of the respondent",
    "female",
    "",
    "3 (50.0%)",
    "Sex of the respondent",
    "male",
    "",
    "3 (50.0%)",
    "value",
    "Mean (SD)",
    "",
    "3.5 (1.9)",
    "value",
    "Median (IQR)",
    "",
    "3.5 (2.5)",
    "value",
    "Range",
    "",
    "1.0 - 6.0"
  )
  ft <- as_flextable(summarizor(labelled_df))
  expect_equal(
    information_data_chunk(ft)$txt,
    expected_txt
  )
})

test_that("spread_first_col preserves variable labels (#724)", {
  data(cars)
  attr(cars$speed, "label") <- "Speed (km/h)"
  attr(cars$dist, "label") <- "Stopping distance (m)"

  expect_no_warning(
    ft <- as_flextable(summarizor(cars), spread_first_col = TRUE)
  )

  chunks <- information_data_chunk(ft)$txt
  expect_true("Speed (km/h)" %in% chunks)
  expect_true("Stopping distance (m)" %in% chunks)
  expect_false(any(is.na(chunks)))
})

test_that("package tables", {
  skip_if_not_installed("tables")
  require("tables", quietly = TRUE)
  x <- tabular(
    (Factor(gear, "Gears") + 1) *
      ((n = 1) +
        Percent() +
        (RowPct = Percent("row")) +
        (ColPct = Percent("col"))) ~
      (Factor(carb, "Carburetors") + 1) * Format(digits = 1),
    data = mtcars
  )

  ft <- as_flextable(
    x,
    spread_first_col = TRUE,
    row_title = as_paragraph(
      colorize("Gears: ", color = "#666666"),
      colorize(as_b(.row_title), color = "red")
    )
  )
  idc <- information_data_chunk(ft)

  expected_txt <- c(
    "",
    "Carburetors",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "1",
    "2",
    "3",
    "4",
    "6",
    "8",
    "All"
  )
  expect_equal(
    idc[idc$.part %in% "header", ]$txt,
    expected_txt
  )

  expected_txt <- c(
    "Gears: ",
    "3",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "n",
    "3",
    "4",
    "3",
    "5",
    "0",
    "0",
    "15",
    "Percent",
    "9",
    "12",
    "9",
    "16",
    "0",
    "0",
    "47",
    "RowPct",
    "20",
    "27",
    "20",
    "33",
    "0",
    "0",
    "100",
    "ColPct",
    "43",
    "40",
    "100",
    "50",
    "0",
    "0",
    "47"
  )
  expect_equal(
    idc[idc$.part %in% "body" & idc$.row_id < 6, ]$txt,
    expected_txt
  )

  idp <- information_data_paragraph(ft)
  expect_equal(
    idp[idp$.part %in% "header", ]$text.align,
    rep("center", 16)
  )
  expect_equal(
    idp[idp$.part %in% "body" & idp$.row_id < 6 & idp$.row_id > 1, ]$text.align,
    rep("center", 32)
  )
})


test_that("package xtable", {
  skip_if_not_installed("xtable")
  require("xtable", quietly = TRUE)

  tli <- data.frame(
    grade = c(6L, 7L, 5L, 3L, 8L, 5L, 8L, 4L, 6L, 7L),
    sex = factor(c("M", "M", "F", "M", "M", "M", "F", "M", "M", "M")),
    disadvg = factor(c(
      "YES",
      "NO",
      "YES",
      "YES",
      "YES",
      "NO",
      "YES",
      "YES",
      "NO",
      "YES"
    )),
    ethnicty = factor(
      c(
        "HISPANIC",
        "BLACK",
        "HISPANIC",
        "HISPANIC",
        "WHITE",
        "BLACK",
        "HISPANIC",
        "BLACK",
        "WHITE",
        "HISPANIC"
      ),
      levels = c("BLACK", "HISPANIC", "OTHER", "WHITE")
    ),
    tlimth = c(43L, 88L, 34L, 65L, 75L, 74L, 72L, 79L, 88L, 87L)
  )

  tli.table <- xtable(tli)
  align(tli.table) <- rep("r", 6)
  align(tli.table) <- "|r|r|clr|r|"
  ft <- as_flextable(
    tli.table,
    rotate.colnames = TRUE,
    include.rownames = FALSE
  )
  ft <- height(ft, i = 1, part = "header", height = 1)

  idc <- information_data_chunk(ft)

  expected_txt <- c(
    "grade",
    "sex",
    "disadvg",
    "ethnicty",
    "tlimth",
    "6",
    "M",
    "YES",
    "HISPANIC",
    "43",
    "7",
    "M",
    "NO",
    "BLACK",
    "88",
    "5",
    "F",
    "YES",
    "HISPANIC",
    "34",
    "3",
    "M",
    "YES",
    "HISPANIC",
    "65",
    "8",
    "M",
    "YES",
    "WHITE",
    "75",
    "5",
    "M",
    "NO",
    "BLACK",
    "74",
    "8",
    "F",
    "YES",
    "HISPANIC",
    "72",
    "4",
    "M",
    "YES",
    "BLACK",
    "79",
    "6",
    "M",
    "NO",
    "WHITE",
    "88",
    "7",
    "M",
    "YES",
    "HISPANIC",
    "87"
  )
  expect_equal(
    idc$txt,
    expected_txt
  )

  idc <- information_data_cell(ft)
  expect_equal(
    idc$text.direction,
    c(rep("btlr", 5), rep("lrtb", 50))
  )
})

test_that("gam models", {
  skip_if_not_installed("mgcv")
  require("mgcv", quietly = TRUE)

  set.seed(2)
  dat <- gamSim(1, n = 400, dist = "normal", scale = 2)
  b <- gam(y ~ s(x0) + s(x1) + s(x2) + s(x3), data = dat)
  options(show.signif.stars = FALSE)
  ft <- as_flextable(b)
  ft <- delete_part(ft, part = "footer")

  idc <- information_data_chunk(ft)

  expected_txt <- c(
    "Component",
    "Term",
    "Estimate",
    "Std Error",
    "t-value",
    "p-value",
    "A. parametric coefficients",
    "(Intercept)",
    "7.833",
    "0.099",
    "79.303",
    "0.0000",
    "Component",
    "Term",
    "edf",
    "Ref. df",
    "F-value",
    "p-value",
    "B. smooth terms",
    "s(x0)",
    "2.500",
    "3.115",
    "6.921",
    "0.0001",
    "B. smooth terms",
    "s(x1)",
    "2.401",
    "2.984",
    "81.858",
    "0.0000",
    "B. smooth terms",
    "s(x2)",
    "7.698",
    "8.564",
    "88.158",
    "0.0000",
    "B. smooth terms",
    "s(x3)",
    "1.000",
    "1.000",
    "4.343",
    "0.0378"
  )
  expect_equal(
    idc$txt,
    expected_txt
  )

  idp <- information_data_paragraph(ft)
  expect_equal(
    idp$text.align,
    c(
      "left",
      "left",
      "right",
      "right",
      "right",
      "right",
      "left",
      "left",
      "right",
      "right",
      "right",
      "right",
      "left",
      "left",
      "right",
      "right",
      "right",
      "right",
      "left",
      "left",
      "right",
      "right",
      "right",
      "right",
      "left",
      "left",
      "right",
      "right",
      "right",
      "right",
      "left",
      "left",
      "right",
      "right",
      "right",
      "right",
      "left",
      "left",
      "right",
      "right",
      "right",
      "right"
    )
  )
})

# compact_summary ----
test_that("compact_summary with mixed types", {
  df <- data.frame(
    num = c(1.5, 2.3, 3.7),
    int = 1L:3L,
    chr = c("a", "b", "a"),
    fct = factor(c("x", "y", "x")),
    lgl = c(TRUE, FALSE, TRUE),
    stringsAsFactors = FALSE
  )
  cs <- compact_summary(df)
  expect_s3_class(cs, "compact_summary")
  expect_s3_class(cs, "data.frame")
  expect_equal(nrow(cs), 5L)
  expect_equal(cs$col_name, c("num", "int", "chr", "fct", "lgl"))
  expect_equal(
    cs$col_type,
    c("numeric", "integer", "character", "factor", "logical")
  )
  expect_equal(
    cs$data_kind,
    c("numeric", "numeric", "discrete", "discrete", "logical")
  )
})

test_that("compact_summary numeric min max", {
  df <- data.frame(x = c(10, 20, 30))
  cs <- compact_summary(df)
  expect_equal(cs$min_val, 10)
  expect_equal(cs$max_val, 30)
  expect_equal(cs$n, 3L)
})

test_that("compact_summary integer type", {
  df <- data.frame(x = c(5L, 10L, 15L))
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "integer")
  expect_equal(cs$data_kind, "numeric")
  expect_equal(cs$min_val, 5)
  expect_equal(cs$max_val, 15)
})

test_that("compact_summary logical column", {
  df <- data.frame(x = c(TRUE, TRUE, FALSE, TRUE))
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "logical")
  expect_equal(cs$data_kind, "logical")
  expect_equal(cs$n, 4L)
  expect_equal(cs$str_val, "TRUE: 3, FALSE: 1")
})

test_that("compact_summary factor column", {
  df <- data.frame(x = factor(c("a", "b", "c", "a")))
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "factor")
  expect_equal(cs$data_kind, "discrete")
  expect_equal(cs$n, 3L)
  expect_match(cs$str_val, "'a', 'b', 'c'", fixed = TRUE)
})

test_that("compact_summary character column", {
  df <- data.frame(x = c("b", "a", "c", "a"), stringsAsFactors = FALSE)
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "character")
  expect_equal(cs$data_kind, "discrete")
  expect_equal(cs$n, 3L)
  # values should be sorted unique
  expect_match(cs$str_val, "'a', 'b', 'c'", fixed = TRUE)
})

test_that("compact_summary Date column", {
  df <- data.frame(x = as.Date(c("2020-01-01", "2020-06-15", "2020-12-31")))
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "Date")
  expect_equal(cs$data_kind, "range")
  expect_equal(cs$n, 3L)
  expect_match(cs$str_val, "2020-01-01", fixed = TRUE)
  expect_match(cs$str_val, "2020-12-31", fixed = TRUE)
})

test_that("compact_summary POSIXt column", {
  df <- data.frame(
    x = as.POSIXct(c("2020-01-01 10:00:00", "2020-12-31 18:00:00"))
  )
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "POSIXct")
  expect_equal(cs$data_kind, "range")
  expect_equal(cs$n, 2L)
})

test_that("compact_summary difftime column", {
  df <- data.frame(
    x = as.difftime(c(1, 2, 3), units = "hours")
  )
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "hms")
  expect_equal(cs$data_kind, "range")
  expect_equal(cs$n, 3L)
})

test_that("compact_summary NA handling", {
  df <- data.frame(
    x = c(1, NA, 3),
    y = c(NA, NA, NA)
  )
  cs <- compact_summary(df)
  expect_equal(cs$na_count, c(1L, 3L))
  # x: 2 non-NA values

  expect_equal(cs$n[1], 2L)
  expect_equal(cs$min_val[1], 1)
  expect_equal(cs$max_val[1], 3)
  # y: all NA => min_val/max_val remain NA
  expect_equal(cs$n[2], 0L)
  expect_true(is.na(cs$min_val[2]))
  expect_true(is.na(cs$max_val[2]))
})

test_that("compact_summary all NA logical", {
  df <- data.frame(x = c(NA, NA))
  df$x <- as.logical(df$x)
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "logical")
  expect_equal(cs$n, 0L)
  expect_true(is.na(cs$str_val))
})

test_that("compact_summary empty factor", {
  df <- data.frame(x = factor(character(0)))
  cs <- compact_summary(df)
  expect_equal(cs$col_type, "factor")
  expect_equal(cs$n, 0L)
  expect_true(is.na(cs$str_val))
})

test_that("compact_summary max_levels truncates", {
  df <- data.frame(
    x = factor(letters[1:15]),
    stringsAsFactors = FALSE
  )
  cs <- compact_summary(df, max_levels = 5L)
  expect_match(cs$str_val, ", ...", fixed = TRUE)
  # only first 5 should appear
  expect_match(cs$str_val, "'a'", fixed = TRUE)
  expect_match(cs$str_val, "'e'", fixed = TRUE)
  expect_no_match(cs$str_val, "'f'", fixed = TRUE)

  # without truncation
  cs2 <- compact_summary(df, max_levels = 20L)
  expect_no_match(cs2$str_val, "...", fixed = TRUE)
})

test_that("compact_summary show_type and show_na attributes", {
  df <- data.frame(x = 1:3)
  cs <- compact_summary(df, show_type = TRUE, show_na = TRUE)
  expect_true(attr(cs, "show_type"))
  expect_true(attr(cs, "show_na"))

  cs2 <- compact_summary(df)
  expect_false(attr(cs2, "show_type"))
  expect_false(attr(cs2, "show_na"))
})

test_that("compact_summary errors on non data.frame", {
  expect_error(compact_summary(1:10), "data.frame")
  expect_error(compact_summary("text"), "data.frame")
  expect_error(compact_summary(list(a = 1)), "data.frame")
})

test_that("compact_summary unsupported column type returns empty", {
  df <- data.frame(x = 1:3)
  df$z <- list(1:2, 3:4, 5:6)
  cs <- compact_summary(df)
  # list column should be dropped (unsupported)
  expect_equal(nrow(cs), 1L)
  expect_equal(cs$col_name, "x")
})

test_that("compact_summary all unsupported types", {
  df <- data.frame(x = 1)
  df$x <- list(1:2)
  cs <- compact_summary(df)
  expect_equal(nrow(cs), 0L)
  expect_s3_class(cs, "compact_summary")
})

# as_flextable.compact_summary ----
test_that("as_flextable.compact_summary basic", {
  df <- data.frame(
    num = c(1.5, 2.3),
    chr = c("a", "b"),
    stringsAsFactors = FALSE
  )
  cs <- compact_summary(df)
  ft <- as_flextable(cs)
  expect_s3_class(ft, "flextable")
  # default col_keys: col_name, n, Values (no Type, no NA)
  expect_true("col_name" %in% ft$col_keys)
  expect_true("n" %in% ft$col_keys)
  expect_true("Values" %in% ft$col_keys)
  expect_false("col_type" %in% ft$col_keys)
  expect_false("na_count" %in% ft$col_keys)
})

test_that("as_flextable.compact_summary with show_type", {
  cs <- compact_summary(iris, show_type = TRUE)
  ft <- as_flextable(cs)
  expect_true("col_type" %in% ft$col_keys)
  expect_false("na_count" %in% ft$col_keys)
})

test_that("as_flextable.compact_summary with show_na", {
  cs <- compact_summary(iris, show_na = TRUE)
  ft <- as_flextable(cs)
  expect_false("col_type" %in% ft$col_keys)
  expect_true("na_count" %in% ft$col_keys)
})

test_that("as_flextable.compact_summary with both show_type and show_na", {
  cs <- compact_summary(iris, show_type = TRUE, show_na = TRUE)
  ft <- as_flextable(cs)
  expect_true("col_type" %in% ft$col_keys)
  expect_true("na_count" %in% ft$col_keys)
  # check header labels
  hdr <- information_data_chunk(ft)
  hdr_parts <- hdr[hdr$.part == "header", ]
  expect_true("Column" %in% hdr_parts$txt)
  expect_true("Type" %in% hdr_parts$txt)
  expect_true("NA" %in% hdr_parts$txt)
  expect_true("Values" %in% hdr_parts$txt)
})

test_that("as_flextable.compact_summary numeric values rendering", {
  df <- data.frame(x = c(1.123, 99.987))
  cs <- compact_summary(df)
  ft <- as_flextable(cs)
  chunks <- information_data_chunk(ft)
  body_values <- chunks[
    chunks$.part == "body" & chunks$.col_id == "Values",
  ]
  txt <- paste(body_values$txt, collapse = "")
  expect_match(txt, "Min:", fixed = TRUE)
  expect_match(txt, "Max:", fixed = TRUE)
})

test_that("as_flextable.compact_summary discrete values rendering", {
  df <- data.frame(x = c("a", "b"), stringsAsFactors = FALSE)
  cs <- compact_summary(df)
  ft <- as_flextable(cs)
  chunks <- information_data_chunk(ft)
  body_values <- chunks[
    chunks$.part == "body" & chunks$.col_id == "Values",
  ]
  txt <- paste(body_values$txt, collapse = "")
  expect_match(txt, "'a', 'b'", fixed = TRUE)
})

test_that("as_flextable.compact_summary with iris dataset", {
  cs <- compact_summary(iris)
  ft <- as_flextable(cs)
  expect_s3_class(ft, "flextable")
  expect_equal(nrow_part(ft, "body"), 5L)
  expect_equal(nrow_part(ft, "header"), 1L)
})

test_that("as_flextable.compact_summary empty result errors", {
  df <- data.frame(x = 1)
  df$x <- list(1:2)
  cs <- compact_summary(df)
  # all columns unsupported: flextable() cannot handle 0-column result
  expect_error(as_flextable(cs))
})

ft <- flextable(iris[1:5, ])
ft <- footnote(
  x = ft,
  i = 1:3,
  j = 1:3,
  ref_symbols = "a",
  value = as_paragraph("This is footnote one")
)
text_data <- information_data_chunk(ft)
setDT(text_data)


test_that("symbols are inserted at the end", {
  cell_1 <- text_data[
    .row_id %in% 1 & .col_id %in% "Sepal.Length" & .part %in% "body"
  ]
  expect_equal(cell_1$txt, c("5.1", "a"))
  cell_2 <- text_data[
    .row_id %in% 2 & .col_id %in% "Sepal.Width" & .part %in% "body"
  ]
  expect_equal(cell_2$txt, c("3.0", "a"))
  cell_3 <- text_data[
    .row_id %in% 3 & .col_id %in% "Petal.Length" & .part %in% "body"
  ]
  expect_equal(cell_3$txt, c("1.3", "a"))
})


test_that("footnotes are in the footer", {
  cell_1 <- text_data[.part %in% "footer" & .col_id %in% "Sepal.Length"]
  expect_equal(cell_1$txt, c("a", "This is footnote one"))
})


test_that("symbols are not inserted as a rectangular selection", {
  cell_1 <- text_data[
    .row_id %in% 2 & .col_id %in% "Sepal.Length" & .part %in% "body"
  ]
  expect_equal(cell_1$txt, c("4.9"))
})


test_that("footnote with zero matching rows does not error", {
  ft_zero <- flextable(data.frame(a = 1:5))
  expect_silent(
    ft_result <- footnote(
      x = ft_zero,
      i = ~ a > 10,
      j = "a",
      value = as_paragraph("big")
    )
  )
  expect_equal(
    nrow_part(ft_result, "footer"),
    0L
  )
})


test_that("symbol_sep separates multiple footnotes in same cell", {
  ft_sep <- flextable(head(iris))
  ft_sep <- footnote(
    ft_sep,
    i = 1,
    j = 1,
    value = as_paragraph("note one"),
    ref_symbols = "1",
    part = "header"
  )
  ft_sep <- footnote(
    ft_sep,
    i = 1,
    j = 1,
    value = as_paragraph("note two"),
    ref_symbols = "2",
    part = "header",
    symbol_sep = ","
  )
  td <- information_data_chunk(ft_sep)
  setDT(td)
  cell <- td[
    .row_id %in% 1 & .col_id %in% "Sepal.Length" & .part %in% "header"
  ]
  expect_equal(cell$txt, c("Sepal.Length", "1", ",", "2"))
})

test_that("symbol_sep not added on first footnote", {
  ft_sep <- flextable(head(iris))
  ft_sep <- footnote(
    ft_sep,
    i = 1,
    j = 1,
    value = as_paragraph("note one"),
    ref_symbols = "1",
    part = "header",
    symbol_sep = ","
  )
  td <- information_data_chunk(ft_sep)
  setDT(td)
  cell <- td[
    .row_id %in% 1 & .col_id %in% "Sepal.Length" & .part %in% "header"
  ]
  expect_equal(cell$txt, c("Sepal.Length", "1"))
})

test_that("symbol_sep works with three footnotes on same cell", {
  ft_sep <- flextable(head(iris))
  for (k in 1:3) {
    ft_sep <- footnote(
      ft_sep,
      i = 1,
      j = 1,
      value = as_paragraph(paste("note", k)),
      ref_symbols = as.character(k),
      part = "header",
      symbol_sep = ","
    )
  }
  td <- information_data_chunk(ft_sep)
  setDT(td)
  cell <- td[
    .row_id %in% 1 & .col_id %in% "Sepal.Length" & .part %in% "header"
  ]
  expect_equal(
    cell$txt,
    c("Sepal.Length", "1", ",", "2", ",", "3")
  )
})

test_that("symbol_sep does not affect different cells", {
  ft_sep <- flextable(head(iris))
  ft_sep <- footnote(
    ft_sep,
    i = 1,
    j = 1,
    value = as_paragraph("note one"),
    ref_symbols = "a",
    part = "header",
    symbol_sep = ","
  )
  ft_sep <- footnote(
    ft_sep,
    i = 1,
    j = 2,
    value = as_paragraph("note two"),
    ref_symbols = "b",
    part = "header",
    symbol_sep = ","
  )
  td <- information_data_chunk(ft_sep)
  setDT(td)
  cell1 <- td[
    .row_id %in% 1 & .col_id %in% "Sepal.Length" & .part %in% "header"
  ]
  cell2 <- td[
    .row_id %in% 1 & .col_id %in% "Sepal.Width" & .part %in% "header"
  ]
  expect_equal(cell1$txt, c("Sepal.Length", "a"))
  expect_equal(cell2$txt, c("Sepal.Width", "b"))
})


ft <- flextable(iris[1:5, ])
ft <- footnote(
  x = ft,
  i = 1:3,
  j = 1:3,
  ref_symbols = c("a", "b", "c"),
  value = as_paragraph(paste("This is footnote", 1:3))
)
text_data <- information_data_chunk(ft)
setDT(text_data)

test_that("more than a symbol and more than a footnote", {
  cell_1 <- text_data[
    .row_id %in% 1 & .col_id %in% "Sepal.Length" & .part %in% "body"
  ]
  expect_equal(cell_1$txt, c("5.1", "a"))
  cell_2 <- text_data[
    .row_id %in% 2 & .col_id %in% "Sepal.Width" & .part %in% "body"
  ]
  expect_equal(cell_2$txt, c("3.0", "b"))
  cell_3 <- text_data[
    .row_id %in% 3 & .col_id %in% "Petal.Length" & .part %in% "body"
  ]
  expect_equal(cell_3$txt, c("1.3", "c"))

  note_1 <- text_data[
    .row_id %in% 1 & .part %in% "footer" & .col_id %in% "Sepal.Length"
  ]
  expect_equal(note_1$txt, c("a", "This is footnote 1"))
  note_2 <- text_data[
    .row_id %in% 2 & .part %in% "footer" & .col_id %in% "Sepal.Length"
  ]
  expect_equal(note_2$txt, c("b", "This is footnote 2"))
  note_3 <- text_data[
    .row_id %in% 3 & .part %in% "footer" & .col_id %in% "Sepal.Length"
  ]
  expect_equal(note_3$txt, c("c", "This is footnote 3"))
})

#' @export
#' @importFrom rlang eval_tidy enquo quo_name
#' @title Set cell content from paragraph chunks
#' @description Modify flextable displayed values with eventually
#' mixed content paragraphs.
#'
#' Function is handling complex formatting as image insertion with
#' [as_image()], superscript with [as_sup()], formated
#' text with [as_chunk()] and several other *chunk* functions.
#'
#' Function `mk_par` is another name for `compose` as
#' there is an unwanted **conflict with package 'purrr'**.
#'
#' If you only need to add some content at the end
#' or the beginning of paragraphs and keep existing
#' content as it is, functions [append_chunks()] and
#' [prepend_chunks()] should be prefered.
#'
#' @inheritParams args_selectors_with_all
#' @param value a call to function [as_paragraph()].
#' @param use_dot by default `use_dot=FALSE`; if `use_dot=TRUE`,
#' `value` is evaluated within a data.frame augmented of a column named `.`
#' containing the `j`th column.
#' @examples
#' ft_1 <- flextable(head(cars, n = 5), col_keys = c("speed", "dist", "comment"))
#' ft_1 <- mk_par(
#'   x = ft_1, j = "comment",
#'   i = ~ dist > 9,
#'   value = as_paragraph(
#'     colorize(as_i("speed: "), color = "gray"),
#'     as_sup(sprintf("%.0f", speed))
#'   )
#' )
#' ft_1 <- set_table_properties(ft_1, layout = "autofit")
#' ft_1
#'
#' # using `use_dot = TRUE` ----
#' set.seed(8)
#' dat <- iris[sample.int(n = 150, size = 10), ]
#' dat <- dat[order(dat$Species), ]
#'
#'
#' ft_2 <- flextable(dat)
#' ft_2 <- mk_par(ft_2,
#'   j = ~ . - Species,
#'   value = as_paragraph(
#'     minibar(.,
#'       barcol = "white",
#'       height = .1
#'     )
#'   ), use_dot = TRUE
#' )
#' ft_2 <- theme_vader(ft_2)
#' ft_2 <- autofit(ft_2)
#' ft_2
#' @export
#' @family cell_content_composition
#' @seealso [fp_text_default()], [as_chunk()], [as_b()], [as_word_field()], [labelizor()]
compose <- function(
  x,
  i = NULL,
  j = NULL,
  value,
  part = "body",
  use_dot = FALSE
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "compose()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- compose(x = x, i = i, j = j, value = value, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  defused_value <- enquo(value)
  # call_label <- quo_name(defused_value)
  # if(!grepl("as_paragraph", call_label)){
  #   stop("argument `value` is expected to be a call to `as_paragraph()` but the value is: `", call_label, "`")
  # }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  tmp_data <- x[[part]]$dataset[i, , drop = FALSE]
  if (use_dot) {
    for (jcol in x$col_keys[j]) {
      tmp_data$. <- tmp_data[, jcol]
      newcontent <- eval_tidy(defused_value, data = tmp_data)
      newcontent <- as_chunkset_struct(
        l_paragraph = newcontent,
        keys = jcol,
        i = i
      )
      x[[part]]$content <- set_chunkset_struct_element(
        x = x[[part]]$content,
        i = i,
        j = jcol,
        value = newcontent
      )
    }
  } else {
    newcontent <- eval_tidy(defused_value, data = tmp_data)
    newcontent <- as_chunkset_struct(
      l_paragraph = newcontent,
      keys = x$col_keys[j],
      i = i
    )
    x[[part]]$content <- set_chunkset_struct_element(
      x = x[[part]]$content,
      i = i,
      j = j,
      value = newcontent
    )
  }

  x
}

#' @rdname compose
#' @export
mk_par <- compose


#' @export
#' @title Replace displayed text with labels
#' @description
#' `labelizor()` substitutes text values shown in a flextable
#' with human-readable labels. This is useful to turn column
#' values such as variable names, factor levels or coded strings
#' into presentation-ready wording (e.g. `"Sepal.Length"` to
#' `"Sepal Length"`).
#'
#' `labels` can be either a **named character vector** (names
#' identify values to find, values are the replacement labels)
#' or a **function** applied to every text chunk (e.g. [toupper()]).
#'
#' Only the displayed content is affected; the underlying data
#' of the flextable is unchanged.
#' @inheritParams args_selectors_with_all
#' @param labels a named vector whose names will be used to identify
#' values to replace and values will be used as labels.
#' @seealso [mk_par()], [append_chunks()], [prepend_chunks()]
#' @examples
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' z <- summarizor(
#'   x = CO2[-c(1, 4)],
#'   by = "Treatment",
#'   overall_label = "Overall"
#' )
#'
#' ft_1 <- as_flextable(z, separate_with = "variable")
#'
#' ft_1 <- labelizor(
#'   x = ft_1, j = c("stat"),
#'   labels = c(Missing = "Kouign amann")
#' )
#'
#' ft_1 <- labelizor(
#'   x = ft_1, j = c("variable"),
#'   labels = toupper
#' )
#'
#' ft_1
#' @family cell_content_composition
labelizor <- function(x, j = NULL, labels, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "labelizor()"
    ))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- labelizor(x = x, j = j, labels = labels, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part = part) < 1) {
    return(x)
  }

  if (
    !is.function(labels) && (is.null(names(labels)) || !is.character(labels))
  ) {
    stop("`labels` must be a named character vector or a function.")
  }
  if (is.null(j)) {
    j <- x$col_keys
  }

  j <- as_col_keys(x[[part]], j)

  curr_content_columns <- get_fpstruct_elements(
    x = x[[part]]$content,
    j = j
  )

  if (!is.function(labels)) {
    levs <- names(labels)
    labs <- as.character(labels)
    newcontent <- apply(curr_content_columns, 2, function(x) {
      lapply(x, function(x) {
        x$txt[x$txt %in% levs] <- labs[match(x$txt, levs, nomatch = 0)]
        x
      })
    })
  } else {
    newcontent <- apply(curr_content_columns, 2, function(x) {
      lapply(x, function(x) {
        x$txt <- labels(x$txt)
        x
      })
    })
  }
  newcontent <- as_chunkset_struct(
    l_paragraph = do.call(c, newcontent),
    keys = j
  )

  x[[part]]$content <- set_chunkset_struct_element(
    x = x[[part]]$content,
    j = j,
    value = newcontent
  )

  x
}

#' @export
#' @title Add footnotes to flextable
#' @description The function let add footnotes to a flextable object
#' by adding some symbols in the flextable and associated notes in
#' the footer of the flextable.
#'
#' Symbols are added to the cells designated by the selection `i`
#' and `j`. If you use i = c(1,3) and j = c(2,5), then you will
#' add the symbols (or the repeated symbol) to cells `[1,2]`
#' and `[3,5]`.
#'
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/flextable-016.png" alt="add_footer illustration" style="width:100\%;">
#'
#' See https://www.ardata.fr/en/flextable-gallery/2022-06-23-separate-headers/ for the example
#' shown
#' }}
#' @inheritParams args_selectors_without_all
#' @param value a call to function [as_paragraph()].
#' @param ref_symbols character value, symbols to append that will be used
#' as references to notes.
#' @param inline whether to add footnote on same line as previous footnote or not
#' @param sep used only when inline = TRUE, character string to use as
#' a separator between footnotes.
#' @param symbol_sep separator to insert between multiple
#' footnote symbols in the same cell (e.g. `","` to
#' produce `1,2` instead of `12`). Default is `""`
#' (no separator, backward compatible).
#' @examples
#' ft_1 <- flextable(head(iris))
#' ft_1 <- footnote(ft_1,
#'   i = 1, j = 1:3,
#'   value = as_paragraph(
#'     c(
#'       "This is footnote one",
#'       "This is footnote two",
#'       "This is footnote three"
#'     )
#'   ),
#'   ref_symbols = c("a", "b", "c"),
#'   part = "header"
#' )
#' ft_1 <- valign(ft_1, valign = "bottom", part = "header")
#' ft_1 <- autofit(ft_1)
#'
#' ft_2 <- flextable(head(iris))
#' ft_2 <- autofit(ft_2)
#' ft_2 <- footnote(ft_2,
#'   i = 1, j = 1:2,
#'   value = as_paragraph(
#'     c(
#'       "This is footnote one",
#'       "This is footnote two"
#'     )
#'   ),
#'   ref_symbols = c("a", "b"),
#'   part = "header", inline = TRUE
#' )
#' ft_2 <- footnote(ft_2,
#'   i = 1, j = 3:4,
#'   value = as_paragraph(
#'     c(
#'       "This is footnote three",
#'       "This is footnote four"
#'     )
#'   ),
#'   ref_symbols = c("c", "d"),
#'   part = "header", inline = TRUE
#' )
#' ft_2
#'
#' ft_3 <- flextable(head(iris))
#' ft_3 <- autofit(ft_3)
#' ft_3 <- footnote(
#'   x = ft_3, i = 1:3, j = 1:3,
#'   ref_symbols = "a",
#'   value = as_paragraph("This is footnote one")
#' )
#' ft_3
#' @export
#' @importFrom stats update
#' @family cell_content_composition
footnote <- function(
  x,
  i = NULL,
  j = NULL,
  value,
  ref_symbols = NULL,
  part = "body",
  inline = FALSE,
  sep = "; ",
  symbol_sep = ""
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "footnote()"
    ))
  }
  if (!inherits(value, "paragraph")) {
    stop("`value` must be a paragraph.")
  }

  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(i) < 1) {
    return(x)
  }

  if (is.null(ref_symbols)) {
    symbols_str <- as.character(seq_along(value))
  } else {
    symbols_str <- ref_symbols
  }

  if (any(symbols_str %in% "")) {
    long_msg <- c(
      "Usage of empty symbol '' with footnote should not happen, ",
      "use `add_footer_lines()` instead, it does not require any symbol. "
    )
    long_msg <- paste0(long_msg, collapse = "\n")
    stop(long_msg)
  }

  sep_str <- rep(sep, length(value))

  cell_index <- data.frame(i = i, j = j)
  if (length(symbols_str) == 1) {
    symbols_str <- rep(symbols_str, nrow(cell_index))
  }
  # Assert that either one footnote, or that footnote symbol length matches number
  # of cells to tag
  stopifnot(length(symbols_str) == nrow(cell_index))

  for (index_num in seq_len(nrow(cell_index))) {
    i_cell <- cell_index[["i"]][index_num]
    j_cell <- cell_index[["j"]][index_num]
    if (nzchar(symbol_sep)) {
      cell_para <- get_fpstruct_elements(
        x = x[[part]]$content,
        i = i_cell,
        j = x$col_keys[j_cell]
      )[[1, 1]]
      last_valign <- cell_para$vertical.align[
        nrow(cell_para)
      ]
      if (isTRUE(last_valign == "superscript")) {
        x <- append_chunks(
          x,
          i = i_cell,
          j = j_cell,
          part = part,
          as_sup(symbol_sep)
        )
      }
    }
    x <- append_chunks(
      x,
      i = i_cell,
      j = j_cell,
      part = part,
      as_sup(symbols_str[index_num])
    )
  }

  if (inline) {
    # init a new line
    if (nrow_part(x, "footer") < 1) {
      x <- add_footer_lines(x, values = "")
      n_row <- 1
    } else {
      n_row <- nrow_part(x, "footer")
      target_par <- get_fpstruct_elements(
        x = x[["footer"]]$content,
        i = n_row,
        j = 1
      )[[1, 1]]
    }

    paras <- mapply(
      rbind,
      as_paragraph(as_sup(ref_symbols)),
      value,
      as_paragraph(sep_str),
      SIMPLIFY = FALSE
    )
    paras <- rbind_match_columns(paras)
    paras$.chunk_index <- NULL
    for (i in seq_len(nrow(paras))) {
      x[["footer"]]$content <-
        append_chunkset_struct_element(
          x = x[["footer"]]$content,
          i = n_row,
          j = 1,
          chunk_data = paras[i, ]
        )
    }
  } else {
    # init new lines
    n_row <- nrow_part(x, "footer")
    x <- add_footer_lines(x, values = rep("", length(value)))
    newcontent <- as_chunkset_struct(
      l_paragraph = value,
      keys = x$col_keys[1]
    )
    x[["footer"]]$content <- set_chunkset_struct_element(
      x = x[["footer"]]$content,
      i = n_row + seq_along(value),
      j = 1,
      value = newcontent
    )
    for (v in seq_along(value)) {
      x <- prepend_chunks(
        x = x,
        i = n_row + v,
        j = 1,
        part = "footer",
        as_sup(symbols_str[v])
      )
    }
  }
  x
}

# utils -----
#' @importFrom officer fp_tab fp_tabs fp_par_lite
ooxml_ppr <- function(paragraphs_properties, type = "wml") {
  data_ref_pars <- distinct_paragraphs_properties(paragraphs_properties)

  ## par style wml
  fp_par_xml <- data_ref_pars
  classnames <- data_ref_pars$classname
  fp_par_xml <- split(fp_par_xml, classnames)
  fp_par_xml <- vapply(
    fp_par_xml,
    function(x) {
      zz <- as.list(x)
      zz$border.bottom <- fp_border(
        color = zz$border.color.bottom,
        width = zz$border.width.bottom,
        style = zz$border.style.bottom
      )
      zz$border.top <- fp_border(
        color = zz$border.color.top,
        width = zz$border.width.top,
        style = zz$border.style.top
      )
      zz$border.right <- fp_border(
        color = zz$border.color.right,
        width = zz$border.width.right,
        style = zz$border.style.right
      )
      zz$border.left <- fp_border(
        color = zz$border.color.left,
        width = zz$border.width.left,
        style = zz$border.style.left
      )

      if (!is.na(zz$tabs) && !isFALSE(zz$tabs)) {
        fp_tabs_ <- strsplit(zz$tabs, "&")[[1]]
        fp_tabs_ <- lapply(strsplit(fp_tabs_, "_"), function(x) {
          fp_tab(pos = as.numeric(x[2]), style = x[1])
        })
        zz$tabs <- do.call(fp_tabs, fp_tabs_)
      } else {
        zz$tabs <- NULL
      }

      # delete names not in formals
      zz[grepl(
        pattern = "^(border\\.color|border\\.width|border\\.style)",
        names(zz)
      )] <- NULL
      zz$classname <- NULL

      zz <- do.call(fp_par_lite, zz)
      format(zz, type = type)
    },
    FUN.VALUE = "",
    USE.NAMES = FALSE
  )

  style_dat <- data.frame(
    fp_par_xml = fp_par_xml,
    classname = classnames,
    stringsAsFactors = FALSE
  )

  # organise everything
  setDT(paragraphs_properties)

  paragraphs_properties <- merge(
    paragraphs_properties,
    data_ref_pars,
    by = intersect(colnames(paragraphs_properties), colnames(data_ref_pars))
  )
  paragraphs_properties <- merge(
    paragraphs_properties,
    style_dat,
    by = "classname"
  )
  paragraphs_properties <- paragraphs_properties[,
    .SD,
    .SDcols = c(
      ".part",
      ".row_id",
      ".col_id",
      "fp_par_xml"
    )
  ]
  setDF(paragraphs_properties)
  paragraphs_properties
}

copy_border_bottom_to_next_border_top <- function(x, value) {
  first_partname <-
    if (nrow_part(value, "header")) {
      "header"
    } else if (nrow_part(value, "body")) {
      "body"
    } else {
      "footer"
    }

  if (nrow_part(value, "header")) {
    last_partname <- "header"
    max_n <- nrow_part(value, "header")
  }
  if (nrow_part(value, "body")) {
    last_partname <- "body"
    max_n <- nrow_part(value, "body")
  }
  if (nrow_part(value, "footer")) {
    last_partname <- "footer"
    max_n <- nrow_part(value, "footer")
  }

  x[
    !(x$.part %in% first_partname & x$.row_id %in% 1),
    c("border.width.top", "border.color.top", "border.style.top")
  ] <-
    x[
      !(x$.part %in% last_partname & x$.row_id %in% max_n),
      c("border.width.bottom", "border.color.bottom", "border.style.bottom")
    ]
  x
}
#' @importFrom data.table shift
wml_cells <- function(value, cell_data) {
  cell_heights <- fortify_height(value)
  cell_widths <- fortify_width(value)

  cell_data$width <- NULL # need to get rid of originals that are empty, should probably rm them
  cell_data$height <- NULL
  cell_data <- merge(cell_data, cell_widths, by = ".col_id")
  cell_data <- merge(cell_data, cell_heights, by = c(".part", ".row_id"))

  setDT(cell_data)
  setorderv(cell_data, cols = c(".part", ".row_id", ".col_id"))

  # fix for word horiz. borders, copying the bottom props to top props of the next cell
  cell_data <- copy_border_bottom_to_next_border_top(cell_data, value = value)

  data_ref_cells <- distinct_cells_properties(cell_data)

  ## cell style wml
  fp_cell_wml <- data_ref_cells
  classnames <- data_ref_cells$classname
  fp_cell_wml$classname <- NULL

  fp_cell_wml <- split(fp_cell_wml, classnames)
  fp_cell_wml <- vapply(
    fp_cell_wml,
    function(x) {
      zz <- as.list(x)
      zz$border.bottom <- fp_border(
        color = zz$border.color.bottom,
        width = zz$border.width.bottom,
        style = zz$border.style.bottom
      )
      zz$border.top <- fp_border(
        color = zz$border.color.top,
        width = zz$border.width.top,
        style = zz$border.style.top
      )
      zz$border.right <- fp_border(
        color = zz$border.color.right,
        width = zz$border.width.right,
        style = zz$border.style.right
      )
      zz$border.left <- fp_border(
        color = zz$border.color.left,
        width = zz$border.width.left,
        style = zz$border.style.left
      )

      zz[c(
        "border.width.bottom",
        "border.width.top",
        "border.width.left",
        "border.width.right",
        "border.color.bottom",
        "border.color.top",
        "border.color.left",
        "border.color.right",
        "border.style.bottom",
        "border.style.top",
        "border.style.left",
        "border.style.right",
        "width",
        "height",
        "hrule"
      )] <- NULL
      zz$classname <- NULL
      zz <- do.call(fp_cell, zz)
      zz <- format(zz, type = "wml")
      zz
    },
    FUN.VALUE = "",
    USE.NAMES = FALSE
  )

  style_dat <- data.frame(
    fp_cell_wml = fp_cell_wml,
    classname = classnames,
    stringsAsFactors = FALSE
  )

  # organise everything
  cell_data <- merge(
    cell_data,
    data_ref_cells,
    by = intersect(colnames(cell_data), colnames(data_ref_cells))
  )
  cell_data <- merge(cell_data, style_dat, by = "classname")
  cell_data <- cell_data[,
    .SD,
    .SDcols = c(
      ".part",
      ".row_id",
      ".col_id",
      "fp_cell_wml"
    )
  ]
  setDF(cell_data)
  cell_data
}

default_fp_text_wml <- function(value) {
  default_chunks_properties <- information_data_default_chunk(value)
  unique_text_props <- distinct_text_properties(default_chunks_properties)
  rpr <- sapply(
    split(
      unique_text_props[setdiff(colnames(unique_text_props), "classname")],
      unique_text_props$classname
    ),
    function(x) {
      z <- do.call(officer::fp_text_lite, x)
      format(z, type = "wml")
    }
  )

  unique_text_props$fp_txt_default <- unname(rpr[unique_text_props$classname])
  setDT(default_chunks_properties)

  by_cols <- c(
    "color",
    "font.size",
    "bold",
    "italic",
    "underlined",
    "strike",
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "vertical.align",
    "shading.color"
  )
  by_cols <- intersect(by_cols, colnames(default_chunks_properties))
  by_cols <- intersect(by_cols, colnames(unique_text_props))
  default_chunks_properties <- merge(
    default_chunks_properties,
    unique_text_props,
    by = by_cols
  )
  setDF(default_chunks_properties)
  default_chunks_properties <- default_chunks_properties[, c(
    ".part",
    ".row_id",
    ".col_id",
    "fp_txt_default"
  )]
  default_chunks_properties
}


wml_rows <- function(value, split = FALSE, repeat_headers = TRUE) {
  # prepare cells formatting properties and add span information
  cell_attributes <- information_data_cell(value)
  span_data <- fortify_span(value)
  setDT(cell_attributes)
  cell_attributes <- merge(
    cell_attributes,
    span_data,
    by = c(".part", ".row_id", ".col_id")
  )
  setDF(cell_attributes)

  # prepare paragraphs formatting properties
  paragraphs_properties <- information_data_paragraph(value)

  # transform alignments for rotated text
  # and add them back to paragraphs_properties and cell_attributes
  new_pos <- ooxml_rotation_alignments(
    rotation = cell_attributes$text.direction,
    valign = cell_attributes$vertical.align,
    align = paragraphs_properties$text.align
  )
  paragraphs_properties$text.align <- new_pos$align
  cell_attributes$vertical.align <- new_pos$valign

  # get runs in wml format and get hyperlinks and images information
  run_data <- runs_as_wml(value, txt_data = information_data_chunk(value))

  cell_data <- wml_cells(value, cell_attributes)
  cell_heights <- fortify_height(value)
  cell_hrule <- fortify_hrule(value)

  setDT(cell_data)

  default_chunks_properties <- default_fp_text_wml(value)
  tab_data <- merge(
    cell_data,
    ooxml_ppr(paragraphs_properties),
    by = c(".part", ".row_id", ".col_id")
  )
  tab_data <- merge(
    tab_data,
    default_chunks_properties,
    by = c(".part", ".row_id", ".col_id")
  )
  tab_data[,
    c("fp_par_xml", "fp_txt_default") := list(
      paste0(
        substr(.SD$fp_par_xml, 1, nchar(.SD$fp_par_xml) - 8),
        .SD$fp_txt_default,
        "</w:pPr>"
      ),
      NULL
    )
  ]
  tab_data <- merge(tab_data, run_data, by = c(".part", ".row_id", ".col_id"))
  tab_data <- merge(tab_data, span_data, by = c(".part", ".row_id", ".col_id"))
  setorderv(tab_data, cols = c(".part", ".row_id", ".col_id"))

  tab_data[,
    c("wml", "fp_par_xml", "run_openxml") := list(
      paste0("<w:p>", .SD$fp_par_xml, .SD$run_openxml, "</w:p>"),
      NULL,
      NULL
    )
  ]

  tab_data[
    tab_data$colspan < 1,
    c("wml") := list(
      gsub("<w:r\\b[^<]*>[^<]*(?:<[^<]*)*</w:r>", "", .SD$wml)
    )
  ]

  tab_data[,
    c("wml") := list(
      paste0(
        "<w:tc>",
        .SD$fp_cell_wml,
        .SD$wml,
        "</w:tc>"
      )
    )
  ]
  tab_data[tab_data$rowspan < 1, c("wml") := list("")]

  cells <- dcast(
    data = tab_data,
    formula = .part + .row_id ~ .col_id,
    drop = TRUE,
    fill = "",
    value.var = "wml",
    fun.aggregate = I
  )
  setDF(cells)
  col_ids <- setdiff(colnames(cells), c(".part", ".row_id"))
  wml <- do.call(paste0, cells[col_ids])

  split_str <- ""
  if (!split) {
    split_str <- "<w:cantSplit/>"
  }

  hrule <- cell_hrule$hrule
  hrule[hrule %in% "atleast"] <- "atLeast"

  header_str <- rep("", nrow(cell_hrule))
  if (repeat_headers) {
    header_str[cell_hrule$.part %in% "header"] <- "<w:tblHeader/>"
  }

  rows <- paste0(
    "<w:tr><w:trPr>",
    split_str,
    "<w:trHeight",
    " w:val=",
    shQuote(round(cell_heights$height * 72 * 20, 0), type = "cmd"),
    " w:hRule=\"",
    hrule,
    "\"/>",
    header_str,
    "</w:trPr>",
    wml,
    "</w:tr>"
  )

  paste0(rows, collapse = "")
}


# docx_str -----
gen_raw_wml <- function(x, ...) {
  align <- x$properties$align

  align <- match.arg(align, c("center", "left", "right"), several.ok = FALSE)
  align <- c("center" = "center", "left" = "start", "right" = "end")[align]
  align <- as.character(align)

  dims <- dim(x)
  widths <- dims$widths

  out <- paste0(
    "<w:tbl xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" ",
    "xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" ",
    "xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" ",
    "xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" ",
    "xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" ",
    "xmlns:pic=\"http://schemas.openxmlformats.org/drawingml/2006/picture\"",
    ">"
  )
  if (x$properties$layout %in% "autofit") {
    pt <- prop_table(
      style = NULL,
      layout = table_layout(type = "autofit"),
      align = align,
      width = table_width(width = x$properties$width, unit = "pct"),
      colwidths = table_colwidths(double(0L)),
      word_title = x$properties$word_title,
      word_description = x$properties$word_description
    )
  } else {
    pt <- prop_table(
      style = NULL,
      layout = table_layout(type = "fixed"),
      align = align,
      width = table_width(
        unit = "in",
        width = sum(widths, na.rm = TRUE)
      ),
      colwidths = table_colwidths(widths),
      word_title = x$properties$word_title,
      word_description = x$properties$word_description
    )
  }

  properties_str <- to_wml(pt, add_ns = FALSE)

  out <- paste0(out, properties_str)

  tab_str <- wml_rows(
    value = x,
    split = x$properties$opts_word$split,
    repeat_headers = x$properties$opts_word$repeat_headers
  )
  out <- paste0(out, tab_str, "</w:tbl>")

  out
}

#' @keywords internal
#' @title flextable old functions
#' @description The function is maintained for compatibility with old codes
#' mades by users but be aware it produces the same exact object than [flextable()].
#' This function should be deprecated then removed in the next versions.
#' @param data dataset
#' @param col_keys columns names/keys to display. If some column names are not in
#' the dataset, they will be added as blank columns by default.
#' @param cwidth,cheight initial width and height to use for cell sizes in inches.
#' @export
regulartable <- function(
  data,
  col_keys = names(data),
  cwidth = .75,
  cheight = .25
) {
  .Deprecated(new = "flextable")
  flextable(
    data = data,
    col_keys = col_keys,
    cwidth = cwidth,
    cheight = cheight
  )
}

# Collection of functions and data pre-processing to help with testing
library(officer)
library(xml2)

# xml related functions --------------------------------------------------------
get_docx_xml <- function(x) {
  if (inherits(x, "flextable")) {
    docx_file <- tempfile(fileext = ".docx")
    doc <- read_docx()
    doc <- body_add_flextable(doc, value = x)
    print(doc, target = docx_file)
    x <- docx_file
  }
  redoc <- read_docx(x)
  xml_child(docx_body_xml(redoc))
}

get_pptx_xml <- function(x) {
  if (inherits(x, "flextable")) {
    pptx_file <- tempfile(fileext = ".pptx")
    doc <- read_pptx()
    doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
    doc <- ph_with(doc, x, location = ph_location_type(type = "body"))
    print(doc, target = pptx_file)
    x <- pptx_file
  }

  redoc <- read_pptx(x)
  slide <- redoc$slide$get_slide(redoc$cursor)
  xml_child(slide$get())
}

get_html_xml <- function(x) {
  if (inherits(x, "flextable")) {
    html_file <- tempfile(fileext = ".html")
    save_as_html(tab, path = html_file)
    x <- html_file
  }
  doc <- read_html(x)
  xml_child(doc, "body")
}
get_pdf_text <- function(x, extract_fun) {
  stopifnot(grepl("\\.pdf$", x))

  doc <- extract_fun(x)
  txtfile <- tempfile()
  cat(paste0(doc, collapse = "\n"), file = txtfile)
  readLines(txtfile)
}

render_rmd <- function(file, rmd_format) {
  unlink(file, force = TRUE)
  sucess <- FALSE
  tryCatch(
    {
      render(rmd_file,
        output_format = rmd_format,
        output_file = pdf_file,
        envir = new.env(),
        quiet = TRUE
      )
      sucess <- TRUE
    },
    warning = function(e) {
    },
    error = function(e) {
    }
  )
  sucess
}

# Getting snapshots in the _snaps folder for local testing if conditions are met
do_manual_msoffice_snapshot_testing <- FALSE
copy_back_new_snapshots <- FALSE # if snapshots are updated can be rewritten back

# Utility function to manually test local snapshots ----------------------------
skip_if_not_local_testing <- function(min_pandoc_version = "2", check_html = FALSE) {
  skip_on_cran() # When doing manual testing, it should be always skipped on CRAN
  skip_on_ci() # msoffice testing can not be done on ci
  skip_if_not(do_manual_msoffice_snapshot_testing)
  local_edition(3, .env = parent.frame()) # Set the local_edition at 3
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  if (!is.null(min_pandoc_version)) { # Can be turned off with NULL
    skip_if_not(rmarkdown::pandoc_version() >= numeric_version(min_pandoc_version))
  }
  if (isTRUE(check_html)) {
    skip_if_not_installed("webshot2")
  }
  invisible(TRUE)
}

handle_manual_snapshots <- function(snapshot_folder, snapshot_name) {
  skip_if_not_installed("withr")
  skip_if_not(do_manual_msoffice_snapshot_testing)

  snapshot_name <- paste0(snapshot_name, ".png")

  # Folder where the snapshots are stored
  main_inst_folder <- system.file("snapshots_for_manual_tests", package = "flextable", mustWork = TRUE)

  snapshot_file <- file.path(main_inst_folder, snapshot_folder, snapshot_name)

  if (!file.exists(snapshot_file)) {
    stop("Following snapshot file not found in {flextable}:", snapshot_file)
  }

  # Construct the path to the _snaps folder
  path_to_snaps <- file.path("_snaps", snapshot_folder)
  if (!dir.exists("_snaps")) {
    dir.create("_snaps")
  }
  if (!dir.exists(path_to_snaps)) {
    dir.create(path_to_snaps)
  }

  # Main copy
  file.copy(snapshot_file, path_to_snaps, overwrite = TRUE)

  # Copying back and cleaning test folder
  withr::defer(
    {
      snap_file <- file.path(path_to_snaps, snapshot_name)
      if (copy_back_new_snapshots) {
        file.copy(snap_file, dirname(snapshot_file), overwrite = TRUE)
      }
      if (file.exists(snap_file)) {
        file.remove(snap_file)
      }
    },
    envir = parent.frame()
  )
}

defer_cleaning_snapshot_directory <- function(snap_folder_test_file) {
  skip_if_not_installed("withr")
  skip_if_not(do_manual_msoffice_snapshot_testing)
  withr::defer({
    last_folder <- file.path("_snaps", snap_folder_test_file)
    files_not_removed_for_error <- list.files(last_folder)
    if (length(files_not_removed_for_error)) {
      lapply(files_not_removed_for_error, file.remove)
    }
    if (dir.exists("_snaps")) {
      unlink("_snaps", recursive = TRUE)
    }
  })
}

init_flextable_defaults()
snap_folder_test_file <- "borders"
defer_cleaning_snapshot_directory(snap_folder_test_file)

set.seed(2)

USUBJID <- sprintf("01-ABC-%04.0f", 1:200)
VISITS <- c("SCREENING 1", "WEEK 2", "MONTH 3")
LBTEST <- c("Albumin", "Sodium")
VISITNUM <- seq_along(VISITS)
LBBLFL <- rep(NA_character_, length(VISITNUM))
LBBLFL[1] <- "Y"

VISIT <- data.frame(VISIT = VISITS, VISITNUM = VISITNUM, LBBLFL = LBBLFL, stringsAsFactors = FALSE)
labdata <- expand.grid(USUBJID = USUBJID, LBTEST = LBTEST, VISITNUM = VISITNUM, stringsAsFactors = FALSE)
setDT(labdata)
labdata <- merge(labdata, VISIT, by = "VISITNUM")
labdata[, c("LBNRIND") := list(sample(x = c("LOW", "NORMAL", "HIGH"), size = .N, replace = TRUE, prob = c(.03, .9, .07)))]
setDF(labdata)

SHIFT_TABLE <- shift_table(
  x = labdata, cn_visit = "VISIT", cn_grade = "LBNRIND", cn_usubjid = "USUBJID",
  cn_lab_cat = "LBTEST", cn_is_baseline = "LBBLFL", baseline_identifier = "Y", grade_levels = c("LOW", "NORMAL", "HIGH")
)

SHIFT_TABLE_VISIT <- attr(SHIFT_TABLE, "VISIT_N")
SHIFT_TABLE$VISIT <- attr(SHIFT_TABLE, "FUN_VISIT")(SHIFT_TABLE$VISIT)
SHIFT_TABLE$BASELINE <- attr(SHIFT_TABLE, "FUN_GRADE")(SHIFT_TABLE$BASELINE)
SHIFT_TABLE$LBNRIND <- attr(SHIFT_TABLE, "FUN_GRADE")(SHIFT_TABLE$LBNRIND)
SHIFT_TABLE_VISIT$VISIT <- attr(SHIFT_TABLE, "FUN_VISIT")(SHIFT_TABLE_VISIT$VISIT)


tab <- tabulator(
  x = SHIFT_TABLE,
  hidden_data = SHIFT_TABLE_VISIT,
  row_compose = list(
    VISIT = as_paragraph(VISIT, "\n(N=", N_VISIT, ")")
  ),
  rows = c("LBTEST", "VISIT", "BASELINE"), columns = c("LBNRIND"),
  `n` = as_paragraph(N),
  `%` = as_paragraph(as_chunk(PCT, formatter = function(z) {
    formatC(z * 100, digits = 1, format = "f", flag = "0", width = 4)
  }))
)

ft_1 <- as_flextable(
  x = tab, separate_with = "VISIT",
  label_rows = c(
    LBTEST = "Lab Test",
    VISIT = "Visit",
    BASELINE = "Reference\nRange\nIndicator"
  )
)
ft_1 <- width(ft_1, j = 3, width = 1)

test_that("pptx, docx, and html borders", {
  skip_if_not_local_testing(check_html = TRUE)

  # pptx borders
  handle_manual_snapshots(snap_folder_test_file, "pptx-borders")
  doconv::expect_snapshot_doc(
    x = save_as_pptx(ft_1, path = tempfile(fileext = ".pptx")),
    name = "pptx-borders", engine = "testthat"
  )

  # docx borders
  handle_manual_snapshots(snap_folder_test_file, "docx-borders")
  doconv::expect_snapshot_doc(
    x = save_as_docx(ft_1, path = tempfile(fileext = ".docx")),
    name = "docx-borders", engine = "testthat"
  )

  # html borders
  handle_manual_snapshots(snap_folder_test_file, "html-borders")
  path <- save_as_html(ft_1, path = tempfile(fileext = ".html"))
  skip_if_not_installed("chromote")
  suppressMessages(is_there_chrome <- chromote::find_chrome())
  skip_if(is.null(is_there_chrome))
  doconv::expect_snapshot_html(name = "html-borders", path, engine = "testthat")
})


rmd_file_0 <- "rmd/borders.Rmd"
if (!file.exists(rmd_file_0)) { # just for dev purpose
  rmd_file_0 <- "tests/testthat/rmd/borders.Rmd"
}
rmd_file <- tempfile(fileext = ".Rmd")
file.copy(rmd_file_0, rmd_file, overwrite = TRUE)

html_file <- gsub("\\.Rmd$", ".html", rmd_file)
docx_file <- gsub("\\.Rmd$", ".docx", rmd_file)
pdf_file <- gsub("\\.Rmd$", ".pdf", rmd_file)
pptx_file <- gsub("\\.Rmd$", ".pptx", rmd_file)

test_that("pdf and office complex borders", {
  skip_if_not_local_testing(min_pandoc_version = "2.7.3")

  # pdf office complex borders
  render(rmd_file,
    output_format = rmarkdown::pdf_document(latex_engine = "xelatex"),
    output_file = pdf_file,
    envir = new.env(),
    quiet = TRUE
  )
  handle_manual_snapshots(snap_folder_test_file, "pdf-complex-borders")
  doconv::expect_snapshot_doc(name = "pdf-complex-borders", pdf_file, engine = "testthat")

  # office complex borders
  render(rmd_file,
    output_format = "word_document",
    output_file = docx_file,
    envir = new.env(),
    quiet = TRUE
  )
  handle_manual_snapshots(snap_folder_test_file, "docx-complex-borders")
  doconv::expect_snapshot_doc(name = "docx-complex-borders", docx_file, engine = "testthat")

  render(rmd_file,
    output_format = "powerpoint_presentation",
    output_file = pptx_file,
    envir = new.env(),
    quiet = TRUE
  )
  handle_manual_snapshots(snap_folder_test_file, "pptx-complex-borders")
  doconv::expect_snapshot_doc(name = "pptx-complex-borders", pptx_file, engine = "testthat")
})

init_flextable_defaults()

# Typst backend for flextable -----------------------------------
# Modeled on R/html_str.R and runs_as_html(); reuses the same format-agnostic
# helpers: information_data_*, fortify_*, runs_types, colcode0, colalpha,
# format_double.
#
# -- small formatting helpers ------------------------------------------------

typst_pt <- function(x, digits = 1) {
  ifelse(
    is.na(x) | x < 0.001,
    "0pt",
    paste0(format_double(x, digits = digits), "pt")
  )
}

# Typst content-mode escaping. One PCRE pass: a character class + backreference
# prefixes every special char with `\` (`\X` is a literal X in Typst). The class
# must also list the markup *triggers* `. - ' " = + /`, else Typst rewrites cell
# text: `--`/`...` to dashes/ellipsis, `-` before a digit to a minus sign,
# quotes to curly quotes, and a leading `= + /` to a heading/enum/term marker
# (`/` even fails to compile). `-` is last in the class so it is not a range.
typst_escape <- function(x) {
  x <- gsub("([]#$*_`<>@~[.'\"=+/\\\\-])", "\\\\\\1", x, perl = TRUE)
  x <- gsub("\t", "#h(1em)", x, fixed = TRUE)
  x
}

# Directory where path-mode images are written. In a knitr/Quarto run this is
# the chunk figure dir, so the emitted relative path resolves at compile time
# (Typst root = doc dir); outside knitr we fall back to a tempdir.
typst_fig_dir <- function() {
  fp <- tryCatch(fig_path("png"), error = function(e) NULL)
  if (is.null(fp)) {
    return(tempdir())
  }
  d <- dirname(fp)
  dir.create(d, showWarnings = FALSE, recursive = TRUE)
  d
}

# `embed = FALSE` (Quarto/knitr): copy the file under the chunk figure dir and
# emit `#image("relative/path", ...)` (path resolves at compile time).
# `embed = TRUE` (standalone): base64-encode the bytes and emit
# `#image(ft-b64-decode("..."), ...)`, decoded inline by the preamble decoder.
img_as_typst <- function(img_data, width, height, alt = "", embed = FALSE) {
  fig_dir <- if (!embed) typst_fig_dir() else NULL
  mapply(
    function(img, w, h) {
      # obtain a file on disk for this image (render raster, or external path)
      if (inherits(img, "raster")) {
        src <- plot_in_png(
          code = {
            op <- par(mar = rep(0, 4))
            plot(img, interpolate = FALSE, asp = NA)
            par(op)
          },
          width = w,
          height = h,
          res = 300,
          units = "in",
          path = tempfile(pattern = "ftimg", fileext = ".png")
        )
      } else if (is.character(img)) {
        src <- img
      } else {
        stop("unknown image format")
      }

      if (embed) {
        # officer::image_to_base64 returns a data URI; strip it to raw base64.
        # No `format:` is emitted: with a `bytes` source Typst auto-detects the
        # format from the decoded data. The MIME subtype is NOT a valid Typst
        # format string (e.g. "jpeg"/"svg+xml" would be rejected at compile).
        duri <- officer::image_to_base64(src)
        b64 <- sub("^data:[^;]+;base64,", "", duri)
        sprintf(
          '#image(ft-b64-decode("%s"), width: %sin, height: %sin)',
          b64,
          format_double(w, digits = 3),
          format_double(h, digits = 3)
        )
      } else {
        file <- file.path(
          fig_dir,
          basename(tempfile(
            pattern = "ftimg",
            fileext = paste0(".", tools::file_ext(src))
          ))
        )
        file.copy(src, file, overwrite = TRUE)
        sprintf(
          '#image("%s", width: %sin, height: %sin)',
          file,
          format_double(w, digits = 3),
          format_double(h, digits = 3)
        )
      }
    },
    img_data,
    width,
    height,
    SIMPLIFY = TRUE,
    USE.NAMES = FALSE
  )
}

typst_halign <- function(text.align) {
  out <- text.align
  out[text.align %in% "justify"] <- "left"
  out[!text.align %in% c("left", "center", "right")] <- "left"
  out
}

typst_valign <- function(vertical.align) {
  out <- rep("horizon", length(vertical.align))
  out[vertical.align %in% "top"] <- "top"
  out[vertical.align %in% "bottom"] <- "bottom"
  out
}

# one stroke side -> 'top: 1pt + rgb("000000")' or 'top: none'
typst_stroke_side <- function(color, width, style, side) {
  has <- !is.na(width) & width > 0.001 & colalpha(color) > 0
  dash <- rep("", length(width))
  dash[style %in% "dashed"] <- ', dash: "dashed"'
  dash[style %in% "dotted"] <- ', dash: "dotted"'
  ifelse(
    has,
    ifelse(
      nzchar(dash),
      sprintf(
        '%s: (paint: rgb("%s"), thickness: %s%s)',
        side,
        colcode0(color),
        typst_pt(width, 2),
        dash
      ),
      sprintf('%s: %s + rgb("%s")', side, typst_pt(width, 2), colcode0(color))
    ),
    sprintf("%s: none", side)
  )
}

# -- text runs ---------------------------------------------------------------

runs_as_typst <- function(
  x,
  chunk_data = information_data_chunk(x),
  image_mode = c("path", "embed")
) {
  image_mode <- match.arg(image_mode)
  order_columns <- intersect(
    colnames(chunk_data),
    c(".part", ".row_id", ".col_id", ".chunk_index")
  )
  by_columns <- intersect(
    colnames(chunk_data),
    c(".part", ".row_id", ".col_id")
  )
  if (length(by_columns) < 1) {
    by_columns <- NULL
  }

  dat <- as.data.table(chunk_data)
  runs_types_lst <- runs_types(dat)

  # build a #text(...) wrapper from the per-chunk text properties
  text_args <- function(d) {
    weight <- ifelse(d$bold, ', weight: "bold"', "")
    style <- ifelse(d$italic, ', style: "italic"', "")
    size <- d$font.size
    has_supsub <- d$vertical.align %in% c("superscript", "subscript")
    size[has_supsub] <- size[has_supsub] * 0.6
    sizearg <- sprintf(", size: %s", typst_pt(size))
    fill <- ifelse(
      colalpha(d$color) > 0,
      sprintf(', fill: rgb("%s")', colcode0(d$color)),
      ""
    )
    font <- sprintf(', font: "%s"', d$font.family)
    inner <- paste0(weight, style, sizearg, fill, font)
    inner <- sub("^, ", "", inner) # drop leading comma
    sprintf("#text(%s)[%s]", inner, d$.body)
  }

  dat[
    runs_types_lst$is_text == TRUE,
    c(".body") := list(typst_escape(.SD$txt))
  ]
  dat[runs_types_lst$is_soft_return == TRUE, c(".body") := list("#linebreak()")]
  dat[runs_types_lst$is_tab == TRUE, c(".body") := list("#h(1em)")]
  # equations: eq_data is LaTeX math -> mitex inline #mi(`...`).
  # Requires `#import "@preview/mitex:0.2.x": *` in the document preamble.
  dat[
    runs_types_lst$is_equation == TRUE,
    c(".body") := list(sprintf("#mi(`%s`)", .SD$eq_data))
  ]
  # Word fields are an OOXML/RTF-native feature; like HTML/LaTeX/PPTX, Typst
  # degrades them to an empty string (see header comment).
  dat[runs_types_lst$is_word_field == TRUE, c(".body") := list("")]
  # as_qmd: emit a base64 marker resolved by the flextable-qmd Lua filter.
  # `/* ... */` is a Typst comment, so it degrades to nothing if the filter
  # is absent; base64 never contains `*`, so `*/` cannot occur inside it.
  dat[
    runs_types_lst$is_qmd == TRUE,
    c(".body") := list(sprintf(
      "/*tblqmd:%s*/",
      officer::as_base64(.SD$qmd_data)
    ))
  ]
  if (any(runs_types_lst$is_raster)) {
    dat[
      runs_types_lst$is_raster == TRUE,
      c(".body") := list(img_as_typst(
        img_data = .SD$img_data,
        width = .SD$width,
        height = .SD$height,
        alt = .SD$alt,
        embed = identical(image_mode, "embed")
      ))
    ]
  }

  # text run wrapping applies to plain text runs only; qmd markers, images,
  # equations, breaks and tabs pass through unwrapped.
  is_textual <- runs_types_lst$is_text
  dat[is_textual, c("txt") := list(text_args(.SD))]
  # leave the non-textual bodies (markers, linebreak, h, ...) as-is
  dat[!is_textual, c("txt") := list(.SD$.body)]

  # decorations that wrap the #text node
  dat[
    is_textual & dat$underlined,
    c("txt") := list(sprintf("#underline[%s]", .SD$txt))
  ]
  dat[
    is_textual & !is.na(dat$strike) & dat$strike,
    c("txt") := list(sprintf("#strike[%s]", .SD$txt))
  ]
  dat[
    is_textual & dat$vertical.align %in% "superscript",
    c("txt") := list(sprintf("#super[%s]", .SD$txt))
  ]
  dat[
    is_textual & dat$vertical.align %in% "subscript",
    c("txt") := list(sprintf("#sub[%s]", .SD$txt))
  ]
  # links wrap whatever was produced
  dat[
    runs_types_lst$is_hlink == TRUE,
    c("txt") := list(sprintf('#link("%s")[%s]', .SD$url, .SD$txt))
  ]

  setorderv(dat, cols = order_columns)
  dat <- dat[,
    list(span_tag = paste0(get("txt"), collapse = "")),
    by = by_columns
  ]
  setorderv(dat, cols = by_columns)
  setDF(dat)
  dat
}

# -- cells -------------------------------------------------------------------

typst_content_strs <- function(x, image_mode = "path") {
  cell_data <- information_data_cell(x)
  setDT(cell_data)
  par_data <- information_data_paragraph(x)
  setDT(par_data)

  txt_data <- runs_as_typst(
    x,
    chunk_data = information_data_chunk(x),
    image_mode = image_mode
  )
  setDT(txt_data)

  # bring paragraph text.align + padding into the cell record
  cell_data <- merge(
    cell_data,
    par_data[, c(
      ".part",
      ".row_id",
      ".col_id",
      "text.align",
      "padding.top",
      "padding.bottom",
      "padding.left",
      "padding.right",
      "first_line",
      "hanging"
    )],
    by = c(".part", ".row_id", ".col_id")
  )

  # first-line/hanging indents (hanging wins, as in officer). Typst
  # `hanging-indent` indents lines 2+, so the Word semantics (all lines
  # at padding.left, first line outdented by `hanging`) translate to
  # inset left = padding.left - hanging plus par(hanging-indent:).
  cell_data[,
    c("hanging_pt", "first_line_pt") := list(
      fifelse(!is.na(.SD$hanging) & .SD$hanging > 0, .SD$hanging, 0),
      fifelse(
        (is.na(.SD$hanging) | .SD$hanging <= 0) &
          !is.na(.SD$first_line) &
          .SD$first_line > 0,
        .SD$first_line,
        0
      )
    )
  ]
  cell_data[,
    c("padding.left") := list(pmax(0, .SD$padding.left - .SD$hanging_pt))
  ]

  cell_data[, c("width", "height", "hrule") := list(NULL, NULL, NULL)]
  cell_data <- merge(cell_data, fortify_width(x), by = ".col_id")
  span_data <- fortify_span(x)
  cell_data <- merge(
    cell_data,
    span_data,
    by = c(".part", ".row_id", ".col_id")
  )

  # build each table.cell(...)[ ... ]
  align <- sprintf(
    "align: %s + %s",
    typst_halign(cell_data$text.align),
    typst_valign(cell_data$vertical.align)
  )

  fill <- ifelse(
    colalpha(cell_data$background.color) > 0,
    sprintf(', fill: rgb("%s")', colcode0(cell_data$background.color)),
    ""
  )

  inset <- sprintf(
    ", inset: (left: %s, right: %s, top: %s, bottom: %s)",
    typst_pt(cell_data$padding.left),
    typst_pt(cell_data$padding.right),
    typst_pt(cell_data$padding.top),
    typst_pt(cell_data$padding.bottom)
  )

  stroke <- sprintf(
    ", stroke: (%s, %s, %s, %s)",
    typst_stroke_side(
      cell_data$border.color.top,
      cell_data$border.width.top,
      cell_data$border.style.top,
      "top"
    ),
    typst_stroke_side(
      cell_data$border.color.bottom,
      cell_data$border.width.bottom,
      cell_data$border.style.bottom,
      "bottom"
    ),
    typst_stroke_side(
      cell_data$border.color.left,
      cell_data$border.width.left,
      cell_data$border.style.left,
      "left"
    ),
    typst_stroke_side(
      cell_data$border.color.right,
      cell_data$border.width.right,
      cell_data$border.style.right,
      "right"
    )
  )

  # flextable naming: `rowspan` = horizontal extent (Typst colspan),
  # `colspan` = vertical extent (Typst rowspan). Mirrors html_str.R.
  colspan_str <- ifelse(
    cell_data$rowspan > 1,
    sprintf(", colspan: %.0f", cell_data$rowspan),
    ""
  )
  rowspan_str <- ifelse(
    cell_data$colspan > 1,
    sprintf(", rowspan: %.0f", cell_data$colspan),
    ""
  )

  cell_data[,
    c("cell_open") := list(paste0(
      "table.cell(",
      align,
      fill,
      inset,
      stroke,
      colspan_str,
      rowspan_str,
      ")"
    ))
  ]
  # covered cells produce nothing (Typst auto-skips spanned positions)
  cell_data[
    cell_data$rowspan < 1 | cell_data$colspan < 1,
    c("cell_open") := list("")
  ]

  dat <- merge(txt_data, cell_data, by = c(".part", ".row_id", ".col_id"))

  # paragraph properties needing an explicit par() constructor: the
  # inline content of a table cell is not a real `par` element, so a
  # `#set par()` rule has no effect on it. `hanging-indent` implements
  # `hanging`; `justify: true` re-enables justification (disabled at
  # table level, see gen_raw_typst) when text.align is "justify".
  # `first_line` is a #h() shift of the first line.
  par_args_indent <- ifelse(
    dat$hanging_pt > 0,
    sprintf("hanging-indent: %s", typst_pt(dat$hanging_pt)),
    ""
  )
  par_args_justify <- ifelse(dat$text.align %in% "justify", "justify: true", "")
  par_args <- ifelse(
    nzchar(par_args_indent) & nzchar(par_args_justify),
    paste0(par_args_justify, ", ", par_args_indent),
    paste0(par_args_justify, par_args_indent)
  )
  span_tag <- ifelse(
    nzchar(par_args),
    sprintf("#par(%s)[%s]", par_args, dat$span_tag),
    dat$span_tag
  )
  span_tag <- ifelse(
    dat$first_line_pt > 0,
    paste0(sprintf("#h(%s)", typst_pt(dat$first_line_pt)), span_tag),
    span_tag
  )

  # rotated text -> wrap the content in #rotate(..., reflow: true)[ ... ] so
  # the cell sizes to the rotated box. tbrl: top->bottom (90deg clockwise);
  # btlr: bottom->top (counter-clockwise). reflow expands row height/col width.
  rot_open <- ifelse(
    dat$text.direction %in% "tbrl",
    "#rotate(90deg, reflow: true)[",
    ifelse(
      dat$text.direction %in% "btlr",
      "#rotate(-90deg, reflow: true)[",
      ""
    )
  )
  content <- ifelse(
    nzchar(rot_open),
    paste0(rot_open, span_tag, "]"),
    span_tag
  )
  dat[,
    c("cell_str") := list(ifelse(
      nzchar(.SD$cell_open),
      sprintf("%s[%s],", .SD$cell_open, content),
      ""
    ))
  ]

  dat[, c(".col_id") := list(factor(.SD$.col_id, levels = x$col_keys))]
  setorderv(dat, c(".part", ".row_id", ".col_id"))

  # assemble rows then parts
  rows <- dat[,
    list(
      row_str = paste0(
        get("cell_str")[nzchar(get("cell_str"))],
        collapse = "\n  "
      )
    ),
    by = c(".part", ".row_id")
  ]
  setorderv(rows, c(".part", ".row_id"))

  build_part <- function(part) {
    sub <- rows[rows$.part == part, ]
    if (nrow(sub) < 1) {
      return("")
    }
    paste0("  ", sub$row_str, collapse = "\n")
  }
  list(
    header = build_part("header"),
    body = build_part("body"),
    footer = build_part("footer"),
    widths = fortify_width(x)
  )
}

# -- top-level entry point ---------------------------------------------------

gen_raw_typst <- function(x, image_mode = "path") {
  codes <- typst_content_strs(x, image_mode = image_mode)

  widths <- codes$widths
  widths <- widths[match(x$col_keys, widths$.col_id), ]
  fixed <- isTRUE(x$properties$layout %in% "fixed")
  if (fixed) {
    cols <- sprintf(
      "(%s)",
      paste0(
        sprintf("%sin", format_double(widths$width, digits = 3)),
        collapse = ", "
      )
    )
  } else {
    cols <- sprintf("(%s)", paste0(rep("auto", nrow(widths)), collapse = ", "))
  }

  parts <- character()
  if (nzchar(codes$header)) {
    parts <- c(parts, sprintf("  table.header(\n%s\n  ),", codes$header))
  }
  if (nzchar(codes$body)) {
    parts <- c(parts, codes$body)
  }
  if (nzchar(codes$footer)) {
    parts <- c(parts, sprintf("  table.footer(\n%s\n  ),", codes$footer))
  }

  # flextable's text.align semantics: "left" is not "justify". Quarto's
  # Typst template sets `par(justify: true)` document-wide, which would
  # justify (and hyphenate) every cell; scope it off around the table.
  # Cells with text.align "justify" opt back in with a per-cell
  # `#par(justify: true)` (see typst_content_strs).
  paste0(
    "#[\n",
    "#set par(justify: false)\n",
    "#table(\n",
    sprintf("  columns: %s,\n", cols),
    "  stroke: none,\n",
    paste0(parts, collapse = "\n"),
    "\n)\n",
    "]\n"
  )
}

# -- standalone document -----------------------------------------------------

# Pure-Typst base64 decoder, injected in the preamble by save_as_typst() when
# images are embedded. No package, so the generated .typ compiles offline.
typst_b64_decoder <- function() {
  paste(
    "#let ft-b64-decode(s) = {",
    '  let alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"',
    "  let lookup = (:)",
    "  let chars = alphabet.clusters()",
    "  for i in range(chars.len()) { lookup.insert(chars.at(i), i) }",
    '  let npad = if s.ends-with("==") { 2 } else if s.ends-with("=") { 1 } else { 0 }',
    '  let cl = s.replace("=", "A").replace("\\n", "").replace("\\r", "").clusters()',
    "  let out = ()",
    "  let i = 0",
    "  while i < cl.len() {",
    "    let a = lookup.at(cl.at(i))",
    "    let b = lookup.at(cl.at(i + 1))",
    "    let c = lookup.at(cl.at(i + 2))",
    "    let d = lookup.at(cl.at(i + 3))",
    "    out.push(a * 4 + calc.quo(b, 16))",
    "    out.push(calc.rem(b, 16) * 16 + calc.quo(c, 4))",
    "    out.push(calc.rem(c, 4) * 64 + d)",
    "    i = i + 4",
    "  }",
    "  if npad > 0 { out = out.slice(0, out.len() - npad) }",
    "  bytes(out)",
    "}",
    sep = "\n"
  )
}

#' @export
#' @title Save flextable objects in a Typst file
#' @description Writes one or more flextables to a standalone Typst
#' (`.typ`) document. Images are embedded as base64 (decoded inline at
#' compile time), so the file is self-contained and compiles offline with
#' `typst compile`. Equations require the `mitex` Typst package, which is
#' fetched from the Typst package registry on first compilation.
#' @param ... flextable objects, possibly named.
#' @param values a list of flextable objects. If provided, `...` is ignored.
#' @param path Typst file to be created.
#' @param page a Typst `#set page(...)` rule used as document setup. The
#' default produces a tightly cropped page fitting the table(s).
#' @return the path to the generated file, invisibly.
#' @examples
#' ft <- flextable(head(iris))
#' tf <- tempfile(fileext = ".typ")
#' save_as_typst(ft, path = tf)
save_as_typst <- function(
  ...,
  values = NULL,
  path,
  page = "#set page(width: auto, height: auto, margin: 12pt)"
) {
  if (is.null(values)) {
    values <- list(...)
  }
  values <- Filter(function(z) inherits(z, "flextable"), values)
  if (length(values) < 1) {
    stop("no flextable object to save")
  }

  bodies <- vapply(
    values,
    function(z) gen_raw_typst(z, image_mode = "embed"),
    NA_character_
  )

  need_eq <- any(vapply(values, has_equation, logical(1)))
  need_img <- any(vapply(values, has_raster, logical(1)))

  preamble <- c(
    if (need_eq) '#import "@preview/mitex:0.2.7": *',
    if (need_img) typst_b64_decoder(),
    page
  )

  writeLines(c(preamble, "", bodies), con = path)
  invisible(path)
}

text_css_styles <- function(x) {
  shading <- rep("background-color:transparent;", nrow(x))
  has_shading <- colalpha(x$shading.color) > 0
  shading[has_shading] <- sprintf(
    "background-color:%s;",
    colcodecss(x$shading.color[has_shading])
  )

  color <- rep("", nrow(x))
  has_color <- colalpha(x$color) > 0
  color[has_color] <- sprintf("color:%s;", colcodecss(x$color[has_color]))

  family <- sprintf("font-family:'%s';", x$font.family)

  vertical.align <- rep("", nrow(x))
  has_vertical_align <- x$vertical.align %in% c("superscript", "subscript")
  vertical.align[x$vertical.align %in% "superscript"] <- "vertical-align:super;"
  vertical.align[x$vertical.align %in% "subscript"] <- "vertical-align:sub;"
  font.size <- sprintf("font-size:%s;", css_pt(x$font.size))
  font.size[has_vertical_align] <- sprintf(
    "font-size:%s;",
    css_pt(x$font.size[has_vertical_align] * .6)
  )

  bold <- rep("font-weight:normal;", nrow(x))
  bold[x$bold] <- "font-weight:bold;"
  italic <- rep("font-style:normal;", nrow(x))
  italic[x$italic] <- "font-style:italic;"
  underline <- rep("text-decoration:none;", nrow(x))
  underline[x$underlined] <- "text-decoration:underline;"
  has_strike <- !is.na(x$strike) & x$strike
  underline[has_strike] <- "text-decoration:line-through;"

  style_column <- paste0(
    family,
    font.size,
    bold,
    italic,
    underline,
    color,
    shading,
    vertical.align
  )

  paste0(".", x$classname, "{", style_column, "}", collapse = "")
}
runs_as_html <- function(x, chunk_data = information_data_chunk(x)) {
  # data can be (1.) from a df computed by information_data_chunk
  # or (2.) a single paragraph from x$caption$value and there will be only .chunk_index
  order_columns <-
    intersect(
      x = colnames(chunk_data),
      y = c(".part", ".row_id", ".col_id", ".chunk_index")
    )

  # prepare the by col for aggregation
  by_columns <- intersect(
    colnames(chunk_data),
    c(".part", ".row_id", ".col_id")
  )
  if (length(by_columns) < 1) {
    by_columns <- NULL
  }

  unique_text_props <- distinct_text_properties(chunk_data)

  spans_dataset <- as.data.table(chunk_data)
  spans_dataset <- merge(
    spans_dataset,
    unique_text_props,
    by = intersect(colnames(unique_text_props), colnames(chunk_data))
  )

  span_style_str <- text_css_styles(unique_text_props)

  runs_types_lst <- runs_types(spans_dataset)
  spans_dataset[
    runs_types_lst$is_text == TRUE,
    c("txt") := list(sprintf(
      "<span class=\"%s\">%s</span>",
      .SD$classname,
      htmlize(.SD$txt)
    ))
  ]
  spans_dataset[
    runs_types_lst$is_raster == TRUE,
    c("txt") := list(img_as_html(
      img_data = .SD$img_data,
      width = .SD$width,
      height = .SD$height,
      alt = .SD$alt
    ))
  ]
  spans_dataset[runs_types_lst$is_word_field == TRUE, c("txt") := list("")]
  spans_dataset[runs_types_lst$is_soft_return == TRUE, c("txt") := list("<br>")]
  spans_dataset[runs_types_lst$is_tab == TRUE, c("txt") := list("&emsp;")]
  spans_dataset[
    runs_types_lst$is_hlink == TRUE,
    c("txt") := list(paste0("<a href=\"", .SD$url, "\">", .SD$txt, "</a>"))
  ]
  spans_dataset[
    runs_types_lst$is_qmd == TRUE,
    c("txt") := list(
      sprintf(
        "<span data-qmd-base64=\"%s\">%s</span>",
        officer::as_base64(.SD$qmd_data),
        htmlize(.SD$txt)
      )
    )
  ]

  if (
    requireNamespace("equatags", quietly = TRUE) &&
      any(runs_types_lst$is_equation)
  ) {
    transform_mathjax <- getFromNamespace("transform_mathjax", "equatags")
    spans_dataset[
      runs_types_lst$is_equation == TRUE,
      c("txt") := list(
        sprintf(
          "<span class=\"%s\">%s</span>",
          .SD$classname,
          transform_mathjax(.SD$eq_data, to = "html")
        )
      )
    ]
  }

  setorderv(spans_dataset, cols = order_columns)
  spans_dataset <- spans_dataset[,
    list(span_tag = paste0(get("txt"), collapse = "")),
    by = by_columns
  ]
  setorderv(spans_dataset, cols = by_columns)

  setDF(spans_dataset)
  attr(spans_dataset, "css") <- span_style_str
  spans_dataset
}
runs_as_text <- function(x, chunk_data = information_data_chunk(x)) {
  # data can be (1.) from a df computed by information_data_chunk
  # or (2.) a single paragraph from x$caption$value and there will be only .chunk_index
  order_columns <-
    intersect(
      x = colnames(chunk_data),
      y = c(".part", ".row_id", ".col_id", ".chunk_index")
    )

  # prepare the by col for aggregation
  by_columns <- intersect(
    colnames(chunk_data),
    c(".part", ".row_id", ".col_id")
  )
  if (length(by_columns) < 1) {
    by_columns <- NULL
  }

  spans_dataset <- as.data.table(chunk_data)

  runs_types_lst <- runs_types(spans_dataset)

  spans_dataset[runs_types_lst$is_word_field == TRUE, c("txt") := list("")]
  spans_dataset[runs_types_lst$is_raster == TRUE, c("txt") := list("")]
  spans_dataset[runs_types_lst$is_soft_return == TRUE, c("txt") := list("\n")]
  spans_dataset[runs_types_lst$is_tab == TRUE, c("txt") := list("\t")]
  spans_dataset[
    which(runs_types_lst$is_equation == TRUE),
    c("txt") := list(paste("$", .SD$eq_data, "$"))
  ]

  setorderv(spans_dataset, cols = order_columns)
  spans_dataset <- spans_dataset[,
    list(span_txt = paste0(get("txt"), collapse = "")),
    by = by_columns
  ]
  setorderv(spans_dataset, cols = by_columns)

  setDF(spans_dataset)
  spans_dataset
}

#' @importFrom data.table setkeyv
runs_as_latex <- function(
  x,
  chunk_data = information_data_chunk(x),
  ls_df = NULL
) {
  txt_data <- chunk_data

  # data can be (1.) from a df computed by information_data_chunk
  # or (2.) a single paragraph from x$caption$value and there will be only .chunk_index
  order_columns <-
    intersect(
      x = colnames(txt_data),
      y = c(".part", ".row_id", ".col_id", ".chunk_index")
    )

  # prepare the by col for aggregation
  by_columns <- intersect(colnames(txt_data), c(".part", ".row_id", ".col_id"))
  if (length(by_columns) < 1) {
    by_columns <- NULL
  }

  setDT(txt_data)

  if (".col_id" %in% colnames(txt_data)) {
    txt_data$.col_id <- factor(txt_data$.col_id, levels = x$col_keys)
  }

  if (
    !is.null(ls_df) &&
      all(c(".part", ".row_id", ".col_id") %in% colnames(txt_data))
  ) {
    txt_data <- merge(txt_data, ls_df, by = c(".part", ".row_id", ".col_id"))
  } else {
    txt_data$line_spacing <- 1.2
  }

  unique_text_props <- distinct_text_properties(
    as.data.frame(txt_data),
    add_columns = "line_spacing"
  )

  txt_data <- merge(
    x = txt_data,
    y = unique_text_props,
    by = intersect(colnames(unique_text_props), colnames(txt_data))
  )

  span_latexstyle_lr <- as_table_latexstyle_lr(unique_text_props)

  dat <- merge(txt_data, span_latexstyle_lr, by = "classname")
  setkeyv(dat, cols = NULL)

  runs_types_lst <- runs_types(dat)

  dat[
    runs_types_lst$is_text == TRUE,
    c("txt") := list(sanitize_latex_str(.SD$txt))
  ]
  dat[
    runs_types_lst$is_qmd == TRUE,
    c("txt") := list(
      sprintf("\\tblqmd{%s}", officer::as_base64(.SD$qmd_data))
    )
  ]
  dat[
    runs_types_lst$is_equation == TRUE,
    c("txt") := list(paste0("$", .SD$eq_data, "$"))
  ]
  dat[
    runs_types_lst$is_hlink & grepl("^#", dat$url),
    c("txt") := list(paste0(
      "\\hyperref[",
      gsub("^#", "", .SD$url),
      "]{",
      .SD$txt,
      "}"
    ))
  ]
  dat[
    runs_types_lst$is_hlink & !grepl("^#", dat$url),
    c("txt") := list(paste0(
      "\\href{",
      sanitize_latex_str(.SD$url),
      "}{",
      .SD$txt,
      "}"
    ))
  ]
  dat[
    runs_types_lst$is_raster == TRUE,
    c("txt", "left", "right") := list(
      img_to_latex(.SD$img_data, .SD$width, .SD$height),
      "",
      ""
    )
  ]
  dat[
    runs_types_lst$is_word_field == TRUE,
    c("left", "right", "txt") := list("", "", "")
  ]
  dat[runs_types_lst$is_soft_return == TRUE, c("txt") := list("\\linebreak ")]
  dat[runs_types_lst$is_tab == TRUE, c("txt") := list("\\quad ")]

  dat[, c("txt") := list(sprintf("%s%s%s", .SD$left, .SD$txt, .SD$right))]

  setorderv(dat, cols = order_columns)
  dat <- dat[, list(txt = paste0(get("txt"), collapse = "")), by = by_columns]
  setorderv(dat, cols = by_columns)

  setDF(dat)
  dat
}
#' @importFrom officer str_encode_to_rtf
runs_as_rtf <- function(x, chunk_data = information_data_chunk(x)) {
  txt_data <- chunk_data

  # data can be (1.) from a df computed by information_data_chunk
  # or (2.) a single paragraph from x$caption$value and there will be only .chunk_index
  order_columns <-
    intersect(
      x = colnames(txt_data),
      y = c(".part", ".row_id", ".col_id", ".chunk_index")
    )

  # prepare the by col for aggregation
  by_columns <- intersect(colnames(txt_data), c(".part", ".row_id", ".col_id"))
  if (length(by_columns) < 1) {
    by_columns <- NULL
  }

  setDT(txt_data)

  if (".col_id" %in% colnames(txt_data)) {
    txt_data$.col_id <- factor(txt_data$.col_id, levels = x$col_keys)
  }

  unique_text_props <- distinct_text_properties(as.data.frame(txt_data))

  txt_data <- merge(
    x = txt_data,
    y = unique_text_props,
    by = intersect(colnames(unique_text_props), colnames(txt_data))
  )
  span_rtfstyle <- vapply(
    split(
      unique_text_props[setdiff(colnames(unique_text_props), "classname")],
      unique_text_props$classname
    ),
    function(x) {
      z <- do.call(officer::fp_text_lite, x)
      format(z, type = "rtf")
    },
    FUN.VALUE = "",
    USE.NAMES = FALSE
  )
  span_rtfstyle <- data.frame(
    rpr_rtf = span_rtfstyle,
    classname = unique_text_props$classname,
    stringsAsFactors = FALSE
  )

  dat <- merge(txt_data, span_rtfstyle, by = "classname")
  setkeyv(dat, cols = NULL)
  runs_types_lst <- runs_types(dat)

  image_fun <- function(img_data, width, height) {
    mapply(
      function(src, w, h) {
        to_rtf(officer::external_img(src = src, width = w, height = h))
      },
      img_data,
      width,
      height,
      SIMPLIFY = TRUE
    )
  }
  dat[runs_types_lst$is_text, c("txt") := list(str_encode_to_rtf(.SD$txt))]
  dat[runs_types_lst$is_qmd, c("txt") := list(str_encode_to_rtf(.SD$txt))]
  dat[
    runs_types_lst$is_raster == TRUE,
    c("txt") := list(image_fun(
      img_data = .SD$img_data,
      width = .SD$width,
      height = .SD$height
    ))
  ]
  dat[runs_types_lst$is_soft_return == TRUE, c("txt") := list("\\line")]
  dat[runs_types_lst$is_tab == TRUE, c("txt") := list("\\tab")]
  dat[runs_types_lst$is_equation == TRUE, c("txt") := list("")]
  dat[
    runs_types_lst$is_word_field == TRUE,
    c("txt") := list(sprintf("{\\field{\\*\\fldinst %s}}", .SD$word_field_data))
  ]

  dat[, c("shadingl", "shadingr") := list("", "")]
  dat[
    !dat$shading.color %in% "transparent",
    c("shadingl", "shadingr") := list(
      paste0("%ftshading:", .SD$shading.color, "%"),
      "\\highlight0"
    )
  ]

  dat[,
    c("txt") := list(sprintf(
      "{%s%s%s%s}",
      .SD$shadingl,
      .SD$rpr_rtf,
      .SD$txt,
      .SD$shadingr
    ))
  ]
  dat[
    runs_types_lst$is_hlink,
    c("txt") := list(sprintf(
      "{\\field{\\*\\fldinst HYPERLINK \"%s\"}{\\fldrslt %s}}",
      .SD$url,
      .SD$txt
    ))
  ]

  setorderv(dat, cols = order_columns)
  dat <- dat[, list(txt = paste0(get("txt"), collapse = "")), by = by_columns]
  setorderv(dat, cols = by_columns)

  setDF(dat)
  dat
}


#' @importFrom officer officer_url_encode
runs_as_wml <- function(x, txt_data = information_data_chunk(x)) {
  # data can be (1.) from a df computed by information_data_chunk
  # or (2.) a single paragraph from x$caption$value and there will be only .chunk_index
  order_columns <-
    intersect(
      x = colnames(txt_data),
      y = c(".part", ".row_id", ".col_id", ".chunk_index")
    )

  # prepare the by col for aggregation
  by_columns <- intersect(colnames(txt_data), c(".part", ".row_id", ".col_id"))
  if (length(by_columns) < 1) {
    by_columns <- NULL
  }

  unique_text_props <- distinct_text_properties(as.data.frame(txt_data))

  rpr <- vapply(
    split(
      unique_text_props[setdiff(colnames(unique_text_props), "classname")],
      unique_text_props$classname
    ),
    function(x) {
      z <- do.call(officer::fp_text_lite, x)
      format(z, type = "wml")
    },
    FUN.VALUE = "",
    USE.NAMES = FALSE
  )

  style_dat <- data.frame(
    rpr = rpr,
    classname = unique_text_props$classname,
    stringsAsFactors = FALSE
  )

  txt_data <- merge(
    as.data.table(txt_data),
    unique_text_props,
    by = colnames(unique_text_props)[-ncol(unique_text_props)]
  )
  txt_data <- merge(txt_data, style_dat, by = "classname")

  runs_types_lst <- runs_types(txt_data)

  txt_data <- add_raster_as_filecolumn(txt_data, runs_types_lst$is_raster)

  txt_data[runs_types_lst$is_raster, c("run_openxml") := list(.SD$img_str)]
  txt_data[
    runs_types_lst$is_text,
    c("run_openxml") := list(
      paste0("<w:t xml:space=\"preserve\">", htmlEscape(.SD$txt), "</w:t>")
    )
  ]
  txt_data[
    runs_types_lst$is_qmd,
    c("run_openxml") := list(
      sprintf("<!--TBLQMD:%s-->", officer::as_base64(.SD$qmd_data))
    )
  ]
  txt_data[runs_types_lst$is_soft_return, c("run_openxml") := list("<w:br/>")]
  txt_data[runs_types_lst$is_tab, c("run_openxml") := list("<w:tab/>")]
  txt_data[
    runs_types_lst$is_text |
      runs_types_lst$is_tab |
      runs_types_lst$is_soft_return,
    c("run_openxml") := list(
      paste0(sprintf("<w:r %s>", base_ns), .SD$rpr, .SD$run_openxml, "</w:r>")
    )
  ]
  txt_data[
    runs_types_lst$is_word_field,
    c("run_openxml") := list(to_wml_word_field(
      .SD$word_field_data,
      pr_txt = .SD$rpr
    ))
  ]
  txt_data[
    runs_types_lst$is_hlink,
    c("run_openxml") := list(paste0(
      "<w:hyperlink r:id=\"",
      officer_url_encode(.SD$url),
      "\">",
      .SD$run_openxml,
      "</w:hyperlink>"
    ))
  ]
  if (
    requireNamespace("equatags", quietly = TRUE) &&
      any(runs_types_lst$is_equation)
  ) {
    transform_mathjax <- getFromNamespace("transform_mathjax", "equatags")
    txt_data[
      runs_types_lst$is_equation,
      c("run_openxml") := list(transform_mathjax(.SD$eq_data, to = "mml"))
    ]
  }
  setorderv(txt_data, cols = order_columns)
  txt_data <- txt_data[,
    lapply(.SD, function(x) paste0(x, collapse = "")),
    by = by_columns,
    .SDcols = "run_openxml"
  ]
  setorderv(txt_data, cols = by_columns)

  setDF(txt_data)
  txt_data
}

runs_as_pml <- function(value) {
  txt_data <- information_data_chunk(value)
  txt_data$.col_id <- factor(txt_data$.col_id, levels = value$col_keys)

  unique_text_props <- distinct_text_properties(as.data.frame(txt_data))

  rpr <- sapply(
    split(
      unique_text_props[setdiff(colnames(unique_text_props), "classname")],
      unique_text_props$classname
    ),
    function(x) {
      z <- do.call(officer::fp_text_lite, x)
      format(z, type = "pml")
    }
  )
  style_dat <- data.frame(
    rpr = rpr,
    classname = unique_text_props$classname,
    stringsAsFactors = FALSE
  )

  setDT(txt_data)
  txt_data <- merge(
    txt_data,
    unique_text_props,
    by = colnames(unique_text_props)[-ncol(unique_text_props)]
  )
  txt_data <- merge(txt_data, style_dat, by = "classname")

  runs_types_lst <- runs_types(txt_data)

  txt_data[,
    c("text_nodes_str") := list(paste0("<a:t>", htmlEscape(.SD$txt), "</a:t>"))
  ]

  txt_data[
    runs_types_lst$is_word_field == TRUE,
    c("text_nodes_str") := list("<a:t></a:t>")
  ]
  txt_data[
    runs_types_lst$is_raster == TRUE,
    c("text_nodes_str") := list("<a:t></a:t>")
  ]
  txt_data[
    runs_types_lst$is_soft_return == TRUE,
    c("text_nodes_str") := list("")
  ]
  txt_data[
    runs_types_lst$is_tab == TRUE,
    c("text_nodes_str") := list("<a:t>\t</a:t>")
  ]

  # manage hlinks
  link_pr <- ifelse(
    runs_types_lst$is_hlink,
    paste0("<a:hlinkClick r:id=\"", officer_url_encode(txt_data$url), "\"/>"),
    ""
  )
  gmatch <- gregexpr(pattern = "</a:rPr>", txt_data$rpr, fixed = TRUE)
  end_tag <- paste0(link_pr, "</a:rPr>")
  regmatches(txt_data$rpr, gmatch) <- as.list(end_tag)

  # manage formula
  if (
    requireNamespace("equatags", quietly = TRUE) &&
      any(runs_types_lst$is_equation)
  ) {
    transform_mathjax <- getFromNamespace("transform_mathjax", "equatags")
    txt_data[
      runs_types_lst$is_equation == TRUE,
      c("text_nodes_str") := list(
        paste0(
          "<mc:AlternateContent xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\"><mc:Choice xmlns:a14=\"http://schemas.microsoft.com/office/drawing/2010/main\" Requires=\"a14\">",
          "<a14:m xmlns:a14=\"http://schemas.microsoft.com/office/drawing/2010/main\"><m:oMathPara xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\"><m:oMathParaPr><m:jc m:val=\"centerGroup\"/></m:oMathParaPr>",
          transform_mathjax(.SD$eq_data, to = "mml"),
          "</m:oMathPara></a14:m>",
          "</mc:Choice></mc:AlternateContent>"
        )
      )
    ]
  }

  txt_data[, c("opening_tag", "closing_tag") := list("<a:r>", "</a:r>")]
  txt_data[
    runs_types_lst$is_equation,
    c("opening_tag", "closing_tag") := list("", "")
  ]
  txt_data[
    runs_types_lst$is_soft_return,
    c("opening_tag", "closing_tag") := list("<a:br>", "</a:br>")
  ]

  txt_data_is_empty <- txt_data[,
    list(is_empty = all(.SD$text_nodes_str %in% "<a:t></a:t>")),
    by = c(".part", ".row_id", ".col_id")
  ]
  txt_data[,
    c("par_nodes_str") := list(
      paste0(.SD$opening_tag, .SD$rpr, .SD$text_nodes_str, .SD$closing_tag)
    )
  ]

  setorderv(txt_data, cols = c(".part", ".row_id", ".col_id", ".chunk_index"))

  txt_data <- txt_data[,
    lapply(
      .SD,
      function(x) {
        paste0(x, collapse = "")
      }
    ),
    by = c(".part", ".row_id", ".col_id"),
    .SDcols = "par_nodes_str"
  ]
  txt_data <- merge(
    txt_data,
    txt_data_is_empty,
    by = c(".part", ".row_id", ".col_id")
  )
  setDF(txt_data)
  txt_data
}

# utils ----
runs_types <- function(dat) {
  x <- list(
    is_soft_return = dat$txt %in% "<br>",
    is_tab = dat$txt %in% "<tab>",
    is_equation = !is.na(dat$eq_data),
    is_word_field = !is.na(dat$word_field_data),
    is_qmd = !is.na(dat$qmd_data),
    is_hlink = !is.na(dat$url),
    is_raster = sapply(dat$img_data, function(x) {
      inherits(x, "raster") || is.character(x)
    })
  )
  x$is_text <- Reduce(
    function(a, b) a & b,
    list(
      !x$is_soft_return,
      !x$is_tab,
      !x$is_equation,
      !x$is_word_field,
      !x$is_qmd,
      !x$is_raster
    )
  )
  x
}

#' @import data.table
add_raster_as_filecolumn <- function(x, is_raster) {
  whichs_ <- which(is_raster)
  alt_vals <- if (!is.null(x$alt)) x$alt[whichs_] else rep("", length(whichs_))
  files <- mapply(
    function(x, width, height, alt) {
      if (inherits(x, "raster")) {
        file <- plot_in_png(
          code = {
            op <- par(mar = rep(0, 4))
            plot(x, interpolate = FALSE, asp = NA)
            par(op)
          },
          width = width,
          height = height,
          res = 300,
          units = "in",
          path = tempfile(fileext = ".png")
        )
      } else if (is.character(x)) {
        file <- x
      } else {
        stop("unknown image format")
      }

      alt <- if (is.na(alt)) "" else alt
      data.frame(
        file = file,
        img_str = wml_image(file, width, height, alt = alt),
        stringsAsFactors = FALSE
      )
    },
    x$img_data[whichs_],
    x$width[whichs_],
    x$height[whichs_],
    alt_vals,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  files <- rbindlist(files)
  setDF(files)

  x$file <- rep(NA_character_, nrow(x))
  x$img_str <- rep(NA_character_, nrow(x))

  x[whichs_, c("file", "img_str")] <- files

  x
}

#' @export
#' @title Transform a 'gam' model into a flextable
#' @description produce a flextable describing a
#' generalized additive model produced by function `mgcv::gam`.
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/flextable-002.png" alt="as_flextable.gam illustration" style="width:100\%;">
#' }}
#' @param x gam model
#' @param ... unused argument
#' @family as_flextable methods
#' @examples
#' if (require("mgcv")) {
#'   set.seed(2)
#'
#'   # Simulated data
#'   dat <- gamSim(1, n = 400, dist = "normal", scale = 2)
#'
#'   # basic GAM model
#'   b <- gam(y ~ s(x0) + s(x1) + s(x2) + s(x3), data = dat)
#'
#'   ft <- as_flextable(b)
#'   ft
#' }
as_flextable.gam <- function(x, ...) {
  data_t <- tidy_gam(x)
  data_g <- glance_gam(x)

  std_border <- fp_border(
    color = flextable_global$defaults$border.color,
    style = "solid",
    ...
  )

  param.head <- c(
    "Component",
    "Term",
    "Estimate",
    "Std Error",
    "t-value",
    "p-value"
  )
  smooth.head <- c("Component", "Term", "edf", "Ref. df", "F-value", "p-value")
  names(param.head) <- names(data_t$parametric)
  names(smooth.head) <- names(data_t$parametric)

  has_pvalue <- if ("p.value" %in% colnames(data_t)) TRUE else FALSE
  show_signif <- !is.null(getOption("show.signif.stars")) &&
    getOption("show.signif.stars")

  ft <- flextable(
    data_t$parametric,
    col_keys = c(
      names(data_t$parametric),
      if (show_signif) "signif"
    )
  )
  ft <- border_remove(ft)
  ft <- set_header_labels(ft, values = param.head)

  if (nrow(data_t$smooth) > 0) {
    ft <- add_body(ft, Component = NA_character_, top = FALSE)
    new_dat <- data_t$smooth
    names(new_dat) <- names(data_t$parametric)
    if (show_signif) {
      new_dat[["signif"]] <- ""
    }
    ft <- add_body(ft, values = new_dat, top = FALSE)
  }

  ft <- colformat_double(ft, j = 3:5, digits = 3)
  ft <- colformat_double(ft, j = 6, digits = 4)
  if (show_signif) {
    ft <- mk_par(ft, j = "signif", value = as_paragraph(pvalue_format(p.value)))
  }

  if (nrow(data_t$smooth) > 0) {
    ft <- mk_par(
      ft,
      i = nrow(data_t$parametric) + 1,
      value = as_paragraph(c(smooth.head, if (show_signif) ""))
    )
    ft <- hline(ft, i = nrow(data_t$parametric) + c(0, 1), border = std_border)
    ft <- bold(ft, i = nrow(data_t$parametric) + 1)
  }

  ft <- hline_bottom(ft, border = std_border)
  ft <- bold(ft, part = "header")
  ft <- hline_top(ft, border = std_border, part = "header")
  ft <- hline(ft, border = std_border, part = "header")
  ft <- merge_v(ft, j = 1)
  ft <- valign(ft, j = 1, valign = "top")
  ft <- align_nottext_col(ft)
  ft <- align_text_col(ft)
  ft <- fix_border_issues(ft)
  if (show_signif) {
    ft <- add_footer_lines(
      ft,
      values = c(
        "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05",
        ""
      )
    )
  }
  ft <- add_footer_lines(
    ft,
    values = c(
      sprintf(
        "Adjusted R-squared: %s, Deviance explained %s",
        format(data_g$adj.r.squared, digits = 3, format = "f", nsmall = 3),
        format(data_g$deviance, digits = 3, format = "f", nsmall = 3)
      ),
      sprintf(
        "%s : %s, Scale est: %s, N: %d",
        data_g$method,
        format(data_g$sp.crit, format = "f", digits = 3, nsmall = 3),
        format(data_g$scale.est, digits = 3, nsmall = 3),
        data_g$nobs
      )
    )
  )
  ft <- align(ft, i = 1, align = "right", part = "footer")
  ft <- autofit(ft, part = c("header", "body"))
  if (show_signif) {
    ft <- width(ft, j = "signif", width = .4)
  }
  ft
}


#' Summarize a(n) gam object
#' @description summarizes information about the components of a model
#' @noRd
tidy_gam <- function(model) {
  ptab <- data.frame(summary(model)$p.table)
  ptab$term <- rownames(ptab)
  rownames(ptab) <- NULL
  ptab$Component <- "A. parametric coefficients"
  ptab <- ptab[, c(6, 5, 1:4)]
  colnames(ptab) <- c(
    "Component",
    "Term",
    "Estimate",
    "Std.Error",
    "t.value",
    "p.value"
  )

  stab <- data.frame(summary(model)$s.table)
  if (nrow(stab) > 0) {
    stab$term <- rownames(stab)
    stab$Component <- "B. smooth terms"
    stab <- stab[, c(6, 5, 1:4)]
    colnames(stab) <- c(
      "Component",
      "Term",
      "edf",
      "Ref. df",
      "F.value",
      "p.value"
    )
    rownames(stab) <- NULL
  }

  list(parametric = ptab, smooth = stab)
}

#' Summarize a(n) gam object
#' @description provides model summaries in one line
#' @noRd
#' @importFrom stats AIC BIC logLik df.residual nobs
glance_gam <- function(model) {
  df <- sum(model$edf)
  if (length(df) < 1) {
    df <- NA_real_
  }
  df.res <- df.residual(model)
  if (length(df.res) < 1) {
    df.res <- NA_real_
  }
  logLik <- as.numeric(logLik(model))
  if (length(logLik) < 1) {
    logLik <- NA_real_
  }
  aic <- AIC(model)
  if (length(aic) < 1) {
    aic <- NA_real_
  }
  bic <- BIC(model)
  if (length(bic) < 1) {
    bic <- NA_real_
  }
  dev <- summary(model)$dev.expl
  if (length(dev) < 1) {
    dev <- NA_real_
  }
  adj.r.squared <- summary(model)$r.sq
  if (length(adj.r.squared) < 1) {
    adj.r.squared <- NA_real_
  }
  scale.est <- summary(model)$scale
  if (length(scale.est) < 1) {
    scale.est <- NA_real_
  }
  sp.criterion <- as.numeric(summary(model)$sp.criterion)
  if (length(sp.criterion) < 1) {
    sp.criterion <- NA_real_
  }

  data.frame(
    df = df,
    df.residual = df.res,
    logLik = logLik,
    AIC = aic,
    BIC = bic,
    adj.r.squared = adj.r.squared,
    deviance = dev,
    nobs = length(model$y),
    method = as.character(summary(model)$method),
    sp.crit = sp.criterion,
    scale.est = scale.est,
    stringsAsFactors = FALSE
  )
}

#' @title Create a flextable from a data frame
#'
#' @description Create a flextable object with function `flextable`.
#'
#' `flextable` are designed to make tabular reporting easier for
#' R users. Functions are available to let you format text, paragraphs and cells;
#' table cells can be merge vertically or horizontally, row headers can easily
#' be defined, rows heights and columns widths can be manually set or automatically
#' computed.
#'
#' If working with 'R Markdown' documents, you should read about knitr
#' chunk options in [knit_print.flextable()] and about setting
#' default values with [set_flextable_defaults()].
#'
#' @section Reuse frequently used parameters:
#'
#' Some default formatting properties are automatically
#' applied to every flextable you produce.
#'
#' It is highly recommended to use this function because
#' its use will minimize the code. For example, instead of
#' calling the `fontsize()` function over and over again for
#' each new flextable, set the font size default value by
#' calling (before creating the flextables)
#' `set_flextable_defaults(font.size = 11)`. This is also
#' a simple way to have homogeneous arrays and make the
#' documents containing them easier to read.
#'
#' You can change these default values with function
#' [set_flextable_defaults()]. You can reset them
#' with function [init_flextable_defaults()]. You
#' can access these values by calling [get_flextable_defaults()].
#'
#' @section new lines and tabulations:
#'
#' The 'flextable' package will translate for you
#' the new lines expressed in the form `\n` and
#' the tabs expressed in the form `\t`.
#'
#' The new lines will be transformed into "soft-return",
#' that is to say a simple carriage return and not a
#' new paragraph.
#'
#' Tabs are different depending on the output format:
#'
#' - HTML is using entity *em space*
#' - Word - a Word 'tab' element
#' - PowerPoint - a PowerPoint 'tab' element
#' - latex - tag "\\quad "
#' - Typst - horizontal space `#h(1em)`
#' @section flextable parts:
#'
#' A `flextable` is made of 3 parts: header, body and footer.
#'
#' Most functions have an argument named `part` that will be used
#' to specify what part of of the table should be modified.
#' @param data dataset
#' @param col_keys columns names/keys to display. If some column names are not in
#' the dataset, they will be added as blank columns by default.
#' @param cwidth,cheight initial width and height to use for cell sizes in inches.
#' @param defaults,theme_fun deprecated, use [set_flextable_defaults()] instead.
#' @param use_labels Logical; if TRUE, any column labels or value labels
#' present in the dataset will be used for display purposes. Defaults
#' to TRUE.
#' @examples
#' ft <- flextable(head(mtcars))
#' ft
#' @export
#' @importFrom stats setNames
#' @seealso [style()], [autofit()], [theme_booktabs()], [knit_print.flextable()],
#' [compose()], [footnote()], [set_caption()]
#' @family flextable_constructors
flextable <- function(
  data,
  col_keys = names(data),
  cwidth = .75,
  cheight = .25,
  defaults = list(),
  theme_fun = theme_booktabs,
  use_labels = TRUE
) {
  stopifnot(is.data.frame(data), ncol(data) > 0)
  if (any(duplicated(col_keys))) {
    stop(sprintf(
      "duplicated col_keys: %s",
      paste0(unique(col_keys[duplicated(col_keys)]), collapse = ", ")
    ))
  }
  list_lbls <- collect_labels(dataset = data, use_labels = use_labels)

  if (
    inherits(data, "data.table") ||
      inherits(data, "tbl_df") ||
      inherits(data, "tbl")
  ) {
    data <- as.data.frame(data, stringsAsFactors = FALSE)
  }

  blanks <- setdiff(col_keys, names(data))
  if (length(blanks) > 0) {
    blanks_col <- lapply(blanks, function(x, n) character(n), nrow(data))
    blanks_col <- setNames(blanks_col, blanks)
    data[blanks] <- blanks_col
  }

  body <- complex_tabpart(
    data = data,
    col_keys = col_keys,
    cwidth = cwidth,
    cheight = cheight
  )

  # header
  header_data <- setNames(as.list(col_keys), col_keys)
  header_data[blanks] <- as.list(rep("", length(blanks)))
  header_data <- as.data.frame(
    header_data,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  header <- complex_tabpart(
    data = header_data,
    col_keys = col_keys,
    cwidth = cwidth,
    cheight = cheight
  )

  # footer
  footer_data <- header_data[FALSE, , drop = FALSE]
  footer <- complex_tabpart(
    data = footer_data,
    col_keys = col_keys,
    cwidth = cwidth,
    cheight = cheight
  )

  out <- list(
    header = header,
    body = body,
    footer = footer,
    col_keys = col_keys,
    caption = list(value = NULL),
    blanks = blanks
  )
  class(out) <- c("flextable")

  out <- do.call(flextable_global$defaults$theme_fun, list(out))
  out <- set_table_properties(
    x = out,
    layout = flextable_global$defaults$table.layout
  )

  if (length(list_lbls$variables_labels) > 0) {
    out <- labelizor(
      out,
      labels = unlist(list_lbls$variables_labels),
      part = "header"
    )
  }

  apply_labels(out, collected_labels = list_lbls)
}

#' @export
#' @rdname flextable
#' @section qflextable:
#' `qflextable` is a convenient tool to produce quickly
#' a flextable for reporting where layout is fixed (see
#' [set_table_properties()]) and columns
#' widths are adjusted with [autofit()].
qflextable <- function(data) {
  ft <- flextable(data)
  ft <- set_table_properties(ft, layout = "fixed")
  autofit(ft)
}

#' @export
#' @title Set flextable caption
#' @description Set caption value in a flextable. The function
#' can also be used to define formattings that will be applied
#' if possible to Word and HTML outputs.
#'
#' * The caption will be associated with a paragraph style when
#' the output is Word. It can also be numbered as a auto-numbered
#' Word computed value.
#' * The PowerPoint format ignores captions. PowerPoint documents are not
#' structured and do not behave as HTML documents and paginated documents
#' (word, pdf), and it's not possible to know where we should create
#' a shape to contain the caption (technically it can't be in the
#' PowerPoint shape containing the table).
#'
#' When working with 'R Markdown' or 'Quarto', the caption settings
#' defined with `set_caption()` will be prioritized over knitr chunk options.
#'
#' Caption value can be a single string or the result to a call to
#' [as_paragraph()]. With the latter, the caption is made of
#' formatted chunks whereas with the former, caption will not be
#' associated with any formatting.
#'
#' @details
#' The behavior of captions in the 'flextable' package varies depending on the formats
#' and technologies used.
#'
#' The values set by the `set_caption()` function will be prioritized whenever
#' possible, including the caption ID and associated paragraph style.
#' However, it's important to note that the behavior may differ across different tools.
#' Here's what we have observed and attempted to respect, but please inform us
#' if you believe our observations are incorrect:
#'
#' - In Word and HTML documents created with 'rmarkdown' `rmarkdown::word_document()`
#' and `rmarkdown::html_document()`, numbered and cross-referenced captions are not
#' typically expected.
#' - In PDF documents created with 'rmarkdown' `rmarkdown::pdf_document()`, numbers
#' are automatically added before the caption.
#' - In Word and HTML documents created with 'bookdown', numbered and cross-referenced
#' captions are expected. 'bookdown' handles this functionality, but due to
#' technical reasons, the caption should not be defined within an HTML or XML block.
#' Therefore, when using 'flextable', the ability to format the caption content is
#' lost (this limitation does not apply to PDF documents).
#' - HTML and PDF documents created with Quarto handle captions and
#' cross-references differently. Quarto replaces captions with 'tbl-cap' and 'label'
#' values.
#' - Word documents created with Quarto present another specific case.
#' Currently, Quarto does not inject captions using the 'tbl-cap' and label values.
#' However, this is a temporary situation that is expected to change in the future.
#' The 'flextable' package will adapt accordingly as Quarto evolves.
#' - When using the `body_add_flextable()` function, all the options
#' specified with `set_caption()` will be enabled.
#'
#' Using [body_add_flextable()] enable all options specified with `set_caption()`.
#'
#' @section R Markdown:
#'
#' flextable captions can be defined from R Markdown documents by using
#' `knitr::opts_chunk$set()`. User don't always have to call `set_caption()`
#' to set a caption, he can use knitr chunk options instead. A typical call
#' would be:
#'
#' ``````
#' ```{r}
#' #| tab.id: bookmark_id
#' #| tab.cap: caption text
#' flextable(head(cars))
#' ```
#' ``````
#'
#' `tab.id` is the caption id or bookmark, `tab.cap` is the caption
#' text. There are many options that can replace `set_caption()`
#' features. The following knitr chunk options are available:
#'
#' | **label**                                               |    **name**     | **value**  |
#' |:--------------------------------------------------------|:---------------:|:----------:|
#' | Word stylename to use for table captions.               | tab.cap.style   |    NULL    |
#' | caption id/bookmark                                     | tab.id          |    NULL    |
#' | caption                                                 | tab.cap         |    NULL    |
#' | display table caption on top of the table or not        | tab.topcaption  |    TRUE    |
#' | caption table sequence identifier.                      | tab.lp          |   "tab:"   |
#' | prefix for numbering chunk (default to   "Table ").     | tab.cap.pre     |   Table    |
#' | suffix for numbering chunk (default to   ": ").         | tab.cap.sep     |    " :"    |
#' | title number depth                                      | tab.cap.tnd     |      0     |
#' | separator to use between title number and table number. | tab.cap.tns     |     "-"    |
#' | caption prefix formatting properties                    | tab.cap.fp_text | fp_text_lite(bold = TRUE) |
#'
#'
#' See [knit_print.flextable] for more details.
#'
#' @section Formatting the caption:
#'
#' To create captions in R Markdown using the 'flextable' package and 'officer'
#' package, you can utilize the `as_paragraph()` function. This approach is
#' recommended when your captions require complex content, such as a combination of
#' different text styles or the inclusion of images and equations.
#'
#' The caption is constructed as a paragraph consisting of multiple chunks. Each
#' chunk represents a specific portion of the caption with its desired formatting,
#' such as red bold text or Arial italic text.
#'
#' By default, if no specific formatting is specified (using either "a string" or
#' `as_chunk("a string")`), the `fp_text_default()` function sets the font settings
#' for the caption, including the font family, boldness, italics, color, etc. The
#' default values can be modified using the `set_flextable_defaults()` function.
#' However, it is recommended to explicitly use `as_chunk()` to define the desired
#' formatting.
#'
#' It's important to note that the style properties of the caption will not
#' override the formatting of the individual elements within it. Therefore, you
#' need to explicitly specify the font to be used for the caption.
#'
#' Here's an example of how to set a caption for a flextable in R Markdown using
#' the 'officer' package:
#'
#'
#' ```
#' library(flextable)
#' library(officer)
#'
#' ftab <- flextable(head(cars)) %>%
#'   set_caption(
#'     as_paragraph(
#'       as_chunk("caption", props = fp_text_default(font.family = "Cambria"))
#'     ), word_stylename = "Table Caption"
#'   )
#'
#' print(ftab, preview = "docx")
#' ```
#'
#' In this example, the `set_caption()` function sets the caption for the
#' flextable. The caption is created using `as_paragraph()` with a single chunk
#' created using `as_chunk("caption", props = fp_text_default(font.family =
#' "Cambria"))`. The `word_stylename` parameter is used to specify the table
#' caption style in the resulting Word document. Finally, the `print()` function
#' generates the flextable with the caption, and `preview = "docx"` displays a
#' preview of the resulting Word document.
#'
#' @section Using 'Quarto':
#'
#' In 'Quarto', captions and cross-references are handled differently
#' compared to 'R Markdown', where flextable takes care of the job.
#' In Quarto, the responsibility for managing captions lies with the Quarto
#' framework itself. Consequently, the `set_caption()` function in 'flextable'
#' is not as useful in a 'Quarto' document. The formatting and numbering of
#' captions are determined by Quarto rather than flextable. Please refer to
#' the Quarto documentation for more information on how to work with captions
#' in Quarto.
#'
#' @inheritParams args_x_only
#' @param caption caption value. The caption can be either a string either
#' a call to [as_paragraph()]. In the latter case, users are free to format
#' the caption with colors, italic fonts, also mixed with images or
#' equations. Note that Quarto does not allow the use of this feature.
#'
#' Caption as a string does not support 'Markdown' syntax. If you want to
#' add a bold text in the caption, use `as_paragraph('a ', as_b('bold'), ' text')`
#' when providing caption.
#' @param autonum an autonum representation. See [officer::run_autonum()].
#' This has an effect only when the output is "Word" (in which case the object
#' is used to define the Word auto-numbering), "html" and "pdf" (in which case only
#' the bookmark identifier will be used). If used, the caption is preceded
#' by an auto-number sequence.
#' @param word_stylename,style 'Word' style name to associate with caption paragraph. These names are available with
#' function [officer::styles_info()] when output is Word. Argument `style`
#' is deprecated in favor of `word_stylename`. If the caption is defined with
#' `as_paragraph()`, some of the formattings of the paragraph style will be
#' replaced by the formattings associated with the chunks (such as the font).
#' @param fp_p paragraph formatting properties associated with the caption, see [officer::fp_par()].
#' It applies when possible, i.e. in HTML and 'Word' but not with bookdown.
#' @param align_with_table if TRUE, caption is aligned as the flextable, if FALSE,
#' `fp_p` will not be updated and alignement is as defined with `fp_p`.
#' It applies when possible, i.e. in HTML and 'Word' but not with bookdown.
#' @param html_classes css class(es) to apply to associate with caption paragraph
#' when output is 'Word'.
#' @param html_escape should HTML entities be escaped so that it can be safely
#' included as text or an attribute value within an HTML document.
#' @examples
#' ftab <- flextable(head(iris))
#' ftab <- set_caption(ftab, "my caption")
#' ftab
#'
#' library(officer)
#' autonum <- run_autonum(seq_id = "tab", bkm = "mtcars")
#' ftab <- flextable(head(mtcars))
#' ftab <- set_caption(ftab, caption = "mtcars data", autonum = autonum)
#' ftab
#' @importFrom officer run_autonum
#' @importFrom htmltools htmlEscape
#' @seealso [flextable()], [knit_print.flextable()]
set_caption <- function(
  x,
  caption = NULL,
  autonum = NULL,
  word_stylename = "Table Caption",
  style = word_stylename,
  fp_p = fp_par(padding = 3),
  align_with_table = TRUE,
  html_classes = NULL,
  html_escape = TRUE
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "set_caption()"
    ))
  }

  caption_value <- NULL
  simple_caption <- TRUE
  if (!is.null(caption) && !inherits(caption, "paragraph")) {
    caption_value <- caption
  } else if (!is.null(caption) && inherits(caption, "paragraph")) {
    simple_caption <- FALSE
    caption_value <- caption[[1]]

    by_columns <- c(
      "font.size",
      "italic",
      "bold",
      "underlined",
      "strike",
      "color",
      "shading.color",
      "font.family",
      "hansi.family",
      "eastasia.family",
      "cs.family",
      "vertical.align"
    )
    default_fp_t <- fp_text_default()
    for (j in by_columns) {
      caption_value[[j]][is.na(caption_value[[j]])] <- default_fp_t[[j]]
    }
  }
  if (!is.null(caption) && !simple_caption) {
    caption_value <- expand_special_char(
      caption_value,
      what = "\n",
      with = "<br>"
    )
    caption_value <- expand_special_char(
      caption_value,
      what = "\t",
      with = "<tab>"
    )
  }

  x$caption <- list(
    value = caption_value,
    simple_caption = simple_caption,
    align_with_table = align_with_table
  )

  if (!is.null(autonum) && inherits(autonum, "run_autonum")) {
    x$caption$autonum <- autonum
  }
  x$caption$fp_p <- fp_p
  x$caption$style <- style
  x$caption$word_stylename <- word_stylename
  x$caption$html_classes <- if (!is.null(html_classes)) {
    paste(html_classes, collapse = " ")
  } else {
    NULL
  }

  x
}
update_caption <- function(
  x,
  caption = NULL,
  autonum = NULL,
  word_stylename = NULL,
  fp_p = NULL,
  align_with_table = NULL,
  html_classes = NULL
) {
  if (!is.null(caption)) {
    if (inherits(caption, "paragraph")) {
      x$caption$simple_caption <- FALSE
    } else {
      x$caption$simple_caption <- TRUE
    }
    x$caption$value <- caption
  }

  if (!is.null(autonum)) {
    x$caption$autonum <- autonum
  }

  if (!is.null(fp_p)) {
    x$caption$fp_p <- fp_p
  }

  if (!is.null(word_stylename)) {
    x$caption$word_stylename <- word_stylename
  }
  if (!is.null(html_classes)) {
    x$caption$html_classes <- html_classes
  }
  if (!is.null(align_with_table)) {
    x$caption$align_with_table <- align_with_table
  }

  x
}


#' @export
knit_print.run_reference <- function(x, ...) {
  is_quarto <- is_in_quarto()
  title <- ""
  if (is_quarto) {
    title <- opts_current_table()$cap.pre
  }
  if (grepl("docx", opts_knit$get("rmarkdown.pandoc.to"))) {
    knit_print(asis_output(
      paste(title, "`", to_wml(x), "`{=openxml}", sep = "")
    ))
  } else if (is_quarto) {
    knit_print(asis_output(
      paste("@", x$id, sep = "")
    ))
  } else {
    knit_print(asis_output(
      paste("\\@ref(tab:", x$id, ")", sep = "")
    ))
  }
}

test_that("padding overwrite all paddings", {
  ft <- flextable(data.frame(a = c("", ""), stringsAsFactors = FALSE))
  ft <- padding(ft, padding = 5)

  new_paddings <- c(
    ft$body$styles$pars$padding.bottom$data[, ],
    ft$body$styles$pars$padding.top$data[, ],
    ft$body$styles$pars$padding.left$data[, ],
    ft$body$styles$pars$padding.right$data[, ]
  )
  new_paddings <- unique(new_paddings)

  expect_equal(new_paddings, 5)
})

test_that("padding overwrite all paddings but not missing", {
  ft <- flextable(iris)
  ft <- padding(ft, padding = 5, padding.top = 20)
  new_paddings <- c(
    ft$body$styles$pars$padding.bottom$data[, ],
    ft$body$styles$pars$padding.top$data[, ],
    ft$body$styles$pars$padding.left$data[, ],
    ft$body$styles$pars$padding.right$data[, ]
  )
  new_paddings <- unique(new_paddings)

  expect_equal(new_paddings, c(5, 20))
})

context("check keep with next")

skip_on_cran()
skip_if_not_installed("doconv")
library(doconv)
skip_if_not(doconv::msoffice_available())
skip_if_not(pandoc_version() >= numeric_version("2"))

init_flextable_defaults()

ft_1 <- qflextable(head(cars, n = 20))

test_that("docx-keep-with-next", {
  local_edition(3)

  docx <- read_docx()
  docx <- body_add_par(docx, value = "a text")
  docx <- body_add_flextable(docx, ft_1, keepnext = FALSE)
  docx <- body_add_flextable(docx, ft_1, keepnext = FALSE)

  expect_snapshot_doc(name = "docx-keep-with-next-false", x = docx, engine = "testthat")

  docx <- read_docx()
  docx <- body_add_par(docx, value = "a text")
  docx <- body_add_flextable(docx, ft_1, keepnext = TRUE)
  docx <- body_add_flextable(docx, ft_1, keepnext = TRUE)

  expect_snapshot_doc(name = "docx-keep-with-next-true", x = docx, engine = "testthat")
})


rmd_file_0 <- "rmd/captions.Rmd"
if (!file.exists(rmd_file_0)) {#just for dev purpose
  rmd_file_0 <- "tests/testthat/rmd/captions.Rmd"
}
rmd_file <- tempfile(fileext = ".Rmd")
file.copy(rmd_file_0, rmd_file, overwrite = TRUE)

test_that("rdocx_document and keep-with-next", {
  local_edition(3)
  require(doconv)
  rmd_file <- tempfile(fileext = ".Rmd")
  outfile <- tempfile(fileext = ".docx")
  file.copy("rmd/keep-next.Rmd", rmd_file)
  render(rmd_file, output_file = outfile, output_format = "officedown::rdocx_document", envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(x = outfile, name = "rdocx_document", engine = "testthat")
  outfile <- tempfile(fileext = ".docx")
  render(rmd_file, output_format = "word_document", output_file = outfile, envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(x = outfile, name = "word_document", engine = "testthat")
})

init_flextable_defaults()

init_flextable_defaults()

test_that("to_rtf produces valid RTF string", {
  ft <- flextable(head(iris, 3))
  ft <- autofit(ft)

  rtf_str <- to_rtf(ft)
  expect_type(rtf_str, "character")
  expect_length(rtf_str, 1)
  expect_true(nchar(rtf_str) > 0)
  # RTF row markers
  expect_true(grepl("\\trowd", rtf_str, fixed = TRUE))
  expect_true(grepl("\\row", rtf_str, fixed = TRUE))
  expect_true(grepl("\\cell", rtf_str, fixed = TRUE))
})

test_that("to_rtf handles caption at top", {
  ft <- flextable(head(iris, 2))
  ft <- set_caption(ft, caption = "Iris dataset")

  rtf_top <- to_rtf(ft, topcaption = TRUE)
  rtf_bottom <- to_rtf(ft, topcaption = FALSE)

  expect_type(rtf_top, "character")
  expect_type(rtf_bottom, "character")
  # caption text should appear in both
  expect_true(grepl("Iris dataset", rtf_top, fixed = TRUE))
  expect_true(grepl("Iris dataset", rtf_bottom, fixed = TRUE))
})

test_that("to_rtf respects layout fixed vs autofit", {
  ft_fixed <- flextable(head(iris, 2))
  ft_fixed <- set_table_properties(ft_fixed, layout = "fixed")

  ft_auto <- flextable(head(iris, 2))
  ft_auto <- set_table_properties(ft_auto, layout = "autofit")

  rtf_fixed <- to_rtf(ft_fixed)
  rtf_auto <- to_rtf(ft_auto)

  expect_true(grepl("\\trautofit0", rtf_fixed, fixed = TRUE))
  expect_true(grepl("\\trautofit1", rtf_auto, fixed = TRUE))
})

test_that("to_rtf handles merged cells", {
  ft <- flextable(head(mtcars, 4)[, 1:3])
  ft <- merge_at(ft, i = 1:2, j = 1)

  rtf_str <- to_rtf(ft)
  expect_type(rtf_str, "character")
  expect_true(grepl("\\trowd", rtf_str, fixed = TRUE))
})

test_that("to_rtf handles styling", {
  ft <- flextable(head(iris, 2))
  ft <- bold(ft, part = "header")
  ft <- color(ft, color = "red", part = "body")

  rtf_str <- to_rtf(ft)
  expect_type(rtf_str, "character")
  # bold marker in header
  expect_true(grepl("\\b\\", rtf_str, fixed = TRUE))
  # color marker in body
  expect_true(grepl("%ftcolor:red%", rtf_str, fixed = TRUE))
})

test_that("rtf_add.flextable works with officer", {
  ft <- flextable(head(iris, 2))
  ft <- autofit(ft)

  doc <- officer::rtf_doc()
  doc <- officer::rtf_add(doc, ft)

  expect_s3_class(doc, "rtf")
  expect_true(length(doc$content) > 0)

  # write and verify file exists
  rtf_file <- tempfile(fileext = ".rtf")
  print(doc, target = rtf_file)
  expect_true(file.exists(rtf_file))
  content <- readLines(rtf_file, warn = FALSE)
  content <- paste(content, collapse = "\n")
  expect_true(grepl("\\trowd", content, fixed = TRUE))
  unlink(rtf_file)
})

test_that("to_rtf handles table alignment", {
  ft <- flextable(head(iris, 2))

  ft_left <- set_table_properties(ft, align = "left")
  ft_center <- set_table_properties(ft, align = "center")
  ft_right <- set_table_properties(ft, align = "right")

  expect_true(grepl("\\trql", to_rtf(ft_left), fixed = TRUE))
  expect_true(grepl("\\trqc", to_rtf(ft_center), fixed = TRUE))
  expect_true(grepl("\\trqr", to_rtf(ft_right), fixed = TRUE))
})

init_flextable_defaults()

init_flextable_defaults()

iris_sum <- summarizor(iris, by = "Species")
ft_1 <- as_flextable(iris_sum, sep_w = 0)
ft_1 <- set_caption(ft_1, "a caption")

test_that("docx-keep-with-next", {
  tmp_file <- tempfile(fileext = ".docx")
  docx <- read_docx()
  docx <- body_add_flextable(docx, ft_1)
  for (i in 1:10) {
    docx <- body_add_par(docx, value = "")
  }
  ft_1 <- paginate(ft_1, init = TRUE, hdr_ftr = TRUE)
  docx <- body_add_flextable(docx, ft_1)
  print(docx, target = tmp_file)

  doc <- read_docx(path = tmp_file)
  body_xml <- docx_body_xml(doc)

  expect_length(
    xml_find_all(body_xml, "//w:tbl[1]/w:tr/w:tc/w:p/w:pPr/w:keepNext"),
    n = 0
  )

  expect_length(
    xml_find_all(body_xml, "//w:tbl[2]/w:tr/w:tc/w:p/w:pPr/w:keepNext"),
    n = 65
  )
})

test_that("paginate with default init uses flextable defaults", {
  init_flextable_defaults()
  ft <- flextable(head(iris, 10))

  ft_paged <- paginate(ft, hdr_ftr = FALSE)
  par_data <- information_data_paragraph(ft_paged)
  default_kwn <- get_flextable_defaults()$keep_with_next
  expect_true(
    all(par_data$keep_with_next == default_kwn)
  )
})

test_that("paginate hdr_ftr binds header and footer to body", {
  dat <- data.frame(a = 1:3, b = 4:6)
  ft <- flextable(dat)
  ft <- add_footer_lines(ft, values = "a footer note")

  ft_paged <- paginate(ft, init = FALSE, hdr_ftr = TRUE)
  par_data <- information_data_paragraph(ft_paged)

  # header rows should all have keep_with_next = TRUE
  header_kwn <- par_data$keep_with_next[par_data$.part == "header"]
  expect_true(all(header_kwn))

  # last body row should have keep_with_next = TRUE (bound to footer)
  body_kwn <- par_data$keep_with_next[par_data$.part == "body"]
  body_nrow <- nrow_part(ft_paged, "body")
  last_body_col_count <- sum(
    par_data$.part == "body" &
      par_data$.row_id == body_nrow
  )
  last_body_kwn <- tail(body_kwn, last_body_col_count)
  expect_true(all(last_body_kwn))

  # footer: all TRUE except last row
  footer_kwn <- par_data[par_data$.part == "footer", ]
  footer_nrow <- nrow_part(ft_paged, "footer")
  kwn_not_last <- footer_kwn$keep_with_next[
    footer_kwn$.row_id < footer_nrow
  ]
  expect_true(all(kwn_not_last))
  kwn_last <- footer_kwn$keep_with_next[
    footer_kwn$.row_id == footer_nrow
  ]
  expect_true(all(!kwn_last))
})

test_that("paginate hdr_ftr FALSE does not alter header/footer binding", {
  dat <- data.frame(a = 1:3, b = 4:6)
  ft <- flextable(dat)
  ft <- add_footer_lines(ft, values = "a footer note")

  ft_paged <- paginate(ft, init = FALSE, hdr_ftr = FALSE)
  par_data <- information_data_paragraph(ft_paged)

  # with init=FALSE and hdr_ftr=FALSE, all keep_with_next should be FALSE
  expect_true(all(!par_data$keep_with_next))
})

test_that("paginate with group_def nonempty", {
  dat <- data.frame(
    grp = c("A", "", "", "B", "", "C"),
    val = 1:6,
    stringsAsFactors = FALSE
  )
  ft <- flextable(dat)

  ft_paged <- paginate(
    ft,
    init = FALSE,
    hdr_ftr = FALSE,
    group = "grp",
    group_def = "nonempty"
  )
  par_data <- information_data_paragraph(ft_paged)
  body_data <- par_data[par_data$.part == "body", ]

  # check keep_with_next per row (same for all columns in a row)
  kwn_by_row <- tapply(body_data$keep_with_next, body_data$.row_id, unique)

  # Groups: rows 1-3 (A), rows 4-5 (B), row 6 (C)
  # Last row of each group should have keep_with_next = FALSE
  # Row 3 (last of group A) -> FALSE
  # Row 5 (last of group B) -> FALSE
  # Row 6 (last of group C) -> FALSE
  # Rows 1, 2, 4 -> TRUE
  expect_true(kwn_by_row[["1"]])
  expect_true(kwn_by_row[["2"]])
  expect_false(kwn_by_row[["3"]])
  expect_true(kwn_by_row[["4"]])
  expect_false(kwn_by_row[["5"]])
  expect_false(kwn_by_row[["6"]])
})

test_that("paginate nonempty errors on missing group column", {
  dat <- data.frame(a = 1:3, b = 4:6)
  ft <- flextable(dat)
  expect_error(
    paginate(ft, group = "nonexistent", group_def = "nonempty"),
    "could not find"
  )
})

test_that("paginate with group_def rle", {
  dat <- data.frame(
    grp = c("A", "A", "A", "B", "B", "C"),
    val = 1:6,
    stringsAsFactors = FALSE
  )
  ft <- flextable(dat)

  ft_paged <- paginate(
    ft,
    init = FALSE,
    hdr_ftr = FALSE,
    group = "grp",
    group_def = "rle"
  )
  par_data <- information_data_paragraph(ft_paged)
  body_data <- par_data[par_data$.part == "body", ]

  kwn_by_row <- tapply(
    body_data$keep_with_next,
    body_data$.row_id,
    unique
  )

  # Groups: rows 1-3 (A), rows 4-5 (B), row 6 (C)
  # Last row of each group -> FALSE, others -> TRUE
  expect_true(kwn_by_row[["1"]])
  expect_true(kwn_by_row[["2"]])
  expect_false(kwn_by_row[["3"]])
  expect_true(kwn_by_row[["4"]])
  expect_false(kwn_by_row[["5"]])
  expect_false(kwn_by_row[["6"]])
})

test_that("paginate rle errors on missing group column", {
  dat <- data.frame(a = 1:3, b = 4:6)
  ft <- flextable(dat)
  expect_error(
    paginate(ft, group = "nonexistent", group_def = "rle"),
    "could not find"
  )
})

test_that("paginate rle with empty body returns unchanged", {
  dat <- data.frame(a = character(0), b = numeric(0))
  ft <- flextable(dat)
  ft_paged <- paginate(ft, group = "a", group_def = "rle")
  expect_equal(nrow_part(ft_paged, "body"), 0L)
})

test_that("paginate nonempty with empty body returns unchanged", {
  dat <- data.frame(a = character(0), b = numeric(0))
  ft <- flextable(dat)
  ft_paged <- paginate(
    ft,
    group = "a",
    group_def = "nonempty"
  )
  expect_equal(nrow_part(ft_paged, "body"), 0L)
})

init_flextable_defaults()

test_that("row height is valid", {
  ft <- flextable(head(iris))
  pptx_file <- "test.pptx"
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, value = ft, location = ph_location_type(type = "body"))
  doc <- print(doc, target = pptx_file)

  main_folder <- file.path(getwd(), "pptx_folder")
  officer::unpack_folder(pptx_file, folder = main_folder)

  slide_file <- file.path(main_folder, "/ppt/slides/slide1.xml")
  doc <- read_xml(slide_file)

  nodes <- xml_find_all(
    doc,
    "//p:graphicFrame/a:graphic/a:graphicData/a:tbl/a:tr"
  )
  h_values <- sapply(nodes, xml_attr, attr = "h")
  h_values <- as.integer(h_values)
  expect_true(all(is.finite(h_values)))
  expect_true(all(h_values > 0))

  unlink(main_folder, recursive = TRUE, force = TRUE)
  unlink(pptx_file, force = TRUE)
})


test_that("location is correct", {
  ft <- flextable(head(iris))
  pptx_file <- tempfile(fileext = ".pptx")
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, value = ft, location = ph_location(left = 0, top = 0))
  doc <- print(doc, target = pptx_file)

  main_folder <- file.path(getwd(), "pptx_folder")
  unpack_folder(file = pptx_file, folder = main_folder)

  slide_file <- file.path(main_folder, "/ppt/slides/slide1.xml")
  doc <- read_xml(slide_file)

  node <- xml_find_first(doc, "//p:graphicFrame/p:xfrm/a:off")
  expect_equal(xml_attr(node, "x"), "0")
  expect_equal(xml_attr(node, "y"), "0")

  unlink(main_folder, recursive = TRUE, force = TRUE)
})

#' @export
#' @title Transform a 'tables::tabular' object into a flextable
#' @description Produce a flextable from a 'tabular' object
#' produced with function [tables::tabular()].
#'
#' When `as_flextable.tabular=TRUE`, the first column is
#' used as row separator acting as a row title. It can
#' be formated with arguments `fp_p` (the formatting
#' properties of the paragraph) and `row_title` that
#' specifies the content and eventually formattings
#' of the content.
#'
#' Two hidden columns can be used for conditional formatting
#' after the creation of the flextable (use only when
#' `spread_first_col=TRUE`):
#'
#' - The column `.row_title` that contains the title label
#' - The column `.type` that can contain the following values:
#'   - "one_row": Indicates that there is only one row for this group. In this case, the row is not expanded with a title above.
#'   - "list_title": Indicates a row that serves as a title for the data that are displayed after it.
#'   - "list_data": Indicates rows that follow a title and contain data to be displayed.
#'
#' The result is paginated (see [paginate()]).
#' @param x object produced by [tables::tabular()].
#' @param spread_first_col if TRUE, first row is spread as a new line separator
#' instead of being a column. This helps to reduce the width and allows for
#' clear divisions.
#' @param fp_p paragraph formatting properties associated with row titles,
#' see [officer::fp_par()].
#' @param row_title a call to [as_paragraph()] - it
#' will be applied to the row titles if any
#' when `spread_first_col=TRUE`.
#' @param add_tab adds a tab in front of "list_data"
#' label lines (located in column `.type`).
#' @param ... unused argument
#' @family as_flextable methods
#' @examples
#' if (require("tables")) {
#'   set.seed(42)
#'   genders <- c("Male", "Female")
#'   status <- c("low", "medium", "high")
#'   Sex <- factor(sample(genders, 100, rep = TRUE))
#'   Status <- factor(sample(status, 100, rep = TRUE))
#'   z <- rnorm(100) + 5
#'   fmt <- function(x) {
#'     s <- format(x, digits = 2)
#'     even <- ((1:length(s)) %% 2) == 0
#'     s[even] <- sprintf("(%s)", s[even])
#'     s
#'   }
#'   tab <- tabular(
#'     Justify(c) * Heading() * z *
#'       Sex * Heading(Statistic) *
#'       Format(fmt()) *
#'       (mean + sd) ~ Status
#'   )
#'   as_flextable(tab)
#' }
#'
#' if (require("tables")) {
#'   tab <- tabular(
#'     (Species + 1) ~ (n = 1) + Format(digits = 2) *
#'       (Sepal.Length + Sepal.Width) * (mean + sd),
#'     data = iris
#'   )
#'   as_flextable(tab)
#' }
#'
#' if (require("tables")) {
#'   x <- tabular((Factor(gear, "Gears") + 1)
#'   * ((n = 1) + Percent()
#'       + (RowPct = Percent("row"))
#'       + (ColPct = Percent("col")))
#'   ~ (Factor(carb, "Carburetors") + 1)
#'     * Format(digits = 1), data = mtcars)
#'
#'   ft <- as_flextable(
#'     x,
#'     spread_first_col = TRUE,
#'     row_title = as_paragraph(
#'       colorize("Gears: ", color = "#666666"),
#'       colorize(as_b(.row_title), color = "red")
#'     )
#'   )
#'   ft
#' }
#'
#' if (require("tables")) {
#'   tab <- tabular(
#'     (mean + mean) * (Sepal.Length + Sepal.Width) ~ 1,
#'     data = iris
#'   )
#'   as_flextable(tab)
#' }
as_flextable.tabular <- function(
  x,
  spread_first_col = FALSE,
  fp_p = fp_par(text.align = "center", padding.top = 4),
  row_title = as_paragraph(as_chunk(.row_title)),
  add_tab = FALSE,
  ...
) {
  stopifnot(requireNamespace("tables", quietly = TRUE))

  body_data <- fortify_tabular_body(x)
  header_data <- fortify_tabular_header(x)
  text_align <- fortify_tabular_justify(x)
  vmerge_ins <- vmerge_instructions(x)
  hmerge_ins <- hmerge_instructions(x)
  .ncol <- ncol(tables::rowLabels(x))

  row_columns <- colnames(vmerge_ins)
  data_columns <- setdiff(colnames(body_data), row_columns)
  group_colname <- character()

  if (spread_first_col) {
    group_colname <- row_columns[1]
    row_columns <- setdiff(row_columns, group_colname)
    .ncol <- .ncol - length(group_colname)

    .is_list_title <- is_list_title(body_data[[group_colname]])
    .is_one_row <- is_one_row(body_data[[group_colname]])

    body_data <- expand_dataset(
      body_data = body_data,
      is_title = .is_list_title,
      is_single = .is_one_row,
      group_colname = group_colname
    )

    text_align$body <- expand_dataset(
      body_data = as.data.frame(text_align$body),
      is_title = .is_list_title,
      is_single = .is_one_row,
      group_colname = group_colname
    )
  } else {
    body_data$.type <- "list_data"
    body_data$.row_title <- FALSE
  }

  columns_keys <- setdiff(c(row_columns, data_columns), group_colname)

  vmerge_ins <- vmerge_ins[,
    setdiff(colnames(vmerge_ins), group_colname),
    drop = FALSE
  ]
  text_align$header <- text_align$header[,
    setdiff(colnames(text_align$header), group_colname),
    drop = FALSE
  ]
  text_align$body <- text_align$body[,
    setdiff(colnames(text_align$body), group_colname),
    drop = FALSE
  ]

  if (length(group_colname)) {
    body_data$.row_title <- body_data[[group_colname]]
    body_data$.row_title[body_data$.type %in% "one_row"] <-
      body_data[[row_columns]][body_data$.type %in% "one_row"]
  }
  ft <- flextable(body_data, col_keys = columns_keys)

  ft <- set_header_df(ft, mapping = header_data)

  ft <- merge_v(ft, part = "header", j = row_columns)

  for (i in seq_along(hmerge_ins)) {
    hgroups <- split(seq_len(nrow(hmerge_ins)), hmerge_ins[[i]])
    for (hgroup in hgroups) {
      ft <- merge_at(ft, part = "header", j = hgroup + .ncol, i = i)
    }
  }
  if (spread_first_col) {
    for (j in setdiff(names(vmerge_ins), ".type")) {
      vgroups <- split(seq_len(nrow(vmerge_ins)), vmerge_ins[[j]])
      for (vgroup in vgroups) {
        ft <- merge_at(ft, part = "body", j = j, i = vgroup)
      }
    }
  }

  ft <- do.call(get_flextable_defaults()$theme_fun, list(ft))

  for (j in columns_keys) {
    ft <- align(ft, j = j, align = text_align$header[, j], part = "header")
    aligns <- text_align$body[, j]
    if (".type" %in% colnames(text_align$body)) {
      aligns <- aligns[text_align$body$.type %in% c("one_row", "list_data")]
    }
    ft <- align(
      ft,
      j = j,
      i = ~ .type %in% c("one_row", "list_data"),
      align = aligns,
      part = "body"
    )
  }

  ft <- valign(ft, valign = "top", part = "body")
  ft <- valign(ft, valign = "bottom", part = "header")

  # format functions can not be simply called
  # then, it looks easier to call format.tabular
  # and inject its trimmed values. That way we
  # preserve data types and can do conditionnal
  # formatting
  strmat <- format(x)
  colnames(strmat) <- data_columns

  rindex <- body_data$.type %in% c("one_row", "list_data")

  newcontent <- get_chunkset_struct_element(
    x = ft[["body"]]$content,
    i = rindex,
    j = data_columns
  )
  newtext <- trimws(strmat[, data_columns, drop = FALSE])
  for (j in data_columns) {
    for (irow in seq_len(nrow(newtext))) {
      newcontent[[irow, j]]$txt <- newtext[[irow, j]]
    }
  }
  ft[["body"]]$content <- set_chunkset_struct_element(
    x = ft[["body"]]$content,
    i = rindex,
    j = data_columns,
    value = newcontent
  )

  if (any(body_data$.type %in% "list_title")) {
    ft <- merge_h_range(
      ft,
      i = ~ .type %in% c("list_title"),
      j1 = 1L,
      j2 = length(columns_keys)
    )
    ft <- mk_par(
      ft,
      i = body_data$.type %in% c("list_title", "one_row"),
      j = 1,
      value = {{ row_title }}
    )
    ft <- style(
      ft,
      i = body_data$.type %in% c("list_title", "one_row"),
      j = 1,
      pr_p = fp_p,
      part = "body"
    )
    if (add_tab) {
      ft <- prepend_chunks(
        ft,
        i = ~ .type %in% "list_data",
        j = 1,
        as_chunk("\t")
      )
    }
  }

  ft <- fix_border_issues(ft)
  best_widths_ <- dim_pretty(ft)$widths

  best_widths_[setdiff(
    seq_along(columns_keys),
    seq_len(.ncol)
  )] <- max(best_widths_[setdiff(seq_along(columns_keys), seq_len(.ncol))]) *
    1.1
  ft <- width(ft, width = best_widths_)

  if (spread_first_col) {
    ft <- paginate(ft, group = group_colname, group_def = "nonempty")
  } else {
    ft <- paginate(ft, group = row_columns[1], group_def = "nonempty")
  }
  ft$tabular_info <- list(
    row_columns = row_columns,
    data_columns = data_columns
  )

  ft
}

# utils -----

expand_dataset <- function(
  body_data,
  is_title,
  is_single,
  group_colname = "COL1"
) {
  body_data$.fakeid <- seq_len(nrow(body_data))

  title_dat <- body_data[is_title, ]
  title_dat$.fakeid <- title_dat$.fakeid - .1
  title_dat <- title_dat[, c(".fakeid", group_colname)]
  title_dat$.type <- rep("list_title", nrow(title_dat))

  singlerow_dat <- body_data[is_single, ]
  singlerow_dat$.fakeid <- singlerow_dat$.fakeid - .1
  .col_dest <- head(setdiff(colnames(singlerow_dat), group_colname), 1)
  singlerow_dat[[.col_dest]] <- singlerow_dat[[group_colname]]
  singlerow_dat[[group_colname]] <- rep(NA_character_, nrow(singlerow_dat))
  singlerow_dat$.type <- rep("one_row", nrow(singlerow_dat))

  body_data <- body_data[!is_single, ]
  body_data[[group_colname]] <- rep(NA_character_, nrow(body_data))
  body_data$.type <- rep("list_data", nrow(body_data))

  dat <- rbindlist(
    list(title_dat, singlerow_dat, body_data),
    use.names = TRUE,
    fill = TRUE
  )
  dat <- dat[order(dat$.fakeid), ]
  dat$.fakeid <- NULL

  setDF(dat)
  dat
}

vmerge_instructions <- function(x) {
  label_data <- tables::rowLabels(x)
  label_data <- apply(label_data[], 2, function(z) {
    notna <- !is.na(z)
    rleid(cumsum(notna))
  })
  label_data <- as.data.frame(label_data)
  colnames(label_data) <- sprintf("COL%.0f", seq_len(ncol(label_data)))
  label_data
}
hmerge_instructions <- function(x) {
  label_data <- tables::colLabels(x)
  label_data <- apply(label_data[], 1, function(z) {
    notna <- !is.na(z)
    rleid(cumsum(notna))
  })
  label_data <- as.data.frame(label_data)
  label_data
}

fortify_tabular_body <- function(x, ...) {
  row_labels <- as.data.frame(unclass(tables::rowLabels(x)))
  names(row_labels) <- sprintf("COL%.0f", seq_len(ncol(row_labels)))

  dims <- attr(x, "dim")
  celldata <- matrix(x[], nrow = dims[1], ncol = dims[2])
  celldata <- apply(celldata, 2, function(dat) {
    unlist(dat)
  })
  celldata <- as.data.frame(celldata)
  names(celldata) <- sprintf(
    "COL%.0f",
    seq_len(ncol(celldata)) + ncol(row_labels)
  )
  cbind(row_labels, celldata)
}

fortify_tabular_header <- function(x, ...) {
  col_labels <- as.data.frame(unclass(tables::colLabels(x)))
  row_labels <- matrix(
    rep(colnames(tables::rowLabels(x)), nrow(col_labels)),
    nrow = nrow(col_labels),
    byrow = TRUE
  )

  dat <- cbind(row_labels, col_labels)
  dat <- t(dat)
  dat <- as.data.frame(dat)
  dat$col_keys <- sprintf("COL%.0f", seq_len(nrow(dat)))
  dat
}

fortify_tabular_justify <- function(x, justification = "c", ...) {
  justify <- attr(x, "justification")
  rlabels <- tables::rowLabels(x)
  rjustify <- attr(rlabels, "justification")

  clabels <- tables::colLabels(x)
  cjustify <- attr(clabels, "justification")
  corjustify <- matrix(NA, nrow(clabels), ncol(rlabels))
  for (i in seq_len(ncol(rlabels))) {
    corjustify[nrow(clabels), i] <- rjustify[1, i]
  }
  justify <- rbind(
    cbind(corjustify, cjustify),
    cbind(rjustify, justify)
  )
  justify[is.na(justify)] <- justification
  justify[] <- c(l = "left", c = "center", r = "right", n = "left")[justify]
  colnames(justify) <- sprintf("COL%.0f", seq_len(ncol(justify)))

  list(
    header = justify[seq_len(nrow(clabels)), , drop = FALSE],
    body = justify[
      setdiff(seq_len(nrow(justify)), seq_len(nrow(clabels))),
      ,
      drop = FALSE
    ]
  )
}

is_list_title <- function(x) {
  lag_str <- c(tail(x, -1), "")
  !is.na(x) & is.na(lag_str)
}

is_one_row <- function(x) {
  lag_str <- c(tail(x, -1), "")
  !is.na(x) & !is.na(lag_str)
}

#' @importFrom officer ph_with
#' @export
#' @method ph_with flextable
#' @title Add a flextable into a PowerPoint slide
#' @description
#'
#' Add a flextable in a PowerPoint document object produced
#' by [officer::read_pptx()].
#'
#' This function will create a native PowerPoint table from the flextable
#' and the result can be eventually edited.
#' @section caption:
#' Captions are not printed in PowerPoint slides.
#'
#' While captions are useful for document formats like Word, RTF, HTML, or PDF,
#' they aren't directly supported in PowerPoint slides. Unlike documents with
#' a defined layout, PowerPoint slides lack a structured document flow. They
#' don't function like HTML documents or paginated formats (RTF, Word, PDF).
#' This makes it technically challenging to determine the ideal placement for
#' a caption within a slide. Additionally, including a caption within the table
#' itself isn't feasible.
#' @param x a pptx device
#' @param value flextable object
#' @param location a location for a placeholder. See [officer::ph_location_type()]
#' for example.
#' @param ... unused arguments.
#' @family officer_integration
#' @note
#' The width and height of the table can not be set with `location`. Use
#' functions [width()], [height()], [autofit()]
#' and [dim_pretty()] instead. The overall size is resulting from
#' cells, paragraphs and text properties (i.e. padding, font size, border widths).
#' @examples
#' library(officer)
#'
#' ft <- flextable(head(iris))
#'
#' doc <- read_pptx()
#' doc <- add_slide(doc, "Title and Content", "Office Theme")
#' doc <- ph_with(doc, ft, location = ph_location_left())
#'
#' fileout <- tempfile(fileext = ".pptx")
#' print(doc, target = fileout)
ph_with.flextable <- function(x, value, location, ...) {
  stopifnot(inherits(x, "rpptx"))

  value <- flextable_global$defaults$post_process_all(value)
  value <- flextable_global$defaults$post_process_pptx(value)
  value <- fix_border_issues(value)

  graphic_frame <- gen_raw_pml(value)
  ph_with(
    x = x,
    value = as_xml_document(graphic_frame),
    location = location,
    ...
  )
}

colcode <- function(x) {
  rgb(t(col2rgb(x, alpha = TRUE)), maxColorValue = 255)
}
colcodecss <- function(x) {
  args <- as.list(as.data.frame(t(col2rgb(x, alpha = TRUE))))
  args$alpha <- args$alpha / 255
  args$fmt <- "rgba(%.0f, %.0f, %.0f, %.2f)"
  do.call(sprintf, args)
}
colalpha <- function(x) {
  x[x %in% "transparent"] <- "#FFFFFF00"
  rgbvals <- col2rgb(x, alpha = TRUE)
  as.integer(rgbvals[4, ] / 255 * 100000)
}

colcode0 <- function(x) {
  x[x %in% "transparent"] <- "#FFFFFF00"
  substr(rgb(t(col2rgb(x, alpha = TRUE)), maxColorValue = 255), 2, 7)
}

check_choice <- function(value, choices) {
  varname <- as.character(substitute(value))
  if (is.character(value) && length(value) == 1) {
    if (!value %in% choices) {
      stop(
        sprintf(
          "%s is expected to be one of: %s",
          varname,
          paste(shQuote(choices))
        ),
        call. = FALSE
      )
    }
  } else {
    stop(sprintf("`%s` must be a character scalar.", varname), call. = FALSE)
  }
  TRUE
}

cast_borders <- function(value) {
  borders_id <- c("border.bottom", "border.top", "border.left", "border.right")
  borders_side <- c("bottom", "top", "left", "right")

  z <- mapply(
    function(name, fp_b) {
      x <- unclass(fp_b)
      if (is.null(x)) {
        x <- list(
          width = NA_integer_,
          color = NA_character_,
          style = NA_character_
        )
      }
      names(x) <- paste("border", c("width", "color", "style"), name, sep = ".")
      x
    },
    borders_side,
    value[borders_id],
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  z <- Reduce(f = append, z)
  value <- append(value, z)
  value[borders_id] <- NULL
  value
}

as_struct <- function(nrow, keys, pr, fun) {
  args <- pr
  args$nrow <- nrow
  args$keys <- keys

  borders_id <- c("border.bottom", "border.top", "border.left", "border.right")
  borders_side <- c("bottom", "top", "left", "right")
  avail_bdr_id <- borders_id %in% names(pr)
  bdr_id <- borders_id[avail_bdr_id]
  bdr_side <- borders_side[avail_bdr_id]
  z <- mapply(
    function(name, fp_b) {
      x <- unclass(fp_b)
      names(x) <- paste("border", names(x), name, sep = ".")
      x
    },
    bdr_side,
    pr[bdr_id],
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  z <- Reduce(f = append, z)

  args <- append(args, z)
  args[bdr_id] <- NULL

  obj_str <- do.call(fun, args)
  obj_str
}

base_ns <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\""

wml_image <- function(src, width, height, alt = "") {
  str <- paste0(
    sprintf("<w:r %s>", base_ns),
    "<w:rPr/><w:drawing><wp:inline distT=\"0\" distB=\"0\" distL=\"0\" distR=\"0\">",
    sprintf(
      "<wp:extent cx=\"%.0f\" cy=\"%.0f\"/>",
      width * 12700 * 72,
      height * 12700 * 72
    ),
    sprintf("<wp:docPr id=\"\" name=\"\" descr=\"%s\"/>", htmlEscape(alt)),
    "<wp:cNvGraphicFramePr><a:graphicFrameLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/></wp:cNvGraphicFramePr>",
    "<a:graphic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"><a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/picture\"><pic:pic xmlns:pic=\"http://schemas.openxmlformats.org/drawingml/2006/picture\">",
    "<pic:nvPicPr>",
    "<pic:cNvPr id=\"\" name=\"\"/>",
    "<pic:cNvPicPr><a:picLocks noChangeAspect=\"1\" noChangeArrowheads=\"1\"/>",
    "</pic:cNvPicPr></pic:nvPicPr>",
    "<pic:blipFill>",
    sprintf("<a:blip r:embed=\"%s\"/>", src),
    "<a:srcRect/><a:stretch><a:fillRect/></a:stretch></pic:blipFill>",
    "<pic:spPr bwMode=\"auto\"><a:xfrm><a:off x=\"0\" y=\"0\"/>",
    sprintf(
      "<a:ext cx=\"%.0f\" cy=\"%.0f\"/></a:xfrm><a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom><a:noFill/></pic:spPr>",
      width * 12700 * 72,
      height * 12700 * 72
    ),
    "</pic:pic></a:graphicData></a:graphic></wp:inline></w:drawing></w:r>"
  )
  str
}

with_openxml_quotes <- function(x) {
  if (length(x) < 1) {
    x <- ""
  } else if (length(x) > 1) {
    x <- paste(x, collapse = "\n")
  }

  paste("\n\n``````{=openxml}", x, "``````\n\n", sep = "\n")
}

#' @export
#' @title Prevent page breaks inside a flextable
#' @description
#' Prevents breaks between tables rows you want to stay together.
#' This feature only applies to Word and RTF output.
#' @inheritParams args_x_only
#' @param init init value for keep_with_next property, it default
#' value is `get_flextable_defaults()$keep_with_next`.
#' @param hdr_ftr if TRUE (default), prevent breaks between table body
#' and header and between table body and footer.
#' @param group name of a column to use for finding groups (when
#'   `group_def` is `"rle"` or `"nonempty"`) or an integer vector
#'   of row indices that start new groups (when `group_def` is `"starts"`).
#' @param group_def algorithm to be used to identify groups
#' that should not be split into two pages, one of 'rle', 'nonempty',
#' 'starts':
#'
#' - 'rle': runs of equal values are used to define the groups,
#' to be used with [tabulator()].
#' - 'nonempty': non empty value start a new group,
#' to be used with [as_flextable.tabular()].
#' - 'starts': `group` is a numeric vector of body row indices
#' where new groups begin. Page breaks are allowed before these rows.
#' @return updated flextable object
#' @details
#' The pagination of tables allows you to control their position
#' in relation to page breaks.
#'
#' For small tables, a simple setting is usually used that indicates
#' that all rows should be displayed together:
#'
#' ```
#' paginate(x, init = TRUE, hdr_ftr = TRUE)
#' ```
#'
#' For large tables, it is recommended to use a setting that
#' indicates that all rows of the header should be bound to
#' the first row of the table to avoid the case where the header
#' is displayed alone at the bottom of the page and then repeated
#' on the next one:
#'
#' ```
#' paginate(x, init = FALSE, hdr_ftr = TRUE)
#' ```
#'
#' For tables that present groups that you don't want to be presented
#' on two pages, you must use a parameterization involving the notion
#' of group and an algorithm for determining the groups.
#'
#'
#' ```
#' paginate(x, group = "grp", group_def = "rle")
#' ```
#'
#' @examples
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' library(data.table)
#' library(flextable)
#'
#' init_flextable_defaults()
#'
#' multi_fun <- function(x) {
#'   list(mean = mean(x), sd = sd(x))
#' }
#'
#' dat <- as.data.table(ggplot2::diamonds)
#' dat <- dat[clarity %in% c("I1", "SI1", "VS2")]
#'
#' dat <- dat[, unlist(lapply(.SD, multi_fun),
#'   recursive = FALSE
#' ),
#' .SDcols = c("z", "y"),
#' by = c("cut", "color", "clarity")
#' ]
#'
#' tab <- tabulator(
#'   x = dat, rows = c("cut", "color"),
#'   columns = "clarity",
#'   `z stats` = as_paragraph(as_chunk(fmt_avg_dev(z.mean, z.sd, digit2 = 2))),
#'   `y stats` = as_paragraph(as_chunk(fmt_avg_dev(y.mean, y.sd, digit2 = 2)))
#' )
#'
#' ft_1 <- as_flextable(tab)
#' ft_1 <- autofit(x = ft_1, add_w = .05)
#' ft_1 <- paginate(ft_1, group = "cut", group_def = "rle")
#'
#' save_as_docx(ft_1, path = tempfile(fileext = ".docx"))
#' save_as_rtf(ft_1, path = tempfile(fileext = ".rtf"))
#' @seealso [knit_print.flextable()]
#' @family table_structure
paginate <- function(
  x,
  init = NULL,
  hdr_ftr = TRUE,
  group = character(),
  group_def = c("rle", "nonempty", "starts")
) {
  if (is.null(init)) {
    init <- get_flextable_defaults()$keep_with_next
  }
  x <- keep_with_next(x, value = init, part = "all")

  if (hdr_ftr) {
    x <- paginate_hdr_ftr(x)
  }

  g_algo <- match.arg(group_def, several.ok = FALSE)
  if ("starts" %in% g_algo) {
    if (is.numeric(group) && length(group) > 0L) {
      x <- paginate_starts(x = x, starts = group)
    }
  } else if (length(group) == 1 && is.character(group)) {
    if ("nonempty" %in% g_algo) {
      x <- paginate_nonempty(x = x, group = group)
    } else if ("rle" %in% g_algo) {
      x <- paginate_rleid(x = x, group = group)
    }
  }

  x
}

paginate_hdr_ftr <- function(x) {
  nrow_footer <- nrow_part(x, "footer")
  nrow_header <- nrow_part(x, "header")
  if (nrow_part(x, "header")) {
    x <- keep_with_next(x, value = TRUE, part = "header")
  }
  if (nrow_footer > 0 && nrow_part(x, "body")) {
    x <- keep_with_next(
      x,
      value = TRUE,
      part = "body",
      i = nrow_part(x, "body")
    )
  }
  if (nrow_footer > 0) {
    x <- keep_with_next(x, value = TRUE, part = "footer")
    x <- keep_with_next(x, i = nrow_footer, value = FALSE, part = "footer")
  }
  x
}

paginate_nonempty <- function(x, group) {
  if (nrow_part(x, "body") < 1) {
    return(x)
  }

  if (!group %in% colnames(x$body$dataset)) {
    stop("could not find ", shQuote(group))
  }
  grp_val <- x$body$dataset[[group]]
  index_false <- c(
    which(!is.na(grp_val) & !grp_val %in% ""),
    length(grp_val) + 1
  )
  index_false <- index_false - 1
  index_false <- index_false[index_false > 0]
  x <- keep_with_next(x, value = TRUE, part = "body")
  x <- keep_with_next(x, i = index_false, value = FALSE, part = "body")
  x
}

paginate_rleid <- function(x, group) {
  if (nrow_part(x, "body") < 1) {
    return(x)
  }

  if (!group %in% colnames(x$body$dataset)) {
    stop("could not find ", shQuote(group))
  }
  grp_val <- x$body$dataset[[group]]
  index_false <- as.integer(tapply(
    seq_along(grp_val),
    rleid(grp_val),
    function(z) z[length(z)]
  ))
  x <- keep_with_next(x, value = TRUE, part = "body")
  x <- keep_with_next(x, i = index_false, value = FALSE, part = "body")
  x
}

paginate_starts <- function(x, starts) {
  n <- nrow_part(x, "body")
  if (n < 1L) {
    return(x)
  }

  index_false <- starts - 1L
  index_false <- index_false[index_false > 0L & index_false <= n]
  index_false <- unique(c(index_false, n))

  x <- keep_with_next(x, value = TRUE, part = "body")
  x <- keep_with_next(x, i = index_false, value = FALSE, part = "body")
  x
}

#' @title Set column formatter functions
#' @description Apply formatter functions to column keys.
#'
#' Functions should have a single argument (the vector) and should
#' return the formatted values as a character vector.
#' @inheritParams args_x_part_no_all
#' @param ... Name-value pairs of functions, names should be existing col_key values
#' @param values format functions, If values is supplied argument `...` is ignored.
#' - It can be a list of name-value pairs of functions, names
#' should be existing col_key values.
#' - If `values` is a single function, it will be applied to each column.
#' @examples
#' ft <- flextable(head(iris))
#' ft <- set_formatter(
#'   x = ft,
#'   Sepal.Length = function(x) sprintf("%.02f", x),
#'   Sepal.Width = function(x) sprintf("%.04f", x)
#' )
#' ft <- theme_vanilla(ft)
#' ft
#' @export
#' @family cells formatters
set_formatter <- function(x, ..., values = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "set_formatter()"
    ))
  }

  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  if (is.null(values)) {
    values <- list(...)
  } else if (is.function(values)) {
    values <- lapply(x$col_keys, function(...) {
      values
    })
    names(values) <- x$col_keys
  }
  col_keys <- intersect(names(values), x[[part]]$col_keys)

  for (key in col_keys) {
    dat <- x[[part]]$dataset[, key]
    chk <- as_chunk(values[[key]](dat))
    x <- compose(x, j = key, value = as_paragraph(chk), part = part)
  }

  x
}

#' @export
#' @title Format character cells
#' @description Format character cells in a flextable.
#' @inheritParams args_x_i_j
#' @param na_str,nan_str string to be used for NA and NaN values
#' @param prefix,suffix string to be used as prefix or suffix
#' @family cells formatters
#' @examples
#' dat <- iris
#' z <- flextable(head(dat))
#' ft <- colformat_char(
#'   x = z, j = "Species", suffix = "!"
#' )
#' z <- autofit(z)
#' z
colformat_char <- function(
  x,
  i = NULL,
  j = NULL,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
) {
  stopifnot(inherits(x, "flextable"))

  quo_fun <- quo(format_fun.character(
    x,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix
    ),
    get_expr(quo_fun)
  )

  col_keys <- filter_col_keys(x, j, function(x) is.character(x) || is.factor(x))
  docall_display(col_keys, fun_, x, i = i)
}

#' @export
#' @title Format double cells
#' @description Format numeric cells in a flextable using
#' [formatC()] with explicit control over digits and
#' decimal mark.
#' @inheritParams colformat_char
#' @param big.mark,digits,decimal.mark see [formatC()]
#' @family cells formatters
#' @examples
#' dat <- mtcars
#' ft <- flextable(head(dat))
#' ft <- colformat_double(
#'   x = ft,
#'   big.mark = ",", digits = 2, na_str = "N/A"
#' )
#' autofit(ft)
#' @importFrom rlang new_function quo get_expr pairlist2
colformat_double <- function(
  x,
  i = NULL,
  j = NULL,
  big.mark = get_flextable_defaults()$big.mark,
  decimal.mark = get_flextable_defaults()$decimal.mark,
  digits = get_flextable_defaults()$digits,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
) {
  stopifnot(inherits(x, "flextable"))

  col_keys <- filter_col_keys(x, j, function(x) {
    is.double(x) && !inherits(x, "POSIXt") && !inherits(x, "Date")
  })

  quo_fun <- quo(format_fun.double(
    x,
    big.mark = big.mark,
    decimal.mark = decimal.mark,
    digits = digits,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      big.mark = big.mark,
      decimal.mark = decimal.mark,
      digits = digits,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix
    ),
    get_expr(quo_fun)
  )

  docall_display(col_keys, fun_, x, i = i)
}
#' @export
#' @title Format numeric cells with format()
#' @description Format numeric cells in a flextable
#' using R's [format()] function.
#'
#' The function is different from [colformat_double()] on numeric type
#' columns. The function uses the [format()] function of R on numeric
#' type columns. So this is normally what you see on the R console
#' most of the time (but scientific mode is disabled and NA are replaced).
#'
#' @section format call:
#'
#' Function [format()] is called with the following values:
#'
#' * `trim` is set to TRUE,
#' * `scientific` is set to FALSE,
#' * `big.mark` is set to the value of `big.mark` argument,
#' * `decimal.mark` is set to the value of `decimal.mark` argument,
#' * other arguments are passed 'as is' to the format function.
#'
#' argument `digits` is ignored as it is not the same `digits` that users
#' want, this one will be used by [format()] and not [formatC()].
#' To change the digit argument use `options(digits=4)` instead.
#'
#' This argument will not be changed because `colformat_num()`
#' is supposed to format things roughly as what you see on the R console.
#'
#' If these functions does not fit your needs, use [set_formatter()]
#' that lets you use any format function.
#'
#' @inheritParams colformat_char
#' @param big.mark,decimal.mark see [format()]
#' @param ... additional argument for function [format()], `scientific`
#' and `digits` can not be used.
#' @family cells formatters
#' @examples
#' dat <- mtcars
#' dat[2, 1] <- NA
#' ft <- flextable(head(dat))
#' ft <- colformat_num(
#'   x = ft,
#'   big.mark = " ", decimal.mark = ",",
#'   na_str = "N/A"
#' )
#' ft <- autofit(ft)
#' ft
colformat_num <- function(
  x,
  i = NULL,
  j = NULL,
  big.mark = get_flextable_defaults()$big.mark,
  decimal.mark = get_flextable_defaults()$decimal.mark,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = "",
  ...
) {
  stopifnot(inherits(x, "flextable"))
  col_keys <- filter_col_keys(x, j, is.numeric)

  quo_fun <- quo(format_fun.default(
    x,
    big.mark = big.mark,
    decimal.mark = decimal.mark,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix,
    ...
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      big.mark = big.mark,
      decimal.mark = decimal.mark,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix,
      ...
    ),
    get_expr(quo_fun)
  )

  docall_display(col_keys, fun_, x, i = i)
}

#' @title Format date cells
#' @description Format date cells in a flextable.
#' @inheritParams colformat_char
#' @param fmt_date see [strptime()]
#' @family cells formatters
#' @export
#' @examples
#' dat <- data.frame(
#'   z = Sys.Date() + 1:3,
#'   w = Sys.Date() - 1:3
#' )
#' ft <- flextable(dat)
#' ft <- colformat_date(x = ft)
#' ft <- autofit(ft)
#' ft
colformat_date <- function(
  x,
  i = NULL,
  j = NULL,
  fmt_date = get_flextable_defaults()$fmt_date,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
) {
  stopifnot(inherits(x, "flextable"))

  col_keys <- filter_col_keys(x, j, function(x) inherits(x, "Date"))

  quo_fun <- quo(format_fun.Date(
    x,
    fmt_date = fmt_date,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      fmt_date = fmt_date,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix
    ),
    get_expr(quo_fun)
  )

  docall_display(col_keys, fun_, x, i = i)
}

#' @title Format datetime cells
#' @description Format datetime cells in a flextable.
#' @inheritParams colformat_char
#' @param fmt_datetime see [strptime()]
#' @family cells formatters
#' @export
#' @examples
#' dat <- data.frame(
#'   z = Sys.time() + (1:3) * 24,
#'   w = Sys.Date() - (1:3) * 24
#' )
#' ft <- flextable(dat)
#' ft <- colformat_datetime(x = ft)
#' ft <- autofit(ft)
#' ft
colformat_datetime <- function(
  x,
  i = NULL,
  j = NULL,
  fmt_datetime = get_flextable_defaults()$fmt_datetime,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
) {
  stopifnot(inherits(x, "flextable"))

  col_keys <- filter_col_keys(x, j, function(x) inherits(x, "POSIXt"))

  quo_fun <- quo(format_fun.POSIXt(
    x,
    fmt_datetime = fmt_datetime,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      fmt_datetime = fmt_datetime,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix
    ),
    get_expr(quo_fun)
  )

  docall_display(col_keys, fun_, x, i = i)
}

#' @title Format integer cells
#' @description Format integer cells in a flextable.
#' @inheritParams colformat_char
#' @param big.mark see [format()]
#' @family cells formatters
#' @export
#' @examples
#' z <- flextable(head(mtcars))
#' j <- c("vs", "am", "gear", "carb")
#' z <- colformat_int(x = z, j = j, prefix = "# ")
#' z
colformat_int <- function(
  x,
  i = NULL,
  j = NULL,
  big.mark = get_flextable_defaults()$big.mark,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
) {
  stopifnot(inherits(x, "flextable"))

  col_keys <- filter_col_keys(x, j, is.integer)

  quo_fun <- quo(format_fun.integer(
    x,
    big.mark = big.mark,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      big.mark = big.mark,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix
    ),
    get_expr(quo_fun)
  )

  docall_display(col_keys, fun_, x, i = i)
}

#' @title Format logical cells
#' @description Format logical cells in a flextable.
#' @inheritParams colformat_char
#' @param false,true string to be used for logical
#' @family cells formatters
#' @export
#' @examples
#' dat <- data.frame(a = c(TRUE, FALSE), b = c(FALSE, TRUE))
#'
#' z <- flextable(dat)
#' z <- colformat_lgl(x = z, j = c("a", "b"))
#' autofit(z)
colformat_lgl <- function(
  x,
  i = NULL,
  j = NULL,
  true = "true",
  false = "false",
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
) {
  stopifnot(inherits(x, "flextable"))

  col_keys <- filter_col_keys(x, j, is.logical)

  quo_fun <- quo(format_fun.logical(
    x,
    true = true,
    false = false,
    na_str = na_str,
    nan_str = nan_str,
    prefix = prefix,
    suffix = suffix
  ))
  fun_ <- new_function(
    pairlist2(
      x = ,
      true = true,
      false = false,
      na_str = na_str,
      nan_str = nan_str,
      prefix = prefix,
      suffix = suffix
    ),
    get_expr(quo_fun)
  )

  docall_display(col_keys, fun_, x, i = i)
}

#' @title Format cells as images
#' @description Format image paths as images in a flextable.
#' @inheritParams colformat_char
#' @param width,height size of the png file in inches
#' @param alt alternative text for the images (used for accessibility)
#' @family cells formatters
#' @export
#' @examples
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#'
#' dat <- head(iris)
#' dat$Species <- as.character(dat$Species)
#' dat[c(1, 3, 5), "Species"] <- img.file
#'
#' myft <- flextable(dat)
#' myft <- colformat_image(
#'   myft,
#'   i = c(1, 3, 5),
#'   j = "Species", width = .20, height = .15
#' )
#' ft <- autofit(myft)
#' ft
colformat_image <- function(
  x,
  i = NULL,
  j = NULL,
  width,
  height,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = "",
  alt = ""
) {
  stopifnot(inherits(x, "flextable"))

  col_keys <- filter_col_keys(x, j, is.character)

  check_formula_i_and_part(i, "body")
  for (varname in col_keys) {
    x <- compose(
      x = x,
      j = varname,
      i = i,
      value = as_paragraph(
        prefix,
        as_image(get(varname), width = width, height = height, alt = alt),
        suffix
      ),
      part = "body"
    )
  }
  x
}


filter_col_keys <- function(x, j, fun) {
  j <- get_columns_id(x[["body"]], j)
  col_keys <- x$col_keys[j]
  col_keys[vapply(x[["body"]]$dataset[col_keys], fun, FUN.VALUE = NA)]
}

docall_display <- function(col_keys, fun, x, i = NULL, part = "body") {
  check_formula_i_and_part(i, part)
  for (varname in col_keys) {
    x <- mk_par(
      x = x,
      j = varname,
      i = i,
      value = as_paragraph(as_chunk(fun(get(varname)))),
      part = part
    )
  }
  x
}

test_that("surround sets border properties on target cell", {
  ft <- flextable(head(iris, 3))
  ft <- border_remove(ft)
  brd <- officer::fp_border(color = "red", width = 2, style = "solid")
  ft <- surround(ft, i = 2, j = 2, border = brd, part = "body")

  props <- flextable:::information_data_cell(ft)
  cell <- props[
    props$.part == "body" & props$.row_id == 2 & props$.col_id == "Sepal.Width",
  ]

  expect_equal(cell$border.width.top, 2)
  expect_equal(cell$border.width.bottom, 2)
  expect_equal(cell$border.width.left, 2)
  expect_equal(cell$border.width.right, 2)
  expect_equal(cell$border.color.top, "red")
  expect_equal(cell$border.color.bottom, "red")
  expect_equal(cell$border.color.left, "red")
  expect_equal(cell$border.color.right, "red")

  # adjacent cells should not have red borders
  other <- props[
    props$.part == "body" &
      props$.row_id == 1 &
      props$.col_id == "Sepal.Length",
  ]
  expect_equal(other$border.width.top, 0)
})

test_that("surround border shortcut sets all sides", {
  ft <- flextable(head(iris, 3))
  ft <- border_remove(ft)
  brd <- officer::fp_border(color = "blue", width = 1)
  # use cell (2,2) to avoid header/body boundary edge case
  ft <- surround(ft, i = 2, j = 2, border = brd, part = "body")

  props <- flextable:::information_data_cell(ft)
  cell <- props[
    props$.part == "body" & props$.row_id == 2 & props$.col_id == "Sepal.Width",
  ]

  for (side in c("top", "bottom", "left", "right")) {
    expect_equal(cell[[paste0("border.color.", side)]], "blue")
    expect_equal(cell[[paste0("border.width.", side)]], 1)
  }
})

test_that("surround selective sides", {
  ft <- flextable(head(iris, 3))
  ft <- border_remove(ft)
  brd <- officer::fp_border(color = "green", width = 3)
  ft <- surround(
    ft,
    i = 2,
    j = 2,
    border.top = brd,
    border.bottom = brd,
    part = "body"
  )

  props <- flextable:::information_data_cell(ft)
  cell <- props[
    props$.part == "body" & props$.row_id == 2 & props$.col_id == "Sepal.Width",
  ]

  expect_equal(cell$border.width.top, 3)
  expect_equal(cell$border.width.bottom, 3)
  expect_equal(cell$border.width.left, 0)
  expect_equal(cell$border.width.right, 0)
})

test_that("surround renders in HTML", {
  ft <- flextable(head(iris, 2))
  ft <- border_remove(ft)
  brd <- officer::fp_border(color = "red", width = 2, style = "solid")
  ft <- surround(ft, i = 1, j = 1, border = brd, part = "body")

  html <- flextable:::gen_raw_html(ft)
  expect_match(html, "border-top: 2pt solid rgba(255, 0, 0", fixed = TRUE)
  expect_match(html, "border-bottom: 2pt solid rgba(255, 0, 0", fixed = TRUE)
  expect_match(html, "border-left: 2pt solid rgba(255, 0, 0", fixed = TRUE)
  expect_match(html, "border-right: 2pt solid rgba(255, 0, 0", fixed = TRUE)
})

test_that("surround renders in LaTeX", {
  ft <- flextable(head(iris, 2))
  ft <- border_remove(ft)
  brd <- officer::fp_border(color = "red", width = 2, style = "solid")
  ft <- surround(ft, i = 1, j = 1, border = brd, part = "body")

  latex <- flextable:::gen_raw_latex(ft)
  # horizontal lines: ascline with red color (FF0000)
  expect_match(latex, "\\ascline{2pt}{FF0000}{1-1}", fixed = TRUE)
  # vertical lines: vrule with red color
  expect_match(latex, "\\color[HTML]{FF0000}\\vrule width 2pt", fixed = TRUE)
})

test_that("void works as expected", {
  expect_error(void(12, part = "all"))

  ftab <- flextable(head(mtcars))
  ftab <- void(ftab, part = "all")
  expect_true(all(information_data_chunk(ftab)$txt %in% ""))
})

z <-
  structure(
    list(
      name = c(
        "Matthieu Guillou-Poulain",
        "Noémi Pasquier d'Vaillant",
        "Honoré L'Delannoy",
        "Alice L'Bonneau",
        "Adrien Dupuy"
      ),
      birthday = structure(c(25463, 7642, 13950, 23805, 21243), class = "Date"),
      n_children = c(4, 4, 4, 4, 2),
      weight = c(
        83.6459598876536,
        61.8819103203714,
        NA,
        52.6329895108938,
        53.5817482229322
      ),
      height = c(
        163.589849524116,
        171.920463847195,
        169.474969328653,
        164.460310102575,
        167.755981621553
      ),
      n_peanuts = c(821107L, 774581L, 721301L, 1116933L, 1009038L),
      eye_color = structure(
        c(2L, 1L, 1L, 1L, 1L),
        .Label = c("dark", "green"),
        class = "factor"
      )
    ),
    row.names = c(NA, -5L),
    class = "data.frame"
  )
z$timestamp <- as.POSIXct("2011-01-01 09:09:09")

test_that("flextable_defaults values for cell content", {
  set_flextable_defaults(
    decimal.mark = ",",
    big.mark = " ",
    digits = 1,
    na_str = "NA",
    fmt_date = "%d %m %Y",
    fmt_datetime = "%d/%m/%Y %H:%M:%S"
  )

  ft <- flextable(z)
  expected <-
    c(
      "name",
      "birthday",
      "n_children",
      "weight",
      "height",
      "n_peanuts",
      "eye_color",
      "timestamp",
      "Matthieu Guillou-Poulain",
      "19 09 2039",
      "4",
      "83,64596",
      "163,5898",
      "821 107",
      "green",
      "01/01/2011 09:09:09",
      "Noémi Pasquier d'Vaillant",
      "04 12 1990",
      "4",
      "61,88191",
      "171,9205",
      "774 581",
      "dark",
      "01/01/2011 09:09:09",
      "Honoré L'Delannoy",
      "12 03 2008",
      "4",
      "NA",
      "169,4750",
      "721 301",
      "dark",
      "01/01/2011 09:09:09",
      "Alice L'Bonneau",
      "06 03 2035",
      "4",
      "52,63299",
      "164,4603",
      "1 116 933",
      "dark",
      "01/01/2011 09:09:09",
      "Adrien Dupuy",
      "29 02 2028",
      "2",
      "53,58175",
      "167,7560",
      "1 009 038",
      "dark",
      "01/01/2011 09:09:09"
    )
  expect_equal(
    object = information_data_chunk(ft)$txt,
    expected,
    ignore_attr = TRUE
  )

  init_flextable_defaults()
})

test_that("colformat_* functions", {
  dat <- data.frame(
    letters1 = c("a", "b", "b", "c"),
    letters2 = c("d", "e", "b", "b"),
    number = 1:4 * pi,
    count = as.integer(1:4),
    is_something = c(TRUE, FALSE, TRUE, FALSE),
    dt = as.POSIXct("2011-01-01 09:09:09") - 1:4,
    date = as.Date("2011-02-23") + 1:4,
    stringsAsFactors = FALSE
  )

  ft <- flextable(dat)
  ft <- colformat_char(x = ft, prefix = "-", suffix = "-")
  ft <- colformat_date(
    x = ft,
    fmt_date = "%d/%m/%Y",
    prefix = "-",
    suffix = "-"
  )
  ft <- colformat_datetime(
    x = ft,
    fmt_date = "%d/%m/%Y %H%M%S",
    prefix = "-",
    suffix = "-"
  )
  ft <- colformat_double(
    x = ft,
    big.mark = "",
    decimal.mark = ",",
    digits = 3,
    prefix = "-",
    suffix = "-"
  )
  ft <- colformat_int(x = ft, prefix = "-", suffix = "-")
  ft <- colformat_lgl(
    x = ft,
    true = "OUI",
    false = "NON",
    prefix = "-",
    suffix = "-"
  )

  expected <-
    c(
      "letters1",
      "letters2",
      "number",
      "count",
      "is_something",
      "dt",
      "date",
      "-a-",
      "-d-",
      "-3,142-",
      "-1-",
      "-OUI-",
      "-01/01/2011 090908-",
      "-24/02/2011-",
      "-b-",
      "-e-",
      "-6,283-",
      "-2-",
      "-NON-",
      "-01/01/2011 090907-",
      "-25/02/2011-",
      "-b-",
      "-b-",
      "-9,425-",
      "-3-",
      "-OUI-",
      "-01/01/2011 090906-",
      "-26/02/2011-",
      "-c-",
      "-b-",
      "-12,566-",
      "-4-",
      "-NON-",
      "-01/01/2011 090905-",
      "-27/02/2011-"
    )
  expect_equal(
    object = information_data_chunk(ft)$txt,
    expected,
    ignore_attr = TRUE
  )

  ft <- colformat_num(
    x = ft,
    big.mark = "",
    decimal.mark = ".",
    prefix = "+",
    suffix = "+"
  )
  expected <-
    c(
      "letters1",
      "letters2",
      "number",
      "count",
      "is_something",
      "dt",
      "date",
      "-a-",
      "-d-",
      "+3.141593+",
      "+1+",
      "-OUI-",
      "-01/01/2011 090908-",
      "-24/02/2011-",
      "-b-",
      "-e-",
      "+6.283185+",
      "+2+",
      "-NON-",
      "-01/01/2011 090907-",
      "-25/02/2011-",
      "-b-",
      "-b-",
      "+9.424778+",
      "+3+",
      "-OUI-",
      "-01/01/2011 090906-",
      "-26/02/2011-",
      "-c-",
      "-b-",
      "+12.566371+",
      "+4+",
      "-NON-",
      "-01/01/2011 090905-",
      "-27/02/2011-"
    )
  expect_equal(
    object = information_data_chunk(ft)$txt,
    expected,
    ignore_attr = TRUE
  )
})


test_that("append and prepend chunks structure", {
  expect_error(append_chunks(12))
  expect_error(prepend_chunks(12))

  ftab <- flextable(head(cars, n = 3))
  ftab <- append_chunks(
    ftab,
    j = 1,
    as_chunk(" Samurai"),
    colorize(as_i(" Shodown"), color = "magenta")
  )

  expect_equal(
    information_data_chunk(ftab)$txt,
    expected = c(
      "speed",
      "dist",
      "4",
      " Samurai",
      " Shodown",
      "2",
      "4",
      " Samurai",
      " Shodown",
      "10",
      "7",
      " Samurai",
      " Shodown",
      "4"
    )
  )

  ftab <- flextable(head(cars, n = 3))
  ftab <- prepend_chunks(
    ftab,
    j = 1,
    as_chunk("Samurai"),
    colorize(as_i(" Shodown "), color = "magenta")
  )

  expect_equal(
    information_data_chunk(ftab)$txt,
    expected = c(
      "speed",
      "dist",
      "Samurai",
      " Shodown ",
      "4",
      "2",
      "Samurai",
      " Shodown ",
      "4",
      "10",
      "Samurai",
      " Shodown ",
      "7",
      "4"
    )
  )
})

test_that("delete rows and columns", {
  ftab <- flextable(head(iris))
  ftab <- delete_columns(ftab, j = 1:3)
  ftab <- delete_rows(ftab, i = 1:5, part = "body")
  ftab <- autofit(ftab)

  expect_equal(
    information_data_chunk(ftab)$txt,
    expected = c("Petal.Width", "Species", "0.4", "setosa")
  )
  expect_equal(ftab$col_keys, expected = c("Petal.Width", "Species"))
  expect_equal(ftab$body$content$nrow, expected = 1L)
  expect_equal(ftab$body$content$ncol, expected = 2L)
})

test_that("labelizor", {
  zzz <- CO2[, -c(1, 4)]
  zzz[1, 1] <- NA
  zzz[47, 1] <- NA

  z <- summarizor(x = zzz, by = "Treatment", overall_label = "Overall")
  ftab <- as_flextable(z, separate_with = "variable")
  ftab <- labelizor(
    x = ftab,
    j = c("stat", "nonchilled@blah"),
    labels = c(Missing = "Kouign amann", "20 (47.6%)" = "plop")
  )
  chunk_txt <- information_data_chunk(ftab)$txt
  expect_equal(chunk_txt[32], expected = "Kouign amann")
  expect_equal(chunk_txt[18], expected = "plop")
  expect_equal(chunk_txt[26], expected = "plop")
})

#' @export
#' @title Number of rows of a part
#' @family functions for flextable size management
#' @description returns the number of lines in a part of flextable.
#' @inheritParams args_x_part
#' @examples
#' library(flextable)
#' ft <- qflextable(head(cars))
#' nrow_part(ft, part = "body")
nrow_part <- function(x, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "nrow_part()"
    ))
  }
  if (is.null(x[[part]])) {
    0
  } else if (is.null(x[[part]]$dataset)) {
    0
  } else {
    nrow(x[[part]]$dataset)
  }
}

#' @export
#' @title Number of columns
#' @family functions for flextable size management
#' @description returns the number of columns displayed
#' @inheritParams args_x_only
#' @examples
#' library(flextable)
#' ft <- qflextable(head(cars))
#' ncol_keys(ft)
ncol_keys <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "ncol_keys()"
    ))
  }
  length(x$col_keys)
}

#' @export
#' @title data.frame automatic printing as a flextable
#' @description Create a summary from a data.frame as a flextable. This function
#' is to be used in an R Markdown document.
#'
#' To use that function, you must declare it in the part `df_print` of the 'YAML'
#' header of your R Markdown document:
#'
#' ```
#' ---
#' df_print: !expr function(x) flextable::df_printer(x)
#' ---
#' ```
#'
#' We notice an unexpected behavior with bookdown. When using bookdown it
#' is necessary to use [use_df_printer()] instead in a setup run chunk:
#'
#' ```
#' use_df_printer()
#' ```
#'
#' @param dat the data.frame
#' @param ... unused argument
#' @details
#' 'knitr' chunk options are available to customize the output:
#'
#' * `ft_max_row`: The number of rows to print. Default to 10.
#' * `ft_split_colnames`: Should the column names be split
#' (with non alpha-numeric characters). Default to FALSE.
#' * `ft_short_strings`: Should the character column be shorten.
#' Default to FALSE.
#' * `ft_short_size`: Maximum length of character column if
#' `ft_short_strings` is TRUE. Default to 35.
#' * `ft_short_suffix`: Suffix to add when character values are shorten.
#' Default to "...".
#' * `ft_do_autofit`: Use autofit() before rendering the table.
#' Default to TRUE.
#' * `ft_show_coltype`: Show column types.
#' Default to TRUE.
#' * `ft_color_coltype`: Color to use for column types.
#' Default to "#999999".
#' @family flextable_output_export
#' @examples
#' df_printer(head(mtcars))
df_printer <- function(dat, ...) {
  args <- Filter(
    function(z) !is.null(z),
    list(
      max_row = knitr::opts_current$get("ft_max_row"),
      split_colnames = knitr::opts_current$get("ft_split_colnames"),
      short_strings = knitr::opts_current$get("ft_short_strings"),
      short_size = knitr::opts_current$get("ft_short_size"),
      short_suffix = knitr::opts_current$get("ft_short_suffix"),
      do_autofit = knitr::opts_current$get("ft_do_autofit"),
      show_coltype = knitr::opts_current$get("ft_show_coltype"),
      color_coltype = knitr::opts_current$get("ft_color_coltype")
    )
  )
  args$x <- dat

  if (is.null(knitr::pandoc_to())) {
    message("this function is to be used in a knitr context.")
    return(invisible(FALSE))
  }

  knitr::knit_print(
    do.call(as_flextable, args)
  )
}

#' @export
#' @title Set data.frame automatic printing as a flextable
#' @description Define [df_printer()] as data.frame
#' print method in an R Markdown document.
#'
#' In a setup run chunk:
#'
#' ```
#' flextable::use_df_printer()
#' ```
#' @seealso [df_printer()], [flextable()]
#' @family flextable_configuration
use_df_printer <- function() {
  registerS3method("knit_print", "data.frame", df_printer)
  registerS3method("knit_print", "data.table", df_printer)
  registerS3method("knit_print", "tibble", df_printer)
  registerS3method("knit_print", "grouped_df", df_printer)
  registerS3method("knit_print", "spec_tbl_df", df_printer)
  registerS3method("knit_print", "tbl", df_printer)
  registerS3method("knit_print", "table", df_printer)
  invisible()
}

look_like_int <- function(x) {
  (is.numeric(x) && isTRUE(all.equal(x, as.integer(x)))) || is.integer(x)
}

multirow_df_printer <- function(
  dat,
  max_row = 10,
  split_colnames = FALSE,
  short_strings = FALSE,
  short_size = 35,
  short_suffix = "...",
  do_autofit = TRUE,
  show_coltype = TRUE,
  color_coltype = "#999999"
) {
  x <- as.data.frame(dat)
  nro <- nrow(x)

  z <- get_flextable_defaults()

  x <- head(x, n = max_row)
  coltypes <- as.character(sapply(dat, function(x) head(class(x), 1)))

  lli <- sapply(x, look_like_int)
  x[lli] <- lapply(x[lli], as.integer)

  if (!is.null(short_strings) && short_strings) {
    wic <- sapply(x, is.character)
    x[wic] <- lapply(x[wic], function(x) {
      paste0(substring(text = x, first = 1, last = short_size), short_suffix)
    })
  }

  colkeys <- colnames(x)

  ft <- flextable(x, col_keys = colkeys)

  if (split_colnames) {
    labs <- strsplit(colkeys, split = "[^[:alnum:]]+")
    names(labs) <- colkeys
    labs <- lapply(labs, paste, collapse = "\n")
    ft <- set_header_labels(ft, values = labs)
  }

  if (show_coltype) {
    ft <- add_header_row(ft, top = FALSE, values = coltypes)
  }

  ft <- colformat_double(ft)
  ft <- colformat_int(ft)

  ft <- add_footer_lines(ft, values = sprintf("n: %.0f", nro))

  ft <- set_table_properties(ft, layout = z$table.layout)
  if ("fixed" %in% z$table.layout && do_autofit) {
    ft <- autofit(ft)
  }

  ft <- do.call(z$theme_fun, list(ft))

  if (show_coltype) {
    ft <- color(
      ft,
      i = nrow_part(ft, "header"),
      part = "header",
      color = color_coltype
    )
  }
  ft <- align(ft, align = "left", part = "footer")
  ft
}

singlerow_df_printer <- function(
  dat,
  max_row = 10,
  split_colnames = FALSE,
  short_strings = FALSE,
  short_size = 35,
  short_suffix = "...",
  do_autofit = TRUE,
  show_coltype = TRUE,
  color_coltype = "#999999"
) {
  coltypes <- as.character(sapply(dat, function(x) head(class(x), 1)))

  lli <- sapply(dat, look_like_int)
  dat[lli] <- lapply(dat[lli], as.integer)

  x <- data.frame(
    "Col." = colnames(dat),
    "Type" = coltypes,
    "Val." = vapply(dat, format_fun.default, FUN.VALUE = NA_character_)
  )

  z <- get_flextable_defaults()

  x <- head(x, n = max_row)

  if (!is.null(short_strings) && short_strings) {
    wic <- sapply(x, is.character)
    x[wic] <- lapply(x[wic], function(x) {
      paste0(substring(text = x, first = 1, last = short_size), short_suffix)
    })
  }

  colkeys <- c("Col.", "Val.")

  ft <- flextable(x, col_keys = colkeys)
  ft <- delete_part(ft, part = "header")
  ft <- set_table_properties(ft, layout = z$table.layout)
  if ("fixed" %in% z$table.layout && do_autofit) {
    ft <- autofit(ft)
  }
  ft <- do.call(z$theme_fun, list(ft))
  ft <- align(ft, align = "center", part = "all")
  ft <- align(ft, j = 1, align = "right", part = "all")
  ft <- valign(x = ft, valign = "top", part = "body")
  if (show_coltype) {
    ft <- append_chunks(
      x = ft,
      j = "Col.",
      as_chunk("\n"),
      colorize(
        x = as_chunk(
          Type,
          props = fp_text_default(font.size = z$font.size * 2 / 3)
        ),
        color = color_coltype
      )
    )
  }
  ft
}

#' @export
#' @title Transform and summarise a 'data.frame' into a flextable
#' Simple summary of a data.frame as a flextable
#' @description It displays the first rows and shows the column types.
#' If there is only one row, a simplified vertical table is produced.
#' @param x a data.frame
#' @param max_row The number of rows to print. Default to 10.
#' @param split_colnames Should the column names be split
#' (with non alpha-numeric characters). Default to FALSE.
#' @param short_strings Should the character column be shorten.
#' Default to FALSE.
#' @param short_size Maximum length of character column if
#' `short_strings` is TRUE. Default to 35.
#' @param short_suffix Suffix to add when character values are shorten.
#' Default to "...".
#' @param do_autofit Use [autofit()] before rendering the table.
#' Default to TRUE.
#' @param show_coltype Show column types.
#' Default to TRUE.
#' @param color_coltype Color to use for column types.
#' Default to "#999999".
#' @param ... unused arguments
#' @examples
#' as_flextable(mtcars)
#' @family as_flextable methods
as_flextable.data.frame <- function(
  x,
  max_row = 10,
  split_colnames = FALSE,
  short_strings = FALSE,
  short_size = 35,
  short_suffix = "...",
  do_autofit = TRUE,
  show_coltype = TRUE,
  color_coltype = "#999999",
  ...
) {
  if (inherits(x, "data.table")) {
    x <- as.data.frame(x)
  } else if (inherits(x, "tbl_df")) {
    x <- as.data.frame(x)
  }

  if (nrow(x) == 1) {
    singlerow_df_printer(
      dat = x,
      max_row = max_row,
      short_strings = short_strings,
      short_size = short_size,
      short_suffix = short_suffix,
      do_autofit = do_autofit,
      show_coltype = show_coltype,
      color_coltype = color_coltype
    )
  } else {
    multirow_df_printer(
      dat = x,
      max_row = max_row,
      split_colnames = split_colnames,
      short_strings = short_strings,
      short_size = short_size,
      short_suffix = short_suffix,
      do_autofit = do_autofit,
      show_coltype = show_coltype,
      color_coltype = color_coltype
    )
  }
}

#' @export
#' @title Transform a 'table' object into a flextable
#' @description produce a flextable describing a
#' count table produced by function `table()`.
#'
#' This function uses the [proc_freq()] function.
#' @param x table object
#' @param ... arguments used by [proc_freq()].
#' @examples
#' tab <- with(warpbreaks, table(wool, tension))
#' ft <- as_flextable(tab)
#' ft
#' @family as_flextable methods
as_flextable.table <- function(x, ...) {
  x <- as.data.frame(x)
  by <- setdiff(colnames(x), "Freq")
  if (length(by) > 2) {
    stop("table must have maximum two dimensions.")
  } else if (length(by) > 1) {
    proc_freq(x, row = by[1], col = by[2], weight = "Freq", ...)
  } else {
    proc_freq(x, row = by[1], weight = "Freq", ...)
  }
}

# args_x_only -----

#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @name args_x_only
#' @keywords internal
NULL

# args_x_part -----

#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' can be used.
#' @name args_x_part
#' @keywords internal
NULL

# args_x_part_no_all -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' is not allowed by the function.
#' @name args_x_part_no_all
#' @keywords internal
NULL

# args_x_i_part -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param i row selector, see section *Row selection with the `i` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' can be used.
#' @name args_x_i_part
#' @keywords internal
NULL

# args_x_i_part_no_all -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param i row selector, see section *Row selection with the `i` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' is not allowed by the function.
#' @name args_x_i_part_no_all
#' @keywords internal
NULL


# args_x_i_j -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param i row selector, see section *Row selection with the `i` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param j column selector, see section *Column selection with the `j` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @name args_x_i_j
#' @keywords internal
NULL

# args_x_j -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param j column selector, see section *Column selection with the `j` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @name args_x_j
#' @keywords internal
NULL

# args_x_j_part -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param j column selector, see section *Column selection with the `j` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' can be used.
#' @name args_x_j_part
#' @keywords internal
NULL

# args_x_j_part_no_all -----
#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param j column selector, see section *Column selection with the `j` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' is not allowed by the function.
#' @name args_x_j_part_no_all
#' @keywords internal
NULL

# args_selectors_with_all -----

#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param i row selector, see section *Row selection with the `i` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param j column selector, see section *Column selection with the `j` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' can be used.
#' @name args_selectors_with_all
#' @keywords internal
NULL

# args_selectors_without_all -----

#' @title internal utils for roxygen tags reuse
#' @param x a 'flextable' object, see [flextable-package] to learn how to create
#' 'flextable' object.
#' @param i row selector, see section *Row selection with the `i` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param j column selector, see section *Column selection with the `j` parameter*
#' in <[`Selectors in flextable`][flextable_selectors]>.
#' @param part part selector, see section *Part selection with the `part`
#' parameter* in <[`Selectors in flextable`][flextable_selectors]>.
#' Value 'all' is not allowed by the function.
#' @name args_selectors_without_all
#' @keywords internal
NULL

# flextable_selectors: Selectors in flextable -----

#' @title Selectors in flextable
#'
#' @description
#' Selectors are a core feature of flextable that allow you to specify which
#' parts (`part`), rows (`i`) and columns (`j`) should be affected by formatting,
#' styling, or content operations.
#'
#' Many flextable functions support these selectors, including [bg()],
#' [bold()], [color()], [padding()], [fontsize()], [italic()], [align()],
#' [mk_par()], [hline()], [vline()], and many others.
#'
#' Selectors make conditional formatting easy and enable seamless piping of
#' multiple operations using `%>%` or `|>`.
#'
#' @section Row selection with the `i` parameter:
#'
#' The `i` parameter is used to select specific rows in a flextable for
#' formatting, styling, or content operations.
#'
#' When `i = NULL` (the default), operations apply to all rows.
#'
#' ```r
#' i = ~ condition           # Formula (body only)
#' i = 1:5                   # Integer vector
#' i = c(TRUE, FALSE, ...)   # Logical vector
#' i = NULL                  # All rows (default)
#' ```
#'
#' Best practices:
#'
#' - Use formulas (`i = ~ condition`) for conditional selection in body
#' - Use integers for positional selection in any part.
#' - Use `nrow_part()` for dynamic row selection
#'
#'
#' @section Column selection with the `j` parameter:
#' The `j` parameter is used to select specific columns in a flextable for
#' formatting, styling, or content operations.
#'
#' When `j = NULL` (the default), operations apply to all columns.
#'
#' ```r
#' j = ~ col1 + col2         # Formula (select multiple)
#' j = ~ . - col1            # Formula (exclude columns)
#' j = c("col1", "col2")     # Character vector (recommended)
#' j = 1:5                   # Integer vector
#' j = c(TRUE, FALSE, ...)   # Logical vector
#' j = NULL                  # All columns (default)
#' ```
#'
#' Best practices:
#'
#' - Use character vectors (`j = c("col1", "col2")`) for clarity and
#' maintainability.
#' - Use formulas (`j = ~ col1 + col2`) for excluding columns.
#' - Avoid integer positions when possible (less maintainable).
#' - Column selectors work with all parts (header, body, footer, all).
#'
#'
#' @section Part selection with the `part` parameter:
#'
#' The `part` parameter specifies which section of the flextable should be
#' affected by formatting, styling, or content operations.
#'
#' ```r
#' part = "body"    # Data rows (default)
#' part = "header"  # Header rows
#' part = "footer"  # Footer rows
#' part = "all"     # All parts
#' ```
#'
#' - When `part = "body"` (the default), operations apply only to the data rows.
#' - When `part = "all"`, the operation is applied to each part independently.
#'
#' Formula row selectors (`i = ~ condition`) CANNOT be used with:
#'
#' - `part = "header"` - Headers contain only character values
#' - `part = "footer"` - Footers contain only character values
#' - `part = "all"` - Header/footer are character-only while body has original data types
#'
#' This restriction exists because formula selectors evaluate conditions using
#' the actual data types from your dataset (numeric, logical, etc.). Headers and
#' footers store only character representations of values, so conditional
#' expressions like `i = ~ price < 330` cannot be evaluated on them.
#'
#' @section Advanced Programming with Selectors:
#'
#' Function [nrow_part()] returns the number of lines in a part.
#'
#' ```r
#' # Format the last row differently
#' ft <- flextable(iris[48:52,])
#' ft <- bold(ft, i = nrow_part(ft, part = "body"))
#'
#' # Add footnote to last row
#' ft <- footnote(
#'   ft,
#'   i = nrow_part(ft, part = "body"), j = 1:4,
#'   value = as_paragraph("Calculated mean")
#' )
#' ft
#' ```
#'
#' Function [ncol_keys()] returns the number of columns.
#'
#' ```r
#' # Get column count
#' ncol_keys(ft)
#'
#' # Useful for programmatic selection
#' ncols <- ncol_keys(ft)
#' ft <- bg(ft, j = rep(c(TRUE, FALSE), length.out = ncols), bg = "#f0f0f0")
#' ft
#' ```
#' @family selectors in flextable
#' @keywords internal
#' @name flextable_selectors
NULL

#' @export
#' @title Add latex dependencies
#' @description Manually add flextable latex dependencies to
#' the knitr session via [knitr::knit_meta_add()].
#'
#' When enabling caching in 'R Markdown' documents for PDF output,
#' the flextable cached result is used directly. Call `add_latex_dep()` in a
#' non cached chunk so that flextable latex dependencies are added
#' to knitr metadata.
#' @param float load package 'float'
#' @param wrapfig load package 'wrapfig'
#' @return NULL
#' @examples
#' add_latex_dep()
#' @seealso [knit_print.flextable()]
#' @family flextable_configuration
#' @keywords internal
#' @importFrom knitr knit_meta_add
add_latex_dep <- function(float = FALSE, wrapfig = FALSE) {
  knit_meta_add(unname(list_latex_dep(float = float, wrapfig = wrapfig)))

  invisible(NULL)
}

#' @importFrom rmarkdown latex_dependency
list_latex_dep <- function(float = FALSE, wrapfig = FALSE) {
  if (!is_latex_output()) {
    return(list())
  }

  fonts_ignore <- flextable_global$defaults$fonts_ignore
  fontspec_compat <- get_pdf_engine() %in% c("xelatex", "lualatex")
  if (!fonts_ignore && !fontspec_compat) {
    warning(
      "fonts used in `flextable` are ignored because ",
      "the `pdflatex` engine is used and not `xelatex` or ",
      "`lualatex`. You can avoid this warning by using ",
      "`set_flextable_defaults(fonts_ignore=TRUE)` or use a ",
      "compatible engine (e.g. `pdf-engine: xelatex`).",
      call. = FALSE
    )
  }

  x <- list()

  if (fontspec_compat) {
    x$fontspec <- latex_dependency("fontspec")
  }
  x$multirow <- latex_dependency("multirow")
  x$multicol <- latex_dependency("multicol")
  x$colortbl <- latex_dependency("colortbl")
  x$ulem <- latex_dependency("ulem", options = "normalem")
  x$hhline <- latex_dependency(
    name = "hhline",
    extra_lines = c(
      "\\newlength\\Oldarrayrulewidth",
      "\\newlength\\Oldtabcolsep"
    )
  )
  x$longtable <- latex_dependency("longtable")
  x$array <- latex_dependency("array")
  x$hyperref <- latex_dependency("hyperref")

  if (float) {
    x$float <- latex_dependency("float")
  }
  if (wrapfig) {
    x$wrapfig <- latex_dependency("wrapfig")
  }
  x
}


gen_raw_latex <- function(
  x,
  lat_container = latex_container_none(),
  caption = "",
  topcaption = TRUE,
  quarto = FALSE
) {
  dims <- dim(x)
  column_sizes <- dims$widths
  column_sizes_df <- data.frame(
    column_size = column_sizes,
    .col_id = factor(x$col_keys, levels = x$col_keys),
    stringsAsFactors = FALSE
  )

  properties_df <- merge_table_properties(x)
  linespacing_df <- properties_df[, c(
    ".part",
    ".row_id",
    ".col_id",
    "line_spacing"
  )]
  dat <- runs_as_latex(
    x = x,
    chunk_data = information_data_chunk(x),
    ls_df = linespacing_df
  )

  # hhlines and vborders ----
  latex_borders_data_str <- latex_gridlines(properties_df)
  properties_df <- merge(
    properties_df,
    latex_borders_data_str$vlines,
    by = c(".part", ".col_id", ".row_id")
  )
  properties_df[,
    setdiff(
      grep("^border\\.", colnames(properties_df), value = TRUE),
      "border.width.left"
    ) := NULL
  ]
  setorderv(properties_df, c(".part", ".col_id", ".row_id"))

  # cell background color -----
  properties_df[,
    c("background_color") := list(
      latex_cell_bgcolor(.SD$background.color)
    )
  ]
  properties_df[, c("background.color") := NULL]

  # text direction ----
  properties_df[,
    c("text_direction_left", "text_direction_right") := list(
      latex_text_direction(.SD$text.direction, left = TRUE),
      latex_text_direction(.SD$text.direction, left = FALSE)
    )
  ]
  properties_df[, c("text.direction") := NULL]

  # merge prop and text and sizes----
  cell_properties_df <- merge(
    properties_df,
    dat,
    by = c(".part", ".row_id", ".col_id")
  )
  cell_properties_df <- merge(
    cell_properties_df,
    column_sizes_df,
    by = c(".col_id")
  )

  # update colspan -----
  cell_properties_df <- reverse_colspan(cell_properties_df)
  # add col sizes -----
  column_sizes_df <- calc_column_size(cell_properties_df, x$col_keys)
  cell_properties_df[, c("column_size") := NULL]
  cell_properties_df <- merge(
    cell_properties_df,
    column_sizes_df,
    by = c(".part", ".row_id", ".col_id")
  )

  # latex for multicolumn + add vert lines ----
  if ("fixed" %in% x$properties$layout) {
    augment_multicolumn_fixed(cell_properties_df)
  } else {
    augment_multicolumn_autofit(cell_properties_df)
  }

  # latex for multirow ----
  augment_multirow_fixed(cell_properties_df)

  # paragraph indents (hanging wins over first_line, as in officer) ----
  # `\hangindent` indents lines 2+ from the left, so the Word semantics
  # (all lines at padding.left, first line outdented by `hanging`)
  # translate to leftskip = padding.left - hanging plus
  # \hangindent=hanging. `first_line` is a \hspace* on the first line.
  cell_properties_df[,
    c("hanging_pt", "first_line_pt") := list(
      fifelse(!is.na(.SD$hanging) & .SD$hanging > 0, .SD$hanging, 0),
      fifelse(
        (is.na(.SD$hanging) | .SD$hanging <= 0) &
          !is.na(.SD$first_line) &
          .SD$first_line > 0,
        .SD$first_line,
        0
      )
    )
  ]

  # paragraph padding ----
  cell_properties_df[,
    c("padding_left_str", "padding_right_str") := list(
      ifelse(
        abs(.SD$padding.left - .SD$hanging_pt) > 0,
        sprintf(
          "\\advance\\leftskip by %.1fpt\\relax ",
          .SD$padding.left - .SD$hanging_pt
        ),
        ""
      ),
      ifelse(
        .SD$padding.right > 0,
        sprintf("\\advance\\rightskip by %.1fpt\\relax ", .SD$padding.right),
        ""
      )
    )
  ]

  cell_properties_df[,
    c("indent_str") := list(
      paste0(
        ifelse(
          .SD$hanging_pt > 0,
          sprintf("\\hangindent=%.1fpt\\hangafter=1\\relax ", .SD$hanging_pt),
          ""
        ),
        ifelse(
          .SD$first_line_pt > 0,
          sprintf("\\hspace*{%.1fpt}", .SD$first_line_pt),
          ""
        )
      )
    )
  ]
  cell_properties_df[, c("hanging_pt", "first_line_pt") := NULL]

  # paste everything ----
  cell_properties_df[,
    c("txt") := list(
      paste0(
        .SD$multirow_left,
        .SD$padding_left_str,
        .SD$padding_right_str,
        .SD$indent_str,
        .SD$text_direction_left,
        .SD$txt,
        .SD$text_direction_right,
        .SD$multirow_right
      )
    )
  ]

  cell_properties_df[cell_properties_df$colspan < 1, c("txt") := list("")]

  cell_properties_df[,
    c("txt") := list(
      paste0(
        .SD$multicolumn_left,
        .SD$txt,
        .SD$multicolumn_right
      )
    )
  ]
  cell_properties_df[
    cell_properties_df$rowspan < 1,
    c("txt") := list(
      NA_character_
    )
  ]
  setorderv(cell_properties_df, c(".part", ".row_id", ".col_id"))
  txt_data <- cell_properties_df[,
    list(txt = paste0(.SD$txt[!is.na(.SD$txt)], collapse = " & ")),
    by = c(".part", ".row_id")
  ]

  # txt_data is now merged by row ----
  txt_data[, c("txt") := list(paste0(.SD$txt, " \\\\"))]
  setorderv(txt_data, c(".part", ".row_id"))

  # add horiz lines ----
  txt_data <- merge(
    txt_data,
    latex_borders_data_str$hlines,
    by = c(".part", ".row_id"),
    all.x = TRUE,
    all.y = TRUE
  )
  setorderv(txt_data, cols = c(".part", ".row_id"))
  txt_data <- augment_part_separators(
    z = txt_data,
    no_container = inherits(lat_container, "latex_container_none"),
    repeat_footer = x$properties$opts_pdf$footer_repeat
  )

  txt_data[,
    c("txt") := list(paste(
      .SD$hlines_t_strings,
      .SD$txt,
      .SD$hlines_b_strings,
      sep = "\n\n"
    ))
  ]

  if (inherits(lat_container, "latex_container_none")) {
    txt_data$.part <- factor(
      as.character(txt_data$.part),
      levels = c("header", "footer", "body")
    )
  } else {
    txt_data$.part <- factor(
      as.character(txt_data$.part),
      levels = c("header", "body", "footer")
    )
  }
  setorderv(txt_data, c(".part", ".row_id"))

  # finalize ----
  if ("fixed" %in% x$properties$layout) {
    column_sizes_latex <- sprintf("|p{%.2fin}", column_sizes)
  } else {
    column_sizes_latex <- rep("c", length(dims$widths))
  }

  align_tag <- latex_table_align(x)

  table_start <- sprintf(
    "\\begin{longtable}[%s]{%s}",
    align_tag,
    paste(column_sizes_latex, collapse = "")
  )
  table_end <- "\\end{longtable}"

  if (x$properties$opts_pdf$caption_repeat && topcaption) {
    second_caption <- gsub("\\caption", "\\caption[]", caption)
    txt_data$part_sep <- gsub(
      "\\endfirsthead",
      paste("\\endfirsthead", second_caption),
      txt_data$part_sep,
      fixed = TRUE
    )
  }

  latex <- paste0(txt_data$txt, txt_data$part_sep, collapse = "\n\n")

  if (inherits(lat_container, "latex_container_wrap")) {
    topcaption <- FALSE
  }

  latex <- paste(
    cline_cmd,
    table_start,
    if (topcaption && !quarto) caption,
    latex,
    if (!topcaption && !quarto) caption,
    table_end,
    sep = "\n\n"
  )

  latex
}

# Adds empty multirow_left/multirow_right columns for vertically merged cells.
#
# \multirow is no longer used because it calculates vertical offsets assuming
# equal row heights, giving wrong results when rows have different heights
# (#639). Instead, reverse_colspan() places content in the correct row
# (first for top, middle for center, last for bottom) and the column type
# (p{} for top, m{} for center, b{} for bottom) handles alignment naturally.
#' @importFrom data.table fcase
augment_multirow_fixed <- function(properties_df) {
  properties_df[, c("multirow_left", "multirow_right") := list("", "")]
  properties_df
}

latex_colwidth <- function(x) {
  grid_dat <- x[,
    .SD,
    .SDcols = c(
      ".part",
      ".row_id",
      ".col_id",
      "rowspan",
      "border.width.left",
      "column_size"
    )
  ]
  grid_dat[,
    c("hspan_id") := list(calc_grid_span_group(.SD$rowspan)),
    by = c(".part", ".row_id")
  ]

  # bdr sum of width
  bdr_dat <- grid_dat[,
    list(
      .col_id = first(.SD$.col_id),
      border_widths = sum(.SD$border.width.left[-1])
    ),
    by = c(".part", ".row_id", "hspan_id")
  ]
  bdr_dat$hspan_id <- NULL
  bdr_dat <- bdr_dat[bdr_dat$border_widths > 0, ]

  grid_dat <- merge(
    grid_dat,
    bdr_dat,
    by = c(".part", ".row_id", ".col_id"),
    all.x = TRUE
  )

  colwidths <- paste0(
    "\\dimexpr ",
    format_double(grid_dat$column_size, 2),
    "in+",
    format_double((grid_dat$rowspan - 1) * 2, digits = 0),
    "\\tabcolsep"
  )

  # if necessary, add bdr width
  bdr_width_instr <- character(length(colwidths))
  bdr_width_instr[!is.na(grid_dat$border_widths)] <-
    paste0(
      "+",
      format_double(
        grid_dat$border_widths[!is.na(grid_dat$border_widths)],
        digits = 2
      ),
      "pt"
    )

  paste0(colwidths, bdr_width_instr)
}


augment_multicolumn_autofit <- function(properties_df) {
  stopifnot(is.data.table(properties_df))

  properties_df[,
    c("multicolumn_left", "multicolumn_right") := list(
      paste0(
        "\\multicolumn{",
        format_double(.SD$rowspan, digits = 0),
        "}{",
        .SD$vborder_left,
        ">{",
        .SD$background_color,
        "}",
        substr(.SD$text.align, 1, 1),
        .SD$vborder_right,
        "}{"
      ),
      "}"
    )
  ]
  properties_df
}
augment_multicolumn_fixed <- function(properties_df) {
  stopifnot(is.data.table(properties_df))

  properties_df[,
    c("multicolumn_left", "multicolumn_right") := list(
      paste0(
        "\\multicolumn{",
        format_double(.SD$rowspan, digits = 0),
        "}{",
        .SD$vborder_left,
        ">{",
        .SD$background_color,
        c(
          "center" = "\\centering",
          left = "\\raggedright",
          right = "\\raggedleft"
        )[.SD$text.align],
        "}",
        c(center = "m", "top" = "p", "bottom" = "b")[.SD$vertical.align],
        "{",
        latex_colwidth(.SD),
        "}",
        .SD$vborder_right,
        "}{"
      ),
      "}"
    )
  ]
  properties_df
}

augment_part_separators <- function(
  z,
  no_container = TRUE,
  repeat_footer = TRUE
) {
  if (repeat_footer) {
    endfoot <- "\\endfoot"
  } else {
    endfoot <- "\\endlastfoot"
  }

  part_separators <- merge(
    z[, c(".part", ".row_id")],
    merge(
      z[, list(.row_id = max(.SD$.row_id)), by = ".part"],
      data.frame(
        .part = factor(
          c("header", "body", "footer"),
          levels = c("header", "body", "footer")
        ),
        part_sep = if (no_container) {
          c("\\endfirsthead", "", endfoot)
        } else {
          c("\\endfirsthead", "", "")
        },
        stringsAsFactors = FALSE
      ),
      by = c(".part")
    ),
    by = c(".part", ".row_id"),
    all.x = TRUE,
    all.y = FALSE
  )
  part_separators$part_sep[is.na(part_separators$part_sep)] <- ""
  setorderv(part_separators, c(".part", ".row_id"))

  z <- merge(z, part_separators, by = c(".part", ".row_id"))

  if ("header" %in% z$.part) {
    z_header <- z[z$.part %in% "header", ]
    z_header$.row_id <- z_header$.row_id + max(z_header$.row_id)
    z_header$part_sep[nrow(z_header)] <- "\\endhead"
    z <- rbind(
      z[z$.part %in% "header", ],
      z_header,
      z[!z$.part %in% "header", ]
    )
  }

  z
}


# col/row spans -----
#' @importFrom stats na.omit
fill_NA <- function(x) {
  which.na <- c(which(!is.na(x)), length(x) + 1)
  values <- na.omit(x)

  if (which.na[1] != 1) {
    which.na <- c(1, which.na)
    values <- c(values[1], values)
  }
  diffs <- diff(which.na)
  return(rep(values, times = diffs))
}


# Handles row placement of content in vertically merged cells (colspan > 1)
# for LaTeX output.
#
# Context: "colspan" here means vertical span (number of rows merged),
# despite the name. A cell with colspan > 1 is the spanning cell that
# holds the content; cells with colspan < 1 are absorbed (empty).
# By default the spanning cell is in the first row of the merged group.
#
# \multirow is NOT used because it calculates offsets assuming equal row
# heights, which gives wrong positions when rows differ in height (#639).
# Instead, content is placed in the correct row and the column type
# (p{} for top, m{} for center, b{} for bottom) handles alignment:
#
# - "top": content stays in the first row, no action needed.
# - "center": content is moved to the middle row (ceiling(N/2)) by
#     swapping .row_id between the first and middle positions.
# - "bottom": content is moved to the last row by reversing .row_id
#     within the merged group.
#
# Vertical borders (vborder_left, vborder_right) are also swapped/reversed
# so they follow the row position they belong to.
reverse_colspan <- function(df) {
  setorderv(df, cols = c(".part", ".col_id", ".row_id"))

  # Assign a unique id (col_uid) to each cell, then propagate the spanning
  # cell's id to its absorbed cells via fill_NA. This groups all cells
  # belonging to the same vertical merge.
  df[,
    c("col_uid") := list(UUIDgenerate(n = nrow(.SD))),
    by = c(".part", ".row_id")
  ]
  df[df$colspan < 1, c("col_uid") := list(NA_character_)]
  df[, c("col_uid") := list(fill_NA(.SD$col_uid)), by = c(".part", ".col_id")]

  # Bottom: reverse row_ids so content moves from first to last row.
  bottom_uids <- unique(df$col_uid[
    df$colspan > 1 & df$vertical.align == "bottom"
  ])
  if (length(bottom_uids) > 0) {
    df[
      df$col_uid %in% bottom_uids,
      c(
        ".row_id",
        "vborder_left",
        "vborder_right"
      ) := list(
        rev(.SD$.row_id),
        rev(.SD$vborder_left),
        rev(.SD$vborder_right)
      ),
      by = c("col_uid")
    ]
  }

  # Center: swap row_ids between first and middle positions so content
  # moves to the middle row. ceiling(N/2) gives the middle index:
  # N=2 -> 1 (no swap), N=3 -> 2, N=4 -> 2, N=5 -> 3, etc.
  center_uids <- unique(df$col_uid[
    df$colspan > 1 & df$vertical.align == "center"
  ])
  if (length(center_uids) > 0) {
    for (uid in center_uids) {
      idx <- which(df$col_uid == uid)
      mid <- ceiling(length(idx) / 2)
      if (mid > 1) {
        for (col_name in c(".row_id", "vborder_left", "vborder_right")) {
          tmp <- df[[col_name]][idx[1]]
          data.table::set(df, idx[1], col_name, df[[col_name]][idx[mid]])
          data.table::set(df, idx[mid], col_name, tmp)
        }
      }
    }
  }

  df[, c("col_uid") := NULL]
  setorderv(df, cols = c(".part", ".row_id", ".col_id"))
  df
}

calc_column_size <- function(df, levels) {
  z <- df[, c(".col_id", ".part", ".row_id", "rowspan", "column_size")]
  z$.col_id <- factor(z$.col_id, levels = levels)
  setorderv(z, cols = c(".part", ".row_id", ".col_id"))
  z[,
    c("col_uid") := list(UUIDgenerate(n = nrow(.SD))),
    by = c(".part", ".row_id")
  ]
  z[z$rowspan < 1, c("col_uid") := list(NA_character_)]
  z[, c("col_uid") := list(fill_NA(.SD$col_uid)), by = c(".part", ".row_id")]
  z[,
    c("column_size") := list(sum(.SD$column_size, na.rm = TRUE)),
    by = c(".part", ".row_id", "col_uid")
  ]
  z[, c("col_uid", "rowspan") := NULL]
  setorderv(z, cols = c(".part", ".row_id", ".col_id"))
  setDT(z)
  z
}

# tools ----
merge_table_properties <- function(x) {
  cell_data <- information_data_cell(x)
  par_data <- information_data_paragraph(x)
  setDT(cell_data)
  setDT(par_data)

  cell_data[, c("width", "height", "hrule") := NULL]
  cell_data <- merge(cell_data, fortify_width(x), by = ".col_id")
  cell_data <- merge(cell_data, fortify_height(x), by = c(".part", ".row_id"))
  cell_data <- merge(
    cell_data,
    fortify_span(x),
    by = c(".part", ".row_id", ".col_id")
  )

  oldnames <- grep("^border\\.", colnames(cell_data), value = TRUE)
  newnames <- paste0("paragraph.", oldnames)
  setnames(par_data, old = oldnames, new = newnames)
  cell_data <- merge(cell_data, par_data, by = c(".part", ".row_id", ".col_id"))
  cell_data$.col_id <- factor(cell_data$.col_id, levels = x$col_keys)
  setorderv(cell_data, c(".part", ".row_id", ".col_id"))
  cell_data
}

#' @importFrom utils compareVersion packageVersion
get_pdf_engine <- function() {
  if (
    compareVersion(as.character(packageVersion("rmarkdown")), "1.10.14") < 0
  ) {
    stop("package rmarkdown >= 1.10.14 is required to use this function")
  }

  pandoc_args <- knitr::opts_knit$get("rmarkdown.pandoc.args")
  rd <- grep("--pdf-engine", pandoc_args)
  if (length(rd)) {
    engine <- pandoc_args[rd + 1]
  } else if (is_in_quarto()) {
    engine <- NULL

    # Quarto >= 1.8: read the resolved engine from QUARTO_EXECUTE_INFO
    exec_info <- Sys.getenv("QUARTO_EXECUTE_INFO", unset = "")
    if (
      nzchar(exec_info) &&
        file.exists(exec_info) &&
        requireNamespace("jsonlite", quietly = TRUE)
    ) {
      qmd <- jsonlite::read_json(exec_info)
      engine <- qmd$format$pandoc$`pdf-engine`
    }

    # Fallback: rmarkdown::metadata (top-level then nested)
    if (is.null(engine) || engine == "") {
      engine <- rmarkdown::metadata$`pdf-engine`
    }
    if (is.null(engine) || engine == "") {
      engine <- rmarkdown::metadata$format$pdf$`pdf-engine`
    }
    if (is.null(engine) || engine == "") {
      engine <- "xelatex"
    }
  } else {
    engine <- "pdflatex"
  }
  engine
}


latex_table_align <- function(x) {
  ft.align <- x$properties$align
  if ("left" %in% ft.align) {
    align_tag <- "l"
  } else if ("right" %in% ft.align) {
    align_tag <- "r"
  } else {
    align_tag <- "c"
  }
  align_tag
}

# latex_container -----
latex_container_none <- function() {
  x <- list()
  class(x) <- c("latex_container_none", "latex_container")
  x
}
latex_container_float <- function() {
  x <- list()
  class(x) <- c("latex_container_float", "latex_container")
  x
}
latex_container_wrap <- function(placement = "l") {
  stopifnot(
    length(placement) == 1,
    placement %in% c("l", "r", "i", "o")
  )
  x <- list(placement = placement)
  class(x) <- c("latex_container_wrap", "latex_container")
  x
}
latex_container_str <- function(x, latex_container, quarto = FALSE, ...) {
  UseMethod("latex_container_str", latex_container)
}

#' @export
latex_container_str.latex_container_none <- function(
  x,
  latex_container,
  quarto = FALSE,
  ...
) {
  c("", "")
}

#' @export
latex_container_str.latex_container_float <- function(
  x,
  latex_container,
  quarto = FALSE,
  ...
) {
  c("\\begin{table}", "\\end{table}")
}

#' @export
latex_container_str.latex_container_wrap <- function(
  x,
  latex_container,
  quarto = FALSE,
  ...
) {
  str <- paste0("\\begin{wraptable}{", latex_container$placement, "}")

  if (x$properties$layout %in% "fixed") {
    w <- sprintf("%.02fin", flextable_dim(x)$widths)
  } else {
    # wrapfig interprets 0pt as "auto-calculate width from
    # content"; needed because in autofit mode, column widths
    # are determined by LaTeX (columns use 'c' spec), not by R.
    w <- "0pt"
  }
  c(
    paste0(str, "{", w, "}"),
    "\\end{wraptable}"
  )
}

test_that("white spaces are protected", {
  ft <- flextable(data.frame(x = ""))
  ft <- delete_part(ft, part = "header")
  ft <- mk_par(ft, 1, 1, as_paragraph("foo", " ", "bar"))
  str <- flextable:::gen_raw_latex(ft)
  expect_true(grepl("{\\ }", str, fixed = TRUE))
})


test_that("merge_v places content at top of merged range", {
  ft <- flextable(data.frame(a = c("X", "X", "Y", "Y"), b = 1:4))
  ft <- merge_v(ft, j = "a")
  latex_str <- flextable:::gen_raw_latex(ft)

  lines <- unlist(strsplit(latex_str, "\n"))

  # no \multirow should be used (removed in favour of row placement)
  expect_false(any(grepl("\\\\multirow", lines)))

  # "X" should appear in first body row (with value "1")
  x_line <- grep("X", lines)[1]
  expect_true(grepl("1", lines[x_line]))

  # "Y" should appear in first row of its group (with value "3")
  y_line <- grep("Y", lines)[1]
  expect_true(grepl("3", lines[y_line]))
})

test_that("valign works in merged cells with varying row heights", {
  ft <- flextable(data.frame(
    col1 = c("A1", "A1", "A1"),
    col2 = c("B1", "long text", "B3")
  ))
  ft <- merge_at(ft, i = 1:3, j = 1)

  lines_for <- function(va) {
    ft2 <- valign(ft, valign = va)
    latex_str <- flextable:::gen_raw_latex(ft2)
    strsplit(latex_str, "\n")[[1]]
  }

  # top: A1 in first row, no multirow
  top_lines <- lines_for("top")
  top_body <- grep("A1|B1|B3", top_lines, value = TRUE)
  a1_line <- grep("A1", top_lines)[1]
  b1_line <- grep("B1", top_lines)[1]
  expect_equal(a1_line, b1_line) # A1 and B1 on same line
  expect_false(any(grepl("\\\\multirow", top_body)))

  # center: A1 in middle row (row 2), no multirow
  ctr_lines <- lines_for("center")
  a1_line <- grep("A1", ctr_lines)[1]
  b1_line <- grep("B1", ctr_lines)[1]
  b3_line <- grep("B3", ctr_lines)[1]
  expect_true(a1_line > b1_line) # A1 is after B1 (not in first row)
  expect_true(a1_line < b3_line) # A1 is before B3 (not in last row)
  expect_false(any(grepl("\\\\multirow", grep("A1", ctr_lines, value = TRUE))))

  # bottom: A1 in last row, no multirow
  bot_lines <- lines_for("bottom")
  bot_body <- grep("A1|B1|B3", bot_lines, value = TRUE)
  a1_line <- grep("A1", bot_lines)[1]
  b3_line <- grep("B3", bot_lines)[1]
  expect_equal(a1_line, b3_line) # A1 and B3 on same line
  expect_false(any(grepl("\\\\multirow", bot_body)))
})

# knit_to_latex tests ----
test_that("knit_to_latex default float none", {
  ft <- flextable(head(iris, 2))
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_true(is.character(z))
  expect_match(z, "\\\\setlength\\{\\\\tabcolsep\\}", perl = TRUE)
  expect_match(z, "\\\\renewcommand", perl = TRUE)
  expect_match(z, "arrayrulewidth", fixed = TRUE)
  # none container => no \begin{table}
  expect_no_match(z, "\\begin{table}", fixed = TRUE)
  expect_no_match(z, "\\begin{wraptable}", fixed = TRUE)
})

test_that("knit_to_latex float container", {
  ft <- flextable(head(iris, 2))
  ft$properties$opts_pdf$float <- "float"
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_match(z, "\\begin{table}", fixed = TRUE)
  expect_match(z, "\\end{table}", fixed = TRUE)
})

test_that("knit_to_latex wrap-l container", {
  ft <- flextable(head(iris, 2))
  ft$properties$opts_pdf$float <- "wrap-l"
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_match(z, "\\begin{wraptable}{l}", fixed = TRUE)
  expect_match(z, "\\end{wraptable}", fixed = TRUE)
})

test_that("knit_to_latex wrap-r container", {
  ft <- flextable(head(iris, 2))
  ft$properties$opts_pdf$float <- "wrap-r"
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_match(z, "\\begin{wraptable}{r}", fixed = TRUE)
})

test_that("knit_to_latex wrap-i container", {
  ft <- flextable(head(iris, 2))
  ft$properties$opts_pdf$float <- "wrap-i"
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_match(z, "\\begin{wraptable}{i}", fixed = TRUE)
})

test_that("knit_to_latex wrap-o container", {
  ft <- flextable(head(iris, 2))
  ft$properties$opts_pdf$float <- "wrap-o"
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_match(z, "\\begin{wraptable}{o}", fixed = TRUE)
})

test_that("knit_to_latex quarto caption is empty", {
  ft <- flextable(head(iris, 2))
  ft <- set_caption(ft, caption = "A test caption")
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = TRUE
  )
  # quarto mode: caption should not appear in latex
  expect_no_match(z, "A test caption", fixed = TRUE)
})

test_that("knit_to_latex bookdown caption", {
  ft <- flextable(head(iris, 2))
  ft <- set_caption(ft, caption = "My bookdown cap")
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = TRUE,
    quarto = FALSE
  )
  expect_true(is.character(z))
})

test_that("knit_to_latex tabcolsep and arraystretch", {
  ft <- flextable(head(iris, 2))
  ft$properties$opts_pdf$tabcolsep <- 5
  ft$properties$opts_pdf$arraystretch <- 2.0
  z <- flextable:::knit_to_latex(
    ft,
    bookdown = FALSE,
    quarto = FALSE
  )
  expect_match(z, "\\setlength{\\tabcolsep}{5pt}", fixed = TRUE)
  expect_match(z, "\\renewcommand*{\\arraystretch}{2}", fixed = TRUE)
})

test_that("fonts are defined in latex", {
  gdtools::register_liberationsans()
  ft <- flextable::flextable(head(cars, n = 1))
  ft <- flextable::font(ft, fontname = "Liberation Sans", part = "body")

  # R Markdown with pdflatex
  knitr::opts_knit$set("quarto.version" = NULL)
  latex_str <- flextable:::gen_raw_latex(ft, quarto = FALSE)
  expect_no_match(latex_str, regexp = "Liberation Sans", fixed = TRUE)

  knitr::opts_knit$set("rmarkdown.pandoc.args" = c("--pdf-engine", "xelatex"))
  latex_str <- flextable:::gen_raw_latex(ft, quarto = FALSE)
  expect_match(latex_str, regexp = "Liberation Sans", fixed = TRUE)
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  # quarto (defaults to xelatex)
  flextable:::fake_quarto()
  latex_str <- flextable:::gen_raw_latex(ft, quarto = TRUE)
  expect_match(latex_str, regexp = "Liberation Sans", fixed = TRUE)
  knitr::opts_knit$set("quarto.version" = NULL)
})

# helper to temporarily set rmarkdown::metadata
local_rmarkdown_metadata <- function(value, envir = parent.frame()) {
  ns <- asNamespace("rmarkdown")
  unlockBinding("metadata", ns)
  old <- rmarkdown::metadata
  assign("metadata", value, envir = ns)
  withr::defer(
    {
      assign("metadata", old, envir = ns)
      lockBinding("metadata", ns)
    },
    envir = envir
  )
}

# get_pdf_engine tests ----
test_that("get_pdf_engine returns pdflatex for Quarto with top-level pdf-engine", {
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list("pdf-engine" = "pdflatex"))
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  engine <- flextable:::get_pdf_engine()
  expect_equal(engine, "pdflatex")

  knitr::opts_knit$set("quarto.version" = NULL)
})

test_that("get_pdf_engine returns pdflatex for Quarto with nested pdf-engine", {
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list(
    format = list(pdf = list("pdf-engine" = "pdflatex"))
  ))
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  engine <- flextable:::get_pdf_engine()
  expect_equal(engine, "pdflatex")

  knitr::opts_knit$set("quarto.version" = NULL)
})

test_that("get_pdf_engine defaults to xelatex for Quarto with no engine", {
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list())
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  engine <- flextable:::get_pdf_engine()
  expect_equal(engine, "xelatex")

  knitr::opts_knit$set("quarto.version" = NULL)
})

test_that("get_pdf_engine reads QUARTO_EXECUTE_INFO when available", {
  skip_if_not_installed("jsonlite")
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list())
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  # write a temporary JSON file mimicking QUARTO_EXECUTE_INFO
  info <- list(format = list(pandoc = list("pdf-engine" = "pdflatex")))
  tmp <- tempfile(fileext = ".json")
  jsonlite::write_json(info, tmp, auto_unbox = TRUE)
  withr::local_envvar("QUARTO_EXECUTE_INFO" = tmp)

  engine <- flextable:::get_pdf_engine()
  expect_equal(engine, "pdflatex")

  knitr::opts_knit$set("quarto.version" = NULL)
})

test_that("get_pdf_engine QUARTO_EXECUTE_INFO takes priority over metadata", {
  skip_if_not_installed("jsonlite")
  flextable:::fake_quarto()
  # metadata says xelatex, but QUARTO_EXECUTE_INFO says pdflatex
  local_rmarkdown_metadata(list("pdf-engine" = "xelatex"))
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  info <- list(format = list(pandoc = list("pdf-engine" = "pdflatex")))
  tmp <- tempfile(fileext = ".json")
  jsonlite::write_json(info, tmp, auto_unbox = TRUE)
  withr::local_envvar("QUARTO_EXECUTE_INFO" = tmp)

  engine <- flextable:::get_pdf_engine()
  expect_equal(engine, "pdflatex")

  knitr::opts_knit$set("quarto.version" = NULL)
})

# list_latex_dep / fontspec tests ----
test_that("list_latex_dep does not include fontspec when Quarto + pdflatex", {
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list("pdf-engine" = "pdflatex"))
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  deps <- suppressWarnings(withr::with_options(
    list(knitr.in.progress = TRUE),
    {
      knitr::opts_knit$set("out.format" = "latex")
      flextable:::list_latex_dep()
    }
  ))
  dep_names <- vapply(deps, function(d) d$name, character(1))
  expect_false("fontspec" %in% dep_names)

  knitr::opts_knit$set("quarto.version" = NULL, "out.format" = NULL)
})

test_that("list_latex_dep includes fontspec when Quarto + xelatex", {
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list("pdf-engine" = "xelatex"))
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  deps <- withr::with_options(
    list(knitr.in.progress = TRUE),
    {
      knitr::opts_knit$set("out.format" = "latex")
      flextable:::list_latex_dep()
    }
  )
  dep_names <- vapply(deps, function(d) d$name, character(1))
  expect_true("fontspec" %in% dep_names)

  knitr::opts_knit$set("quarto.version" = NULL, "out.format" = NULL)
})

test_that("font warning fires for Quarto + pdflatex", {
  flextable:::fake_quarto()
  local_rmarkdown_metadata(list("pdf-engine" = "pdflatex"))
  knitr::opts_knit$set("rmarkdown.pandoc.args" = NULL)

  expect_warning(
    withr::with_options(
      list(knitr.in.progress = TRUE),
      {
        knitr::opts_knit$set("out.format" = "latex")
        flextable:::list_latex_dep()
      }
    ),
    "fonts used in `flextable` are ignored"
  )

  knitr::opts_knit$set("quarto.version" = NULL, "out.format" = NULL)
})

test_that("padding.left and padding.right produce advance leftskip/rightskip in latex", {
  ft <- flextable(data.frame(a = c("hello", "world"), b = 1:2))
  ft <- padding(ft, padding.left = 0, padding.right = 0, part = "all")
  ft <- padding(ft, i = 1, j = 1, padding.left = 20)
  ft <- padding(ft, i = 2, j = 1, padding.right = 15)
  latex_str <- flextable:::gen_raw_latex(ft)

  # row 1 col 1: \advance\leftskip by 20.0pt\relax
  expect_match(latex_str, "\\advance\\leftskip by 20.0pt\\relax", fixed = TRUE)

  # row 2 col 1: \advance\rightskip by 15.0pt\relax
  expect_match(latex_str, "\\advance\\rightskip by 15.0pt\\relax", fixed = TRUE)

  # only 1 leftskip and 1 rightskip (other cells have padding = 0)
  expect_equal(
    length(gregexpr("advance\\\\leftskip", latex_str, perl = TRUE)[[1]]),
    1L
  )
  expect_equal(
    length(gregexpr("advance\\\\rightskip", latex_str, perl = TRUE)[[1]]),
    1L
  )
})

list(
  rd_family_title = list(
    cell_content_composition = "Other functions to compose cell content: ",
    table_structure = "Other row and column operations: ",
    flextable_constructors = "Other flextable constructors: ",
    flextable_configuration = "Other flextable configuration: ",
    formatting_shortcuts = "Other formatting shortcuts: ",
    default_formatting_properties = "Other default formatting properties:",
    flextable_output_export = "Other functions for flextable output and export: ",
    officer_integration = "Other functions for officer integration: ",
    themes_and_defaults = "Other themes and defaults: "
    )
)

fmt_freq_table <- function(
  pctcol,
  pctrow,
  include.row_percent = TRUE,
  include.column_percent = TRUE
) {
  out_cols <- rep("", length(pctcol))
  out_rows <- out_sep <- out_cols

  if (include.column_percent) {
    out_cols[!is.na(pctcol)] <- fmt_pct(pctcol[!is.na(pctcol)])
  }
  if (include.row_percent) {
    out_rows[!is.na(pctrow)] <- fmt_pct(pctrow[!is.na(pctrow)])
  }
  if (include.column_percent && include.row_percent) {
    out_sep[!is.na(pctcol) | !is.na(pctrow)] <- " ; "
  }

  paste0(out_cols, out_sep, out_rows)
}

add_level_total <- function(sdat, to_na = TRUE, lev = "Total") {
  z <- sdat[, lapply(
    .SD,
    function(x, lev) {
      if (is.factor(x)) {
        old_levs <- levels(x)
        new_levs <- c(old_levs, lev)
        levels(x) <- new_levs
        if (to_na && all(is.na(x))) x[] <- lev
      }
      x
    },
    lev = lev
  )]
  z$grouping <- NULL
  z
}


#' @importFrom data.table groupingsets
fortified_freq <- function(
  dat,
  row = character(),
  column = character(),
  weight = character()
) {
  by <- unique(c(row, column))
  sets <- list()
  if (length(by) > 1) {
    sets <- list(c(row, column), row, column, character())
  } else if (length(by) > 0) {
    sets <- list(by, character())
  }
  dataset <- as.data.table(dat)
  if (length(weight) == 1) {
    setnames(dataset, old = weight, new = ".weight.")
  } else {
    dataset$.weight. <- 1
  }

  # replace na with "Missing"
  dataset[,
    c(by) := lapply(.SD, function(x) {
      if (is.factor(x)) {
        old_levs <- levels(x)
        new_levs <- c(old_levs, "Missing")
        levels(x) <- new_levs
        x[is.na(x)] <- "Missing"
      } else if (is.character(x)) {
        x[is.na(x)] <- "Missing"
      }
      x
    }),
    .SDcols = by
  ]

  freq_data <- groupingsets(
    x = dataset,
    j = c(list(count = sum(.SD$.weight., na.rm = TRUE))),
    by = by,
    sets = sets,
    id = TRUE
  )

  if (length(by) > 1) {
    row_sums <- add_level_total(freq_data[freq_data$grouping %in% 1, ])
    col_sums <- add_level_total(freq_data[freq_data$grouping %in% 2, ])
    all_sums <- add_level_total(freq_data[freq_data$grouping %in% 3, ])
    tab_ct <- add_level_total(freq_data[freq_data$grouping %in% 0, ])
    tab <- rbindlist(list(row_sums, col_sums, all_sums, tab_ct))

    setnames(row_sums, old = "count", new = "total_row")
    row_sums[[column]] <- NULL
    rtab <- merge(tab, row_sums, by = row)
    rtab[,
      c("pct_row", "total_row", "count") := list(
        .SD$count / .SD$total_row,
        NULL,
        NULL
      )
    ]

    setnames(col_sums, old = "count", new = "total_col")
    col_sums[[row]] <- NULL
    ctab <- merge(tab, col_sums, by = column)
    ctab[,
      c("pct_col", "total_col", "count") := list(
        .SD$count / .SD$total_col,
        NULL,
        NULL
      )
    ]

    tab_margins <- merge(ctab, rtab, by = by, all = FALSE)
    tab <- merge(tab, tab_margins, by = by, all = TRUE)
  } else if (length(by) > 0) {
    all_sums <- add_level_total(freq_data[freq_data$grouping %in% 1, ])
    tab_ct <- add_level_total(freq_data[freq_data$grouping %in% 0, ])
    tab <- rbindlist(list(all_sums, tab_ct))
  }
  tab[, c("pct") := list(.SD$count / all_sums$count)]
  setDF(tab)
  tab
}

relayout_freq_data <- function(x, order_by) {
  dat_pct_rowcol <- as.data.table(x)
  dat_pct_rowcol[, c("count", "pct", ".what.") := list(NULL, NULL, "mpct")]

  dat <- as.data.table(x)
  dat$.what. <- "count"
  dat[, c("pct_col", "pct_row") := NULL]

  dat <- rbindlist(list(dat_pct_rowcol, dat), fill = TRUE, use.names = TRUE)
  dat$.what. <- factor(dat$.what., c("count", "mpct"))
  setorderv(dat, c(order_by, ".what."))

  setDF(dat)
  dat
}


#' @title Frequency table
#'
#' @description This function computes a one or two way
#' contingency table and creates a flextable from the result.
#'
#' The function is largely inspired by "PROC FREQ" from "SAS"
#' and was written with the intent to make it
#' as compact as possible.
#' @param x a `data.frame` object containing variable(s) to use for counts.
#' @param row `characer` column names for row
#' @param col `characer` column names for column
#' @param include.row_percent `boolean` whether to include the row percents; defaults to `TRUE`
#' @param include.column_percent `boolean` whether to include the column percents; defaults to `TRUE`
#' @param include.table_percent `boolean` whether to include the table percents; defaults to `TRUE`
#' @param include.table_count `boolean` whether to include the table counts; defaults to `TRUE`
#' @param weight `character` column name for weight
#' @param count_format_fun a function to format the count values,
#' defaults to [fmt_int].
#' @param ... unused arguments
#' @importFrom stats as.formula
#' @examples
#' proc_freq(mtcars, "vs", "gear")
#' proc_freq(mtcars, "gear", "vs", weight = "wt")
#' @export
proc_freq <- function(
  x,
  row = character(),
  col = character(),
  include.row_percent = TRUE,
  include.column_percent = TRUE,
  include.table_percent = TRUE,
  include.table_count = TRUE,
  weight = character(),
  count_format_fun = fmt_int,
  ...
) {
  list_lbls <- collect_labels(dataset = x, use_labels = TRUE)
  for (colname in names(list_lbls$values_labels)) {
    x[[colname]] <- factor(
      x = x[[colname]],
      levels = names(list_lbls$values_labels[[colname]]),
      labels = unname(list_lbls$values_labels[[colname]])
    )
  }

  if (length(row) && !is.factor(x[[row]])) {
    x[[row]] <- as.factor(x[[row]])
  }

  if (length(col) && !is.factor(x[[col]])) {
    x[[col]] <- as.factor(x[[col]])
  }
  by <- unique(c(row, col))
  if (!length(by) %in% 1:2) {
    stop(
      "The `col` and `row` parameters do not allow to define a univariate or bivariate count table. It requires exactly one or two columns."
    )
  }

  dat <- fortified_freq(x, row = row, column = col, weight = weight)
  if (length(by) > 1) {
    dat <- relayout_freq_data(dat, order_by = by)
    dat <- dat[!(dat[[row]] %in% "Total" & dat[[".what."]] %in% c("mpct")), ]
    dat$.coltitle. <- col

    count <- pct <- pct_col <- pct_row <- NULL

    rows_set <- c(row, ".what.")
    first_vline <- 2
    fnote_lab <- NA_character_
    margins_label <- "Mar. pct"

    if (include.column_percent && include.row_percent) {
      fnote_lab <- " Columns and rows percentages"
    } else if (include.column_percent && !include.row_percent) {
      margins_label <- "Col. pct"
    } else if (!include.column_percent && include.row_percent) {
      margins_label <- "Row pct"
    } else {
      dat <- dat[!(dat[[".what."]] %in% c("mpct")), ]
      rows_set <- row
      first_vline <- 1
    }

    table_label <- "Count"
    if (!include.table_count && include.table_percent) {
      table_label <- "Percent"
    } else if (!include.table_count && !include.table_percent) {
      stop("At least one of the include.table_*  parameters must be TRUE.")
    }
    tab <- tabulator(
      x = dat,
      rows = rows_set,
      columns = c(".coltitle.", col),
      stat = as_paragraph(
        if (include.table_count & include.table_percent) {
          as_chunk(fmt_n_percent(count, pct))
        } else if (include.table_count) {
          as_chunk(count, formatter = count_format_fun)
        } else if (include.table_percent) {
          as_chunk(fmt_pct(pct))
        },
        as_chunk(fmt_freq_table(
          pct_col,
          pct_row,
          include.column_percent = include.column_percent,
          include.row_percent = include.row_percent
        ))
      )
    )

    ft <- as_flextable(tab, columns_alignment = "center", sep_w = 0)
    if (include.column_percent || include.row_percent) {
      ft <- labelizor(
        x = ft,
        labels = c(.what. = "", "count" = table_label, "mpct" = margins_label),
        j = ".what."
      )
      if (!is.na(fnote_lab)) {
        ft <- footnote(
          ft,
          ref_symbols = " (1)",
          j = ".what.",
          i = ~ .what. %in% "mpct" & !duplicated(.what.),
          value = as_paragraph(fnote_lab),
          part = "body"
        )
      }
    }
    ft <- hline(ft, i = as.formula(sprintf("~before(`%s`, 'Total')", row)))
    ft <- vline(ft, j = ncol_keys(ft) - 1)
    ft <- vline(ft, j = first_vline)
    ft <- autofit(ft)
  } else {
    dat <- dat[do.call(order, dat[by]), ]
    ft <- flextable(dat)
    ft <- hline(ft, i = as.formula(sprintf("~before(`%s`, 'Total')", by)))
    ft <- set_formatter(ft, pct = fmt_pct, count = fmt_int)
    ft <- set_header_labels(ft, count = "Count", pct = "Percent")
    ft <- autofit(ft)
  }

  ft
}

context("check proc_freq")

library(utils)
library(xml2)

test_that("proc_freq executes without errors", {
  dummy_df <- data.frame( values = rep(letters[1:3], each = 2), groups = rep(letters[1:3], each = 2), stringsAsFactors = FALSE )
  ft <- proc_freq(dummy_df, "values", "groups")
  expect_equivalent(class(ft), "flextable")
})

test_that("proc_freq can take options", {
  dummy_df <- data.frame( values = rep(letters[1:3], each = 2), groups = rep(letters[1:3], each = 2), stringsAsFactors = FALSE )
  ft <- proc_freq(dummy_df, "values", "groups", include.row_percent = FALSE, include.column_percent = FALSE, include.table_percent = FALSE, include.column_total = FALSE, include.row_total = FALSE, include.header_row = FALSE)
  expect_equivalent(class(ft), "flextable")
})

#' @title Rename column labels in the header
#'
#' @description Change the display labels in the bottom row of the header.
#' Unlike [set_header_df()] which replaces the entire header structure,
#' this function only modifies column labels in the last header row.
#'
#' @inheritParams args_x_only
#' @param ... named arguments (names are data colnames), each element is a single character
#' value specifying label to use.
#' @param values a named list (names are data colnames), each element is a single character
#' value specifying label to use. If provided, argument `...` will be ignored.
#' It can also be a unamed character vector, in that case, it must have the same
#' length than the number of columns of the flextable.
#' @examples
#' ft <- flextable(head(iris))
#' ft <- set_header_labels(ft,
#'   Sepal.Length = "Sepal length",
#'   Sepal.Width = "Sepal width", Petal.Length = "Petal length",
#'   Petal.Width = "Petal width"
#' )
#'
#' ft <- flextable(head(iris))
#' ft <- set_header_labels(ft,
#'   values = list(
#'     Sepal.Length = "Sepal length",
#'     Sepal.Width = "Sepal width",
#'     Petal.Length = "Petal length",
#'     Petal.Width = "Petal width"
#'   )
#' )
#' ft
#'
#' ft <- flextable(head(iris))
#' ft <- set_header_labels(
#'   x = ft,
#'   values = c(
#'     "Sepal length",
#'     "Sepal width", "Petal length",
#'     "Petal width", "Species")
#' )
#' ft
#' @export
#' @family table_structure
set_header_labels <- function(x, ..., values = NULL) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "set_header_labels()"
    ))
  }

  if (is.null(values)) {
    values <- list(...)
    if (nrow_part(x, "header") < 1) {
      stop("there is no header row to be replaced")
    }
  }

  names_ <- names(values)
  if (is.null(names_)) {
    if (length(values) != ncol_keys(x)) {
      stop("if unamed, the labels must have the same length than col_keys.")
    }

    newcontent <- as_paragraph(as_chunk(values))
    newcontent <- as_chunkset_struct(
      l_paragraph = newcontent,
      keys = x$col_keys
    )
    x$header$content <- set_chunkset_struct_element(
      x = x$header$content,
      i = nrow_part(x, "header"),
      j = x$col_keys,
      value = newcontent
    )
  } else {
    names_ <- names_[names_ %in% x$col_keys]
    values <- values[names_]
    if (length(values) < 1) {
      return(x)
    }
    newcontent <- lapply(
      values,
      function(x) {
        as_paragraph(as_chunk(x, formatter = format_fun.default))
      }
    )

    newcontent <- as_chunkset_struct(
      l_paragraph = do.call(c, newcontent),
      keys = names_
    )

    x$header$content <- set_chunkset_struct_element(
      x = x$header$content,
      i = nrow_part(x, "header"),
      j = names_,
      value = newcontent
    )
  }

  x
}


#' @export
#' @title Delete flextable part
#'
#' @description indicate to not print a part of
#' the flextable, i.e. an header, footer or the body.
#'
#' @inheritParams args_x_part_no_all
#' @family table_structure
#' @examples
#' ft <- flextable(head(iris))
#' ft <- delete_part(x = ft, part = "header")
#' ft
delete_part <- function(x, part = "header") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "delete_part()"
    ))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)
  nrow_ <- nrow(x[[part]]$dataset)
  x[[part]] <- complex_tabpart(
    data = x[[part]]$dataset[-seq_len(nrow_), , drop = FALSE],
    col_keys = x$col_keys,
    cwidth = x[[part]]$colwidths,
    cheight = x[[part]]$rowheights
  )
  x
}

delete_rows_from_part <- function(x, i) {
  # dataset rows
  x$dataset <- x$dataset[-i, ]
  # heights
  x$rowheights <- x$rowheights[-i]
  # hrules
  x$hrule <- x$hrule[-i]
  # spans — only reset if vertical spans are broken
  runs <- split(i, cumsum(c(1, diff(i) != 1)))
  has_broken_vspan <- any(vapply(
    runs,
    function(r) {
      any(colSums(x$spans$columns[r, , drop = FALSE]) != length(r))
    },
    logical(1)
  ))

  x$spans$rows <- x$spans$rows[-i, , drop = FALSE]
  x$spans$columns <- x$spans$columns[-i, , drop = FALSE]
  if (has_broken_vspan) {
    x <- span_free(x)
  }

  # styles
  x$styles$cells <- delete_style_row(x$styles$cells, i)
  x$styles$pars <- delete_style_row(x$styles$pars, i)
  x$styles$text <- delete_style_row(x$styles$text, i)

  # content
  x$content <- delete_row_from_fpstruct(x$content, i)

  x
}

#' @export
#' @title Delete flextable rows
#' @family table_structure
#' @description The function removes one or more rows
#' from a 'flextable'.
#' @details
#' Deleting one or more rows will result in the deletion
#' of any span parameters that may have been set previously.
#' They will have to be redone after this operation or
#' performed only after this deletion.
#' @inheritParams args_x_i_part
#' @examples
#' ft <- flextable(head(iris))
#' ft <- delete_rows(ft, i = 1:5, part = "body")
#' ft
delete_rows <- function(x, i = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "delete_rows()"
    ))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- delete_rows(x = x, i = i, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  if (length(i) < 1) {
    return(x)
  }
  x[[part]] <- delete_rows_from_part(x[[part]], i = i)
  x
}


delete_colums_from_part <- function(x, j) {
  j_idx <- which(x$col_keys %in% j)
  # heights
  x$colwidths <- x$colwidths[!x$col_keys %in% j]
  # spans — only reset if horizontal spans are broken
  runs <- split(j_idx, cumsum(c(1, diff(j_idx) != 1)))
  has_broken_hspan <- any(vapply(
    runs,
    function(r) {
      any(rowSums(x$spans$rows[, r, drop = FALSE]) != length(r))
    },
    logical(1)
  ))

  x$spans$rows <- x$spans$rows[, -j_idx, drop = FALSE]
  x$spans$columns <- x$spans$columns[, -j_idx, drop = FALSE]
  if (has_broken_hspan) {
    x <- span_free(x)
  }

  # styles
  x$styles$cells <- delete_style_col(x$styles$cells, j)
  x$styles$pars <- delete_style_col(x$styles$pars, j)
  x$styles$text <- delete_style_col(x$styles$text, j)

  # content
  x$content <- delete_col_from_fpstruct(x$content, j)
  x$col_keys <- x$content$keys
  x
}

#' @export
#' @title Delete flextable columns
#'
#' @description The function removes one or more columns
#' from a 'flextable'.
#' @details
#' Deleting one or more columns will result in the deletion
#' of any span parameters that may have been set previously.
#' They will have to be redone after this operation or
#' performed only after this deletion.
#' @inheritParams args_x_j
#' @family table_structure
#' @examples
#' ft <- flextable(head(iris))
#' ft <- delete_columns(ft, j = "Species")
#' ft
delete_columns <- function(x, j = NULL) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "delete_columns()"
    ))
  }
  cols <- NULL

  parts <- c("header", "body", "footer")
  for (part in parts) {
    if (nrow_part(x, part) > 0) {
      if (is.null(cols)) {
        cols <- get_columns_id(x[[part]], j)
        cols <- x$col_keys[cols]
      }
      x[[part]] <- delete_colums_from_part(x[[part]], j = cols)
    }
  }
  x$col_keys <- setdiff(x$col_keys, cols)
  x
}


as_new_data <- function(x, ..., values = NULL) {
  if (is.null(values)) {
    values <- list(...)
  } else if (is.atomic(values)) {
    values <- as.list(values)
  }

  if (!is.list(values)) {
    stop("argument 'values' must be a list")
  }

  args_ <- lapply(
    x$col_keys,
    function(x, n) rep(NA, n),
    n = length(values[[1]])
  )
  names(args_) <- x$col_keys
  args_[names(values)] <- values

  data.frame(as.list(args_), check.names = FALSE, stringsAsFactors = FALSE)
}

add_rows_to_tabpart <- function(x, rows, first = FALSE) {
  data <- x$dataset
  spans <- x$spans
  ncol <- length(x$col_keys)
  nrow <- nrow(rows)

  x$styles$cells <- add_rows_to_struct(
    x$styles$cells,
    nrows = nrow,
    first = first
  )
  x$styles$pars <- add_rows_to_struct(
    x$styles$pars,
    nrows = nrow,
    first = first
  )
  x$styles$text <- add_rows_to_struct(
    x$styles$text,
    nrows = nrow,
    first = first
  )
  x$content <- add_rows_to_chunkset_struct(
    x$content,
    nrows = nrow,
    first = first,
    rows
  )

  span_new <- matrix(1, ncol = ncol, nrow = nrow)
  rowheights <- x$rowheights
  hrule <- x$hrule

  if (!first) {
    data <- rbind_match_columns(list(data, rows))
    spans$rows <- rbind(spans$rows, span_new)
    spans$columns <- rbind(spans$columns, span_new)
    rowheights <- c(rowheights, rep(rev(rowheights)[1], nrow(rows)))
    hrule <- c(hrule, rep(rev(hrule)[1], nrow(rows)))
  } else {
    data <- rbind_match_columns(list(rows, data))
    spans$rows <- rbind(span_new, spans$rows)
    spans$columns <- rbind(span_new, spans$columns)
    rowheights <- c(rep(rowheights[1], nrow(rows)), rowheights)
    hrule <- c(rep(hrule[1], nrow(rows)), hrule)
  }
  x$rowheights <- rowheights
  x$dataset <- data
  x$spans <- spans
  x$hrule <- hrule
  x
}

# add header/footer content ----

#' @export
#' @title Add body rows with one value per column
#'
#' @description
#' Add new rows to the body where each value maps to a named column,
#' preserving the original column data types.
#' Unlike [add_body_row()] where labels can span multiple columns,
#' here each value fills exactly one column.
#'
#' If some columns are not provided, they will be replaced by
#' `NA` and displayed as empty.
#' @inheritParams args_x_only
#' @param top should the rows be inserted at the top or the bottom.
#' @param ... named arguments (names are data colnames) of values
#' to add. It is important to insert data of the same type as the
#' original data, otherwise it will be transformed (probably
#' into strings if you add a `character` where a `double` is expected).
#' This makes possible to still format cell contents with the `colformat_*`
#' functions, for example [colformat_num()].
#' @param values a list of name-value pairs of labels or values,
#' names should be existing col_key values. This argument can be used
#' instead of `...` for programming purpose (If `values` is
#' supplied argument `...` is ignored).
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c(
#'     "Species", "Sepal.Length", "Petal.Length",
#'     "Sepal.Width", "Petal.Width"
#'   )
#' )
#'
#' ft <- add_body(
#'   x = ft, Sepal.Length = 1:5,
#'   Sepal.Width = 1:5 * 2, Petal.Length = 1:5 * 3,
#'   Petal.Width = 1:5 + 10, Species = "Blah", top = FALSE
#' )
#'
#' ft <- theme_booktabs(ft)
#' ft
#' @family table_structure
#' @seealso [flextable()]
add_body <- function(x, top = TRUE, ..., values = NULL) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_body()"
    ))
  }

  new_data <- as_new_data(x = x, ..., values = values)
  x$body <- add_rows_to_tabpart(x$body, new_data, first = top)
  x
}


#' @export
#' @title Add header rows with one value per column
#'
#' @description
#' Add new rows to the header where each value maps to a named column.
#' Unlike [add_header_row()] where labels can span multiple columns,
#' here each value fills exactly one column.
#'
#' If some columns are not provided, they will be replaced by
#' `NA` and displayed as empty.
#'
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/flextable-016.png" alt="add_header illustration" style="width:100\%;">
#' }}
#'
#' @note
#' when repeating values, they can be merged together with
#' function [merge_h()] and [merge_v()].
#' @inheritParams add_body
#' @examples
#' library(flextable)
#'
#' fun <- function(x) {
#'   paste0(
#'     c("min: ", "max: "),
#'     formatC(range(x))
#'   )
#' }
#' new_row <- list(
#'   Sepal.Length = fun(iris$Sepal.Length),
#'   Sepal.Width =  fun(iris$Sepal.Width),
#'   Petal.Width =  fun(iris$Petal.Width),
#'   Petal.Length = fun(iris$Petal.Length)
#' )
#'
#' ft_1 <- flextable(data = head(iris))
#' ft_1 <- add_header(ft_1, values = new_row, top = FALSE)
#' ft_1 <- append_chunks(ft_1, part = "header", i = 2, )
#' ft_1 <- theme_booktabs(ft_1, bold_header = TRUE)
#' ft_1 <- align(ft_1, align = "center", part = "all")
#' ft_1
#' @family table_structure
add_header <- function(x, top = TRUE, ..., values = NULL) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_header()"
    ))
  }

  header_data <- as_new_data(x = x, ..., values = values)
  x$header <- add_rows_to_tabpart(x$header, header_data, first = top)

  x
}

#' @export
#' @title Add footer rows with one value per column
#'
#' @description
#' Add new rows to the footer where each value maps to a named column.
#' Unlike [add_footer_row()] where labels can span multiple columns,
#' here each value fills exactly one column.
#'
#' If some columns are not provided, they will be replaced by
#' `NA` and displayed as empty.
#' @inheritParams add_body
#' @examples
#' new_row <- as.list(colMeans(iris[, -5]))
#' new_row$Species <- "Means"
#'
#' formatter <- function(x) sprintf("%.1f", x)
#'
#' ft <- flextable(data = head(iris))
#' ft <- add_footer(ft, values = new_row)
#'
#' # cosmetics
#' ft <- compose(
#'   x = ft, j = 1:4,
#'   value = as_paragraph(
#'     as_chunk(., formatter = formatter)
#'   ),
#'   part = "footer", use_dot = TRUE
#' )
#' ft <- align(ft, part = "footer", align = "right", j = 1:4)
#' ft
#' @family table_structure
add_footer <- function(x, top = TRUE, ..., values = NULL) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_footer()"
    ))
  }

  footer_data <- as_new_data(x = x, ..., values = values)

  if (nrow_part(x, "footer") < 1) {
    x$footer <- complex_tabpart(
      data = footer_data,
      col_keys = x$col_keys,
      cwidth = .75,
      cheight = .25
    )
  } else {
    x$footer <- add_rows_to_tabpart(x$footer, footer_data, first = top)
  }

  x
}

# add spanned labels as a single row ----

#' @export
#' @title Add a body row with spanning labels
#'
#' @description Add a single row to the body where labels can span
#' multiple columns (merged cells) via the `colwidths` argument.
#'
#' Labels are associated with a number of columns
#' to merge that default to one if not specified.
#' In this case, you have to make sure that the
#' number of labels is equal to the number of columns displayed.
#'
#' The function can add only one single row by call.
#'
#' Labels can also be formatted with [as_paragraph()].
#'
#' @inheritParams args_x_only
#' @param top should the row be inserted at the top or the bottom.
#' @param values values to add. It can be a `list`, a `character()` vector
#' or a call to [as_paragraph()].
#'
#' If it is a list, it can be a named list with the names of the columns of the
#' original data.frame or the `colkeys`; this is the recommended method because
#' it allows to keep the original data types and therefore allows to perform
#' conditional formatting. If a character, columns of the original data.frame
#' stored in the flextable object are changed to `character()`; this is often
#' not an issue with footer and header but can be inconvenient if adding
#' rows into body as it will change data types to character and prevent
#' efficient conditional formatting.
#'
#' @param colwidths the number of columns to merge in the row for each label
#'
#' @examples
#' library(flextable)
#'
#' ft01 <- fp_text_default(color = "red")
#' ft02 <- fp_text_default(color = "orange")
#'
#' pars <- as_paragraph(
#'   as_chunk(c("(1)", "(2)"), props = ft02), " ",
#'   as_chunk(
#'     c(
#'       "My tailor is rich",
#'       "My baker is rich"
#'     ),
#'     props = ft01
#'   )
#' )
#'
#' ft_1 <- flextable(head(mtcars))
#' ft_1 <- add_body_row(ft_1,
#'   values = pars,
#'   colwidths = c(5, 6), top = FALSE
#' )
#' ft_1 <- add_body_row(ft_1,
#'   values = pars,
#'   colwidths = c(3, 8), top = TRUE
#' )
#' ft_1 <- theme_box(ft_1)
#' ft_1
#'
#' ft_2 <- flextable(head(airquality))
#' ft_2 <- add_body_row(ft_2,
#'   values = c("blah", "bleeeh"),
#'   colwidths = c(4, 2), top = TRUE
#' )
#' ft_2 <- theme_box(ft_2)
#' ft_2
#' @family table_structure
#' @seealso [flextable()], [set_caption()]
add_body_row <- function(
  x,
  top = TRUE,
  values = list(),
  colwidths = integer(0)
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_body_row()"
    ))
  }

  if (length(colwidths) < 1) {
    colwidths <- rep(1L, length(x$col_keys))
  }

  if (sum(colwidths) != length(x$col_keys)) {
    stop(sprintf(
      "sum of colwidths elements must be equal to the number of col_keys: %.0f.",
      length(x$col_keys)
    ))
  }
  if (is.atomic(values)) {
    values <- as.list(values)
  }

  row_span <- unlist(lapply(colwidths, function(x) {
    z <- integer(x)
    z[1] <- x
    z
  }))
  col_j <- which(row_span > 0)

  body_data <- x$body$dataset[nrow(x$body$dataset) + 1, ]

  if (inherits(values, "paragraph")) {} else if (
    !is.null(attr(values, "names"))
  ) {
    body_data[, names(values)] <- values
  } else {
    body_data[, which(row_span > 0)] <- values
  }

  if (nrow_part(x, "body") < 1) {
    x$body <- complex_tabpart(
      data = body_data,
      col_keys = x$col_keys,
      cwidth = dim(x)$widths,
      cheight = .25
    )
  } else {
    x$body <- add_rows_to_tabpart(x$body, body_data, first = top)
  }

  i <- ifelse(top, 1, nrow_part(x, "body"))
  x$body$spans$rows[i, ] <- row_span

  if (inherits(values, "paragraph")) {
    for (j in seq_along(col_j)) {
      x <- mk_par(x, i = i, j = col_j[j], value = values[j], part = "body")
    }
  }

  x
}

#' @export
#' @title Add a header row with spanning labels
#'
#' @description Add a single row to the header where labels can span
#' multiple columns (merged cells) via the `colwidths` argument.
#'
#' Labels are associated with a number of columns
#' to merge that default to one if not specified.
#' In this case, you have to make sure that the
#' number of labels is equal to the number of columns displayed.
#'
#' The function can add only one single row by call.
#'
#' Labels can also be formatted with [as_paragraph()].
#'
#' @inheritParams args_x_only
#' @param top should the row be inserted at the
#' top or the bottom. Default to TRUE.
#' @param values values to add, a character vector (as header rows
#' contains only character values/columns), a list
#' or a call to [as_paragraph()].
#' @param colwidths the number of columns used for each label
#' @examples
#' library(flextable)
#'
#' ft01 <- fp_text_default(color = "red")
#' ft02 <- fp_text_default(color = "orange")
#'
#' pars <- as_paragraph(
#'   as_chunk(c("(1)", "(2)"), props = ft02), " ",
#'   as_chunk(c(
#'     "My tailor is rich",
#'     "My baker is rich"
#'   ), props = ft01)
#' )
#'
#' ft_1 <- flextable(head(mtcars))
#' ft_1 <- add_header_row(ft_1,
#'   values = pars,
#'   colwidths = c(5, 6), top = FALSE
#' )
#' ft_1 <- add_header_row(ft_1,
#'   values = pars,
#'   colwidths = c(3, 8), top = TRUE
#' )
#' ft_1
#'
#' ft_2 <- flextable(head(airquality))
#' ft_2 <- add_header_row(ft_2,
#'   values = c("Measure", "Time"),
#'   colwidths = c(4, 2), top = TRUE
#' )
#' ft_2 <- theme_box(ft_2)
#' ft_2
#' @family table_structure
#' @seealso [flextable()], [set_caption()]
add_header_row <- function(
  x,
  top = TRUE,
  values = character(0),
  colwidths = integer(0)
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_header_row()"
    ))
  }

  if (length(colwidths) < 1) {
    colwidths <- rep(1L, length(x$col_keys))
  }

  if (sum(colwidths) != length(x$col_keys)) {
    stop(sprintf(
      "sum of colwidths elements must be equal to the number of col_keys: %.0f.",
      length(x$col_keys)
    ))
  }

  if (inherits(values, "paragraph")) {
    str_values <- vapply(
      values,
      function(x) paste0(x$txt, collapse = ""),
      FUN.VALUE = ""
    )
  } else {
    str_values <- values
  }
  header_data <- data_from_char(
    values = str_values,
    colwidths = colwidths,
    col_keys = x$col_keys
  )

  if (nrow_part(x, "header") < 1) {
    x$header <- complex_tabpart(
      data = header_data,
      col_keys = x$col_keys,
      cwidth = dim(x)$widths,
      cheight = .25
    )
  } else {
    x$header <- add_rows_to_tabpart(x$header, header_data, first = top)
  }

  row_span <- unlist(lapply(colwidths, function(x) {
    z <- integer(x)
    z[1] <- x
    z
  }))
  i <- ifelse(top, 1, nrow(x$header$dataset))
  x$header$spans$rows[i, ] <- row_span

  if (inherits(values, "paragraph")) {
    if (top) {
      i <- 1
    } else {
      i <- nrow_part(x, "header")
    }

    col_j <- which(row_span > 0)

    for (j in seq_along(col_j)) {
      x <- mk_par(x, i = i, j = col_j[j], value = values[j], part = "header")
    }
  }

  x
}


#' @export
#' @title Add a footer row with spanning labels
#'
#' @description Add a single row to the footer where labels can span
#' multiple columns (merged cells) via the `colwidths` argument.
#'
#' Labels are associated with a number of columns
#' to merge that default to one if not specified.
#' In this case, you have to make sure that the
#' number of labels is equal to the number of columns displayed.
#'
#' The function can add only one single row by call.
#'
#' Labels can be formatted with [as_paragraph()].
#' @inheritParams add_body_row
#' @family table_structure
#' @seealso [flextable()], [set_caption()]
#' @examples
#' library(flextable)
#'
#' ft01 <- fp_text_default(color = "red")
#' ft02 <- fp_text_default(color = "orange")
#'
#' pars <- as_paragraph(
#'   as_chunk(c("(1)", "(2)"), props = ft02), " ",
#'   as_chunk(
#'     c(
#'       "My tailor is rich",
#'       "My baker is rich"
#'     ),
#'     props = ft01
#'   )
#' )
#'
#' ft_1 <- flextable(head(mtcars))
#' ft_1 <- add_footer_row(ft_1,
#'   values = pars,
#'   colwidths = c(5, 6), top = FALSE
#' )
#' ft_1 <- add_footer_row(ft_1,
#'   values = pars,
#'   colwidths = c(3, 8), top = TRUE
#' )
#' ft_1
#'
#' ft_2 <- flextable(head(airquality))
#' ft_2 <- add_footer_row(ft_2,
#'   values = c("Measure", "Time"),
#'   colwidths = c(4, 2), top = TRUE
#' )
#' ft_2 <- theme_box(ft_2)
#' ft_2
add_footer_row <- function(
  x,
  top = TRUE,
  values = character(0),
  colwidths = integer(0)
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_footer_row()"
    ))
  }

  if (sum(colwidths) != length(x$col_keys)) {
    stop(sprintf(
      "`colwidths` argument specify room for %.0f columns but %.0f are expected.",
      sum(colwidths),
      length(x$col_keys)
    ))
  }
  if (is.atomic(values)) {
    values <- as.list(values)
  }

  if (inherits(values, "paragraph")) {
    str_values <- vapply(
      values,
      function(x) paste0(x$txt, collapse = ""),
      FUN.VALUE = ""
    )
  } else {
    str_values <- values
  }
  footer_data <- data_from_char(
    values = str_values,
    colwidths = colwidths,
    col_keys = x$col_keys
  )

  if (nrow_part(x, "footer") < 1) {
    x$footer <- complex_tabpart(
      data = footer_data,
      col_keys = x$col_keys,
      cwidth = dim(x)$widths,
      cheight = .25
    )
  } else {
    x$footer <- add_rows_to_tabpart(x$footer, footer_data, first = top)
  }

  row_span <- unlist(lapply(colwidths, function(x) {
    z <- integer(x)
    z[1] <- x
    z
  }))
  i <- ifelse(top, 1, nrow_part(x, "footer"))
  x$footer$spans$rows[i, ] <- row_span
  col_j <- which(row_span > 0)

  if (inherits(values, "paragraph")) {
    for (j in seq_along(col_j)) {
      x <- mk_par(x, i = i, j = col_j[j], value = values[j], part = "footer")
    }
  }

  x
}

data_from_char <- function(values, colwidths, col_keys) {
  if (is.list(values)) {
    values <- as.character(unlist(values))
  } else if (!is.atomic(values)) {
    stop("'values' must be an atomic vector or a list.")
  }

  values_ <- inverse.rle(structure(
    list(lengths = colwidths, values = values),
    class = "rle"
  ))
  values_ <- as.list(values_)

  names(values_) <- col_keys
  as.data.frame(values_, check.names = FALSE, stringsAsFactors = FALSE)
}

# add labels as spanned rows ----

#' @export
#' @title Add full-width rows to the header
#'
#' @description Add one or more rows to the header where each label
#' spans all columns (all cells merged into one). Useful for adding
#' titles or subtitles above the column headers.
#'
#' @inheritParams args_x_only
#' @param values a character vector or a call to [as_paragraph()]
#' to get formated content, each element will
#' be added as a new row.
#' @param top should the row be inserted at the top
#' or the bottom. Default to TRUE.
#' @family table_structure
#' @examples
#' # ex 1----
#' ft_1 <- flextable(head(iris))
#' ft_1 <- add_header_lines(ft_1, values = "blah blah")
#' ft_1 <- add_header_lines(ft_1, values = c("blah 1", "blah 2"))
#' ft_1 <- autofit(ft_1)
#' ft_1
#'
#' # ex 2----
#' ft01 <- fp_text_default(color = "red")
#' ft02 <- fp_text_default(color = "orange")
#' ref <- c("(1)", "(2)")
#' pars <- as_paragraph(
#'   as_chunk(ref, props = ft02), " ",
#'   as_chunk(rep("My tailor is rich", length(ref)), props = ft01)
#' )
#'
#' ft_2 <- flextable(head(mtcars))
#' ft_2 <- add_header_lines(ft_2, values = pars, top = FALSE)
#' ft_2 <- add_header_lines(ft_2, values = ref, top = TRUE)
#' ft_2 <- add_footer_lines(ft_2, values = "blah", top = TRUE)
#' ft_2 <- add_footer_lines(ft_2, values = pars, top = TRUE)
#' ft_2 <- add_footer_lines(ft_2, values = ref, top = FALSE)
#' ft_2 <- autofit(ft_2)
#' ft_2
add_header_lines <- function(x, values = character(0), top = TRUE) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_header_lines()"
    ))
  }
  values_map <- values
  if (top) {
    values_map <- rev(values_map)
  }

  if (inherits(values, "paragraph")) {
    for (ivalue in seq_len(length(values_map))) {
      x <- add_header_row(
        x,
        values = "",
        colwidths = length(x$col_keys),
        top = top
      )
      if (top) {
        i <- 1
      } else {
        i <- nrow_part(x, "header")
      }
      x <- mk_par(x, i = i, j = 1, value = values_map[ivalue], part = "header")
    }
  } else {
    for (value in values_map) {
      x <- add_header_row(
        x,
        values = value,
        colwidths = length(x$col_keys),
        top = top
      )
    }
  }

  x
}


#' @export
#' @title Add full-width rows to the footer
#'
#' @description Add one or more rows to the footer where each label
#' spans all columns (all cells merged into one). Useful for adding
#' footnotes or source notes below the table.
#' @inheritParams add_header_lines
#' @family table_structure
#' @examples
#' ft_1 <- flextable(head(iris))
#' ft_1 <- add_footer_lines(ft_1,
#'   values = c("blah 1", "blah 2")
#' )
#' ft_1
add_footer_lines <- function(x, values = character(0), top = FALSE) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "add_footer_lines()"
    ))
  }

  ivalues <- seq_len(length(values))
  if (top) {
    ivalues <- rev(ivalues)
  }

  if (inherits(values, "paragraph")) {
    for (ivalue in ivalues) {
      x <- add_footer_row(
        x,
        values = "",
        colwidths = length(x$col_keys),
        top = top
      )
      if (top) {
        i <- 1
      } else {
        i <- nrow_part(x, "footer")
      }
      x <- mk_par(x, i = i, j = 1, value = values[ivalue], part = "footer")
    }
  } else {
    for (ivalue in ivalues) {
      x <- add_footer_row(
        x,
        values = values[ivalue],
        colwidths = length(x$col_keys),
        top = top
      )
    }
  }

  x
}


# add header/footer with reference table ----

set_part_df <- function(x, mapping = NULL, key = "col_keys", part) {
  keys <- data.frame(
    col_keys = x$col_keys,
    stringsAsFactors = FALSE
  )
  names(keys) <- key

  part_data <- merge(
    keys,
    mapping,
    by = key,
    all.x = TRUE,
    all.y = FALSE,
    sort = FALSE
  )
  part_data <- part_data[match(keys[[key]], part_data[[key]]), ]

  part_data[[key]] <- NULL

  part_data <- do.call(rbind, part_data)
  dimnames(part_data) <- NULL

  part_data <- as.data.frame(part_data, stringsAsFactors = FALSE)
  names(part_data) <- x$col_keys

  if (length(x$blanks)) {
    blank_ <- character(nrow(part_data))
    replace_ <- lapply(x$blanks, function(x, bl) bl, blank_)
    names(replace_) <- x$blanks
    part_data[x$blanks] <- replace_
  }

  colwidths <- x[[part]]$colwidths
  x[[part]] <- eval(
    call(
      class(x[[part]]),
      data = part_data,
      col_keys = x$col_keys,
      cwidth = .75,
      cheight = .25
    )
  )
  cheight <- optimal_sizes(x[[part]])$heights

  x[[part]]$colwidths <- colwidths
  x[[part]]$rowheights <- cheight
  x
}

#' @export
#' @rdname set_header_footer_df
#' @name set_header_footer_df
#' @title Replace the entire header or footer from a data frame
#'
#' @description Replace all header or footer rows using a mapping data frame.
#' Unlike [set_header_labels()] which only renames the bottom header row,
#' this function rebuilds the entire header (or footer) structure.
#'
#' The data.frame must contain a column whose values match flextable
#' `col_keys` argument, this column will be used as join key. The
#' other columns will be displayed as header or footer rows. The leftmost column
#' is used as the top header/footer row and the rightmost column
#' is used as the bottom header/footer row.
#'
#' @inheritParams args_x_only
#' @param mapping a `data.frame` specyfing for each colname
#' content of the column.
#' @param key column to use as key when joigning data_mapping.
#' @examples
#' typology <- data.frame(
#'   col_keys = c(
#'     "Sepal.Length", "Sepal.Width", "Petal.Length",
#'     "Petal.Width", "Species"
#'   ),
#'   what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
#'   measure = c("Length", "Width", "Length", "Width", "Species"),
#'   stringsAsFactors = FALSE
#' )
#'
#' ft_1 <- flextable(head(iris))
#' ft_1 <- set_header_df(ft_1, mapping = typology, key = "col_keys")
#' ft_1 <- merge_h(ft_1, part = "header")
#' ft_1 <- merge_v(ft_1, j = "Species", part = "header")
#' ft_1 <- theme_vanilla(ft_1)
#' ft_1 <- fix_border_issues(ft_1)
#' ft_1
#' @family table_structure
set_header_df <- function(x, mapping = NULL, key = "col_keys") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "set_header_df()"
    ))
  }

  set_part_df(x, mapping = mapping, key = key, part = "header")
}

#' @rdname set_header_footer_df
#' @export
#' @examples
#'
#'
#' typology <- data.frame(
#'   col_keys = c(
#'     "Sepal.Length", "Sepal.Width", "Petal.Length",
#'     "Petal.Width", "Species"
#'   ),
#'   unit = c("(cm)", "(cm)", "(cm)", "(cm)", ""),
#'   stringsAsFactors = FALSE
#' )
#' ft_2 <- set_footer_df(ft_1, mapping = typology, key = "col_keys")
#' ft_2 <- italic(ft_2, italic = TRUE, part = "footer")
#' ft_2 <- theme_booktabs(ft_2)
#' ft_2 <- fix_border_issues(ft_2)
#' ft_2
set_footer_df <- function(x, mapping = NULL, key = "col_keys") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "set_footer_df()"
    ))
  }

  set_part_df(x, mapping = mapping, key = key, part = "footer")
}

#' @importFrom data.table tstrsplit
#' @export
#' @title Split column names using a separator into multiple rows
#' @description This function is used to separate and place individual
#' labels in their own rows if your variable names contain multiple
#' delimited labels.
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/flextable-016.png" alt="add_header illustration" style="width:100\%;">
#' }}
#' @inheritParams args_x_only
#' @param opts Optional treatments to apply to the resulting header part.
#' This should be a character vector with support for multiple values.
#'
#' Supported values include:
#'
#' - "span-top": This operation spans empty cells with the first non-empty
#' cell, applied column by column.
#' - "center-hspan": Center the cells that are horizontally spanned.
#' - "bottom-vspan": Aligns to the bottom the cells treated at the  when "span-top" is applied.
#' - "default-theme": Applies the theme set in `set_flextable_defaults(theme_fun = ...)` to the new header part.
#' @param split a regular expression (unless `fixed = TRUE`)
#' to use for splitting.
#' @param fixed logical. If TRUE match `split` exactly,
#' otherwise use regular expressions.
#' @family table_structure
#' @examples
#' library(flextable)
#'
#' x <- data.frame(
#'   Species = as.factor(c("setosa", "versicolor", "virginica")),
#'   Sepal.Length_mean = c(5.006, 5.936, 6.588),
#'   Sepal.Length_sd = c(0.35249, 0.51617, 0.63588),
#'   Sepal.Width_mean = c(3.428, 2.77, 2.974),
#'   Sepal.Width_sd = c(0.37906, 0.3138, 0.3225),
#'   Petal.Length_mean = c(1.462, 4.26, 5.552),
#'   Petal.Length_sd = c(0.17366, 0.46991, 0.55189),
#'   Petal.Width_mean = c(0.246, 1.326, 2.026),
#'   Petal.Width_sd = c(0.10539, 0.19775, 0.27465)
#' )
#'
#' ft_1 <- flextable(x)
#' ft_1 <- colformat_double(ft_1, digits = 2)
#' ft_1 <- theme_box(ft_1)
#' ft_1 <- separate_header(
#'   x = ft_1,
#'   opts = c("span-top", "bottom-vspan")
#' )
#' ft_1
separate_header <- function(
  x,
  opts = c(
    "span-top",
    "center-hspan",
    "bottom-vspan",
    "default-theme"
  ),
  split = "[_\\.]",
  fixed = FALSE
) {
  if (nrow_part(x, "header") > 1) {
    stop(
      "the flextable object already have additional row(s),\nrun `separate_header()` before any header row augmentation"
    )
  }

  ref_list <- tstrsplit(x$col_keys, split = split, fill = "", fixed = fixed)
  ref_list <- rev(ref_list)
  last_names <- ref_list[[1]]
  names(last_names) <- x$col_keys
  x <- set_header_labels(x, values = last_names)
  add_labels <- ref_list[-1]
  for (labels in add_labels) {
    # dont span ""
    tmp_labels <- labels
    tmp_labels[tmp_labels %in% ""] <-
      paste0("dummy-", seq_len(sum(tmp_labels %in% "")))
    rle_labs <- rle(tmp_labels)
    rle_labs$values[grepl("^dummy\\-", rle_labs$values)] <- ""

    x <- add_header_row(
      x = x,
      top = TRUE,
      values = rle_labs$values,
      colwidths = rle_labs$lengths
    )
  }
  names(ref_list) <- letters[seq_along(ref_list)]
  ref_list <- matrix(unlist(ref_list) %in% "", ncol = length(ref_list))

  if ("default-theme" %in% opts) {
    tmp <- do.call(
      get_flextable_defaults()$theme_fun,
      list(x)
    )
    x$header <- tmp$header
  }

  if ("span-top" %in% opts) {
    for (j in seq_len(nrow(ref_list))) {
      if (ref_list[j, 1]) {
        to <- rle(ref_list[j, ])$lengths[1] + 1
        if (all(x$header$spans$rows[seq(1, to), j] %in% 1)) {
          #can be v-merged
          x <- merge_at(
            x = x,
            i = seq(1, to),
            j = j,
            part = "header"
          )
        }

        if ("bottom-vspan" %in% opts) {
          x <- valign(
            x = x,
            i = seq(1, to),
            j = j,
            valign = "bottom",
            part = "header"
          )
        }
      }
    }
  }

  if ("center-hspan" %in% opts) {
    nr <- nrow(x[["header"]]$spans$rows)
    header_spans <- fortify_span(x, parts = "header")
    header_spans <- header_spans[header_spans$rowspan > 1, ]
    header_spans <- split(header_spans$.col_id, header_spans$.row_id)
    for (i in seq_along(header_spans)) {
      x <- align(
        x = x,
        i = i,
        j = as.character(header_spans[[i]]),
        align = "center",
        part = "header"
      )
    }
  }
  fix_border_issues(x)
}

test_that("set_header_labels", {
  col_keys <- c(
    "Species",
    "sep1",
    "Sepal.Length",
    "Sepal.Width",
    "sep2",
    "Petal.Length",
    "Petal.Width"
  )
  ft <- flextable(head(iris), col_keys = col_keys)
  ft <- set_header_labels(
    ft,
    Sepal.Length = "Sepal length",
    Sepal.Width = "Sepal width",
    Petal.Length = "Petal length",
    Petal.Width = "Petal width"
  )

  docx_file <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- body_add_flextable(doc, value = ft)
  doc <- print(doc, target = docx_file)

  main_folder <- file.path(getwd(), "docx_folder")
  unpack_folder(file = docx_file, folder = main_folder)

  doc_file <- file.path(main_folder, "/word/document.xml")
  doc <- read_xml(doc_file)

  colnodes <- xml_find_all(doc, "w:body/w:tbl/w:tr[w:trPr/w:tblHeader]/w:tc")

  expect_equal(
    xml_text(colnodes),
    c(
      "Species",
      "",
      "Sepal length",
      "Sepal width",
      "",
      "Petal length",
      "Petal width"
    )
  )

  unlink(main_folder, recursive = TRUE, force = TRUE)

  ft <- flextable(mtcars)
  ft <- set_header_labels(ft, values = letters[1:ncol(mtcars)])
  ft <- delete_part(ft, part = "body")
  expect_equal(
    information_data_chunk(ft)$txt,
    letters[1:ncol(mtcars)]
  )
})

test_that("add_header", {
  data_ref <- structure(
    list(
      Sepal.Length = c("Sepal", "s", "(cm)"),
      Sepal.Width = c("Sepal", "", "(cm)"),
      Petal.Length = c("Petal", "", "(cm)"),
      Petal.Width = c("Petal", "", "(cm)"),
      Species = c("Species", "", "(cm)")
    ),
    .Names = c(
      "Sepal.Length",
      "Sepal.Width",
      "Petal.Length",
      "Petal.Width",
      "Species"
    ),
    row.names = c(NA, -3L),
    class = "data.frame"
  )

  ft <- flextable(iris[1:6, ])
  ft <- set_header_labels(
    ft,
    Sepal.Length = "Sepal",
    Sepal.Width = "Sepal",
    Petal.Length = "Petal",
    Petal.Width = "Petal",
    Species = "Species"
  )
  ft <- add_header(ft, Sepal.Length = "s", top = FALSE)
  ft <- add_header(
    ft,
    Sepal.Length = "(cm)",
    Sepal.Width = "(cm)",
    Petal.Length = "(cm)",
    Petal.Width = "(cm)",
    Species = "(cm)",
    top = FALSE
  )
  has_ <- flextable:::fortify_content(
    ft$header$content,
    default_chunk_fmt = ft$header$styles$text
  )$txt
  expect_equal(has_, as.character(unlist(data_ref)))

  ft <- flextable(iris[1:6, ])
  ft <- set_header_labels(
    ft,
    Sepal.Length = "Sepal",
    Sepal.Width = "Sepal",
    Petal.Length = "Petal",
    Petal.Width = "Petal",
    Species = "Species"
  )
  ft <- add_header(ft, Sepal.Length = "s", top = FALSE)
  ft <- add_header(
    ft,
    Sepal.Length = "(cm)",
    Sepal.Width = "(cm)",
    Petal.Length = "(cm)",
    Petal.Width = "(cm)",
    Species = "(cm)",
    top = FALSE
  )
  has_ <- flextable:::fortify_content(
    ft$header$content,
    default_chunk_fmt = ft$header$styles$text
  )$txt
  expect_equal(has_, as.character(unlist(data_ref)))
})

test_that("set_header_df", {
  typology <- data.frame(
    col_keys = c(
      "Sepal.Length",
      "Sepal.Width",
      "Petal.Length",
      "Petal.Width",
      "Species"
    ),
    what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
    measure = c("Length", "Width", "Length", "Width", "Species"),
    stringsAsFactors = FALSE
  )
  data <- iris[c(1:3, 51:53, 101:104), ]

  ft <- flextable(
    data,
    col_keys = c(
      "Species",
      "sep_1",
      "Sepal.Length",
      "Sepal.Width",
      "sep_2",
      "Petal.Length",
      "Petal.Width"
    )
  )
  ft <- set_header_df(ft, mapping = typology, key = "col_keys")

  data_ref <- structure(
    list(
      Species = c("Species", "Species"),
      sep_1 = c("", ""),
      Sepal.Length = c("Sepal", "Length"),
      Sepal.Width = c("Sepal", "Width"),
      sep_2 = c("", ""),
      Petal.Length = c("Petal", "Length"),
      Petal.Width = c("Petal", "Width")
    ),
    .Names = c(
      "Species",
      "sep_1",
      "Sepal.Length",
      "Sepal.Width",
      "sep_2",
      "Petal.Length",
      "Petal.Width"
    ),
    row.names = c(NA, -2L),
    class = "data.frame"
  )
  expect_ <- as.character(unlist(data_ref))
  has_ <- flextable:::fortify_content(
    ft$header$content,
    default_chunk_fmt = ft$header$styles$text
  )$txt
  expect_equal(has_, expect_)
})

test_that("add_footer", {
  data_ref <- structure(
    list(
      Sepal.Length = c("Sepal", "s", "(cm)"),
      Sepal.Width = c("Sepal", "", "(cm)"),
      Petal.Length = c("Petal", "", "(cm)"),
      Petal.Width = c("Petal", "", "(cm)"),
      Species = c("Species", "", "(cm)")
    ),
    .Names = c(
      "Sepal.Length",
      "Sepal.Width",
      "Petal.Length",
      "Petal.Width",
      "Species"
    ),
    row.names = c(NA, -3L),
    class = "data.frame"
  )

  ft <- flextable(iris[1:6, ])
  ft <- add_footer(
    ft,
    Sepal.Length = "Sepal",
    Sepal.Width = "Sepal",
    Petal.Length = "Petal",
    Petal.Width = "Petal",
    Species = "Species"
  )
  ft <- add_footer(ft, Sepal.Length = "s", top = FALSE)
  ft <- add_footer(
    ft,
    Sepal.Length = "(cm)",
    Sepal.Width = "(cm)",
    Petal.Length = "(cm)",
    Petal.Width = "(cm)",
    Species = "(cm)",
    top = FALSE
  )
  has_ <- flextable:::fortify_content(
    ft$footer$content,
    default_chunk_fmt = ft$footer$styles$text
  )$txt
  expect_equal(has_, as.character(unlist(data_ref)))

  ft <- flextable(iris[1:6, ])
  ft <- add_footer_row(
    ft,
    values = c("Sepal", "Petal", "Species"),
    colwidths = c(2, 2, 1)
  )
  ft <- add_footer_lines(ft, "s", top = FALSE)
  ft <- add_footer_row(ft, values = "(cm)", colwidths = 5, top = FALSE)
  has_ <- flextable:::fortify_content(
    ft$footer$content,
    default_chunk_fmt = ft$footer$styles$text
  )$txt

  ref <- c(
    "Sepal",
    "s",
    "(cm)",
    "Sepal",
    "s",
    "(cm)",
    "Petal",
    "s",
    "(cm)",
    "Petal",
    "s",
    "(cm)",
    "Species",
    "s",
    "(cm)"
  )
  expect_equal(has_, ref)
})

test_that("nrow_part or ncol_keys checks", {
  expect_error(nrow_part(12))
  expect_error(ncol_keys(12))
  ft <- flextable(head(iris))
  expect_equal(nrow_part(ft, part = "footer"), 0)
  expect_equal(nrow_part(ft, part = "body"), 6)
  expect_equal(ncol_keys(ft), 5)
})

test_that("add lines", {
  ft <- flextable(head(iris))

  newvals <- c("A", "B", "C", "D")

  ft <- add_header_lines(
    x = ft,
    values = newvals,
    top = TRUE
  )
  expect_equal(nrow_part(ft, part = "header"), 5)

  ft <- add_footer_lines(
    x = ft,
    values = newvals,
    top = FALSE
  )
  expect_equal(nrow_part(ft, part = "footer"), 4)

  x <- information_data_chunk(ft)

  header_sel <- x[x$.part %in% "header", ]
  expect_equal(
    header_sel$txt,
    c(
      rep(newvals, each = 5),
      colnames(iris)
    )
  )
  footer_sel <- x[x$.part %in% "footer", ]
  expect_equal(
    footer_sel$txt,
    rep(newvals, each = 5)
  )
})

test_that("separate_header", {
  x <- data.frame(
    Species = as.factor(c("setosa", "versicolor", "virginica")),
    Sepal.Length_mean_zzz = c(5.006, 5.936, 6.588),
    Sepal.Length_sd = c(0.35249, 0.51617, 0.63588),
    Sepal.Width_mean = c(3.428, 2.77, 2.974),
    Sepal.Width_sd_sfsf_dsfsdf = c(0.37906, 0.3138, 0.3225),
    Petal.Length_mean = c(1.462, 4.26, 5.552),
    Petal.Length_sd = c(0.17366, 0.46991, 0.55189),
    Petal.Width_mean = c(0.246, 1.326, 2.026),
    Petal.Width_sd = c(0.10539, 0.19775, 0.27465)
  )

  ft_1 <- flextable(x)
  ft_1 <- separate_header(x = ft_1, opts = c("span-top", "bottom-vspan"))
  header_txt <- information_data_chunk(ft_1)
  header_txt <- subset(header_txt, .part %in% "header")
  expect_equal(
    object = header_txt$txt,
    expected = c(
      "Species",
      "Sepal",
      "Sepal",
      "Sepal",
      "Sepal",
      "Petal",
      "Petal",
      "Petal",
      "Petal",
      "",
      "Length",
      "Length",
      "Width",
      "Width",
      "Length",
      "Length",
      "Width",
      "Width",
      "",
      "mean",
      "sd",
      "mean",
      "sd",
      "mean",
      "sd",
      "mean",
      "sd",
      "",
      "zzz",
      "",
      "",
      "sfsf",
      "",
      "",
      "",
      "",
      "",
      "",
      "",
      "",
      "dsfsdf",
      "",
      "",
      "",
      ""
    )
  )
})


test_that("add part rows", {
  ft01 <- fp_text_default(color = "red")
  ft02 <- fp_text_default(color = "orange")

  pars <- as_paragraph(
    as_chunk(c("(1)", "(2)"), props = ft02),
    " ",
    as_chunk(
      c(
        "My tailor is rich",
        "My baker is rich"
      ),
      props = ft01
    )
  )

  ft_1 <- flextable(head(mtcars))
  ft_1 <- add_header_row(ft_1, values = pars, colwidths = c(5, 6), top = FALSE)
  ft_1 <- add_body_row(ft_1, values = pars, colwidths = c(5, 6), top = TRUE)
  ft_1 <- add_footer_row(ft_1, values = pars, colwidths = c(3, 8), top = FALSE)

  x <- information_data_chunk(ft_1)

  new_header_sel <- x[
    x$.part %in% "header" & x$.row_id %in% 2 & x$.col_id %in% "mpg",
  ]
  expect_equal(new_header_sel$txt, c("(1)", " ", "My tailor is rich"))
  expect_equal(new_header_sel$color, c("orange", "black", "red"))
  new_header_sel <- x[
    x$.part %in% "header" & x$.row_id %in% 2 & x$.col_id %in% "wt",
  ]
  expect_equal(new_header_sel$txt, c("(2)", " ", "My baker is rich"))
  expect_equal(new_header_sel$color, c("orange", "black", "red"))
  spans <- flextable:::fortify_span(ft_1, parts = "header")
  expect_equal(
    spans$rowspan,
    c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 5, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0)
  )
  expect_true(all(spans$colspan %in% 1))
  expect_equal(
    colSums(is.na(ft_1$header$dataset)),
    rep(0L, ncol(mtcars)),
    ignore_attr = TRUE
  )

  new_body_sel <- x[
    x$.part %in% "body" & x$.row_id %in% 1 & x$.col_id %in% "mpg",
  ]
  expect_equal(new_body_sel$txt, c("(1)", " ", "My tailor is rich"))
  expect_equal(new_body_sel$color, c("orange", "black", "red"))
  new_body_sel <- x[
    x$.part %in% "body" & x$.row_id %in% 1 & x$.col_id %in% "wt",
  ]
  expect_equal(new_body_sel$txt, c("(2)", " ", "My baker is rich"))
  expect_equal(new_body_sel$color, c("orange", "black", "red"))
  spans <- flextable:::fortify_span(ft_1, parts = "body")
  spans <- spans[spans$.row_id %in% 1, ]
  expect_equal(
    spans$rowspan,
    c(5, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0)
  )
  expect_true(all(spans$colspan %in% 1))
  expect_equal(
    colSums(is.na(ft_1$body$dataset)),
    rep(1L, ncol(mtcars)),
    ignore_attr = TRUE
  )

  new_footer_sel <- x[
    x$.part %in% "footer" & x$.row_id %in% 1 & x$.col_id %in% "mpg",
  ]
  expect_equal(new_footer_sel$txt, c("(1)", " ", "My tailor is rich"))
  new_footer_sel <- x[
    x$.part %in% "footer" & x$.row_id %in% 1 & x$.col_id %in% "hp",
  ]
  expect_equal(new_header_sel$txt, c("(2)", " ", "My baker is rich"))
  spans <- flextable:::fortify_span(ft_1, parts = "footer")
  expect_equal(
    spans$rowspan,
    c(3, 0, 0, 8, 0, 0, 0, 0, 0, 0, 0)
  )
  expect_true(all(spans$colspan %in% 1))
  expect_equal(
    colSums(is.na(ft_1$footer$dataset)),
    rep(0L, ncol(mtcars)),
    ignore_attr = TRUE
  )
})

test_that("add rows", {
  ft <- flextable(
    head(iris),
    col_keys = c(
      "Species",
      "Sepal.Length",
      "Petal.Length",
      "Sepal.Width",
      "Petal.Width"
    )
  )

  fun <- function(x) {
    paste0(
      c("min: ", "max: "),
      formatC(range(x))
    )
  }
  new_row <- list(
    Sepal.Length = fun(iris$Sepal.Length),
    Sepal.Width = fun(iris$Sepal.Width),
    Petal.Width = fun(iris$Petal.Width),
    Petal.Length = fun(iris$Petal.Length)
  )

  ft <- add_header(ft, values = new_row, top = FALSE)

  ft <- add_body(
    x = ft,
    Sepal.Length = 1:5,
    Sepal.Width = 1:5 * 2,
    Petal.Length = 1:5 * 3,
    Petal.Width = 1:5 + 10,
    Species = "Blah",
    top = FALSE
  )

  x <- information_data_chunk(ft)

  new_row_sel <- x[
    x$.part %in% "body" & x$.row_id %in% 7:11 & x$.col_id %in% "Species",
  ]
  expect_equal(new_row_sel$txt, rep("Blah", 5))

  new_row_sel <- x[
    x$.part %in% "body" & x$.row_id %in% 7:11 & x$.col_id %in% "Sepal.Length",
  ]
  expect_equal(new_row_sel$txt, as.character(1:5))

  expect_true(is.factor(ft$body$dataset[7:11, ]$Species))
  expect_equal(
    levels(ft$body$dataset[7:11, ]$Species),
    c("setosa", "versicolor", "virginica", "Blah")
  )
  expect_equal(as.character(ft$body$dataset[7:11, ]$Species), rep("Blah", 5))
  expect_equal(ft$body$dataset[7:11, ]$Sepal.Length, 1:5)

  new_header_sel <- x[
    x$.part %in% "header" & x$.row_id %in% 2:3 & x$.col_id %in% "Sepal.Width",
  ]
  expect_equal(new_header_sel$txt, c("min: 2", "max: 4.4"))
  new_header_sel <- x[
    x$.part %in% "header" & x$.row_id %in% 2:3 & x$.col_id %in% "Species",
  ]
  expect_equal(new_header_sel$txt, c("", ""))
})

#' @export
#' @title Create a chunk representation suitable for flextable
#' @description This function is to be used by external packages
#' that want to provide an object that can be inserted as a chunk
#' object in paragraphs of a flextable object.
#' @param ... values to set.
#' @section  text pattern with default values:
#'
#' ```
#' chunk_dataframe(txt = c("any text", "other text"))
#' ```
#' @section  text pattern with bold set to TRUE:
#'
#' ```
#' chunk_dataframe(
#'   txt = c("any text", "other text"),
#'   bold = c(TRUE, TRUE))
#' ```
#' @section  text pattern with control over all formatting properties:
#'
#' ```
#' chunk_dataframe(
#'   txt = c("any text", "other text"),
#'   font.size = c(12, 10),
#'   italic = c(FALSE, TRUE),
#'   bold = c(FALSE, TRUE),
#'   underlined = c(FALSE, TRUE),
#'   strike = c(FALSE, TRUE),
#'   color = c("black", "red"),
#'   shading.color = c("transparent", "yellow"),
#'   font.family = c("Arial", "Arial"),
#'   hansi.family = c("Arial", "Arial"),
#'   eastasia.family = c("Arial", "Arial"),
#'   cs.family = c("Arial", "Arial"),
#'   vertical.align = c("top", "bottom") )
#' ```
#' @section  text with url pattern:
#' ```
#' chunk_dataframe(
#'   txt = c("any text", "other text"),
#'   url = rep("https://www.google.fr", 2),
#'   font.size = c(12, 10),
#'   italic = c(FALSE, TRUE),
#'   bold = c(FALSE, TRUE),
#'   underlined = c(FALSE, TRUE),
#'   strike = c(FALSE, TRUE),
#'   color = c("black", "red"),
#'   shading.color = c("transparent", "yellow"),
#'   font.family = c("Arial", "Arial"),
#'   hansi.family = c("Arial", "Arial"),
#'   eastasia.family = c("Arial", "Arial"),
#'   cs.family = c("Arial", "Arial"),
#'   vertical.align = c("top", "bottom") )
#' ```
#' @section  images pattern:
#' ```
#' chunk_dataframe(width = width, height = height, img_data = files )
#' ```
#' @keywords internal
#' @return a data.frame with an additional class "chunk" that makes it
#' suitable for beeing used in [as_paragraph()]
chunk_dataframe <- function(...) {
  x <- list(...)

  img_data <- x$img_data
  x$img_data <- NULL
  x <- as.data.frame(x, stringsAsFactors = FALSE)
  def_chr <- rep(NA_character_, nrow(x))
  def_dbl <- rep(NA_real_, nrow(x))
  def_lgl <- as.logical(rep(NA_integer_, nrow(x)))
  def_lst <- rep(list(NULL), nrow(x))
  data0 <- data.frame(
    txt = def_chr,
    font.size = def_dbl,
    italic = def_lgl,
    bold = def_lgl,
    underlined = def_lgl,
    strike = def_lgl,
    color = def_chr,
    shading.color = def_chr,
    font.family = def_chr,
    hansi.family = def_chr,
    eastasia.family = def_chr,
    cs.family = def_chr,
    vertical.align = def_chr,
    width = def_dbl,
    height = def_dbl,
    url = def_chr,
    eq_data = def_chr,
    word_field_data = def_chr,
    qmd_data = def_chr,
    stringsAsFactors = FALSE
  )
  data0$img_data <- def_lst

  data0[names(x)] <- x
  if (!is.null(img_data)) {
    data0$img_data <- img_data
  }
  class(data0) <- c("chunk", "data.frame")
  data0
}

default_fptext_prop <- structure(
  list(
    font.size = NA_real_,
    bold = as.logical(NA_integer_),
    italic = as.logical(NA_integer_),
    underlined = as.logical(NA_integer_),
    strike = as.logical(NA_integer_),
    color = NA_character_,
    shading.color = NA_character_,
    font.family = NA_character_,
    hansi.family = NA_character_,
    eastasia.family = NA_character_,
    cs.family = NA_character_,
    vertical.align = NA_character_
  ),
  class = "fp_text"
)


#' @export
#' @title Text chunk
#' @description The function lets add formated text in flextable
#' cells.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' It should be used inside a call to [as_paragraph()].
#' @param x text or any element that can be formatted as text
#' with function provided in argument `formatter`.
#' @param props an [fp_text_default()] or [officer::fp_text()] object to be used to format the text.
#' If not specified, it will be the default value corresponding to the cell.
#' @param formatter a function that will format x as a character vector.
#' @param ... additional arguments for `formatter` function.
#' @family chunk elements for paragraph
#' @examples
#' library(officer)
#'
#' ft <- flextable(head(iris))
#'
#' ft <- compose(ft,
#'   j = "Sepal.Length",
#'   value = as_paragraph(
#'     "Sepal.Length value is ",
#'     as_chunk(Sepal.Length, props = fp_text(color = "red"))
#'   ),
#'   part = "body"
#' )
#' ft <- color(ft, color = "gray40", part = "all")
#' ft <- autofit(ft)
#' ft
as_chunk <- function(x, props = NULL, formatter = format_fun, ...) {
  if (is.function(x)) {
    stop("argument `x` in function `as_chunk` cannot be a function")
  }
  if (!is.function(formatter)) {
    stop("argument `formatter` in function `as_chunk` should be a function")
  }

  text <- formatter(x, ...)

  if (is.null(props)) {
    props <- default_fptext_prop
  } else if (!inherits(props, "fp_text")) {
    stop("argument `props` should be an object of class 'fp_text'.")
  }

  if (inherits(props, "fp_text")) {
    props <- rep(list(props), length(text))
  }

  if (length(props) > 0 && is.list(props)) {
    if (!all(sapply(props, inherits, "fp_text"))) {
      stop("`props` should be a list of `fp_text` object(s).")
    }
    if (length(props) != length(text)) {
      stop(sprintf(
        "`props` should be a list of same length than `x`: %.0f.",
        length(text)
      ))
    }
  }
  data <- chunk_dataframe(
    txt = text,
    font.size = sapply(props, function(x) x$font.size),
    italic = sapply(props, function(x) x$italic),
    bold = sapply(props, function(x) x$bold),
    underlined = sapply(props, function(x) x$underlined),
    strike = sapply(props, function(x) x$strike),
    color = sapply(props, function(x) x$color),
    shading.color = sapply(props, function(x) x$shading.color),
    font.family = sapply(props, function(x) x$font.family),
    hansi.family = sapply(props, function(x) x$hansi.family),
    eastasia.family = sapply(props, function(x) x$eastasia.family),
    cs.family = sapply(props, function(x) x$cs.family),
    vertical.align = sapply(props, function(x) x$vertical.align)
  )

  data
}


#' @export
#' @title Subscript chunk
#' @description The function is producing a chunk with
#' subscript vertical alignment.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#' @param x value, if a chunk, the chunk will be updated
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris), col_keys = c("dummy"))
#'
#' ft <- compose(ft,
#'   i = 1, j = "dummy", part = "header",
#'   value = as_paragraph(
#'     as_sub("Sepal.Length"),
#'     " anything "
#'   )
#' )
#'
#' ft <- autofit(ft)
#' ft
as_sub <- function(x) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }
  x$vertical.align <- "subscript"
  x
}
#' @export
#' @title Superscript chunk
#' @description The function is producing a chunk with
#' superscript vertical alignment.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @inheritParams as_sub
#' @note
#' This is a sugar function that ease the composition of complex
#' labels made of different formattings. It should be used inside a
#' call to [as_paragraph()].
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris), col_keys = c("dummy"))
#'
#' ft <- compose(ft,
#'   i = 1, j = "dummy", part = "header",
#'   value = as_paragraph(
#'     " anything ",
#'     as_sup("Sepal.Width")
#'   )
#' )
#'
#' ft <- autofit(ft)
#' ft
as_sup <- function(x) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }
  x$vertical.align <- "superscript"
  x
}


#' @export
#' @title Bold chunk
#' @description The function is producing a chunk with
#' bold font.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @inheritParams as_sub
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c("Sepal.Length", "dummy")
#' )
#'
#' ft <- compose(ft,
#'   j = "dummy",
#'   value = as_paragraph(
#'     as_b(Sepal.Length)
#'   )
#' )
#'
#' ft
as_b <- function(x) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }
  x$bold <- TRUE
  x
}

#' @export
#' @title Strikethrough chunk
#' @description The function is producing a chunk with
#' strikethrough font.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @inheritParams as_sub
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c("Sepal.Length", "dummy")
#' )
#'
#' ft <- compose(ft,
#'   j = "dummy",
#'   value = as_paragraph(
#'     as_strike(Sepal.Length)
#'   )
#' )
#'
#' ft
as_strike <- function(x) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }
  x$strike <- TRUE
  x
}

#' @export
#' @title Italic chunk
#' @description The function is producing a chunk with
#' italic font.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @inheritParams as_sub
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c("Sepal.Length", "dummy")
#' )
#'
#' ft <- compose(ft,
#'   j = "dummy",
#'   value = as_paragraph(as_i(Sepal.Length))
#' )
#'
#' ft
as_i <- function(x) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }
  x$italic <- TRUE
  x
}

#' @export
#' @title Colorize chunk
#' @description The function is producing a chunk with
#' a font in color.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @param color color to use as text highlighting color as character vector.
#' @inheritParams as_sub
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c("Sepal.Length", "dummy")
#' )
#'
#' ft <- compose(ft,
#'   j = "dummy",
#'   value = as_paragraph(colorize(Sepal.Length, color = "red"))
#' )
#'
#' ft
colorize <- function(x, color) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }

  x$color <- color
  x
}

#' @export
#' @title Highlight chunk
#' @description The function is producing a chunk with
#' an highlight chunk.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @param color color to use as text highlighting color as character vector.
#' @inheritParams as_sub
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c("Sepal.Length", "dummy")
#' )
#'
#' ft <- compose(ft,
#'   j = "dummy",
#'   value = as_paragraph(as_highlight(Sepal.Length, color = "yellow"))
#' )
#'
#' ft
as_highlight <- function(x, color) {
  if (!inherits(x, "chunk")) {
    x <- as_chunk(x, formatter = format_fun)
  }
  x$shading.color <- color
  x
}

#' @export
#' @title Bracket chunk
#' @description The function is producing a chunk by
#' pasting values and add the result in brackets.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' @param ... text and column names
#' @param sep separator
#' @param p prefix, default to '('
#' @param s suffix, default to ')'
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris),
#'   col_keys = c("Species", "Sepal", "Petal")
#' )
#' ft <- set_header_labels(ft, Sepal = "Sepal", Petal = "Petal")
#' ft <- compose(ft,
#'   j = "Sepal",
#'   value = as_paragraph(as_bracket(Sepal.Length, Sepal.Width))
#' )
#' ft <- compose(ft,
#'   j = "Petal",
#'   value = as_paragraph(as_bracket(Petal.Length, Petal.Width))
#' )
#' ft
as_bracket <- function(..., sep = ", ", p = "(", s = ")") {
  x <- list(...)
  x <- lapply(x, formatC)
  x$sep <- sep
  x <- do.call(paste, x)
  x <- paste0(p, x, s)
  as_chunk(x)
}

#' @export
#' @title Hyperlink chunk
#' @description The function lets add hyperlinks within flextable
#' objects.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' URL are not encoded, they are preserved 'as is'.
#'
#' @note
#' This chunk option requires package officedown in a R Markdown
#' context with Word output format. With Quarto (`format: docx`) or
#' `rmarkdown::word_document()`, the resulting file must be repaired
#' with [repair_docx()].
#' @inheritParams as_chunk
#' @param url url to be used
#' @seealso [compose()]
#' @examples
#' dat <- data.frame(
#'   col = "Google it",
#'   href = "https://www.google.fr/search?source=hp&q=flextable+R+package",
#'   stringsAsFactors = FALSE
#' )
#'
#' ftab <- flextable(dat)
#' ftab <- compose(
#'   x = ftab, j = "col",
#'   value = as_paragraph(
#'     "This is a link: ",
#'     hyperlink_text(x = col, url = href)
#'   )
#' )
#' ftab
#' @family chunk elements for paragraph
hyperlink_text <- function(x, props = NULL, formatter = format_fun, url, ...) {
  zz <- data.frame(x = x, url = url, stringsAsFactors = FALSE)
  x <- as_chunk(x = zz$x, props = props, formatter = formatter, ...)
  x$url <- zz$url
  x
}

#' @export
#' @title Equation chunk
#' @description This function is used to insert equations into
#' flextable.
#'
#' It is used to add it to the content of a cell of the
#' flextable with the functions [compose()], [append_chunks()]
#' or [prepend_chunks()].
#'
#' To use this function, package 'equatags' is required;
#' also `equatags::mathjax_install()` must be executed only once
#' to install necessary dependencies.
#'
#' @param x values containing the 'MathJax' equations
#' @param props an [fp_text_default()] or [officer::fp_text()] object to be used to format the text.
#' If not specified, it will be the default value corresponding to the cell.
#' @param width,height size of the resulting equation
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @family chunk elements for paragraph
#' @examples
#' library(flextable)
#' if (require("equatags")) {
#'   eqs <- c(
#'     "(ax^2 + bx + c = 0)",
#'     "a \\ne 0",
#'     "x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}"
#'   )
#'   df <- data.frame(formula = eqs)
#'   df
#'
#'
#'   ft <- flextable(df)
#'   ft <- compose(
#'     x = ft, j = "formula",
#'     value = as_paragraph(as_equation(formula, width = 2, height = .5))
#'   )
#'   ft <- align(ft, align = "center", part = "all")
#'   ft <- width(ft, width = 2)
#'   ft
#' }
as_equation <- function(x, width = 1, height = .2, unit = "in", props = NULL) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (length(x) > 1) {
    if (length(width) == 1) {
      width <- rep(width, length(x))
    }
    if (length(height) == 1) height <- rep(height, length(x))
  }
  if (is.null(props)) {
    props <- default_fptext_prop
  }

  if (inherits(props, "fp_text")) {
    props <- rep(list(props), length(x))
  }

  if (length(props) > 0 && is.list(props)) {
    if (!all(sapply(props, inherits, "fp_text"))) {
      stop("`props` should be a list of `fp_text` object(s).")
    }
    if (length(props) != length(x)) {
      stop(sprintf(
        "`props` should be a list of same length than `x`: %.0f.",
        length(x)
      ))
    }
  }

  x <- chunk_dataframe(
    width = as.double(width),
    height = as.double(height),
    eq_data = x,
    font.size = sapply(props, function(x) x$font.size),
    italic = sapply(props, function(x) x$italic),
    bold = sapply(props, function(x) x$bold),
    underlined = sapply(props, function(x) x$underlined),
    strike = sapply(props, function(x) x$strike),
    color = sapply(props, function(x) x$color),
    shading.color = sapply(props, function(x) x$shading.color),
    font.family = sapply(props, function(x) x$font.family),
    hansi.family = sapply(props, function(x) x$hansi.family),
    eastasia.family = sapply(props, function(x) x$eastasia.family),
    cs.family = sapply(props, function(x) x$cs.family),
    vertical.align = sapply(props, function(x) x$vertical.align)
  )
  class(x) <- c("chunk", "data.frame")
  x
}


#' @export
#' @title Word dynamic field chunk
#' @description
#' `as_word_field()` inserts a Word field code (e.g. page
#' numbers, dates, cross-references) as a chunk inside a
#' flextable cell. Field codes are Word's mechanism for
#' auto-computed values; see
#' [Microsoft's field-code reference](https://support.microsoft.com/en-US/Word/list-of-field-codes-in-word)
#' for the available codes.
#'
#' The chunk is used with [compose()], [append_chunks()]
#' or [prepend_chunks()]. It only has an effect in Word
#' (docx) output; other formats ignore it. To apply it
#' conditionally, use the post-processing step (see
#' `set_flextable_defaults(post_process_docx = ...)`).
#'
#' **Important**: fields are inserted but not computed.
#' After opening the document in Word, select all text
#' and press `F9` (on macOS: `Fn + F9`) to refresh the
#' field values.
#' @param x computed field strings
#' @param props text properties (see [fp_text_default()] or [officer::fp_text()])
#' object to be used to format the text. If not specified, it will use
#' the default text properties of the cell(s).
#' @param width,height size computed field
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @family chunk elements for paragraph
#' @examples
#' library(flextable)
#'
#' # define some default values ----
#' set_flextable_defaults(font.size = 22, border.color = "gray")
#'
#' # an example with append_chunks ----
#' pp_docx <- function(x) {
#'   x <- add_header_lines(x, "Page ")
#'   x <- append_chunks(
#'     x = x, i = 1, part = "header", j = 1,
#'     as_word_field(x = "Page")
#'   )
#'   align(x, part = "header", align = "left")
#' }
#' ft_1 <- flextable(cars)
#' ft_1 <- autofit(ft_1)
#' ft_1 <- pp_docx(ft_1)
#'
#' ## or:
#' # set_flextable_defaults(post_process_docx = pp_docx)
#' ## to prevent this line addition when output is not docx
#'
#' # print(ft_1, preview = "docx")
#'
#' # an example with compose ----
#'
#' library(officer)
#' ft_2 <- flextable(head(cars))
#' ft_2 <- add_footer_lines(ft_2, "temp text")
#' ft_2 <- compose(
#'   x = ft_2, part = "footer", i = 1, j = 1,
#'   as_paragraph(
#'     "p. ",
#'     as_word_field(x = "Page", width = .05),
#'     " on ", as_word_field(x = "NumPages", width = .05)
#'   )
#' )
#' ft_2 <- autofit(ft_2, part = c("header", "body"))
#'
#' doc <- read_docx()
#' doc <- body_add_flextable(doc, ft_2)
#' doc <- body_add_break(doc)
#' doc <- body_add_flextable(doc, ft_2)
#' outfile <- print(doc, target = tempfile(fileext = ".docx"))
#'
#' # reset default values ----
#' init_flextable_defaults()
as_word_field <- function(
  x,
  props = NULL,
  width = .1,
  height = .15,
  unit = "in"
) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (is.null(props)) {
    props <- default_fptext_prop
  }

  if (inherits(props, "fp_text")) {
    props <- rep(list(props), length(x))
  }

  if (length(props) > 0 && is.list(props)) {
    if (!all(sapply(props, inherits, "fp_text"))) {
      stop("`props` should be a list of `fp_text` object(s).")
    }
    if (length(props) != length(x)) {
      stop(sprintf(
        "`props` should be a list of same length than `x`: %.0f.",
        length(x)
      ))
    }
  }

  x <- chunk_dataframe(
    width = as.double(width),
    height = as.double(height),
    word_field_data = x,
    font.size = sapply(props, function(x) x$font.size),
    italic = sapply(props, function(x) x$italic),
    bold = sapply(props, function(x) x$bold),
    underlined = sapply(props, function(x) x$underlined),
    strike = sapply(props, function(x) x$strike),
    color = sapply(props, function(x) x$color),
    shading.color = sapply(props, function(x) x$shading.color),
    font.family = sapply(props, function(x) x$font.family),
    hansi.family = sapply(props, function(x) x$hansi.family),
    eastasia.family = sapply(props, function(x) x$eastasia.family),
    cs.family = sapply(props, function(x) x$cs.family),
    vertical.align = sapply(props, function(x) x$vertical.align)
  )
  class(x) <- c("chunk", "data.frame")
  x
}
to_wml_word_field <- function(x, pr_txt) {
  xml_elt_1 <- paste0(
    "<w:r>",
    pr_txt,
    "<w:fldChar w:fldCharType=\"begin\" w:dirty=\"true\"/>",
    "</w:r>"
  )
  xml_elt_2 <- paste0(
    "<w:r>",
    pr_txt,
    sprintf(
      "<w:instrText xml:space=\"preserve\" w:dirty=\"true\">%s</w:instrText>",
      x
    ),
    "</w:r>"
  )
  xml_elt_3 <- paste0(
    "<w:r>",
    pr_txt,
    "<w:fldChar w:fldCharType=\"end\" w:dirty=\"true\"/>",
    "</w:r>"
  )
  out <- paste0(xml_elt_1, xml_elt_2, xml_elt_3)

  out
}

#' @export
#' @title Quarto inline markdown chunk
#' @description
#' `as_qmd()` creates a chunk for inline Quarto markdown
#' content (text-level) that fits within a table
#' cell paragraph. This enables cross-references
#' (`@fig-xxx`, `@tbl-xxx`), links, bold/italic, math,
#' inline code, shortcodes and other inline Quarto markdown
#' features inside flextable cells.
#'
#' It is not designed for block-level elements such as
#' headings, bullet lists or fenced code blocks.
#'
#' The chunk is used with [compose()], [append_chunks()]
#' or [prepend_chunks()]. It requires the `flextable-qmd` Lua
#' filter extension (see [use_flextable_qmd()]) and works with
#' HTML, PDF, Typst and Word (docx) Quarto output formats.
#'
#' @section Setup:
#'
#' 1. Install the extension once per project:
#'
#' ```r
#' flextable::use_flextable_qmd()
#' ```
#'
#' 2. Add the filter to your Quarto document YAML.
#' For HTML and PDF, a single line is enough:
#'
#' ```yaml
#' filters:
#'   - flextable-qmd
#' ```
#'
#' For Word (docx), an additional post-render filter
#' removes the wrapper table that Quarto adds around
#' labelled flextables:
#'
#' ```yaml
#' filters:
#'   - flextable-qmd
#'   - at: post-render
#'     path: _extensions/flextable-qmd/unwrap-float.lua
#' ```
#'
#' @section Supported markdown:
#'
#' - Cross-references: `@fig-xxx`, `@tbl-xxx`
#' - Bold / italic: `**bold**`, `*italic*`
#' - Inline code: `` `code` ``
#' - Links: `[text](url)` (internal and external)
#' - Math: `$\\alpha + \\beta$`
#' - Shortcodes and other Quarto markdown constructs
#'
#' @section Limitations:
#'
#' Each table cell in a flextable contains a single paragraph
#' built from inline chunks (see [as_paragraph()]). There is no
#' mechanism to insert block-level structures (multiple
#' paragraphs, lists, headings, fenced code blocks, callouts,
#' etc.) inside a cell. Because `as_qmd()` produces one of
#' these inline chunks, only inline markdown is supported.
#'
#' @param x character vector of Quarto markdown content.
#' @param display character vector of display text used
#' as fallback when the Lua filter is not active.
#' Defaults to `x`.
#' @family chunk elements for paragraph
#' @seealso [use_flextable_qmd()] to install the Lua filter extension,
#' [knit_print.flextable()] for rendering options in knitr documents.
#' @examples
#' library(flextable)
#'
#' dat <- data.frame(
#'   label = c("Bold", "Link", "Code"),
#'   content = c(
#'     "This is **bold** text",
#'     "Visit [Quarto](https://quarto.org)",
#'     "Use `print()` here"
#'   )
#' )
#' ft <- flextable(dat)
#' ft <- mk_par(ft, j = "content",
#'   value = as_paragraph(as_qmd(content)))
#' ft
as_qmd <- function(x, display = x) {
  x <- as.character(x)
  display <- as.character(display)
  if (length(display) != length(x)) {
    stop(
      sprintf("`display` must be the same length as `x`: %.0f.", length(x)),
      call. = FALSE
    )
  }
  chunk_dataframe(
    txt = display,
    qmd_data = x
  )
}

#' @export
#' @title Install the flextable-qmd Quarto extension
#' @description
#' Copies the `flextable-qmd` Quarto extension (bundled with
#' flextable) into the `_extensions/` directory of a
#' Quarto project. The extension provides Lua filters
#' that resolve Quarto markdown content produced by
#' [as_qmd()] inside flextable cells for HTML, PDF,
#' Typst and Word (docx) output formats.
#'
#' After installation, add the filter to your document
#' or project YAML:
#'
#' ```yaml
#' filters:
#'   - flextable-qmd
#' ```
#'
#' This single filter is all that is required for HTML,
#' PDF and Typst (`format: typst`) output.
#'
#' For Word (docx) output with labelled flextable chunks
#' (e.g. `#| label: tbl-xxx`), add the post-render filter
#' to remove the wrapper table Quarto creates around the
#' flextable (this extra filter is not needed for Typst,
#' where the table is embedded directly in the figure):
#'
#' ```yaml
#' filters:
#'   - flextable-qmd
#'   - at: post-render
#'     path: _extensions/flextable-qmd/unwrap-float.lua
#' ```
#' @param path Path to the Quarto project root. Defaults
#'   to the current working directory.
#' @param quiet If `TRUE`, suppress informational messages.
#' @return The path to the installed extension (invisibly).
#' @seealso [as_qmd()] for creating Quarto markdown chunks,
#' [knit_print.flextable()] for rendering options in knitr documents.
#' @family flextable_configuration
#' @examples
#' \dontrun{
#' use_flextable_qmd()
#' }
use_flextable_qmd <- function(path = ".", quiet = FALSE) {
  src <- system.file(
    "_extensions",
    "flextable-qmd",
    package = "flextable",
    mustWork = TRUE
  )
  dest <- file.path(path, "_extensions", "flextable-qmd")
  if (!dir.exists(dirname(dest))) {
    dir.create(dirname(dest), recursive = TRUE)
  }
  if (dir.exists(dest)) {
    unlink(dest, recursive = TRUE)
  }
  file.copy(src, dirname(dest), recursive = TRUE)
  if (!quiet) {
    message(
      "Installed flextable-qmd extension in ",
      dest,
      ".\n",
      "Add `filters: [flextable-qmd]` to your document YAML."
    )
  }
  invisible(dest)
}

#' @export
#' @title Build a paragraph from chunks
#' @description
#' `as_paragraph()` assembles one or more chunks into a single paragraph
#' that defines the content of a flextable cell. Each cell in a flextable
#' contains exactly one paragraph; a paragraph is an ordered sequence of
#' chunks.
#'
#' Chunks are the smallest content units and can be created with
#' [as_chunk()] (formatted text), [as_b()] / [as_i()] (bold / italic
#' shortcuts), [minibar()] (inline bar), [as_image()] (image),
#' [gg_chunk()] (ggplot), [as_equation()] (equation) or
#' [hyperlink_text()] (link). Plain character strings passed to
#' `as_paragraph()` are automatically converted to chunks via
#' [as_chunk()].
#'
#' The resulting paragraph is passed to the `value` argument of
#' [compose()], [mk_par()], [add_header_lines()],
#' [add_footer_lines()] or [footnote()] to set cell content.
#'
#' @param ... chunk elements that are defining the paragraph content.
#' If a character is used, it is transformed to a chunk object with
#' function [as_chunk()].
#' @param list_values a list of chunk elements that are defining
#' the paragraph content. If specified argument `...` is unused.
#' @family cell_content_composition
#' @seealso [as_chunk()], [minibar()],
#' [as_image()], [hyperlink_text()]
#' @examples
#' library(flextable)
#' ft <- flextable(airquality[sample.int(150, size = 10), ])
#' ft <- compose(ft,
#'   j = "Wind",
#'   value = as_paragraph(
#'     as_chunk(Wind, props = fp_text_default(color = "orange")),
#'     " ",
#'     minibar(value = Wind, max = max(airquality$Wind), barcol = "orange", bg = "black", height = .15)
#'   ),
#'   part = "body"
#' )
#' ft <- autofit(ft)
#' ft
as_paragraph <- function(..., list_values = NULL) {
  if (is.null(list_values)) {
    list_values <- list(...)
  }

  is_atomic <- sapply(list_values, is.atomic)
  if (any(is_atomic)) {
    list_values[is_atomic] <- lapply(list_values[is_atomic], function(x) {
      as_chunk(x, formatter = format_fun)
    })
  }

  if (!all(sapply(list_values, inherits, "chunk"))) {
    stop("argument ... should only contain objects of class 'chunk'.")
  }

  nrows <- sapply(list_values, nrow)
  exp_nrow <- max(nrows)

  id_recycl <- which(nrows == 1)
  if (length(nrows) != 1 && length(id_recycl) > 0) {
    list_values[id_recycl] <- lapply(
      list_values[id_recycl],
      function(x, n) {
        z <- rbind_match_columns(rep(list(x), n))
        z
      },
      exp_nrow
    )
  }

  nrows <- sapply(list_values, nrow)

  if (length(unique(nrows)) != 1 && 1 %in% nrows) {
    which_ <- which(nrows %in% 1)
    list_values[which_] <- lapply(
      list_values[which_],
      function(x, n) {
        rbind_match_columns(rep(list(x), n))
      },
      n = max(nrows, na.rm = TRUE)
    )
  }
  nrows <- sapply(list_values, nrow)

  if (length(unique(nrows)) != 1) {
    stop("paragraph elements should all have the same length")
  }

  data <- mapply(
    function(x, index) {
      x$.chunk_index <- rep(index, nrow(x))
      x
    },
    list_values,
    seq_along(list_values),
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  data <- rbind_match_columns(data)

  data <- split(
    data,
    rep(seq_len(nrow(list_values[[1]])), length((list_values)))
  )
  names(data) <- NULL
  class(data) <- c("paragraph")
  data
}

img_file <- file.path(R.home("doc"), "html", "logo.jpg")

ft_img <- function() {
  ft <- flextable(data.frame(a = "x"))
  compose(
    ft,
    j = "a",
    value = as_paragraph(as_image(img_file, width = .5, height = .4))
  )
}

test_that("knit_to_wml warns when images can not be embedded by pandoc", {
  old <- knitr::opts_knit$get("rmarkdown.pandoc.to")
  knitr::opts_knit$set("rmarkdown.pandoc.to" = "docx")
  on.exit(knitr::opts_knit$set("rmarkdown.pandoc.to" = old))

  expect_warning(
    flextable:::knit_to_wml(ft_img(), quarto = TRUE),
    "repair_docx"
  )
  expect_no_warning(
    flextable:::knit_to_wml(flextable(head(cars, 2)), quarto = TRUE)
  )

  # hyperlinks are also unresolved references in pandoc-written docx
  ft_url <- flextable(data.frame(a = "flextable"))
  ft_url <- compose(
    ft_url,
    j = "a",
    value = as_paragraph(
      hyperlink_text(x = a, url = "https://davidgohel.github.io/flextable/")
    )
  )
  expect_true(flextable:::has_hlink(ft_url))
  expect_false(flextable:::has_hlink(flextable(head(cars, 2))))
  expect_warning(
    flextable:::knit_to_wml(ft_url, quarto = TRUE),
    "repair_docx"
  )

  # no warning when officedown post-processes the document
  knitr::opts_current$set(is_rdocx_document = TRUE)
  on.exit(knitr::opts_current$set(is_rdocx_document = NULL), add = TRUE)
  expect_no_warning(flextable:::knit_to_wml(ft_img()))
})

test_that("repair_docx registers image references", {
  # simulate a pandoc-written docx: image referenced by path, not
  # registered in the package
  path <- tempfile(fileext = ".docx")
  z <- officer::read_docx()
  z <- officer::body_add_par(z, "placeholder")
  print(z, target = path)

  dir <- tempfile()
  officer::unpack_folder(path, dir)
  doc_xml_path <- file.path(dir, "word", "document.xml")
  doc <- readLines(doc_xml_path, warn = FALSE)
  run <- flextable:::wml_image(img_file, width = .5, height = .4)
  doc <- sub(
    "placeholder",
    paste0("placeholder", "</w:t></w:r>", run, "<w:r><w:t>"),
    doc,
    fixed = TRUE
  )
  writeLines(doc, doc_xml_path, useBytes = TRUE)
  broken <- officer::pack_folder(dir, tempfile(fileext = ".docx"))

  fixed <- tempfile(fileext = ".docx")
  repair_docx(broken, target = fixed)

  out <- tempfile()
  officer::unpack_folder(fixed, out)
  xml <- paste(
    readLines(file.path(out, "word", "document.xml"), warn = FALSE),
    collapse = ""
  )
  expect_match(xml, "r:embed=\"rId[0-9]+\"")
  expect_no_match(xml, "r:embed=\"[^\"]*logo\\.jpg\"")
  expect_length(list.files(file.path(out, "word", "media")), 1)

  expect_error(repair_docx(tempfile(fileext = ".docx")), "existing")
  expect_error(repair_docx(img_file), "docx")
})

data <- data.frame(
  code = c("X01", "X02"),
  name = c("X Number 1", "X Number 2"),
  stringsAsFactors = FALSE
)
url_base <- "https://example.com?/path&project=%s"
ft <- flextable(data)
ft <- mk_par(
  x = ft,
  j = ~code,
  value = as_paragraph(
    hyperlink_text(code, url = sprintf(url_base, code))
  )
)


test_that("URL are preserved in docx", {
  outfile <- tempfile(fileext = ".docx")
  save_as_docx(ft, path = outfile)
  doc <- read_docx(path = outfile)
  body <- docx_body_xml(doc)
  rid <- xml_attr(xml_find_all(body, "//w:hyperlink"), "id")
  rels <- doc$doc_obj$rel_df()
  urls <- rels[rels$id %in% rid, "target"]
  expect_equal(urls, sprintf(url_base, data$code), ignore_attr = TRUE)
})

test_that("URL are preserved in pptx", {
  outfile <- tempfile(fileext = ".pptx")
  save_as_pptx(ft, path = outfile)
  doc <- read_pptx(path = outfile)
  xml_slide <- doc$slide$get_slide(1)$get()
  rid <- xml_attr(xml_find_all(xml_slide, "//a:hlinkClick"), "id")
  rels <- doc$slide$get_slide(1)$rel_df()
  urls <- rels[rels$id %in% rid, "target"]
  expect_equal(urls, sprintf(url_base, data$code), ignore_attr = TRUE)
})

test_that("URL are preserved in html", {
  str_ <- flextable:::gen_raw_html(ft)
  str_ <- gsub("<style>(.*)</style>", "", str_)
  str_ <- gsub("<script>(.*)</script>", "", str_)
  str_ <- gsub("<template id=\"[0-9a-z\\-]+\">", "", str_)
  str_ <- gsub("</div></template(.*)", "", str_)
  doc <- read_html(str_)
  urls <- xml_attr(xml_find_all(doc, "//a"), "href")
  expect_equal(urls, sprintf(url_base, data$code), ignore_attr = TRUE)
})

ft1 <- flextable(data.frame(a = "1 < 3", stringsAsFactors = FALSE))

get_xml_doc <- function(tab, main_folder = "docx_folder") {
  docx_file <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- body_add_flextable(doc, value = tab)
  print(doc, target = docx_file)

  main_folder <- file.path(getwd(), main_folder)
  unlink(main_folder, recursive = TRUE, force = TRUE)

  unpack_folder(file = docx_file, folder = main_folder)
  doc_file <- file.path(main_folder, "/word/document.xml")
  read_xml(doc_file)
}

get_xml_ppt <- function(tab, main_folder = "pptx_folder") {
  pptx_file <- tempfile(fileext = ".pptx")
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, tab, location = ph_location_type(type = "body"))
  print(doc, target = pptx_file)

  main_folder <- file.path(getwd(), main_folder)
  unlink(main_folder, recursive = TRUE, force = TRUE)
  unpack_folder(file = pptx_file, folder = main_folder)
  doc_file <- file.path(main_folder, "/ppt/slides/slide1.xml")
  read_xml(doc_file)
}

test_that("docx - string are html encoded", {
  main_folder <- "docx_folder"

  doc <- get_xml_doc(tab = ft1, main_folder = main_folder)
  text_ <- xml_text(xml_find_first(
    doc,
    "w:body/w:tbl[1]/w:tr[2]/w:tc/w:p/w:r/w:t"
  ))
  expect_equal(text_, c("1 < 3"))

  unlink(main_folder, recursive = TRUE, force = TRUE)
})

test_that("pptx - string are html encoded", {
  main_folder <- "pptx_folder"

  doc <- get_xml_ppt(tab = ft1, main_folder = main_folder)
  text_ <- xml_text(xml_find_first(
    doc,
    "//a:tbl/a:tr[2]/a:tc/a:txBody/a:p/a:r/a:t"
  ))
  expect_equal(text_, c("1 < 3"))

  unlink(main_folder, recursive = TRUE, force = TRUE)
})

test_that("html - string are html encoded", {
  str_ <- flextable:::gen_raw_html(ft1)
  str_ <- gsub("<style>(.*)</style>", "", str_)
  str_ <- gsub("<script>(.*)</script>", "", str_)
  str_ <- gsub("<template id=\"[0-9a-z\\-]+\">", "", str_)
  str_ <- gsub("</div></template(.*)", "", str_)
  doc <- read_xml(str_)
  text_ <- xml_text(xml_find_first(doc, "//tbody/tr/td/p/span"))
  expect_equal(text_, c("1 < 3"))
})
test_that("utf8 is preserved", {
  ft <- flextable(data.frame(a = c("\U2265 4.7")))
  str_ <- flextable:::gen_raw_html(ft)
  str_ <- gsub("<style>(.*)</style>", "", str_)
  str_ <- gsub("<script>(.*)</script>", "", str_)
  str_ <- gsub("<template id=\"[0-9a-z\\-]+\">", "", str_)
  str_ <- gsub("</div></template(.*)", "", str_)
  doc <- read_xml(str_)
  text_ <- xml_text(xml_find_first(doc, "//tbody/tr/td/p/span"))
  expect_equal(text_, c("\U2265 4.7"))
})

test_that("NA managment", {
  x <- data.frame(zz = c("a", NA_character_), stringsAsFactors = FALSE)
  ft1 <- flextable(x)

  str_ <- flextable:::gen_raw_html(ft1)
  str_ <- paste0("<div>", str_, "</div>")
  doc <- read_xml(str_)
  text_ <- xml_text(xml_find_all(doc, "//table/tbody/tr/td/p"))
  expect_equal(text_, c("a", ""))
})

test_that("newlines and tabulations expand correctly", {
  z <- flextable(data.frame(a = c("")))
  z <- compose(
    x = z,
    i = 1,
    j = 1,
    value = as_paragraph(
      "this\nis\nit",
      "\n\t",
      "This",
      "\n",
      "is",
      "\n",
      "it",
      "\n",
      "this\tis\tit\nthatsit\n"
    )
  )
  z <- delete_part(z, part = "header")

  chunks_txt <- information_data_chunk(z)$txt

  expect_equal(
    chunks_txt,
    c(
      "this",
      "<br>",
      "is",
      "<br>",
      "it",
      "<br>",
      "<tab>",
      "This",
      "<br>",
      "is",
      "<br>",
      "it",
      "<br>",
      "this",
      "<tab>",
      "is",
      "<tab>",
      "it",
      "<br>",
      "thatsit",
      "<br>"
    )
  )
})

test_that("word superscript and subscript", {
  ft <- flextable(data.frame(a = ""), col_keys = c("dummy"))
  ft <- delete_part(ft, part = "header")

  ft <- compose(
    ft,
    i = 1,
    j = "dummy",
    part = "body",
    value = as_paragraph(
      as_sub("Sepal.Length")
    )
  )
  runs <- information_data_chunk(ft)
  expect_equal(runs$vertical.align[1], "subscript")
  openxml <- flextable:::runs_as_wml(ft, txt_data = runs)$run_openxml
  expect_match(openxml, "<w:vertAlign w:val=\"subscript\"/>", fixed = TRUE)

  ft <- compose(
    ft,
    i = 1,
    j = "dummy",
    part = "body",
    value = as_paragraph(
      as_sup("Sepal.Length")
    )
  )
  runs <- information_data_chunk(ft)
  expect_equal(runs$vertical.align[1], "superscript")
  openxml <- flextable:::runs_as_wml(ft, txt_data = runs)$run_openxml
  expect_match(openxml, "<w:vertAlign w:val=\"superscript\"/>", fixed = TRUE)
})

test_that("highlight", {
  ft <- flextable(data.frame(test = "test"))
  ft <- add_footer_lines(ft, values = "test")
  ft <- mk_par(
    ft,
    j = "test",
    part = "body",
    value = as_paragraph(
      as_highlight(test, color = "yellow")
    )
  )
  ft <- mk_par(
    ft,
    j = "test",
    part = "header",
    value = as_paragraph(
      as_highlight(test, color = "red")
    )
  )
  ft <- mk_par(
    ft,
    j = "test",
    part = "footer",
    value = as_paragraph(
      as_highlight(test, color = "purple")
    )
  )

  runs <- information_data_chunk(ft)
  expect_equal(runs$shading.color, c("red", "yellow", "purple"))

  openxml <- flextable:::runs_as_wml(ft, txt_data = runs)$run_openxml
  expect_match(
    openxml[1],
    "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"FF0000\"/>",
    fixed = TRUE
  )
  expect_match(
    openxml[2],
    "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"FFFF00\"/>",
    fixed = TRUE
  )
  expect_match(
    openxml[3],
    "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"A020F0\"/>",
    fixed = TRUE
  )

  openxml <- flextable:::runs_as_pml(ft)$par_nodes_str
  expect_match(
    openxml[1],
    "<a:highlight><a:srgbClr val=\"FF0000\">",
    fixed = TRUE
  )
  expect_match(
    openxml[2],
    "<a:highlight><a:srgbClr val=\"FFFF00\">",
    fixed = TRUE
  )
  expect_match(
    openxml[3],
    "<a:highlight><a:srgbClr val=\"A020F0\">",
    fixed = TRUE
  )

  html_info <- flextable:::runs_as_html(ft)
  css <- attr(html_info, "css")
  expect_match(css, "background-color:rgba(255, 0, 0, 1.00);}", fixed = TRUE)
  expect_match(css, "background-color:rgba(255, 255, 0, 1.00);}", fixed = TRUE)
  expect_match(css, "background-color:rgba(160, 32, 240, 1.00);}", fixed = TRUE)

  latex_str <- flextable:::runs_as_latex(ft)$txt
  expect_match(latex_str[1], "\\colorbox[HTML]{FF0000}", fixed = TRUE)
  expect_match(latex_str[2], "\\colorbox[HTML]{FFFF00}", fixed = TRUE)
  expect_match(latex_str[3], "\\colorbox[HTML]{A020F0}", fixed = TRUE)

  rtf_str <- flextable:::runs_as_rtf(ft)$txt
  expect_match(rtf_str[1], "%ftshading:red%", fixed = TRUE)
  expect_match(rtf_str[2], "%ftshading:yellow%", fixed = TRUE)
  expect_match(rtf_str[3], "%ftshading:purple%", fixed = TRUE)
})

test_that("as_bracket", {
  ft <- flextable(head(iris), col_keys = c("Species", "what"))
  ft <- mk_par(
    ft,
    j = "what",
    part = "body",
    value = as_paragraph(
      as_bracket(Sepal.Length, Sepal.Width, sep = "-")
    )
  )

  runs <- information_data_chunk(ft)
  runs <- runs[runs$.part %in% "body", ]
  expect_equal(
    runs$txt,
    c(
      "setosa",
      "(5.1-3.5)",
      "setosa",
      "(4.9-3)",
      "setosa",
      "(4.7-3.2)",
      "setosa",
      "(4.6-3.1)",
      "setosa",
      "(5-3.6)",
      "setosa",
      "(5.4-3.9)"
    )
  )
})

test_that("as_equation", {
  skip_if_not_installed("equatags")

  eqs <- c(
    "(ax^2 + bx + c = 0)",
    "a \\ne 0",
    "x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}"
  )
  df <- data.frame(formula = eqs)

  ft <- flextable(df)
  ft <- mk_par(
    ft,
    j = "formula",
    part = "body",
    value = as_paragraph(as_equation(formula, width = 2, height = .5))
  )

  runs <- information_data_chunk(ft)
  openxml <- flextable:::runs_as_wml(ft, txt_data = runs)$run_openxml
  expect_match(openxml[2], "</m:oMath>", fixed = TRUE)
  expect_match(openxml[3], "</m:oMath>", fixed = TRUE)
  expect_match(openxml[4], "</m:oMath>", fixed = TRUE)

  openxml <- flextable:::runs_as_pml(ft)$par_nodes_str
  expect_match(openxml[2], "</m:oMath>", fixed = TRUE)
  expect_match(openxml[3], "</m:oMath>", fixed = TRUE)
  expect_match(openxml[4], "</m:oMath>", fixed = TRUE)

  latex_str <- flextable:::runs_as_latex(ft)$txt
  expect_match(latex_str[2], eqs[1], fixed = TRUE)
  expect_match(latex_str[3], eqs[2], fixed = TRUE)
  expect_match(latex_str[4], eqs[3], fixed = TRUE)

  runs <- information_data_chunk(ft)
  html_str <- flextable:::runs_as_html(ft)$span_tag
  expect_match(html_str[2], "<span class=\"katex\">", fixed = TRUE)
  expect_match(html_str[3], "<span class=\"katex\">", fixed = TRUE)
  expect_match(html_str[4], "<span class=\"katex\">", fixed = TRUE)
})

test_that("as_word_field", {
  ft <- flextable(head(cars))
  ft <- add_footer_lines(ft, "temp text")
  ft <- compose(
    x = ft,
    part = "footer",
    i = 1,
    j = 1,
    as_paragraph(
      as_word_field(x = "Page", width = .05)
    )
  )

  runs <- information_data_chunk(ft)
  wml_str <- flextable:::runs_as_wml(ft)
  wml_str <- wml_str[wml_str$.part %in% "footer", ]$run_openxml[1]
  expect_match(
    wml_str,
    "<w:instrText xml:space=\"preserve\" w:dirty=\"true\">Page</w:instrText>",
    fixed = TRUE
  )
})

test_that("strike", {
  ft <- flextable(data.frame(test = "test"))
  ft <- add_footer_lines(ft, values = "test")
  ft <- mk_par(
    ft,
    j = "test",
    part = "body",
    value = as_paragraph(
      as_strike(test)
    )
  )
  ft <- mk_par(
    ft,
    j = "test",
    part = "header",
    value = as_paragraph(
      as_strike(test)
    )
  )
  ft <- mk_par(
    ft,
    j = "test",
    part = "footer",
    value = as_paragraph(
      as_strike(test)
    )
  )

  runs <- information_data_chunk(ft)
  expect_equal(runs$strike, c(TRUE, TRUE, TRUE))

  openxml <- flextable:::runs_as_wml(ft, txt_data = runs)$run_openxml
  expect_match(openxml[1], "<w:strike w:val=\"true\"", fixed = TRUE)
  expect_match(openxml[2], "<w:strike w:val=\"true\"", fixed = TRUE)
  expect_match(openxml[3], "<w:strike w:val=\"true\"", fixed = TRUE)

  openxml <- flextable:::runs_as_pml(ft)$par_nodes_str
  expect_match(openxml[1], "strike=\"sngStrike\"", fixed = TRUE)
  expect_match(openxml[2], "strike=\"sngStrike\"", fixed = TRUE)
  expect_match(openxml[3], "strike=\"sngStrike\"", fixed = TRUE)

  html_info <- flextable:::runs_as_html(ft)
  css <- attr(html_info, "css")
  expect_match(css, "text-decoration:line-through;", fixed = TRUE)

  latex_str <- flextable:::runs_as_latex(ft)$txt
  expect_match(latex_str[1], "\\sout{", fixed = TRUE)
  expect_match(latex_str[2], "\\sout{", fixed = TRUE)
  expect_match(latex_str[3], "\\sout{", fixed = TRUE)

  rtf_str <- flextable:::runs_as_rtf(ft)$txt
  expect_match(rtf_str[1], "\\strike", fixed = TRUE)
  expect_match(rtf_str[2], "\\strike", fixed = TRUE)
  expect_match(rtf_str[3], "\\strike", fixed = TRUE)
})

skip_on_cran()

library(officer)
library(xml2)

# -- test data ---------------------------------------------------------------
dat <- data.frame(
  label = c("bold", "link", "code", "xref"),
  content = c(
    "This is **bold** text",
    "Visit [Quarto](https://quarto.org)",
    "Use `print()` here",
    "See @tbl-results"
  ),
  stringsAsFactors = FALSE
)

ft_qmd <- flextable(dat)
ft_qmd <- mk_par(ft_qmd, j = "content", value = as_paragraph(as_qmd(content)))

# -- as_qmd() object creation ------------------------------------------------
test_that("as_qmd creates a chunk with qmd_data", {
  chunk <- as_qmd("**bold**")
  expect_true(inherits(chunk, "data.frame"))
  expect_true("qmd_data" %in% colnames(chunk))
  expect_equal(chunk$qmd_data, "**bold**")
  expect_equal(chunk$txt, "**bold**")
})

test_that("as_qmd uses display as fallback text", {
  chunk <- as_qmd("@tbl-results", display = "Table 1")
  expect_equal(chunk$txt, "Table 1")
  expect_equal(chunk$qmd_data, "@tbl-results")
})

# -- DOCX markers (<!--TBLQMD:base64-->) ------------------------------------
test_that("DOCX output contains TBLQMD markers", {
  docx_file <- tempfile(fileext = ".docx")
  save_as_docx(ft_qmd, path = docx_file)
  doc <- read_docx(docx_file)
  body <- docx_body_xml(doc)

  # structure: header + data rows
  rows <- xml_find_all(body, "w:body/w:tbl/w:tr")
  expect_length(rows, nrow(dat) + 1)

  # label column: regular text
  labels <- xml_text(xml_find_all(
    body,
    "w:body/w:tbl/w:tr[position()>1]/w:tc[1]"
  ))
  expect_equal(labels, dat$label)

  # content column: TBLQMD markers in raw XML
  doc_str <- as.character(body)

  markers <- regmatches(
    doc_str,
    gregexpr("<!--TBLQMD:([A-Za-z0-9+/=]+)-->", doc_str)
  )[[1]]
  expect_length(markers, nrow(dat))

  # markers should NOT be wrapped in <w:r>...</w:r>
  expect_false(grepl("<w:r><!--TBLQMD:", doc_str))

  # decode each marker and verify content
  for (i in seq_along(markers)) {
    encoded <- sub("<!--TBLQMD:(.+)-->", "\\1", markers[i])
    expect_equal(officer::from_base64(encoded), dat$content[i])
  }
})

# -- HTML markers (<span data-qmd-base64="...">) ----------------------------
test_that("HTML output contains data-qmd-base64 spans", {
  html_str <- flextable:::gen_raw_html(ft_qmd)

  # each cell should produce a data-qmd-base64 span
  spans <- regmatches(
    html_str,
    gregexpr('data-qmd-base64="[^"]+"', html_str)
  )[[1]]
  expect_length(spans, nrow(dat))

  # extract and decode the first marker
  encoded <- sub('data-qmd-base64="([^"]+)"', "\\1", spans[1])
  decoded <- officer::from_base64(encoded)
  expect_equal(decoded, dat$content[1])
})

# -- LaTeX markers (\tblqmd{base64}) ----------------------------------------
test_that("LaTeX output contains tblqmd markers", {
  latex_str <- flextable:::knit_to_latex(
    ft_qmd,
    bookdown = FALSE,
    quarto = TRUE
  )

  # each cell should produce a \tblqmd{...} marker
  markers <- regmatches(
    latex_str,
    gregexpr("\\\\tblqmd\\{([A-Za-z0-9+/=]+)\\}", latex_str)
  )[[1]]
  expect_length(markers, nrow(dat))

  # decode the first marker
  encoded <- sub("\\\\tblqmd\\{(.+)\\}", "\\1", markers[1])
  decoded <- officer::from_base64(encoded)
  expect_equal(decoded, dat$content[1])
})

# -- use_flextable_qmd() installs the extension -----------------------------
test_that("use_flextable_qmd copies extension files", {
  tmp_project <- tempfile("qmd_project_")
  dir.create(tmp_project)

  path <- use_flextable_qmd(path = tmp_project, quiet = TRUE)

  ext_dir <- file.path(tmp_project, "_extensions", "flextable-qmd")
  expect_true(dir.exists(ext_dir))
  expect_true(file.exists(file.path(ext_dir, "_extension.yml")))
  expect_true(file.exists(file.path(ext_dir, "flextable-qmd.lua")))
  expect_true(file.exists(file.path(ext_dir, "unwrap-float.lua")))

  unlink(tmp_project, recursive = TRUE, force = TRUE)
})

# -- mixed content: as_qmd alongside regular chunks -------------------------
test_that("as_qmd works alongside other chunk types in DOCX", {
  mixed <- data.frame(
    a = c("row1", "row2"),
    b = c("plain text", "also plain"),
    stringsAsFactors = FALSE
  )
  ft_mix <- flextable(mixed)
  ft_mix <- mk_par(
    ft_mix,
    i = 1,
    j = "a",
    value = as_paragraph(
      as_chunk("prefix: "),
      as_qmd("**bold ref** to @tbl-x")
    )
  )

  docx_file <- tempfile(fileext = ".docx")
  save_as_docx(ft_mix, path = docx_file)
  doc <- read_docx(docx_file)
  doc_str <- as.character(docx_body_xml(doc))

  # regular chunk produces text run
  expect_true(grepl("prefix: ", doc_str))
  # as_qmd chunk produces TBLQMD marker
  expect_true(grepl("<!--TBLQMD:", doc_str))
  # only one marker (one as_qmd cell)
  marker_count <- lengths(regmatches(
    doc_str,
    gregexpr("<!--TBLQMD:", doc_str)
  ))
  expect_equal(marker_count, 1)
})

gen_raw_html <- function(
  x,
  class = "tabwid",
  caption = "",
  topcaption = TRUE,
  manual_css = ""
) {
  align <- x$properties$align
  # for ubiquity and other packages that dump old flextable
  if (is.null(align)) {
    align <- "center"
  }

  fixed_layout <- x$properties$layout %in% "fixed"
  if (!fixed_layout) {
    if (x$properties$width > 0) {
      # setting width will contraint columns'widths and make word breaks
      tbl_width <- paste0("width:", formatC(x$properties$width * 100), "%;")
    } else {
      tbl_width <- ""
    }
    tabcss <- paste0("table-layout:auto;", tbl_width)
  } else {
    tabcss <- ""
  }

  codes <- html_content_strs(x)

  classname <- UUIDgenerate(n = 1, use.time = TRUE)
  classname <- gsub("(^[[:alnum:]]+)(.*)$", "cl-\\1", classname)
  tabcss <- paste0(".", classname, "{", tabcss, "}")

  if (is.null(manual_css) || "" %in% manual_css) {
    manual_css_str <- ""
  } else {
    manual_css_str <- manual_css
  }

  if (!is.null(x$properties$opts_html$scroll)) {
    freeze_first_column <- x$properties$opts_html$scroll$freeze_first_column
    if (isTRUE(freeze_first_column)) {
      freeze_first_css_th <- paste0(
        ".",
        classname,
        " thead > tr > th:first-child {position:sticky;left:0;z-index:5;}"
      )
      freeze_first_css_td <- paste0(
        ".",
        classname,
        " tbody > tr > td:first-child {position:sticky;left:0;z-index:3;}"
      )
      freeze_first_css <- paste0(
        c(freeze_first_css_th, freeze_first_css_td),
        collapse = ""
      )

      x$properties$opts_html$extra_css <- paste0(
        freeze_first_css,
        x$properties$opts_html$extra_css
      )
    }

    if (!is.null(x$properties$opts_html$scroll$height)) {
      x$properties$opts_html$extra_css <- paste0(
        ".",
        classname,
        " th {position: sticky;top: 0;z-index: 4;}.",
        classname,
        " {border-collapse: separate !important;}",
        x$properties$opts_html$extra_css
      )
    }
  }

  if (!is.null(x$properties$opts_html$extra_class)) {
    classname <- c(
      classname,
      x$properties$opts_html$extra_class
    )
    classname <- paste0(classname, collapse = " ")
  }

  html <- paste0(
    "<style>",
    tabcss,
    codes$css,
    x$properties$opts_html$extra_css,
    manual_css_str,
    "</style>",
    sprintf(
      "<table data-quarto-disable-processing='true' class='%s'>",
      classname
    ),
    caption,
    codes$html,
    "</table>"
  )

  if (is.null(align)) {
    align <- "center"
  }
  if ("left" %in% align) {
    tab_class <- paste0(class, " tabwid_left")
  } else if ("right" %in% align) {
    tab_class <- paste0(class, " tabwid_right")
  } else {
    tab_class <- class
  }

  if (!topcaption) {
    tab_class <- paste0(tab_class, " tabwid-caption-bottom")
  }

  style_div <- ""
  if (!is.null(x$properties$opts_html$scroll)) {
    style_div <- do.call(scrollbox, x$properties$opts_html$scroll)
    style_div <- paste0(" style=\"", style_div, "\"")
  }

  html <- paste0(
    "<div class=\"",
    tab_class,
    "\"",
    style_div,
    ">",
    html,
    "</div>"
  )
  html
}


scrollbox <- function(height = NULL, add_css = "", ...) {
  str <- "overflow-x:auto;width:100%;"
  if (!is.null(height)) {
    if (is.numeric(height)) {
      height <- sprintf("%.0fpx", height)
    }
    str <- paste0(str, "overflow-y:auto;height:", height, ";")
  }
  str <- paste0(str, add_css)
  str
}

# to html/css  ----
#' @importFrom data.table setnames setorderv := setcolorder setDT setDF dcast
html_content_strs <- function(x) {
  cell_data_f <- information_data_cell(x)
  setDT(cell_data_f)
  par_data_f <- information_data_paragraph(x)
  setDT(par_data_f)

  txt_data <- runs_as_html(x, chunk_data = information_data_chunk(x))
  span_style_str <- attr(txt_data, "css")
  setDT(txt_data)

  cell_data <- merge(
    x = cell_data_f,
    y = par_data_f[, c(".part", ".row_id", ".col_id", "text.align")],
    by = c(".part", ".row_id", ".col_id")
  )
  par_data <- merge(
    x = par_data_f,
    y = cell_data_f[, c(
      ".part",
      ".row_id",
      ".col_id",
      "text.direction",
      "vertical.align"
    )],
    by = c(".part", ".row_id", ".col_id")
  )

  # get rid of originals that are empty
  cell_data[, c("width", "height", "hrule") := list(NULL, NULL, NULL)]
  cell_data <- merge(cell_data, fortify_width(x), by = ".col_id")
  cell_data <- merge(cell_data, fortify_height(x), by = c(".part", ".row_id"))
  cell_data <- merge(cell_data, fortify_hrule(x), by = c(".part", ".row_id"))

  span_data <- fortify_span(x)

  data_ref_pars <- distinct_paragraphs_properties(par_data)
  setDT(data_ref_pars)
  data_ref_cells <- distinct_cells_properties(cell_data)
  setDT(data_ref_cells)

  par_data <- merge(
    par_data,
    data_ref_pars,
    by = setdiff(colnames(data_ref_pars), "classname")
  )
  par_data <- par_data[,
    list(p_tag = paste0("<p class=\"", get("classname"), "\">")),
    by = c(".part", ".row_id", ".col_id")
  ]

  by_columns <- intersect(colnames(cell_data), colnames(data_ref_cells))
  cell_data <- merge(cell_data, data_ref_cells, by = by_columns)
  cell_data <- merge(
    cell_data,
    span_data,
    by = c(".part", ".row_id", ".col_id")
  )

  cell_data <- cell_data[,
    list(
      td_tag = paste0(
        ifelse(get(".part") %in% "header", "<th ", "<td "),
        paste0(
          ifelse(
            get("rowspan") > 1,
            paste0(" colspan=\"", get("rowspan"), "\""),
            ""
          ),
          ifelse(
            get("colspan") > 1,
            paste0(" rowspan=\"", get("colspan"), "\""),
            ""
          )
        ),
        "class=\"",
        get("classname"),
        "\">"
      )
    ),
    by = c(".part", ".row_id", ".col_id")
  ]

  dat <- merge(txt_data, par_data, by = c(".part", ".row_id", ".col_id"))
  dat$p_tag <- paste0(dat$p_tag, dat$span_tag, "</p>")
  dat <- merge(dat, cell_data, by = c(".part", ".row_id", ".col_id"))
  dat$td_tag <- paste0(
    dat$td_tag,
    dat$p_tag,
    ifelse(dat$.part %in% "header", "</th>", "</td>")
  )

  data_hrule <- fortify_hrule(x)
  data_hrule$tr_tag <- "<tr>"
  data_hrule$tr_tag[
    !data_hrule$hrule %in% "exact"
  ] <- "<tr style=\"overflow-wrap:break-word;\">"

  rows_data <- data_hrule[c(".part", ".row_id", "tr_tag")]

  dat <- merge(dat, span_data, by = c(".part", ".row_id", ".col_id"))
  dat$td_tag[dat$rowspan < 1 | dat$colspan < 1] <- ""

  z <- dcast(
    dat,
    .part + .row_id ~ .col_id,
    drop = TRUE,
    fill = "",
    value.var = "td_tag",
    fun.aggregate = I
  )
  z <- merge(z, rows_data, by = c(".part", ".row_id"))
  setorderv(z, c(".part", ".row_id"))
  z$tr_end <- "</tr>"

  parts <- z$.part
  header_start <- head(which(parts %in% "header"), n = 1)
  header_end <- tail(which(parts %in% "header"), n = 1)
  body_start <- head(which(parts %in% "body"), n = 1)
  body_end <- tail(which(parts %in% "body"), n = 1)
  footer_start <- head(which(parts %in% "footer"), n = 1)
  footer_end <- tail(which(parts %in% "footer"), n = 1)

  z$tpart_start <- ""
  z$tpart_end <- ""
  z$tpart_start[header_start] <- "<thead>"
  z$tpart_end[header_end] <- "</thead>"
  z$tpart_start[body_start] <- "<tbody>"
  z$tpart_end[body_end] <- "</tbody>"
  z$tpart_start[footer_start] <- "<tfoot>"
  z$tpart_end[footer_end] <- "</tfoot>"
  setcolorder(z, neworder = c("tpart_start", "tr_tag"))

  z[, c(".part", ".row_id") := NULL]

  html <- apply(as.matrix(z), 1, paste0, collapse = "")
  html <- paste0(html, collapse = "")

  par_style_str <- par_css_styles(data_ref_pars)
  cell_style_str <- cell_css_styles(
    data_ref_cells,
    add_widths = x$properties$layout %in% "fixed"
  )
  list(
    html = html,
    css = paste0(span_style_str, par_style_str, cell_style_str)
  )
}


# html chunks ----
htmlize <- function(x) {
  x <- htmlEscape(x)
  x <- gsub("\t", "&emsp;", x)
  x
}

#' @importFrom officer image_to_base64 plot_in_png
img_as_html <- function(img_data, width, height, alt = "") {
  img_data <- str_raster <- mapply(
    function(img_raster, width, height) {
      if (inherits(img_raster, "raster")) {
        img_raster <- plot_in_png(
          code = {
            op <- par(mar = rep(0, 4))
            plot(img_raster, interpolate = FALSE, asp = NA)
            par(op)
          },
          width = width,
          height = height,
          res = 300,
          units = "in",
          path = tempfile(fileext = ".png")
        )
      }
      img_raster
    },
    img_data,
    width,
    height,
    SIMPLIFY = TRUE,
    USE.NAMES = FALSE
  )
  base64_strings <- image_to_base64(img_data)
  alt <- ifelse(is.na(alt), "", alt)
  sprintf(
    "<img style=\"vertical-align:baseline;width:%.0fpx;height:%.0fpx;\" src=\"%s\" alt=\"%s\" />",
    width * 72,
    height * 72,
    base64_strings,
    htmlEscape(alt)
  )
}

# css ----

css_pt <- function(x, digits = 1) {
  x <- ifelse(
    is.na(x),
    "inherit",
    ifelse(x < 0.001, "0", paste0(format_double(x, digits = digits), "pt"))
  )
  x
}
css_no_unit <- function(x, digits = 0) {
  x <- ifelse(
    is.na(x),
    "inherit",
    ifelse(x < 0.001, "0", format_double(x, digits = digits))
  )
  x
}

border_css <- function(color, width, style, side) {
  style[!style %in% c("dotted", "dashed", "solid")] <- "solid"
  x <- sprintf(
    "border-%s: %s %s %s;",
    side,
    css_pt(width, 2),
    style,
    colcodecss(color)
  )
  x
}


css_align <- function(text.direction, align) {
  textdir <- rep("", length(text.direction))
  textdir[
    text.direction %in% "btlr"
  ] <- "writing-mode: vertical-rl;transform: rotate(180deg);-ms-writing-mode:bt-lr;-webkit-writing-mode:vertical-rl;"
  textdir[
    text.direction %in% "tbrl"
  ] <- "writing-mode: vertical-rl;-ms-writing-mode:tb-rl;-webkit-writing-mode:vertical-rl;"

  textalign <- sprintf("text-align:%s;", align)

  textalign_margins <- rep("", length(text.direction))
  textalign_margins[
    text.direction %in% "tbrl" & align %in% "center"
  ] <- "margin-left:auto;margin-right:auto;"
  textalign_margins[
    text.direction %in% "tbrl" & align %in% "left"
  ] <- "margin-right:auto;"
  textalign_margins[
    text.direction %in% "tbrl" & align %in% "right"
  ] <- "margin-left:auto;"

  textalign_margins[
    text.direction %in% "btlr" & align %in% "center"
  ] <- "margin-left:auto;margin-right:auto;"
  textalign_margins[
    text.direction %in% "btlr" & align %in% "left"
  ] <- "margin-right:auto;"
  textalign_margins[
    text.direction %in% "btlr" & align %in% "right"
  ] <- "margin-left:auto;"

  paste0(textalign, textalign_margins, textdir)
}

par_css_styles <- function(x) {
  shading <- rep("background-color:transparent;", nrow(x))
  has_shading <- colalpha(x$shading.color) > 0
  shading[has_shading] <- sprintf(
    "background-color:%s;",
    colcodecss(x$shading.color[has_shading])
  )

  textalign <- css_align(x$text.direction, x$text.align)

  bb <- border_css(
    color = x$border.color.bottom,
    width = x$border.width.bottom,
    style = x$border.style.bottom,
    side = "bottom"
  )
  bt <- border_css(
    color = x$border.color.top,
    width = x$border.width.top,
    style = x$border.style.top,
    side = "top"
  )
  bl <- border_css(
    color = x$border.color.left,
    width = x$border.width.left,
    style = x$border.style.left,
    side = "left"
  )
  br <- border_css(
    color = x$border.color.right,
    width = x$border.width.right,
    style = x$border.style.right,
    side = "right"
  )

  padding.bottom <- sprintf("padding-bottom:%s;", css_pt(x$padding.bottom))
  padding.top <- sprintf("padding-top:%s;", css_pt(x$padding.top))
  padding.left <- sprintf("padding-left:%s;", css_pt(x$padding.left))
  padding.right <- sprintf("padding-right:%s;", css_pt(x$padding.right))

  line_spacing <- sprintf("line-height: %s;", css_no_unit(x$line_spacing, 2))

  # first-line/hanging indents: CSS `text-indent` only moves the first
  # line, so hanging is a negative value relative to `padding-left`
  # (hanging wins over first_line, as in officer's `fp_par()`).
  text_indent <- rep("", nrow(x))
  has_first <- !is.na(x$first_line) & x$first_line > 0
  text_indent[has_first] <- sprintf(
    "text-indent:%s;",
    css_pt(x$first_line[has_first])
  )
  has_hanging <- !is.na(x$hanging) & x$hanging > 0
  text_indent[has_hanging] <- sprintf(
    "text-indent:-%s;",
    css_pt(x$hanging[has_hanging])
  )

  style_column <- paste0(
    "margin:0;",
    textalign,
    bb,
    bt,
    bl,
    br,
    padding.bottom,
    padding.top,
    padding.left,
    padding.right,
    line_spacing,
    text_indent,
    shading
  )
  paste0(".", x$classname, "{", style_column, "}", collapse = "")
}

cell_css_styles <- function(x, add_widths = TRUE) {
  background.color <- ifelse(
    colalpha(x$background.color) > 0,
    sprintf("background-color:%s;", colcodecss(x$background.color)),
    "background-color:transparent;"
  )
  width <- rep("", nrow(x))
  if (add_widths) {
    width[!is.na(x$width)] <- sprintf(
      "width:%sin;",
      css_no_unit(x$width[!is.na(x$width)], digits = 3)
    )
  }

  height <- rep("", nrow(x))
  hsel <- !is.na(x$height) & x$hrule %in% c("exact", "atleast")
  height[hsel] <- sprintf(
    "height:%sin;",
    css_no_unit(x$height[hsel], digits = 3)
  )

  vertical.align <- rep("vertical-align: middle;", nrow(x))
  vertical.align[x$vertical.align %in% "top"] <- "vertical-align: top;"
  vertical.align[x$vertical.align %in% "bottom"] <- "vertical-align: bottom;"

  bb <- border_css(
    color = x$border.color.bottom,
    width = x$border.width.bottom,
    style = x$border.style.bottom,
    side = "bottom"
  )
  bt <- border_css(
    color = x$border.color.top,
    width = x$border.width.top,
    style = x$border.style.top,
    side = "top"
  )
  bl <- border_css(
    color = x$border.color.left,
    width = x$border.width.left,
    style = x$border.style.left,
    side = "left"
  )
  br <- border_css(
    color = x$border.color.right,
    width = x$border.width.right,
    style = x$border.style.right,
    side = "right"
  )

  margin.bottom <- sprintf("margin-bottom:%s;", css_pt(x$margin.bottom))
  margin.top <- sprintf("margin-top:%s;", css_pt(x$margin.top))
  margin.left <- sprintf("margin-left:%s;", css_pt(x$margin.left))
  margin.right <- sprintf("margin-right:%s;", css_pt(x$margin.right))

  style_column <- paste0(
    width,
    height,
    background.color,
    vertical.align,
    bb,
    bt,
    bl,
    br,
    margin.bottom,
    margin.top,
    margin.left,
    margin.right
  )
  paste0(".", x$classname, "{", style_column, "}", collapse = "")
}

# htmlDependency ----
#' @importFrom htmltools htmlDependency
#' @export
#' @title htmlDependency for flextable objects
#' @description When using loops in an R Markdown for HTML document, the
#' htmlDependency object for flextable must also be added at least once.
#' @examples
#' if (require("htmltools")) {
#'   div(flextable_html_dependency())
#' }
#' @keywords internal
flextable_html_dependency <- function() {
  htmlDependency(
    "tabwid",
    "1.1.3",
    src = system.file(package = "flextable", "web_1.1.3"),
    stylesheet = "tabwid.css",
    script = "tabwid.js"
  )
}

# URL of the KaTeX stylesheet that `equatags`/`katex` emits when rendering
# equations to HTML. Extracted from the actual output so the version always
# matches the markup (no hard-coded version, no version skew). The extraction
# spins up a V8 rendering, so the result is cached for the session in
# `flextable_global`; it only depends on the installed katex version, which
# cannot change in a loaded session. Failures (no equatags, no match) are not
# cached so a later call can recover.
katex_css_href <- function() {
  if (!is.null(flextable_global$katex_css_href)) {
    return(flextable_global$katex_css_href)
  }
  if (!requireNamespace("equatags", quietly = TRUE)) {
    return(NULL)
  }
  transform_mathjax <- getFromNamespace("transform_mathjax", "equatags")
  html <- transform_mathjax("1", to = "html")
  href <- regmatches(
    html,
    regexpr("https://[^\"']+katex[^\"']+\\.css", html)
  )
  if (length(href) < 1) {
    return(NULL)
  }
  flextable_global$katex_css_href <- href[1]
  href[1]
}

# `data-external="1"` keeps pandoc --self-contained/--embed-resources from
# downloading and inlining the stylesheet: inlining would break its relative
# `url(fonts/...)` references and require internet at render time.
katex_css_link <- function() {
  href <- katex_css_href()
  if (is.null(href)) {
    return(NULL)
  }
  sprintf(
    "<link rel=\"stylesheet\" href=\"%s\" data-external=\"1\">",
    href
  )
}

# flextable renders HTML tables inside a shadow DOM (see web_1.1.3/tabwid.js),
# which moves the table and the KaTeX stylesheet injected in cells into a
# shadow root. `@font-face` rules declared inside a shadow root are not
# registered by the browser, so the KaTeX math fonts (KaTeX_SansSerif,
# KaTeX_Typewriter, KaTeX_Fraktur, KaTeX_AMS) never load and the corresponding
# variants (`\mathsf`, `\mathtt`, `\mathfrak`, `\mathbb`) silently fall back to
# the default font (issue #622). Adding the same stylesheet as a document-level
# dependency registers the fonts in the light DOM, where they become usable by
# the shadow DOM content too.
katex_html_dependency <- function() {
  link <- katex_css_link()
  if (is.null(link)) {
    return(NULL)
  }
  version <- sub(".*katex@([0-9.]+).*", "\\1", link)
  if (!grepl("^[0-9.]+$", version)) {
    version <- "0.0.0"
  }
  # The stylesheet is served from a CDN; remote `href` dependencies are not
  # supported by the knitr meta path ("not disk-based"), so emit the link tag
  # via `head` while pointing `src` at an existing on-disk directory and
  # copying nothing (`all_files = FALSE`).
  htmlDependency(
    name = "katex",
    version = version,
    src = system.file(package = "flextable", "web_1.1.3"),
    all_files = FALSE,
    head = link
  )
}

html_dependencies_list <- function(x) {
  list_deps <- list(flextable_html_dependency())
  list_deps <- append(
    list_deps,
    lapply(avail_gfonts(x), gdtools::gfontHtmlDependency)
  )
  if (has_equation(x)) {
    katex_dep <- katex_html_dependency()
    if (!is.null(katex_dep)) {
      list_deps <- append(list_deps, list(katex_dep))
    }
  }
  list_deps
}

#' @importFrom gdtools installed_gfonts
avail_gfonts <- function(x) {
  z <- character()
  if (nrow_part(x, part = "body") > 0) {
    z <- append(z, unique(as.vector(x$body$styles$text$font.family$data)))
  }
  if (nrow_part(x, part = "footer") > 0) {
    z <- append(z, unique(as.vector(x$footer$styles$text$font.family$data)))
  }
  if (nrow_part(x, part = "header") > 0) {
    z <- append(z, unique(as.vector(x$header$styles$text$font.family$data)))
  }
  families <- unique(as.character(z))
  intersect(families, installed_gfonts())
}

#' @importFrom rmarkdown html_document render pandoc_available
render_htmltag <- function(x, path, title, lang = "en") {
  sucess <- FALSE
  rmd_file <- tmp_rmd(title = paste0(title, collapse = "\n"), lang = lang)
  html_file <- gsub("\\.Rmd$", ".html", rmd_file)
  tryCatch(
    {
      render(
        rmd_file,
        output_format = html_document(
          self_contained = TRUE,
          theme = NULL,
          mathjax = NULL
        ),
        # output_file = basename(path),
        envir = new.env(),
        quiet = TRUE,
        params = list(x = x)
      )
      sucess <- file.copy(html_file, path, overwrite = TRUE)
    },
    warning = function(e) {},
    error = function(e) {}
  )
  sucess
}

tmp_rmd <- function(title, lang = "en") {
  stopifnot(pandoc_available())
  file <- tempfile(fileext = ".Rmd")
  writeLines(
    c(
      "---",
      sprintf("title: \"%s\"", title),
      sprintf("lang: %s", lang),
      "params:",
      "  x: ''",
      "---",
      "```{r include=FALSE}",
      "library(knitr)",
      "opts_chunk$set(echo = FALSE)",
      "```",
      "",
      "```{r}",
      "params$x",
      "```"
    ),
    file,
    useBytes = TRUE
  )
  file
}

#' @title Method to transform objects into flextables
#' @description This is a convenient function
#' to let users create flextable bindings
#' from any objects. Users should consult documentation
#' of corresponding method to understand the details and
#' see what arguments can be used.
#' @param x object to be transformed as flextable
#' @param ... arguments for custom methods
#' @export
#' @family flextable_constructors
as_flextable <- function(x, ...) {
  UseMethod("as_flextable")
}


#' @title Insert group-label rows into a data frame
#'
#' @description Repeated consecutive values of group columns will
#' be used to define the title of the groups and will
#' be added as a row title.
#'
#' @param x dataset
#' @param groups columns names to be used as row separators.
#' @param columns columns names to keep
#' @param expand_single if FALSE, groups with only one
#' row will not be expanded with a title row. If TRUE (the
#' default), single row groups and multi-row groups are all
#' restructured.
#' @examples
#' # as_grouped_data -----
#' library(data.table)
#' CO2 <- CO2
#' setDT(CO2)
#' CO2$conc <- as.integer(CO2$conc)
#'
#' data_co2 <- dcast(CO2, Treatment + conc ~ Type,
#'   value.var = "uptake", fun.aggregate = mean
#' )
#' data_co2
#' data_co2 <- as_grouped_data(x = data_co2, groups = c("Treatment"))
#' data_co2
#' @seealso [as_flextable.grouped_data()]
#' @export
as_grouped_data <- function(x, groups, columns = NULL, expand_single = TRUE) {
  if (
    inherits(x, "data.table") ||
      inherits(x, "tbl_df") ||
      inherits(x, "tbl") ||
      is.matrix(x)
  ) {
    x <- as.data.frame(x, stringsAsFactors = FALSE)
  }

  stopifnot(is.data.frame(x), ncol(x) > 0)

  if (is.null(columns)) {
    columns <- setdiff(names(x), groups)
  }

  z <- x[, c(groups, columns), drop = FALSE]
  setDT(z)

  z[, c("rleid") := list(do.call(rleid, as.list(.SD))), .SDcols = groups]
  z <- merge(
    z,
    z[, list(rlen = .N), by = "rleid"],
    by = "rleid"
  )

  subsets <- list()
  for (grp_i in seq_along(groups)) {
    grp_comb <- groups[seq_len(grp_i)]
    if (!expand_single) {
      subdat <- unique(
        z[z$rlen > 1, .SD, .SDcols = c(grp_comb, "rleid")],
        by = "rleid"
      )
    } else {
      subdat <- unique(z[, .SD, .SDcols = c(grp_comb, "rleid")], by = "rleid")
    }
    subdat[, c("rleid") := list(.SD$rleid - 1 + grp_i * .1)]
    void_cols <- setdiff(colnames(subdat), c(groups[grp_i], "rleid"))
    if (length(void_cols)) {
      subdat[,
        c(void_cols) := lapply(.SD, function(w) {
          w[] <- NA
          w
        }),
        .SDcols = void_cols
      ]
    }
    subsets[[length(subsets) + 1]] <- subdat
  }

  if (!expand_single) {
    z[
      z$rlen > 1,
      c(groups) := lapply(.SD, function(w) {
        w[] <- NA
        w
      }),
      .SDcols = groups
    ]
  } else {
    z[,
      c(groups) := lapply(.SD, function(w) {
        w[] <- NA
        w
      }),
      .SDcols = groups
    ]
  }
  z$rlen <- NULL

  subsets[[length(subsets) + 1]] <- z

  x <- rbindlist(subsets, use.names = TRUE, fill = TRUE)
  setorderv(x, cols = "rleid")
  x$rleid <- NULL
  setDF(x)
  class(x) <- c("grouped_data", class(x))
  attr(x, "groups") <- groups
  attr(x, "columns") <- columns
  x
}

#' @export
#' @title Transform a 'grouped_data' object into a flextable
#' @description Produce a flextable from a table
#' produced by function [as_grouped_data()].
#' @param x 'grouped_data' object to be transformed into a "flextable"
#' @param col_keys columns names/keys to display. If some column names are not in
#' the dataset, they will be added as blank columns by default.
#' @param hide_grouplabel if TRUE, group label will not be rendered, only
#' level/value will be rendered.
#' @param ... unused argument
#' @examples
#' library(data.table)
#' CO2 <- CO2
#' setDT(CO2)
#' CO2$conc <- as.integer(CO2$conc)
#'
#' data_co2 <- dcast(CO2, Treatment + conc ~ Type,
#'   value.var = "uptake", fun.aggregate = mean
#' )
#' data_co2 <- as_grouped_data(x = data_co2, groups = c("Treatment"))
#'
#' ft <- as_flextable(data_co2)
#' ft <- add_footer_lines(ft, "dataset CO2 has been used for this flextable")
#' ft <- add_header_lines(ft, "mean of carbon dioxide uptake in grass plants")
#' ft <- set_header_labels(ft, conc = "Concentration")
#' ft <- autofit(ft)
#' ft <- width(ft, width = c(1, 1, 1))
#' ft
#' @family as_flextable methods
#' @seealso [as_grouped_data()]
as_flextable.grouped_data <- function(
  x,
  col_keys = NULL,
  hide_grouplabel = FALSE,
  ...
) {
  if (is.null(col_keys)) {
    col_keys <- attr(x, "columns")
  }
  groups <- attr(x, "groups")
  if (hide_grouplabel) {
    col_keys <- setdiff(col_keys, groups)
  }
  z <- flextable(x, col_keys = col_keys)

  j2 <- length(col_keys)
  for (grp_name in groups) {
    i <- !is.na(x[[grp_name]])
    gnames <- x[[grp_name]][i]
    if (!hide_grouplabel) {
      z <- compose(
        z,
        i = i,
        j = 1,
        value = as_paragraph(as_chunk(grp_name), ": ", as_chunk(gnames))
      )
    } else {
      z <- compose(z, i = i, j = 1, value = as_paragraph(as_chunk(gnames)))
    }

    z <- merge_h_range(z, i = i, j1 = 1, j2 = j2)
    z <- align(z, i = i, align = "left")
  }

  z
}


pvalue_format <- function(x) {
  z <- cut(
    x,
    breaks = c(-Inf, 0.001, 0.01, 0.05, 0.1, Inf),
    labels = c("***", " **", "  *", "  .", "   ")
  )
  z <- as.character(z)
  z[is.na(x)] <- ""
  z
}

#' @export
#' @importFrom stats naprint quantile
#' @importFrom utils tail
#' @title Transform a 'glm' object into a flextable
#' @description produce a flextable describing a
#' generalized linear model produced by function `glm`.
#'
#' You can remove significance stars by setting options
#' `options(show.signif.stars = FALSE)`.
#' @param x glm model
#' @param ... unused argument
#' @examples
#' if (require("broom")) {
#'   dat <- attitude
#'   dat$high.rating <- (dat$rating > 70)
#'   probit.model <- glm(high.rating ~ learning + critical +
#'     advance, data = dat, family = binomial(link = "probit"))
#'   ft <- as_flextable(probit.model)
#'   ft
#' }
#' @family as_flextable methods
as_flextable.glm <- function(x, ...) {
  if (!requireNamespace("broom", quietly = TRUE)) {
    stop(sprintf(
      "'%s' package should be installed to create a flextable from an object of type '%s'.",
      "broom",
      "glm"
    ))
  }

  data_t <- broom::tidy(x)
  sum_obj <- summary(x)
  show_signif <- !is.null(getOption("show.signif.stars")) &&
    getOption("show.signif.stars")

  ft <- flextable(
    data_t,
    col_keys = c(
      "term",
      "estimate",
      "std.error",
      "statistic",
      "p.value",
      if (show_signif) "signif"
    )
  )
  ft <- colformat_double(
    ft,
    j = c(
      "estimate",
      "std.error",
      "statistic"
    ),
    digits = 3
  )
  ft <- colformat_double(ft, j = c("p.value"), digits = 4) # nolint
  if (show_signif) {
    ft <- mk_par(ft, j = "signif", value = as_paragraph(pvalue_format(p.value)))
  }

  ft <- set_header_labels(
    ft,
    term = "",
    estimate = "Estimate",
    std.error = "Standard Error",
    statistic = "z value",
    p.value = "Pr(>|z|)"
  )

  digits <- max(3L, getOption("digits") - 3L)

  ft <- add_footer_lines(
    ft,
    values = c(
      if (show_signif) {
        "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05"
      },
      if (show_signif) " ",
      paste(
        "(Dispersion parameter for ",
        x$family$family,
        " family taken to be ",
        format(sum_obj$dispersion),
        ")",
        sep = ""
      ),
      sprintf(
        "Null deviance: %s on %s degrees of freedom",
        formatC(sum_obj$null.deviance),
        formatC(sum_obj$df.null)
      ),
      sprintf(
        "Residual deviance: %s on %s degrees of freedom",
        formatC(sum_obj$deviance),
        formatC(sum_obj$df.residual)
      ),
      {
        mess <- naprint(sum_obj$na.action)
        if (nzchar(mess)) {
          paste("  (", mess, ")\n", sep = "")
        } else {
          character(0)
        }
      }
    )
  )
  if (show_signif) {
    ft <- align(ft, i = 1, align = "right", part = "footer")
    ft <- italic(ft, i = 1, italic = TRUE, part = "footer")
  }
  ft <- hrule(ft, rule = "auto")
  ft <- autofit(ft, part = c("header", "body"))
  if (show_signif) {
    ft <- width(ft, j = "signif", width = .4)
  }

  ft
}


#' @export
#' @title Transform a 'lm' object into a flextable
#' @description produce a flextable describing a
#' linear model produced by function `lm`.
#'
#' You can remove significance stars by setting options
#' `options(show.signif.stars = FALSE)`.
#' @param x lm model
#' @param ... unused argument
#' @examples
#' if (require("broom")) {
#'   lmod <- lm(rating ~ complaints + privileges +
#'     learning + raises + critical, data = attitude)
#'   ft <- as_flextable(lmod)
#'   ft
#' }
#' @family as_flextable methods
as_flextable.lm <- function(x, ...) {
  if (!requireNamespace("broom", quietly = TRUE)) {
    stop(sprintf(
      "'%s' package should be installed to create a flextable from an object of type '%s'.",
      "broom",
      "lm"
    ))
  }

  data_t <- broom::tidy(x)
  data_g <- broom::glance(x)

  show_signif <- !is.null(getOption("show.signif.stars")) &&
    getOption("show.signif.stars")
  col_keys <- c(
    "term",
    "estimate",
    "std.error",
    "statistic",
    "p.value",
    if (show_signif) "signif"
  )

  ft <- flextable(data_t, col_keys = col_keys)
  ft <- colformat_double(
    ft,
    j = c("estimate", "std.error", "statistic"),
    digits = 3
  )
  ft <- colformat_double(ft, j = c("p.value"), digits = 4)

  if (show_signif) {
    ft <- mk_par(ft, j = "signif", value = as_paragraph(pvalue_format(p.value)))
  }

  ft <- set_header_labels(
    ft,
    term = "",
    estimate = "Estimate",
    std.error = "Standard Error",
    statistic = "t value",
    p.value = "Pr(>|t|)",
    signif = ""
  )
  dimpretty <- dim_pretty(ft, part = "all")

  ft <- add_footer_lines(
    ft,
    values = c(
      if (show_signif) {
        "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05"
      },
      if (show_signif) "",
      sprintf(
        "Residual standard error: %s on %.0f degrees of freedom",
        formatC(data_g$sigma),
        data_g$df.residual
      ),
      sprintf(
        "Multiple R-squared: %s, Adjusted R-squared: %s",
        formatC(data_g$r.squared),
        formatC(data_g$adj.r.squared)
      ),
      sprintf(
        "F-statistic: %s on %.0f and %.0f DF, p-value: %.4f",
        formatC(data_g$statistic),
        data_g$df.residual,
        data_g$df,
        data_g$p.value
      )
    )
  )

  if (show_signif) {
    ft <- align(ft, i = 1, align = "right", part = "footer")
    ft <- italic(ft, i = 1, italic = TRUE, part = "footer")
  }
  ft <- hrule(ft, rule = "auto")
  ft <- autofit(ft, part = c("header", "body"))
  if (show_signif) {
    ft <- width(ft, j = "signif", width = .4)
  }
  ft
}


#' @export
#' @title Transform a 'htest' object into a flextable
#' @description produce a flextable describing an
#' object oof class `htest`.
#' @param x htest object
#' @param ... unused argument
#' @examples
#' if (require("stats")) {
#'   M <- as.table(rbind(c(762, 327, 468), c(484, 239, 477)))
#'   dimnames(M) <- list(
#'     gender = c("F", "M"),
#'     party = c("Democrat", "Independent", "Republican")
#'   )
#'   ft_1 <- as_flextable(chisq.test(M))
#'   ft_1
#' }
#' @family as_flextable methods
as_flextable.htest <- function(x, ...) {
  ret <- x[c("estimate", "statistic", "p.value", "parameter")]
  if (length(ret$estimate) > 1) {
    names(ret$estimate) <- paste0("estimate", seq_along(ret$estimate))
    ret <- c(ret$estimate, ret)
    ret$estimate <- NULL
    if (x$method == "Welch Two Sample t-test") {
      ret <- c(
        estimate = ret$estimate1 - ret$estimate2,
        ret
      )
    }
  }
  if (length(x$parameter) > 1) {
    ret$parameter <- NULL
    if (is.null(names(x$parameter))) {
      warning("Multiple unnamed parameters in hypothesis test; dropping them")
    } else {
      message(
        "Multiple parameters; naming those columns ",
        paste(make.names(names(x$parameter)), collapse = ", ")
      )
      ret <- append(ret, x$parameter, after = 1)
    }
  }
  ret <- Filter(Negate(is.null), ret)
  if (!is.null(x$conf.int)) {
    ret <- c(ret, conf.low = x$conf.int[1], conf.high = x$conf.int[2])
  }
  if (!is.null(x$method)) {
    ret <- c(ret, method = as.character(x$method))
  }
  if (!is.null(x$alternative)) {
    ret <- c(ret, alternative = as.character(x$alternative))
  }
  dat <- as.data.frame(ret, stringsAsFactors = FALSE)
  z <- flextable(dat)
  z <- colformat_double(z)

  show_signif <- !is.null(getOption("show.signif.stars")) &&
    getOption("show.signif.stars")

  if ("p.value" %in% colnames(dat)) {
    z <- colformat_double(z, j = "p.value", digits = 4)

    if (show_signif) {
      z <- append_chunks(
        x = z,
        j = "p.value",
        part = "body",
        dumb = as_chunk(p.value, formatter = pvalue_format)
      )
      z <- add_footer_lines(
        z,
        values = c(
          "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05"
        )
      )
    }
  }
  z <- autofit(z)
  z
}


#' @export
#' @title Summarize continuous variables as a flextable
#' @description create a data.frame summary for continuous variables
#' @param dat a data.frame
#' @param columns continuous variables to be summarized. If NULL all
#' continuous variables are summarized.
#' @param by discrete variables to use as groups when summarizing.
#' @param hide_grouplabel if TRUE, group label will not be rendered, only
#' level/value will be rendered.
#' @param digits the desired number of digits after the decimal point
#' @examples
#' ft_1 <- continuous_summary(iris, names(iris)[1:4],
#'   by = "Species",
#'   hide_grouplabel = FALSE
#' )
#' ft_1
continuous_summary <- function(
  dat,
  columns = NULL,
  by = character(0),
  hide_grouplabel = TRUE,
  digits = 3
) {
  if (!is.data.table(dat)) {
    x <- as.data.table(dat)
  }
  if (is.null(columns)) {
    columns <- colnames(dat)[sapply(dat, function(z) {
      is.double(z) || is.integer(z)
    })]
  }

  fun_list <- c(
    "N",
    "MIN",
    "Q1",
    "MEDIAN",
    "Q3",
    "MAX",
    "MEAN",
    "SD",
    "MAD",
    "NAS"
  )
  agg <- x[,
    c(
      lapply(.SD, N),
      lapply(.SD, MIN),
      lapply(.SD, Q1),
      lapply(.SD, MEDIAN),
      lapply(.SD, Q3),
      lapply(.SD, MAX),
      lapply(.SD, MEAN),
      lapply(.SD, SD),
      lapply(.SD, MAD),
      lapply(.SD, NAS)
    ),
    .SDcols = columns,
    by = by
  ]

  gen_cn <- lapply(fun_list, function(fun, col) paste0(col, "_", fun), columns)
  colnames(agg) <- c(by, unlist(gen_cn))

  agg <- melt(agg, measure = c(gen_cn), value.name = fun_list)
  levels(x = agg[["variable"]]) <- columns
  z <- as_grouped_data(
    agg,
    groups = "variable",
    columns = setdiff(names(agg), "variable")
  )
  is_label <- !is.na(z$variable)
  ft <- as_flextable(z, hide_grouplabel = hide_grouplabel)

  ft <- colformat_int(ft, j = c("N", "NAS"))
  ft <- colformat_double(
    ft,
    j = setdiff(fun_list, c("N", "NAS")),
    digits = digits
  )
  ft <- set_header_labels(
    ft,
    values = c(
      "MIN" = "min.",
      "MAX" = "max.",
      "Q1" = "q1",
      "Q3" = "q3",
      "MEDIAN" = "median",
      "MEAN" = "mean",
      "SD" = "sd",
      "MAD" = "mad",
      "NAS" = "# na"
    )
  )
  ft <- hline(ft, i = is_label, border = officer::fp_border(width = .5))
  ft <- italic(ft, italic = TRUE, i = is_label)
  ft <- merge_v(ft, j = by)
  ft <- valign(ft, j = by, valign = "top")
  ft <- vline(
    ft,
    j = length(by),
    border = officer::fp_border(width = .5),
    part = "body"
  )
  ft <- vline(
    ft,
    j = length(by),
    border = officer::fp_border(width = .5),
    part = "header"
  )
  fix_border_issues(ft)
}


#' @export
#' @title Transform a 'merMod' or 'lme' object into a flextable
#' @description produce a flextable describing a
#' mixed model. The function is only using package 'broom.mixed'
#' that provides the data presented in the resulting flextable.
#'
#' You can remove significance stars by setting options
#' `options(show.signif.stars = FALSE)`.
#' @param x a mixed model
#' @param add.random TRUE or FALSE, if TRUE random effects are
#' added to the table.
#' @param ... unused argument
#' @examples
#' if (require("broom.mixed") && require("nlme")) {
#'   m1 <- lme(distance ~ age, data = Orthodont)
#'   ft <- as_flextable(m1)
#'   ft
#' }
#' @family as_flextable methods
as_flextable.merMod <- function(x, add.random = TRUE, ...) {
  if (!requireNamespace("broom.mixed", quietly = TRUE)) {
    stop(sprintf(
      "'%s' package should be installed to create a flextable from an object of type '%s'.",
      "broom.mixed",
      "mixed model"
    ))
  }

  data_t <- broom::tidy(x)
  data_t$effect[data_t$effect %in% "fixed"] <- "Fixed effects"
  data_t$effect[
    data_t$effect %in% c("ran_pars", "ran_vals", "ran_coefs")
  ] <- "Random effects"

  data_g <- broom::glance(x)

  has_pvalue <- if ("p.value" %in% colnames(data_t)) TRUE else FALSE
  show_signif <- !is.null(getOption("show.signif.stars")) &&
    getOption("show.signif.stars")

  col_keys <- c(
    "effect",
    "group",
    "term",
    "estimate",
    "std.error",
    "df",
    "statistic",
    if (has_pvalue) c("p.value"),
    if (has_pvalue && show_signif) "signif"
  )

  if (add.random) {
    data_t <- as_grouped_data(x = data_t, groups = "effect", )
    ft <- as_flextable(data_t, col_keys = col_keys, hide_grouplabel = TRUE)
  } else {
    data_t <- data_t[data_t$effect %in% "Fixed effects", ]
    ft <- flextable(data_t, col_keys = setdiff(col_keys, c("effect", "group")))
  }

  ft <- colformat_double(
    ft,
    j = c("estimate", "std.error", "statistic"),
    digits = 3
  )
  ft <- colformat_double(ft, j = c("df"), digits = 0)
  if (has_pvalue) {
    ft <- colformat_double(ft, j = "p.value", digits = 4)
  }
  ft <- set_header_labels(
    ft,
    term = "",
    estimate = "Estimate",
    std.error = "Standard Error",
    p.value = "p-value"
  )
  ft <- autofit(ft, part = c("header", "body"))

  if (has_pvalue && show_signif) {
    ft <- mk_par(ft, j = "signif", value = as_paragraph(pvalue_format(p.value)))
    ft <- width(ft, j = "signif", width = .4)
  }
  ft <- align(ft, i = ~ !is.na(effect), align = "center")

  if (has_pvalue && show_signif) {
    ft <- add_footer_lines(
      ft,
      values = c(
        "Signif. codes: 0 <= '***' < 0.001 < '**' < 0.01 < '*' < 0.05",
        ""
      )
    )
  }

  mod_stats <- c("sigma", "logLik", "AIC", "BIC")
  mod_gl <- data.frame(
    stat = mod_stats,
    value = as.double(unlist(data_g[mod_stats])),
    labels = c(
      "square root of the estimated residual variance",
      "data's log-likelihood under the model",
      "Akaike Information Criterion",
      "Bayesian Information Criterion"
    )
  )
  mod_qual <- paste0(
    c(
      "square root of the estimated residual variance",
      "data's log-likelihood under the model",
      "Akaike Information Criterion",
      "Bayesian Information Criterion"
    ),
    ": ",
    format_fun(unlist(data_g[mod_stats]))
  )
  ft <- add_footer_lines(ft, values = mod_qual)
  ft <- align(ft, align = "left", part = "footer")
  if (has_pvalue && show_signif) {
    ft <- align(ft, i = 1, align = "right", part = "footer")
  }
  ft <- hrule(ft, rule = "auto")
  ft
}

#' @export
#' @rdname as_flextable.merMod
as_flextable.lme <- as_flextable.merMod

#' @export
#' @rdname as_flextable.merMod
as_flextable.gls <- as_flextable.merMod

#' @export
#' @rdname as_flextable.merMod
as_flextable.nlme <- as_flextable.merMod

#' @export
#' @rdname as_flextable.merMod
as_flextable.brmsfit <- as_flextable.merMod

#' @export
#' @rdname as_flextable.merMod
as_flextable.glmmTMB <- as_flextable.merMod

#' @export
#' @rdname as_flextable.merMod
as_flextable.glmmadmb <- as_flextable.merMod

#' @export
#' @title Transform a 'kmeans' object into a flextable
#' @description produce a flextable describing a
#' kmeans object. The function is only using package 'broom'
#' that provides the data presented in the resulting flextable.
#' @param x a [kmeans()] object
#' @param digits number of digits for the numeric columns
#' @param ... unused argument
#' @examples
#' if (require("stats")) {
#'   cl <- kmeans(scale(mtcars[1:7]), 5)
#'   ft <- as_flextable(cl)
#'   ft
#' }
#' @importFrom rlang sym
#' @family as_flextable methods
as_flextable.kmeans <- function(x, digits = 4, ...) {
  if (!requireNamespace("broom", quietly = TRUE)) {
    stop(sprintf(
      "'%s' package should be installed to create a flextable from an object of type '%s'.",
      "broom",
      "kmeans"
    ))
  }

  ## kmeans body ----
  clusters_stat <- broom::tidy(x)
  setDT(clusters_stat)
  keys <- c(
    "withinss",
    "size",
    setdiff(colnames(clusters_stat), c("cluster", "withinss", "size"))
  )
  key_type <- rep("Centers", length(keys))
  key_type[1:2] <- "Statistics"
  clusters_stat[, c(keys) := lapply(.SD, as.double), .SDcols = keys]

  clusters_stat <- melt.data.table(
    data = clusters_stat,
    id.vars = "cluster",
    measure.vars = setdiff(colnames(clusters_stat), c("cluster")),
    value.name = "value"
  )
  clusters_stat$variable <- factor(clusters_stat$variable, levels = keys)
  setorderv(clusters_stat, cols = c("variable"))
  setDF(clusters_stat)

  ## kmeans footer ----
  data_g <- broom::glance(x)
  w_labels <- c(
    "Total sum of squares",
    "Total within-cluster sum of squares",
    "Total between-cluster sum of squares",
    "BSS/TSS ratio",
    "Number of iterations"
  )
  totss <- data_g$totss
  tot.withinss <- data_g$tot.withinss
  betweenss <- data_g$betweenss
  ratio <- betweenss / totss
  w_labels <- paste0(
    w_labels,
    ": ",
    c(
      format_fun(totss),
      format_fun(tot.withinss),
      format_fun(betweenss),
      format_fun(ratio * 100, suffix = "%"),
      as.character(x$iter)
    )
  )

  ## tabulate ----
  ct <- tabulator(
    x = clusters_stat,
    rows = c("variable"),
    columns = "cluster",
    hidden_data = data.frame(
      variable = keys,
      key_type = key_type
    ),
    zz = as_paragraph(as_chunk(value))
  )

  ## flextable ----
  ft <- as_flextable(ct)
  ft <- add_footer_lines(ft, c("(*) Centers", w_labels))

  zz_labs <- tabulator_colnames(ct, type = "columns", columns = "zz")
  value_labq <- tabulator_colnames(ct, type = "hidden", columns = "value")

  for (j in seq_along(zz_labs)) {
    sym_val <- sym(value_labq[j])
    ft <- mk_par(
      ft,
      i = ~ variable %in% "size",
      j = zz_labs[j],
      value = as_paragraph(
        as_chunk(
          !!sym_val,
          formatter = function(x) sprintf("%.0f", x)
        )
      )
    )
    ft <- mk_par(
      ft,
      i = ~ !variable %in% c("size", "withinss"),
      j = zz_labs[j],
      value = as_paragraph(
        as_chunk(
          !!sym_val,
          formatter = function(x) format_fun(x, digits = digits)
        )
      )
    )
  }
  ft <- append_chunks(
    ft,
    j = 1,
    part = "body",
    i = ~ key_type %in% "Centers",
    as_chunk("*")
  )
  ft <- hline(
    ft,
    j = c("variable", zz_labs),
    i = ~ variable %in% "size",
    border = fp_border_default()
  )
  ft <- autofit(ft, part = c("header", "body"))
  ft <- align(ft, align = "right", part = "footer")
  ft <- align(ft, align = "left", i = 1, part = "footer")
  ft <- hrule(ft, rule = "auto")
  ft <- bold(ft, part = "header", bold = TRUE)
  ft
}

#' @export
#' @title Transform a 'pam' object into a flextable
#' @description produce a flextable describing a
#' pam object. The function is only using package 'broom'
#' that provides the data presented in the resulting flextable.
#' @param x a [cluster::pam()] object
#' @param digits number of digits for the numeric columns
#' @param ... unused argument
#' @examples
#' if (require("cluster")) {
#'   dat <- as.data.frame(scale(mtcars[1:7]))
#'   cl <- pam(dat, 3)
#'   ft <- as_flextable(cl)
#'   ft
#' }
#' @family as_flextable methods
as_flextable.pam <- function(x, digits = 4, ...) {
  if (!requireNamespace("broom", quietly = TRUE)) {
    if (!requireNamespace("broom", quietly = TRUE)) {
      stop(sprintf(
        "'%s' package should be installed to create a flextable from an object of type '%s'.",
        "broom",
        "pam"
      ))
    }
  }

  clus_stat_names <- c(
    "size",
    "max.diss",
    "avg.diss",
    "diameter",
    "separation",
    "avg.width"
  )

  ## kmeans body ----
  clusters_stat <- broom::tidy(x)
  setDT(clusters_stat)
  keys <- colnames(clusters_stat)
  clusters_stat <- melt.data.table(
    data = clusters_stat,
    id.vars = "cluster",
    measure.vars = setdiff(colnames(clusters_stat), c("cluster")),
    value.name = "value"
  )

  ## tabulate ----
  ct <- tabulator(
    x = clusters_stat,
    rows = c("variable"),
    columns = "cluster",
    zz = as_paragraph(as_chunk(value))
  )

  ## flextable ----
  ft <- as_flextable(ct)
  ft <- add_footer_lines(ft, "(*) Centers")

  zz_labs <- tabulator_colnames(ct, type = "columns", columns = "zz")
  value_labq <- tabulator_colnames(ct, type = "hidden", columns = "value")

  for (j in seq_along(zz_labs)) {
    sym_val <- sym(value_labq[j])
    ft <- mk_par(
      ft,
      i = ~ variable %in% "size",
      j = zz_labs[j],
      value = as_paragraph(
        as_chunk(
          !!sym_val,
          formatter = function(x) sprintf("%.0f", x)
        )
      )
    )
    ft <- mk_par(
      ft,
      i = ~ !variable %in%
        c(
          "size",
          "max.diss",
          "avg.diss",
          "diameter",
          "separation",
          "avg.width"
        ),
      j = zz_labs[j],
      value = as_paragraph(
        as_chunk(
          !!sym_val,
          formatter = function(x) {
            format_fun(x, digits = digits)
          }
        )
      )
    )
  }

  data_g <- broom::glance(x)

  ft <- append_chunks(
    ft,
    j = 1,
    part = "body",
    i = ~ !variable %in%
      c(
        "size",
        "max.diss",
        "avg.diss",
        "diameter",
        "separation",
        "avg.width"
      ),
    as_chunk("*")
  )
  ft <- hline(
    ft,
    j = c("variable", zz_labs),
    i = ~ variable %in% "avg.width",
    border = fp_border_default()
  )

  ft <- autofit(ft, part = c("header", "body"))
  ## kmeans footer ----
  ft <- add_footer_lines(
    x = ft,
    values = paste0(
      "The average silhouette width is ",
      formatC(data_g$avg.silhouette.width)
    )
  )

  ft <- align(ft, j = 1, align = "left", part = "footer")
  ft <- hrule(ft, rule = "auto")
  ft <- bold(ft, part = "header", bold = TRUE)
  ft
}


#' @export
#' @title Set automatic flextable printing for models
#' @description Define [as_flextable()] as
#' print method in an R Markdown document for models
#' of class:
#'
#' * lm
#' * glm
#' * models from package 'lme' and 'lme4'
#' * htest (t.test, chisq.test, ...)
#' * gam
#' * kmeans and pam
#'
#'
#' In a setup run chunk:
#'
#' ```
#' flextable::use_model_printer()
#' ```
#' @seealso [use_df_printer()], [flextable()]
#' @family flextable_configuration
use_model_printer <- function() {
  fun <- function(x, ...) knitr::knit_print(as_flextable(x))
  registerS3method("knit_print", "lm", fun)
  registerS3method("knit_print", "glm", fun)
  registerS3method("knit_print", "lme", fun)
  registerS3method("knit_print", "htest", fun)
  registerS3method("knit_print", "merMod", fun)
  registerS3method("knit_print", "gls", fun)
  registerS3method("knit_print", "nlme", fun)
  registerS3method("knit_print", "brmsfit", fun)
  registerS3method("knit_print", "glmmTMB", fun)
  registerS3method("knit_print", "glmmadmb", fun)
  registerS3method("knit_print", "gam", fun)
  registerS3method("knit_print", "pam", fun)
  registerS3method("knit_print", "kmeans", fun)
  invisible()
}

utils::globalVariables(
  c(
    "p.value",
    "value",
    "Type",
    ".row_title",
    "min_val",
    "max_val",
    "str_val"
  )
)

context("check as_flextable")

test_that("data.frame", {
  dummy_df <- data.frame(
    A = rep(letters[1:3], each = 2),
    B = seq(0, 1, length = 6)
  )
  ft <- as_flextable(dummy_df)
  expect_equal(flextable:::fortify_run(ft)$txt,
               c("A", "B", "character", "numeric", "a", "0.0", "a", "0.2",
                 "b", "0.4", "b", "0.6", "c", "0.8", "c", "1.0"))
  ft <- as_flextable(dummy_df[1,])
  expect_equal(flextable:::fortify_run(ft)$txt,
               c("A", "<br>", "character", "a", "B", "<br>", "numeric", "0"))
})

#' @export
#' @title Fix border issues when cells are merged
#' @description When cells are merged, the rendered borders will be those
#' of the first cell. If a column is made of three merged cells, the bottom
#' border that will be seen will be the bottom border of the first cell in the
#' column. From a user point of view, this is wrong, the bottom should be the one
#' defined for cell 3. This function modify the border values to avoid that effect.
#'
#' Note since version `0.9.7` that the function is called automatically
#' before rendering, user should not have to call this function anymore.
#' @inheritParams args_x_part
#' @examples
#' library(officer)
#' dat <- data.frame(a = 1:5, b = 6:10)
#' ft <- flextable(dat)
#' ft <- theme_box(ft)
#' ft <- merge_at(ft, i = 4:5, j = 1, part = "body")
#' ft <- hline(ft,
#'   i = 5, part = "body",
#'   border = fp_border(color = "red", width = 5)
#' )
#' print(ft)
#' ft <- fix_border_issues(ft)
#' print(ft)
#' @keywords internal
fix_border_issues <- function(x, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "fix_border_issues()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- fix_border_issues(x = x, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  x[[part]] <- correct_h_border(x[[part]])
  x[[part]] <- correct_v_border(x[[part]])
  x
}

correct_h_border <- function(x) {
  span_cols <- as.list(as.data.frame(x$spans$columns))

  bool_to_be_corrected <- lapply(span_cols, function(x) x > 1)
  l_apply_bottom_border <- lapply(span_cols, function(x) {
    rle_ <- rle(x)
    from <- cumsum(rle_$lengths)[rle_$values < 1]
    to <- cumsum(rle_$lengths)[rle_$values > 1]
    list(from = from, to = to, dont = length(to) < 1)
  })

  for (j in seq_len(ncol(x$spans$columns))) {
    apply_bottom_border <- l_apply_bottom_border[[j]]

    if (apply_bottom_border$dont) {
      next
    }

    for (i in seq_along(apply_bottom_border$from)) {
      i_from <- apply_bottom_border$from[i]
      i_to <- apply_bottom_border$to[i]

      x$styles$cells$border.color.bottom$data[
        i_to,
        x$col_keys[j]
      ] <- x$styles$cells$border.color.bottom$data[i_from, x$col_keys[j]]
      x$styles$cells$border.width.bottom$data[
        i_to,
        x$col_keys[j]
      ] <- x$styles$cells$border.width.bottom$data[i_from, x$col_keys[j]]
      x$styles$cells$border.style.bottom$data[
        i_to,
        x$col_keys[j]
      ] <- x$styles$cells$border.style.bottom$data[i_from, x$col_keys[j]]
    }
  }

  x
}
correct_v_border <- function(x) {
  span_rows <- as.list(as.data.frame(t(x$spans$rows)))

  l_apply_right_border <- lapply(span_rows, function(x) {
    rle_ <- rle(x)
    from <- cumsum(rle_$lengths)[rle_$values < 1]
    to <- cumsum(rle_$lengths)[rle_$values > 1]
    list(from = from, to = to, dont = length(to) < 1)
  })

  for (i in seq_along(l_apply_right_border)) {
    apply_right_border <- l_apply_right_border[[i]]

    if (apply_right_border$dont) {
      next
    }

    for (j in seq_along(apply_right_border$from)) {
      colkeyto <- x$col_keys[apply_right_border$to[j]]
      colkeyfrom <- x$col_keys[apply_right_border$from[j]]
      x$styles$cells$border.color.right$data[
        i,
        colkeyto
      ] <- x$styles$cells$border.color.right$data[i, colkeyfrom]
      x$styles$cells$border.width.right$data[
        i,
        colkeyto
      ] <- x$styles$cells$border.width.right$data[i, colkeyfrom]
      x$styles$cells$border.style.right$data[
        i,
        colkeyto
      ] <- x$styles$cells$border.style.right$data[i, colkeyfrom]
    }
  }

  x
}

def_fonts <- if (Sys.info()["sysname"] == "Windows") {
  "Arial"
} else if (Sys.info()["sysname"] == "Darwin") {
  "Helvetica"
} else {
  "DejaVu Sans"
}

flextable_global <- new.env(parent = emptyenv())
default_flextable_settings <- list(
  font.family = def_fonts,
  cs.family = def_fonts,
  eastasia.family = def_fonts,
  hansi.family = def_fonts,
  font.size = 11,
  font.color = "black",
  text.align = "left",
  padding.bottom = 5,
  padding.top = 5,
  padding.left = 5,
  padding.right = 5,
  line_spacing = 1,
  border.color = "#666666",
  border.width = .75,
  background.color = "transparent",
  table.layout = "fixed",
  decimal.mark = ".",
  big.mark = ",",
  digits = 1,
  pct_digits = 1,
  na_str = "",
  nan_str = "",
  fmt_date = "%Y-%m-%d",
  fmt_datetime = "%Y-%m-%d %H:%M:%S",
  extra_css = "",
  scroll = NULL,
  table_align = "center",
  split = TRUE,
  keep_with_next = FALSE,
  tabcolsep = 2,
  arraystretch = 1.5,
  float = "none",
  fonts_ignore = FALSE,
  theme_fun = "theme_booktabs",
  post_process_all = function(x) x,
  post_process_pdf = function(x) x,
  post_process_docx = function(x) x,
  post_process_html = function(x) x,
  post_process_pptx = function(x) x
)

flextable_global$defaults <- default_flextable_settings


#' @export
#' @title Modify flextable defaults formatting properties
#'
#' @description The current formatting properties (see [get_flextable_defaults()])
#' are automatically applied to every flextable you produce.
#' Use `set_flextable_defaults()` to override them. Use `init_flextable_defaults()`
#' to re-init all values with the package defaults.
#' @param font.family single character value. When format is Word, it specifies the font to
#' be used to format characters in the Unicode range (U+0000-U+007F). If you
#' want to use non ascii characters in Word, you should also set `hansi.family`
#' to the same family name.
#' @param cs.family optional and only for Word. Font to be used to format
#' characters in a complex script Unicode range. For example, Arabic
#' text might be displayed using the "Arial Unicode MS" font.
#' @param eastasia.family optional and only for Word. Font to be used to
#' format characters in an East Asian Unicode range. For example,
#' Japanese text might be displayed using the "MS Mincho" font.
#' @param hansi.family optional and only for Word. Font to be used to format
#' characters in a Unicode range which does not fall into one of the
#' other categories.
#' @param font.size font size (in point) - 0 or positive integer value.
#' @param font.color font color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param text.align text alignment - a single character value, expected value
#' is one of 'left', 'right', 'center', 'justify'.
#' @param padding.bottom,padding.top,padding.left,padding.right paragraph paddings - 0 or
#' positive integer value.
#' @param padding padding (shortcut for top, bottom, left and right padding)
#' @param line_spacing space between lines of text, 1 is single line spacing, 2 is double line spacing.
#' @param border.color border color - single character value
#' (e.g. "#000000" or "black").
#' @param border.width border width in points.
#' @param background.color cell background color - a single character value specifying a
#' valid color (e.g. "#000000" or "black").
#' @param table.layout 'autofit' or 'fixed' algorithm. Default to 'autofit'.
#' @param digits [formatC] argument used by [colformat_double()].
#' @param pct_digits number of digits for percentages.
#' @param decimal.mark,big.mark,na_str,nan_str [formatC] arguments used by [colformat_num()],
#' [colformat_double()], and [colformat_int()].
#' @param fmt_date,fmt_datetime formats for date and datetime columns as
#' documented in [strptime()]. Default to '%Y-%m-%d' and '%Y-%m-%d %H:%M:%S'.
#' @param fonts_ignore if TRUE, pdf-engine pdflatex can be used instead of
#' xelatex or lualatex. If pdflatex is used, fonts will be ignored because they are
#' not supported by pdflatex, whereas with the xelatex and lualatex engines they are.
#' @param extra_css css instructions to be integrated with the table.
#' @param scroll NULL or a list if you want to add a scroll-box.
#' See **scroll** element of argument `opts_html` in function [set_table_properties()].
#' @param table_align default flextable alignment, supported values are 'left', 'center'
#' and 'right'.
#' @param split Word option 'Allow row to break across pages' can be
#' activated when TRUE.
#' @param keep_with_next default initialization value used by the [paginate()]
#' function corresponding to the Word option "keep rows together" that will
#' be defined in the array.
#' @param tabcolsep space between the text and the left/right border of its containing
#' cell.
#' @param arraystretch height of each row relative to its default
#' height, the default value is 1.5.
#' @param float type of floating placement in the PDF document, one of:
#' * 'none' (the default value), table is placed after the preceding
#' paragraph.
#' * 'float', table can float to a place in the text where it fits best
#' * 'wrap-r', wrap text around the table positioned to the right side of the text
#' * 'wrap-l', wrap text around the table positioned to the left side of the text
#' * 'wrap-i', wrap text around the table positioned inside edge-near the binding
#' * 'wrap-o', wrap text around the table positioned outside edge-far from the binding
#' @param theme_fun a single character value (the name of the theme function
#' to be applied) or a theme function (input is a flextable, output is a flextable).
#' @param post_process_all Post-processing function
#' that will allow you to customize the the table. It will be executed before
#' call to post_process_pdf(), post_process_docx(),
#' post_process_html(), post_process_pptx().
#' @param post_process_pdf,post_process_docx,post_process_html,post_process_pptx Post-processing functions
#' that will allow you to customize the display by output type (pdf, html, docx, pptx).
#' They are executed just before printing the table.
#' @param ... unused or deprecated arguments
#' @return a list containing previous default values.
#' @examples
#' ft_1 <- qflextable(head(airquality))
#' ft_1
#'
#' old <- set_flextable_defaults(
#'   font.color = "#AA8855",
#'   border.color = "#8855AA"
#' )
#' ft_2 <- qflextable(head(airquality))
#' ft_2
#'
#' do.call(set_flextable_defaults, old)
#' @family themes_and_defaults
#' @importFrom utils modifyList
set_flextable_defaults <- function(
  font.family = NULL,
  font.size = NULL,
  font.color = NULL,
  text.align = NULL,
  padding = NULL,
  padding.bottom = NULL,
  padding.top = NULL,
  padding.left = NULL,
  padding.right = NULL,
  border.color = NULL,
  border.width = NULL,
  background.color = NULL,
  line_spacing = NULL,
  table.layout = NULL,
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  decimal.mark = NULL,
  big.mark = NULL,
  digits = NULL,
  pct_digits = NULL,
  na_str = NULL,
  nan_str = NULL,
  fmt_date = NULL,
  fmt_datetime = NULL,
  extra_css = NULL,
  scroll = NULL,
  table_align = "center",
  split = NULL,
  keep_with_next = NULL,
  tabcolsep = NULL,
  arraystretch = NULL,
  float = NULL,
  fonts_ignore = NULL,
  theme_fun = NULL,
  post_process_all = NULL,
  post_process_pdf = NULL,
  post_process_docx = NULL,
  post_process_html = NULL,
  post_process_pptx = NULL,
  ...
) {
  x <- list()

  if (!is.null(padding)) {
    if (is.null(padding.top)) {
      padding.top <- padding
    }
    if (is.null(padding.bottom)) {
      padding.bottom <- padding
    }
    if (is.null(padding.left)) {
      padding.left <- padding
    }
    if (is.null(padding.right)) padding.right <- padding
  }

  if (!is.null(font.family)) {
    x$font.family <- font.family
  }
  if (!is.null(cs.family)) {
    x$cs.family <- cs.family
  }
  if (!is.null(eastasia.family)) {
    x$eastasia.family <- eastasia.family
  }
  if (!is.null(hansi.family)) {
    x$hansi.family <- hansi.family
  }

  if (!is.null(font.size) && is.numeric(font.size) && !(font.size < 0)) {
    x$font.size <- font.size
  }

  if (!is.null(font.color)) {
    x$font.color <- font.color
  }

  if (!is.null(table_align) && table_align %in% c("left", "right", "center")) {
    x$table_align <- table_align
  }

  if (
    !is.null(text.align) &&
      text.align %in% c("left", "right", "center", "justify")
  ) {
    x$text.align <- text.align
  }

  if (
    !is.null(padding.bottom) &&
      is.numeric(padding.bottom) &&
      !(padding.bottom < 0)
  ) {
    x$padding.bottom <- padding.bottom
  }
  if (!is.null(padding.top) && is.numeric(padding.top) && !(padding.top < 0)) {
    x$padding.top <- padding.top
  }
  if (
    !is.null(padding.left) && is.numeric(padding.left) && !(padding.left < 0)
  ) {
    x$padding.left <- padding.left
  }
  if (
    !is.null(padding.right) && is.numeric(padding.right) && !(padding.right < 0)
  ) {
    x$padding.right <- padding.right
  }
  if (
    !is.null(line_spacing) && is.numeric(line_spacing) && !(line_spacing < 0)
  ) {
    x$line_spacing <- line_spacing
  }

  if (!is.null(border.color)) {
    x$border.color <- border.color
  }
  if (!is.null(border.width)) {
    x$border.width <- border.width
  }
  if (!is.null(background.color)) {
    x$background.color <- background.color
  }
  if (!is.null(table.layout)) {
    x$table.layout <- table.layout
  }
  if (!is.null(big.mark)) {
    x$big.mark <- big.mark
  }
  if (!is.null(decimal.mark)) {
    x$decimal.mark <- decimal.mark
  }
  if (!is.null(digits)) {
    x$digits <- digits
  }
  if (!is.null(pct_digits)) {
    x$pct_digits <- pct_digits
  }
  if (!is.null(na_str)) {
    x$na_str <- na_str
  }
  if (!is.null(nan_str)) {
    x$nan_str <- nan_str
  }
  if (!is.null(fonts_ignore)) {
    x$fonts_ignore <- fonts_ignore
  }
  if (!is.null(split)) {
    x$split <- split
  }
  if (!is.null(keep_with_next)) {
    x$keep_with_next <- keep_with_next
  }
  if (!is.null(tabcolsep)) {
    x$tabcolsep <- tabcolsep
  }
  if (!is.null(arraystretch)) {
    x$arraystretch <- arraystretch
  }
  if (!is.null(float)) {
    x$float <- float
  }

  if (!is.null(scroll)) {
    x$scroll <- scroll
  }
  if (!is.null(extra_css)) {
    x$extra_css <- extra_css
  }
  if (!is.null(fmt_date)) {
    x$fmt_date <- fmt_date
  }
  if (!is.null(fmt_datetime)) {
    x$fmt_datetime <- fmt_datetime
  }

  if (
    !is.null(theme_fun) && is.character(theme_fun) && length(theme_fun) == 1
  ) {
    x$theme_fun <- theme_fun
  }
  if (!is.null(theme_fun) && is.function(theme_fun)) {
    x$theme_fun <- theme_fun
  }

  if (!is.null(post_process_all) && is.function(post_process_all)) {
    x$post_process_all <- post_process_all
  }
  if (!is.null(post_process_pdf) && is.function(post_process_pdf)) {
    x$post_process_pdf <- post_process_pdf
  }
  if (!is.null(post_process_docx) && is.function(post_process_docx)) {
    x$post_process_docx <- post_process_docx
  }
  if (!is.null(post_process_html) && is.function(post_process_html)) {
    x$post_process_html <- post_process_html
  }
  if (!is.null(post_process_pptx) && is.function(post_process_pptx)) {
    x$post_process_pptx <- post_process_pptx
  }

  flextable_defaults <- flextable_global$defaults

  flextable_new_defaults <- modifyList(flextable_defaults, x)
  flextable_global$defaults <- flextable_new_defaults
  invisible(flextable_defaults)
}

#' @export
#' @rdname set_flextable_defaults
init_flextable_defaults <- function() {
  x <- default_flextable_settings
  flextable_global$defaults <- x
  class(x) <- "flextable_defaults"
  invisible(x)
}


#' @export
#' @title Get flextable defaults formatting properties
#'
#' @description The current formatting properties
#' are automatically applied to every flextable you produce.
#' These default values are returned by this function.
#' @return a list containing default values.
#' @examples
#' get_flextable_defaults()
#' @family themes_and_defaults
get_flextable_defaults <- function() {
  x <- flextable_global$defaults
  class(x) <- "flextable_defaults"
  x
}

#' @export
print.flextable_defaults <- function(x, ...) {
  cat("## style properties\n")
  styles <- c(
    "font.family",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "font.size",
    "font.color",
    "text.align",
    "padding.bottom",
    "padding.top",
    "padding.left",
    "padding.right",
    "line_spacing",
    "border.color",
    "border.width",
    "background.color"
  )
  df <- data.frame(
    property = styles,
    value = unlist(x[styles]),
    stringsAsFactors = FALSE
  )
  row.names(df) <- NULL
  print(df)
  cat("\n")

  cat("## cell content settings\n")
  contents <- c(
    "decimal.mark",
    "big.mark",
    "digits",
    "pct_digits",
    "na_str",
    "nan_str",
    "fmt_date",
    "fmt_datetime"
  )
  df <- data.frame(
    property = contents,
    value = unlist(x[contents]),
    stringsAsFactors = FALSE
  )
  row.names(df) <- NULL
  print(df)
  cat("\n")

  cat("## table.layout is:", x$table.layout, "\n")
  if (is.character(x$theme_fun)) {
    cat("## default theme is:", x$theme_fun, "\n")
  }

  invisible(NULL)
}


#' @export
#' @inheritParams officer::fp_text
#' @title Create text formatting with flextable defaults
#'
#' @description Create a [officer::fp_text()] object whose unspecified
#' arguments inherit from [set_flextable_defaults()] (font family, size, color).
#'
#' Use `fp_text_default()` instead of `officer::fp_text()` when building
#' chunks with [as_chunk()]: only override the properties you need,
#' and the rest will match the table's default style.
#'
#' See also [set_flextable_defaults()] to modify the inherited values.
#' @examples
#' library(flextable)
#'
#' set_flextable_defaults(
#'   font.size = 11, font.color = "#303030",
#'   padding = 3, table.layout = "autofit"
#' )
#' z <- flextable(head(cars))
#'
#' z <- compose(
#'   x = z,
#'   i = ~ speed < 6,
#'   j = "speed",
#'   value = as_paragraph(
#'     as_chunk("slow... ", props = fp_text_default(color = "red")),
#'     as_chunk(speed, props = fp_text_default(italic = TRUE))
#'   )
#' )
#' z
#'
#' init_flextable_defaults()
#'
#' @family default_formatting_properties
#' @seealso [as_chunk()], [compose()], [append_chunks()], [prepend_chunks()]
fp_text_default <- function(
  color = flextable_global$defaults$font.color,
  font.size = flextable_global$defaults$font.size,
  bold = FALSE,
  italic = FALSE,
  underlined = FALSE,
  strike = FALSE,
  font.family = flextable_global$defaults$font.family,
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  vertical.align = "baseline",
  shading.color = "transparent"
) {
  fp_text(
    color = color,
    font.size = font.size,
    bold = bold,
    italic = italic,
    underlined = underlined,
    font.family = font.family,
    cs.family = cs.family,
    eastasia.family = eastasia.family,
    hansi.family = hansi.family,
    vertical.align = vertical.align,
    shading.color = shading.color,
    strike = strike
  )
}
#' @export
#' @inheritParams officer::fp_border
#' @title Create border formatting with flextable defaults
#'
#' @description Create a [officer::fp_border()] object whose unspecified
#' arguments inherit from [set_flextable_defaults()] (border color, width).
#'
#' Use `fp_border_default()` instead of `officer::fp_border()` so that
#' borders automatically match the table's default style.
#' @examples
#' library(flextable)
#'
#' set_flextable_defaults(
#'   border.color = "orange"
#' )
#'
#' z <- flextable(head(cars))
#' z <- theme_vanilla(z)
#' z <- vline(
#'   z,
#'   j = 1, part = "all",
#'   border = officer::fp_border()
#' )
#' z <- vline(
#'   z,
#'   j = 2, part = "all",
#'   border = fp_border_default()
#' )
#' z
#'
#' init_flextable_defaults()
#'
#' @family default_formatting_properties
#' @seealso [hline()], [vline()]
fp_border_default <- function(
  color = flextable_global$defaults$border.color,
  style = "solid",
  width = flextable_global$defaults$border.width
) {
  fp_border(
    color = color,
    style = style,
    width = width
  )
}

context("check markdown captions")

skip_on_cran()
skip_if_not_installed("doconv")
library(doconv)
skip_if_not(doconv::msoffice_available())
skip_if_not(pandoc_version() >= numeric_version("2"))
skip_if_not_installed("webshot2")

library(rmarkdown)
library(xml2)
library(officer)

init_flextable_defaults()

test_that("rmarkdown caption", {
  local_edition(3)

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/rmarkdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".pdf")
  render(rmd_file, output_file = outfile, output_format = rmarkdown::pdf_document(latex_engine = "xelatex"),
         envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(x = outfile, name = "rmarkdown_pdf_document", engine = "testthat")

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/rmarkdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".docx")
  render(
    input = rmd_file,
    output_format = "rmarkdown::word_document",
    output_file = outfile,
    envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(name = "rmarkdown_word_document", x = outfile, engine = "testthat")

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/rmarkdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".html")
  render(rmd_file, output_format = "rmarkdown::html_document", output_file = outfile, envir = new.env(), quiet = TRUE)
  expect_snapshot_html(name = "rmarkdown_html_document", outfile, engine = "testthat",
                       zoom = 3, expand = 10)
})

test_that("bookdown caption", {
  local_edition(3)

  skip_if_not_installed("bookdown")

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/bookdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".pdf")
  render(rmd_file, output_file = outfile, output_format = bookdown::pdf_document2(latex_engine = "xelatex"),
         envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(x = outfile, name = "bookdown_pdf_document2", engine = "testthat")

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/bookdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".docx")
  render(
    input = rmd_file,
    output_format = "bookdown::word_document2",
    output_file = outfile,
    envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(name = "bookdown_word_document2", x = outfile, engine = "testthat")

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/bookdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".html")
  render(rmd_file, output_format = "bookdown::html_document2", output_file = outfile, envir = new.env(), quiet = TRUE)
  expect_snapshot_html(name = "bookdown_html_document2", outfile, engine = "testthat",
                       zoom = 3, expand = 10)
})

test_that("rdocx caption", {
  local_edition(3)

  skip_if_not_installed("bookdown")
  skip_if_not_installed("officedown")

  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/bookdown.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".docx")
  render(
    input = rmd_file,
    output_format = bookdown::markdown_document2(base_format = "officedown::rdocx_document"),
    output_file = outfile,
    envir = new.env(), quiet = TRUE)
  expect_snapshot_doc(name = "officedown_word_document2", x = outfile, engine = "testthat")
})


init_flextable_defaults()

#' @export
#' @title Apply borderless theme
#' @description Apply theme borderless to a flextable.
#' All borders are removed. Header text is bold,
#' text columns are left aligned, other columns are
#' right aligned.
#' @inheritParams args_x_only
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_borderless(ft)
#' ft
theme_borderless <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_borderless()"
    ))
  }
  x <- border_remove(x)
  x <- bold(x = x, bold = TRUE, part = "header")
  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)
  x
}

#' @importFrom officer fp_border fp_par
#' @export
#' @title Apply vanilla theme
#' @description Apply theme vanilla to a flextable:
#' The external horizontal lines of the different parts of
#' the table (body, header, footer) are black 2 points thick,
#' the external horizontal lines of the different parts
#' are black 0.5 point thick. Header text is bold,
#' text columns are left aligned, other columns are
#' right aligned.
#' @section behavior:
#' Theme functions are not like 'ggplot2' themes. They are applied to the existing
#' table **immediately**. If you add a row in the footer, the new row is not formatted
#' with the theme. The theme function applies the theme only to existing elements
#' when the function is called.
#'
#' That is why theme functions should be applied after all elements of the table
#' have been added (mainly additionnal header or footer rows).
#'
#' If you want to automatically apply a theme function to each flextable,
#' you can use the `theme_fun` argument of [set_flextable_defaults()]; be
#' aware that this theme function is applied as the last instruction when
#' calling `flextable()` - so if you add headers or footers to the array,
#' they will not be formatted with the theme.
#'
#' You can also use the `post_process_html` argument
#' of [set_flextable_defaults()] (or `post_process_pdf`,
#' `post_process_docx`, `post_process_pptx`) to specify a theme
#' to be applied systematically before the `flextable()` is printed;
#' in this case, don't forget to take care that the theme doesn't
#' override any formatting done before the print statement.
#' @inheritParams args_x_only
#' @family themes_and_defaults
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_vanilla(ft)
#' ft
theme_vanilla <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_vanilla()"
    ))
  }

  std_b <- fp_border(
    width = flextable_global$defaults$border.width * 2,
    color = flextable_global$defaults$border.color
  )
  thin_b <- fp_border(
    width = flextable_global$defaults$border.width,
    color = flextable_global$defaults$border.color
  )

  x <- border_remove(x)

  x <- hline(x, border = thin_b, part = "all")
  x <- hline_top(x, border = std_b, part = "header")
  x <- hline_bottom(x, border = std_b, part = "header")
  x <- hline_bottom(x, border = std_b, part = "body")
  x <- bold(x = x, bold = TRUE, part = "header")
  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)
  fix_border_issues(x)
}

#' @importFrom officer fp_border fp_par
#' @export
#' @title Apply box theme
#' @description Apply theme box to a flextable
#' @inheritParams args_x_only
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_box(ft)
#' ft
theme_box <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_box()"
    ))
  }

  x <- border_remove(x)

  fp_bdr <- fp_border(
    width = flextable_global$defaults$border.width,
    color = flextable_global$defaults$border.color
  )

  x <- border_outer(x, part = "all", border = fp_bdr)
  x <- border_inner_h(x, border = fp_bdr, part = "all")
  x <- border_inner_v(x, border = fp_bdr, part = "all")

  x <- bold(x = x, bold = TRUE, part = "header")
  x <- italic(x = x, italic = TRUE, part = "footer")
  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)
  fix_border_issues(x)
}

#' @importFrom officer fp_border fp_par
#' @export
#' @title Apply alafoli theme
#' @description Apply alafoli theme
#' @inheritParams args_x_only
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_alafoli(ft)
#' ft
theme_alafoli <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_alafoli()"
    ))
  }

  fp_bdr <- fp_border(
    width = flextable_global$defaults$border.width,
    color = flextable_global$defaults$border.color
  )

  x <- border_remove(x)
  x <- bg(x, bg = "transparent", part = "all")
  x <- color(x, color = "#666666", part = "all")
  x <- bold(x = x, bold = FALSE, part = "all")
  x <- italic(x = x, italic = FALSE, part = "all")
  x <- padding(x = x, padding = 3, part = "all")

  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)
  x <- hline_bottom(x, part = "header", border = fp_bdr)
  x <- hline_top(x, part = "body", border = fp_bdr)
  fix_border_issues(x)
}

#' @export
#' @title Apply Sith Lord Darth Vader theme
#' @description Apply Sith Lord Darth Vader theme to a flextable
#' @inheritParams args_x_only
#' @param ... unused
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_vader(ft)
#' ft
theme_vader <- function(x, ...) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_vader()"
    ))
  }

  x <- border_remove(x)
  x <- bg(x, bg = "#242424", part = "all")
  x <- color(x, color = "#dfdfdf", part = "all")
  x <- bold(x = x, bold = FALSE, part = "all")
  x <- bold(x = x, bold = TRUE, part = "header")
  x <- italic(x = x, italic = FALSE, part = "all")

  big_border <- fp_border(
    color = "#ff0000",
    width = flextable_global$defaults$border.width * 2
  )

  h_nrow <- nrow_part(x, "header")
  b_nrow <- nrow_part(x, "body")
  f_nrow <- nrow_part(x, "footer")

  if (h_nrow > 0) {
    x <- hline_bottom(x, border = big_border, part = "header")
  }

  if (b_nrow > 0 || h_nrow < 1) {
    x <- hline_top(x, border = big_border, part = "body")
  }
  if (f_nrow > 0) {
    x <- hline_top(x, border = big_border, part = "footer")
  }

  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)
  fix_border_issues(x)
}

#' @export
#' @title Apply zebra theme
#' @description Apply theme zebra to a flextable
#' @inheritParams args_x_only
#' @param odd_header,odd_body,even_header,even_body odd/even colors for table header and body
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_zebra(ft)
#' ft
theme_zebra <- function(
  x,
  odd_header = "#CFCFCF",
  odd_body = "#EFEFEF",
  even_header = "transparent",
  even_body = "transparent"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_zebra()"
    ))
  }

  h_nrow <- nrow_part(x, "header")
  f_nrow <- nrow_part(x, "footer")
  b_nrow <- nrow_part(x, "body")

  x <- border_remove(x)
  x <- align(x = x, align = "center", part = "header")

  if (h_nrow > 0) {
    even <- seq_len(h_nrow) %% 2 == 0
    odd <- !even

    x <- bg(x = x, i = odd, bg = odd_header, part = "header")
    x <- bg(x = x, i = even, bg = even_header, part = "header")
    x <- bold(x = x, bold = TRUE, part = "header")
  }
  if (f_nrow > 0) {
    even <- seq_len(f_nrow) %% 2 == 0
    odd <- !even

    x <- bg(x = x, i = odd, bg = odd_header, part = "footer")
    x <- bg(x = x, i = even, bg = even_header, part = "footer")
    x <- bold(x = x, bold = TRUE, part = "footer")
  }
  if (b_nrow > 0) {
    even <- seq_len(b_nrow) %% 2 == 0
    odd <- !even

    x <- bg(x = x, i = odd, bg = odd_body, part = "body")
    x <- bg(x = x, i = even, bg = even_body, part = "body")
  }
  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)

  x
}

#' @export
#' @title Apply tron legacy theme
#' @description Apply theme tron legacy to a flextable
#' @inheritParams args_x_only
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_tron_legacy(ft)
#' ft
theme_tron_legacy <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_tron_legacy()"
    ))
  }

  h_nrow <- nrow_part(x, "header")
  f_nrow <- nrow_part(x, "footer")
  b_nrow <- nrow_part(x, "body")

  x <- border(
    x = x,
    border = fp_border(
      width = flextable_global$defaults$border.width,
      color = "#6FC3DF"
    ),
    part = "all"
  )
  x <- align(x = x, align = "right", part = "all")
  x <- bg(x = x, bg = "#0C141F", part = "all")

  if (h_nrow > 0) {
    x <- bold(x = x, bold = TRUE, part = "header")
    x <- color(x = x, color = "#DF740C", part = "header")
  }
  if (f_nrow > 0) {
    x <- color(x = x, color = "#DF740C", part = "footer")
  }
  if (b_nrow > 0) {
    x <- color(x = x, color = "#FFE64D", part = "body")
  }
  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)

  fix_border_issues(x)
}

#' @export
#' @title Apply tron theme
#' @description Apply theme tron to a flextable
#' @inheritParams args_x_only
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_tron(ft)
#' ft
theme_tron <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_tron()"
    ))
  }

  h_nrow <- nrow_part(x, "header")
  f_nrow <- nrow_part(x, "footer")
  b_nrow <- nrow_part(x, "body")

  x <- border(
    x = x,
    border = fp_border(
      width = flextable_global$defaults$border.width,
      color = "#a4cee5"
    ),
    part = "all"
  )
  x <- align(x = x, align = "right", part = "all")
  x <- bg(x = x, bg = "#000000", part = "all")

  if (h_nrow > 0) {
    x <- bold(x = x, bold = TRUE, part = "header")
    x <- color(x = x, color = "#ec9346", part = "header")
  }
  if (f_nrow > 0) {
    x <- color(x = x, color = "#ec9346", part = "footer")
  }
  if (b_nrow > 0) {
    x <- color(x = x, color = "#a4cee5", part = "body")
  }
  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)

  fix_border_issues(x)
}

#' @export
#' @title Apply booktabs theme
#' @description Apply theme booktabs to a flextable
#' @inheritParams args_x_only
#' @param bold_header header will be bold if TRUE.
#' @param ... unused
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(airquality))
#' ft <- theme_booktabs(ft)
#' ft
theme_booktabs <- function(x, bold_header = FALSE, ...) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_booktabs()"
    ))
  }

  big_border <- fp_border(
    width = flextable_global$defaults$border.width * 2,
    color = flextable_global$defaults$border.color
  )
  std_border <- update(
    big_border,
    width = flextable_global$defaults$border.width
  )

  h_nrow <- nrow_part(x, "header")
  f_nrow <- nrow_part(x, "footer")
  b_nrow <- nrow_part(x, "body")

  x <- border_remove(x)

  if (h_nrow > 0) {
    x <- hline_top(x, border = big_border, part = "header")
    x <- hline(x, border = std_border, part = "header")
    x <- hline_bottom(x, border = big_border, part = "header")
    x <- bold(x, bold = bold_header, part = "header")
  } else if (b_nrow > 0) {
    x <- hline_top(x, border = big_border, part = "body")
  } else if (f_nrow > 0) {
    x <- hline_top(x, border = big_border, part = "body")
  }

  if (f_nrow > 0) {
    x <- hline_bottom(x, border = big_border, part = "footer")
  }
  if (b_nrow > 0) {
    x <- hline_bottom(x, border = big_border, part = "body")
  }

  x <- align_text_col(x, align = "left", header = TRUE)
  x <- align_nottext_col(x, align = "right", header = TRUE)
  fix_border_issues(x)
}

#' @export
#' @title Apply APA theme
#' @description Apply theme APA (the stylistic style of the American
#'              Psychological Association) to a flextable
#' @inheritParams args_x_only
#' @param ... unused
#' @family themes_and_defaults
#' @inheritSection theme_vanilla behavior
#' @examples
#' ft <- flextable(head(mtcars * 22.22))
#' ft <- theme_apa(ft)
#' ft
theme_apa <- function(x, ...) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "theme_apa()"
    ))
  }
  apa.border <- list(
    "width" = flextable_global$defaults$border.width,
    color = "black",
    style = "solid"
  )
  x <- font(x, part = "all", fontname = "Times New Roman")
  x <- line_spacing(x, space = 2, part = "all")
  x <- hline_top(x, part = "head", border = apa.border)
  x <- hline_bottom(x, part = "head", border = apa.border)
  x <- hline_top(x, part = "body", border = apa.border)
  x <- hline_bottom(x, part = "body", border = apa.border)
  x <- align(x, align = "center", part = "all")
  x <- valign(x, valign = "center", part = "all")
  x <- colformat_double(x, digits = 2)
  fix_border_issues(x)
}

gdtools::register_liberationsans()

init_flextable_defaults()
set_flextable_defaults(
  font.family = "Liberation Sans",
  border.color = "#333333"
)

make_ft <- function() {
  autofit(flextable(head(airquality)))
}

make_ft_with_footer <- function() {
  ft <- make_ft()
  ft <- add_footer_lines(ft, values = "a note")
  autofit(ft)
}

# -- theme_borderless --------------------------------------------------------

test_that("theme_borderless returns a flextable", {
  ft <- theme_borderless(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_borderless rejects non-flextable", {
  expect_error(theme_borderless("not a ft"), "flextable objects")
})

test_that("theme_borderless renders with ragg", {
  ft <- theme_borderless(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_borderless works with footer", {
  ft <- theme_borderless(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- theme_apa ---------------------------------------------------------------

test_that("theme_apa returns a flextable", {
  ft <- theme_apa(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_apa rejects non-flextable", {
  expect_error(theme_apa("not a ft"), "flextable objects")
})

test_that("theme_apa renders with ragg", {
  ft <- theme_apa(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_apa works with footer", {
  ft <- theme_apa(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- theme_tron --------------------------------------------------------------

test_that("theme_tron returns a flextable", {
  ft <- theme_tron(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_tron rejects non-flextable", {
  expect_error(theme_tron("not a ft"), "flextable objects")
})

test_that("theme_tron renders with ragg", {
  ft <- theme_tron(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_tron works with footer", {
  ft <- theme_tron(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- theme_tron_legacy -------------------------------------------------------

test_that("theme_tron_legacy returns a flextable", {
  ft <- theme_tron_legacy(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_tron_legacy rejects non-flextable", {
  expect_error(theme_tron_legacy("not a ft"), "flextable objects")
})

test_that("theme_tron_legacy renders with ragg", {
  ft <- theme_tron_legacy(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_tron_legacy works with footer", {
  ft <- theme_tron_legacy(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- theme_zebra -------------------------------------------------------------

test_that("theme_zebra returns a flextable", {
  ft <- theme_zebra(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_zebra rejects non-flextable", {
  expect_error(theme_zebra("not a ft"), "flextable objects")
})

test_that("theme_zebra custom colors", {
  ft <- theme_zebra(
    make_ft(),
    odd_header = "#FF0000",
    odd_body = "#00FF00",
    even_header = "#0000FF",
    even_body = "#FFFF00"
  )
  expect_s3_class(ft, "flextable")
})

test_that("theme_zebra renders with ragg", {
  ft <- theme_zebra(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_zebra works with footer", {
  ft <- theme_zebra(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- theme_vader -------------------------------------------------------------

test_that("theme_vader returns a flextable", {
  ft <- theme_vader(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_vader rejects non-flextable", {
  expect_error(theme_vader("not a ft"), "flextable objects")
})

test_that("theme_vader renders with ragg", {
  ft <- theme_vader(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_vader works with footer", {
  ft <- theme_vader(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- theme_alafoli -----------------------------------------------------------

test_that("theme_alafoli returns a flextable", {
  ft <- theme_alafoli(make_ft())
  expect_s3_class(ft, "flextable")
})

test_that("theme_alafoli rejects non-flextable", {
  expect_error(theme_alafoli("not a ft"), "flextable objects")
})

test_that("theme_alafoli renders with ragg", {
  ft <- theme_alafoli(make_ft())
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("theme_alafoli works with footer", {
  ft <- theme_alafoli(make_ft_with_footer())
  expect_s3_class(ft, "flextable")
  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)
  save_as_image(ft, path = tf, res = 150)
  expect_gt(file.info(tf)$size, 1000)
})

# -- grob structure checks ---------------------------------------------------

test_that("theme_borderless grob has no border segments", {
  ft <- theme_borderless(make_ft())
  gr <- gen_grob(ft)
  bdr <- gr$children[[1]]$children$borders
  expect_length(bdr$children, 0)
})

test_that("all themes produce valid grobs", {
  ft <- make_ft()
  themes <- list(
    theme_borderless,
    theme_apa,
    theme_tron,
    theme_tron_legacy,
    theme_zebra,
    theme_vader,
    theme_alafoli
  )
  for (thm in themes) {
    themed <- thm(ft)
    gr <- gen_grob(themed)
    expect_s3_class(gr, "grob")
    expect_true(length(gr$children) > 0)
  }
})

init_flextable_defaults()

#' @noRd
#' @title make border top and bottom restructured
#' as hline. If two borders overlap, the largest is
#' choosen.
fortify_latex_borders <- function(x) {
  properties_df <- x[,
    .SD,
    .SDcols = c(
      ".part",
      ".col_id",
      ".row_id",
      "colspan",
      "rowspan",
      "border.width.top",
      "border.color.top",
      "border.width.bottom",
      "border.color.bottom",
      "border.width.left",
      "border.color.left",
      "border.width.right",
      "border.color.right",
      "background.color"
    )
  ]
  col_id_levels <- levels(properties_df$.col_id)
  properties_df[,
    c("vspan_id") := list(rleid(cumsum(.SD$colspan))),
    by = c(".part", ".col_id")
  ]
  properties_df[,
    c("draw_hline") := list({
      z <- logical(nrow(.SD))
      z[length(z)] <- TRUE
      z
    }),
    by = c(".part", ".col_id", "vspan_id")
  ]
  properties_df[,
    c("draw_hline_top") := list({
      z <- logical(nrow(.SD))
      z[1] <- TRUE
      z
    }),
    by = c(".part", ".col_id", "vspan_id")
  ]

  top <- dcast(
    properties_df,
    .part + .row_id ~ .col_id,
    value.var = "border.width.top"
  )
  bottom <- dcast(
    properties_df,
    .part + .row_id ~ .col_id,
    value.var = "border.width.bottom"
  )
  draw_hline <- dcast(
    properties_df,
    .part + .row_id ~ .col_id,
    value.var = "draw_hline"
  )
  draw_hline_top <- dcast(
    properties_df,
    .part + .row_id ~ .col_id,
    value.var = "draw_hline_top"
  )
  top_mat <- as.matrix(top[, 3:ncol(top)])
  bot_mat <- as.matrix(bottom[, 3:ncol(top)])

  # don't want the merged columns widths inside
  draw_hline_mat <- as.matrix(draw_hline[, 3:ncol(top)])
  draw_hline_top_mat <- as.matrix(draw_hline_top[, 3:ncol(top)])
  bot_mat <- bot_mat * (draw_hline_mat > 0)
  top_mat <- top_mat * (draw_hline_top_mat > 0)

  new_row_n <- nrow(top) + 1

  if (new_row_n > 2) {
    # at least 3 rows
    # hlinemat is the top border and all bottom borders
    hlinemat <- matrix(0.0, nrow = new_row_n, ncol = ncol(top_mat))

    # add top border values
    hlinemat[1, ] <- top_mat[1, , drop = FALSE]
    # add bottom border values
    hlinemat[nrow(hlinemat), ] <- bot_mat[nrow(bot_mat), , drop = FALSE]

    # lines width in between must all have the same width, i.e. max observed width

    hlinemat[setdiff(seq_len(new_row_n), c(1, new_row_n)), ] <- pmax(
      bot_mat[-nrow(bot_mat), , drop = FALSE],
      top_mat[-1, , drop = FALSE]
    )

    # now lets replace values
    bottom[, 3:ncol(top)] <- as.data.table(hlinemat[-1, ])
    top[1, 3:ncol(top)] <- as.data.table(hlinemat[1, , drop = FALSE])
    top[2:nrow(top), 3:ncol(top)] <- 0.0

    top <- melt(
      top,
      id.vars = c(".part", ".row_id"),
      variable.name = ".col_id",
      value.name = "border.width.top",
      variable.factor = FALSE
    )
    top$.col_id <- factor(top$.col_id, levels = col_id_levels)
    bottom <- melt(
      bottom,
      id.vars = c(".part", ".row_id"),
      variable.name = ".col_id",
      value.name = "border.width.bottom",
      variable.factor = FALSE
    )
    bottom$.col_id <- factor(bottom$.col_id, levels = col_id_levels)

    properties_df$border.width.bottom <- NULL
    properties_df$border.width.top <- NULL

    properties_df <- merge(
      x = properties_df,
      y = top,
      by = c(".part", ".row_id", ".col_id")
    )
    properties_df <- merge(
      x = properties_df,
      y = bottom,
      by = c(".part", ".row_id", ".col_id")
    )
  }
  properties_df
}

#' @importFrom data.table fifelse
latex_gridlines <- function(properties_df) {
  stopifnot(is.data.table(properties_df))

  x <- fortify_latex_borders(properties_df)

  # init computed latex instructions
  x[,
    c(
      "vborder_left",
      "vborder_right",
      "hborder_bottom",
      "hborder_top",
      "hborder_bottom_pre_vline",
      "hborder_bottom_post_vline"
    ) := list(
      "",
      "",
      "~",
      "~",
      "",
      ""
    )
  ]
  x[
    has_border(x, "left"),
    c("vborder_left") := list(
      vline_token(w = .SD$border.width.left, cols = .SD$border.color.left)
    )
  ]
  x[
    has_border(x, "right"),
    c("vborder_right") := list(
      fcase(
        (as.integer(.SD$.col_id) + .SD$rowspan) == (nlevels(.SD$.col_id) + 1L) ,
        vline_token(w = .SD$border.width.right, cols = .SD$border.color.right) ,
        default = ""
      )
    )
  ]
  vlines <- x[,
    .SD,
    .SDcols = c(".part", ".col_id", ".row_id", "vborder_left", "vborder_right")
  ]
  setDF(vlines)

  is_transparent <- !has_background(x)

  if (is_transparent) {
    fun_hborder <- cline_token
  } else {
    fun_hborder <- hline_token
  }

  # generate hborder_top only for the first row
  x[
    x$.row_id %in% 1 & as.integer(x$.part) == min(as.integer(x$.part)),
    c("hborder_top") := list(fun_hborder(
      w = .SD$border.width.top,
      cols = .SD$border.color.top
    ))
  ]
  # generate hborder_bottom for those that have bottom borders
  x[
    has_border(x, "bottom"),
    c("hborder_bottom") := list(fun_hborder(
      w = .SD$border.width.bottom,
      cols = .SD$border.color.bottom
    ))
  ]
  if (is_transparent) {
    x[
      !x$draw_hline,
      c("hborder_bottom") := list({
        rep("ascline{0pt}{FFFFFF}", nrow(.SD))
      }),
      by = c(".part", ".col_id", "vspan_id")
    ]

    x[x$hborder_top %in% "~", c("hborder_top") := list("ascline{0pt}{FFFFFF}")]
    hlines <- x[,
      list(
        hlines_b_strings = cline_intruction(.SD$hborder_bottom),
        hlines_t_strings = cline_intruction(.SD$hborder_top)
      ),
      by = c(".part", ".row_id")
    ]
    setDF(hlines)
  } else {
    # set hborder_bottom_pre_vline to '|' for the first column where there are bottom and left borders
    x[
      has_border(x, "bottom") &
        has_border(x, "left") &
        x$.col_id %in% head(levels(x$.col_id), 1),
      c("hborder_bottom_pre_vline") := list("|")
    ]
    # set hborder_bottom_post_vline to '|' where there are bottom and right borders
    x[,
      c("has_bdr_right") := list(
        shift(.SD[["border.width.left"]], type = "lead") > 0 &
          colalpha(shift(.SD[["border.color.left"]], type = "lead")) > 0
      ),
      by = c(".part", ".row_id")
    ]
    x[
      x[["has_bdr_right"]],
      c("hborder_bottom_post_vline") := list("|")
    ]
    # if cells are vertically merged, dont draw bottom borders nor their vertical columns/joins
    x[,
      c("hborder_bottom", "hborder_bottom_post_vline") := list(
        data.table::fifelse(c(.SD$colspan[-1], 1) < 1, "~", .SD$hborder_bottom),
        data.table::fifelse(
          c(.SD$colspan[-1], 1) < 1,
          "",
          .SD$hborder_bottom_post_vline
        )
      ),
      by = c(".part", ".col_id")
    ]

    # reinit color and line size before drawing new h borders
    x_rulecolor_start <- x[x$.col_id %in% head(levels(x$.col_id), 1), ]
    x_rulecolor_start <- sprintf(
      "\\noalign{\\global\\arrayrulewidth %spt}\\arrayrulecolor[HTML]{%s}\n\n",
      format_double(x_rulecolor_start$border.width.left, 2),
      colcode0(x_rulecolor_start$border.color.left)
    )

    hlines <- x[,
      list(
        hlines_b_strings = hhline_instruction(
          .SD$hborder_bottom,
          pre_str = .SD$hborder_bottom_pre_vline,
          post_str = .SD$hborder_bottom_post_vline
        ),
        hlines_t_strings = hhline_instruction(.SD$hborder_top)
      ),
      by = c(".part", ".row_id")
    ]
    hlines$hlines_b_strings <- paste0(
      x_rulecolor_start,
      hlines$hlines_b_strings
    )
    setDF(hlines)
  }

  list(hlines = hlines, vlines = vlines)
}


# utils ----
## cline command for hline when no background
cline_cmd <- paste0(
  "\\providecommand{\\ascline}[3]{",
  "\\noalign{\\global\\arrayrulewidth #1}",
  "\\arrayrulecolor[HTML]{#2}\\cline{#3}}"
)

has_border <- function(x, side = c("left", "right", "top", "bottom")) {
  side_ <- match.arg(side, several.ok = FALSE)
  size_name <- paste0("border.width.", side_)
  col_name <- paste0("border.color.", side_)
  x[[size_name]] > 0 & colalpha(x[[col_name]]) > 0
}
has_background <- function(x) {
  any(colalpha(x$background.color) > 0)
}
vline_token <- function(w, cols, digits = 2) {
  size <- format_double(w, digits = digits)
  cols <- colcode0(cols)
  z <- sprintf("!{\\color[HTML]{%s}\\vrule width %spt}", cols, size)
  z
}

cline_token <- function(w, cols, digits = 2) {
  size <- format_double(w, digits = digits)
  colchar <- colcode0(cols)
  z <- sprintf("ascline{%spt}{%s}", w, colchar)
  z[w < .001 | colalpha(cols) < 1] <- "ascline{0pt}{FFFFFF}"
  z
}
hline_token <- function(w, cols, digits = 2) {
  size <- format_double(w, digits = digits)
  cols <- colcode0(cols)
  z <- sprintf(
    ">{\\arrayrulecolor[HTML]{%s}\\global\\arrayrulewidth=%spt}%s",
    cols,
    size,
    "-"
  )
  z
}

cline_intruction <- function(token) {
  rle_ <- rle(token)

  lengths <- rle_$lengths
  values <- rle_$values
  end_at <- cumsum(lengths)
  start_at <- end_at - (lengths - 1)

  keep <- !values %in% c("ascline{0pt}{FFFFFF}", "~")
  if (sum(keep) < 1) {
    return("")
  }

  values <- values[keep]
  start_at <- start_at[keep]
  end_at <- end_at[keep]

  paste(sprintf("\\%s{%.0f-%.0f}", values, start_at, end_at), collapse = "")
}

hhline_instruction <- function(
  x,
  pre_str = character(length(x)),
  post_str = character(length(x))
) {
  if (all(x %in% c("~", "|"))) {
    return("")
  }
  z <- paste0(pre_str, x, post_str)

  paste0("\\hhline{", paste0(z, collapse = ""), "}")
}

#' @export
#' @title Transform an rtables object into a flextable
#' @description produce a flextable from a `TableTree` or
#' `ElementaryTable` object produced with the
#' 'rtables' package.
#'
#' The conversion uses [formatters::matrix_form()] to extract the
#' formatted content, column spans, alignments, indentation
#' and footnotes, then maps them to flextable features.
#'
#' Indentation of row labels is rendered with left padding.
#'
#' Label rows (`LabelRow`) are displayed in bold in the first column.
#' Content rows (`ContentRow`) are displayed entirely in bold.
#'
#' When `LabelRow` groups exist, [paginate()] is applied so
#' that all rows belonging to the same group are kept together
#' on the same page in Word and RTF output.
#'
#' To paginate the resulting flextable into multiple pages, use
#' [split_to_pages()], [split_rows()], or [split_columns()] after
#' calling `as_flextable()`.
#' @param x a `TableTree` or `ElementaryTable` object produced
#'   by the 'rtables' package.
#' @param indent_padding base left padding in points per indentation
#'   level. Default is `4`.
#' @param ... unused arguments
#' @return a flextable object.
#' @seealso [split_to_pages()], [split_rows()], [split_columns()]
#' @examples
#' if (require("rtables", character.only = TRUE, quietly = TRUE)) {
#'   library(rtables)
#'
#'   lyt <- basic_table(title = "Demographic Summary") %>%
#'     split_cols_by("ARM") %>%
#'     split_rows_by("SEX") %>%
#'     analyze("AGE", afun = mean, format = "xx.x")
#'
#'   tbl <- build_table(lyt, DM)
#'   as_flextable(tbl)
#' }
as_flextable.TableTree <- function(
  x,
  indent_padding = 4,
  ...
) {
  rtables_to_flextable(
    x,
    indent_padding = indent_padding,
    ...
  )
}

#' @export
#' @rdname as_flextable.TableTree
as_flextable.ElementaryTable <- function(
  x,
  indent_padding = 4,
  ...
) {
  rtables_to_flextable(
    x,
    indent_padding = indent_padding,
    ...
  )
}


rtables_to_flextable <- function(
  x,
  indent_padding = 4,
  ...
) {
  stopifnot(
    requireNamespace("formatters", quietly = TRUE),
    requireNamespace("rtables", quietly = TRUE)
  )

  mf <- formatters::matrix_form(x, indent_rownames = TRUE)
  indent_size <- mf$indent_size
  str_mat <- formatters::mf_strings(mf)
  spans_mat <- formatters::mf_spans(mf)
  aligns_mat <- formatters::mf_aligns(mf)
  rinfo <- formatters::mf_rinfo(mf)
  nlh <- formatters::mf_nlheader(mf)

  nc <- ncol(str_mat)
  nr <- nrow(str_mat)
  n_body <- nr - nlh

  col_keys <- paste0("col_", seq_len(nc))

  header_rows <- seq_len(nlh)
  body_rows <- seq(nlh + 1, nr)

  base_padding <- get_flextable_defaults()$padding.left

  # --- body ---
  body_mat <- str_mat[body_rows, , drop = FALSE]
  body_df <- as.data.frame(body_mat, stringsAsFactors = FALSE)
  colnames(body_df) <- col_keys

  # strip leading spaces added by indent_rownames
  indents <- rinfo$indent
  body_df[[1]] <- trimws(body_df[[1]], which = "left")

  ft <- flextable(body_df, col_keys = col_keys)

  # apply left padding for indentation
  for (lvl in unique(indents[indents > 0])) {
    rows <- which(indents == lvl)
    ft <- padding(
      ft,
      i = rows,
      j = 1,
      padding.left = base_padding + lvl * indent_padding
    )
  }

  # --- header ---
  header_mat <- str_mat[header_rows, , drop = FALSE]

  # detect indent levels from leading spaces in column 1
  header_spaces <- nchar(header_mat[, 1]) -
    nchar(trimws(header_mat[, 1], which = "left"))
  header_indents <- header_spaces %/% indent_size
  header_mat[, 1] <- trimws(header_mat[, 1], which = "left")

  mapping <- as.data.frame(t(header_mat), stringsAsFactors = FALSE)
  colnames(mapping) <- paste0("V", seq_len(nlh))
  mapping$col_keys <- col_keys
  ft <- set_header_df(ft, mapping = mapping, key = "col_keys")

  # apply left padding for header indentation
  for (lvl in unique(header_indents[header_indents > 0])) {
    hdr_rows <- which(header_indents == lvl)
    ft <- padding(
      ft,
      i = hdr_rows,
      j = 1,
      padding.left = base_padding + lvl * indent_padding,
      part = "header"
    )
  }

  # --- header spans ---
  for (i in seq_len(nlh)) {
    spans_row <- spans_mat[i, ]
    j <- 1L
    while (j <= nc) {
      span <- spans_row[j]
      if (span > 1L) {
        ft <- merge_at(ft, part = "header", i = i, j = j:(j + span - 1L))
        j <- j + span
      } else {
        j <- j + 1L
      }
    }
  }

  # --- body spans ---
  for (i in seq_len(n_body)) {
    spans_row <- spans_mat[nlh + i, ]
    j <- 1L
    while (j <= nc) {
      span <- spans_row[j]
      if (span > 1L) {
        ft <- merge_at(ft, part = "body", i = i, j = j:(j + span - 1L))
        j <- j + span
      } else {
        j <- j + 1L
      }
    }
  }

  ft <- do.call(get_flextable_defaults()$theme_fun, list(ft))

  # --- alignments ---
  align_map <- c(
    "left" = "left",
    "right" = "right",
    "center" = "center",
    "decimal" = "right",
    "dec_right" = "right",
    "dec_left" = "left"
  )
  ft <- align(ft, align = align_map[aligns_mat[header_rows, ]], part = "header")
  ft <- align(ft, align = align_map[aligns_mat[body_rows, ]], part = "body")

  # --- style by node_class ---
  label_rows <- which(rinfo$node_class == "LabelRow")
  if (length(label_rows) > 0L) {
    ft <- bold(ft, i = label_rows, j = 1, part = "body")
    ft <- merge_h_range(ft, i = label_rows, j1 = 1, j2 = ncol_keys(ft))
  }

  # --- trailing separators (horizontal lines between groups) ---
  trailing_seps <- rinfo$trailing_sep
  sep_rows <- which(!is.na(trailing_seps))
  if (length(sep_rows) > 0L) {
    ft <- hline(ft, i = sep_rows, part = "body", border = fp_border_default())
  }

  # --- titles ---
  main_ttl <- formatters::main_title(mf)
  sub_ttls <- formatters::subtitles(mf)

  if (length(main_ttl) > 0L && any(nzchar(main_ttl))) {
    ft <- set_caption(ft, caption = paste(main_ttl, collapse = " "))
  }
  if (length(sub_ttls) > 0L && any(nzchar(sub_ttls))) {
    ft <- add_header_lines(ft, values = sub_ttls, top = TRUE)
  }

  # --- footnotes ---
  ref_fnotes <- formatters::mf_rfnotes(mf)
  if (length(ref_fnotes) > 0L) {
    ft <- add_footer_lines(ft, values = ref_fnotes)
  }

  main_ftr <- formatters::main_footer(mf)
  if (length(main_ftr) > 0L && any(nzchar(main_ftr))) {
    ft <- add_footer_lines(ft, values = main_ftr)
  }

  prov_ftr <- formatters::prov_footer(mf)
  if (length(prov_ftr) > 0L && any(nzchar(prov_ftr))) {
    ft <- add_footer_lines(ft, values = prov_ftr)
  }

  # --- finish ---
  ft <- autofit(ft)

  # --- paginate hints for Word/RTF ---
  if (length(label_rows) > 0L) {
    rows_grp_start <- intersect(which(indents < 1), label_rows)
    ft <- paginate(
      ft,
      hdr_ftr = TRUE,
      init = FALSE,
      group = rows_grp_start,
      group_def = "starts"
    )
  }

  ft
}

# main ----
#' @export
#' @title Convert a flextable to an HTML object
#'
#' @description get a [htmltools::div()] from a flextable object.
#' This can be used in a shiny application. For an output within
#' "R Markdown" document, use [knit_print.flextable].
#' @return an object marked as [htmltools::HTML] ready to be used within
#' a call to `shiny::renderUI` for example.
#' @inheritParams args_x_only
#' @param ft.align flextable alignment, supported values are 'left', 'center' and 'right'.
#' @param ft.shadow deprecated.
#' @param extra_dependencies a list of HTML dependencies to
#' add in the HTML output.
#' @family flextable_output_export
#' @examples
#' htmltools_value(flextable(iris[1:5, ]))
#' @importFrom htmltools tagList attachDependencies tags
htmltools_value <- function(
  x,
  ft.align = NULL,
  ft.shadow = NULL,
  extra_dependencies = NULL
) {
  x <- flextable_global$defaults$post_process_all(x)
  x <- flextable_global$defaults$post_process_html(x)
  x <- fix_border_issues(x)

  if (!is.null(ft.align)) {
    x$properties$align <- ft.align
  }

  caption <- caption_default_html(x)
  manual_css <- attr(caption, "css")

  list_deps <- html_dependencies_list(x)

  html_o <- tagList(
    if (length(extra_dependencies) > 0) {
      attachDependencies(
        x = tags$style(""),
        extra_dependencies
      )
    },
    attachDependencies(
      x = tags$style(""),
      list_deps
    ),
    HTML(gen_raw_html(
      x,
      class = "tabwid",
      caption = caption,
      manual_css = manual_css
    ))
  )
  html_o
}

#' @importFrom knitr knit_child
#' @export
#' @title Print a flextable inside knitr loops and conditionals
#'
#' @description Print flextable in R Markdown or Quarto documents
#' within `for` loop or `if` statement.
#'
#' The function is particularly useful when you want
#' to generate flextable in a loop from a R Markdown document.
#'
#' Inside R Markdown document, chunk option `results` must be
#' set to 'asis'.
#'
#' See [knit_print.flextable] for more details.
#' @inheritParams args_x_only
#' @param ... unused argument
#' @family flextable_output_export
#' @examples
#' \dontrun{
#' library(rmarkdown)
#' if (pandoc_available() &&
#'   pandoc_version() > numeric_version("2")) {
#'   demo_loop <- system.file(
#'     package = "flextable",
#'     "examples/rmd",
#'     "loop_with_flextable.Rmd"
#'   )
#'   rmd_file <- tempfile(fileext = ".Rmd")
#'   file.copy(demo_loop, to = rmd_file, overwrite = TRUE)
#'   render(
#'     input = rmd_file, output_format = "html_document",
#'     output_file = "loop_with_flextable.html"
#'   )
#' }
#' }
flextable_to_rmd <- function(x, ...) {
  is_bookdown <- is_in_bookdown()
  is_quarto <- is_in_quarto()

  x <- knitr_update_properties(x, bookdown = is_bookdown, quarto = is_quarto)

  if (is_quarto) {
    tmp_file <- tempfile(fileext = ".qmd")
  } else {
    tmp_file <- tempfile(fileext = ".Rmd")
  }

  writeLines(
    c("```{r echo=FALSE}", "x", "```", ""),
    tmp_file,
    useBytes = TRUE
  )

  z <- knit_child(
    input = tmp_file,
    options = list(
      fig.path = tempfile(),
      eval = isTRUE(knitr::opts_current$get("eval"))
    ),
    envir = environment(),
    quiet = TRUE
  )

  cat(z, sep = '\n')

  invisible("")
}

#' @export
#' @title Get HTML code as a string
#' @description Generate HTML code of corresponding
#' flextable as an HTML table or an HTML image.
#' @inheritParams args_x_only
#' @param type output type. one of "table" or "img".
#' @param ... unused
#' @return If `type='img'`, the result will be a string
#' containing HTML code of an image tag, otherwise, the
#' result will be a string containing HTML code of
#' a table tag.
#' @family flextable_output_export
#' @examples
#' library(officer)
#' library(flextable)
#' x <- to_html(as_flextable(cars))
to_html.flextable <- function(x, type = c("table", "img"), ...) {
  type_output <- match.arg(type)

  x <- knitr_update_properties(x)

  is_bookdown <- is_in_bookdown()
  is_quarto <- is_in_quarto()

  if ("table" %in% type_output) {
    x <- flextable_global$defaults$post_process_all(x)
    x <- flextable_global$defaults$post_process_html(x)
    x <- fix_border_issues(x)
    manual_css <- readLines(
      system.file(package = "flextable", "web_1.1.3", "tabwid.css"),
      encoding = "UTF-8"
    )
    str <- gen_raw_html(
      x,
      class = "tabwid",
      caption = "",
      manual_css = paste0(manual_css, collapse = "\n")
    )
    # `to_html()` returns a standalone string that does not carry html
    # dependencies; like `tabwid.css` above, the KaTeX stylesheet (math
    # fonts, issue #622) must travel inline with the markup.
    if (has_equation(x)) {
      katex_link <- katex_css_link()
      if (!is.null(katex_link)) {
        str <- paste0(katex_link, str)
      }
    }
    str
  } else {
    gr <- gen_grob(x, fit = "fixed", just = "center")
    dims <- dim(gr)
    expand <- 10
    width <- dims$width + expand / 72
    height <- dims$height + expand / 72

    tmp <- plot_in_png(
      code = plot(gr),
      width = width,
      height = height,
      res = 200,
      units = "in",
      path = tempfile(fileext = ".png")
    )

    base64_string <- image_to_base64(tmp)

    sprintf(
      "<img style=\"width:%.3fin;height:%.3fin;\" src=\"%s\" />",
      width,
      height,
      base64_string
    )
  }
}

#' @export
to_wml.flextable <- function(x, ...) {
  x <- flextable_global$defaults$post_process_all(x)
  x <- flextable_global$defaults$post_process_docx(x)
  x <- fix_border_issues(x)
  x <- knitr_update_properties(x)
  gen_raw_wml(x)
}


#' @noRd
#' @title flextable HTML string
#' @description get a string for HTML output with pandoc.
#' @inheritParams args_x_only
#' @param bookdown `TRUE` or `FALSE` (default) to support cross referencing with bookdown.
#' @param quarto `TRUE` or `FALSE` (default).
#' @examples
#' knit_to_html(flextable(iris[1:5, ]))
#' @importFrom knitr raw_html
knit_to_html <- function(x, bookdown = FALSE, quarto = FALSE) {
  x <- flextable_global$defaults$post_process_all(x)
  x <- flextable_global$defaults$post_process_html(x)
  x <- fix_border_issues(x)

  tab_props <- opts_current_table()
  topcaption <- tab_props$topcaption

  manual_css <- ""
  if (bookdown) {
    caption_str <- caption_bookdown_html(x)
    manual_css <- attr(caption_str, "css")
  } else if (quarto) {
    caption_str <- ""
  } else {
    caption_str <- caption_default_html(x)
    manual_css <- attr(caption_str, "css")
  }

  table_str <- gen_raw_html(
    x,
    caption = caption_str,
    topcaption = topcaption,
    manual_css = manual_css
  )

  table_str
}


#' @noRd
#' @title flextable Office Open XML string for Word
#' @description get openxml raw code for Word
#' from a flextable object.
#' @importFrom officer opts_current_table run_autonum to_wml
knit_to_wml <- function(x, bookdown = FALSE, quarto = FALSE) {
  x <- flextable_global$defaults$post_process_all(x)
  x <- flextable_global$defaults$post_process_docx(x)
  x <- fix_border_issues(x)

  for (part in c("body", "header", "footer")) {
    if (nrow_part(x, part) > 0L) {
      x[[part]]$styles$pars$word_style$data[,] <- NA_character_
    }
  }

  is_rdocx_document <- opts_current$get("is_rdocx_document")
  if (is.null(is_rdocx_document)) {
    is_rdocx_document <- FALSE
  }

  # images and hyperlinks are referenced by file path/URL in the raw
  # OOXML; only officer can register them in the docx package (when the
  # docx is written or re-written by officer). When pandoc writes the
  # final file (Quarto `format: docx`, rmarkdown `word_document`), the
  # references stay unresolved and Word refuses to open the document
  # (#711).
  if (!is_rdocx_document && (has_raster(x) || has_hlink(x))) {
    warning(
      "This flextable contains images or hyperlinks and the document ",
      "is not post-processed by officer: the resulting 'docx' file ",
      "will be invalid. Use `officedown::rdocx_document()` ",
      "(R Markdown) or repair the file after rendering with ",
      "`flextable::repair_docx()` (with Quarto, call it from a ",
      "post-render script).",
      call. = FALSE
    )
  }

  tab_props <- opts_current_table()
  topcaption <- tab_props$topcaption

  if (!is.null(tab_props$alt.title)) {
    x$properties$word_title <- tab_props$alt.title
  }
  if (!is.null(tab_props$alt.description)) {
    x$properties$word_description <- tab_props$alt.description
  }

  apply_cap_kwn <- FALSE
  if (topcaption) {
    apply_cap_kwn <- TRUE
  }

  word_autonum <- FALSE
  if (is_rdocx_document || quarto) {
    word_autonum <- TRUE
  }
  if (bookdown) {
    caption <- caption_bookdown_docx_md(x)
  } else if (quarto) {
    caption <- ""
  } else {
    caption <- caption_default_docx_openxml(
      x,
      keep_with_next = apply_cap_kwn,
      allow_autonum = word_autonum
    )
  }

  table_str <- gen_raw_wml(x = x)

  if (bookdown) {
    out <- c(
      if (topcaption) caption,
      with_openxml_quotes(table_str),
      if (!topcaption) caption
    )
  } else {
    out <- with_openxml_quotes(
      c(
        if (topcaption) caption,
        table_str,
        if (!topcaption) caption
      )
    )
  }

  out
}

#' @noRd
#' @title flextable latex string for PDF
#'
#' @description get latex raw code for PDF
#' from a flextable object.
#' @inheritParams args_x_only
#' @param ft.align flextable alignment, supported values are 'left', 'center' and 'right'.
#' @param tabcolsep space between the text and the left/right border of its containing
#' cell, the default value is 8 points.
#' @param ft.arraystretch height of each row relative to its default
#' height, the default value is 1.5.
#' @param ft.latex.float 'none' (the default value), 'float', 'wrap-r',
#' 'wrap-l', 'wrap-i', 'wrap-o'.
#' @param bookdown `TRUE` or `FALSE` (default) to support cross referencing with bookdown.
#' @examples
#' knit_to_latex(flextable(airquality[1:5, ]))
#' @importFrom officer opts_current_table run_autonum to_wml
#' @importFrom knitr raw_block raw_latex
knit_to_latex <- function(x, bookdown, quarto = FALSE) {
  align <- x$properties$align
  tabcolsep <- x$properties$opts_pdf$tabcolsep
  ft.arraystretch <- x$properties$opts_pdf$arraystretch
  ft.latex.float <- x$properties$opts_pdf$float

  x <- flextable_global$defaults$post_process_all(x)
  x <- flextable_global$defaults$post_process_pdf(x)
  x <- fix_border_issues(x)

  if ("none" %in% ft.latex.float) {
    lat_container <- latex_container_none()
  } else if ("float" %in% ft.latex.float) {
    lat_container <- latex_container_float()
  } else if ("wrap-l" %in% ft.latex.float) {
    lat_container <- latex_container_wrap(placement = "l")
  } else if ("wrap-r" %in% ft.latex.float) {
    lat_container <- latex_container_wrap(placement = "r")
  } else if ("wrap-i" %in% ft.latex.float) {
    lat_container <- latex_container_wrap(placement = "i")
  } else if ("wrap-o" %in% ft.latex.float) {
    lat_container <- latex_container_wrap(placement = "o")
  } else {
    lat_container <- latex_container_none()
  }

  tab_props <- opts_current_table()
  topcaption <- tab_props$topcaption
  if (quarto) {
    caption_str <- ""
  } else if (bookdown) {
    caption_str <- caption_bookdown_latex(x)
  } else {
    caption_str <- caption_default_latex(x)
  }

  container_str <- latex_container_str(
    x = x,
    latex_container = lat_container,
    quarto = quarto
  )

  str <- gen_raw_latex(
    x,
    lat_container = lat_container,
    caption = caption_str,
    topcaption = topcaption,
    quarto = quarto
  )
  latex <- paste(
    container_str[1],
    str,
    container_str[2],
    sep = "\n\n"
  )
  z <- paste(
    "\\global\\setlength{\\Oldarrayrulewidth}{\\arrayrulewidth}",
    "\\global\\setlength{\\Oldtabcolsep}{\\tabcolsep}",
    sprintf(
      "\\setlength{\\tabcolsep}{%spt}",
      format_double(x$properties$opts_pdf$tabcolsep, 0)
    ),
    sprintf(
      "\\renewcommand*{\\arraystretch}{%s}",
      format_double(x$properties$opts_pdf$arraystretch, 2)
    ),
    latex,
    sprintf(
      "\\arrayrulecolor[HTML]{%s}",
      colcode0(x$properties$opts_pdf$default_line_color)
    ),
    "\\global\\setlength{\\arrayrulewidth}{\\Oldarrayrulewidth}",
    "\\global\\setlength{\\tabcolsep}{\\Oldtabcolsep}",
    "\\renewcommand*{\\arraystretch}{1}",
    sep = "\n\n"
  )
  z
}

#' @importFrom knitr raw_block
knit_to_pml <- function(x) {
  left <- opts_current$get("ft.left")
  top <- opts_current$get("ft.top")

  if (is.null(left)) {
    left <- 1
  }
  if (is.null(top)) {
    top <- 2
  }
  x <- flextable_global$defaults$post_process_all(x)
  x <- flextable_global$defaults$post_process_pptx(x)
  x <- fix_border_issues(x)

  uid <- as.integer(runif(n = 1) * 10^9)

  str <- gen_raw_pml(x, uid = uid, offx = left, offy = top, cx = 10, cy = 6)
  with_openxml_quotes(str)
}

#' @noRd
#' @title flextable Typst string
#' @description get Typst raw code from a flextable object, to be
#' embedded in a Quarto document rendered with `format: typst`.
knit_to_typst <- function(x, bookdown = FALSE, quarto = FALSE) {
  x <- flextable_global$defaults$post_process_all(x)
  x <- fix_border_issues(x)

  str <- gen_raw_typst(x)

  # equations rely on the mitex Typst package; emit the import once,
  # right before a table that contains at least one equation. Users can
  # also declare it in the document preamble.
  if (has_equation(x)) {
    str <- paste0("#import \"@preview/mitex:0.2.7\": *\n", str)
  }

  str
}

#' @importFrom htmltools HTML browsable
#' @export
#' @title Print a flextable
#'
#' @description print a flextable object to format `html`, `docx`,
#' `pptx` or as text (not for display but for informative purpose).
#' This function is to be used in an interactive context.
#'
#' @note
#' When argument `preview` is set to `"docx"` or `"pptx"`, an
#' external client linked to these formats (Office is installed) is used to
#' edit a document. The document is saved in the temporary directory of
#' the R session and will be removed when R session will be ended.
#'
#' When argument `preview` is set to `"html"`, an
#' external client linked to these HTML format is used to display the table.
#' If RStudio is used, the Viewer is used to display the table.
#'
#' Note also that a print method is used when flextable are used within
#' R markdown documents. See [knit_print.flextable()].
#' @inheritParams args_x_only
#' @param preview preview type, one of c("html", "pptx", "docx", "rtf", "pdf, "log").
#' When `"log"` is used, a description of the flextable is printed.
#' @param align left, center (default) or right. Only for docx/html/pdf.
#' @param ... arguments for 'pdf_document' call when preview is "pdf".
#' @family flextable_output_export
#' @importFrom utils browseURL str
#' @importFrom rmarkdown render pdf_document
#' @importFrom officer read_pptx add_slide read_docx
print.flextable <- function(x, preview = "html", align = "center", ...) {
  if (is_in_pkgdown()) {
    return(htmltools_value(x, ft.align = align))
  } else if (!interactive() || "log" %in% preview) {
    cat("a flextable object.\n")
    cat("col_keys:", paste0("`", x$col_keys, "`", collapse = ", "), "\n")
    cat("header has", nrow(x$header$dataset), "row(s)", "\n")
    cat("body has", nrow(x$body$dataset), "row(s)", "\n")
    cat("original dataset sample:", "\n")
    str(x$body$dataset)
  } else if (preview == "html") {
    print(browsable(htmltools_value(x = x, ft.align = align)))
  } else if (preview == "pptx") {
    doc <- read_pptx()
    doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
    doc <- ph_with(doc, value = x, location = ph_location_type(type = "body"))
    file_out <- print(doc, target = tempfile(fileext = ".pptx"))
    browseURL(file_out)
  } else if (preview == "docx") {
    doc <- read_docx()
    doc <- body_add_flextable(doc, value = x, align = align)
    file_out <- print(doc, target = tempfile(fileext = ".docx"))
    browseURL(file_out)
  } else if (preview == "rtf") {
    file_out <- tempfile(fileext = ".rtf")
    save_as_rtf(x, path = file_out)
    browseURL(file_out)
  } else if (preview == "pdf") {
    rmd <- tempfile(fileext = ".Rmd")
    cat(
      sprintf("```{r echo=FALSE, ft.align=\"%s\"}\nx\n```\n", align),
      file = rmd
    )
    render(rmd, output_format = pdf_document(...), quiet = TRUE)
    file_out <- gsub("\\.Rmd$", ".pdf", rmd)
    browseURL(file_out)
  }

  invisible(NULL)
}


#' @title Render flextable in knitr documents
#' @description This function is called automatically by knitr to display
#' a flextable in R Markdown and Quarto documents. You do not need to call
#' it directly.
#'
#' Supported output formats: HTML, Word (docx), PDF, PowerPoint (pptx) and
#' Typst (`format: typst`, Quarto only). For other formats (e.g.,
#' `github_document`, `beamer_presentation`), the table is rendered as a
#' PNG image.
#'
#' @section Getting started:
#'
#' No special setup is needed: place a flextable object in a code chunk
#' and it will be rendered in the output document.
#'
#' Add a caption with [set_caption()]:
#'
#' ```r
#' ft <- set_caption(ft, caption = "My table caption")
#' ```
#'
#' In Quarto documents, use chunk options `tbl-cap` and `label` instead:
#'
#' ````
#' ```{r}
#' #| label: tbl-mytable
#' #| tbl-cap: "My table caption"
#' ft
#' ```
#' ````
#'
#' @section Captions:
#'
#' **Recommended method:** use [set_caption()] to define the caption
#' directly on the flextable object. When `set_caption()` is used,
#' chunk options related to captions are ignored.
#'
#' **Alternative (R Markdown only):** use knitr chunk options `tab.cap`
#' and `tab.id`:
#'
#' | **Description**                   |  **Chunk option** | **Default** |
#' |:----------------------------------|:-----------------:|:-----------:|
#' | Caption text                      | tab.cap           |    NULL     |
#' | Caption id/bookmark               | tab.id            |    NULL     |
#' | Caption on top of the table       | tab.topcaption    |    TRUE     |
#' | Caption sequence identifier       | tab.lp            |   "tab:"   |
#' | Word style for captions           | tab.cap.style     |    NULL     |
#'
#' **Bookdown:** cross-references use the pattern
#' `\@ref(tab:chunk_label)`. The usual bookdown numbering applies.
#'
#' **Quarto:** cross-references use `@tbl-chunk_label`. To embed
#' cross-references or other Quarto markdown inside flextable cells,
#' use [as_qmd()] with the `flextable-qmd` extension
#' (see [use_flextable_qmd()]).
#'
#' @section Chunk options:
#'
#' Use `knitr::opts_chunk$set(...)` to set defaults for the whole document.
#'
#' **All formats:**
#' - `ft.align`: table alignment, one of `'left'`, `'center'` (default)
#'   or `'right'`.
#'
#' **HTML:**
#' - `ft.htmlscroll`: `TRUE` or `FALSE` (default) to enable horizontal
#'   scrolling. See [set_table_properties()] for finer control.
#'
#' **Word:**
#' - `ft.split`: allow rows to break across pages (`TRUE` by default).
#'
#' **PDF:**
#' - `ft.tabcolsep`: space between text and cell borders (default 0 pt).
#' - `ft.arraystretch`: row height multiplier (default 1.5).
#' - `ft.latex.float`: floating placement. One of `'none'` (default),
#'   `'float'`, `'wrap-r'`, `'wrap-l'`, `'wrap-i'`, `'wrap-o'`.
#'
#' **PowerPoint:**
#' - `ft.left`, `ft.top`: top-left coordinates of the table placeholder
#'   in inches (defaults: 1 and 2).
#'
#' @section Word with officedown:
#'
#' When using `officedown::rdocx_document()`, additional caption options
#' are available:
#'
#' | **Description**                          |  **Chunk option** | **Default**               |
#' |:-----------------------------------------|:-----------------:|:-------------------------:|
#' | Numbering prefix                         | tab.cap.pre       | "Table "                  |
#' | Numbering suffix                         | tab.cap.sep       | ": "                      |
#' | Title number depth                       | tab.cap.tnd       | 0                         |
#' | Caption prefix formatting                | tab.cap.fp_text   | `fp_text_lite(bold=TRUE)` |
#' | Title number / table number separator    | tab.cap.tns       | "-"                       |
#'
#' @section Quarto:
#'
#' flextable works natively in Quarto documents for HTML, PDF, Word and
#' Typst.
#'
#' The `flextable-qmd` Lua filter extension enables Quarto markdown
#' inside flextable cells: cross-references (`@tbl-xxx`, `@fig-xxx`),
#' bold/italic, links, math, inline code and shortcodes.
#' See [as_qmd()] and [use_flextable_qmd()] for setup instructions.
#'
#' @section PDF accessibility (PDF/UA-2):
#'
#' Quarto's `pdf-standard: ua-2` injects `\DocumentMetadata{tagging=on}`
#' in the LaTeX preamble, activating 'tagpdf'.
#' This code patches LaTeX commands at compile time to insert
#' PDF structure tags. Neither Quarto nor flextable control this process.
#'
#' The tagging code does not yet support `longtable` + `colortbl`,
#' `booktabs` rules, and `cline`, which flextable relies on. Compilation
#' fails with `\ERRORtbl/row`. When these upstream issues are resolved,
#' flextable PDF output will support tagging without changes. Other formats are
#' not affected.
#'
#' @section PDF limitations:
#'
#' The following properties are not supported in PDF output:
#' `padding.top`, `padding.bottom`, `line_spacing` and row `height`.
#' Justified text is converted to left-aligned.
#'
#' To use system fonts, set `latex_engine: xelatex` in the YAML
#' header (the default `pdflatex` engine does not support them).
#'
#' See [add_latex_dep()] when caching flextable results.
#'
#' @section PowerPoint limitations:
#'
#' PowerPoint only supports fixed table layout. Use [autofit()] to
#' adjust column widths. Images inside table cells are not supported
#' (this is a PowerPoint limitation).
#'
#' @section HTML note:
#'
#' HTML output uses Shadow DOM to isolate table styles from the rest
#' of the page.
#'
#' @inheritParams args_x_only
#' @param ... unused.
#' @export
#' @importFrom utils getFromNamespace
#' @importFrom htmltools HTML div
#' @importFrom knitr knit_print asis_output opts_knit opts_current fig_path is_html_output is_latex_output include_graphics pandoc_to
#' @importFrom rmarkdown pandoc_version
#' @importFrom stats runif
#' @importFrom graphics plot par
#' @family flextable_output_export
#' @seealso [set_caption()], [as_qmd()], [use_flextable_qmd()],
#' [paginate()]
#' @examples
#' \dontrun{
#' library(rmarkdown)
#' if (pandoc_available() &&
#'   pandoc_version() > numeric_version("2")) {
#'   demo_loop <- system.file(
#'     package = "flextable",
#'     "examples/rmd",
#'     "demo.Rmd"
#'   )
#'   rmd_file <- tempfile(fileext = ".Rmd")
#'   file.copy(demo_loop, to = rmd_file, overwrite = TRUE)
#'   render(
#'     input = rmd_file, output_format = "html_document",
#'     output_file = "demo.html"
#'   )
#' }
#' }
knit_print.flextable <- function(x, ...) {
  is_bookdown <- is_in_bookdown()
  is_quarto <- is_in_quarto()

  x <- knitr_update_properties(x, bookdown = is_bookdown, quarto = is_quarto)

  if (is.null(pandoc_to())) {
    str <- to_html(x, type = "table")
    str <- asis_output(str)
  } else if (!is.null(getOption("xaringan.page_number.offset"))) {
    # xaringan
    str <- knit_to_html(x, bookdown = FALSE, quarto = FALSE)
    str <- asis_output(str, meta = html_dependencies_list(x))
  } else if (
    is_html_output(excludes = "gfm") &&
      isTRUE(knitr::opts_knit$get("is.paged.js"))
  ) {
    x$properties$opts_html$extra_class <- c(
      x$properties$opts_html$extra_class,
      "no-shadow-dom"
    )
    str <- knit_to_html(x, bookdown = FALSE, quarto = is_quarto)
    str <- raw_html(str, meta = html_dependencies_list(x))
  } else if (is_html_output(excludes = "gfm")) {
    # html
    str <- knit_to_html(x, bookdown = is_bookdown, quarto = is_quarto)
    str <- raw_html(str, meta = html_dependencies_list(x))
  } else if (is_latex_output()) {
    # latex
    str <- knit_to_latex(x, bookdown = is_bookdown, quarto = is_quarto)
    str <- raw_latex(
      x = str,
      meta = unname(list_latex_dep(float = TRUE, wrapfig = TRUE))
    )
  } else if (grepl("docx", opts_knit$get("rmarkdown.pandoc.to"))) {
    # docx
    if (pandoc_version() < numeric_version("2")) {
      stop("pandoc version >= 2 required for printing flextable in docx")
    }
    str <- knit_to_wml(x, bookdown = is_bookdown, quarto = is_quarto)
    str <- asis_output(str)
  } else if (grepl("pptx", opts_knit$get("rmarkdown.pandoc.to"))) {
    # pptx
    if (pandoc_version() < numeric_version("2.4")) {
      stop("pandoc version >= 2.4 required for printing flextable in pptx")
    }
    str <- knit_to_pml(x)
    str <- asis_output(str)
  } else if (grepl("typst", opts_knit$get("rmarkdown.pandoc.to"))) {
    # typst (Quarto `format: typst`)
    str <- knit_to_typst(x, bookdown = is_bookdown, quarto = is_quarto)
    str <- raw_block(str, "typst")
  } else {
    plot_counter <- getFromNamespace("plot_counter", "knitr")
    in_base_dir <- getFromNamespace("in_base_dir", "knitr")
    tmp <- fig_path("png", number = plot_counter())
    in_base_dir({
      dir.create(dirname(tmp), showWarnings = FALSE, recursive = TRUE)
      save_as_image(x, path = tmp, expand = 0)
    })
    str <- include_graphics(tmp)
  }
  str
}

#' @export
#' @title Save flextable objects in an 'HTML' file
#' @description save a flextable in an 'HTML' file. This function
#' is useful to save the flextable in 'HTML' file without using
#' R Markdown (it is highly recommanded to use R Markdown
#' instead).
#' @param ... flextable objects, objects, possibly named. If named objects, names are
#' used as titles.
#' @param values a list (possibly named), each element is a flextable object. If named objects, names are
#' used as titles. If provided, argument `...` will be ignored.
#' @param path HTML file to be created
#' @param title page title
#' @param lang language of the document using IETF language tags
#' @return a string containing the full name of the generated file
#' @examples
#' ft1 <- flextable(head(iris))
#' tf1 <- tempfile(fileext = ".html")
#' if (rmarkdown::pandoc_available()) {
#'   save_as_html(ft1, path = tf1)
#'   # browseURL(tf1)
#' }
#'
#' ft2 <- flextable(head(mtcars))
#' tf2 <- tempfile(fileext = ".html")
#' if (rmarkdown::pandoc_available()) {
#'   save_as_html(
#'     `iris table` = ft1,
#'     `mtcars table` = ft2,
#'     path = tf2,
#'     title = "rhoooo"
#'   )
#'   # browseURL(tf2)
#' }
#' @family flextable_output_export
#' @importFrom htmltools save_html
save_as_html <- function(
  ...,
  values = NULL,
  path,
  lang = "en",
  title = "&#32;"
) {
  if (is.null(values)) {
    values <- list(...)
  }
  values <- Filter(function(x) inherits(x, "flextable"), values)
  values <- lapply(values, flextable_global$defaults$post_process_all)
  values <- lapply(values, flextable_global$defaults$post_process_html)
  values <- lapply(values, fix_border_issues)
  values <- lapply(values, htmltools_value)
  titles <- names(values)
  show_names <- !is.null(titles)
  if (show_names) {
    old_values <- values
    values <- rep(list(NULL), length(old_values) * 2)
    values[seq_along(old_values) * 2] <- old_values
    values[seq_along(old_values) * 2 - 1] <- lapply(titles, htmltools::h2)
  }
  values <- do.call(htmltools::tagList, values)

  is_succes <- render_htmltag(
    x = values,
    path = absolute_path(path),
    title = title,
    lang = lang
  )
  if (!is_succes) {
    stop("could not write the file ", shQuote(path))
  }
  invisible(path)
}


#' @export
#' @importFrom officer ph_location_type
#' @title Save flextable objects in a 'PowerPoint' file
#' @description sugar function to save flextable objects in
#' an PowerPoint file.
#'
#' This feature is available to simplify the work of users by avoiding
#' the need to use the 'officer' package. If it doesn't suit your needs,
#' then use the API offered by 'officer' which allows simple and
#' complicated things.
#' @note
#' The PowerPoint format ignores captions (see [set_caption()]).
#' @param ... flextable objects, objects, possibly named. If named objects, names are
#' used as slide titles.
#' @param values a list (possibly named), each element is a flextable object. If named objects, names are
#' used as slide titles. If provided, argument `...` will be ignored.
#' @param path PowerPoint file to be created
#' @return a string containing the full name of the generated file
#' @examples
#' ft1 <- flextable(head(iris))
#' tf <- tempfile(fileext = ".pptx")
#' save_as_pptx(ft1, path = tf)
#'
#' ft2 <- flextable(head(mtcars))
#' tf <- tempfile(fileext = ".pptx")
#' save_as_pptx(`iris table` = ft1, `mtcars table` = ft2, path = tf)
#' @family flextable_output_export
save_as_pptx <- function(..., values = NULL, path) {
  if (is.null(values)) {
    values <- list(...)
  }

  values <- Filter(function(x) inherits(x, "flextable"), values)
  titles <- names(values)
  show_names <- !is.null(titles)
  z <- read_pptx()
  for (i in seq_along(values)) {
    z <- add_slide(z, "Title and Content")
    if (show_names) {
      z <- ph_with(z, titles[i], location = ph_location_type(type = "title"))
    }
    z <- ph_with(z, values[[i]], location = ph_location_type(type = "body"))
  }
  print(z, target = path)
  invisible(path)
}


#' @export
#' @title Save flextable objects in a 'Word' file
#' @description sugar function to save flextable objects in an Word file.
#' @param ... flextable objects, objects, possibly named. If named objects, names are
#' used as titles.
#' @param values a list (possibly named), each element is a flextable object. If named objects, names are
#' used as titles. If provided, argument `...` will be ignored.
#' @param path Word file to be created
#' @param pr_section a [officer::prop_section] object that can be used to define page
#' layout such as orientation, width and height.
#' @param align left, center (default) or right.
#' @return a string containing the full name of the generated file
#' @examples
#'
#' tf <- tempfile(fileext = ".docx")
#'
#' library(officer)
#' ft1 <- flextable(head(iris))
#' save_as_docx(ft1, path = tf)
#'
#'
#' ft2 <- flextable(head(mtcars))
#' sect_properties <- prop_section(
#'   page_size = page_size(
#'     orient = "landscape",
#'     width = 8.3, height = 11.7
#'   ),
#'   type = "continuous",
#'   page_margins = page_mar()
#' )
#' save_as_docx(
#'   `iris table` = ft1, `mtcars table` = ft2,
#'   path = tf, pr_section = sect_properties
#' )
#' @family flextable_output_export
#' @seealso [paginate()]
#' @importFrom officer body_add_par prop_section body_set_default_section page_size page_mar
save_as_docx <- function(
  ...,
  values = NULL,
  path,
  pr_section = NULL,
  align = "center"
) {
  if (is.null(values)) {
    values <- list(...)
  }

  values <- Filter(function(x) inherits(x, "flextable"), values)

  titles <- names(values)
  show_names <- !is.null(titles)

  if (is.null(pr_section)) {
    pr_section <- prop_section(
      type = "continuous"
    )
  }

  if (!inherits(pr_section, "prop_section")) {
    stop("pr_section is not a prop_section object, use officer::prop_section.")
  }

  z <- read_docx()

  for (i in seq_along(values)) {
    if (show_names) {
      z <- body_add_par(z, titles[i], style = "heading 2")
    } else {
      z <- body_add_par(z, "")
    }
    z <- body_add_flextable(z, values[[i]], align = align)
  }
  z <- body_set_default_section(z, pr_section)
  print(z, target = path)
  invisible(path)
}

#' @export
#' @title Repair flextables in a 'Word' document produced by 'pandoc'
#' @description
#' Word documents produced by Quarto (`format: docx`) or by
#' `rmarkdown::word_document()` are written by 'pandoc', which copies the
#' raw OOXML of flextable objects verbatim: images and hyperlinks are
#' left as unresolved references (file paths and URLs instead of
#' relationship identifiers) and **Word refuses to open the file**.
#'
#' This function reads such a document and rewrites it with the package
#' 'officer', whose writer resolves these references: images and SVG
#' files are embedded, hyperlinks are registered, and the other
#' officer-flavored contents are processed as well (sections, footnotes,
#' comments, custom style names, list markers, documents poured with
#' `officer::block_pour_docx()`).
#'
#' It is not needed when the document is produced with
#' `officedown::rdocx_document()` or [save_as_docx()], where the file is
#' already written by 'officer'.
#'
#' @section What to do with Quarto:
#'
#' The repair must run after `quarto render`. Quarto can do it for you
#' with a *post-render script*, but this can **only** be declared in a
#' `_quarto.yml` file - the `project:` key is ignored in the YAML header
#' of a `.qmd` file.
#'
#' 1. Next to your `.qmd` file, create a file `_quarto.yml` containing:
#'
#' ```yaml
#' project:
#'   post-render: repair.R
#' ```
#'
#' 2. In the same directory, create the file `repair.R` containing:
#'
#' ```r
#' files <- strsplit(Sys.getenv("QUARTO_PROJECT_OUTPUT_FILES"), "\n")[[1]]
#' for (f in files[grepl("\\.docx$", files, ignore.case = TRUE)]) {
#'   flextable::repair_docx(f)
#' }
#' ```
#'
#' 3. Render as usual, either the whole project (`quarto render`) or a
#' single file (`quarto render doc.qmd`); the script runs in both cases
#' and only processes the files that were rendered.
#'
#' Without a `_quarto.yml`, call `repair_docx("doc.docx")` manually
#' after each render.
#'
#' @section What to do with R Markdown:
#'
#' Prefer `officedown::rdocx_document()` as output format: the file is
#' post-processed by 'officer' and no repair is needed. With
#' `rmarkdown::word_document()`, call
#' `repair_docx()` on the output file after `rmarkdown::render()`.
#'
#' @section Requirements and limits:
#'
#' - The images referenced in the document must still exist, at paths
#' resolvable **from the directory where `repair_docx()` runs**: a
#' Quarto post-render script runs at the project root, so images
#' referenced with paths relative to a sub-directory document will not
#' be found - use absolute paths (e.g. `here::here()`) in that case.
#' - Mini graphics generated by [minibar()], [linerange()],
#' [gg_chunk()], [plot_chunk()] and [grid_chunk()] are written to
#' temporary 'PNG' files that no longer exist once rendering has
#' completed: **these can not be repaired afterwards**. With Quarto or
#' `rmarkdown::word_document()`, use image files you manage yourself
#' ([as_image()], [colformat_image()]); mini graphics require
#' `officedown::rdocx_document()` or [save_as_docx()].
#' - When a referenced image can not be found, the rewrite fails with
#' an error `cannot open the connection` and a warning naming the
#' missing file.
#'
#' @param path path of the 'docx' file to repair. The file is modified
#' in place unless `target` is provided.
#' @param target path of the resulting 'docx' file. Defaults to `path`.
#' @return the path of the repaired file, invisibly.
#' @examples
#' \dontrun{
#' repair_docx("document.docx")
#' }
#' @family flextable_output_export
repair_docx <- function(path, target = path) {
  if (!file.exists(path) || !grepl("\\.docx$", path, ignore.case = TRUE)) {
    stop("`path` must be an existing 'docx' file.", call. = FALSE)
  }
  z <- read_docx(path)
  print(z, target = target)
  invisible(target)
}

#' @export
#' @title Save flextable objects in an 'RTF' file
#' @description sugar function to save flextable objects in an 'RTF' file.
#' @param ... flextable objects, objects, possibly named. If named objects, names are
#' used as titles.
#' @param values a list (possibly named), each element is a flextable object. If named objects, names are
#' used as titles. If provided, argument `...` will be ignored.
#' @param path Word file to be created
#' @param pr_section a [officer::prop_section] object that can be used to define page
#' layout such as orientation, width and height.
#' @return a string containing the full name of the generated file
#' @family flextable_output_export
#' @seealso [paginate()]
#' @examples
#'
#' tf <- tempfile(fileext = ".rtf")
#'
#' library(officer)
#' ft1 <- flextable(head(iris))
#' save_as_rtf(ft1, path = tf)
#'
#'
#' ft2 <- flextable(head(mtcars))
#' sect_properties <- prop_section(
#'   page_size = page_size(
#'     orient = "landscape",
#'     width = 8.3, height = 11.7
#'   ),
#'   type = "continuous",
#'   page_margins = page_mar(),
#'   header_default = block_list(
#'     fpar(ftext("text for default page header")),
#'     qflextable(data.frame(a = 1L))
#'   )
#' )
#' tf <- tempfile(fileext = ".rtf")
#' save_as_rtf(
#'   `iris table` = ft1, `mtcars table` = ft2,
#'   path = tf, pr_section = sect_properties
#' )
#' @importFrom officer rtf_doc fp_text_lite fpar
save_as_rtf <- function(..., values = NULL, path, pr_section = NULL) {
  if (is.null(values)) {
    values <- list(...)
  }

  values <- Filter(function(x) inherits(x, "flextable"), values)

  titles <- names(values)
  show_names <- !is.null(titles)
  if (is.null(pr_section)) {
    pr_section <- prop_section(
      type = "continuous"
    )
  }

  if (!inherits(pr_section, "prop_section")) {
    stop("pr_section is not a prop_section object, use officer::prop_section.")
  }

  z <- rtf_doc(def_sec = pr_section)

  for (i in seq_along(values)) {
    if (show_names) {
      z <- rtf_add(
        z,
        value = fpar(
          titles[i],
          fp_p = fp_par(
            text.align = "left",
            padding.top = 6,
            padding.bottom = 6
          ),
          fp_t = fp_text_lite(
            bold = TRUE,
            font.size = 18,
            font.family = "Arial"
          )
        )
      )
    }
    z <- rtf_add(z, values[[i]])
  }
  print(z, target = path)
  invisible(path)
}


#' @export
#' @title Save a flextable in a 'png' or 'svg' file
#' @description Save a flextable as a png or svg image.
#' This function uses R graphic system to create an image from the flextable,
#' allowing for high-quality image output. See [gen_grob()] for more options.
#'
#' @section caption:
#' It's important to note that captions are not part of the table itself.
#' This means when exporting a table to PNG or SVG formats (image formats),
#' the caption won't be included. Captions are intended for document outputs
#' like Word, HTML, or PDF, where tables are embedded within the document
#' itself.
#' @inheritParams args_x_only
#' @param path image file to be created. It should end with '.png'
#' or '.svg'.
#' @param expand space in pixels to add around the table.
#' @param res The resolution of the device
#' @param ... unused arguments
#' @return a string containing the full name of the generated file
#' @examples
#' library(gdtools)
#' register_liberationsans()
#' set_flextable_defaults(font.family = "Liberation Sans")
#'
#' ft <- flextable(head(mtcars))
#' ft <- autofit(ft)
#' tf <- tempfile(fileext = ".png")
#' save_as_image(x = ft, path = tf)
#'
#' init_flextable_defaults()
#' @family flextable_output_export
#' @importFrom ragg agg_png agg_capture
save_as_image <- function(x, path, expand = 10, res = 200, ...) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      as.character(sys.call()[[1]])
    ))
  }

  file_ext_pos <- regexpr("\\.([[:alnum:]]+)$", path)
  file_ext <- substring(path, file_ext_pos + 1L)

  file_ext <- match.arg(
    tolower(file_ext),
    choices = c("png", "svg"),
    several.ok = FALSE
  )

  gr <- gen_grob(x, fit = "fixed", just = "center")
  dims <- dim(gr)

  if (file_ext %in% "png") {
    plot_in_png(
      code = base::plot(gr),
      width = dims$width + expand / 72,
      height = dims$height + expand / 72,
      res = res,
      units = "in",
      path = path
    )
  } else if (file_ext %in% "svg") {
    if (!requireNamespace("svglite", quietly = TRUE)) {
      stop(sprintf(
        "'%s' package should be installed to save a flextable in a '%s' image.",
        "svglite",
        "svg"
      ))
    }
    svglite::svglite(
      filename = path,
      width = dims$width + expand / 72,
      height = dims$height + expand / 72,
      bg = "transparent"
    )
    tryCatch(
      {
        base::plot(gr)
      },
      finally = {
        dev.off()
      }
    )
  }

  path
}


#' @export
#' @title Plot a flextable
#' @description plots a flextable as a grid grob object
#' and display the result in a new graphics window.
#' 'ragg' or 'svglite' or 'ggiraph' graphical device drivers
#' should be used to ensure a correct rendering.
#' @inheritParams args_x_only
#' @param ... additional arguments passed to [gen_grob()].
#' @inheritSection save_as_image caption
#' @examples
#' library(gdtools)
#' library(ragg)
#' register_liberationsans()
#' set_flextable_defaults(font.family = "Liberation Sans")
#' ftab <- as_flextable(cars)
#'
#' \dontshow{
#' cap <- ragg::agg_capture(width = 7, height = 6, units = "in", res = 150)
#' grDevices::dev.control("enable")
#' }
#' plot(ftab)
#' \dontshow{
#' raster <- cap()
#' dev.off()
#' plot(as.raster(raster))
#' init_flextable_defaults()
#' }
#' @family flextable_output_export
#' @importFrom grid grid.newpage grid.draw viewport pushViewport popViewport
plot.flextable <- function(x, ...) {
  grid.newpage()
  # leave a 5pt margin around the plot
  pushViewport(viewport(
    width = unit(1, "npc") - unit(10, "pt"),
    height = unit(1, "npc") - unit(10, "pt")
  ))
  grid.draw(gen_grob(x, ...))
  popViewport()
  invisible()
}


# utils ----
is_in_bookdown <- function() {
  is_rdocx_document <- opts_current$get("is_rdocx_document")
  if (is.null(is_rdocx_document)) {
    is_rdocx_document <- FALSE
  }

  isTRUE(opts_knit$get("bookdown.internal.label")) &&
    isTRUE(!is_rdocx_document)
}
is_in_quarto <- function() {
  if (getRversion() >= numeric_version("4.4.0")) {
    isTRUE(knitr::opts_knit$get("quarto.version") > 0)
  } else {
    isTRUE(knitr::opts_knit$get("quarto.version") > numeric_version("0"))
  }
}
fake_quarto <- function() {
  if (getRversion() >= numeric_version("4.4.0")) {
    knitr::opts_knit$set("quarto.version" = 1)
  } else {
    knitr::opts_knit$set("quarto.version" = numeric_version("1.0"))
  }
}
is_in_pkgdown <- function() {
  identical(Sys.getenv("IN_PKGDOWN"), "true") &&
    requireNamespace("pkgdown", quietly = TRUE)
}

#' @importFrom knitr opts_current
knitr_update_properties <- function(x, bookdown = FALSE, quarto = FALSE) {
  # global properties
  ft.align <- opts_current$get("ft.align")
  ft.split <- opts_current$get("ft.split")
  ft.tabcolsep <- opts_current$get("ft.tabcolsep")
  ft.arraystretch <- opts_current$get("ft.arraystretch")
  ft.latex.float <- mcoalesce_options(
    opts_current$get("ft.latex.float"),
    opts_current$get("ft-latex-float")
  )
  ft.left <- opts_current$get("ft.left")
  ft.top <- opts_current$get("ft.top")
  ft.htmlscroll <- opts_current$get("ft.htmlscroll")

  if (!is.null(ft.align)) {
    x$properties$align <- ft.align
  }
  if (isTRUE(ft.htmlscroll)) {
    x$properties$opts_html$scroll <- list()
  }
  # word chunk options
  if (!is.null(ft.split)) {
    x$properties$opts_word$split <- ft.split
  }
  # latex chunk options
  if (!is.null(ft.tabcolsep)) {
    x$properties$opts_pdf$tabcolsep <- ft.tabcolsep
  }
  if (!is.null(ft.arraystretch)) {
    x$properties$opts_pdf$arraystretch <- ft.arraystretch
  }
  if (!is.null(ft.latex.float)) {
    x$properties$opts_pdf$float <- ft.latex.float
  }

  # captions properties
  knitr_tab_opts <- opts_current_table()
  if (is.null(knitr_tab_opts$cap.style)) {
    knitr_tab_opts$cap.style <- "Table Caption"
  }
  x <- update_caption(
    x = x,
    caption = knitr_tab_opts$cap,
    word_stylename = knitr_tab_opts$cap.style
  )
  if (
    has_caption(x) &&
      !has_autonum(x) &&
      !is.null(knitr_tab_opts$id)
  ) {
    autonum <- run_autonum(
      seq_id = gsub(":$", "", knitr_tab_opts$tab.lp),
      pre_label = knitr_tab_opts$cap.pre,
      post_label = knitr_tab_opts$cap.sep,
      bkm = knitr_tab_opts$id,
      bkm_all = FALSE,
      tnd = knitr_tab_opts$cap.tnd,
      tns = knitr_tab_opts$cap.tns,
      prop = knitr_tab_opts$cap.fp_text
    )
    x <- update_caption(x = x, autonum = autonum)
  }

  x
}

#' @export
#' @title Add flextable into a Word document
#' @description Add a flextable into a Word document
#' created with 'officer'.
#' @family officer_integration
#'
#' @details
#' Use the [paginate()] function to define whether the table should
#' be displayed on one or more pages, and whether the header should be
#' displayed with the first lines of the table body on the same page.
#'
#' Use the [set_caption()] function to define formatted captions
#' (with [as_paragraph()]) or simple captions (with a string).
#' `topcaption` can be used to insert the caption before the table
#' (default) or after the table (use `FALSE`).
#' @param x an rdocx object
#' @param value `flextable` object
#' @param align left, center (default) or right.
#' The `align` parameter is still supported for the time being, but
#' we recommend using `set_flextable_defaults(table_align = "center")` instead
#' that will set this default alignment for all flextables during
#' the R session, or to define alignement for each table with
#' `set_table_properties(align = "center")`.
#' @param split set to TRUE if you want to activate Word
#' option 'Allow row to break across pages'.
#' This argument is still supported for the time being, but
#' we recommend using `set_flextable_defaults(split = TRUE)` instead
#' that will set this as default setting for all flextables during
#' the R session, or to define alignement for each table with
#' [set_table_properties()] with argument `opts_word=list(split = TRUE)`
#' instead.
#' @param keepnext Defunct in favor of [paginate()]. The default value
#' used for `keep_with_next` is set with
#' `set_flextable_defaults(keep_with_next = TRUE)`.
#' @param pos where to add the flextable relative to the cursor,
#' one of "after", "before", "on" (end of line).
#' @param topcaption if TRUE caption is added before the table, if FALSE,
#' caption is added after the table.
#' @importFrom officer body_add_xml
#' @seealso [knit_print.flextable()], [save_as_docx()]
#' @examples
#' \dontshow{
#' init_flextable_defaults()
#' }
#' library(officer)
#'
#' # define global settings
#' set_flextable_defaults(
#'   split = TRUE,
#'   table_align = "center",
#'   table.layout = "autofit"
#' )
#'
#' # produce 3 flextable
#' ft_1 <- flextable(head(airquality, n = 20))
#' ft_1 <- color(ft_1, i = ~ Temp > 70, color = "red", j = "Temp")
#' ft_1 <- highlight(ft_1, i = ~ Wind < 8, color = "yellow", j = "Wind")
#' ft_1 <- set_caption(
#'   x = ft_1,
#'   autonum = run_autonum(seq_id = "tab"),
#'   caption = "Daily air quality measurements"
#' )
#' ft_1 <- paginate(ft_1, init = TRUE, hdr_ftr = TRUE)
#'
#' ft_2 <- proc_freq(mtcars, "vs", "gear")
#' ft_2 <- set_caption(
#'   x = ft_2,
#'   autonum = run_autonum(seq_id = "tab", bkm = "mtcars"),
#'   caption = as_paragraph(
#'     as_b("mtcars"), " ",
#'     colorize("table", color = "orange")
#'   ),
#'   fp_p = fp_par(keep_with_next = TRUE)
#' )
#' ft_2 <- paginate(ft_2, init = TRUE, hdr_ftr = TRUE)
#'
#' ft_3 <- summarizor(iris, by = "Species")
#' ft_3 <- as_flextable(ft_3, spread_first_col = TRUE)
#' ft_3 <- set_caption(
#'   x = ft_3,
#'   autonum = run_autonum(seq_id = "tab"),
#'   caption = "iris summary"
#' )
#' ft_3 <- paginate(ft_3, init = TRUE, hdr_ftr = TRUE)
#'
#' # add the 3 flextable in a new Word document
#' doc <- read_docx()
#' doc <- body_add_flextable(doc, value = ft_1)
#' doc <- body_add_par(doc, value = "")
#' doc <- body_add_flextable(doc, value = ft_2)
#' doc <- body_add_par(doc, value = "")
#' doc <- body_add_flextable(doc, value = ft_3)
#'
#' fileout <- tempfile(fileext = ".docx")
#' print(doc, target = fileout)
#' \dontshow{
#' init_flextable_defaults()
#' }
body_add_flextable <- function(
  x,
  value,
  align = NULL,
  pos = "after",
  split = NULL,
  topcaption = TRUE,
  keepnext = NULL
) {
  stopifnot(
    inherits(x, "rdocx"),
    inherits(value, "flextable")
  )

  if (!is.null(align)) {
    value$properties$align <- align
  }
  if (!is.null(split)) {
    value$properties$opts_word$split <- split
  }

  value <- flextable_global$defaults$post_process_all(value)
  value <- flextable_global$defaults$post_process_docx(value)
  value <- fix_border_issues(value)

  caption_str <- NULL
  if (has_caption(value)) {
    if (topcaption) {
      apply_cap_kwn <- TRUE
    } else {
      value <- keep_with_next(value, part = "all", value = TRUE)
      apply_cap_kwn <- FALSE
    }
    caption_str <- caption_default_docx_openxml(
      x = value,
      keep_with_next = apply_cap_kwn,
      allow_autonum = TRUE
    )
    if ("" %in% caption_str) caption_str <- NULL
  }

  if (topcaption && !is.null(caption_str)) {
    x <- body_add_xml(x = x, str = caption_str, pos = pos)
  }
  out <- gen_raw_wml(value, doc = x)

  x <- body_add_xml(x = x, str = out, pos = pos)

  if (!topcaption && !is.null(caption_str)) {
    x <- body_add_xml(x = x, str = caption_str, pos = pos)
  }

  x
}

#' @export
#' @title Add flextable at bookmark location in a Word document
#' @family officer_integration
#' @description
#' Use this function if you want to replace a paragraph containing
#' a bookmark with a flextable. As a side effect, the bookmark will be lost.
#' @param x an rdocx object
#' @param value `flextable` object
#' @param bookmark bookmark id
#' @param align left, center (default) or right.
#' @param split set to TRUE if you want to activate Word
#' option 'Allow row to break across pages'.
#' @importFrom officer cursor_bookmark
#' @importFrom xml2 as_xml_document
body_replace_flextable_at_bkm <- function(
  x,
  bookmark,
  value,
  align = "center",
  split = FALSE
) {
  x <- cursor_bookmark(x, bookmark)
  x <- body_add_flextable(
    x = x,
    value = value,
    pos = "on",
    align = align,
    split = split
  )
  x
}

#' @export
#' @title Constrain table width by shrinking font size
#' @description Decrease font size incrementally until the table fits
#' within `max_width`.
#'
#' To constrain width by wrapping text instead, see [fit_columns()].
#' To size columns to their content without a width constraint,
#' see [autofit()].
#' @inheritParams args_x_only
#' @param max_width maximum width to fit in inches
#' @param inc the font size decrease for each step
#' @param max_iter maximum iterations
#' @param unit unit for max_width, one of "in", "cm", "mm".
#' @examples
#' ft_1 <- qflextable(head(mtcars))
#' ft_1 <- width(ft_1, width = 1)
#' ft_1
#'
#' ft_2 <- fit_to_width(ft_1, max_width = 4)
#' ft_2
#' @family functions for flextable size management
fit_to_width <- function(x, max_width, inc = 1L, max_iter = 20, unit = "in") {
  max_width <- convin(unit = unit, x = max_width)

  for (i in seq_len(max_iter)) {
    fdim <- flextable_dim(x)
    do_stop <- FALSE
    if (fdim$widths > max_width) {
      if (nrow_part(x, part = "body") > 0) {
        new_font_sizes <- x$body$styles$text$font.size$data - inc
        if (all(new_font_sizes > 0)) {
          x$body$styles$text$font.size$data[] <- new_font_sizes
        } else {
          do_stop <- TRUE
        }
      }
      if (nrow_part(x, part = "footer") > 0) {
        new_font_sizes <- x$footer$styles$text$font.size$data - inc
        if (all(new_font_sizes > 0)) {
          x$footer$styles$text$font.size$data[] <- new_font_sizes
        } else {
          do_stop <- TRUE
        }
      }
      if (nrow_part(x, part = "header") > 0) {
        new_font_sizes <- x$header$styles$text$font.size$data - inc
        if (all(new_font_sizes > 0)) {
          x$header$styles$text$font.size$data[] <- new_font_sizes
        } else {
          do_stop <- TRUE
        }
      }
      if (do_stop) {
        warning(
          "The function has to be stopped because it results in negative font sizes."
        )
        return(x)
      }
      x <- autofit(x, add_w = 0.0, add_h = 0.0)
    } else {
      return(x)
    }
  }
  x
}


#' @export
#' @title Constrain table width by wrapping text
#' @description Shrink column widths so that the total table width
#' does not exceed `max_width`. Font sizes are unchanged; text wraps
#' inside the narrower cells.
#'
#' Columns that cannot shrink below their longest word (plus padding)
#' are clamped at that floor; remaining space is iteratively
#' redistributed among unclamped columns.
#'
#' Columns listed in `no_wrap` keep their optimal width (as computed
#' by [dim_pretty()]) and are not compressed.
#'
#' To constrain width by reducing font size instead, see
#' [fit_to_width()]. To size columns to their content without a
#' width constraint, see [autofit()].
#' @inheritParams args_x_only
#' @param max_width maximum total table width (in inches by default)
#' @param no_wrap column names or indices that must keep their
#'   optimal width (no compression, no text wrapping).
#' @param unit unit for `max_width`, one of "in", "cm", "mm".
#' @examples
#' ft <- qflextable(head(mtcars))
#' ft <- fit_columns(ft, max_width = 5)
#' flextable_dim(ft)$widths
#' ft
#' @family functions for flextable size management
#' @importFrom rlang %||%
fit_columns <- function(x, max_width, no_wrap = NULL, unit = "in") {
  stopifnot(inherits(x, "flextable"))
  max_width <- convin(unit = unit, x = max_width)

  # compute both optimal and floor widths (single fortify_content per part)
  metrics <- fit_columns_metrics(x)
  pretty_w <- metrics$pretty_w
  floor_w <- metrics$floor_w
  n_col <- length(pretty_w)

  # resolve no_wrap to integer indices
  if (!is.null(no_wrap)) {
    if (is.character(no_wrap)) {
      no_wrap <- match(no_wrap, x$col_keys)
      if (anyNA(no_wrap)) {
        stop(
          "Column(s) not found in flextable: ",
          paste(x$col_keys[is.na(no_wrap)], collapse = ", "),
          call. = FALSE
        )
      }
    }
    no_wrap <- as.integer(no_wrap)
    stopifnot(all(no_wrap >= 1L & no_wrap <= n_col))
  }

  # if already fits, just apply pretty widths and return
  if (sum(pretty_w) <= max_width) {
    x <- width(x, width = pretty_w)
    return(x)
  }

  # protected columns
  protected <- no_wrap %||% integer(0)
  protected_width <- sum(pretty_w[protected])

  if (protected_width >= max_width) {
    warning(
      "Protected (no_wrap) columns already exceed max_width; ",
      "returning the table unchanged.",
      call. = FALSE
    )
    return(x)
  }

  compressible <- setdiff(seq_len(n_col), protected)
  available <- max_width - protected_width

  # Proportional allocation with iterative floor clamping.
  #
  # Columns whose proportional share falls below their floor width
  # (longest word + padding) are fixed at that floor. The remaining
  # space is redistributed proportionally among still-free columns.
  # The loop repeats until no new column is clamped.
  #
  # The loop runs at most length(compressible) iterations (one column
  # clamped per iteration in the worst case), each iteration doing
  # only vector arithmetic (negligible cost).
  comp_pretty <- pretty_w[compressible]
  comp_floor <- floor_w[compressible]
  target <- rep(0, length(compressible))
  free <- rep(TRUE, length(compressible))

  repeat {
    free_pretty <- comp_pretty[free]
    total_free <- sum(free_pretty)
    if (total_free <= 0) {
      break
    }

    remaining <- available - sum(comp_floor[!free])
    target[free] <- free_pretty / total_free * remaining
    target[free] <- pmax(target[free], comp_floor[free])

    newly_clamped <- free & (target == comp_floor)
    if (!any(newly_clamped)) {
      break
    }

    free[newly_clamped] <- FALSE
    if (!any(free)) break
  }

  # warn if floors exceed available space
  if (sum(target) > available) {
    warning(
      "Column floor widths exceed max_width; some text wrapping ",
      "at word boundaries may not be achievable.",
      call. = FALSE
    )
  }

  # assemble les largeurs finales qui sont ensuite utilisee comme largeurs
  final_w <- pretty_w
  final_w[protected] <- pretty_w[protected]
  final_w[compressible] <- target
  x <- width(x, width = final_w)

  x
}


# compute both pretty and floor widths with a single fortify_content per part
#' @importFrom data.table copy
fit_columns_metrics <- function(x) {
  parts <- c("header", "body", "footer")
  all_pretty <- list()
  all_floor <- list()

  for (p in parts) {
    if (nrow_part(x, p) < 1) {
      next
    }
    part_obj <- x[[p]]

    txt_data <- fortify_content(
      x = part_obj$content,
      default_chunk_fmt = part_obj$styles$text
    )

    # --- pretty widths (text_metric + optimal_sizes logic) ---
    pretty_sizes <- .text_metric_from_data(part_obj, copy(txt_data))

    nr <- nrow(part_obj$spans$rows)
    rowspan <- data.frame(
      .col_id = rep(part_obj$col_keys, each = nr),
      .row_id = rep(seq_len(nr), length(part_obj$col_keys)),
      rowspan = as.vector(part_obj$spans$rows),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
    pretty_sizes <- merge(pretty_sizes, rowspan, by = c(".col_id", ".row_id"))
    pretty_sizes[pretty_sizes$rowspan != 1, "width"] <- 0

    pretty_sizes$.col_id <- factor(
      pretty_sizes$.col_id,
      levels = part_obj$col_keys
    )
    pretty_sizes <- pretty_sizes[
      order(pretty_sizes$.col_id, pretty_sizes$.row_id),
    ]
    widths_mat <- as_wide_matrix_(
      data = pretty_sizes[, c(".col_id", "width", ".row_id")],
      idvar = ".row_id",
      timevar = ".col_id"
    )
    dimnames(widths_mat)[[2]] <- gsub(
      "^width\\.",
      "",
      dimnames(widths_mat)[[2]]
    )

    par_dim <- dim_paragraphs(part_obj)
    widths_mat <- widths_mat + par_dim$widths

    cell_dim <- dim_cells(part_obj)
    widths_mat <- widths_mat + cell_dim$widths

    all_pretty[[p]] <- apply(widths_mat, 2, max, na.rm = TRUE)

    # --- floor widths (min_col_widths logic) ---
    words <- strsplit(txt_data$txt, "(?<=[- ])", perl = TRUE)
    n_words <- vapply(words, length, integer(1))

    word_txt <- unlist(words, use.names = FALSE)
    word_col <- rep(txt_data$.col_id, n_words)
    word_fname <- rep(txt_data$font.family, n_words)
    word_fsize <- rep(txt_data$font.size, n_words)
    word_bold <- rep(txt_data$bold, n_words)
    word_ital <- rep(txt_data$italic, n_words)

    if (length(word_txt) > 0) {
      word_sizes <- gdtools::strings_sizes(
        word_txt,
        fontname = word_fname,
        fontsize = word_fsize,
        bold = word_bold,
        italic = word_ital
      )
      word_df <- data.frame(
        .col_id = word_col,
        width = word_sizes$width,
        stringsAsFactors = FALSE
      )
      agg <- tapply(word_df$width, word_df$.col_id, max, na.rm = TRUE)
      all_floor[[p]] <- agg
    }
  }

  all_cols <- x$col_keys

  # combine pretty widths across parts
  if (length(all_pretty) > 0) {
    pretty_mat <- do.call(rbind, all_pretty)
    pretty_w <- as.numeric(apply(pretty_mat, 2, max, na.rm = TRUE))
  } else {
    pretty_w <- rep(0, length(all_cols))
  }

  # combine floor widths across parts
  min_text_w <- setNames(rep(0, length(all_cols)), all_cols)
  for (agg in all_floor) {
    matched <- intersect(names(agg), all_cols)
    min_text_w[matched] <- pmax(min_text_w[matched], agg[matched])
  }

  pad_w <- rep(0, length(all_cols))
  names(pad_w) <- all_cols
  cell_w <- pad_w

  for (p in parts) {
    if (nrow_part(x, p) < 1) {
      next
    }
    part_obj <- x[[p]]

    par_dim <- dim_paragraphs(part_obj)
    pad_col <- apply(par_dim$widths, 2, max, na.rm = TRUE)
    pad_col <- setNames(
      as.numeric(pad_col),
      gsub("^width\\.", "", names(pad_col))
    )
    matched <- intersect(names(pad_col), all_cols)
    pad_w[matched] <- pmax(pad_w[matched], pad_col[matched])

    cell_dim <- dim_cells(part_obj)
    cell_col <- apply(cell_dim$widths, 2, max, na.rm = TRUE)
    cell_col <- setNames(
      as.numeric(cell_col),
      gsub("^width\\.", "", names(cell_col))
    )
    matched <- intersect(names(cell_col), all_cols)
    cell_w[matched] <- pmax(cell_w[matched], cell_col[matched])
  }

  floor_w <- as.numeric(min_text_w + pad_w + cell_w)

  list(pretty_w = pretty_w, floor_w = floor_w)
}

# text_metric logic accepting pre-computed txt_data
.text_metric_from_data <- function(x, txt_data) {
  widths <- txt_data$width
  heights <- txt_data$height

  setDT(txt_data)

  txt_data[, c("width", "height") := list(NULL, NULL)]

  txt_data[,
    c("fake_row_id") := list(
      fcase(
        .SD$txt %in% "<br>" , 1L ,
        default = 0L
      )
    )
  ]
  txt_data[,
    c("fake_row_id") := list(
      cumsum(.SD$fake_row_id)
    ),
    by = c(".row_id", ".col_id")
  ]
  txt_data[,
    c("fake_row_id") := list(
      rleid(.SD$fake_row_id)
    ),
    by = c(".row_id", ".col_id")
  ]

  txt_data$txt[txt_data$txt %in% "<br>"] <- ""
  txt_data$txt[txt_data$txt %in% "<tab>"] <- "mmmm"

  # line_spacing per cell for minimum line height
  ls_data <- par_struct_to_df(x$styles$pars)[, c(".row_id", ".col_id")]
  ls_data$line_spacing <- as.vector(x$styles$pars[["line_spacing"]]$data)
  txt_data <- merge(txt_data, ls_data, by = c(".row_id", ".col_id"))

  fontsize <- txt_data$font.size
  not_baseline <- !(txt_data$vertical.align %in% "baseline")
  fontsize[not_baseline] <- fontsize[not_baseline] / 2

  extents_values <- gdtools::strings_sizes(
    txt_data$txt,
    fontname = txt_data$font.family,
    fontsize = fontsize,
    bold = txt_data$bold,
    italic = txt_data$italic
  )
  glyph_height <- extents_values$ascent + extents_values$descent
  line_height <- fontsize * txt_data$line_spacing * 1.2 / 72
  extents_values$height <- pmax(glyph_height, line_height, 0, na.rm = TRUE)
  extents_values$ascent <- NULL
  extents_values$descent <- NULL
  colnames(extents_values) <- c("width", "height")

  extents_values$width <- ifelse(
    !is.na(widths),
    widths,
    extents_values$width
  )
  extents_values$height <- ifelse(
    !is.na(heights),
    heights,
    extents_values$height
  )

  txt_data <- cbind(txt_data, extents_values)

  td_data <- cell_struct_to_df(x$styles$cells)[, c(
    ".row_id",
    ".col_id",
    "text.direction"
  )]
  txt_data <- merge(txt_data, td_data, by = c(".row_id", ".col_id"))
  txt_data[
    txt_data$text.direction %in% c("tbrl", "btlr"),
    c("width", "height") := list(.SD$height, .SD$width)
  ]

  txt_data <- txt_data[,
    c(list(
      width = sum(.SD$width, na.rm = TRUE),
      height = max(.SD$height, na.rm = TRUE)
    )),
    by = c(".row_id", "fake_row_id", ".col_id")
  ]
  txt_data <- txt_data[,
    c(list(
      width = max(.SD$width, na.rm = TRUE),
      height = sum(.SD$height, na.rm = TRUE)
    )),
    by = c(".row_id", ".col_id")
  ]
  setDF(txt_data)
  txt_data
}


# minimum column widths: longest word per column + padding + margins
min_col_widths <- function(x) {
  parts <- c("header", "body", "footer")
  all_min_w <- list()

  for (p in parts) {
    if (nrow_part(x, p) < 1) {
      next
    }

    part_obj <- x[[p]]

    txt_data <- fortify_content(
      x = part_obj$content,
      default_chunk_fmt = part_obj$styles$text
    )

    # split text into words at spaces and hyphens (keep delimiter)
    words <- strsplit(txt_data$txt, "(?<=[- ])", perl = TRUE)
    n_words <- vapply(words, length, integer(1))

    word_txt <- unlist(words, use.names = FALSE)
    word_col <- rep(txt_data$.col_id, n_words)
    word_fname <- rep(txt_data$font.family, n_words)
    word_fsize <- rep(txt_data$font.size, n_words)
    word_bold <- rep(txt_data$bold, n_words)
    word_ital <- rep(txt_data$italic, n_words)

    if (length(word_txt) == 0) {
      next
    }

    # measure each word
    word_sizes <- gdtools::strings_sizes(
      word_txt,
      fontname = word_fname,
      fontsize = word_fsize,
      bold = word_bold,
      italic = word_ital
    )

    # max word width per column
    word_df <- data.frame(
      .col_id = word_col,
      width = word_sizes$width,
      stringsAsFactors = FALSE
    )
    agg <- tapply(word_df$width, word_df$.col_id, max, na.rm = TRUE)
    all_min_w[[p]] <- agg
  }

  if (length(all_min_w) == 0) {
    return(rep(0, length(x$col_keys)))
  }

  # combine across parts: max per column
  all_cols <- x$col_keys
  min_text_w <- setNames(rep(0, length(all_cols)), all_cols)
  for (agg in all_min_w) {
    matched <- intersect(names(agg), all_cols)
    min_text_w[matched] <- pmax(min_text_w[matched], agg[matched])
  }

  # add padding and margins (take max across rows per column)
  pad_w <- rep(0, length(all_cols))
  names(pad_w) <- all_cols
  cell_w <- pad_w

  for (p in parts) {
    if (nrow_part(x, p) < 1) {
      next
    }
    part_obj <- x[[p]]

    par_dim <- dim_paragraphs(part_obj)
    pad_col <- apply(par_dim$widths, 2, max, na.rm = TRUE)
    pad_col <- setNames(
      as.numeric(pad_col),
      gsub("^width\\.", "", names(pad_col))
    )
    matched <- intersect(names(pad_col), all_cols)
    pad_w[matched] <- pmax(pad_w[matched], pad_col[matched])

    cell_dim <- dim_cells(part_obj)
    cell_col <- apply(cell_dim$widths, 2, max, na.rm = TRUE)
    cell_col <- setNames(
      as.numeric(cell_col),
      gsub("^width\\.", "", names(cell_col))
    )
    matched <- intersect(names(cell_col), all_cols)
    cell_w[matched] <- pmax(cell_w[matched], cell_col[matched])
  }

  as.numeric(min_text_w + pad_w + cell_w)
}


#' @export
#' @title Set columns width
#' @description Defines the widths of one or more columns in the
#' table. This function will have no effect if you have
#' used `set_table_properties(layout = "autofit")`.
#'
#' [set_table_properties()] can provide an alternative to fixed-width layouts
#' that is supported with HTML and Word output that can be set
#' with `set_table_properties(layout = "autofit")`.
#'
#'
#' @inheritParams args_x_j
#' @param width width in inches
#' @param unit unit for width, one of "in", "cm", "mm".
#' @details
#' Heights are not used when flextable is been rendered into HTML.
#' @examples
#'
#' ft <- flextable(head(iris))
#' ft <- width(ft, width = 1.5)
#' ft
#' @family functions for flextable size management
width <- function(x, j = NULL, width, unit = "in") {
  width <- convin(unit = unit, x = width)
  j <- get_columns_id(x[["body"]], j)

  stopifnot(length(j) == length(width) || length(width) == 1)

  if (length(width) == 1) {
    width <- rep(width, length(j))
  }

  x$header$colwidths[j] <- width
  x$footer$colwidths[j] <- width
  x$body$colwidths[j] <- width

  x
}

#' @export
#' @title Set flextable rows height
#' @description control rows height for a part of the flextable when the line
#' height adjustment is "atleast" or "exact" (see [hrule()]).
#' @note
#' This function has no effect when the rule for line height is set to
#' "auto" (see [hrule()]), which is the default case, except with PowerPoint
#' which does not support this automatic line height adjustment feature.
#' @inheritParams args_x_i_part_no_all
#' @param height height in inches
#' @param unit unit for height, one of "in", "cm", "mm".
#' @examples
#' ft_1 <- flextable(head(iris))
#' ft_1 <- height(ft_1, height = .5)
#' ft_1 <- hrule(ft_1, rule = "exact")
#' ft_1
#' @family functions for flextable size management
height <- function(x, i = NULL, height, part = "body", unit = "in") {
  height <- convin(unit = unit, x = height)

  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  if (inherits(i, "formula") && any(c("header", "footer") %in% part)) {
    stop("formula in argument i cannot adress part 'header' or 'footer'.")
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  i <- get_rows_id(x[[part]], i)
  if (!(length(i) == length(height) || length(height) == 1)) {
    stop(sprintf("height should be of length 1 or %.0f.", length(i)))
  }

  x[[part]]$rowheights[i] <- height

  x
}

#' @export
#' @title Set how row heights are determined
#' @description
#' `hrule()` controls whether row heights are automatic,
#' minimum or fixed. This only affects Word and PowerPoint
#' outputs; it has no effect on HTML or PDF.
#'
#' * `"auto"` (default): the row height adjusts to fit the
#'   content; any value set by [height()] is ignored.
#' * `"atleast"`: the row is at least as tall as the value
#'   set by [height()], but can grow if the content is taller.
#' * `"exact"`: the row is exactly the height set by
#'   [height()]; content that overflows is clipped.
#'
#' For PDF see the `ft.arraystretch` chunk option.
#' @inheritParams args_x_i_part
#' @param rule specify the meaning of the height. Possible values
#' are "atleast" (height should be at least the value specified), "exact"
#' (height should be exactly the value specified), or the default value "auto"
#' (height is determined based on the height of the contents, so the value is ignored).
#' @examples
#'
#' ft_1 <- flextable(head(iris))
#' ft_1 <- width(ft_1, width = 1.5)
#' ft_1 <- height(ft_1, height = 0.75, part = "header")
#' ft_1 <- hrule(ft_1, rule = "exact", part = "header")
#' ft_1
#'
#' ft_2 <- hrule(ft_1, rule = "auto", part = "header")
#' ft_2
#' @family functions for flextable size management
hrule <- function(x, i = NULL, rule = "auto", part = "body") {
  part <- match.arg(
    part,
    c("body", "header", "footer", "all"),
    several.ok = FALSE
  )

  if ("all" %in% part) {
    for (i in c("body", "header", "footer")) {
      x <- hrule(x, rule = rule, part = i)
    }
    return(x)
  }

  if (inherits(i, "formula") && any(c("header", "footer") %in% part)) {
    stop("formula in argument i cannot adress part 'header' or 'footer'.")
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  i <- get_rows_id(x[[part]], i)
  if (!(length(i) == length(height) || length(height) == 1)) {
    stop(sprintf("height should be of length 1 or %.0f.", length(i)))
  }

  x[[part]]$hrule[i] <- rule
  x
}


#' @export
#' @rdname height
#' @section height_all:
#' `height_all` is a convenient function for
#' setting the same height to all rows (selected
#' with argument `part`).
#' @examples
#'
#'
#' ft_2 <- flextable(head(iris))
#' ft_2 <- height_all(ft_2, height = 1)
#' ft_2 <- hrule(ft_2, rule = "exact")
#' ft_2
height_all <- function(x, height, part = "all", unit = "in") {
  height <- convin(unit = unit, x = height)

  part <- match.arg(
    part,
    c("body", "header", "footer", "all"),
    several.ok = FALSE
  )
  if (length(height) != 1 || !is.numeric(height) || height < 0.0) {
    stop("height should be a single positive numeric value", call. = FALSE)
  }

  if ("all" %in% part) {
    for (i in c("body", "header", "footer")) {
      x <- height_all(x, height = height, part = i)
    }
  }

  if (nrow_part(x, part) > 0) {
    i <- seq_len(nrow(x[[part]]$dataset))
    x[[part]]$rowheights[i] <- height
  }

  x
}

#' @export
#' @title Get overall width and height of a flextable
#' @description Returns the width, height and
#' aspect ratio of a flextable in a named list.
#' The aspect ratio is the ratio corresponding to `height/width`.
#'
#' Names of the list are `widths`, `heights` and `aspect_ratio`.
#' @inheritParams args_x_only
#' @param unit unit for returned values, one of "in", "cm", "mm".
#' @examples
#' ftab <- flextable(head(iris))
#' flextable_dim(ftab)
#' ftab <- autofit(ftab)
#' flextable_dim(ftab)
#' @family functions for flextable size management
flextable_dim <- function(x, unit = "in") {
  dims <- lapply(dim(x), function(x) convin(unit = unit, x = sum(x)))
  dims$aspect_ratio <- dims$heights / dims$widths
  dims
}


#' @title Get column widths and row heights of a flextable
#' @description returns widths and heights for each table columns and rows.
#' Values are expressed in inches.
#' @inheritParams args_x_only
#' @family functions for flextable size management
#' @examples
#' ftab <- flextable(head(iris))
#' dim(ftab)
#' @export
dim.flextable <- function(x) {
  max_widths <- list()
  max_heights <- list()
  for (j in c("header", "body", "footer")) {
    if (nrow_part(x, j) > 0) {
      max_widths[[j]] <- x[[j]]$colwidths
      max_heights[[j]] <- x[[j]]$rowheights
    }
  }

  mat_widths <- do.call("rbind", max_widths)
  if (is.null(mat_widths)) {
    out_widths <- numeric(0)
  } else {
    out_widths <- apply(mat_widths, 2, max)
    names(out_widths) <- x$col_keys
  }

  out_heights <- as.double(unlist(max_heights))
  list(widths = out_widths, heights = out_heights)
}

#' @export
#' @title Calculate optimal column widths and row heights
#' @description return minimum estimated widths and heights for
#' each table columns and rows in inches.
#' @inheritParams args_x_part
#' @param unit unit for returned values, one of "in", "cm", "mm".
#' @param hspans specifies how cells that are horizontally are included in the calculation.
#' It must be one of the following values "none", "divided" or "included". If
#' "none", widths of horizontally spanned cells is set to 0 (then do not affect the
#' widths); if "divided", widths of horizontally spanned cells is divided by
#' the number of spanned cells; if "included", all widths (included horizontally
#' spanned cells) will be used in the calculation.
#' @examples
#' ftab <- flextable(head(mtcars))
#' dim_pretty(ftab)
#' @family functions for flextable size management
dim_pretty <- function(x, part = "all", unit = "in", hspans = "none") {
  stopifnot(length(hspans) == 1, hspans %in% c("none", "divided", "included"))

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = TRUE
  )
  if ("all" %in% part) {
    part <- c("header", "body", "footer")
  }
  dimensions <- list()
  for (j in part) {
    if (nrow_part(x, j) > 0) {
      dimensions[[j]] <- optimal_sizes(x[[j]], hspans = hspans)
    } else {
      dimensions[[j]] <- list(
        widths = rep(0, length(x$col_keys)),
        heights = numeric(0)
      )
    }
  }
  widths <- lapply(dimensions, function(x) x$widths)
  widths <- as.numeric(apply(do.call(rbind, widths), 2, max, na.rm = TRUE))

  heights <- lapply(dimensions, function(x) x$heights)
  heights <- as.numeric(do.call(c, heights))

  list(
    widths = convmeters(unit = unit, x = widths),
    heights = convmeters(unit = unit, x = heights)
  )
}


#' @export
#' @title Adjust columns to their content size
#' @description Compute and apply the minimum widths and heights needed
#' to display each cell's content on a single line, with an optional
#' extra margin (`add_w`, `add_h`).
#'
#' This function sizes columns to fit their content. It does not
#' constrain the table to a given total width. To enforce a maximum
#' width, use [fit_columns()] (wraps text) or [fit_to_width()]
#' (shrinks font size).
#'
#' Note that this function is not related to 'Microsoft Word'
#' *Autofit* feature.
#'
#' There is an alternative to fixed-width layouts that works
#' well with HTML and Word output that can be set
#' with `set_table_properties(layout = "autofit")`, see
#' [set_table_properties()].
#' @inheritParams args_x_part
#' @param add_w extra width to add in inches
#' @param add_h extra height to add in inches
#' @param unit unit for add_h and add_w, one of "in", "cm", "mm".
#' @param hspans specifies how cells that are horizontally are included in the calculation.
#' It must be one of the following values "none", "divided" or "included". If
#' "none", widths of horizontally spanned cells is set to 0 (then do not affect the
#' widths); if "divided", widths of horizontally spanned cells is divided by
#' the number of spanned cells; if "included", all widths (included horizontally
#' spanned cells) will be used in the calculation.
#' @examples
#' ft_1 <- flextable(head(mtcars))
#' ft_1
#' ft_2 <- autofit(ft_1)
#' ft_2
#' @family functions for flextable size management
autofit <- function(
  x,
  add_w = 0.1,
  add_h = 0.1,
  part = c("body", "header"),
  unit = "in",
  hspans = "none"
) {
  add_w <- convin(unit = unit, x = add_w)
  add_h <- convin(unit = unit, x = add_h)

  stopifnot(inherits(x, "flextable"))
  stopifnot(length(hspans) == 1, hspans %in% c("none", "divided", "included"))

  parts <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = TRUE
  )
  if ("all" %in% parts) {
    parts <- c("header", "body", "footer")
  }

  dimensions_ <- dim_pretty(x, part = parts, hspans = hspans)
  names(dimensions_$widths) <- x$col_keys

  nrows <- lapply(parts, function(j) {
    nrow_part(x, j)
  })
  heights <- list(lengths = unlist(nrows), values = parts)
  class(heights) <- "rle"
  heights <- data.frame(
    part = inverse.rle(heights),
    height = dimensions_$heights + add_h,
    stringsAsFactors = FALSE
  )
  heights <- split(heights$height, heights$part)

  for (j in names(heights)) {
    x[[j]]$colwidths <- dimensions_$widths + add_w
    x[[j]]$rowheights <- heights[[j]]
  }
  x
}


#' @importFrom gdtools strings_sizes
optimal_sizes <- function(x, hspans = "none") {
  sizes <- text_metric(x)

  nr <- nrow(x$spans$rows)
  rowspan <- data.frame(
    .col_id = rep(x$col_keys, each = nr),
    .row_id = rep(seq_len(nr), length(x$col_keys)),
    rowspan = as.vector(x$spans$rows),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
  sizes <- merge(sizes, rowspan, by = c(".col_id", ".row_id"))

  if (hspans %in% "none") {
    sizes[sizes$rowspan != 1, "width"] <- 0
  } else if (hspans %in% "divided") {
    sizes[sizes$rowspan > 1, "width"] <- sizes[sizes$rowspan > 1, "width"] /
      sizes[sizes$rowspan > 1, "rowspan"]
    sizes[sizes$rowspan < 1, "width"] <- 0
  }

  sizes$.col_id <- factor(sizes$.col_id, levels = x$col_keys)
  sizes <- sizes[order(sizes$.col_id, sizes$.row_id), ]
  widths <- as_wide_matrix_(
    data = sizes[, c(".col_id", "width", ".row_id")],
    idvar = ".row_id",
    timevar = ".col_id"
  )
  dimnames(widths)[[2]] <- gsub("^width\\.", "", dimnames(widths)[[2]])
  heights <- as_wide_matrix_(
    data = sizes[, c(".col_id", "height", ".row_id")],
    idvar = ".row_id",
    timevar = ".col_id"
  )
  dimnames(heights)[[2]] <- gsub("^height\\.", "", dimnames(heights)[[2]])

  par_dim <- dim_paragraphs(x)
  widths <- widths + par_dim$widths
  heights <- heights + par_dim$heights

  cell_dim <- dim_cells(x)
  widths <- widths + cell_dim$widths
  heights <- heights + cell_dim$heights

  list(
    widths = apply(widths, 2, max, na.rm = TRUE),
    heights = apply(heights, 1, max, na.rm = TRUE)
  )
}


# utils ----
#' @importFrom stats reshape
as_wide_matrix_ <- function(data, idvar, timevar = "col_key") {
  x <- reshape(
    data = data,
    idvar = idvar,
    timevar = timevar,
    direction = "wide"
  )
  x[[idvar]] <- NULL
  as.matrix(x)
}


dim_paragraphs <- function(x) {
  par_dim <- par_struct_to_df(x$styles$pars)

  par_dim$width <- as.vector(
    x$styles$pars[["padding.right"]]$data +
      x$styles$pars[["padding.left"]]$data
  ) *
    (4 / 3) /
    72
  par_dim$height <- as.vector(
    x$styles$pars[["padding.top"]]$data +
      x$styles$pars[["padding.bottom"]]$data
  ) /
    72

  selection_ <- c(".row_id", ".col_id", "width", "height")
  par_dim[, selection_]

  list(
    widths = as_wide_matrix_(
      par_dim[, c(".col_id", "width", ".row_id")],
      idvar = ".row_id",
      timevar = ".col_id"
    ),
    heights = as_wide_matrix_(
      par_dim[, c(".col_id", "height", ".row_id")],
      idvar = ".row_id",
      timevar = ".col_id"
    )
  )
}

dim_cells <- function(x) {
  cell_dim <- cell_struct_to_df(x$styles$cells)
  cell_dim$width <- as.vector(
    x$styles$cells[["margin.right"]]$data + x$styles$cells[["margin.left"]]$data
  ) *
    (4 / 3) /
    72
  cell_dim$height <- as.vector(
    x$styles$cells[["margin.top"]]$data + x$styles$cells[["margin.bottom"]]$data
  ) /
    72 +
    as.vector(
      x$styles$cells[["border.width.top"]]$data +
        x$styles$cells[["border.width.bottom"]]$data
    ) /
      2 /
      72
  selection_ <- c(".row_id", ".col_id", "width", "height")
  cell_dim <- cell_dim[, selection_]

  cellwidths <- as_wide_matrix_(
    cell_dim[, c(".col_id", "width", ".row_id")],
    idvar = ".row_id",
    timevar = ".col_id"
  )
  cellheights <- as_wide_matrix_(
    cell_dim[, c(".col_id", "height", ".row_id")],
    idvar = ".row_id",
    timevar = ".col_id"
  )

  list(widths = cellwidths, heights = cellheights)
}


text_metric <- function(x) {
  txt_data <- fortify_content(
    x = x$content,
    default_chunk_fmt = x$styles$text
  )

  widths <- txt_data$width
  heights <- txt_data$height

  setDT(txt_data)

  txt_data[, c("width", "height") := list(NULL, NULL)]

  # build a fake_row_id so that new lines are considered
  txt_data[,
    c("fake_row_id") := list(
      fcase(
        .SD$txt %in% "<br>" , 1L ,
        default = 0L
      )
    )
  ]
  txt_data[,
    c("fake_row_id") := list(
      cumsum(.SD$fake_row_id)
    ),
    by = c(".row_id", ".col_id")
  ]
  txt_data[,
    c("fake_row_id") := list(
      rleid(.SD$fake_row_id)
    ),
    by = c(".row_id", ".col_id")
  ]

  # set new lines to blank
  txt_data$txt[txt_data$txt %in% "<br>"] <- ""
  # set tabs to 4 'm'
  txt_data$txt[txt_data$txt %in% "<tab>"] <- "mmmm"

  # line_spacing per cell for minimum line height
  ls_data <- par_struct_to_df(x$styles$pars)[, c(".row_id", ".col_id")]
  ls_data$line_spacing <- as.vector(x$styles$pars[["line_spacing"]]$data)
  txt_data <- merge(txt_data, ls_data, by = c(".row_id", ".col_id"))

  fontsize <- txt_data$font.size
  not_baseline <- !(txt_data$vertical.align %in% "baseline")
  fontsize[not_baseline] <- fontsize[not_baseline] / 2

  extents_values <- gdtools::strings_sizes(
    txt_data$txt,
    fontname = txt_data$font.family,
    fontsize = fontsize,
    bold = txt_data$bold,
    italic = txt_data$italic
  )
  glyph_height <- extents_values$ascent + extents_values$descent
  line_height <- fontsize * txt_data$line_spacing * 1.2 / 72
  extents_values$height <- pmax(glyph_height, line_height, 0, na.rm = TRUE)
  extents_values$ascent <- NULL
  extents_values$descent <- NULL
  colnames(extents_values) <- c("width", "height")

  extents_values$width <- ifelse(
    !is.na(widths),
    widths,
    extents_values$width
  )
  extents_values$height <- ifelse(
    !is.na(heights),
    heights,
    extents_values$height
  )

  txt_data <- cbind(txt_data, extents_values)

  # swap width/height when cell is rotated
  td_data <- cell_struct_to_df(x$styles$cells)[, c(
    ".row_id",
    ".col_id",
    "text.direction"
  )]
  txt_data <- merge(txt_data, td_data, by = c(".row_id", ".col_id"))
  txt_data[
    txt_data$text.direction %in% c("tbrl", "btlr"),
    c("width", "height") := list(.SD$height, .SD$width)
  ]

  txt_data <- txt_data[,
    c(list(
      width = sum(.SD$width, na.rm = TRUE),
      height = max(.SD$height, na.rm = TRUE)
    )),
    by = c(".row_id", "fake_row_id", ".col_id")
  ]
  txt_data <- txt_data[,
    c(list(
      width = max(.SD$width, na.rm = TRUE),
      height = sum(.SD$height, na.rm = TRUE)
    )),
    by = c(".row_id", ".col_id")
  ]
  setDF(txt_data)
  txt_data
}

test_that("fit_columns returns unchanged widths when table fits", {
  ft <- qflextable(head(mtcars))
  original_w <- dim_pretty(ft)$widths
  ft2 <- fit_columns(ft, max_width = 20)
  expect_equal(unname(dim(ft2)$widths), unname(original_w))
})

test_that("fit_columns compresses to max_width", {
  ft <- qflextable(head(mtcars))
  ft2 <- suppressWarnings(fit_columns(ft, max_width = 4))
  expect_lte(sum(dim(ft2)$widths), sum(dim_pretty(ft)$widths))
})

test_that("fit_columns applies same widths to all parts", {
  ft <- qflextable(head(iris))
  ft2 <- suppressWarnings(fit_columns(ft, max_width = 3))
  w <- dim(ft2)$widths
  expect_true(all(is.finite(w)))
  expect_true(all(w > 0))
})

test_that("fit_columns no_wrap protects columns by name", {
  ft <- qflextable(head(iris))
  protected <- "Species"
  pretty_w <- dim_pretty(ft)$widths
  ft2 <- suppressWarnings(fit_columns(ft, max_width = 3, no_wrap = protected))
  idx <- match(protected, ft$col_keys)
  expect_equal(unname(dim(ft2)$widths[idx]), unname(pretty_w[idx]))
})

test_that("fit_columns no_wrap protects columns by index", {
  ft <- qflextable(head(iris))
  pretty_w <- dim_pretty(ft)$widths
  ft2 <- suppressWarnings(fit_columns(ft, max_width = 3, no_wrap = 5L))
  expect_equal(unname(dim(ft2)$widths[5]), unname(pretty_w[5]))
})

test_that("fit_columns errors on invalid column names", {
  ft <- qflextable(head(iris))
  expect_error(
    fit_columns(ft, max_width = 3, no_wrap = "nonexistent"),
    "not found"
  )
})

test_that("fit_columns warns when protected columns exceed max_width", {
  ft <- qflextable(head(iris))
  expect_warning(
    fit_columns(ft, max_width = 0.01, no_wrap = names(iris)),
    "exceed max_width"
  )
})

test_that("fit_columns warns when floors exceed available space", {
  ft <- qflextable(head(mtcars))
  expect_warning(
    fit_columns(ft, max_width = 2),
    "floor widths exceed"
  )
})

test_that("fit_columns respects unit conversion", {
  ft <- qflextable(head(iris))
  ft_in <- suppressWarnings(fit_columns(ft, max_width = 4, unit = "in"))
  ft_cm <- suppressWarnings(fit_columns(ft, max_width = 4 * 2.54, unit = "cm"))
  expect_equal(dim(ft_in)$widths, dim(ft_cm)$widths, tolerance = 1e-6)
})

test_that("fit_columns iterative clamping redistributes correctly", {
  dat <- data.frame(
    A = "Supercalifragilisticexpialidocious",
    B = "a b c",
    C = "x y z",
    stringsAsFactors = FALSE
  )
  ft <- flextable(dat)
  ft <- autofit(ft, add_w = 0, add_h = 0)

  pretty_w <- dim_pretty(ft)$widths
  max_w <- sum(pretty_w) * 0.5
  ft2 <- suppressWarnings(fit_columns(ft, max_width = max_w))

  w <- dim(ft2)$widths
  expect_true(all(is.finite(w)))
  expect_true(all(w > 0))
})

test_that("dimensions are valid", {
  dummy_df <- data.frame(
    my_col = rep(letters[1:3], each = 2),
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  dims <- dim(ft)
  expect_length(dims$widths, 1)
  expect_length(dims$heights, 7)

  ft <- add_header(ft, my_col = "second row header")
  dims <- dim(ft)
  expect_length(dims$widths, 1)
  expect_length(dims$heights, 8)

  ft <- height_all(ft, height = .15, part = "all")
  dims <- dim(ft)
  expect_true(all(is.finite(dims$widths)))
  expect_true(all(is.finite(dims$heights)))

  typology <- data.frame(
    col_keys = c(
      "Sepal.Length",
      "Sepal.Width",
      "Petal.Length",
      "Petal.Width",
      "Species"
    ),
    what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
    measure = c("Length", "Width", "Length", "Width", "Species"),
    stringsAsFactors = FALSE
  )
  ft <- flextable(head(iris))
  ft <- set_header_df(ft, mapping = typology, key = "col_keys")
  dims <- dim(ft)
  expect_length(dims$widths, 5)
  expect_length(dims$heights, 8)
})

test_that("autofit and dim_pretty usage", {
  typology <- data.frame(
    col_keys = c(
      "Sepal.Length",
      "Sepal.Width",
      "Petal.Length",
      "Petal.Width",
      "Species"
    ),
    what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
    measure = c("Length", "Width", "Length", "Width", "Species"),
    stringsAsFactors = FALSE
  )
  ft <- flextable(head(iris))
  ft <- set_header_df(ft, mapping = typology, key = "col_keys")
  ft <- autofit(ft, add_w = 0.0, add_h = 0.0)
  dims <- dim(ft)
  expect_true(all(is.finite(dims$widths)))
  expect_true(all(is.finite(dims$heights)))

  dims <- dim_pretty(ft)
  expect_true(all(is.finite(dims$widths)))
  expect_true(all(is.finite(dims$heights)))
})


test_that("height usage", {
  dummy_df <- data.frame(
    my_col1 = rep(letters[1:3], each = 2),
    my_col2 = rep(letters[4:6], each = 2),
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  dims <- dim_pretty(ft)
  expect_silent({
    ft <- height_all(ft, height = .25, part = "all")
  })
  expect_error({
    ft <- height(ft, height = dims$heights[-1], part = "all")
  })
  expect_error({
    ft <- height(ft, height = 1:3, part = "body")
  })
})

test_that("width usage", {
  dummy_df <- data.frame(
    my_col1 = rep(letters[1:3], each = 2),
    my_col2 = rep(letters[4:6], each = 2),
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  expect_silent(width(ft, j = "my_col1", width = 1))
  expect_silent(width(ft, j = 1:2, width = .7))
  expect_error(width(ft, j = 1:2, width = rep(.7, 3)))
})

test_that("autofit with horizontal spans", {
  dat <-
    data.frame(
      `Header should span 2 cols` = c("Whoooo", "Whaaaat", "Whyyyyy"),
      dummy_title = c("Whoooo", "Whaaaat", "Whyyyyy"),
      check.names = FALSE
    )

  ft <- flextable(dat)
  ft <- merge_at(
    x = ft,
    i = 1,
    j = 1:2,
    part = "header"
  )

  dims_divided <- dim_pretty(ft, hspans = "divided")
  dims_included <- dim_pretty(ft, hspans = "included")
  expect_gt(dims_included$widths[1], dims_divided$widths[1])

  dims_none <- dim_pretty(ft, hspans = "none")
  dims_included_no_header <- dim_pretty(ft, hspans = "included", part = "body")
  expect_equal(dims_none$widths, dims_included_no_header$widths)
})

test_that("HTML table width when autofit layout", {
  x <- data.frame(
    x = c(
      "[-0.36, -0.01]",
      "[0, 0]",
      "000, 000",
      "0000 0000"
    )
  )

  ft <- flextable(x)
  ft <- set_table_properties(
    x = ft,
    layout = "autofit"
  )
  str <- flextable:::gen_raw_html(ft)
  expect_false(grepl("table-layout:auto;width:", str, fixed = TRUE))

  ft <- set_table_properties(
    x = ft,
    layout = "autofit",
    width = .1
  )
  str <- flextable:::gen_raw_html(x = ft)
  expect_true(grepl("table-layout:auto;width:", str, fixed = TRUE))
})

#' @export
#' @title Set formatting properties on a flextable selection
#' @description
#' `style()` applies text, paragraph and cell formatting properties
#' to a selection of rows and columns in one call. It is a lower-level
#' function that bundles what the convenience functions do individually:
#'
#' * text (`pr_t`): font family, size, color, bold, italic, etc.,
#'   same properties as [bold()], [italic()], [color()], [fontsize()],
#'   [font()], [highlight()].
#' * paragraph (`pr_p`): alignment, padding, line spacing,
#'   same properties as [align()], [padding()], [line_spacing()].
#' * cell (`pr_c`): background colour, borders, vertical alignment,
#'   same properties as [bg()], [border()], [valign()].
#'
#' Use `style()` when you need to set several property types at once
#' on the same selection; use the convenience functions when you
#' only need to change one aspect.
#'
#' Note that `style()` modifies an existing flextable.
#' To change the initial formatting applied to every new
#' flextable, use [set_flextable_defaults()] instead.
#' @inheritParams args_selectors_with_all
#' @param pr_t an [officer::fp_text()] or [officer::fp_text_lite()] object
#' defining text formatting (font, size, color, bold, ...).
#' @param pr_p an [officer::fp_par()] or [officer::fp_par_lite()] object
#' defining paragraph formatting (alignment, padding, line spacing, ...).
#' @param pr_c an [officer::fp_cell()] object defining cell formatting
#' (background, borders, vertical alignment, ...).
#' @family formatting_shortcuts
#' @importFrom stats terms
#' @examples
#' library(officer)
#' def_cell <- fp_cell(border = fp_border(color = "wheat"))
#'
#' def_par <- fp_par(text.align = "center")
#'
#' ft <- flextable(head(mtcars))
#'
#' ft <- style(ft, pr_c = def_cell, pr_p = def_par, part = "all")
#' ft <- style(ft, ~ drat > 3.5, ~ vs + am + gear + carb,
#'   pr_t = fp_text(color = "red", italic = TRUE)
#' )
#'
#' ft
style <- function(
  x,
  i = NULL,
  j = NULL,
  pr_t = NULL,
  pr_p = NULL,
  pr_c = NULL,
  part = "body"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "style()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    args <- list()
    for (p in c("header", "body", "footer")) {
      args$x <- x
      args$i <- i
      args$j <- j
      args$pr_t <- pr_t
      args$pr_p <- pr_p
      args$pr_c <- pr_c
      args$part <- p
      x <- do.call(style, args)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (!is.null(pr_t)) {
    x[[part]]$styles$text <- set_text_struct_values(
      x = x[[part]]$styles$text,
      i = i,
      j = j,
      value = pr_t
    )
  }

  if (!is.null(pr_p)) {
    if (
      !is.null(pr_p$tabs) && !isFALSE(pr_p$tabs) && !is.character(pr_p$tabs)
    ) {
      pr_p$tabs <- as.character(pr_p$tabs)
    } else {
      pr_p$tabs <- NULL
    }

    pr_p <- cast_borders(pr_p)
    for (property in intersect(names(pr_p), names(x[[part]]$styles$pars))) {
      if (!is.null(pr_p[[property]]) && !is.na(pr_p[[property]])) {
        x[[part]]$styles$pars[[property]]$data[i, j] <- pr_p[[property]]
      }
    }
  }

  if (!is.null(pr_c)) {
    pr_c <- cast_borders(pr_c)
    for (property in intersect(names(pr_c), names(x[[part]]$styles$cells))) {
      x[[part]]$styles$cells[[property]]$data[i, j] <- pr_c[[property]]
    }
  }

  x
}


# text format ----

#' @export
#' @title Set bold font
#' @description Change the font weight of selected rows and columns of a flextable.
#' @inheritParams args_selectors_with_all
#' @param bold boolean value
#' @family formatting_shortcuts
#' @examples
#' ft <- flextable(head(iris))
#' ft <- bold(ft, bold = TRUE, part = "header")
bold <- function(x, i = NULL, j = NULL, bold = TRUE, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "bold()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- bold(x = x, i = i, j = j, bold = bold, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(bold) == length(j)) {
    bold <- rep(bold, each = length(i))
  }

  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "bold",
    value = bold
  )

  x
}

#' @export
#' @title Set font size
#' @description Change the font size of selected rows and columns of a flextable.
#' @inheritParams args_selectors_with_all
#' @param size integer value (points)
#' @family formatting_shortcuts
#' @examples
#' ft <- flextable(head(iris))
#' ft <- fontsize(ft, size = 14, part = "header")
#' ft <- fontsize(ft, size = 14, j = 2)
#' ft <- fontsize(ft, size = 7, j = 3)
#' ft
fontsize <- function(x, i = NULL, j = NULL, size = 11, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "fontsize()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- fontsize(x = x, i = i, j = j, size = size, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "font.size",
    value = size
  )

  x
}

#' @export
#' @title Set italic font
#' @description Change the font decoration of selected rows and columns of a flextable.
#' @inheritParams args_selectors_with_all
#' @param italic boolean value
#' @family formatting_shortcuts
#' @examples
#' ft <- flextable(head(mtcars))
#' ft <- italic(ft, italic = TRUE, part = "header")
italic <- function(x, i = NULL, j = NULL, italic = TRUE, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "italic()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- italic(x = x, i = i, j = j, italic = italic, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(italic) == length(j)) {
    italic <- rep(italic, each = length(i))
  }

  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "italic",
    value = italic
  )

  x
}

#' @export
#' @title Set text highlight color
#' @description Change the text highlight color of selected rows and
#' columns of a flextable. A function can be used instead of
#' fixed colors.
#'
#' When `color` is a function, it is possible to color cells based on values
#' located in other columns; using hidden columns (those not used by
#' argument `colkeys`) is a common use case. The argument `source`
#' must be used to define the columns to be used for the color
#' definition, and the argument `j` must be used to define where to apply
#' the colors and only accepts values from `colkeys`.

#' @inheritParams args_selectors_with_all
#' @param color color to use as text highlighting color.
#' If a function, the function must return a character vector of colors.
#' @param source if color is a function, source specifies the dataset column to be used
#' as an argument to `color`. This is only useful when j is colored with values contained in
#' other columns.
#' @family formatting_shortcuts
#' @examples
#' my_color_fun <- function(x) {
#'   out <- rep("yellow", length(x))
#'   out[x < quantile(x, .75)] <- "pink"
#'   out[x < quantile(x, .50)] <- "wheat"
#'   out[x < quantile(x, .25)] <- "gray90"
#'   out
#' }
#' ft <- flextable(head(mtcars, n = 10))
#' ft <- highlight(ft, j = "disp", i = ~ disp > 200, color = "yellow")
#' ft <- highlight(ft, j = ~ drat + wt + qsec, color = my_color_fun)
#' ft
highlight <- function(
  x,
  i = NULL,
  j = NULL,
  color = "yellow",
  part = "body",
  source = j
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "highlight()"
    ))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- highlight(x = x, i = i, j = j, color = color, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)

  if (is.function(color)) {
    source <- as_col_keys(x[[part]], source, blanks = x$blanks)
    source_dataset <- x[[part]]$dataset[source]
    source_dataset <- source_dataset[get_rows_id(x[[part]], i), ]
    color <- data_colors(source_dataset, color)
  }

  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(color) == length(j)) {
    color <- rep(color, each = length(i))
  }

  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "shading.color",
    value = color
  )

  x
}

#' @export
#' @title Set font color
#' @description Change the text color of selected rows and
#' columns of a flextable. A function can be used instead of
#' fixed colors.
#'
#' When `color` is a function, it is possible to color cells based on values
#' located in other columns; using hidden columns (those not used by
#' argument `colkeys`) is a common use case. The argument `source`
#' must be used to define the columns to be used for the color
#' definition, and the argument `j` must be used to define where to apply
#' the colors and only accepts values from `colkeys`.
#'
#' @inheritParams args_selectors_with_all
#' @param color color to use as font color. If a function, the function must return
#' a character vector of colors.
#' @param source if color is a function, source specifies the dataset column to be used
#' as an argument to `color`. This is only useful when j is colored with values contained in
#' other columns.
#' @family formatting_shortcuts
#' @examples
#' ft <- flextable(head(mtcars))
#' ft <- color(ft, color = "orange", part = "header")
#' ft <- color(ft,
#'   color = "red",
#'   i = ~ qsec < 18 & vs < 1
#' )
#' ft
#'
#' if (require("scales")) {
#'   scale <- scales::col_numeric(domain = c(-1, 1), palette = "RdBu")
#'   x <- as.data.frame(cor(iris[-5]))
#'   x <- cbind(
#'     data.frame(
#'       colname = colnames(x),
#'       stringsAsFactors = FALSE
#'     ),
#'     x
#'   )
#' }
color <- function(x, i = NULL, j = NULL, color, part = "body", source = j) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "color()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- color(x = x, i = i, j = j, color = color, part = p, source = source)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)

  if (is.function(color)) {
    source <- as_col_keys(x[[part]], source, blanks = x$blanks)
    source_dataset <- x[[part]]$dataset[source]
    source_dataset <- source_dataset[get_rows_id(x[[part]], i), ]
    color <- data_colors(source_dataset, color)
  }

  i <- get_rows_id(x[[part]], i)
  j <- x$col_keys[get_columns_id(x[[part]], j)]

  if (length(color) == length(j)) {
    color <- rep(color, each = length(i))
  }

  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "color",
    value = color
  )

  x
}

#' @export
#' @title Set font
#' @description Change the font of selected rows and columns of a flextable.
#'
#' Fonts impact the readability and aesthetics of the table. Font families
#' refer to a set of typefaces that share common design features, such as 'Arial'
#' and 'Open Sans'.
#'
#' 'Google Fonts' is a popular library of free web fonts that can be
#' easily integrated into flextable with the [gdtools::register_gfont()] function.
#' When the output is HTML, the font will be automatically added to the HTML
#' document.
#'
#' @inheritParams args_selectors_with_all
#' @param fontname single character value, the font family name.
#' With Word and PowerPoint output, this value specifies the font to
#' be used for formatting characters in the Unicode range (U+0000-U+007F).
#' @param cs.family Optional font to be used for formatting
#' characters in a complex script Unicode range. For example, Arabic
#' text might be displayed using the "Arial Unicode MS" font.
#' Used only with Word and PowerPoint outputs. The default value is the value
#' of `fontname`.
#' @param eastasia.family Optional font to be used for
#' formatting characters in an East Asian Unicode range. For example,
#' Japanese text might be displayed using the "MS Mincho" font.
#' Used only with Word and PowerPoint outputs. The default value is the value
#' of `fontname`.
#' @param hansi.family Optional font to be used for formatting
#' characters in a Unicode range that does not fall into one of the
#' other categories.
#' Used only with Word and PowerPoint outputs. The default value is the value
#' of `fontname`.
#' @family formatting_shortcuts
#' @examples
#' library(gdtools)
#' fontname <- "Brush Script MT"
#'
#' if (font_family_exists(fontname)) {
#'   ft_1 <- flextable(head(iris))
#'   ft_2 <- font(ft_1, fontname = fontname, part = "header")
#'   ft_2 <- font(ft_2, fontname = fontname, j = 5)
#'   ft_2
#' }
font <- function(
  x,
  i = NULL,
  j = NULL,
  fontname,
  part = "body",
  cs.family = fontname,
  hansi.family = fontname,
  eastasia.family = fontname
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "font()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- font(x = x, i = i, j = j, fontname = fontname, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(fontname) == length(j)) {
    fontname <- rep(fontname, each = length(i))
  }
  if (length(cs.family) == length(j)) {
    cs.family <- rep(cs.family, each = length(i))
  }
  if (length(hansi.family) == length(j)) {
    hansi.family <- rep(hansi.family, each = length(i))
  }
  if (length(eastasia.family) == length(j)) {
    eastasia.family <- rep(eastasia.family, each = length(i))
  }

  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "font.family",
    value = fontname
  )
  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "cs.family",
    value = cs.family
  )
  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "hansi.family",
    value = hansi.family
  )
  x[[part]]$styles$text <- set_text_struct_values(
    x = x[[part]]$styles$text,
    i = i,
    j = j,
    property = "eastasia.family",
    value = eastasia.family
  )
  x
}

# paragraphs format ----

#' @export
#' @title Set paragraph paddings
#' @description Change the padding of selected rows and columns of a flextable.
#' @note
#' In PDF output, only `padding.left` and `padding.right` are supported.
#' `padding.top` and `padding.bottom` are ignored due to LaTeX limitations.
#' Global horizontal spacing can also be set with
#' `set_table_properties(opts_pdf = list(tabcolsep = 1))`.
#' @inheritParams args_selectors_with_all
#' @param padding padding (shortcut for top, bottom, left and right), unit is pts (points).
#' @param padding.top padding top, unit is pts (points).
#' @param padding.bottom padding bottom, unit is pts (points).
#' @param padding.left padding left, unit is pts (points).
#' @param padding.right padding right, unit is pts (points).
#' @family formatting_shortcuts
#' @examples
#' ft_1 <- flextable(head(iris))
#' ft_1 <- theme_vader(ft_1)
#' ft_1 <- padding(ft_1, padding.top = 4, part = "all")
#' ft_1 <- padding(ft_1, j = 1, padding.right = 40)
#' ft_1 <- padding(ft_1, i = 3, padding.top = 40)
#' ft_1 <- padding(ft_1, padding.top = 10, part = "header")
#' ft_1 <- padding(ft_1, padding.bottom = 10, part = "header")
#' ft_1 <- autofit(ft_1)
#' ft_1
padding <- function(
  x,
  i = NULL,
  j = NULL,
  padding = NULL,
  padding.top = NULL,
  padding.bottom = NULL,
  padding.left = NULL,
  padding.right = NULL,
  part = "body"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "padding()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (!is.null(padding)) {
    if (is.null(padding.top)) {
      padding.top <- padding
    }
    if (is.null(padding.bottom)) {
      padding.bottom <- padding
    }
    if (is.null(padding.left)) {
      padding.left <- padding
    }
    if (is.null(padding.right)) padding.right <- padding
  }
  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- padding(
        x = x,
        i = i,
        j = j,
        padding.top = padding.top,
        padding.bottom = padding.bottom,
        padding.left = padding.left,
        padding.right = padding.right,
        part = p
      )
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (!is.null(padding.top)) {
    x[[part]]$styles$pars <- set_par_struct_values(
      x = x[[part]]$styles$pars,
      i = i,
      j = j,
      property = "padding.top",
      value = padding.top
    )
  }
  if (!is.null(padding.bottom)) {
    x[[part]]$styles$pars <- set_par_struct_values(
      x = x[[part]]$styles$pars,
      i = i,
      j = j,
      property = "padding.bottom",
      value = padding.bottom
    )
  }
  if (!is.null(padding.left)) {
    x[[part]]$styles$pars <- set_par_struct_values(
      x = x[[part]]$styles$pars,
      i = i,
      j = j,
      property = "padding.left",
      value = padding.left
    )
  }
  if (!is.null(padding.right)) {
    x[[part]]$styles$pars <- set_par_struct_values(
      x = x[[part]]$styles$pars,
      i = i,
      j = j,
      property = "padding.right",
      value = padding.right
    )
  }

  x
}

#' @export
#' @title Set paragraph indents
#' @description Change the first-line or hanging indent of paragraphs in
#' selected rows and columns of a flextable.
#'
#' A *first-line* indent moves only the first line of the paragraph to the
#' right. A *hanging* indent moves all lines **but** the first one to the
#' right; the indented lines start at the paragraph left padding plus the
#' hanging value, so `hanging` should not exceed the left padding
#' (see [padding()]).
#' @inheritParams args_selectors_with_all
#' @param first_line first-line indent in pts (points). Use NA to unset.
#' @param hanging hanging indent in pts (points). Use NA to unset. If both
#' `first_line` and `hanging` are provided, `hanging` wins.
#' @family formatting_shortcuts
#' @examples
#' ft_1 <- flextable(head(iris))
#' ft_1 <- mk_par(
#'   x = ft_1, i = 1, j = 1, part = "header",
#'   value = as_paragraph("A rather long header that will wrap on two lines")
#' )
#' ft_1 <- padding(ft_1, i = 1, j = 1, padding.left = 30, part = "header")
#' ft_1 <- indentation(ft_1, i = 1, j = 1, hanging = 20, part = "header")
#' ft_1 <- width(ft_1, j = 1, width = 1.2)
#' ft_1
indentation <- function(
  x,
  i = NULL,
  j = NULL,
  first_line = NULL,
  hanging = NULL,
  part = "body"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "indentation()"
    ))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- indentation(
        x = x,
        i = i,
        j = j,
        first_line = first_line,
        hanging = hanging,
        part = p
      )
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (!is.null(first_line)) {
    if (!is.numeric(first_line) && !all(is.na(first_line))) {
      stop("`first_line` must be numeric (pts) or NA.", call. = FALSE)
    }
    x[[part]]$styles$pars <- set_par_struct_values(
      x = x[[part]]$styles$pars,
      i = i,
      j = j,
      property = "first_line",
      value = as.double(first_line)
    )
  }
  if (!is.null(hanging)) {
    if (!is.numeric(hanging) && !all(is.na(hanging))) {
      stop("`hanging` must be numeric (pts) or NA.", call. = FALSE)
    }
    x[[part]]$styles$pars <- set_par_struct_values(
      x = x[[part]]$styles$pars,
      i = i,
      j = j,
      property = "hanging",
      value = as.double(hanging)
    )
  }

  x
}


#' @export
#' @title Set text alignment
#' @description Change the text alignment of selected rows and columns of a flextable.
#' @inheritParams args_selectors_with_all
#' @param align text alignment - a single character value, or a vector of
#' character values equal in length to the number of columns selected by `j`.
#' Expected values must be from the set ('left', 'right', 'center', or 'justify').
#'
#' If the number of columns is a multiple of the length of the `align` parameter,
#' then the values in `align` will be recycled across the remaining columns.
#' @family formatting_shortcuts
#' @examples
#' # Table of 6 columns
#' ft_car <- flextable(head(mtcars)[, 2:7])
#'
#' # All 6 columns right aligned
#' align(ft_car, align = "right", part = "all")
#'
#' # Manually specify alignment of each column
#' align(
#'   ft_car,
#'   align = c("left", "right", "left", "center", "center", "right"),
#'   part = "all"
#' )
#'
#' # Center-align column 2 and left-align column 5
#' align(ft_car, j = c(2, 5), align = c("center", "left"), part = "all")
#'
#' # Alternate left and center alignment across columns 1-4 for header only
#' align(ft_car, j = 1:4, align = c("left", "center"), part = "header")
align <- function(
  x,
  i = NULL,
  j = NULL,
  align = "left",
  part = c("body", "header", "footer", "all")
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "align()"))
  }
  part <- match.arg(
    part,
    c("body", "header", "footer", "all"),
    several.ok = FALSE
  )

  allowed_alignments <- c("left", "center", "right", "justify")
  if (!all(align %in% allowed_alignments)) {
    stop(
      "Invalid `align` argument(s) provided to `align()`: \"",
      paste(setdiff(align, allowed_alignments), collapse = "\", \""),
      "\".\n  `align` should be one of: \"",
      paste(allowed_alignments, collapse = "\", \""),
      "\"."
    )
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- align(x = x, i = i, j = j, align = align, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(j) %% length(align) == 0) {
    align <- rep(align, each = length(i))
  }

  x[[part]]$styles$pars <- set_par_struct_values(
    x = x[[part]]$styles$pars,
    i = i,
    j = j,
    property = "text.align",
    value = align
  )
  x
}

#' @export
#' @title Set Word 'Keep with next' instructions
#' @description The 'Keep with next' functionality in 'Word', applied
#' to the rows of a table, ensures that rows with that attribute
#' stay together and do not break across multiple pages.
#'
#' This function provides better control of page breaks than
#' the global `keep_with_next` parameter.
#' @inheritParams args_selectors_without_all
#' @param value TRUE or FALSE. When applied to a group, all rows
#' except the last one should be flagged with the 'Keep with next' attribute.
#' @family formatting_shortcuts
#' @seealso [paginate()]
#' @examples
#' library(flextable)
#' dat <- iris[c(1:25, 51:75, 101:125), ]
#' ft <- qflextable(dat)
#' ft <- keep_with_next(
#'   x = ft,
#'   i = c(1:24, 26:49, 51:74),
#'   value = TRUE
#' )
#'
#' save_as_docx(ft, path = tempfile(fileext = ".docx"))
keep_with_next <- function(x, i = NULL, value = TRUE, part = "body") {
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- keep_with_next(x = x, i = i, value = value, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  x[[part]]$styles$pars <- set_par_struct_values(
    x = x[[part]]$styles$pars,
    i = i,
    j = NULL,
    property = "keep_with_next",
    value = value
  )

  x
}

#' @export
#' @title Set tabulation marks configuration
#' @description Define tabulation marks configuration.
#' Specifying the positions and types of tabulation marks in table
#' paragraphs helps organize content, especially in clinical tables,
#' by aligning numbers properly.
#' @inheritParams args_selectors_with_all
#' @param value an object generated by [officer::fp_tabs()].
#' @family formatting_shortcuts
#' @examples
#' library(officer)
#' library(flextable)
#'
#' z <- data.frame(
#'   Statistic = c("Median (Q1 ; Q3)", "Min ; Max"),
#'   Value = c(
#'     "\t999.99\t(99.9 ; 99.9)",
#'     "\t9.99\t(9999.9 ; 99.9)"
#'   )
#' )
#'
#' ts <- fp_tabs(
#'   fp_tab(pos = 0.4, style = "decimal"),
#'   fp_tab(pos = 1.4, style = "decimal")
#' )
#'
#' zz <- flextable(z)
#' zz <- tab_settings(zz, j = 2, value = ts)
#' zz <- width(zz, width = c(1.5, 2))
#'
#'
#' save_as_docx(zz, path = tempfile(fileext = ".docx"))
tab_settings <- function(x, i = NULL, j = NULL, value = TRUE, part = "body") {
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- tab_settings(x = x, i = i, j = j, value = value, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  x[[part]]$styles$pars <- set_par_struct_values(
    x = x[[part]]$styles$pars,
    i = i,
    j = j,
    property = "tabs",
    value = as.character(value)
  )

  x
}

#' @export
#' @title Set line spacing
#' @description Change the line spacing of selected rows and columns of a flextable.
#' @inheritParams args_selectors_with_all
#' @param space space between lines of text; 1 is single line spacing, 2 is double line spacing.
#' @family formatting_shortcuts
#' @examples
#' ft <- flextable(head(mtcars)[, 3:6])
#' ft <- line_spacing(ft, space = 1.6, part = "all")
#' ft <- set_table_properties(ft, layout = "autofit")
#' ft
line_spacing <- function(x, i = NULL, j = NULL, space = 1, part = "body") {
  if (!inherits(x, "flextable")) {
    stop("align supports only flextable objects.")
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- line_spacing(x = x, i = i, j = j, space = space, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  x[[part]]$styles$pars <- set_par_struct_values(
    x = x[[part]]$styles$pars,
    i = i,
    j = j,
    property = "line_spacing",
    value = space
  )

  x
}

#' @export
#' @rdname align
#' @param header should the header be aligned with the body
#' @param footer should the footer be aligned with the body
#' @examples
#' ftab <- flextable(mtcars)
#' ftab <- align_text_col(ftab, align = "left")
#' ftab <- align_nottext_col(ftab, align = "right")
#' ftab
align_text_col <- function(x, align = "left", header = TRUE, footer = TRUE) {
  which_j <- which(sapply(x$body$dataset[x$col_keys], function(x) {
    is.character(x) | is.factor(x)
  }))
  x <- align(x, j = which_j, align = align, part = "body")
  if (header) {
    x <- align(x, j = which_j, align = align, part = "header")
  }
  if (footer) {
    x <- align(x, j = which_j, align = align, part = "footer")
  }
  x
}

#' @export
#' @rdname align
align_nottext_col <- function(
  x,
  align = "right",
  header = TRUE,
  footer = TRUE
) {
  which_j <- which(
    !sapply(x$body$dataset[x$col_keys], function(x) {
      is.character(x) | is.factor(x)
    })
  )
  x <- align(x, j = which_j, align = align, part = "body")
  if (header) {
    x <- align(x, j = which_j, align = align, part = "header")
  }
  if (footer) {
    x <- align(x, j = which_j, align = align, part = "footer")
  }
  x
}

# cells format ----

#' @export
#' @title Set background color
#' @description Change the background color of selected rows and
#' columns of a flextable. A function can be used instead of
#' fixed colors.
#'
#' When `bg` is a function, it is possible to color cells based on values
#' located in other columns; using hidden columns (those not used by
#' argument `colkeys`) is a common use case. The argument `source`
#' must be used to define the columns to be used for the color
#' definition, and the argument `j` must be used to define where to apply
#' the colors and only accepts values from `colkeys`.
#'
#' @inheritParams args_selectors_with_all
#' @param bg color to use as background color. If a function, the function must return
#' a character vector of colors.
#' @param source if bg is a function, source specifies the dataset column to be used
#' as an argument to `bg`. This is only useful when j is colored with values contained in
#' other columns.
#' @family formatting_shortcuts
#' @note
#' Word does not allow you to apply transparency to table cells or paragraph shading.
#' @examples
#' ft_1 <- flextable(head(mtcars))
#' ft_1 <- bg(ft_1, bg = "wheat", part = "header")
#' ft_1 <- bg(ft_1, i = ~ qsec < 18, bg = "#EFEFEF", part = "body")
#' ft_1 <- bg(ft_1, j = "drat", bg = "#606060", part = "all")
#' ft_1 <- color(ft_1, j = "drat", color = "white", part = "all")
#' ft_1
#'
#' if (require("scales")) {
#'   ft_2 <- flextable(head(iris))
#'   colourer <- col_numeric(
#'     palette = c("wheat", "red"),
#'     domain = c(0, 7)
#'   )
#'   ft_2 <- bg(ft_2,
#'     j = c(
#'       "Sepal.Length", "Sepal.Width",
#'       "Petal.Length", "Petal.Width"
#'     ),
#'     bg = colourer, part = "body"
#'   )
#'   ft_2
#' }
bg <- function(x, i = NULL, j = NULL, bg, part = "body", source = j) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "bg()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- bg(x = x, i = i, j = j, bg = bg, part = p, source = source)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)

  if (is.function(bg)) {
    source <- as_col_keys(x[[part]], source, blanks = x$blanks)
    source_dataset <- x[[part]]$dataset[source]
    source_dataset <- source_dataset[get_rows_id(x[[part]], i), ]
    bg <- data_colors(source_dataset, bg)
  }

  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (length(bg) == length(j)) {
    bg <- rep(bg, each = length(i))
  }

  x[[part]]$styles$cells[["background.color"]]$data[i, j] <- bg

  x
}


data_colors <- function(dataset, fun) {
  out <- tryCatch(
    {
      lbg <- lapply(dataset, fun)
      matrix(unlist(lbg), ncol = length(lbg))
    },
    error = function(cond) {
      msg <- paste0(
        "an error occurred while using color function: ",
        cond$message
      )
      stop(msg, call. = FALSE)
    }
  )
  if (anyNA(out)) {
    stop("colors can not contain missing values")
  }
  return(out)
}


#' @export
#' @title Set vertical alignment
#' @description Change the vertical alignment of selected rows and columns of a flextable.
#' @inheritParams args_selectors_with_all
#' @param valign vertical alignment of paragraph within cell,
#' one of "center" or "top" or "bottom".
#' @family formatting_shortcuts
#' @examples
#' ft_1 <- flextable(iris[c(1:3, 51:53, 101:103), ])
#' ft_1 <- theme_box(ft_1)
#' ft_1 <- merge_v(ft_1, j = 5)
#' ft_1
#'
#' ft_2 <- valign(ft_1, j = 5, valign = "top", part = "all")
#' ft_2
valign <- function(x, i = NULL, j = NULL, valign = "center", part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "valign()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- valign(x = x, i = i, j = j, valign = valign, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  x[[part]]$styles$cells[["vertical.align"]]$data[i, j] <- valign

  x
}


#' @export
#' @title Rotate cell text
#' @description It can be useful to change the text direction
#' when table headers are large. For example, header labels can
#' be rendered as "tbrl" (top to bottom and right to left), corresponding
#' to a 90-degree rotation, or "btlr", corresponding to a 270-degree
#' rotation.

#' This function changes cell text direction. By default, it is
#' "lrtb", which means from left to right and top to bottom.
#'
#' 'Word' and 'PowerPoint' do not handle automatic height with rotated headers.
#' Therefore, you need to set header heights (with the [height()] function)
#' and set the rule to "exact" for row heights (with the [hrule()] function);
#' otherwise, Word and PowerPoint outputs will have insufficient height
#' to properly display the text.
#'
#' flextable does not rotate text by arbitrary angles. It only
#' rotates by right angles (90-degree increments). This choice ensures
#' consistent rendering across Word, PowerPoint (limited to angles 0, 270, and 90),
#' HTML, and PDF.
#'
#' @inheritParams args_selectors_with_all
#' @param rotation one of "lrtb", "tbrl", "btlr".
#' @param align vertical alignment of paragraph within cell,
#' one of "center" or "top" or "bottom".
#' @details
#' When the [autofit()] function is used, rotation will be
#' ignored. In that case, use [dim_pretty] and [width] instead
#' of `autofit()`.
#' @family formatting_shortcuts
#' @examples
#' library(flextable)
#'
#' ft_1 <- flextable(head(iris))
#'
#' ft_1 <- rotate(ft_1, j = 1:4, align = "bottom", rotation = "tbrl", part = "header")
#' ft_1 <- rotate(ft_1, j = 5, align = "bottom", rotation = "btlr", part = "header")
#'
#' # if output is docx or pptx, think about (1) set header heights
#' # and (2) set rule "exact" for rows heights because Word
#' # and PowerPoint don't handle auto height with rotated headers
#' ft_1 <- height(ft_1, height = 1.2, part = "header")
#' ft_1 <- hrule(ft_1, i = 1, rule = "exact", part = "header")
#'
#' ft_1
#'
#' dat <- data.frame(
#'   a = c("left-top", "left-middle", "left-bottom"),
#'   b = c("center-top", "center-middle", "center-bottom"),
#'   c = c("right-top", "right-middle", "right-bottom")
#' )
#'
#' ft_2 <- flextable(dat)
#' ft_2 <- theme_box(ft_2)
#' ft_2 <- height_all(x = ft_2, height = 1.3, part = "body")
#' ft_2 <- hrule(ft_2, rule = "exact")
#' ft_2 <- rotate(ft_2, rotation = "tbrl")
#' ft_2 <- width(ft_2, width = 1.3)
#'
#' ft_2 <- align(ft_2, j = 1, align = "left")
#' ft_2 <- align(ft_2, j = 2, align = "center")
#' ft_2 <- align(ft_2, j = 3, align = "right")
#'
#' ft_2 <- valign(ft_2, i = 1, valign = "top")
#' ft_2 <- valign(ft_2, i = 2, valign = "center")
#' ft_2 <- valign(ft_2, i = 3, valign = "bottom")
#'
#' ft_2
rotate <- function(
  x,
  i = NULL,
  j = NULL,
  rotation,
  align = NULL,
  part = "body"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "rotate()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- rotate(x = x, i = i, j = j, rotation = rotation, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  x[[part]]$styles$cells[["text.direction"]]$data[i, j] <- rotation
  if (!is.null(align)) {
    x[[part]]$styles$cells[["vertical.align"]]$data[i, j] <- align
  }

  x
}


# misc. ----

#' @title Make blank columns transparent
#' @description Blank columns are set as transparent. This is a shortcut function
#' that deletes top and bottom borders, changes the background color to
#' transparent, displays empty content, and sets blank column widths.
#' @inheritParams args_selectors_with_all
#' @param width width of blank columns (.1 inch by default).
#' @param unit unit for width, one of "in", "cm", "mm".
#' @family formatting_shortcuts
#' @examples
#' typology <- data.frame(
#'   col_keys = c(
#'     "Sepal.Length", "Sepal.Width", "Petal.Length",
#'     "Petal.Width", "Species"
#'   ),
#'   what = c("Sepal", "Sepal", "Petal", "Petal", " "),
#'   measure = c("Length", "Width", "Length", "Width", "Species"),
#'   stringsAsFactors = FALSE
#' )
#' typology
#'
#' ftab <- flextable(head(iris), col_keys = c(
#'   "Species",
#'   "break1", "Sepal.Length", "Sepal.Width",
#'   "break2", "Petal.Length", "Petal.Width"
#' ))
#' ftab <- set_header_df(ftab, mapping = typology, key = "col_keys")
#' ftab <- merge_h(ftab, part = "header")
#' ftab <- theme_vanilla(ftab)
#' ftab <- empty_blanks(ftab)
#' ftab <- width(ftab, j = c(2, 5), width = .1)
#' ftab
#' @export
empty_blanks <- function(x, width = .05, unit = "in", part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "empty_blanks()"
    ))
  }
  if (length(x$blanks) < 1) {
    return(x)
  }

  x <- border(
    x,
    j = x$blanks,
    border.top = fp_border_default(width = 0),
    border.bottom = fp_border_default(width = 0),
    part = part
  )
  x <- bg(x, j = x$blanks, bg = "transparent", part = part)
  x <- void(x, j = x$blanks, part = part)
  x <- width(x, j = x$blanks, width = width, unit = unit)
  x
}

dat <- data.frame(
  a = c("left-top", "left-middle", "left-bottom"),
  b = c("center-top", "center-middle", "center-bottom"),
  c = c("right-top", "right-middle", "right-bottom")
)

ft_1 <- flextable(dat)
ft_1 <- theme_box(ft_1)
ft_1 <- height_all(x = ft_1, height = 1.3)
ft_1 <- hrule(ft_1, rule = "exact")
ft_1 <- rotate(ft_1, rotation = "tbrl")
ft_1 <- delete_part(ft_1)

ft_1 <- align(ft_1, j = 1, align = "left")
ft_1 <- align(ft_1, j = 2, align = "center")
ft_1 <- align(ft_1, j = 3, align = "right")

ft_1 <- valign(ft_1, i = 1, valign = "top")
ft_1 <- valign(ft_1, i = 2, valign = "center")
ft_1 <- valign(ft_1, i = 3, valign = "bottom")


test_that("docx rotations", {
  tmp_docx <- tempfile(fileext = ".docx")
  save_as_docx(ft_1, path = tmp_docx)

  doc <- read_docx(tmp_docx)

  docx <- docx_body_xml(doc)
  valign_val <- xml_find_all(docx, "w:body/w:tbl/w:tr/w:tc/w:tcPr/w:vAlign")
  valign_val <- xml_attr(valign_val, "val")

  text_direction_val <- xml_find_all(docx, "w:body/w:tbl/w:tr/w:tc/w:tcPr/w:textDirection")
  text_direction_val <- xml_attr(text_direction_val, "val")

  align_val <- xml_find_all(docx, "w:body/w:tbl/w:tr/w:tc/w:p/w:pPr/w:jc")
  align_val <- xml_attr(align_val, "val")

  align <- c("left", "center", "right")
  valign <- c("bottom", "center", "top")

  expect_equal(valign_val, rep(valign, 3))
  expect_equal(text_direction_val, rep("tbRl", 9))
  expect_equal(align_val, rep(align, each = 3))
})

test_that("pptx rotations", {
  tmp_pptx <- tempfile(fileext = ".pptx")
  save_as_pptx(ft_1, path = tmp_pptx)

  x <- read_pptx(tmp_pptx)
  slide <- x$slide$get_slide(x$cursor)

  doc <- slide$get()

  cell_prs <- xml_find_all(doc, "//a:tbl/a:tr/a:tc/a:tcPr")
  par_prs <- xml_find_all(doc, "//a:tbl/a:tr/a:tc/a:txBody/a:p/a:pPr")

  text_direction_val <- xml_attr(cell_prs, "vert")
  valign_val <- xml_attr(cell_prs, "anchor")
  align_val <- xml_attr(par_prs, "algn")

  align <- c("l", "ctr", "r")
  valign <- c("b", "ctr", "t")

  expect_equal(valign_val, rep(valign, 3))
  expect_equal(text_direction_val, rep("vert", 9))
  expect_equal(align_val, rep(align, each = 3))
})

test_that("rows selections", {
  dummy_df <- data.frame(
    my_col = rep(letters[1:3], each = 2),
    row.names = letters[21:26],
    stringsAsFactors = FALSE
  )
  ft <- flextable(dummy_df)
  expect_error(bold(ft, i = ~ my_col %in% "a", part = "header"))
  expect_error(bold(ft, i = 1L:8L), "invalid row selection")
  expect_error(bold(ft, i = -9), "invalid row selection")
  expect_error(bold(ft, i = rep(TRUE, 10)), "invalid row selection")
  expect_error(bold(ft, i = c("m", "n")), "invalid row selection")
})

test_that("columns selections", {
  ft <- flextable(iris)
  expect_error(bold(ft, j = ~ Sepalsd.Length + Species), "Sepalsd.Length")
  expect_error(bold(ft, j = 1:6), "invalid columns selection")
  expect_error(bold(ft, j = c("Sepalsd.Length")), "Sepalsd.Length")
})


test_that("part=header and formula selection for rows", {
  ft <- flextable(head(mtcars, n = 10))
  def_cell <- fp_cell(border = fp_border(color = "#00FFFF"))
  def_par <- fp_par(text.align = "center")
  expect_error(style(
    ft,
    i = ~ mpg < 20,
    pr_c = def_cell,
    pr_p = def_par,
    part = "all"
  ))
  expect_error(bg(ft, i = ~ mpg < 20, bg = "#DDDDDD", part = "header"))
  expect_error(bold(ft, i = ~ mpg < 20, bold = TRUE, part = "header"))
  expect_error(fontsize(ft, i = ~ mpg < 20, size = 10, part = "header"))
  expect_error(italic(ft, i = ~ mpg < 20, italic = TRUE, part = "header"))
  expect_error(color(ft, i = ~ mpg < 20, color = "red", part = "header"))
  expect_error(padding(ft, i = ~ mpg < 20, padding = 3, part = "header"))
  expect_error(align(ft, i = ~ mpg < 20, align = "center", part = "header"))
  expect_error(border(
    ft,
    i = ~ mpg < 20,
    border = fp_border(color = "orange"),
    part = "header"
  ))
  expect_error(rotate(
    ft,
    i = ~ mpg < 20,
    rotation = "lrtb",
    align = "top",
    part = "header"
  ))
})

get_indent_ft <- function() {
  ft <- flextable(head(cars, 2))
  ft <- padding(ft, i = 1, j = 1, padding.left = 30, part = "header")
  ft <- indentation(ft, i = 1, j = 1, hanging = 20, part = "header")
  ft <- indentation(ft, i = 1, j = 2, first_line = 15, part = "header")
  ft
}

test_that("indentation() stores first_line and hanging", {
  ft <- get_indent_ft()
  dat <- information_data_paragraph(ft)
  header <- dat[dat$.part %in% "header", ]
  expect_equal(header$hanging[header$.col_id %in% "speed"], 20)
  expect_equal(header$first_line[header$.col_id %in% "dist"], 15)
  body <- dat[dat$.part %in% "body", ]
  expect_true(all(is.na(body$hanging)))
  expect_true(all(is.na(body$first_line)))

  expect_error(indentation(ft, hanging = "z"), "must be numeric")
  expect_error(indentation(iris), "supports only flextable")
})

test_that("indents can be set with style(pr_p = fp_par(...))", {
  ft <- flextable(head(cars, 2))
  ft <- style(
    ft,
    i = 1,
    j = 1,
    pr_p = officer::fp_par(padding.left = 30, hanging = 12),
    part = "header"
  )
  dat <- information_data_paragraph(ft)
  header <- dat[dat$.part %in% "header" & dat$.row_id %in% 1, ]
  expect_equal(header$hanging[header$.col_id %in% "speed"], 12)
})

test_that("indents in wml, pml and html", {
  ft <- get_indent_ft()

  wml <- flextable:::gen_raw_wml(ft)
  expect_match(wml, "w:hanging=\"400\"", fixed = TRUE)
  expect_match(wml, "w:firstLine=\"300\"", fixed = TRUE)

  pml <- flextable:::gen_raw_pml(ft)
  expect_match(pml, "indent=\"-254000\"", fixed = TRUE)
  expect_match(pml, "indent=\"190500\"", fixed = TRUE)

  html <- flextable:::gen_raw_html(ft)
  expect_match(html, "text-indent:-20pt;", fixed = TRUE)
  expect_match(html, "text-indent:15pt;", fixed = TRUE)

  html0 <- flextable:::gen_raw_html(flextable(head(cars, 2)))
  expect_no_match(html0, "text-indent", fixed = TRUE)
})

test_that("indents in rtf", {
  ft <- get_indent_ft()
  path <- tempfile(fileext = ".rtf")
  save_as_rtf(ft, path = path)
  rtf <- paste(readLines(path, warn = FALSE), collapse = "\n")
  expect_match(rtf, "\\fi-400", fixed = TRUE)
  expect_match(rtf, "\\fi300", fixed = TRUE)
})

test_that("indents in latex", {
  ft <- get_indent_ft()
  ltx <- paste(flextable:::gen_raw_latex(ft), collapse = "\n")
  # leftskip is reduced by the hanging value: 30 - 20
  expect_match(ltx, "\\advance\\leftskip by 10.0pt", fixed = TRUE)
  expect_match(ltx, "\\hangindent=20.0pt\\hangafter=1", fixed = TRUE)
  expect_match(ltx, "\\hspace*{15.0pt}", fixed = TRUE)

  ltx0 <- paste(
    flextable:::gen_raw_latex(flextable(head(cars, 2))),
    collapse = "\n"
  )
  expect_no_match(ltx0, "hangindent", fixed = TRUE)
})

test_that("indents in typst", {
  ft <- get_indent_ft()
  ty <- paste(flextable:::gen_raw_typst(ft), collapse = "\n")
  expect_match(ty, "#par(hanging-indent: 20pt)[", fixed = TRUE)
  expect_match(ty, "#h(15pt)", fixed = TRUE)
  # inset left is reduced by the hanging value: 30 - 20
  expect_match(ty, "inset: (left: 10pt", fixed = TRUE)

  ty0 <- paste(
    flextable:::gen_raw_typst(flextable(head(cars, 2))),
    collapse = "\n"
  )
  expect_no_match(ty0, "hanging-indent", fixed = TRUE)
})

test_that("indents in grid output", {
  skip_if_not_installed("gdtools")
  gdtools::font_set_liberation()

  ft <- flextable(data.frame(species = "x", island = "y"))
  ft <- mk_par(
    ft,
    part = "header",
    i = 1,
    j = "species",
    value = as_paragraph("species with a long trailing text that wraps")
  )
  ft <- padding(ft, i = 1, j = 1, padding.left = 30, part = "header")
  ft <- align(ft, align = "left", part = "all")
  ft <- width(ft, j = 1, width = 1.6)

  ft_hang <- indentation(ft, i = 1, j = 1, hanging = 24, part = "header")
  gr1 <- gen_grob(ft_hang, fit = "fixed")

  # first line has no indent column, continuation rows carry a leading
  # indent column of 24/72 in
  rows <- gr1$children[[1]]$children$contents$children
  expect_gt(length(rows), 1)
  first_widths <- as.numeric(rows[[1]]$vp$layout$widths)
  expect_false(isTRUE(all.equal(first_widths[1], 24 / 72)))
  for (row_id in seq_along(rows)[-1]) {
    widths <- as.numeric(rows[[row_id]]$vp$layout$widths)
    expect_equal(widths[1], 24 / 72, tolerance = 1e-6)
  }

  # first_line: only the first row is indented
  ft_first <- indentation(ft, i = 1, j = 1, first_line = 24, part = "header")
  gr2 <- gen_grob(ft_first, fit = "fixed")
  rows2 <- gr2$children[[1]]$children$contents$children
  expect_equal(
    as.numeric(rows2[[1]]$vp$layout$widths)[1],
    24 / 72,
    tolerance = 1e-6
  )
  widths2 <- as.numeric(rows2[[2]]$vp$layout$widths)
  expect_false(isTRUE(all.equal(widths2[1], 24 / 72)))
})

test_that("shortcut functions", {
  ft <- flextable(head(mtcars, n = 2))

  ft <- bg(x = ft, bg = "red", j = 1, part = "all")
  ft <- bold(x = ft, bold = TRUE, j = 2, part = "all")
  ft <- fontsize(x = ft, size = 9, j = 3, part = "all")
  ft <- italic(x = ft, italic = TRUE, j = 4, part = "all")
  ft <- color(x = ft, color = "gray", j = 5, part = "all")
  ft <- padding(x = ft, padding = 5, j = 6, part = "all")
  ft <- align(x = ft, align = "left", j = 7, part = "all")
  ft <- border(
    x = ft,
    border = fp_border(color = "orange"),
    j = 8,
    part = "all"
  )

  docx_file <- tempfile(fileext = ".docx")

  doc <- read_docx()
  doc <- body_add_flextable(doc, value = ft)
  doc <- print(doc, target = docx_file)

  main_folder <- file.path(getwd(), "docx_folder")
  unpack_folder(file = docx_file, folder = main_folder)

  doc_file <- file.path(main_folder, "/word/document.xml")
  doc <- read_xml(doc_file)

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[1]/w:tcPr/w:shd")
  expect_equal(xml_attr(nodes, "fill"), c("FF0000", "FF0000", "FF0000"))

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[2]/w:p/w:r/w:rPr/w:b")
  expect_length(nodes, 3)

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[3]/w:p/w:r/w:rPr/w:sz")
  expect_equal(xml_attr(nodes, "val"), c("18", "18", "18"))

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[4]/w:p/w:r/w:rPr/w:i")
  expect_length(nodes, 3)

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[5]/w:p/w:r/w:rPr/w:color")
  expect_equal(xml_attr(nodes, "val"), c("BEBEBE", "BEBEBE", "BEBEBE"))

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[6]/w:p/w:pPr/w:spacing")
  expect_equal(xml_attr(nodes, "after"), rep("100", 3))
  expect_equal(xml_attr(nodes, "before"), rep("100", 3))
  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[6]/w:p/w:pPr/w:ind")
  expect_equal(xml_attr(nodes, "left"), rep("100", 3))
  expect_equal(xml_attr(nodes, "right"), rep("100", 3))

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[7]/w:p/w:pPr/w:jc")
  expect_equal(xml_attr(nodes, "val"), rep("left", 3))

  nodes <- xml_find_all(doc, "w:body/w:tbl/w:tr/w:tc[8]/w:tcPr/w:tcBorders/*")
  expect_true(all(xml_attr(nodes, "color") == "FFA500"))

  unlink(main_folder, recursive = TRUE, force = TRUE)
})

tab <- data.frame(x = c("Row1", "Row2"), y = c(1, 2))

ft <- flextable(tab)
ft <- border_remove(ft)
ft <- hline(ft, i = 1:2, j = 2, part = "body")
ft <- delete_part(ft, part = "header")

test_that("borders with office docs are sanitized", {
  docx_file <- tempfile(fileext = ".docx")
  pptx_file <- tempfile(fileext = ".pptx")
  save_as_docx(ft, path = docx_file)
  save_as_pptx(ft, path = pptx_file)

  docx <- read_docx(docx_file)
  doc <- docx_body_xml(docx)

  top_nodes <- xml_find_all(
    doc,
    "w:body/w:tbl/w:tr/w:tc/w:tcPr/w:tcBorders/w:top"
  )
  bot_nodes <- xml_find_all(
    doc,
    "w:body/w:tbl/w:tr/w:tc/w:tcPr/w:tcBorders/w:bottom"
  )
  expect_equal(
    xml_attr(top_nodes, "color"),
    c("000000", "000000", "000000", "666666")
  )
  expect_equal(
    xml_attr(bot_nodes, "color"),
    c("000000", "666666", "000000", "666666")
  )

  pptx <- read_pptx(pptx_file)
  slide <- pptx$slide$get_slide(1)$get()

  top_nodes <- xml_find_all(slide, "//a:tbl/a:tr/a:tc/a:tcPr/a:lnT")
  bot_nodes <- xml_find_all(slide, "//a:tbl/a:tr/a:tc/a:tcPr/a:lnB")
  expect_equal(xml_attr(top_nodes, "w"), c("0", "0", "0", "12700"))
  expect_equal(xml_attr(bot_nodes, "w"), c("0", "12700", "0", "12700"))
})

test_that("align() accepts default align argument when columns is not a multiple of 4", {
  ft <- flextable(head(mtcars, n = 2)[, 1:6])
  ft1 <- align(ft, part = "all")
  expect_equal(
    rep("left", 18),
    information_data_paragraph(ft1)$text.align
  )
})

test_that("align() accepts combinations of align arguments.", {
  ft <- flextable(head(mtcars, n = 2)[, 1:6])

  # All columns right-aligned
  ft2 <- align(ft, align = "right", part = "all")
  expect_equal(
    rep("right", 18),
    information_data_paragraph(ft2)$text.align
  )

  # Custom alignment for each column
  custom_alignment <- c("left", "right", "left", "center", "center", "right")
  ft3 <- align(ft, align = custom_alignment, part = "all")
  expect_equal(
    rep(custom_alignment, 3),
    information_data_paragraph(ft3)$text.align
  )

  # Custom alignment for only columns 3 and 5 in body only
  custom_alignment <- c("center", "left")
  ft4 <- align(
    ft,
    j = c("disp", "drat"),
    align = custom_alignment,
    part = "body"
  )
  subdat <- information_data_paragraph(ft4)
  subdat <- subdat[subdat$.col_id %in% c("disp", "drat"), ]
  subdat <- subdat[subdat$.part %in% c("body"), ]
  expect_equal(
    rep(custom_alignment, 2),
    subdat$text.align
  )

  # Custom alignment for only columns 3 and 5 in body only (using default body arg)
  ft4b <- align(ft, j = c("disp", "drat"), align = custom_alignment)
  subdat <- information_data_paragraph(ft4b)
  subdat <- subdat[subdat$.col_id %in% c("disp", "drat"), ]
  subdat <- subdat[subdat$.part %in% c("body"), ]
  expect_equal(
    rep(custom_alignment, 2),
    subdat$text.align
  )

  # Center alignment for only columns 3 and 5
  ft5 <- align(ft, j = c("disp", "drat"), align = "center", part = "all")
  subdat <- information_data_paragraph(ft5)
  subdat <- subdat[subdat$.col_id %in% c("disp", "drat"), ]
  expect_equal(
    rep("center", 6),
    subdat$text.align
  )

  # Alternate left and center alignment across columns 1-4 for header only
  ft6 <- align(ft, j = 1:4, align = c("left", "center"), part = "header")
  subdat <- information_data_paragraph(ft6)
  subdat <- subdat[subdat$.part %in% c("header"), ]
  expect_equal(
    c(rep(c("left", "center"), 2), "right", "right"),
    subdat$text.align
  )
})

test_that("align() will error if invalid align and part arguments are supplied", {
  ft <- flextable(head(mtcars, n = 2)[, 1:6])

  # Invalid "part" argument
  expect_error(align(
    ft,
    align = c("left", "center", "right"),
    part = "everything"
  ))

  # Invalid "align" argument
  expect_error(align(ft, align = "top"))

  # Invalid "align" argument mixed in with valid arguments throws warning
  expect_error(align(ft, align = c("top", "left")), "Invalid `align` argument")
})

# Liberation Sans is registered so the emitted `font: "..."` is deterministic
# across machines (mirrors test-latex.R).
gdtools::register_liberationsans()

# build a flextable with a fixed font on every part
ft_lib <- function(df) {
  ft <- flextable(df)
  font(ft, fontname = "Liberation Sans", part = "all")
}

# a small real PNG file (path-based image, no raster rendering needed)
make_png <- function() {
  f <- tempfile(fileext = ".png")
  grDevices::png(f, width = 60, height = 40)
  old <- graphics::par(mar = c(0, 0, 0, 0))
  graphics::plot.new()
  graphics::rect(0, 0, 1, 1, col = "grey")
  graphics::par(old)
  grDevices::dev.off()
  f
}

# -- table structure ---------------------------------------------------------
test_that("gen_raw_typst emits a Typst table skeleton", {
  str <- flextable:::gen_raw_typst(ft_lib(head(iris, 2)))
  expect_type(str, "character")
  expect_match(str, "#table(", fixed = TRUE)
  expect_match(str, "columns:", fixed = TRUE)
  expect_match(str, "stroke: none", fixed = TRUE)
  expect_match(str, "table.cell(", fixed = TRUE)
  expect_match(str, "table.header(", fixed = TRUE)
})

test_that("font family is emitted from the cell property", {
  str <- flextable:::gen_raw_typst(ft_lib(head(iris, 1)))
  expect_match(str, 'font: "Liberation Sans"', fixed = TRUE)
})

# -- fills, borders, alignment ----------------------------------------------
test_that("background colour becomes fill: rgb(...)", {
  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- bg(ft, j = "a", bg = "#FFEEAA", part = "body")
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, 'fill: rgb("FFEEAA")', fixed = TRUE)
})

test_that("borders become per-side stroke entries", {
  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- hline(
    ft,
    j = "a",
    border = fp_border_default(color = "red", width = 2),
    part = "body"
  )
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "stroke: (", fixed = TRUE)
  expect_match(str, 'rgb("FF0000")', fixed = TRUE)
  # a side without border is emitted as none
  expect_match(str, "left: none", fixed = TRUE)
})

test_that("alignment maps to Typst align", {
  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- align(ft, j = "a", align = "center", part = "body")
  ft <- valign(ft, j = "a", valign = "top", part = "body")
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "align: center + top", fixed = TRUE)
})

# -- text run styling --------------------------------------------------------
test_that("bold / italic / colour render as #text args", {
  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- bold(ft, j = "a", part = "body")
  ft <- italic(ft, j = "a", part = "body")
  ft <- color(ft, j = "a", color = "red", part = "body")
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, 'weight: "bold"', fixed = TRUE)
  expect_match(str, 'style: "italic"', fixed = TRUE)
  expect_match(str, 'fill: rgb("FF0000")', fixed = TRUE)
})

test_that("underline, strike and super/subscript wrap the run", {
  mk <- function(props) {
    ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
    ft <- compose(
      ft,
      j = "a",
      value = as_paragraph(as_chunk("x", props = props))
    )
    flextable:::gen_raw_typst(ft)
  }
  expect_match(
    mk(fp_text_default(underlined = TRUE)),
    "#underline[",
    fixed = TRUE
  )
  expect_match(mk(fp_text_default(strike = TRUE)), "#strike[", fixed = TRUE)
  expect_match(
    mk(fp_text_default(vertical.align = "superscript")),
    "#super[",
    fixed = TRUE
  )
  expect_match(
    mk(fp_text_default(vertical.align = "subscript")),
    "#sub[",
    fixed = TRUE
  )
})

test_that("special characters are escaped in text", {
  ft <- ft_lib(data.frame(a = "x#y*z", stringsAsFactors = FALSE))
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "x\\#y\\*z", fixed = TRUE)
})

test_that("markup shorthand triggers are escaped (no dash/ellipsis/quote subst)", {
  # Typst would otherwise turn --/---/... into en/em dash/ellipsis, `-` before
  # a digit into a minus sign, and '/" into curly quotes. Each trigger char is
  # escaped individually so the literal text is preserved.
  mk <- function(txt) {
    flextable:::gen_raw_typst(ft_lib(data.frame(
      a = txt,
      stringsAsFactors = FALSE
    )))
  }
  expect_match(mk("a--b"), "a\\-\\-b", fixed = TRUE)
  expect_match(mk("a---b"), "a\\-\\-\\-b", fixed = TRUE)
  expect_match(mk("x...y"), "x\\.\\.\\.y", fixed = TRUE)
  expect_match(mk("-5"), "\\-5", fixed = TRUE)
  expect_match(mk("q\"u'v"), "q\\\"u\\'v", fixed = TRUE)
})

test_that("line-start markers are escaped (no heading/enum/term marker)", {
  # `=`/`+`/`/` are markup markers at a line start; `/` even fails to compile.
  # Escaping every occurrence keeps the literal text and is safe mid-line too.
  mk <- function(txt) {
    flextable:::gen_raw_typst(ft_lib(data.frame(
      a = txt,
      stringsAsFactors = FALSE
    )))
  }
  expect_match(mk("= Titre"), "\\= Titre", fixed = TRUE)
  expect_match(mk("+ item"), "\\+ item", fixed = TRUE)
  expect_match(mk("/ terme"), "\\/ terme", fixed = TRUE)
  expect_match(mk("a/b"), "a\\/b", fixed = TRUE)
})

# -- spans -------------------------------------------------------------------
test_that("horizontal merge produces colspan", {
  ft <- ft_lib(data.frame(a = "x", b = "y", stringsAsFactors = FALSE))
  ft <- merge_at(ft, i = 1, j = 1:2, part = "header")
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "colspan: 2", fixed = TRUE)
})

test_that("vertical merge produces rowspan", {
  ft <- ft_lib(data.frame(a = c("X", "X"), b = 1:2, stringsAsFactors = FALSE))
  ft <- merge_v(ft, j = "a")
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "rowspan: 2", fixed = TRUE)
})

# -- equations ---------------------------------------------------------------
test_that("equations become mitex #mi(`...`)", {
  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- compose(ft, j = "a", value = as_paragraph(as_equation("\\alpha")))
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "#mi(`\\alpha`)", fixed = TRUE)
})

# -- as_qmd ------------------------------------------------------------------
test_that("as_qmd emits a base64 marker, not wrapped in #text", {
  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- compose(ft, j = "a", value = as_paragraph(as_qmd("**bold**")))
  str <- flextable:::gen_raw_typst(ft)
  expected <- sprintf("/*tblqmd:%s*/", officer::as_base64("**bold**"))
  expect_match(str, expected, fixed = TRUE)
  # marker sits directly inside the cell content (no #text wrapper)
  expect_match(str, "[/*tblqmd:", fixed = TRUE)
})

# -- rotation ----------------------------------------------------------------
test_that("text rotation wraps content in #rotate(..., reflow: true)", {
  ft <- ft_lib(data.frame(tb = 1, bt = 2, stringsAsFactors = FALSE))
  ft <- rotate(ft, j = "tb", rotation = "tbrl", part = "header")
  ft <- rotate(ft, j = "bt", rotation = "btlr", part = "header")
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, "#rotate(90deg, reflow: true)[", fixed = TRUE)
  expect_match(str, "#rotate(-90deg, reflow: true)[", fixed = TRUE)
})

# -- images: path mode (Quarto) vs embed mode (standalone) -------------------
test_that("image path mode emits #image(\"path\")", {
  # path mode writes the image under fig_path()'s dir ("figure/" by default);
  # run in a throwaway working dir so nothing leaks into the package sources.
  withr::local_dir(withr::local_tempdir())
  img <- make_png()
  ft <- ft_lib(data.frame(a = "", stringsAsFactors = FALSE))
  ft <- compose(
    ft,
    j = "a",
    value = as_paragraph(as_image(img, width = .5, height = .3))
  )
  str <- flextable:::gen_raw_typst(ft)
  expect_match(str, '#image("', fixed = TRUE)
  expect_no_match(str, "ft-b64-decode", fixed = TRUE)
})

test_that("image embed mode emits base64 decoder call", {
  img <- make_png()
  ft <- ft_lib(data.frame(a = "", stringsAsFactors = FALSE))
  ft <- compose(
    ft,
    j = "a",
    value = as_paragraph(as_image(img, width = .5, height = .3))
  )
  str <- flextable:::gen_raw_typst(ft, image_mode = "embed")
  expect_match(str, "#image(ft-b64-decode(\"", fixed = TRUE)
  # no `format:` is emitted: a bytes source is auto-detected by Typst, and the
  # mime subtype ("jpeg"/"svg+xml") is not a valid Typst format string.
  expect_no_match(str, "format:", fixed = TRUE)
})

# -- knit_to_typst -----------------------------------------------------------
test_that("knit_to_typst returns character and injects mitex on equations", {
  z <- flextable:::knit_to_typst(ft_lib(head(iris, 2)))
  expect_type(z, "character")
  expect_no_match(z, "@preview/mitex", fixed = TRUE)

  ft <- ft_lib(data.frame(a = "x", stringsAsFactors = FALSE))
  ft <- compose(ft, j = "a", value = as_paragraph(as_equation("\\alpha")))
  z <- flextable:::knit_to_typst(ft)
  expect_match(z, "@preview/mitex", fixed = TRUE)
})

# -- save_as_typst -----------------------------------------------------------
test_that("save_as_typst writes a standalone document", {
  tf <- tempfile(fileext = ".typ")
  res <- save_as_typst(ft_lib(head(iris, 2)), path = tf)
  expect_equal(res, tf)
  lines <- readLines(tf)
  doc <- paste(lines, collapse = "\n")
  expect_match(doc, "#set page", fixed = TRUE)
  expect_match(doc, "#table(", fixed = TRUE)
  # no image / no equation -> no decoder, no mitex import
  expect_no_match(doc, "ft-b64-decode", fixed = TRUE)
  expect_no_match(doc, "@preview/mitex", fixed = TRUE)
})

test_that("save_as_typst injects the base64 decoder when images are embedded", {
  img <- make_png()
  ft <- ft_lib(data.frame(a = "", stringsAsFactors = FALSE))
  ft <- compose(
    ft,
    j = "a",
    value = as_paragraph(as_image(img, width = .5, height = .3))
  )
  tf <- tempfile(fileext = ".typ")
  save_as_typst(ft, path = tf)
  doc <- paste(readLines(tf), collapse = "\n")
  expect_match(doc, "#let ft-b64-decode", fixed = TRUE)
  expect_match(doc, "#image(ft-b64-decode(\"", fixed = TRUE)
})

test_that("save_as_typst errors without a flextable", {
  expect_error(save_as_typst(1, path = tempfile()), "no flextable")
})

test_that("cells are not justified unless text.align is justify", {
  ft <- ft_lib(data.frame(a = "some long text", stringsAsFactors = FALSE))
  ty <- paste(flextable:::gen_raw_typst(ft), collapse = "\n")
  # Quarto sets par(justify: true) document-wide; the table opts out
  expect_match(ty, "#set par(justify: false)", fixed = TRUE)
  expect_no_match(ty, "justify: true", fixed = TRUE)

  ft <- align(ft, align = "justify", part = "body")
  ty <- paste(flextable:::gen_raw_typst(ft), collapse = "\n")
  expect_match(ty, "#par(justify: true)[", fixed = TRUE)
})

# note_docx_images -----

#' @title internal utils for roxygen tags reuse
#' @note
#' This chunk option requires package officedown in a R Markdown
#' context with Word output format. With Quarto (`format: docx`) or
#' `rmarkdown::word_document()`, the resulting file must be repaired
#' with [repair_docx()].
#'
#' PowerPoint cannot mix images and text in a paragraph, images
#' are removed when outputing to PowerPoint format.
#' @name note_docx_images
#' @keywords internal
NULL

#' @importFrom grDevices as.raster
#' @export
#' @title Image chunk
#' @description The function lets add images within flextable
#' objects with functions:
#' - [compose()] and [as_paragraph()],
#' - [append_chunks()],
#' - [prepend_chunks()]
#' \if{html}{\out{
#' <img src="https://www.ardata.fr/img/flextable-imgs/flextable-006.png" alt="as_image illustration" style="width:100\%;">
#' }}
#' @param src image filename
#' @param width,height size of the image file. It can be ignored
#' if parameter `guess_size=TRUE`, see parameter `guess_size`.
#' @param guess_size If package 'magick' is installed, this option
#' can be used (set it to `TRUE` and don't provide values for paramters
#' `width` and `height`). When the flextable will be printed,
#' the images will be read and width and height will be guessed. This
#' should be avoid if possible as it can be an extensive task when
#' several images.
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param alt alternative text for the image (used for accessibility)
#' @param ... unused argument
#' @family chunk elements for paragraph
#' @inherit note_docx_images note
#' @seealso [compose()], [as_paragraph()]
#' @examples
#' img.file <- file.path(
#'   R.home("doc"),
#'   "html", "logo.jpg"
#' )
#' if (require("magick")) {
#'   myft <- flextable(head(iris))
#'   myft <- compose(myft,
#'     i = 1:3, j = 1,
#'     value = as_paragraph(
#'       as_image(src = img.file),
#'       " ",
#'       as_chunk(Sepal.Length,
#'         props = fp_text_default(color = "red")
#'       )
#'     ),
#'     part = "body"
#'   )
#'   ft <- autofit(myft)
#'   ft
#' }
as_image <- function(
  src,
  width = NULL,
  height = NULL,
  unit = "in",
  guess_size = TRUE,
  alt = "",
  ...
) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (length(src) > 1) {
    if (length(width) == 1) {
      width <- rep(width, length(src))
    }
    if (length(height) == 1) height <- rep(height, length(src))
  }

  if (guess_size && (is.null(width) || is.null(height))) {
    if (!requireNamespace("magick", quietly = TRUE)) {
      stop("package magick is required when using `guess_size` option.")
    }
    sizes <- lapply(src, function(x) {
      z <- magick::image_read(x)
      z <- magick::image_data(z)
      attr(z, "dim")[-1]
    })
    sizes <- do.call(rbind, sizes)
    width <- sizes[, 1] / 72
    height <- sizes[, 2] / 72
  }

  alt <- rep_len(alt, length(src))
  data <- chunk_dataframe(
    width = as.double(width),
    height = as.double(height),
    img_data = src,
    alt = alt
  )
  class(data) <- c("img_src", class(data))
  data
}


#' @export
#' @title Mini barplot chunk
#' @description This function is used to insert bars into
#' flextable with functions:
#' - [compose()] and [as_paragraph()],
#' - [append_chunks()],
#' - [prepend_chunks()].
#' @param value values containing the bar size
#' @param max max bar size
#' @param barcol bar color
#' @param bg background color
#' @param width,height size of the resulting png file in inches
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param alt alternative text for the image (used for accessibility)
#' @inherit note_docx_images note
#' @family chunk elements for paragraph
#' @examples
#' ft <- flextable(head(iris, n = 10))
#'
#' ft <- compose(ft,
#'   j = 1,
#'   value = as_paragraph(
#'     minibar(value = Sepal.Length, max = max(Sepal.Length))
#'   ),
#'   part = "body"
#' )
#'
#' ft <- autofit(ft)
#' ft
#' @importFrom grDevices as.raster col2rgb rgb
#' @seealso [compose()], [as_paragraph()]
minibar <- function(
  value,
  max = NULL,
  barcol = "#CCCCCC",
  bg = "transparent",
  width = 1,
  height = .2,
  unit = "in",
  alt = ""
) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (all(is.na(value))) {
    max <- 1
  }
  if (all(is.na(value)) || all(value == 0, na.rm = TRUE)) {
    max <- 1
  }

  value[is.na(value)] <- 0

  if (is.null(max)) {
    max <- max(value, na.rm = TRUE)
  }

  stopifnot(value >= 0, max >= 0)
  barcol <- rgb(t(col2rgb(barcol)) / 255)
  bg <- ifelse(bg == "transparent", bg, rgb(t(col2rgb(bg)) / 255))
  if (any(value > max)) {
    warning("value > max, truncate to max")
    value[value > max] <- max
  }
  width_ <- as.integer(width * 72)
  value <- as.integer((value / max) * width_)
  n_empty <- width_ - value

  rasters <- mapply(
    function(count_on, count_off, bg_on, bg_off) {
      as.raster(matrix(
        c(rep(bg_on, count_on), rep(bg_off, count_off)),
        nrow = 1
      ))
    },
    value,
    n_empty,
    bg_on = barcol,
    bg_off = bg,
    SIMPLIFY = FALSE
  )
  alt <- rep_len(alt, length(value))
  z <- chunk_dataframe(
    width = as.double(rep(width, length(value))),
    height = as.double(rep(height, length(value))),
    img_data = rasters,
    alt = alt
  )

  class(z) <- c("img_chunk", class(z))
  z
}

#' @export
#' @title Mini linerange chunk
#' @description This function is used to insert lineranges into
#' flextable with functions:
#' - [compose()] and [as_paragraph()],
#' - [append_chunks()],
#' - [prepend_chunks()].
#' @param value values containing the bar size
#' @param min min bar size. Default min of value
#' @param max max bar size. Default max of value
#' @param rangecol bar color
#' @param stickcol jauge color
#' @param bg background color
#' @param width,height size of the resulting png file in inches
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param alt alternative text for the image (used for accessibility)
#' @param raster_width number of pixels used as width
#' when interpolating value.
#' @inherit note_docx_images note
#' @family chunk elements for paragraph
#' @examples
#' myft <- flextable(head(iris, n = 10))
#'
#' myft <- compose(myft,
#'   j = 1,
#'   value = as_paragraph(
#'     linerange(value = Sepal.Length)
#'   ),
#'   part = "body"
#' )
#'
#' autofit(myft)
#' @importFrom grDevices as.raster col2rgb rgb
#' @importFrom stats approx
#' @seealso [compose()], [as_paragraph()]
linerange <- function(
  value,
  min = NULL,
  max = NULL,
  rangecol = "#CCCCCC",
  stickcol = "#FF0000",
  bg = "transparent",
  width = 1,
  height = .2,
  raster_width = 30,
  unit = "in",
  alt = ""
) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (all(is.na(value))) {
    min <- 0
    max <- 1
  }

  if (raster_width < 2) {
    stop("raster_width must be greater than 1")
  }

  raster_nrow <- 9
  raster_center <- 5

  if (is.null(max)) {
    max <- max(value, na.rm = TRUE)
  }
  if (is.null(min)) {
    min <- min(value, na.rm = TRUE)
  }

  value[!is.finite(value)] <- max + 1 # to be sure not displayed

  stopifnot(!is.null(value), !is.null(min), !is.null(min))

  # transform color code
  stickcol <- rgb(t(col2rgb(stickcol)) / 255)
  rangecol <- rgb(t(col2rgb(rangecol)) / 255)
  bg <- ifelse(bg == "transparent", bg, rgb(t(col2rgb(bg)) / 255))

  # get value approx on range 1,raster_width
  stick_pos <- as.integer(
    approx(x = c(min, max), y = c(1, raster_width), xout = value)$y
  )
  base <- matrix(bg, nrow = raster_nrow, ncol = raster_width)
  base[, 1] <- rangecol
  base[, raster_width] <- rangecol
  base[raster_center, ] <- rangecol

  rasters <- lapply(
    stick_pos,
    function(val, def_mat, col) {
      newmat <- def_mat
      newmat[, val] <- col
      as.raster(newmat)
    },
    base,
    stickcol
  )

  alt <- rep_len(alt, length(value))
  z <- chunk_dataframe(
    width = as.double(rep(width, length(value))),
    height = as.double(rep(height, length(value))),
    img_data = rasters,
    alt = alt
  )

  class(z) <- c("img_chunk", class(z))
  z
}

#' @export
#' @title Mini plot chunk
#' @description This function is used to insert mini plots into
#' flextable with functions:
#' - [compose()] and [as_paragraph()],
#' - [append_chunks()],
#' - [prepend_chunks()].
#'
#' Available plots are 'box', 'line', 'points', 'density'.
#' @param value a numeric vector, stored in a list column.
#' @param width,height size of the resulting png file in inches
#' @param type type of the plot: 'box', 'line', 'points' or 'density'.
#' @param free_scale Should scales be free (TRUE or FALSE, the default value).
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param alt alternative text for the image (used for accessibility)
#' @param ... arguments sent to plot functions (see [par()])
#' @inherit note_docx_images note
#' @family chunk elements for paragraph
#' @examples
#' library(data.table)
#' library(flextable)
#'
#' z <- as.data.table(iris)
#' z <- z[, list(
#'   Sepal.Length = mean(Sepal.Length, na.rm = TRUE),
#'   z = list(.SD$Sepal.Length)
#' ), by = "Species"]
#'
#' ft <- flextable(z,
#'   col_keys = c("Species", "Sepal.Length", "box", "density")
#' )
#' ft <- mk_par(ft, j = "box", value = as_paragraph(
#'   plot_chunk(
#'     value = z, type = "box",
#'     border = "red", col = "transparent"
#'   )
#' ))
#' ft <- mk_par(ft, j = "density", value = as_paragraph(
#'   plot_chunk(value = z, type = "dens", col = "red")
#' ))
#' ft <- set_table_properties(ft, layout = "autofit", width = .6)
#' ft <- set_header_labels(ft, box = "boxplot", density = "density")
#' theme_vanilla(ft)
#' @importFrom grDevices dev.off dev.list
#' @importFrom graphics boxplot
#' @importFrom stats density
plot_chunk <- function(
  value,
  width = 1,
  height = .2,
  type = "box",
  free_scale = FALSE,
  unit = "in",
  alt = "",
  ...
) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  type <- match.arg(
    arg = type,
    choices = c("box", "line", "points", "density"),
    several.ok = FALSE
  )

  width <- as.double(rep(width, length(value)))
  height <- as.double(rep(height, length(value)))

  params <- list(...)
  params <- append(
    params,
    list(
      xlab = "",
      ylab = "",
      main = "",
      axes = FALSE
    )
  )

  lims <- NULL
  if (!free_scale) {
    lims <- range(unlist(value), na.rm = TRUE)
  }

  files <- mapply(
    function(x, width, height, type) {
      # pointsize = 12 matches agg_png's previous default (plot_in_png uses 11)
      plot_in_png(
        code = {
          par(mar = rep(0, 4))
          parcall <- params

          if ("box" %in% type) {
            parcall$x <- x
            parcall$horizontal <- TRUE
            parcall$ylim <- lims
            do.call(boxplot, parcall)
          } else if ("line" %in% type) {
            parcall$x <- seq_along(x)
            parcall$y <- x
            parcall$type <- "l"
            parcall$ylim <- lims
            do.call(plot, parcall)
          } else if ("points" %in% type) {
            parcall$x <- seq_along(x)
            parcall$y <- x
            parcall$ylim <- lims
            do.call(plot, parcall)
          } else if ("density" %in% type) {
            parcall$x <- density(x)
            parcall$xlim <- lims
            do.call(plot, parcall)
          }
        },
        width = width,
        height = height,
        res = 200,
        units = "in",
        pointsize = 12,
        path = tempfile(fileext = ".png")
      )
    },
    x = value,
    width = width,
    height = height,
    SIMPLIFY = FALSE,
    MoreArgs = list(type = type)
  )

  files <- as.character(unlist(files))

  alt <- rep_len(alt, length(value))
  z <- chunk_dataframe(
    width = width,
    height = height,
    img_data = files,
    alt = alt
  )
  class(z) <- c("img_chunk", class(z))
  z
}


#' @export
#' @title ggplot chunk
#' @description This function is used to insert mini gg plots into
#' flextable with functions:
#' - [compose()] and [as_paragraph()],
#' - [append_chunks()],
#' - [prepend_chunks()].
#'
#' @param value gg objects, stored in a list column; or a list
#' of 'ggplot' objects.
#' @param width,height size of the resulting png file.
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param res resolution of the png image in ppi
#' @param alt alternative text for the image (used for accessibility)
#' @inherit note_docx_images note
#' @family chunk elements for paragraph
#' @examples
#' library(data.table)
#' library(flextable)
#' if (require("ggplot2")) {
#'   my_cor_plot <- function(x) {
#'     cols <- colnames(x)[sapply(x, is.numeric)]
#'     x <- x[, .SD, .SDcols = cols]
#'     cormat <- as.data.table(cor(x))
#'     cormat$var1 <- colnames(cormat)
#'     cormat <- melt(cormat,
#'       id.vars = "var1", measure.vars = cormat$var1,
#'       variable.name = "var2", value.name = "correlation"
#'     )
#'     ggplot(data = cormat, aes(x = var1, y = var2, fill = correlation)) +
#'       geom_tile() +
#'       coord_equal() +
#'       scale_fill_gradient2(
#'         low = "blue",
#'         mid = "white", high = "red", limits = c(-1, 1),
#'         guide = "none"
#'       ) +
#'       theme_void()
#'   }
#'   z <- as.data.table(iris)
#'   z <- z[, list(gg = list(my_cor_plot(.SD))), by = "Species"]
#'   ft <- flextable(z)
#'   ft <- mk_par(ft,
#'     j = "gg",
#'     value = as_paragraph(
#'       gg_chunk(value = gg, width = 1, height = 1)
#'     )
#'   )
#'   ft
#' }
gg_chunk <- function(
  value,
  width = 1,
  height = .2,
  unit = "in",
  res = 300,
  alt = ""
) {
  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  width <- as.double(rep(width, length(value)))
  height <- as.double(rep(height, length(value)))

  files <- mapply(
    function(x, width, height) {
      plot_in_png(
        ggobj = x,
        width = width,
        height = height,
        units = "in",
        res = res,
        path = tempfile(fileext = ".png")
      )
    },
    x = value,
    width = width,
    height = height,
    SIMPLIFY = FALSE
  )

  files <- as.character(unlist(files))

  alt <- rep_len(alt, length(value))
  z <- chunk_dataframe(
    width = width,
    height = height,
    img_data = files,
    alt = alt
  )
  class(z) <- c("img_chunk", class(z))
  z
}

#' @export
#' @title Grid Graphics chunk
#' @description This function is used to insert grid objects into
#' flextable with functions:
#' - [compose()] and [as_paragraph()],
#' - [append_chunks()],
#' - [prepend_chunks()].
#'
#' @param value grid objects, stored in a list column; or a list
#' of grid objects.
#' @param width,height size of the resulting png file
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param res resolution of the png image in ppi
#' @param alt alternative text for the image (used for accessibility)
#' @inherit note_docx_images note
#' @family chunk elements for paragraph
#' @examples
#' library(flextable)
#' ft_1 <- flextable(head(cars))
#' if (require("grid")) {
#'   ft_1 <- prepend_chunks(
#'     x = ft_1, i = 2, j = 2,
#'     grid_chunk(
#'       list(
#'         circleGrob(gp = gpar(
#'           fill = "#ec11c2",
#'           col = "transparent"
#'         ))
#'       ),
#'       width = .15, height = .15
#'     )
#'   )
#' }
#' ft_1
grid_chunk <- function(
  value,
  width = 1,
  height = .2,
  unit = "in",
  res = 300,
  alt = ""
) {
  if (!requireNamespace("grid", quietly = TRUE)) {
    stop(sprintf(
      "'%s' package should be installed to create a flextable from an object of type '%s'.",
      "grid",
      "grid"
    ))
  }

  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  width <- as.double(rep(width, length(value)))
  height <- as.double(rep(height, length(value)))
  files <- mapply(
    function(x, width, height) {
      plot_in_png(
        code = grid::grid.draw(x),
        width = width,
        height = height,
        units = "in",
        res = res,
        path = tempfile(fileext = ".png")
      )
    },
    x = value,
    width = width,
    height = height,
    SIMPLIFY = FALSE
  )

  files <- as.character(unlist(files))

  alt <- rep_len(alt, length(value))
  z <- chunk_dataframe(
    width = width,
    height = height,
    img_data = files,
    alt = alt
  )
  class(z) <- c("img_chunk", class(z))
  z
}

data <- iris[c(1:3, 51:53, 101:104), ]
col_keys <- c(
  "Species",
  "sep_1",
  "Sepal.Length",
  "Sepal.Width",
  "sep_2",
  "Petal.Length",
  "Petal.Width"
)
img.file <- file.path(R.home("doc"), "html", "logo.jpg")

rlogo <- tempfile(fileext = ".jpg")
file.copy(img.file, rlogo)

test_that("images", {
  ft <- flextable(data, col_keys = col_keys)
  ft <- compose(
    ft,
    j = "Sepal.Length",
    value = as_paragraph(
      as_chunk("blah blah "),
      as_image(rlogo, width = .3, height = 0.23),
      " ",
      as_chunk(
        sprintf("val: %.1f", Sepal.Length),
        props = fp_text(color = "orange", vertical.align = "superscript")
      )
    )
  )
  ft <- compose(
    ft,
    j = "sep_1",
    value = as_paragraph(
      as_image(rlogo, width = .3, height = 0.23)
    )
  )
  ft <- compose(
    ft,
    j = "Petal.Length",
    value = as_paragraph(
      "blah blah ",
      as_chunk(
        Sepal.Length,
        props = fp_text(color = "orange", vertical.align = "superscript")
      )
    )
  )
  ft <- style(
    ft,
    pr_c = fp_cell(margin = 0, border = fp_border(width = 0)),
    pr_p = fp_par(padding = 0, border = fp_border(width = 0)),
    pr_t = fp_text(font.size = 10),
    part = "all"
  )
  ft <- autofit(ft, add_w = 0, add_h = 0)

  dims <- ft$body$colwidths
  expect_equal(as.vector(dims["sep_1"]), .3, tolerance = .00001)

  docx_file <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- body_add_flextable(doc, value = ft)
  expect_error(
    {
      print(doc, target = docx_file)
    },
    NA
  )
})

plot1 <- tempfile(fileext = ".png")
plot2 <- tempfile(fileext = ".png")
ragg::agg_png(filename = plot1, width = 300, height = 300, units = "px")
plot(1:15, 1:15)
dev.off()
ragg::agg_png(filename = plot2, width = 300, height = 300, units = "px")
plot(1:150, 1:150)
dev.off()

df <- data.frame(
  plot = c(plot1, plot2)
)

test_that("multiple images", {
  skip_if_not_installed("magick")

  ft <- flextable(df)
  ft <- mk_par(
    ft,
    j = "plot",
    value = as_paragraph(as_image(rlogo, width = .3, height = 0.23)),
    part = "header"
  )
  ft <- mk_par(
    ft,
    j = "plot",
    value = as_paragraph(as_image(plot, guess_size = TRUE))
  )
  chunk_info <- flextable::information_data_chunk(ft)
  expect_equal(chunk_info$img_data, c(rlogo, df$plot))
  expect_equal(chunk_info$width, c(.3, 300 / 72, 300 / 72))
  expect_equal(chunk_info$height, c(.23, 300 / 72, 300 / 72))

  docx_path <- save_as_docx(ft, path = tempfile(fileext = ".docx"))
  doc <- read_docx(docx_path)
  images_path <- doc$doc_obj$relationship()$get_images_path()
  expect_equal(
    gsub("([a-z0-9]+)(\\.png|\\.jpg)$", "\\2", basename(images_path)),
    c(".jpg", ".png", ".png")
  )

  html_path <- save_as_html(ft, path = tempfile(fileext = ".html"))
  doc <- read_html(html_path)
  all_imgs <- xml_find_all(doc, "//img")
  src_imgs <- xml_attr(all_imgs, "src")
  expect_length(src_imgs, 3)
  if (length(src_imgs) == 3) {
    expect_match(
      src_imgs[1],
      "data:image/jpeg",
      fixed = TRUE
    )
    expect_match(
      src_imgs[2],
      "data:image/png",
      fixed = TRUE
    )
    expect_match(
      src_imgs[3],
      "data:image/png",
      fixed = TRUE
    )
  }

  zz <- gen_grob(ft)
  expect_s3_class(
    zz$children$cell_1_1$children$contents$ftgrobs[[1]],
    "rastergrob"
  )
  expect_s3_class(
    zz$children$cell_2_1$children$contents$ftgrobs[[1]],
    "rastergrob"
  )
  expect_s3_class(
    zz$children$cell_3_1$children$contents$ftgrobs[[1]],
    "rastergrob"
  )

  ft <- flextable(df)
  ft <- colformat_image(ft, j = "plot", width = 300 / 72, height = 300 / 72)
  zz <- gen_grob(ft)
  expect_s3_class(zz$children$cell_1_1$children$contents$ftgrobs[[1]], "text")
  expect_s3_class(
    zz$children$cell_2_1$children$contents$ftgrobs[[1]],
    "rastergrob"
  )
  expect_s3_class(
    zz$children$cell_3_1$children$contents$ftgrobs[[1]],
    "rastergrob"
  )
})

test_that("minibar", {
  ft <- flextable(data.frame(n = 1:2))

  ft <- mk_par(
    ft,
    j = 1,
    value = as_paragraph(
      minibar(value = n, max = 10, width = .5, barcol = "red", bg = "yellow")
    ),
    part = "body"
  )
  minibar1 <- flextable::information_data_chunk(ft)$img_data[[2]]
  expect_s3_class(minibar1, "raster")
  expect_equal(nrow(minibar1), 1)
  expect_equal(ncol(minibar1), 36)
  expect_equal(minibar1[1:3], rep("#FF0000", 3))
  expect_equal(minibar1[4:36], rep("#FFFF00", 33))

  minibar2 <- flextable::information_data_chunk(ft)$img_data[[3]]
  expect_s3_class(minibar2, "raster")
  expect_equal(nrow(minibar2), 1)
  expect_equal(ncol(minibar2), 36)
  expect_equal(minibar2[1:7], rep("#FF0000", 7))
  expect_equal(minibar2[8:36], rep("#FFFF00", 29))
})

# linerange ----
test_that("linerange basic", {
  ft <- flextable(head(iris, 5))
  ft <- compose(
    ft,
    j = "Sepal.Length",
    value = as_paragraph(
      linerange(value = Sepal.Length)
    ),
    part = "body"
  )
  chunk_info <- information_data_chunk(ft)
  body_chunks <- chunk_info[chunk_info$.part == "body", ]
  sl_chunks <- body_chunks[body_chunks$.col_id == "Sepal.Length", ]
  expect_equal(nrow(sl_chunks), 5)
  expect_true(all(sl_chunks$width == 1))
  expect_true(all(sl_chunks$height == 0.2))
  expect_true(
    all(vapply(sl_chunks$img_data, inherits, logical(1), "raster"))
  )
})

test_that("linerange with min max and colors", {
  vals <- c(2, 5, 8)
  ft <- flextable(data.frame(x = vals))
  ft <- compose(
    ft,
    j = "x",
    value = as_paragraph(
      linerange(
        value = x,
        min = 0,
        max = 10,
        rangecol = "blue",
        stickcol = "green",
        bg = "white",
        width = 2,
        height = .3
      )
    ),
    part = "body"
  )
  chunk_info <- information_data_chunk(ft)
  body_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "x",
  ]
  expect_equal(nrow(body_chunks), 3)
  expect_true(all(body_chunks$width == 2))
  expect_true(all(body_chunks$height == 0.3))
})

test_that("linerange all NA values", {
  ft <- flextable(data.frame(x = c(NA_real_, NA_real_)))
  ft <- compose(
    ft,
    j = "x",
    value = as_paragraph(
      linerange(value = x)
    ),
    part = "body"
  )
  chunk_info <- information_data_chunk(ft)
  body_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "x",
  ]
  expect_equal(nrow(body_chunks), 2)
})

test_that("linerange raster_width < 2 errors", {
  expect_error(
    linerange(value = 1:3, raster_width = 1),
    "raster_width"
  )
})

test_that("linerange unit cm", {
  vals <- c(3, 6)
  ft <- flextable(data.frame(x = vals))
  ft <- compose(
    ft,
    j = "x",
    value = as_paragraph(
      linerange(value = x, width = 2, height = 0.5, unit = "cm")
    ),
    part = "body"
  )
  chunk_info <- information_data_chunk(ft)
  body_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "x",
  ]
  # cm -> in conversion: 2cm / 2.54 ~ 0.787
  expect_true(all(abs(body_chunks$width - 2 / 2.54) < 0.01))
})

# plot_chunk ----
test_that("plot_chunk box type", {
  library(data.table)
  z <- as.data.table(iris)
  z <- z[, list(z = list(.SD$Sepal.Length)), by = "Species"]
  ft <- flextable(z, col_keys = c("Species", "box"))
  ft <- mk_par(
    ft,
    j = "box",
    value = as_paragraph(
      plot_chunk(value = z, type = "box")
    )
  )
  chunk_info <- information_data_chunk(ft)
  box_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "box",
  ]
  expect_equal(nrow(box_chunks), 3)
  expect_true(all(file.exists(unlist(box_chunks$img_data))))
})

test_that("plot_chunk line type", {
  library(data.table)
  z <- as.data.table(iris)
  z <- z[, list(z = list(.SD$Sepal.Length)), by = "Species"]
  ft <- flextable(z, col_keys = c("Species", "pl"))
  ft <- mk_par(
    ft,
    j = "pl",
    value = as_paragraph(
      plot_chunk(value = z, type = "line")
    )
  )
  chunk_info <- information_data_chunk(ft)
  pl_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "pl",
  ]
  expect_equal(nrow(pl_chunks), 3)
  expect_true(all(file.exists(unlist(pl_chunks$img_data))))
})

test_that("plot_chunk points type", {
  library(data.table)
  z <- as.data.table(iris)
  z <- z[, list(z = list(.SD$Sepal.Length)), by = "Species"]
  ft <- flextable(z, col_keys = c("Species", "pl"))
  ft <- mk_par(
    ft,
    j = "pl",
    value = as_paragraph(
      plot_chunk(value = z, type = "points")
    )
  )
  chunk_info <- information_data_chunk(ft)
  pl_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "pl",
  ]
  expect_equal(nrow(pl_chunks), 3)
  expect_true(all(file.exists(unlist(pl_chunks$img_data))))
})

test_that("plot_chunk density type", {
  library(data.table)
  z <- as.data.table(iris)
  z <- z[, list(z = list(.SD$Sepal.Length)), by = "Species"]
  ft <- flextable(z, col_keys = c("Species", "pl"))
  ft <- mk_par(
    ft,
    j = "pl",
    value = as_paragraph(
      plot_chunk(value = z, type = "density")
    )
  )
  chunk_info <- information_data_chunk(ft)
  pl_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "pl",
  ]
  expect_equal(nrow(pl_chunks), 3)
  expect_true(all(file.exists(unlist(pl_chunks$img_data))))
})

test_that("plot_chunk free_scale", {
  library(data.table)
  z <- as.data.table(iris)
  z <- z[, list(z = list(.SD$Sepal.Length)), by = "Species"]
  ft <- flextable(z, col_keys = c("Species", "pl"))
  ft <- mk_par(
    ft,
    j = "pl",
    value = as_paragraph(
      plot_chunk(
        value = z,
        type = "box",
        free_scale = TRUE
      )
    )
  )
  chunk_info <- information_data_chunk(ft)
  pl_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "pl",
  ]
  expect_equal(nrow(pl_chunks), 3)
})

# gg_chunk ----
test_that("gg_chunk", {
  skip_if_not_installed("ggplot2")
  library(ggplot2)
  library(data.table)

  z <- as.data.table(iris)
  z <- z[,
    list(
      gg = list(
        ggplot(.SD, aes(Sepal.Length, Sepal.Width)) +
          geom_point() +
          theme_void()
      )
    ),
    by = "Species"
  ]

  ft <- flextable(z)
  ft <- mk_par(
    ft,
    j = "gg",
    value = as_paragraph(
      gg_chunk(value = gg, width = 1, height = 1)
    )
  )
  chunk_info <- information_data_chunk(ft)
  gg_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "gg",
  ]
  expect_equal(nrow(gg_chunks), 3)
  expect_true(all(file.exists(unlist(gg_chunks$img_data))))
  expect_true(all(gg_chunks$width == 1))
  expect_true(all(gg_chunks$height == 1))
})

test_that("gg_chunk unit cm", {
  skip_if_not_installed("ggplot2")
  library(ggplot2)

  plots <- list(
    ggplot(iris, aes(Sepal.Length)) +
      geom_histogram() +
      theme_void()
  )
  ft <- flextable(data.frame(g = "a"))
  ft <- mk_par(
    ft,
    j = "g",
    value = as_paragraph(
      gg_chunk(
        value = plots,
        width = 3,
        height = 2,
        unit = "cm"
      )
    )
  )
  chunk_info <- information_data_chunk(ft)
  gg_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "g",
  ]
  expect_true(abs(gg_chunks$width - 3 / 2.54) < 0.01)
  expect_true(abs(gg_chunks$height - 2 / 2.54) < 0.01)
})

# grid_chunk ----
test_that("grid_chunk", {
  library(grid)

  grobs <- list(
    circleGrob(gp = gpar(fill = "red", col = "transparent")),
    rectGrob(gp = gpar(fill = "blue", col = "transparent"))
  )
  ft <- flextable(data.frame(a = c("circle", "rect")))
  ft <- mk_par(
    ft,
    j = "a",
    value = as_paragraph(
      grid_chunk(
        value = grobs,
        width = .5,
        height = .5
      )
    )
  )
  chunk_info <- information_data_chunk(ft)
  grid_chunks <- chunk_info[
    chunk_info$.part == "body" &
      chunk_info$.col_id == "a",
  ]
  expect_equal(nrow(grid_chunks), 2)
  expect_true(all(file.exists(unlist(grid_chunks$img_data))))
  expect_true(all(grid_chunks$width == 0.5))
  expect_true(all(grid_chunks$height == 0.5))
})

test_that("grid_chunk in docx output", {
  library(grid)

  grobs <- list(
    circleGrob(gp = gpar(fill = "green", col = "transparent"))
  )
  ft <- flextable(data.frame(a = "dot"))
  ft <- mk_par(
    ft,
    j = "a",
    value = as_paragraph(
      grid_chunk(value = grobs, width = .3, height = .3)
    )
  )
  docx_file <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- body_add_flextable(doc, value = ft)
  expect_no_error(print(doc, target = docx_file))
  expect_true(file.exists(docx_file))
})

test_that("as_image alt text in DOCX", {
  ft <- flextable(head(iris, 2))
  ft <- compose(
    ft,
    i = 1,
    j = 1,
    value = as_paragraph(
      as_image(src = rlogo, width = 0.3, height = 0.23, alt = "R logo")
    ),
    part = "body"
  )
  doc <- read_docx()
  doc <- body_add_flextable(doc, ft)
  out <- print(doc, target = tempfile(fileext = ".docx"))
  unpack_dir <- tempfile()
  officer::unpack_folder(out, unpack_dir)
  doc_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "document.xml")
    ),
    collapse = ""
  )
  expect_true(grepl('descr="R logo"', doc_xml))
})

test_that("as_image alt text in HTML", {
  ft <- flextable(head(iris, 2))
  ft <- compose(
    ft,
    i = 1,
    j = 1,
    value = as_paragraph(
      as_image(src = rlogo, width = 0.3, height = 0.23, alt = "R logo")
    ),
    part = "body"
  )
  html <- as.character(htmltools::tagList(htmltools_value(ft)))
  expect_true(grepl('alt="R logo"', html))
})

test_that("as_image without alt has empty alt", {
  ft <- flextable(head(iris, 2))
  ft <- compose(
    ft,
    i = 1,
    j = 1,
    value = as_paragraph(
      as_image(src = rlogo, width = 0.3, height = 0.23)
    ),
    part = "body"
  )
  html <- as.character(htmltools::tagList(htmltools_value(ft)))
  expect_true(grepl('alt=""', html))
})

test_that("paragraph borders restored via style()", {
  border <- fp_border_default(width = 0.75, color = "black")
  ft <- qflextable(head(iris, 3)) |> border_remove()
  ft <- add_header_row(
    ft,
    values = c("Sepal", "Petal", "Species"),
    colwidths = c(2, 2, 1)
  )
  ft <- style(
    x = ft,
    part = "header",
    i = 1,
    j = c(1, 3),
    pr_p = officer::fp_par(
      border.bottom = border,
      text.align = "center"
    )
  )
  doc <- read_docx()
  doc <- body_add_flextable(doc, ft)
  out <- print(doc, target = tempfile(fileext = ".docx"))
  unpack_dir <- tempfile()
  officer::unpack_folder(out, unpack_dir)
  doc_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "document.xml")
    ),
    collapse = ""
  )
  expect_true(grepl("w:pBdr", doc_xml))
})

#' @export
#' @title Append chunks to flextable content
#' @description append chunks (for example chunk [as_chunk()])
#' in a flextable.
#' @inheritParams args_selectors_without_all
#' @param ... chunks to be appened, see [as_chunk()], [gg_chunk()] and other
#' chunk elements for paragraph.
#' @seealso [as_chunk()], [as_sup()], [as_sub()], [colorize()]
#' @family cell_content_composition
#' @examples
#' library(flextable)
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#'
#' ft_1 <- flextable(head(cars))
#'
#' ft_1 <- append_chunks(ft_1,
#'   # where to append
#'   i = c(1, 3, 5),
#'   j = 1,
#'   # what to append
#'   as_chunk(" "),
#'   as_image(src = img.file, width = .20, height = .15)
#' )
#' ft_1 <- set_table_properties(ft_1, layout = "autofit")
#' ft_1
append_chunks <- function(x, ..., i = NULL, j = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "append_chunks()"
    ))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  j <- x$col_keys[j]

  col_exprs <- enquos(...)

  tmp_data <- x[[part]]$dataset[i, , drop = FALSE]

  for (col_expr in col_exprs) {
    value <- eval_tidy(col_expr, data = tmp_data)
    x[[part]]$content <- append_chunkset_struct_element(
      x = x[[part]]$content,
      i = i,
      j = j,
      chunk_data = value
    )
  }

  x
}
#' @export
#' @title Prepend chunks to flextable content
#' @description prepend chunks (for example chunk [as_chunk()])
#' in a flextable.
#' @inheritParams args_selectors_without_all
#' @param ... chunks to be prepended, see [as_chunk()], [gg_chunk()] and other
#' chunk elements for paragraph.
#' @family cell_content_composition
#' @examples
#' x <- flextable(head(iris))
#' x <- prepend_chunks(
#'   x,
#'   i = 1, j = 1,
#'   colorize(as_b("Hello "), color = "red"),
#'   colorize(as_i("World"), color = "magenta")
#' )
#' x
prepend_chunks <- function(x, ..., i = NULL, j = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "prepend_chunks()"
    ))
  }
  part <- match.arg(part, c("body", "header", "footer"), several.ok = FALSE)

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  col_exprs <- enquos(...)

  tmp_data <- x[[part]]$dataset[i, , drop = FALSE]
  col_exprs <- rev(col_exprs)
  for (col_expr in col_exprs) {
    value <- eval_tidy(col_expr, data = tmp_data)
    x[[part]]$content <- append_chunkset_struct_element(
      x = x[[part]]$content,
      i = i,
      j = j,
      chunk_data = value,
      last = FALSE
    )
  }
  x
}

format_fun <- function(x, ...) {
  UseMethod("format_fun")
}

#' @export
format_fun.default <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    big.mark = flextable_global$defaults$big.mark,
    decimal.mark = flextable_global$defaults$decimal.mark,
    digits = flextable_global$defaults$digits,
    fmt_date = flextable_global$defaults$fmt_date,
    fmt_datetime = flextable_global$defaults$fmt_datetime,
    prefix = "",
    suffix = "",
    ...
  ) {
    if (inherits(x, "Date")) {
      out <- format_fun.Date(
        x,
        fmt_date = fmt_date,
        na_str = na_str,
        nan_str = nan_str,
        prefix = prefix,
        suffix = suffix
      )
    } else if (inherits(x, "POSIXt")) {
      out <- format_fun.POSIXt(
        x,
        fmt_datetime = fmt_datetime,
        na_str = na_str,
        nan_str = nan_str,
        prefix = prefix,
        suffix = suffix
      )
    } else if (is.double(x)) {
      out <- format_fun_defaultnum(
        x,
        na_str = na_str,
        nan_str = nan_str,
        big.mark = big.mark,
        decimal.mark = decimal.mark,
        digits = digits,
        prefix = prefix,
        suffix = suffix,
        ...
      )
    } else if (is.integer(x)) {
      out <- format_fun.integer(
        x,
        na_str = na_str,
        nan_str = nan_str,
        big.mark = big.mark,
        digits = digits,
        prefix = prefix,
        suffix = suffix
      )
    } else if (is.character(x)) {
      out <- format_fun.character(
        x,
        na_str = na_str,
        nan_str = nan_str,
        prefix = prefix,
        suffix = suffix
      )
    } else if (is.factor(x)) {
      out <- format_fun.factor(
        x,
        na_str = na_str,
        nan_str = nan_str,
        prefix = prefix,
        suffix = suffix
      )
    } else if (is.logical(x)) {
      out <- format_fun.logical(
        x,
        na_str = na_str,
        nan_str = nan_str,
        prefix = prefix,
        suffix = suffix
      )
    } else if (is.list(x)) {
      what <- sapply(x, function(x) head(class(x), n = 1))
      out <- sprintf("[[%s]]", what)
    } else {
      out <- format(
        x,
        trim = TRUE,
        big.mark = big.mark,
        decimal.mark = decimal.mark,
        justify = "none",
        scientific = FALSE,
        digits = digits,
        ...
      )
      out[is.na(x)] <- na_str
      out[is.nan(x)] <- nan_str
    }
    out
  }

#' @export
format_fun.character <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    ...
  ) {
    out <- paste0(prefix, as.character(x), suffix)
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }

#' @export
format_fun.factor <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    ...
  ) {
    out <- paste0(prefix, as.character(x), suffix)
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }

#' @export
format_fun.logical <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    true = "true",
    false = "false",
    ...
  ) {
    out <- ifelse(x, true, false)
    out <- paste0(prefix, out, suffix)
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }


format_fun_defaultnum <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    big.mark = flextable_global$defaults$big.mark,
    decimal.mark = flextable_global$defaults$decimal.mark,
    digits = flextable_global$defaults$digits,
    ...
  ) {
    out <- paste0(
      prefix,
      format(
        x,
        trim = TRUE,
        scientific = FALSE,
        big.mark = big.mark,
        decimal.mark = decimal.mark,
        ...
      ),
      suffix
    )
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }

#' @export
format_fun.double <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    big.mark = flextable_global$defaults$big.mark,
    decimal.mark = flextable_global$defaults$decimal.mark,
    digits = flextable_global$defaults$digits,
    ...
  ) {
    out <- paste0(
      prefix,
      formatC(
        x,
        format = "f",
        big.mark = big.mark,
        digits = digits,
        decimal.mark = decimal.mark
      ),
      suffix
    )
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }
#' @export
format_fun.pct <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    big.mark = flextable_global$defaults$big.mark,
    decimal.mark = flextable_global$defaults$decimal.mark,
    pct_digits = flextable_global$defaults$pct_digits,
    ...
  ) {
    if (is.null(x)) {
      return("")
    }
    wisna <- is.na(x)
    wisnan <- is.nan(x)
    out <- paste0(
      formatC(
        x * 100,
        format = "f",
        big.mark = big.mark,
        digits = pct_digits,
        decimal.mark = decimal.mark
      ),
      "%"
    )
    out[wisna] <- na_str
    out[wisnan] <- nan_str
    out
  }

#' @export
format_fun.integer <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    big.mark = flextable_global$defaults$big.mark,
    ...
  ) {
    out <- paste0(
      prefix,
      formatC(x, format = "f", big.mark = big.mark, digits = 0),
      suffix
    )
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }

#' @export
format_fun.Date <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    fmt_date = flextable_global$defaults$fmt_date,
    ...
  ) {
    out <- paste0(prefix, format(x, fmt_date), suffix)
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }

#' @export
format_fun.POSIXt <-
  function(
    x,
    na_str = flextable_global$defaults$na_str,
    nan_str = flextable_global$defaults$nan_str,
    prefix = "",
    suffix = "",
    fmt_datetime = flextable_global$defaults$fmt_datetime,
    ...
  ) {
    out <- paste0(prefix, format(x, fmt_datetime), suffix)
    out[is.na(x)] <- na_str
    out[is.nan(x)] <- nan_str
    out
  }

# flextable grob ----------------------------------------------------------

#' @export
#' @title Render a flextable as a graphic object
#' @description
#' `gen_grob()` converts a flextable into a Grid Graphics
#' object (`grob`) that can be drawn on any R graphic device.
#' This is the function behind [save_as_image()] and the
#' patchwork integration ([wrap_flextable()]).
#'
#' Typical uses:
#' * embed a flextable in a `ggplot2` plot (via
#'   [wrap_flextable()] or cowplot)
#' * export a flextable as a PNG or SVG image (via
#'   [save_as_image()])
#'
#' Text wrapping and scaling are supported. The `fit`
#' argument controls how the table adapts to the available
#' space (fixed size, auto-fit width, or fill the device).
#'
#' Not recommended for very large tables because the
#' grid calculations can be slow.
#'
#' Limitations: equations ([as_equation()]) and hyperlinks
#' ([officer::hyperlink_ftext()]) are not rendered.
#'
#' Use a 'ragg', 'svglite' or 'ggiraph' device for correct
#' rendering.
#' @inheritSection save_as_image caption
#' @section size:
#'
#' The size of the flextable can be known by using the method
#' \link[=dim.flextableGrob]{dim} on the grob.
#'
#' @inheritParams args_x_only
#'
#' @param fit Determines the fitting/scaling of the grob on its parent viewport.
#' One of `auto`, `width`, `fixed`, `TRUE`, `FALSE`:
#' \itemize{
#'   \item `auto` or `TRUE` (default): The grob is resized to fit in the parent viewport.
#'   The table row heights and column widths are resized proportionally.
#'   \item `width`: The grob is resized horizontally to fit the width
#'   of the parent viewport. The column widths are resized proportionally.
#'   The row heights are unaffected and the table height may be smaller or larger
#'   than the height of the parent viewport.
#'   \item `fixed` or `FALSE`: The grob will have fixed dimensions,
#'   as determined by the column widths and the row heights.
#' }
#'
#' @param scaling Determines the scaling of the table contents.
#' One of `min`, `full`, `fixed`, `TRUE`, `FALSE`:
#' \itemize{
#'   \item `min` or `TRUE` (default):
#'   When the parent viewport is smaller than the necessary,
#'   the various content sizes (text font size, line width and image dimensions)
#'   will decrease accordingly so that the content can still fit.
#'   When the parent viewport is larger than the necessary,
#'   the content sizes will remain the same, they will not increase.
#'   \item `full`: Same as `min`, except that the content sizes are scaled fully,
#'   they will increase or decrease, according to the size of the drawing surface.
#'   \item `fixed` or `FALSE`: The content sizes will not be scaled.
#' }
#'
#' @param wrapping Determines the soft wrapping (line breaking) method
#' for the table cell contents. One of `TRUE`, `FALSE`:
#' \itemize{
#'   \item `TRUE`: Text content may wrap into separate lines at normal word break points
#'   (such as a space or tab character between two words) or at newline characters
#'   anywhere in the text content. If a word does not fit in the available cell width,
#'   then the text content may wrap at any character.
#'   Non-text content (such as images) is also wrapped into new lines,
#'   according to the available cell width.
#'   \item `FALSE`: Text content may wrap only with a newline character.
#'   Non-text content is not wrapped.
#' }
#'
#' Superscript and subscript chunks do not wrap.
#' Newline and tab characters are removed from these chunk types.
#'
#' @param autowidths If `TRUE` (default) the column widths are adjusted
#' in order to fit the contents of the cells (taking into account the `wrapping` setting).
#'
#' @param just Justification of viewport layout,
#' same as `just` argument in [grid::grid.layout()].
#' When set to `NULL` (default), it is determined according to the `fit` argument.
#'
#' @param ... Reserved for extra arguments
#'
#' @return a grob (gTree) object made with package `grid`
#' @examples
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' library(ragg)
#' library(gdtools)
#' register_liberationsans()
#'
#' set_flextable_defaults(font.family = "Liberation Sans")
#'
#' ft <- flextable(head(mtcars))
#' gr <- gen_grob(ft)
#' \dontshow{
#' cap <- ragg::agg_capture(width = 7, height = 4, units = "in", res = 150)
#' grDevices::dev.control("enable")
#' }
#' plot(gr)
#' \dontshow{
#' raster <- cap()
#' dev.off()
#' plot(as.raster(raster))
#' }
#' @family flextable_output_export
#' @importFrom grid gTree
gen_grob <- function(
  x,
  ...,
  fit = c("auto", "width", "fixed"),
  scaling = c("min", "full", "fixed"),
  wrapping = TRUE,
  autowidths = TRUE,
  just = NULL
) {
  x <- flextable_global$defaults$post_process_all(x)
  x <- fix_border_issues(x)

  dots <- list(...)
  debug <- isTRUE(dots$debug)
  if (identical(fit, TRUE)) {
    fit <- "auto"
  }
  if (identical(fit, FALSE)) {
    fit <- "fixed"
  }
  if (identical(scaling, TRUE)) {
    scaling <- "min"
  }
  if (identical(scaling, FALSE)) {
    scaling <- "fixed"
  }
  fit <- match.arg(fit)
  scaling <- match.arg(scaling)
  wrapping <- isTRUE(wrapping)
  autowidths <- isTRUE(autowidths)
  if (fit %in% "fixed") {
    scaling <- "fixed"
  }

  # generate grid data
  dat <- get_grid_data(x, autowidths, wrapping)
  dat <- grid_data_add_cell_info(dat, x)
  dat <- grid_data_add_par_info(dat, x)
  dat <- grid_data_add_chunk_info(dat, x, autowidths, wrapping)
  setorderv(dat, cols = c("row_index", "col_index"))

  # create grob
  gr <- make_flextable_grob(dat, fit, scaling, wrapping, debug = debug)

  # calculate real row heights/column widths
  dims <- dim(x)
  widths <- calc_grob_widths(gr, dat, dims)
  heights <- calc_grob_heights(gr, dat, dims)

  # set parameter list
  gr$ftpar <- list(
    fit = fit,
    scaling = scaling,
    wrapping = wrapping,
    just = just,
    dims = dims,
    widths = widths,
    heights = heights,
    debug = debug
  )

  # set data
  gr$ftdat <- dat[, c("row_index", "col_index", "rowspan", "colspan")]

  # done
  invisible(gr)
}

#' @importFrom grid viewport grid.layout gList vpStack
make_flextable_grob <- function(dat, fit, scaling, wrapping, debug = FALSE) {
  gTree(
    cl = "flextableGrob",
    children = do.call(
      gList,
      mapply(
        function(
          row_index,
          col_index,
          cell_vp,
          cell_width,
          cell_height,
          cell_params,
          background,
          borders,
          contents_vp,
          chunk_data,
          chunk_part_data
        ) {
          # single cell grob
          gTree(
            children = gList(
              # background
              make_background_grob(background),
              # borders
              make_borders_grob(borders),
              # debug contents rect
              if (debug) {
                rectGrob(
                  gp = gpar(col = "red", fill = "yellow", alpha = 0.15),
                  vp = viewport(
                    width = unit(1, "npc") - unit(cell_params$paddingx, "in"),
                    height = unit(1, "npc") - unit(cell_params$paddingy, "in")
                  )
                )
              },
              # contents
              make_contents_grob(
                row_index = row_index,
                col_index = col_index,
                chunk_data = chunk_data,
                chunk_part_data = chunk_part_data,
                vp = contents_vp,
                params = cell_params,
                width = cell_width,
                height = cell_height,
                fit = fit,
                scaling = scaling,
                wrapping = wrapping,
                debug = debug
              ),
              # debug text
              if (debug) {
                textGrob(
                  label = paste(row_index, col_index, sep = "_"),
                  x = cell_params$just[[1]],
                  y = cell_params$just[[2]],
                  gp = gpar(col = "red", fontsize = 8),
                  just = cell_params$justr,
                  rot = cell_params$angle
                )
              }
            ),
            vp = do.call(viewport, cell_vp),
            name = paste("cell", row_index, col_index, sep = "_")
          )
        },
        row_index = dat$row_index,
        col_index = dat$col_index,
        cell_vp = dat$cell_vp,
        cell_width = dat$cell_width,
        cell_height = dat$cell_height,
        cell_params = dat$cell_params,
        background = dat$background,
        borders = dat$borders,
        contents_vp = dat$contents_vp,
        chunk_data = dat$chunk_data,
        chunk_part_data = dat$chunk_part_data,
        SIMPLIFY = FALSE
      )
    )
  )
}

#' Calculate column widths for a flextableGrob
#' @param x The flextable grob
#' @param dat The grid_data
#' @param dims The flextable dimensions
#' @return A numeric vector with the widths
#' @noRd
calc_grob_widths <- function(x, dat, dims) {
  # get the minimum width of each column
  col_dat <- dat[,
    list(width = min(.SD$cell_width, na.rm = TRUE)),
    by = "col_index"
  ]
  # dims contain all columns, but some of them might have been merged
  # completely into others and they should get a 0 size.
  col_dat <- merge(
    data.table(col_index = seq_along(dims$widths)),
    col_dat,
    by = "col_index",
    all.x = TRUE
  )
  col_dat[is.na(col_dat$width), "width" := 0]
  col_dat$width
}

#' Calculate the row heights for a flextableGrob
#' @param x The flextable grob
#' @param dat The grid_data
#' @param dims The flextable dimensions
#' @return A numeric vector with the heights
#' @noRd
#' @importFrom grid is.grob
calc_grob_heights <- function(x, dat, dims) {
  # get height info from the grob contents
  content_heights <- mapply(
    function(row_index, col_index, colspan) {
      cell_name <- paste("cell", row_index, col_index, sep = "_")
      gr <- x$children[[cell_name]]$children[["contents"]]
      if (is.grob(gr)) {
        list(
          row_index = gr$ftpar$row_index,
          col_index = gr$ftpar$col_index,
          colspan = colspan,
          height = gr$ftpar$height
        )
      }
    },
    dat$row_index,
    dat$col_index,
    dat$colspan,
    SIMPLIFY = FALSE
  )
  height_dat <- rbindlist(content_heights)
  # keep only the rows with minimum colspan
  height_dat <- merge(
    height_dat,
    dat[, list(colspan = min(.SD$colspan)), by = c("row_index")],
    by = c("row_index", "colspan")
  )
  # get the maximum height of each row
  height_dat <- height_dat[,
    list(height = max(.SD$height, na.rm = TRUE)),
    by = "row_index"
  ]
  # dims contain all rows, but some of them might have been merged
  # completely into others and they should get a 0 size.
  row_dat <- merge(
    data.table(row_index = seq_along(dims$heights)),
    height_dat,
    by = "row_index",
    all.x = TRUE
  )
  row_dat[is.na(row_dat$height), "height" := 0]
  row_dat$height
}

#' @importFrom grid makeContext unit convertWidth convertHeight
#' @export
makeContext.flextableGrob <- function(x) {
  params <- x$ftpar
  dat <- x$ftdat
  cex <- 1
  lex <- 1

  # dimensions of the table
  widths <- params$widths
  table_width <- sum(widths)
  heights <- params$heights
  table_height <- sum(heights)

  if (params$fit %in% c("auto", "width")) {
    # dimensions of drawing area
    norm_width <- convertWidth(unit(1, "npc"), unitTo = "in", valueOnly = TRUE)
    norm_height <- convertHeight(
      unit(1, "npc"),
      unitTo = "in",
      valueOnly = TRUE
    )
    # calculate table scale factor
    norm_surface <- norm_width * norm_height
    table_surface <- table_width * table_height
    surface_factor <- sqrt(norm_surface / table_surface)
    width_factor <- norm_width / table_width
    height_factor <- norm_height / table_height

    # set cex/lex as the smallest of the three
    scale_factor <- min(surface_factor, width_factor, height_factor)
    if (
      params$scaling %in%
        "full" ||
        (params$scaling %in% "min" && scale_factor < 1)
    ) {
      lex <- cex <- scale_factor
    }

    # rearrange the contents with new parameters
    cell_names <- paste("cell", dat$row_index, dat$col_index, sep = "_")
    for (cell_name in cell_names) {
      gr <- x$children[[cell_name]]$children[["contents"]]
      if (is.grob(gr)) {
        x$children[[cell_name]]$children[["contents"]] <- arrange_contents_grob(
          gr,
          width_factor = width_factor,
          height_factor = height_factor,
          cex = cex
        )
      }
    }

    # recalculate the heights
    heights <- calc_grob_heights(x, dat, params$dims)
    table_height <- sum(heights)

    # flex_body: redistribute body row heights to fill the panel
    if (isTRUE(params$flex_body)) {
      n_h <- params$n_header_rows
      n_b <- params$n_body_rows
      n_f <- params$n_footer_rows
      header_h <- if (n_h > 0) sum(heights[seq_len(n_h)]) else 0
      footer_h <- if (n_f > 0) {
        sum(heights[seq.int(n_h + n_b + 1L, length(heights))])
      } else {
        0
      }
      body_h <- norm_height - header_h - footer_h
      if (body_h > 0 && n_b > 0) {
        heights[seq.int(n_h + 1L, n_h + n_b)] <- body_h / n_b
        table_height <- sum(heights)
      }
    }

    # flex_cols: redistribute data column widths to fill the panel
    if (isTRUE(params$flex_cols)) {
      n_rh <- params$n_row_header_cols
      n_dc <- params$n_data_cols
      rh_w <- if (n_rh > 0) sum(widths[seq_len(n_rh)]) else 0
      data_w <- norm_width - rh_w
      if (data_w > 0 && n_dc > 0) {
        widths[seq.int(n_rh + 1L, n_rh + n_dc)] <- data_w / n_dc
        table_width <- sum(widths)
      }
    }
  }

  # create viewport
  layout_just <- params$just
  if (params$fit %in% "auto") {
    # convert to relative widths/heights
    layout_widths <- unit(widths / table_width, "npc")
    layout_heights <- unit(heights / table_height, "npc")
    if (is.null(layout_just)) layout_just <- c(0.5, 0.5)
  } else if (params$fit %in% "width") {
    layout_widths <- unit(widths / table_width, "npc")
    layout_heights <- unit(heights, "in")
    if (is.null(layout_just)) layout_just <- c(0.5, 1)
  } else {
    layout_widths <- unit(widths, "in")
    layout_heights <- unit(heights, "in")
    if (is.null(layout_just)) layout_just <- c(0, 1)
  }
  layout_vp <- viewport(
    gp = gpar(
      cex = cex,
      lex = lex
    ),
    layout = grid.layout(
      nrow = length(heights),
      ncol = length(widths),
      widths = layout_widths,
      heights = layout_heights,
      just = layout_just
    )
  )
  if (is.null(x$vp)) {
    x$vp <- layout_vp
  } else {
    x$vp <- vpStack(x$vp, layout_vp)
  }

  # done
  x
}

#' plot a flextable grob
#'
#' @param x a flextableGrob object
#' @param ... additional arguments passed to other functions
#' @importFrom grid grid.newpage grid.draw pushViewport popViewport
#' @export
plot.flextableGrob <- function(x, ...) {
  grid.newpage()
  # leave a 5pt margin around the plot
  pushViewport(viewport(
    width = unit(1, "npc") - unit(10, "pt"),
    height = unit(1, "npc") - unit(10, "pt")
  ))
  grid.draw(x)
  popViewport()
  invisible(x)
}

#' @title Get optimal width and height of a flextable grob
#' @description returns the optimal width and height for the grob,
#' according to the grob generation parameters.
#' @param x a flextableGrob object
#' @return a named list with two elements, `width` and `height`.
#' Values are expressed in inches.
#' @examples
#' ftab <- flextable(head(iris))
#' gr <- gen_grob(ftab)
#' dim(gr)
#' @export
dim.flextableGrob <- function(x) {
  list(
    width = sum(x$ftpar$widths, na.rm = TRUE),
    height = sum(x$ftpar$heights, na.rm = TRUE)
  )
}

# contents grob -----------------------------------------------------------

#' @importFrom data.table is.data.table .GRP
make_contents_grob <- function(
  row_index,
  col_index,
  chunk_data,
  chunk_part_data,
  vp,
  params,
  ...
) {
  if (
    !is.data.table(chunk_data) ||
      !nrow(chunk_data) ||
      !is.data.table(chunk_part_data) ||
      !nrow(chunk_part_data)
  ) {
    return(NULL)
  }

  dots <- list(...)
  params <- c(
    list(
      row_index = row_index,
      col_index = col_index
    ),
    params,
    dots
  )

  # create contents grob
  gr <- gTree(
    vp = do.call(viewport, vp),
    name = "contents"
  )

  # create chunk grobs
  gr$ftgrobs <- chunk_data[,
    list(grobs = list(make_chunk_grob(.GRP, .SD, params))),
    by = "chunk_index"
  ]$grobs

  # set data
  gr$ftdat <- chunk_part_data

  # set parameters
  gr$ftpar <- params

  # arrange grob to get the real height
  # if the fit is fixed, make the changes permanent
  gr <- arrange_contents_grob(gr, do_update_grob = params$fit %in% "fixed")

  # done
  gr
}

#' @importFrom grid setChildren
arrange_contents_grob <- function(
  x,
  width_factor = 1,
  height_factor = 1,
  cex = 1,
  do_update_grob = TRUE
) {
  params <- x$ftpar
  dat <- x$ftdat

  # first-line/hanging indents: honored for horizontal, left-justified
  # content only. The contents viewport is already placed at the
  # first-line position (see grid_data_add_par_info), so the first line
  # uses the full width and continuation lines are indented by `hanging`
  # (`first_line` is the reverse: only the first line is indented).
  hanging_in <- params$hanging
  if (is.null(hanging_in) || is.na(hanging_in)) {
    hanging_in <- 0
  }
  first_line_in <- params$first_line
  if (is.null(first_line_in) || is.na(first_line_in)) {
    first_line_in <- 0
  }
  if (params$angle != 0 || !isTRUE(params$just[[1]] == 0)) {
    hanging_in <- 0
    first_line_in <- 0
  }

  if (nrow(dat) > 1) {
    # calculate real max width for the contents
    if (params$angle == 0) {
      max_width <- (params$width - params$paddingx) * width_factor
    } else {
      max_width <- (max(
        params$height,
        params$min_height,
        params$max_height,
        na.rm = TRUE
      ) -
        params$paddingy) *
        height_factor
    }
    indented_width <- max_width - max(hanging_in, first_line_in)

    child_count <- .chunk_index <- width <- NULL
    if (params$wrapping) {
      filtered <- c(
        unlist(dat[
          child_count > 0 & width * cex <= indented_width,
          "children_index"
        ]),
        unlist(dat[
          child_count > 0 & width * cex > indented_width,
          ".chunk_index"
        ])
      )
      dat <- dat[!.chunk_index %in% filtered, , drop = FALSE]
    }
    n <- nrow(dat)
    if (n > 1) {
      # loop the chunk parts and arrange them in rows/columns
      # according to their width
      row_index <- integer(n)
      col_index <- integer(n)
      cur_width <- 0
      cur_row <- 1L
      cur_col <- 1L
      cur_chunk <- 0L
      for (i in seq_len(n)) {
        item <- dat[i, ]
        cur_width <- cur_width + (item$width * cex)
        # available width depends on the line: `first_line` narrows the
        # first line, `hanging` narrows the continuation lines
        row_width <- max_width -
          (if (cur_row == 1) first_line_in else hanging_in)
        if (item$is_newline) {
          # if it's the first item in first row, add it, else ignore it
          if (cur_row == 1 && cur_col == 1) {
            row_index[[i]] <- cur_row
            col_index[[i]] <- cur_col
          }
          cur_row <- cur_row + 1L
          cur_col <- 1L
          cur_width <- 0
          cur_chunk <- 0L
        } else if (params$wrapping && cur_width > row_width) {
          if (cur_col == 1) {
            # if it's the first item in row, add it and go to next row
            row_index[[i]] <- cur_row
            col_index[[i]] <- cur_col
            cur_row <- cur_row + 1L
            cur_col <- 1L
            cur_width <- 0
            cur_chunk <- 0L
          } else {
            # else go to next row and then add it
            cur_row <- cur_row + 1L
            cur_col <- 1L
            if (item$is_whitespace) {
              # ignore whitespace in line break
              cur_width <- 0
              cur_chunk <- 0L
            } else {
              row_index[[i]] <- cur_row
              col_index[[i]] <- cur_col
              cur_col <- cur_col + 1L
              cur_width <- (item$width * cex)
              cur_chunk <- item$chunk_index
            }
          }
        } else {
          if (cur_row > 1L && cur_col == 1L && item$is_whitespace) {
            # ignore whitespace
            cur_width <- 0
            cur_chunk <- 0L
          } else {
            # set the row
            row_index[[i]] <- cur_row
            # if the item is from a different chunk than the previous
            if (item$chunk_index != cur_chunk) {
              # set the column
              col_index[[i]] <- cur_col
              # and go to next column
              cur_col <- cur_col + 1L
            } else {
              # set it to the previous column
              col_index[[i]] <- cur_col - 1
            }
            cur_chunk <- item$chunk_index
          }
        }
      }
      dat$row_index <- row_index
      dat$col_index <- col_index
      dat <- dat[row_index > 0, ]
    }
  }

  make_row_grob <- function(row_index, grobs, layout) {
    # when there are more than one grobs in a row
    # we need to align them to their baseline
    # because they might have different font sizes
    # use flextableRowPartGrob class for that
    if (length(grobs) > 1) {
      grob_names <- sapply(grobs, `[[`, "name")
      grobs <- lapply(grobs, function(gr) {
        class(gr) <- c("flextableRowPartGrob", class(gr))
        gr$ftsiblings <- grob_names
        names(gr$ftsiblings) <- grob_names
        gr
      })
    }

    if (params$debug) {
      grobs <- c(
        grobs,
        list(
          rectGrob(
            gp = gpar(lwd = 0.5, col = "red", fill = NA)
          )
        )
      )
    }

    gTree(
      children = do.call(gList, grobs),
      vp = viewport(
        layout.pos.row = row_index,
        layout.pos.col = 1,
        layout = layout
      ),
      name = paste("chunk_row", row_index, sep = "_")
    )
  }

  make_col_grob <- function(
    col_index,
    chunk_index,
    txt_data,
    from_.chunk_index,
    to_.chunk_index,
    width,
    height
  ) {
    gr <- x$ftgrobs[[chunk_index]]
    params <- gr$ftpar
    if (inherits(gr, "text") && params$txt_data != txt_data) {
      gr$label <- calc_grid_label(
        x = txt_data,
        valign = params$valign,
        is_underlined = params$is_underlined,
        is_bold = params$is_bold,
        is_italic = params$is_italic
      )
      gr$name <- paste(
        gr$name,
        "seq",
        from_.chunk_index,
        "to",
        to_.chunk_index,
        sep = "_"
      )
    }
    gr$vp <- viewport(
      layout.pos.row = 1,
      layout.pos.col = col_index
    )
    if (inherits(gr, "flextableRasterChunk")) {
      gr$vp <- vpStack(
        gr$vp,
        viewport(
          width = unit(width * cex, "in"),
          height = unit(height * cex, "in")
        )
      )
    }
    gr
  }

  if (nrow(dat) > 1) {
    # group data by row, column and chunk
    col_dat <- dat[,
      list(
        from = min(.SD$.chunk_index),
        to = max(.SD$.chunk_index),
        txt_data = paste0(.SD$txt_data, collapse = ""),
        width = sum(.SD$width),
        height = max(.SD$ascent) + max(.SD$descent)
      ),
      by = c("row_index", "col_index", "chunk_index")
    ]

    # per-line indent, materialized as an extra leading layout column
    # (first line for `first_line`, continuation lines for `hanging`)
    col_dat[,
      "row_indent" := fifelse(
        .SD$row_index == 1L,
        first_line_in,
        hanging_in
      )
    ]
    col_dat[, "col_index" := .SD$col_index + (.SD$row_indent > 0)]

    if (do_update_grob) {
      # create grobs for columns
      col_dat[,
        "grob" := mapply(
          make_col_grob,
          .SD$col_index,
          .SD$chunk_index,
          .SD$txt_data,
          .SD$from,
          .SD$to,
          .SD$width,
          .SD$height,
          SIMPLIFY = FALSE
        )
      ]
    }

    # group data by row
    row_dat <- col_dat[,
      {
        row_indent <- first(.SD$row_indent)
        list(
          grobs = list(.SD$grob),
          layout = list(grid.layout(
            nrow = 1,
            ncol = nrow(.SD) + (row_indent > 0),
            widths = unit(
              c(if (row_indent > 0) row_indent, .SD$width * cex),
              "in"
            ),
            # heights = unit(max(.SD$height * cex), "in"),
            just = params$justr
          )),
          width = row_indent + sum(.SD$width * cex),
          height = max(.SD$height * cex)
        )
      },
      by = "row_index"
    ]

    if (do_update_grob) {
      # create grobs for rows
      row_dat[,
        "grob" := mapply(
          make_row_grob,
          .SD$row_index,
          .SD$grobs,
          .SD$layout,
          SIMPLIFY = FALSE
        )
      ]
      children <- do.call(gList, row_dat$grob)
    }
    widths <- row_dat$width
    heights <- row_dat$height
  } else {
    if (do_update_grob) {
      children <- gList(make_col_grob(
        1L,
        dat$chunk_index,
        dat$txt_data,
        dat$.chunk_index,
        dat$.chunk_index,
        dat$width,
        dat$height
      ))
    }
    widths <- dat$width * cex
    heights <- dat$height * cex
  }

  if (do_update_grob) {
    # set viewport layout
    x$vp$layout <- grid.layout(
      nrow = length(children),
      ncol = 1,
      heights = unit(heights, "in"),
      widths = unit(max(widths), "in"),
      just = params$justr
    )
  }

  # calculate real height of the contents
  if (params$angle == 0) {
    height <- sum(heights) + params$paddingy
  } else {
    height <- max(widths) + params$paddingy
  }
  height <- min(
    max(height, params$min_height, na.rm = TRUE),
    params$max_height,
    na.rm = TRUE
  )

  # set the height info
  x$ftpar$height <- height

  if (do_update_grob) {
    # set the children and return the updated grob
    setChildren(x, children)
  } else {
    # just return the grob
    x
  }
}

# misc grobs --------------------------------------------------------------

#' @importFrom grid textGrob rasterGrob
make_chunk_grob <- function(chunk_index, dat, params) {
  if (dat$is_text) {
    gr <- textGrob(
      label = calc_grid_label(
        x = paste(dat$txt_data[[1]], collapse = ""),
        valign = dat$valign,
        is_underlined = dat$is_underlined,
        is_bold = dat$is_bold,
        is_italic = dat$is_italic
      ),
      gp = dat$gp[[1]]
    )
    # store info for label
    gr$ftpar <- list(
      txt_data = dat$txt_data[[1]],
      valign = dat$valign,
      is_underlined = dat$is_underlined,
      is_bold = dat$is_bold,
      is_italic = dat$is_italic
    )
  } else if (dat$is_raster) {
    if (inherits(dat$img_data[[1]], "flextableRasterChunk")) {
      gr <- rasterGrob(
        image = dat$img_data[[1]],
        width = unit(1, "npc"),
        height = unit(1, "npc"),
        gp = dat$gp[[1]]
      )
      class(gr) <- c("flextableRasterChunk", class(gr))
    } else {
      gr <- rasterGrob(
        image = dat$img_data[[1]],
        gp = dat$gp[[1]]
      )
    }
  } else {
    gr <- rasterGrob(
      image = blank_raster_image,
      width = unit(1, "npc"),
      height = unit(1, "npc"),
      gp = dat$gp[[1]]
    )
  }
  if (!dat$shading_color %in% "transparent" || params$debug) {
    class(gr) <- c("flextableHighlightedGrob", class(gr))
    gr$highlight_gp <- gpar(
      lwd = 0.5,
      fill = dat$shading_color,
      col = ifelse(params$debug, "black", dat$shading_color)
    )
  }
  gr$name <- paste("chunk", chunk_index, sep = "_")
  gr
}

#' @importFrom grid rectGrob
make_background_grob <- function(args) {
  if (length(args) > 0) {
    args$name <- "background"
    do.call(rectGrob, args)
  }
}

#' @importFrom grid segmentsGrob
make_segment_grob <- function(args) {
  if (length(args) > 0) {
    do.call(segmentsGrob, args)
  }
}

make_borders_grob <- function(args) {
  if (length(args) > 0) {
    segments <- lapply(args, make_segment_grob)
    if (length(segments) > 0) {
      gTree(
        children = do.call(gList, segments),
        name = "borders"
      )
    }
  }
}

#' @importFrom grid drawDetails grid.rect grobWidth grobHeight grobDescent
#' @export
drawDetails.flextableHighlightedGrob <- function(x, recording) {
  if (!is.null(x$highlight_gp)) {
    # draw a rect around the grob, taking into account the descent
    # as grobHeight does not include it
    descent <- grobDescent(x)
    if (inherits(x, "text") && is.expression(x$label)) {
      yadd <- descent
      yoffset <- unit(0, "npc")
    } else {
      yadd <- descent * 2
      yoffset <- descent / 2
    }
    grid.rect(
      x = x$x,
      y = x$y - yoffset,
      just = x$just,
      width = grobWidth(x),
      height = grobHeight(x) + yadd,
      gp = x$highlight_gp
    )
  }
  NextMethod()
}

#' @export
drawDetails.flextableRowPartGrob <- function(x, recording) {
  # When different chunks are parts of a row, they must be aligned in their baseline.
  # At this point they are aligned in their middle (as vjust is 0.5).
  # So we must offset them according to the tallest grob.
  # We use grobHeight from grid because it is most accurate and it excludes ascent and descent.
  # x$ftsiblings contains the names of all the grobs in the row
  if (length(x$ftsiblings) > 1) {
    # calculate half height of the grobs
    halfheights <- sapply(x$ftsiblings, function(z) {
      convertHeight(grobHeight(z) / 2, unitTo = "in", valueOnly = TRUE)
    })
    # offset y, so that it is aligned with the tallest grob
    x$y <- x$y - unit(max(halfheights) - halfheights[[x$name]], "in")
    # offset further for subscripts
    if (inherits(x, "text") && x$ftpar$valign %in% "subscript") {
      x$y <- x$y - grobDescent(x)
    }
  }
  NextMethod()
}

# borders format ----

#' @export
#' @title Set cell borders
#' @description change borders of selected rows and columns of a flextable.
#' This function is **not to be used by end user** (it requires careful
#' settings to avoid overlapping borders) but only for programming purposes.
#'
#' If you need to add borders, use instead other functions:
#'
#' These set borders for the whole table : [border_outer()],
#' [border_inner_h()] and [border_inner_v()].
#'
#' To add horizontal or vertical lines in the table at specific location,
#' use:
#'
#' - [hline()]: set bottom borders (inner horizontal)
#' - [vline()]: set right borders (inner vertical)
#'
#' To add following horizontal or vertical lines at
#' beginning or end of the table, use:
#'
#' - [hline_top()]: set the top border (outer horizontal)
#' - [hline_bottom()]: set the bottom border (outer horizontal)
#' - [vline_left()]: set the left border (outer vertical)
#' - [vline_right()]: set the right border (outer vertical)
#'
#' If you want to highlight specific cells with some borders, use
#' [surround()].
#' @note
#' pdf and pptx outputs do not support `border()` usage.
#' @inheritParams args_selectors_with_all
#' @param border border (shortcut for top, bottom, left and right)
#' @param border.top border top
#' @param border.bottom border bottom
#' @param border.left border left
#' @param border.right border right
#' @examples
#' library(officer)
#' ftab <- flextable(head(mtcars))
#' ftab <- border(ftab, border.top = fp_border(color = "orange"))
#' ftab
#' @keywords internal
border <- function(
  x,
  i = NULL,
  j = NULL,
  border = NULL,
  border.top = NULL,
  border.bottom = NULL,
  border.left = NULL,
  border.right = NULL,
  part = "body"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "border()"))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (!is.null(border)) {
    if (is.null(border.top)) {
      border.top <- border
    }
    if (is.null(border.bottom)) {
      border.bottom <- border
    }
    if (is.null(border.left)) {
      border.left <- border
    }
    if (is.null(border.right)) border.right <- border
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- border(
        x = x,
        i = i,
        j = j,
        border.top = border.top,
        border.bottom = border.bottom,
        border.left = border.left,
        border.right = border.right,
        part = p
      )
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (!is.null(border.top)) {
    x[[part]]$styles$cells[["border.style.top"]]$data[i, j] <- border.top$style
    x[[part]]$styles$cells[["border.color.top"]]$data[i, j] <- border.top$color
    x[[part]]$styles$cells[["border.width.top"]]$data[i, j] <- border.top$width
  }
  if (!is.null(border.bottom)) {
    x[[part]]$styles$cells[["border.style.bottom"]]$data[
      i,
      j
    ] <- border.bottom$style
    x[[part]]$styles$cells[["border.color.bottom"]]$data[
      i,
      j
    ] <- border.bottom$color
    x[[part]]$styles$cells[["border.width.bottom"]]$data[
      i,
      j
    ] <- border.bottom$width
  }
  if (!is.null(border.left)) {
    x[[part]]$styles$cells[["border.style.left"]]$data[
      i,
      j
    ] <- border.left$style
    x[[part]]$styles$cells[["border.color.left"]]$data[
      i,
      j
    ] <- border.left$color
    x[[part]]$styles$cells[["border.width.left"]]$data[
      i,
      j
    ] <- border.left$width
  }
  if (!is.null(border.right)) {
    x[[part]]$styles$cells[["border.style.right"]]$data[
      i,
      j
    ] <- border.right$style
    x[[part]]$styles$cells[["border.color.right"]]$data[
      i,
      j
    ] <- border.right$color
    x[[part]]$styles$cells[["border.width.right"]]$data[
      i,
      j
    ] <- border.right$width
  }

  x
}


#' @title Remove borders
#' @description The function is deleting all borders of the flextable object.
#' @inheritParams args_x_only
#' @examples
#' dat <- iris[c(1:5, 51:55, 101:105), ]
#' ft_1 <- flextable(dat)
#' ft_1 <- theme_box(ft_1)
#' ft_1
#'
#' # remove all borders
#' ft_2 <- border_remove(x = ft_1)
#' ft_2
#' @export
#' @family borders management
border_remove <- function(x) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "border_remove()"
    ))
  }
  x <- border(x = x, border = fp_border(width = 0), part = "all")
  x
}

#' @title Set outer borders
#' @description The function is applying a border to outer cells of one
#' or all parts of a flextable.
#' @inheritParams args_x_part
#' @param border border properties defined by a call to [officer::fp_border()]
#' @export
#' @examples
#' library(officer)
#' big_border <- fp_border(color = "red", width = 2)
#'
#' dat <- iris[c(1:5, 51:55, 101:105), ]
#' ft <- flextable(dat)
#' ft <- border_remove(x = ft)
#'
#' # add outer borders
#' ft <- border_outer(ft, part = "all", border = big_border)
#' ft
#' @family borders management
border_outer <- function(x, border = NULL, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "border_outer()"
    ))
  }

  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- border_outer(x = x, border = border, part = p)
    }
    return(x)
  }
  if (nrow_part(x, part) < 1) {
    return(x)
  }

  x <- hline_top(x, border = border, part = part)
  x <- hline_bottom(x, border = border, part = part)
  x <- vline_right(x, border = border, part = part)
  x <- vline_left(x, border = border, part = part)

  x
}

#' @export
#' @title Set inner horizontal borders
#' @description The function is applying a border to inner content of one
#' or all parts of a flextable.
#' @inheritParams border_outer
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "orange", width = 1)
#'
#' dat <- iris[c(1:5, 51:55, 101:105), ]
#' ft <- flextable(dat)
#' ft <- border_remove(x = ft)
#'
#' # add inner horizontal borders
#' ft <- border_inner_h(ft, border = std_border)
#' ft
#' @family borders management
border_inner_h <- function(x, border = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "border_inner_h()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- border_inner_h(x = x, border = border, part = p)
    }
    return(x)
  }
  if ((nl <- nrow_part(x, part)) < 1) {
    return(x)
  }
  at <- seq_len(nl)
  at <- at[-length(at)]
  x <- hline(x, i = at, border = border, part = part)

  x
}

#' @export
#' @title Set inner vertical borders
#' @description The function is applying a vertical border to inner content of one
#' or all parts of a flextable.
#' @inheritParams border_outer
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "orange", width = 1)
#'
#' dat <- iris[c(1:5, 51:55, 101:105), ]
#' ft <- flextable(dat)
#' ft <- border_remove(x = ft)
#'
#' # add inner vertical borders
#' ft <- border_inner_v(ft, border = std_border)
#' ft
#' @family borders management
border_inner_v <- function(x, border = NULL, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "border_inner_v()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- border_inner_v(x = x, border = border, part = p)
    }
    return(x)
  }
  if (nrow_part(x, part) < 1) {
    return(x)
  }
  at <- seq_along(x$col_keys)
  at <- at[-length(at)]
  x <- vline(x, j = at, border = border, part = part)

  x
}

#' @export
#' @title Set all inner borders
#' @description The function is applying a vertical and horizontal borders to inner content of one
#' or all parts of a flextable.
#' @inheritParams border_outer
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "orange", width = 1)
#'
#' dat <- iris[c(1:5, 51:55, 101:105), ]
#' ft <- flextable(dat)
#' ft <- border_remove(x = ft)
#'
#' # add inner vertical borders
#' ft <- border_inner(ft, border = std_border)
#' ft
#' @family borders management
border_inner <- function(x, border = NULL, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "border_inner()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- border_inner(x = x, border = border, part = p)
    }
    return(x)
  }
  if ((nl <- nrow_part(x, part)) < 1) {
    return(x)
  }

  # v
  at <- seq_along(x$col_keys)
  at <- at[-length(at)]
  x <- vline(x, j = at, border = border, part = part)

  # h
  at <- seq_len(nl)
  at <- at[-length(at)]
  x <- hline(x, i = at, border = border, part = part)

  x
}

#' @export
#' @title Set horizontal borders below selected rows
#' @description
#' `hline()` draws a horizontal line **below** each selected row by
#' setting the bottom border of cells at row `i` (and the top border
#' of cells at row `i + 1` so that the line renders consistently
#' across output formats).
#'
#' Use the `i` selector to target specific rows (e.g. a formula
#' such as `~ before(col, "Total")`). When `i` is `NULL` (the
#' default) the border is added below every row, which produces
#' a full grid of inner horizontal lines.
#'
#' For the **outer** edges of the table, use [hline_top()] and
#' [hline_bottom()] instead; those always target the very first
#' or very last row of a part.
#' @inheritParams args_selectors_with_all
#' @param border border properties defined by a call to
#' [officer::fp_border()]
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "gray")
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#'
#' # add horizontal borders below every row
#' ft <- hline(ft, part = "all", border = std_border)
#' ft
#' @family borders management
hline <- function(x, i = NULL, j = NULL, border = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "hline()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- hline(
        x = x,
        i = i,
        j = j,
        border = border,
        part = p
      )
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  x <- border(x, i = i, j = j, border.bottom = border, part = part)

  ii <- i + 1
  ii <- ii[ii > 1 & ii <= nrow_part(x, part)]

  if (length(ii) > 0) {
    x <- border(x, i = ii, j = j, border.top = border, part = part)
  }
  x
}

#' @export
#' @title Set the top border of a table part
#' @description
#' `hline_top()` draws a horizontal line at the **very top** of a
#' table part. It does not accept a row selector `i` because it
#' always targets the first row.
#'
#' When the part above exists (e.g. header above body), the line
#' is stored as the bottom border of that adjacent part so that
#' it renders seamlessly.
#'
#' Unlike [hline()], which adds inner lines below arbitrary rows,
#' `hline_top()` is meant for the outer top edge of a part.
#' @inheritParams args_x_j_part
#' @param border border properties defined by a call to
#' [officer::fp_border()]
#' @examples
#' library(officer)
#' big_border <- fp_border(color = "orange", width = 3)
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#'
#' # add a thick line on top of each part
#' ft <- hline_top(ft, part = "all", border = big_border)
#' ft
#' @family borders management
hline_top <- function(x, j = NULL, border = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "hline_top()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- hline_top(x = x, j = j, border = border, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  j <- get_columns_id(x[[part]], j)

  if (part %in% "body" && nrow_part(x, "header") > 0) {
    x <- border(
      x,
      i = nrow_part(x, "header"),
      j = j,
      border.bottom = border,
      part = "header"
    )
  } else if (part %in% "footer" && nrow_part(x, "body") > 0) {
    x <- border(
      x,
      i = nrow_part(x, "body"),
      j = j,
      border.bottom = border,
      part = "body"
    )
  } else {
    x <- border(x, i = 1, j = j, border.top = border, part = part)
  }

  x
}

#' @export
#' @title Set the bottom border of a table part
#' @description
#' `hline_bottom()` draws a horizontal line at the **very bottom**
#' of a table part. It does not accept a row selector `i` because
#' it always targets the last row.
#'
#' Unlike [hline()], which adds inner lines below arbitrary rows,
#' `hline_bottom()` is meant for the outer bottom edge of a part.
#' @inheritParams hline_top
#' @examples
#' library(officer)
#' big_border <- fp_border(color = "orange", width = 3)
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#'
#' # add a thick line at the bottom of the body
#' ft <- hline_bottom(ft, part = "body", border = big_border)
#' ft
#' @family borders management
hline_bottom <- function(x, j = NULL, border = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "hline_bottom()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- hline_bottom(x = x, j = j, border = border, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  j <- get_columns_id(x[[part]], j)
  x <- border(
    x,
    i = nrow_part(x, part),
    j = j,
    border.bottom = border,
    part = part
  )
  x
}

#' @export
#' @title Set vertical borders to the right of selected columns
#' @description
#' `vline()` draws a vertical line to the **right** of each
#' selected column by setting the right border of cells at
#' column `j` (and the left border of cells at column `j + 1`
#' so that the line renders consistently across output formats).
#'
#' Use the `j` selector to target specific columns. When `j` is
#' `NULL` (the default) the border is added to the right of every
#' column, producing a full grid of inner vertical lines.
#'
#' For the **outer** edges of the table, use [vline_left()] and
#' [vline_right()] instead; those always target the very first
#' or very last column.
#' @inheritParams args_selectors_with_all
#' @param border border properties defined by a call to
#' [officer::fp_border()]
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "orange")
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#'
#' # add vertical borders to the right of every column
#' ft <- vline(ft, border = std_border)
#' ft
#' @family borders management
vline <- function(x, i = NULL, j = NULL, border = NULL, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "vline()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- vline(x = x, i = i, j = j, border = border, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)
  x <- border(x, i = i, j = j, border.right = border, part = part)
  j <- setdiff(j, length(x$col_keys))
  if (length(j) > 0) {
    x <- border(x, i = i, j = j + 1, border.left = border, part = part)
  }
  x
}

#' @export
#' @title Set the left border of the table
#' @description
#' `vline_left()` draws a vertical line along the **left edge**
#' of the table by setting the left border of the first column.
#' It does not accept a column selector `j` because it always
#' targets column 1.
#'
#' An optional row selector `i` lets you restrict the line to
#' specific rows (e.g. only the body, or only certain rows).
#'
#' Unlike [vline()], which adds inner lines to the right of
#' arbitrary columns, `vline_left()` is meant for the outer
#' left edge of the table.
#' @inheritParams args_x_i_part
#' @param border border properties defined by a call to
#' [officer::fp_border()]
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "orange")
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#'
#' # add a border on the left edge of the table
#' ft <- vline_left(ft, border = std_border)
#' ft
#' @family borders management
vline_left <- function(x, i = NULL, border = NULL, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "vline_left()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- vline_left(x = x, i = i, border = border, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  x <- border(x, j = 1, i = i, border.left = border, part = part)
  x
}

#' @export
#' @title Set the right border of the table
#' @description
#' `vline_right()` draws a vertical line along the **right edge**
#' of the table by setting the right border of the last column.
#' It does not accept a column selector `j` because it always
#' targets the last column.
#'
#' An optional row selector `i` lets you restrict the line to
#' specific rows.
#'
#' Unlike [vline()], which adds inner lines to the right of
#' arbitrary columns, `vline_right()` is meant for the outer
#' right edge of the table.
#' @inheritParams vline_left
#' @examples
#' library(officer)
#' std_border <- fp_border(color = "orange")
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#'
#' # add a border on the right edge of the table
#' ft <- vline_right(ft, border = std_border)
#' ft
#' @family borders management
vline_right <- function(x, i = NULL, border = NULL, part = "all") {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "vline_right()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (is.null(border)) {
    border <- fp_border(color = flextable_global$defaults$border.color)
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- vline_right(x = x, i = i, border = border, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  x <- border(
    x,
    j = length(x$col_keys),
    i = i,
    border.right = border,
    part = part
  )
  x
}


#' @export
#' @title Detect rows before a given value
#' @description Returns a logical vector indicating which elements
#' of `x` appear before the first occurrence of any of the
#' `entries` values. Useful as a row selector in [hline()] to
#' insert a border above a summary row such as `"Total"`.
#' @param x an atomic vector of values to be tested
#' @param entries a sequence of items to be searched in `x`.
#' @seealso [hline()]
#' @examples
#' library(flextable)
#' library(officer)
#'
#' dat <- data.frame(
#'   stringsAsFactors = FALSE,
#'   check.names = FALSE,
#'   Level = c("setosa", "versicolor", "virginica", "<NA>", "Total"),
#'   Freq = as.integer(c(50, 50, 50, 0, 150)),
#'   `% Valid` = c(
#'     100 / 3,
#'     100 / 3, 100 / 3, NA, 100
#'   ),
#'   `% Valid Cum.` = c(100 / 3, 100 * 2 / 3, 100, NA, 100),
#'   `% Total` = c(
#'     100 / 3,
#'     100 / 3, 100 / 3, 0, 100
#'   ),
#'   `% Total Cum.` = c(
#'     100 / 3,
#'     100 * 2 / 3, 100, 100, 100
#'   )
#' )
#'
#' ft <- flextable(dat)
#' ft <- hline(ft,
#'   i = ~ before(Level, "Total"),
#'   border = fp_border_default(width = 2)
#' )
#' ft
before <- function(x, entries) {
  z <- rep(FALSE, length(x))
  index <- which(x %in% entries) - 1
  index <- index[is.finite(index) & index > 0]
  z[index] <- TRUE
  z
}

#' @export
#' @title Surround cells with borders
#' @description
#' `surround()` draws borders around specific cells, highlighting
#' them individually.
#'
#' To set borders for the whole table, use [border_outer()],
#' [border_inner_h()] and [border_inner_v()].
#'
#' All the following functions also support the
#' row and column selector `i` and `j`:
#'
#' * [hline()]: set bottom borders (inner horizontal)
#' * [vline()]: set right borders (inner vertical)
#' * [hline_top()]: set the top border (outer horizontal)
#' * [vline_left()]: set the left border (outer vertical)
#'
#' @inheritParams border
#' @family borders management
#' @examples
#' library(officer)
#' library(flextable)
#'
#' # cell to highlight
#' vary_i <- 1:3
#' vary_j <- 1:3
#'
#' std_border <- fp_border(color = "orange")
#'
#' ft <- flextable(head(iris))
#' ft <- border_remove(x = ft)
#' ft <- border_outer(x = ft, border = std_border)
#'
#' for (id in seq_along(vary_i)) {
#'   ft <- bg(
#'     x = ft,
#'     i = vary_i[id],
#'     j = vary_j[id], bg = "yellow"
#'   )
#'   ft <- surround(
#'     x = ft,
#'     i = vary_i[id],
#'     j = vary_j[id],
#'     border.left = std_border,
#'     border.right = std_border,
#'     part = "body"
#'   )
#' }
#'
#' ft <- autofit(ft)
#' ft
#' # # render
#' # print(ft, preview = "pptx")
#' # print(ft, preview = "docx")
#' # print(ft, preview = "pdf")
#' # print(ft, preview = "html")
surround <- function(
  x,
  i = NULL,
  j = NULL,
  border = NULL,
  border.top = NULL,
  border.bottom = NULL,
  border.left = NULL,
  border.right = NULL,
  part = "body"
) {
  if (!inherits(x, "flextable")) {
    stop(sprintf(
      "Function `%s` supports only flextable objects.",
      "surround()"
    ))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (!is.null(border)) {
    if (is.null(border.top)) {
      border.top <- border
    }
    if (is.null(border.bottom)) {
      border.bottom <- border
    }
    if (is.null(border.left)) {
      border.left <- border
    }
    if (is.null(border.right)) border.right <- border
  }

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- surround(
        x = x,
        i = i,
        j = j,
        border.top = border.top,
        border.bottom = border.bottom,
        border.left = border.left,
        border.right = border.right,
        part = p
      )
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  check_formula_i_and_part(i, part)
  i <- get_rows_id(x[[part]], i)
  j <- get_columns_id(x[[part]], j)

  if (!is.null(border.top)) {
    itop <- setdiff(i, 1) - 1
    x <- hline(x, i = itop, j = j, part = part, border = border.top)
    if (1 %in% i) {
      x <- hline_top(x, j = j, part = part, border = border.top)
    }
  }
  if (!is.null(border.bottom)) {
    x <- hline(x, i = i, j = j, part = part, border = border.bottom)
  }
  if (!is.null(border.left)) {
    jleft <- setdiff(j, 1) - 1
    x <- vline(x, i = i, j = jleft, part = part, border = border.left)
    if (1 %in% j) {
      x <- vline_left(x, i = i, part = part, border = border.left)
    }
  }
  if (!is.null(border.right)) {
    x <- vline(x, i = i, j = j, part = part, border = border.right)
  }

  x
}

#' @importFrom officer fp_cell fp_par fp_text fp_border
complex_tabpart <- function(
  data,
  col_keys = names(data),
  default_pr_text = fp_text(
    font.family = flextable_global$defaults$font.family,
    cs.family = flextable_global$defaults$cs.family,
    eastasia.family = flextable_global$defaults$eastasia.family,
    hansi.family = flextable_global$defaults$hansi.family,
    font.size = flextable_global$defaults$font.size,
    color = flextable_global$defaults$font.color
  ),
  default_pr_par = fp_par(
    text.align = flextable_global$defaults$text.align,
    padding.left = flextable_global$defaults$padding.left,
    padding.right = flextable_global$defaults$padding.right,
    padding.bottom = flextable_global$defaults$padding.bottom,
    padding.top = flextable_global$defaults$padding.top,
    line_spacing = flextable_global$defaults$line_spacing
  ),
  default_pr_cell = fp_cell(
    background.color = flextable_global$defaults$background.color,
    border = fp_border(color = "transparent", width = 0)
  ),
  cwidth = NULL,
  cheight = NULL
) {
  pr_cell_init <- as_struct(nrow(data), col_keys, default_pr_cell, cell_struct)
  pr_par_init <- as_struct(nrow(data), col_keys, default_pr_par, par_struct)
  pr_text_init <- as_struct(nrow(data), col_keys, default_pr_text, text_struct)

  content <- new_chunkset_struct(col_keys = col_keys, data = data)

  span_init <- matrix(1L, nrow = nrow(data), ncol = length(col_keys))
  spans <- list(rows = span_init, columns = span_init)

  if (length(cwidth) == length(col_keys)) {
    colwidths <- cwidth
  } else {
    colwidths <- rep(cwidth, length(col_keys))
  }

  rowheights <- rep(cheight, nrow(data))

  out <- list(
    dataset = data,
    content = content,
    col_keys = col_keys,
    colwidths = colwidths,
    rowheights = rowheights,
    hrule = rep("auto", nrow(data)),
    spans = spans,
    styles = list(
      cells = pr_cell_init,
      pars = pr_par_init,
      text = pr_text_init
    )
  )
  class(out) <- "complex_tabpart"
  out
}

#' @title Wrap a flextable for use with patchwork
#'
#' @description
#' This function wraps a flextable as a patchwork-compliant patch, similar
#' to what [patchwork::wrap_table()] does for gt tables. It allows flextable
#' objects to be combined with ggplot2 plots in a patchwork layout, with
#' optional alignment of table headers and body with plot panel areas.
#'
#' Note this is experimental and may change in the future.
#' @param x A flextable object.
#' @param panel What portion of the table should be aligned with the panel
#' region? `"body"` means that header and footer will be placed outside the
#' panel region. `"full"` means that the whole table will be placed inside the
#' panel region. `"rows"` keeps all rows inside the panel but is otherwise
#' equivalent to `"body"`. `"cols"` places all columns within the panel region
#' but keeps column headers on top.
#' @param space How should the dimension of the table influence the final
#' composition? `"fixed"` means that the table width and height will set the
#' dimensions of the area it occupies. `"free"` means the table dimensions will
#' not influence the sizing. `"free_x"` and `"free_y"` allow freeing either
#' direction.
#' @param n_row_headers Number of leading columns to treat as row headers.
#' These columns will be placed outside the panel region and will not
#' participate in alignment with the plot axes.
#' @param flex_body If `TRUE`, the table body row heights become flexible:
#' the adjacent ggplot determines the panel height and the body rows
#' stretch equally to fill it. Header and footer keep their fixed size.
#' This is useful to align table rows with discrete bars or categories
#' in a neighbouring plot. Implies `free_y` for `space`.
#' @param flex_cols If `TRUE`, the data column widths (all columns after
#' `n_row_headers`) become flexible: the adjacent ggplot determines the
#' panel width and the data columns stretch to fill it. Row header
#' columns keep their fixed width. This is useful to align table columns
#' with discrete x-axis categories in a neighbouring plot.
#' Implies `free_x` for `space`.
#' @param expand Expansion value matching the ggplot discrete axis expansion
#' (`ggplot2::expansion(add = expand)`). Default is `0.6`, which is the
#' ggplot2 default for discrete axes. Only used when `flex_cols = TRUE`.
#' @param just Horizontal alignment of the table within its patchwork panel.
#' One of `"left"` (default), `"right"`, or `"center"`. Useful when the
#' table is narrower than the available panel width.
#' Ignored when `flex_cols = TRUE` (columns fill the panel).
#' @return A patchwork-compliant object that can be combined with ggplot2 plots
#' using `+`, `|`, or `/` operators.
#' @export
#' @examplesIf requireNamespace("patchwork", quietly = TRUE) && requireNamespace("ggplot2", quietly = TRUE) && requireNamespace("ragg", quietly = TRUE)
#' library(gdtools)
#' font_set_liberation()
#' library(ggplot2)
#' library(patchwork)
#'
#' set_flextable_defaults(
#'   font.family = "Liberation Sans",
#'   font.size = 10,
#'   big.mark = "",
#'   border.color = "grey60"
#' )
#'
#' # Adapted from <https://r-graph-gallery.com/web-dumbell-chart.html>
#'
#' dataset <- data.frame(
#'   team = c(
#'     "FC Bayern Munchen", "SV Werder Bremen", "Borussia Dortmund",
#'     "VfB Stuttgart", "Borussia M'gladbach", "Hamburger SV",
#'     "Eintracht Frankfurt", "FC Schalke 04", "1. FC Koln",
#'     "Bayer 04 Leverkusen"
#'   ),
#'   matches = c(2000, 1992, 1924, 1924, 1898, 1866, 1856, 1832, 1754, 1524),
#'   won     = c(1206,  818,  881,  782,  763,  746,  683,  700,  674,  669),
#'   lost    = c( 363,  676,  563,  673,  636,  625,  693,  669,  628,  447)
#' )
#' dataset$win_pct  <- dataset$won  / dataset$matches * 100
#' dataset$loss_pct <- dataset$lost / dataset$matches * 100
#' dataset$team <- factor(dataset$team, levels = rev(dataset$team))
#'
#' # -- dumbbell chart --
#' pal <- c(lost = "#EFAC00", won = "#28A87D")
#' df_long <- reshape(dataset, direction = "long",
#'   varying = list(c("loss_pct", "win_pct")),
#'   v.names = "pct", timevar = "type",
#'   times = c("lost", "won"), idvar = "team"
#' )
#'
#' p <- ggplot(df_long, aes(x = pct / 100, y = team)) +
#'   stat_summary(
#'     geom = "linerange", fun.min = "min", fun.max = "max",
#'     linewidth = .7, color = "grey60"
#'   ) +
#'   geom_point(aes(fill = type), size = 4, shape = 21,
#'     stroke = .8, color = "white"
#'   ) +
#'   scale_x_continuous(
#'     labels = scales::percent,
#'     expand = expansion(add = c(.02, .02))
#'   ) +
#'   scale_y_discrete(name = NULL, guide = "none") +
#'   scale_fill_manual(
#'     values = pal,
#'     labels = c(lost = "Lost", won = "Won")
#'   ) +
#'   labs(x = NULL, fill = NULL) +
#'   theme_minimal(base_family = "Liberation Sans", base_size = 10) +
#'   theme(
#'     legend.position = "top",
#'     legend.justification = "left",
#'     panel.grid.minor = element_blank(),
#'     panel.grid.major.y = element_blank()
#'   )
#'
#' # -- flextable --
#' ft_dat <- dataset[, c("matches", "win_pct", "loss_pct", "team")]
#' ft_dat$team <- as.character(ft_dat$team)
#'
#' ft <- flextable(ft_dat)
#' ft <- border_remove(ft)
#' ft <- bold(ft, part = "header")
#' ft <- colformat_double(ft, j = c("win_pct", "loss_pct"),
#'   digits = 1, suffix = "%"
#' )
#' ft <- set_header_labels(ft,
#'   team = "Team", matches = "GP",
#'   win_pct = "", loss_pct = ""
#' )
#' ft <- color(ft, color = "#28A87D", j = 2)
#' ft <- color(ft, color = "#EFAC00", j = 3)
#' ft <- bold(ft, bold = TRUE, j = 2:3)
#' ft <- italic(ft, italic = TRUE, j = 4)
#' ft <- align(ft, align = "right", part = "all")
#' ft <- autofit(ft)
#'
#' \dontshow{
#' cap <- ragg::agg_capture(width = 7, height = 6, units = "in", res = 150)
#' grDevices::dev.control("enable")
#' }
#' print(
#'   wrap_flextable(ft, flex_body = TRUE, just = "right") +
#'     p + plot_layout(widths = c(1.1, 2))
#' )
#' \dontshow{
#' raster <- cap()
#' dev.off()
#' plot(as.raster(raster))
#' init_flextable_defaults()
#' }
#' @family flextable_output_export
wrap_flextable <- function(
  x,
  panel = c("body", "full", "rows", "cols"),
  space = c("free", "free_x", "free_y", "fixed"),
  n_row_headers = 0L,
  flex_body = FALSE,
  flex_cols = FALSE,
  expand = 0.6,
  just = c("left", "right", "center")
) {
  if (!requireNamespace("patchwork", quietly = TRUE)) {
    stop("Package 'patchwork' is required for wrap_flextable().", call. = FALSE)
  }
  if (!requireNamespace("gtable", quietly = TRUE)) {
    stop("Package 'gtable' is required for wrap_flextable().", call. = FALSE)
  }
  panel <- match.arg(panel)
  space <- match.arg(space)
  just <- match.arg(just)
  flex_body <- isTRUE(flex_body)
  flex_cols <- isTRUE(flex_cols)

  if (flex_body) {
    attr(x, ".patchwork_flex_body") <- TRUE
  }
  if (flex_cols) {
    attr(x, ".patchwork_flex_cols") <- TRUE
    attr(x, ".patchwork_n_row_headers") <- as.integer(n_row_headers)
    attr(x, ".patchwork_flex_cols_expand") <- expand
  }
  attr(x, ".patchwork_just") <- just

  wrapped <- patchwork::wrap_elements(panel = x, ignore_tag = FALSE)
  attr(wrapped, "patch_settings")$panel <- panel
  attr(wrapped, "patch_settings")$n_row_headers <- as.integer(n_row_headers)
  attr(wrapped, "patch_settings")$space <- c(
    if (flex_cols) TRUE else space %in% c("free", "free_x"),
    if (flex_body) TRUE else space %in% c("free", "free_y")
  )
  class(wrapped) <- c("wrapped_table", class(wrapped))
  wrapped
}

# S3 method: ggplot_add.flextable -------------------------------------------
# Enables: ggplot_obj + flextable_obj
ggplot_add.flextable <- function(object, plot, object_name) {
  plot + wrap_flextable(object)
}

# S3 method: as_patch.flextable ---------------------------------------------
# Called by patchwork when it needs to convert a flextable to a grob.
# Returns a gtable with named components so that patchwork's
# patchGrob.wrapped_table() can split header/body/footer.
#' @importFrom grid unit viewport grobWidth grobHeight
as_patch.flextable <- function(x, ...) {
  if (!requireNamespace("gtable", quietly = TRUE)) {
    stop("Package 'gtable' is required.", call. = FALSE)
  }

  flex_body <- isTRUE(attr(x, ".patchwork_flex_body"))
  flex_cols <- isTRUE(attr(x, ".patchwork_flex_cols"))
  n_row_hdrs <- as.integer(attr(x, ".patchwork_n_row_headers") %||% 0L)

  n_header <- nrow_part(x, "header")
  n_body <- nrow_part(x, "body")
  n_footer <- nrow_part(x, "footer")
  n_rows <- n_header + n_body + n_footer

  if (flex_body || flex_cols) {
    grob <- gen_grob(x, fit = "auto", scaling = "fixed")
  } else {
    grob <- gen_grob(x, fit = "fixed", scaling = "fixed")
  }
  widths <- grob$ftpar$widths
  heights <- grob$ftpar$heights
  n_cols <- length(widths)

  # --- row heights ---
  if (flex_body && n_body > 0) {
    grob$ftpar$flex_body <- TRUE
    grob$ftpar$n_header_rows <- n_header
    grob$ftpar$n_body_rows <- n_body
    grob$ftpar$n_footer_rows <- n_footer

    row_heights <- unit(heights, "in")
    body_seq <- seq.int(n_header + 1L, n_header + n_body)
    row_heights[body_seq] <- unit(rep(1, n_body), "null")
  } else {
    row_heights <- unit(heights, "in")
  }

  # --- column widths ---
  n_data_cols <- n_cols - n_row_hdrs
  if (flex_cols && n_data_cols > 0) {
    grob$ftpar$flex_cols <- TRUE
    grob$ftpar$n_row_header_cols <- n_row_hdrs
    grob$ftpar$n_data_cols <- n_data_cols

    col_widths <- unit(widths, "in")
    data_seq <- seq.int(n_row_hdrs + 1L, n_cols)
    col_widths[data_seq] <- unit(rep(1, n_data_cols), "null")
  } else {
    col_widths <- unit(widths, "in")
  }

  # Create the gtable skeleton
  gt <- gtable::gtable(
    widths = col_widths,
    heights = row_heights
  )

  # Add the flextable grob spanning the entire table
  gt <- gtable::gtable_add_grob(
    gt,
    grobs = list(grob),
    t = 1L,
    l = 1L,
    b = n_rows,
    r = n_cols,
    clip = "off",
    name = "table"
  )

  # Add a zero-size grob for "table_body" so patchwork can find boundaries
  body_top <- n_header + 1L
  body_bottom <- n_header + n_body
  if (n_body > 0) {
    gt <- gtable::gtable_add_grob(
      gt,
      grobs = list(grid::nullGrob()),
      t = body_top,
      l = 1L,
      b = body_bottom,
      r = n_cols,
      clip = "off",
      name = "table_body"
    )
  }

  # --- viewport height ---
  if (flex_body && n_header > 0) {
    header_height <- sum(unit(heights[seq_len(n_header)], "in"))
    vp_height <- unit(1, "npc") + header_height
  } else if (flex_body) {
    vp_height <- unit(1, "npc")
  } else {
    vp_height <- sum(row_heights)
  }

  # --- viewport width and horizontal position ---
  if (flex_cols && n_data_cols > 0) {
    expand_val <- as.numeric(
      attr(x, ".patchwork_flex_cols_expand") %||% 0.6
    )
    # clamp: below 0.5, viewport would overflow the panel
    expand_eff <- max(expand_val, 0.5)
    range_units <- n_data_cols - 1 + 2 * expand_eff
    # data columns span n_dc/range of the panel; margins are empty
    cat_fraction <- n_data_cols / range_units
    vp_x <- unit((expand_eff - 0.5) / range_units, "npc")
    if (n_row_hdrs > 0) {
      row_hdr_width <- sum(unit(widths[seq_len(n_row_hdrs)], "in"))
      vp_width <- unit(cat_fraction, "npc") + row_hdr_width
    } else {
      vp_width <- unit(cat_fraction, "npc")
    }
    vp_just_x <- 0
  } else {
    halign <- attr(x, ".patchwork_just") %||% "left"
    if (halign == "right") {
      vp_x <- unit(1, "npc")
      vp_just_x <- 1
    } else if (halign == "center") {
      vp_x <- unit(0.5, "npc")
      vp_just_x <- 0.5
    } else {
      vp_x <- unit(0, "npc")
      vp_just_x <- 0
    }
    vp_width <- sum(col_widths)
  }

  gt$vp <- viewport(
    x = vp_x,
    y = 1,
    width = vp_width,
    height = vp_height,
    default.units = "npc",
    just = c(vp_just_x, 1)
  )

  gt
}

# .onLoad registration ------------------------------------------------------
# Register S3 methods for generics from other packages (ggplot2 and patchwork)
# so that flextable objects are handled natively.
.onLoad_patchwork <- function() {
  if (requireNamespace("ggplot2", quietly = TRUE)) {
    registerS3method(
      "ggplot_add",
      "flextable",
      ggplot_add.flextable,
      envir = asNamespace("ggplot2")
    )
  }
  if (requireNamespace("patchwork", quietly = TRUE)) {
    registerS3method(
      "as_patch",
      "flextable",
      as_patch.flextable,
      envir = asNamespace("patchwork")
    )
  }
}

skip_if_not_installed("patchwork")
skip_if_not_installed("ggplot2")
skip_if_not_installed("gtable")

gdtools::register_liberationsans()

init_flextable_defaults()
set_flextable_defaults(
  font.family = "Liberation Sans",
  border.color = "#333333"
)

make_ft <- function(n = 3) {
  dat <- data.frame(
    label = head(LETTERS, n),
    value = seq_len(n)
  )
  autofit(flextable(dat))
}

make_ft_with_footer <- function() {
  ft <- make_ft()
  ft <- add_footer_lines(ft, values = "a note")
  autofit(ft)
}

# -- wrap_flextable ----------------------------------------------------------

test_that("wrap_flextable returns a wrapped_table", {
  ft <- make_ft()
  w <- wrap_flextable(ft)
  expect_s3_class(w, "wrapped_table")
  expect_s3_class(w, "wrapped_patch")
})

test_that("wrap_flextable panel argument is validated", {
  ft <- make_ft()
  expect_error(wrap_flextable(ft, panel = "invalid"), "should be one of")
  for (p in c("body", "full", "rows", "cols")) {
    w <- wrap_flextable(ft, panel = p)
    expect_equal(attr(w, "patch_settings")$panel, p)
  }
})

test_that("wrap_flextable space argument sets free axes", {
  ft <- make_ft()

  w_fixed <- wrap_flextable(ft, space = "fixed")
  expect_equal(attr(w_fixed, "patch_settings")$space, c(FALSE, FALSE))

  w_free <- wrap_flextable(ft, space = "free")
  expect_equal(attr(w_free, "patch_settings")$space, c(TRUE, TRUE))

  w_fx <- wrap_flextable(ft, space = "free_x")
  expect_equal(attr(w_fx, "patch_settings")$space, c(TRUE, FALSE))

  w_fy <- wrap_flextable(ft, space = "free_y")
  expect_equal(attr(w_fy, "patch_settings")$space, c(FALSE, TRUE))
})

test_that("wrap_flextable flex_body forces free y", {
  ft <- make_ft()
  w <- wrap_flextable(ft, space = "fixed", flex_body = TRUE)
  space <- attr(w, "patch_settings")$space
  expect_true(space[2])
})

test_that("wrap_flextable flex_cols forces free x", {
  ft <- make_ft()
  w <- wrap_flextable(ft, space = "fixed", flex_cols = TRUE, n_row_headers = 1)
  space <- attr(w, "patch_settings")$space
  expect_true(space[1])
  expect_equal(attr(w, "patch_settings")$n_row_headers, 1L)
})

test_that("wrap_flextable just is forwarded to as_patch viewport", {
  ft <- make_ft()
  expected_x <- c(left = 0, right = 1, center = 0.5)
  for (j in c("left", "right", "center")) {
    w <- wrap_flextable(ft, just = j)
    # The just attribute is on the inner flextable; verify via as_patch
    ft_inner <- ft
    attr(ft_inner, ".patchwork_just") <- j
    gt <- flextable:::as_patch.flextable(ft_inner)
    expect_equal(as.numeric(gt$vp$x), expected_x[[j]])
  }
})

# -- as_patch.flextable (S3, explicit name for covr) -------------------------

test_that("as_patch.flextable returns a gtable", {
  ft <- make_ft()
  gt <- flextable:::as_patch.flextable(ft)
  expect_s3_class(gt, "gtable")
  expect_true(!is.null(gt$vp))
})

test_that("as_patch.flextable dimensions match flextable", {
  ft <- make_ft()
  gt <- flextable:::as_patch.flextable(ft)

  n_header <- nrow_part(ft, "header")
  n_body <- nrow_part(ft, "body")
  expected_rows <- n_header + n_body
  expect_equal(nrow(gt), expected_rows)
  expect_equal(ncol(gt), ncol_keys(ft))
})

test_that("as_patch.flextable with footer adds rows", {
  ft <- make_ft_with_footer()
  gt <- flextable:::as_patch.flextable(ft)

  n_header <- nrow_part(ft, "header")
  n_body <- nrow_part(ft, "body")
  n_footer <- nrow_part(ft, "footer")
  expect_equal(nrow(gt), n_header + n_body + n_footer)
})

test_that("as_patch.flextable has table_body grob", {
  ft <- make_ft()
  gt <- flextable:::as_patch.flextable(ft)
  grob_names <- gt$layout$name
  expect_true("table" %in% grob_names)
  expect_true("table_body" %in% grob_names)
})

test_that("as_patch.flextable body grob spans correct rows", {
  ft <- make_ft()
  gt <- flextable:::as_patch.flextable(ft)

  body_layout <- gt$layout[gt$layout$name == "table_body", ]
  n_header <- nrow_part(ft, "header")
  n_body <- nrow_part(ft, "body")
  expect_equal(body_layout$t, n_header + 1L)
  expect_equal(body_layout$b, n_header + n_body)
})

test_that("as_patch.flextable flex_body uses null units for body rows", {
  ft <- make_ft()
  attr(ft, ".patchwork_flex_body") <- TRUE
  gt <- flextable:::as_patch.flextable(ft)

  n_header <- nrow_part(ft, "header")
  n_body <- nrow_part(ft, "body")
  body_seq <- seq.int(n_header + 1L, n_header + n_body)

  row_units <- grid::unitType(gt$heights)
  for (i in body_seq) {
    expect_equal(row_units[i], "null")
  }

  if (n_header > 0) {
    for (i in seq_len(n_header)) {
      expect_equal(row_units[i], "inches")
    }
  }
})

test_that("as_patch.flextable flex_cols uses null units for data columns", {
  ft <- make_ft()
  attr(ft, ".patchwork_flex_cols") <- TRUE
  attr(ft, ".patchwork_n_row_headers") <- 1L
  attr(ft, ".patchwork_flex_cols_expand") <- 0.6
  gt <- flextable:::as_patch.flextable(ft)

  n_cols <- ncol_keys(ft)
  data_seq <- seq.int(2L, n_cols)

  col_units <- grid::unitType(gt$widths)
  for (i in data_seq) {
    expect_equal(col_units[i], "null")
  }
  expect_equal(col_units[1], "inches")
})

test_that("as_patch.flextable just controls viewport x", {
  ft <- make_ft()

  attr(ft, ".patchwork_just") <- "left"
  gt_left <- flextable:::as_patch.flextable(ft)
  expect_equal(as.numeric(gt_left$vp$x), 0)

  attr(ft, ".patchwork_just") <- "right"
  gt_right <- flextable:::as_patch.flextable(ft)
  expect_equal(as.numeric(gt_right$vp$x), 1)

  attr(ft, ".patchwork_just") <- "center"
  gt_center <- flextable:::as_patch.flextable(ft)
  expect_equal(as.numeric(gt_center$vp$x), 0.5)
})

# -- ggplot_add.flextable (S3, explicit name for covr) -----------------------

test_that("ggplot_add.flextable wraps and adds to plot", {
  ft <- make_ft()
  p <- ggplot2::ggplot(data.frame(x = 1:3, y = 1:3), ggplot2::aes(x, y)) +
    ggplot2::geom_point()

  result <- flextable:::ggplot_add.flextable(ft, p, "ft")
  expect_s3_class(result, "patchwork")
})

# -- rendering with ragg (end-to-end) ----------------------------------------

test_that("patchwork composition renders to PNG with ragg", {
  ft <- make_ft()
  p <- ggplot2::ggplot(data.frame(x = 1:3, y = 1:3), ggplot2::aes(x, y)) +
    ggplot2::geom_point() +
    ggplot2::theme_minimal(base_family = "Liberation Sans")

  pw <- wrap_flextable(ft) + p

  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)

  ragg::agg_png(filename = tf, width = 8, height = 4, units = "in", res = 150)
  print(pw)
  dev.off()

  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("patchwork with flex_body renders to PNG with ragg", {
  ft <- make_ft()
  p <- ggplot2::ggplot(
    data.frame(x = 1:3, y = factor(LETTERS[1:3], levels = rev(LETTERS[1:3]))),
    ggplot2::aes(x, y)
  ) +
    ggplot2::geom_col() +
    ggplot2::theme_minimal(base_family = "Liberation Sans")

  pw <- wrap_flextable(ft, flex_body = TRUE) + p

  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)

  ragg::agg_png(filename = tf, width = 8, height = 4, units = "in", res = 150)
  print(pw)
  dev.off()

  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("patchwork with flex_cols renders to PNG with ragg", {
  ft <- make_ft()
  p <- ggplot2::ggplot(
    data.frame(x = factor(c("label", "value")), y = c(10, 20)),
    ggplot2::aes(x, y)
  ) +
    ggplot2::geom_col() +
    ggplot2::theme_minimal(base_family = "Liberation Sans")

  pw <- wrap_flextable(ft, flex_cols = TRUE, n_row_headers = 0) / p

  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)

  ragg::agg_png(filename = tf, width = 6, height = 6, units = "in", res = 150)
  print(pw)
  dev.off()

  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

test_that("patchwork with just right renders to PNG with ragg", {
  ft <- make_ft()
  p <- ggplot2::ggplot(data.frame(x = 1:3, y = 1:3), ggplot2::aes(x, y)) +
    ggplot2::geom_point() +
    ggplot2::theme_minimal(base_family = "Liberation Sans")

  pw <- wrap_flextable(ft, just = "right") + p

  tf <- tempfile(fileext = ".png")
  on.exit(unlink(tf), add = TRUE)

  ragg::agg_png(filename = tf, width = 8, height = 4, units = "in", res = 150)
  print(pw)
  dev.off()

  expect_true(file.exists(tf))
  expect_gt(file.info(tf)$size, 1000)
})

init_flextable_defaults()

#' @title Split a flextable into pages by columns
#'
#' @description Split a flextable into a list of flextables whose columns
#' fit within a given width (in inches). This is useful for paginating
#' wide tables in Word or PowerPoint output.
#'
#' @inheritParams args_x_only
#' @param max_width Maximum width for each page (in inches by default).
#' @param rep_cols Columns to repeat on every page. Can be a character
#'   vector of column names or an integer vector of column positions.
#'   `NULL` (default) means no repetition. Repeated columns appear at
#'   the beginning of each page in the order specified.
#' @param unit Unit for `max_width`, one of `"in"`, `"cm"`, `"mm"`.
#' @return A list of flextable objects. If no splitting is needed, a
#' single-element list is returned.
#' @family table_structure
#' @examples
#' ft <- flextable(head(mtcars))
#' ft_pages <- split_columns(ft, max_width = 5)
#' length(ft_pages)
#' @export
split_columns <- function(x, max_width, rep_cols = NULL, unit = "in") {
  max_width <- convin(unit = unit, x = max_width)
  all_keys <- x$col_keys

  if (is.null(rep_cols)) {
    rep_keys <- character(0)
  } else {
    ref_part <- if (nrow_part(x, "header") > 0L) x$header else x$body
    # as_col_keys validates and resolves formulas/logicals/indices to names
    resolved <- as_col_keys(ref_part, j = rep_cols, blanks = character())
    # preserve user-specified order (as_col_keys reorders to dataset order)
    if (is.character(rep_cols)) {
      rep_keys <- rep_cols[rep_cols %in% resolved]
    } else if (is.numeric(rep_cols)) {
      rep_keys <- all_keys[rep_cols[
        rep_cols >= 1L & rep_cols <= length(all_keys)
      ]]
    } else {
      rep_keys <- resolved
    }
  }

  data_keys <- setdiff(all_keys, rep_keys)

  widths <- dim_pretty(x)$widths
  names(widths) <- all_keys
  rep_width <- sum(widths[rep_keys])
  data_widths <- widths[data_keys]
  available <- max_width - rep_width

  if (available <= 0) {
    warning(
      "max_width is too small to fit the repeated columns, no pagination performed."
    )
    return(list(x))
  }

  # greedy packing
  pages <- list()
  current <- integer(0)
  current_sum <- 0
  for (k in seq_along(data_keys)) {
    if (current_sum + data_widths[k] > available && length(current) > 0L) {
      pages <- c(pages, list(current))
      current <- integer(0)
      current_sum <- 0
    }
    current <- c(current, k)
    current_sum <- current_sum + data_widths[k]
  }
  if (length(current) > 0L) {
    pages <- c(pages, list(current))
  }

  if (length(pages) <= 1L) {
    return(list(x))
  }

  lapply(pages, function(idx) {
    keep <- c(rep_keys, data_keys[idx])
    drop <- setdiff(all_keys, keep)
    if (length(drop) > 0L) {
      out <- delete_columns(x, j = drop)
    } else {
      out <- x
    }
    out$col_keys <- keep
    for (pt in c("header", "body", "footer")) {
      if (nrow_part(out, pt) > 0L) {
        out[[pt]]$col_keys <- keep
      }
    }
    out
  })
}

#' @title Split a flextable into pages by rows
#'
#' @description Split a flextable into a list of flextables whose body
#' rows fit within a given height (in inches). Header and footer are
#' repeated on every page. An optional `group` argument keeps row groups
#' together (no page break inside a group).
#'
#' @note Footnotes are currently repeated on every page, even when
#' they reference rows that only appear on a specific page. This
#' limitation will be resolved in a future version when footnotes
#' are restructured to track their association with body rows.
#'
#' @inheritParams args_x_only
#' @param max_height Maximum height for each page, including
#' header and footer (in inches by default).
#' @param group Integer vector of body row indices that start a new group.
#' Rows belonging to the same group are kept together on a single page.
#' Default is `integer(0)` (no grouping, every row is independent).
#' @param unit Unit for `max_height`, one of `"in"`, `"cm"`, `"mm"`.
#' @return A list of flextable objects. If no splitting is needed, a
#' single-element list is returned.
#' @family table_structure
#' @examples
#' ft <- flextable(iris)
#' ft_pages <- split_rows(ft, max_height = 3)
#' length(ft_pages)
#' @export
split_rows <- function(x, max_height, group = integer(0), unit = "in") {
  max_height <- convin(unit = unit, x = max_height)
  heights <- dim_pretty(x)$heights
  n_hdr <- nrow_part(x, "header")
  n_ftr <- nrow_part(x, "footer")
  n_body <- nrow_part(x, "body")

  hdr_height <- sum(heights[seq_len(n_hdr)])
  if (n_ftr > 0L) {
    ftr_height <- sum(heights[seq(n_hdr + n_body + 1L, length(heights))])
  } else {
    ftr_height <- 0
  }
  body_heights <- heights[seq(n_hdr + 1L, n_hdr + n_body)]
  available <- max_height - hdr_height - ftr_height

  if (available <= 0) {
    warning(
      "max_height is too small to fit header and footer, no pagination performed."
    )
    return(list(x))
  }

  # Build groups from group starts (indices of group starts in body)
  if (length(group) > 0L) {
    if (is.unsorted(group, strictly = TRUE)) {
      stop(
        "group must be a strictly increasing vector of row indices.",
        call. = FALSE
      )
    }
    if (any(group < 1L | group > n_body)) {
      stop("group indices must be between 1 and ", n_body, ".", call. = FALSE)
    }
    group_ids <- findInterval(seq_len(n_body), group)
  } else {
    group_ids <- seq_len(n_body)
  }

  # Greedy packing of groups into pages
  pages <- list()
  current <- integer(0)
  current_height <- 0
  for (g in unique(group_ids)) {
    rows_in_group <- which(group_ids == g)
    group_height <- sum(body_heights[rows_in_group])
    if (current_height + group_height > available && length(current) > 0L) {
      pages <- c(pages, list(current))
      current <- integer(0)
      current_height <- 0
    }
    current <- c(current, rows_in_group)
    current_height <- current_height + group_height
  }
  if (length(current) > 0L) {
    pages <- c(pages, list(current))
  }

  if (length(pages) <= 1L) {
    return(list(x))
  }

  # Split via delete_rows (spans preserved when no vertical span is broken)
  lapply(pages, function(keep_rows) {
    drop_rows <- setdiff(seq_len(n_body), keep_rows)
    if (length(drop_rows) > 0L) {
      delete_rows(x, i = drop_rows, part = "body")
    } else {
      x
    }
  })
}

#' @export
#' @title Split a flextable into pages by rows and columns
#' @description Split a flextable into a list of flextables that fit
#' within the given height and width constraints. Rows are split
#' first, then columns are split within each row page.
#'
#' This is a convenience wrapper around [split_rows()] and
#' [split_columns()].
#' @inheritParams args_x_only
#' @param max_width Maximum width per page (in inches by default).
#'   `NULL` means no column splitting.
#' @param max_height Maximum height per page, including header and
#'   footer (in inches by default). `NULL` means no row splitting.
#' @param rep_cols Columns to repeat on every horizontal page. Can be
#'   a character vector of column names, an integer vector of column
#'   positions, or `NULL` (default, no repetition).
#' @param group Integer vector of body row indices that start a new
#'   group. Default is `integer(0)`.
#' @param unit Unit for `max_width` and `max_height`, one of
#'   `"in"`, `"cm"`, `"mm"`.
#' @return A list of flextable objects.
#' @family table_structure
#' @examples
#' ft <- flextable(iris)
#' ft_pages <- split_to_pages(ft, max_width = 4, max_height = 5)
#' length(ft_pages)
split_to_pages <- function(
  x,
  max_width = NULL,
  max_height = NULL,
  rep_cols = NULL,
  group = integer(0),
  unit = "in"
) {
  ft_list <- list(x)

  if (!is.null(max_height)) {
    ft_list <- split_rows(
      x,
      max_height = max_height,
      group = group,
      unit = unit
    )
  }

  if (!is.null(max_width)) {
    ft_list <- unlist(
      lapply(
        ft_list,
        split_columns,
        max_width = max_width,
        rep_cols = rep_cols,
        unit = unit
      ),
      recursive = FALSE
    )
  }

  ft_list
}

# --- split_columns ---

test_that("split_columns returns single page when table fits", {
  ft <- flextable(head(mtcars[, 1:3]))
  pages <- split_columns(ft, max_width = 20)
  expect_length(pages, 1)
})

test_that("split_columns splits wide table", {
  ft <- flextable(head(mtcars))
  pages <- split_columns(ft, max_width = 5)
  expect_gt(length(pages), 1)
  for (p in pages) {
    expect_s3_class(p, "flextable")
  }
})

test_that("split_columns rep_cols by name repeats columns", {
  ft <- flextable(head(mtcars))
  pages <- split_columns(ft, max_width = 5, rep_cols = c("mpg", "cyl"))
  for (p in pages) {
    expect_true(all(c("mpg", "cyl") %in% p$col_keys))
  }
})

test_that("split_columns rep_cols by index repeats columns", {
  ft <- flextable(head(mtcars))
  pages <- split_columns(ft, max_width = 5, rep_cols = 1:2)
  for (p in pages) {
    expect_true(all(c("mpg", "cyl") %in% p$col_keys))
  }
})

test_that("split_columns preserves rep_cols order", {
  ft <- flextable(head(mtcars))
  pages <- split_columns(ft, max_width = 5, rep_cols = c("am", "mpg"))
  for (p in pages) {
    idx_am <- match("am", p$col_keys)
    idx_mpg <- match("mpg", p$col_keys)
    expect_lt(idx_am, idx_mpg)
  }
})

test_that("split_columns warns on invalid column name", {
  ft <- flextable(head(mtcars))
  expect_warning(
    split_columns(ft, max_width = 5, rep_cols = "nonexistent"),
    "can not be found"
  )
})

test_that("split_columns warns when rep_cols exceed max_width", {
  ft <- flextable(head(mtcars))
  expect_warning(
    split_columns(ft, max_width = 0.01, rep_cols = names(mtcars)),
    "too small"
  )
})

test_that("split_columns respects unit conversion", {
  ft <- flextable(head(mtcars))
  p_in <- split_columns(ft, max_width = 5, unit = "in")
  p_cm <- split_columns(ft, max_width = 5 * 2.54, unit = "cm")
  expect_equal(length(p_in), length(p_cm))
})

test_that("split_columns preserves header on every page", {
  ft <- flextable(head(mtcars))
  pages <- split_columns(ft, max_width = 5)
  for (p in pages) {
    expect_gt(nrow_part(p, "header"), 0)
  }
})

# --- split_rows ---

test_that("split_rows returns single page when table fits", {
  ft <- flextable(head(iris))
  pages <- split_rows(ft, max_height = 20)
  expect_length(pages, 1)
})

test_that("split_rows splits tall table", {
  ft <- flextable(iris)
  pages <- split_rows(ft, max_height = 3)
  expect_gt(length(pages), 1)
  for (p in pages) {
    expect_s3_class(p, "flextable")
  }
})

test_that("split_rows keeps group together", {
  ft <- flextable(iris)
  # groups start at rows 1, 51, 101
  pages <- split_rows(ft, max_height = 3, group = c(1L, 51L, 101L))
  # each page should only contain rows from contiguous groups
  total_rows <- sum(sapply(pages, nrow_part, part = "body"))
  expect_equal(total_rows, 150)
})

test_that("split_rows errors on unsorted group", {
  ft <- flextable(iris)
  expect_error(
    split_rows(ft, max_height = 3, group = c(51L, 1L)),
    "strictly increasing"
  )
})

test_that("split_rows errors on out-of-range group", {
  ft <- flextable(iris)
  expect_error(
    split_rows(ft, max_height = 3, group = c(0L, 50L)),
    "between 1 and"
  )
})

test_that("split_rows warns when max_height too small", {
  ft <- flextable(iris)
  ft <- add_footer_lines(ft, values = rep("footer line", 50))
  expect_warning(
    split_rows(ft, max_height = 0.01),
    "too small"
  )
})

test_that("split_rows respects unit conversion", {
  ft <- flextable(iris)
  p_in <- split_rows(ft, max_height = 3, unit = "in")
  p_cm <- split_rows(ft, max_height = 3 * 2.54, unit = "cm")
  expect_equal(length(p_in), length(p_cm))
})

test_that("split_rows repeats footer on every page", {
  ft <- flextable(iris)
  ft <- add_footer_lines(ft, values = "a footnote")
  pages <- split_rows(ft, max_height = 3)
  for (p in pages) {
    expect_equal(nrow_part(p, "footer"), 1)
  }
})

# --- split_to_pages ---

test_that("split_to_pages with max_height only equals split_rows", {
  ft <- flextable(iris)
  p_pages <- split_to_pages(ft, max_height = 3)
  p_rows <- split_rows(ft, max_height = 3)
  expect_equal(length(p_pages), length(p_rows))
})

test_that("split_to_pages with max_width only equals split_columns", {
  ft <- flextable(head(mtcars))
  p_pages <- split_to_pages(ft, max_width = 5)
  p_cols <- split_columns(ft, max_width = 5)
  expect_equal(length(p_pages), length(p_cols))
})

test_that("split_to_pages with both produces more pages", {
  ft <- flextable(iris)
  p_h <- split_to_pages(ft, max_height = 3)
  p_w <- split_to_pages(ft, max_width = 4)
  p_both <- split_to_pages(ft, max_width = 4, max_height = 3)
  expect_gte(length(p_both), length(p_h))
  expect_gte(length(p_both), length(p_w))
})

test_that("split_to_pages with no constraints returns single-element list", {
  ft <- flextable(head(iris))
  pages <- split_to_pages(ft)
  expect_true(is.list(pages))
  expect_length(pages, 1)
  expect_s3_class(pages[[1]], "flextable")
})

test_that("split_to_pages always returns a list", {
  ft <- flextable(head(iris))
  # even when table fits
  pages <- split_to_pages(ft, max_width = 20, max_height = 20)
  expect_true(is.list(pages))
  expect_length(pages, 1)
})

test_that("split_to_pages every element is a flextable", {
  ft <- flextable(iris)
  pages <- split_to_pages(ft, max_width = 4, max_height = 3)
  for (p in pages) {
    expect_s3_class(p, "flextable")
  }
})

#' @title Clear the displayed content of selected columns
#' @description
#' `void()` replaces the visible text of the selected columns
#' with an empty string. The columns themselves (and their
#' headers) remain in the table, but the cell values are no
#' longer displayed.
#'
#' This is useful when a column should stay in the layout
#' (e.g. to preserve its width or to keep its header label)
#' but its body values should be hidden, for instance
#' after using [compose()] to build a richer display in a
#' neighbouring column that already incorporates those values.
#'
#' The underlying dataset is not modified; only the displayed
#' content is affected. To remove a column entirely, use
#' the `col_keys` argument of [flextable()] instead.
#' @inheritParams args_x_j_part
#' @examples
#' ftab <- flextable(head(mtcars))
#' ftab <- void(ftab, ~ vs + am + gear + carb)
#' ftab
#' @export
#' @family cell_content_composition
void <- function(x, j = NULL, part = "body") {
  if (!inherits(x, "flextable")) {
    stop(sprintf("Function `%s` supports only flextable objects.", "void()"))
  }
  part <- match.arg(
    part,
    c("all", "body", "header", "footer"),
    several.ok = FALSE
  )

  if (part == "all") {
    for (p in c("header", "body", "footer")) {
      x <- void(x = x, j = j, part = p)
    }
    return(x)
  }

  if (nrow_part(x, part) < 1) {
    return(x)
  }

  j <- get_columns_id(x[[part]], j)
  j <- x$col_keys[j]
  newcontent <- as_chunkset_struct(
    l_paragraph = rep(
      as_paragraph(as_chunk(x = "", fp_text_default())),
      nrow_part(x, part)
    ),
    keys = j,
    i = seq_len(nrow_part(x, part))
  )
  x[[part]]$content <- set_chunkset_struct_element(
    x = x[[part]]$content,
    j = j,
    value = newcontent
  )

  x
}

#' @export
#' @title Add flextable at a bookmark location in document's header
#' @description Function is now defunct, use [officer::prop_section()] instead.
#' @param ... unused
#' @keywords internal
headers_flextable_at_bkm <- function(...) {
  .Defunct(new = "prop_section", package = "officer")
}

#' @export
#' @title Add flextable at a bookmark location in document's footer
#' @description Function is now defunct, use [officer::prop_section()] instead.
#' @param ... unused
#' @keywords internal
footers_flextable_at_bkm <- function(...) {
  .Defunct(new = "prop_section", package = "officer")
}


#' @export
#' @title Transform a flextable into a raster
#' @description Function is now defunct, use [gen_grob()] instead.
#' @param ... unused
#' @keywords internal
as_raster <- function(...) {
  .Defunct(new = "gen_grob", package = "flextable")
}


#' @export
#' @title Mini lollipop chart chunk wrapper
#' @description Function is now defunct, use [gg_chunk()] instead.
#' @keywords internal
lollipop <- function(...) {
  .Defunct(new = "gg_chunk", package = "flextable")
}


#' @export
#' @title Set Formatter by Types of Columns
#' @description Function is now defunct, use `colformat_` functions instead.
#' @keywords internal
set_formatter_type <- function(...) {
  .Defunct(new = "colformat_*", package = "flextable")
}

test_that("use_model_printer and use_df_printer works", {
  skip_if_not(rmarkdown::pandoc_available("1.12.3"))
  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy("rmd/use-printer.Rmd", rmd_file)
  outfile <- tempfile(fileext = ".html")

  options(show.signif.stars = TRUE)

  rmarkdown::render(
    rmd_file,
    output_file = outfile,
    output_format = "html_document",
    envir = new.env(),
    quiet = TRUE
  )

  doc <- read_html(outfile)
  table_nodes <- xml_find_all(doc, "body//*/table")
  testthat::expect_length(table_nodes, n = 3)

  # check last table that should be a summary table
  table_node_3 <- table_nodes[[3]]
  expect_length(xml_children(xml_child(table_node_3, search = "/thead")), 2)
  expect_length(xml_children(xml_child(table_node_3, search = "/tbody")), 10)

  tfoot_content <- xml_child(table_node_3, search = "/tfoot")
  expect_length(xml_children(tfoot_content), 1)
  expect_equal(xml_text(tfoot_content), "n: 153")

  # check first table and last column should have stars
  first_tr <- xml_find_first(doc, "//table/tbody/tr")
  expect_length(xml_children(first_tr), 6)
  expect_equal(xml_text(xml_child(first_tr, 6)), "***")

  # check second table has 5 columns
  first_tr <- xml_find_first(doc, "//*[@id='no-stars']//table/tbody/tr")
  expect_length(xml_children(first_tr), 5)
})

---
title: "Using flextable"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using flextable}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

The user manual is available here: 

https://ardata-fr.github.io/flextable-book/

Help pages of functions are available here: 

https://davidgohel.github.io/flextable/reference/index.html

---
title: "flextable example"
---

```{r setup, include=FALSE}
library(knitr)
library(flextable)
library(officer)
library(magrittr)
library(data.table)
opts_chunk$set(echo = FALSE, message = FALSE)
```

```{r}
col_palette <- c("#D73027", "#F46D43", "#FDAE61", "#FEE08B", 
  "#D9EF8B", "#A6D96A", "#66BD63", "#1A9850")
cor_matrix <- cor(mtcars)
mycut <- cut( cor_matrix, 
  breaks = c(-1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1), 
  include.lowest = TRUE, label = FALSE)
mycolors <- col_palette[mycut]
data <- data.frame(rowname = row.names(cor_matrix), stringsAsFactors = FALSE) %>%
  cbind(cor_matrix)
```

## An example 

```{r}
flextable(data) %>%
  bg(j = colnames(cor_matrix), bg = mycolors) %>%
  align(align = "center", part = "all") %>%
  compose(i = 1, j = 1, value = as_paragraph(""), part = "header") %>% 
  colformat_num(j = colnames(cor_matrix), digits = 2) %>% 
  autofit()
```

## Another example 

```{r}
ft <- flextable(head(iris))

# measure column widths but only for the body part
w_body <- dim_pretty(ft, part = "body")$widths
# measure column widths only for the header part and get the max
# as height value for rotated text
h_header <- max( dim_pretty(ft, part = "header")$widths )

ft <- rotate(ft, j = 1:4, rotation="btlr",part="header")
ft <- rotate(ft, j = 5, rotation="tbrl",part="header")

ft <- valign(ft, valign = "center", part = "header")
ft <- align(ft, align = "center", part = "all")

# Manage header height
ft <- height(ft, height = h_header * 1.1, part = "header")
# ... mainly because Word don't handle auto height with rotated headers
ft <- hrule(ft, i = 1, rule = "exact", part = "header")

ft <- set_caption(ft, caption = "iris dataset", autonum = run_autonum(bkm = "iris"))
ft
```


## flextable - demo as_grouped_data

```{r tab.cap='mean of carbon dioxide uptake in grass plants', tab.id='CO2-id', label='CO2-label'}
data_CO2 <- dcast(as.data.table(CO2), 
  Treatment + conc ~ Type, value.var = "uptake", fun.aggregate = mean)
data_CO2 <- as_grouped_data(x = data_CO2, groups = c("Treatment"))

as_flextable( data_CO2 ) %>% 
  bold(j = 1, i = ~ !is.na(Treatment), bold = TRUE, part = "body" ) %>% 
  bold(part = "header", bold = TRUE ) %>% 
  width(width = 1.5) %>% 
  compose(i = ~ !is.na(conc), j = "conc", 
          value = as_paragraph(
            as_chunk(conc, formatter = function(x) sprintf("%.0f", x))
          )
  ) %>% add_footer_lines("dataset CO2 has been used for this flextable") %>% 
  bg(bg = "#FFFFFF", part = "footer") %>% 
  set_header_labels(conc = "Concentration") %>% 
  width(width = c(1.5, 1, 1)) 
```

---
title: "R Markdown loop with flextable for HTML"
output: html_document
---

```{r setup, include=FALSE}
library(flextable)
library(htmltools)
library(knitr)

opts_chunk$set(echo = TRUE)

ft_list <- list()

ft1 <- qflextable( head( airquality ))
ft1 <- set_caption(ft1,"airquality table")
ft_list[[1]] <- ft1

ft2 <- qflextable( tail( mtcars ))
ft2 <- set_caption(ft2,"mtcars table")
ft_list[[2]] <- ft2
```

```{r echo = FALSE}
div(flextable_html_dependency())
```

```{r results='asis', echo = FALSE}
for(i in seq_along(ft_list)){
  flextable_to_rmd(ft_list[[i]])
}
```

---
title: "flextable caption example"
---

```{r include=FALSE}
library(knitr)
library(flextable)
library(magrittr)
library(data.table)
opts_chunk$set(echo = FALSE)

data_CO2 <- dcast(as.data.table(CO2), 
  Treatment + conc ~ Type, value.var = "uptake", fun.aggregate = mean)
data_CO2 <- as_grouped_data(x = data_CO2, groups = c("Treatment"))
```

```{r}
ft <- as_flextable( data_CO2 ) %>% 
  bold(j = 1, i = ~ !is.na(Treatment), bold = TRUE, part = "body" ) %>% 
  bold(part = "header", bold = TRUE ) %>% 
  width(width = 1.5) %>% 
  compose(i = ~ !is.na(conc), j = "conc", 
          value = as_paragraph(
            as_chunk(conc, formatter = function(x) sprintf("%.0f", x))
          )
  ) %>% add_footer_lines("dataset CO2 has been used for this flextable") %>% 
  bg(bg = "#FFFFFF", part = "footer") %>% 
  set_header_labels(conc = "Concentration") %>% 
  width(width = c(1.5, 1, 1)) 
```

```{r}
ft <- as_flextable( data_CO2 ) %>% 
  bold(j = 1, i = ~ !is.na(Treatment), bold = TRUE, part = "body" ) %>% 
  bold(part = "header", bold = TRUE ) %>% 
  width(width = 1.5) %>% 
  compose(i = ~ !is.na(conc), j = "conc", 
          value = as_paragraph(
            as_chunk(conc, formatter = function(x) sprintf("%.0f", x))
          )
  ) %>% add_footer_lines("dataset CO2 has been used for this flextable") %>% 
  bg(bg = "#FFFFFF", part = "footer") %>% 
  set_header_labels(conc = "Concentration") %>% 
  width(width = c(1.5, 1, 1)) 
```

## Captions

Now, let's add a caption with `set_caption`:

```{r ft.align="left"}
library(officer)
set_caption(ft, 
  caption = "mean of carbon dioxide uptake in grass plants", 
  style = "Table Caption", 
  autonum = run_autonum(seq_id = "tab", bkm = "tab1"))
```

Or use kintr chunk options:

```{r ft.align="center", tab.cap='mean of carbon dioxide uptake in grass plants', tab.id='tab2', label='tab2'}
ft
```

## Cross-references

* This is a reference to flextable \@ref(tab:tab1).
* This is a reference to flextable \@ref(tab:tab2).
* This is a link to [flextable](#tab:tab1)

---
title: "bookdown examples"
---

```{r setup, include=FALSE}
set.seed(2)

knitr::opts_chunk$set(echo = FALSE, ft.shadow = FALSE, ft.latex.float = 'float')
library(officer)
library(flextable)

set_flextable_defaults(
  font.family = "Arial", 
  fonts_ignore = TRUE,
  font.size = 10, 
  font.color = "#222222", border.color = "#222222")

str1 <- "Lorem ipsum dolor sit amet, ac id condimentum cras urna velit, "
str2 <- "fringilla nec nostra in. "
str3 <- "Iaculis sit sed in quam cubilia orci dui eget maximus ullamcorper."

fpt1 <- fp_text_default(color = "orange")
fpt2 <- fp_text_default(color = "#068282", italic = TRUE)

fpp1 <- fp_par(
  padding.left = 15, 
  padding.right = 25, 
  padding.top = 5, 
  padding.bottom = 10, 
  text.align = "left")

ft <- qflextable(airquality[sample.int(150, size = 5), ]) |>
  set_caption(
    fp_p = fpp1,
    align_with_table = FALSE,
    autonum = run_autonum(seq_id = "tab", bkm = "tbl-flextable", bkm_all = TRUE, prop = fpt1),
    caption = as_paragraph(
      as_chunk(str1, props = fpt1), str2, as_chunk(str3, props = fpt2)
    )
  )
```

# some examples

## example 1

```{r}
#| tab.topcaption: false
ft
```

## example 2

```{r}
#| tab.id: tbl-coco
#| tab.cap: Caption zzzz
#| tab.topcaption: true
flextable(cars[1:4, ])
```

## example 3

This is a reference to table: `r run_reference("tbl-flextable")`.

---
title: 'Example keep with next'
---

```{r include=FALSE}
library(knitr)
library(officedown)
library(flextable)

opts_chunk$set(
  echo = FALSE,
  ft.keepnext = FALSE
)
table_caption <- 'airquality dataset'
```


```{r table1, tab.cap = table_caption}
qflextable(mtcars)
```


```{r table2, tab.cap = table_caption}
qflextable(mtcars)
```

blop

blop

blop

blop

blop

blop

blop

```{r include=FALSE}
opts_chunk$set(ft.keepnext = TRUE)
```


```{r table3, tab.cap = table_caption}
qflextable(head(mtcars))
```

---
title: "rmarkdown examples"
---

```{r setup, include=FALSE}
set.seed(2)

knitr::opts_chunk$set(echo = FALSE, ft.shadow = FALSE, ft.latex.float = 'float')
library(officer)
library(flextable)

set_flextable_defaults(
  font.family = "Arial", 
  fonts_ignore = TRUE,
  font.size = 10, 
  font.color = "#222222", border.color = "#222222")

str1 <- "Lorem ipsum dolor sit amet, ac id condimentum cras urna velit, "
str2 <- "fringilla nec nostra in. "
str3 <- "Iaculis sit sed in quam cubilia orci dui eget maximus ullamcorper."

fpt1 <- fp_text_default(color = "orange")
fpt2 <- fp_text_default(color = "#068282", italic = TRUE)

fpp1 <- fp_par(
  padding.left = 15, 
  padding.right = 25, 
  padding.top = 5, 
  padding.bottom = 10, 
  text.align = "left")

ft <- qflextable(airquality[sample.int(150, size = 5), ]) |>
  set_caption(
    fp_p = fpp1,
    align_with_table = FALSE,
    autonum = run_autonum(seq_id = "tab", bkm = "tbl-flextable", bkm_all = TRUE, prop = fpt1),
    caption = as_paragraph(
      as_chunk(str1, props = fpt1), str2, as_chunk(str3, props = fpt2)
    )
  )
```

# some examples

## example 1

```{r}
#| tab.topcaption: false
ft
```

## example 2

```{r}
#| tab.id: tbl-coco
#| tab.cap: Caption zzzz
#| tab.topcaption: true
flextable(cars[1:4, ])
```

---
title: "use printers"
---

```{r setup, include=FALSE}
library(knitr)
opts_chunk$set(echo = FALSE)
library(flextable)
use_model_printer()
use_df_printer()
```

# a model

```{r}
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100),
    lot1 = c(118,58,42,35,27,25,21,19,18),
    lot2 = c(69,35,26,21,18,16,13,12,12))
model <- glm(lot1 ~ log(u), data = clotting, family = Gamma)
model
```

# no stars

```{r}
options(show.signif.stars = FALSE)
model
```

## example 2

```{r}
airquality
```

---
title: none
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE)
library(flextable)
library(magrittr)
library(officer)
caption_md <- "This is a caption"

knitr::opts_chunk$set(ft.shadow = FALSE,
  tab.lp = "tab:", 
  tab.cap.style = "Table Caption")

ft <- flextable(head(cars)) %>% 
  theme_vanilla() %>% 
  autofit()
```


Cross-reference is there: \@ref(tab:id1)

## A caption, no id

```{r tab.cap=caption_md}
ft
```


## A caption and an id

```{r tab.cap=caption_md, label="id1"}
ft
```

## No caption and an id

```{r tab.id="tab3"}
ft
```

## Using set caption

```{r table2}
#| tab.topcaption: true
set_caption(
  x = ft, 
  caption = as_paragraph(
      as_chunk("azerty ", props = fp_text_default(color = "cyan")),
      as_chunk("querty", props = fp_text_default(color = "orange"))
  ),
  autonum = run_autonum(seq_id = "tab", bkm = "id2"),
  fp_p = fp_par(
    padding = 10, 
    border = fp_border_default(color = "red", width = 1))
)
```

## Using set caption and chunk options

```{r label="id3"}
#| tab.topcaption: false
set_caption(
  x = ft, 
  caption = as_paragraph(
      as_chunk("azerty ", props = fp_text_default(color = "cyan")),
      as_chunk("querty", props = fp_text_default(color = "orange"))
  )
)
```

---
title: 'borders examples'
---

```{r include=FALSE}
library(knitr)
library(officedown)
library(flextable)

opts_chunk$set(
  echo = FALSE,
  ft.tabcolsep = 3
)
```

## Blah 

```{r}
dat <- tibble::tribble(
    ~S,   ~T,   ~U,   ~V,   ~W,   ~X,   ~Y,   ~Z,
  "AA", "AA", "AA", "AA", "AA", "AA", "AA", "AA",
  "BB", "AA", "AA", "AA", "AA", "AA", "AA", "AA",
  "CC", "CC", "AA", "AA", "AA", "AA", "AA", "AA",
  "DD", "DD", "DD", "AA", "AA", "AA", "AA", "AA",
  "EE", "EE", "EE", "EE", "AA", "AA", "AA", "AA",
  "FF", "FF", "FF", "FF", "FF", "AA", "AA", "AA",
  "GG", "GG", "GG", "GG", "GG", "GG", "AA", "AA",
  "HH", "HH", "HH", "HH", "HH", "HH", "HH", "AA"
  )
ft <- flextable(dat)
ft <- theme_vader(ft)
ft <- border_outer(ft, border = fp_border_default(width = 2.5, color = "red"))
ft <- border_inner_v(ft, border = fp_border_default(width = 1, color = "cyan"))
ft <- border_inner_h(ft, border = fp_border_default(width = 1, color = "orange"))
ft <- merge_h(ft)
ft <- fix_border_issues(ft)
ft
```


## Blah

```{r}
ft <- flextable(head(iris, n = 2))
ft <- border_inner_v(ft, part = "all")
ft <- hline(ft,
    i = 1, j = 2:4, part = "body",
    border = officer::fp_border("red")
  )
ft <- hline(ft, 
    i = 1, j = 1:1, part = "body",
    border = officer::fp_border("orange")
  )
ft <- autofit(ft)
ft
```

\pagebreak

## Blah

```{r}
dat <- data.frame(
  s = c("a", "a", "a", "a", "b", "b", "b", "b"),
  t = c("AA", "AA", "AA", "BB", "BB", "BB", "CC", "CC"),
  u = c("DD", "DD", "DD", "DD", "DD", "DD", "DD", "DD"),
  v = c("DD", "EE", "EE", "DD", "EE", "EE", "EE", "DD"),
  w = c("DD", "EE", "EE", "DD", "EE", "EE", "EE", "DD"),
  x = c("DD", "DD", "DD", "DD", "DD", "DD", "DD", "DD")
)
ft <- flextable(dat)
ft <- border_outer(ft, border = fp_border_default(width = 2.5, color = "red"))
ft <- border_inner_v(ft, border = fp_border_default(width = 1, color = "cyan"))
ft <- border_inner_h(ft, border = fp_border_default(width = 1, color = "orange"))
ft <- merge_v(ft, j = 1:2)
ft <- merge_h(ft)
ft <- fix_border_issues(ft)
ft
```

## Blah

```{r}
ft <- flextable(dat)
ft <- theme_vader(ft)
ft
```

\pagebreak

## Blah

```{r}
ft <- border_outer(ft, border = fp_border_default(width = 2.5, color = "red"))
ft <- border_inner_v(ft, border = fp_border_default(width = 1, color = "cyan"))
ft <- border_inner_h(ft, border = fp_border_default(width = 1, color = "orange"))
ft <- merge_v(ft, j = 1:2)
ft <- merge_h(ft)
ft <- fix_border_issues(ft)
ft
```


## Blah

```{r}
dt <- data.frame(
  stringsAsFactors = FALSE,
  V1 = c("OOOOOOOO", "OOOOOOOO", "OOOOOOOO"),
  V2 = c("XXX", "XXX", "YYY"),
  V3 = c("ZZZ", "ZZZ", "ZZZ"),
  V4 = c("XXXX XXXX XXXX XXXX",
         "XXXX XXXX XXXX XXXX",
         "XXXX XXXX XXXX XXXX")
)

dt |>
  flextable() |>
  set_table_properties(
    layout = "autofit", width = .8,
    opts_pdf = list(tabcolsep = 3)
  ) |>
  merge_v(j = ~ V1 + V2) |>
  valign(valign = "top", part = "all") |>
  border_inner_h(part = "all") |>
  border_inner_v(part = "all") |>
  border_outer(part = "all", border = fp_border_default(color = "red")) |>
  autofit() |>
  fix_border_issues(part = "all")
```
## blouh

```{r echo=FALSE}
ft <- as_flextable(cars)
ft <- theme_zebra(ft)
ft <- hline(ft, j = 2, part = "body")
ft
```


## kable

```{r}
kable(head(iris))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.pam}
\alias{as_flextable.pam}
\title{Transform a 'pam' object into a flextable}
\usage{
\method{as_flextable}{pam}(x, digits = 4, ...)
}
\arguments{
\item{x}{a \code{\link[cluster:pam]{cluster::pam()}} object}

\item{digits}{number of digits for the numeric columns}

\item{...}{unused argument}
}
\description{
produce a flextable describing a
pam object. The function is only using package 'broom'
that provides the data presented in the resulting flextable.
}
\examples{
if (require("cluster")) {
  dat <- as.data.frame(scale(mtcars[1:7]))
  cl <- pam(dat, 3)
  ft <- as_flextable(cl)
  ft
}
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nrowcol.R
\name{nrow_part}
\alias{nrow_part}
\title{Number of rows of a part}
\usage{
nrow_part(x, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
returns the number of lines in a part of flextable.
}
\examples{
library(flextable)
ft <- qflextable(head(cars))
nrow_part(ft, part = "body")
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_n_percent}
\alias{fmt_n_percent}
\title{format content for count data}
\usage{
fmt_n_percent(n, pct, digit = 1)
}
\arguments{
\item{n}{count values}

\item{pct}{percent values}

\item{digit}{number of digits for the percentages}
}
\description{
The function formats counts and
percentages as \verb{n (xx.x\%)}. If percentages are
missing, they are not printed.
}
\examples{
library(flextable)

df <- structure(
  list(
    cut = structure(
      .Data = 1:5, levels = c(
        "Fair", "Good", "Very Good", "Premium", "Ideal"),
      class = c("ordered", "factor")),
    n = c(1610L, 4906L, 12082L, 13791L, 21551L),
    pct = c(0.0299, 0.0909, 0.2239, 0.2557, 0.3995)
  ),
  row.names = c(NA, -5L),
  class = "data.frame")

ft_1 <- flextable(df, col_keys = c("cut", "txt"))
ft_1 <- mk_par(
  x = ft_1, j = "txt",
  value = as_paragraph(fmt_n_percent(n, pct)))
ft_1 <- align(ft_1, j = "txt", part = "all", align = "right")
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other text formatter functions: 
\code{\link{fmt_2stats}()},
\code{\link{fmt_avg_dev}()},
\code{\link{fmt_header_n}()}
}
\concept{text formatter functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{save_as_pptx}
\alias{save_as_pptx}
\title{Save flextable objects in a 'PowerPoint' file}
\usage{
save_as_pptx(..., values = NULL, path)
}
\arguments{
\item{...}{flextable objects, objects, possibly named. If named objects, names are
used as slide titles.}

\item{values}{a list (possibly named), each element is a flextable object. If named objects, names are
used as slide titles. If provided, argument \code{...} will be ignored.}

\item{path}{PowerPoint file to be created}
}
\value{
a string containing the full name of the generated file
}
\description{
sugar function to save flextable objects in
an PowerPoint file.

This feature is available to simplify the work of users by avoiding
the need to use the 'officer' package. If it doesn't suit your needs,
then use the API offered by 'officer' which allows simple and
complicated things.
}
\note{
The PowerPoint format ignores captions (see \code{\link[=set_caption]{set_caption()}}).
}
\examples{
ft1 <- flextable(head(iris))
tf <- tempfile(fileext = ".pptx")
save_as_pptx(ft1, path = tf)

ft2 <- flextable(head(mtcars))
tf <- tempfile(fileext = ".pptx")
save_as_pptx(`iris table` = ft1, `mtcars table` = ft2, path = tf)
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{line_spacing}
\alias{line_spacing}
\title{Set line spacing}
\usage{
line_spacing(x, i = NULL, j = NULL, space = 1, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{space}{space between lines of text; 1 is single line spacing, 2 is double line spacing.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Change the line spacing of selected rows and columns of a flextable.
}
\examples{
ft <- flextable(head(mtcars)[, 3:6])
ft <- line_spacing(ft, space = 1.6, part = "all")
ft <- set_table_properties(ft, layout = "autofit")
ft
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compact_summary.R
\name{as_flextable.compact_summary}
\alias{as_flextable.compact_summary}
\title{Transform a 'compact_summary' object into a flextable}
\usage{
\method{as_flextable}{compact_summary}(x, ...)
}
\arguments{
\item{x}{A \code{compact_summary} object produced by
\code{\link[=compact_summary]{compact_summary()}}.}

\item{...}{unused arguments.}
}
\value{
A \code{\link[=flextable]{flextable()}} object.
}
\description{
\code{compact_summary} objects can be transformed into
a flextable with method \code{\link[=as_flextable]{as_flextable()}}.

Numeric columns are formatted with \code{\link[=formatC]{formatC()}} using the
\code{digits} value stored in the object and the current
flextable defaults for \code{big.mark} and \code{decimal.mark}.
}
\examples{
z <- compact_summary(iris, show_type = TRUE, show_na = TRUE)
as_flextable(z)
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{valign}
\alias{valign}
\title{Set vertical alignment}
\usage{
valign(x, i = NULL, j = NULL, valign = "center", part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{valign}{vertical alignment of paragraph within cell,
one of "center" or "top" or "bottom".}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Change the vertical alignment of selected rows and columns of a flextable.
}
\examples{
ft_1 <- flextable(iris[c(1:3, 51:53, 101:103), ])
ft_1 <- theme_box(ft_1)
ft_1 <- merge_v(ft_1, j = 5)
ft_1

ft_2 <- valign(ft_1, j = 5, valign = "top", part = "all")
ft_2
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable_tabulator.R
\name{tabulator_colnames}
\alias{tabulator_colnames}
\title{Column keys of tabulator objects}
\usage{
tabulator_colnames(x, columns, ..., type = NULL)
}
\arguments{
\item{x}{a \code{\link[=tabulator]{tabulator()}} object}

\item{columns}{column names to look for}

\item{...}{any filter conditions that use variables
names, the same than the argument \code{columns} of function \code{\link[=tabulator]{tabulator()}}
(\code{tabulator(columns = c("col1", "col2"))}).}

\item{type}{the type of column to look for, it can be:
\itemize{
\item 'columns': visible columns, corresponding to names provided
in the '...' arguments of your call to 'tabulator()'.
\item 'hidden': unvisible columns, corresponding to names of
the original dataset columns.
\item 'rows': visible columns used as 'row' content
\item 'rows_supp': visible columns used as 'rows_supp' content
\item NULL: any type of column
}}
}
\description{
The function provides a way to get column keys
associated with the flextable corresponding to a \code{\link[=tabulator]{tabulator()}}
object. It helps in customizing or programing with \code{tabulator}.

The function is using column names from the original
dataset, eventually filters and returns the names
corresponding to the selection.
}
\examples{
library(flextable)

cancer_dat <- data.frame(
  count = c(
    9L, 5L, 1L, 2L, 2L, 1L, 9L, 3L, 1L, 10L, 2L, 1L, 1L, 2L, 0L, 3L,
    2L, 1L, 1L, 2L, 0L, 12L, 4L, 1L, 7L, 3L, 1L, 5L, 5L, 3L, 10L,
    4L, 1L, 4L, 2L, 0L, 3L, 1L, 0L, 4L, 4L, 2L, 42L, 28L, 19L, 26L,
    19L, 11L, 12L, 10L, 7L, 10L, 5L, 6L, 5L, 0L, 3L, 4L, 3L, 3L,
    1L, 2L, 3L
  ),
  risktime = c(
    157L, 77L, 21L, 139L, 68L, 17L, 126L, 63L, 14L, 102L, 55L,
    12L, 88L, 50L, 10L, 82L, 45L, 8L, 76L, 42L, 6L, 134L, 71L,
    22L, 110L, 63L, 18L, 96L, 58L, 14L, 86L, 42L, 10L, 66L,
    35L, 8L, 59L, 32L, 8L, 51L, 28L, 6L, 212L, 130L, 101L,
    136L, 72L, 63L, 90L, 42L, 43L, 64L, 21L, 32L, 47L, 14L,
    21L, 39L, 13L, 14L, 29L, 7L, 10L
  ),
  time = rep(as.character(1:7), 3),
  histology = rep(as.character(1:3), 21),
  stage = rep(as.character(1:3), each = 21)
)

datasup_first <- data.frame(
  time = factor(1:7, levels = 1:7),
  zzz = runif(7)
)

z <- tabulator(cancer_dat,
  rows = "time",
  columns = c("histology", "stage"),
  datasup_first = datasup_first,
  n = as_paragraph(as_chunk(count))
)

j <- tabulator_colnames(
  x = z, type = "columns",
  columns = c("n"),
  stage \%in\% 1
)

src <- tabulator_colnames(
  x = z, type = "hidden",
  columns = c("count"),
  stage \%in\% 1
)

if (require("scales")) {
  colourer <- col_numeric(
    palette = c("wheat", "red"),
    domain = c(0, 45)
  )
  ft_1 <- as_flextable(z)
  ft_1 <- bg(
    ft_1,
    bg = colourer, part = "body",
    j = j, source = src
  )
  ft_1
}
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xtable_to_flextable.R
\name{as_flextable.xtable}
\alias{as_flextable.xtable}
\title{get a flextable from a xtable object}
\usage{
\method{as_flextable}{xtable}(
  x,
  text.properties = fp_text_default(),
  format.args = getOption("xtable.format.args", NULL),
  rowname_col = "rowname",
  hline.after = getOption("xtable.hline.after", c(-1, 0, nrow(x))),
  NA.string = getOption("xtable.NA.string", ""),
  include.rownames = TRUE,
  rotate.colnames = getOption("xtable.rotate.colnames", FALSE),
  ...
)
}
\arguments{
\item{x}{\code{xtable} object}

\item{text.properties}{default text formatting properties}

\item{format.args}{List of arguments for the formatC function.
See argument \code{format.args} of \code{print.xtable}. Not yet
implemented.}

\item{rowname_col}{colname used for row names column}

\item{hline.after}{see \code{?print.xtable}.}

\item{NA.string}{see \code{?print.xtable}.}

\item{include.rownames}{see \code{?print.xtable}.}

\item{rotate.colnames}{see \code{?print.xtable}.}

\item{...}{unused arguments}
}
\description{
Get a \code{flextable} object from
a \code{xtable} object.
}
\section{Illustrations}{


\if{html}{\figure{fig_as_flextable.xtable_1.png}{options: width="300"}}

\if{html}{\figure{fig_as_flextable.xtable_2.png}{options: width="300"}}

\if{html}{\figure{fig_as_flextable.xtable_3.png}{options: width="500"}}
}

\examples{
library(officer)
if( require("xtable") ){

  data(tli)
  tli.table <- xtable(tli[1:10, ])
  align(tli.table) <- rep("r", 6)
  align(tli.table) <- "|r|r|clr|r|"
  ft_1 <- as_flextable(
    tli.table,
    rotate.colnames = TRUE,
    include.rownames = FALSE)
  ft_1 <- height(ft_1, i = 1, part = "header", height = 1)
  ft_1

  \donttest{
  Grade3 <- c("A","B","B","A","B","C","C","D","A","B",
    "C","C","C","D","B","B","D","C","C","D")
  Grade6 <- c("A","A","A","B","B","B","B","B","C","C",
    "A","C","C","C","D","D","D","D","D","D")
  Cohort <- table(Grade3, Grade6)
  ft_2 <- as_flextable(xtable(Cohort))
  ft_2 <- set_header_labels(ft_2, rowname = "Grade 3")
  ft_2 <- autofit(ft_2)
  ft_2 <- add_header(ft_2, A = "Grade 6")
  ft_2 <- merge_at(ft_2, i = 1, j = seq_len( ncol(Cohort) ) + 1,
    part = "header" )
  ft_2 <- bold(ft_2, j = 1, bold = TRUE, part = "body")
  ft_2 <- height_all(ft_2, part = "header", height = .4)
  ft_2

  temp.ts <- ts(cumsum(1 + round(rnorm(100), 0)),
    start = c(1954, 7), frequency = 12)
  ft_3 <- as_flextable(x = xtable(temp.ts, digits = 0),
    NA.string = "-")
  ft_3
  }
  detach("package:xtable", unload = TRUE)
}
}
\seealso{
Other as_flextable methods: 
\code{\link{as_flextable.data.frame}()},
\code{\link{as_flextable.glm}()},
\code{\link{as_flextable.grouped_data}()},
\code{\link{as_flextable.htest}()},
\code{\link{as_flextable.kmeans}()},
\code{\link{as_flextable.lm}()},
\code{\link{as_flextable.merMod}()},
\code{\link{as_flextable.pam}()},
\code{\link{as_flextable.summarizor}()},
\code{\link{as_flextable.tabulator}()},
\code{\link{as_flextable}()}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{fit_columns}
\alias{fit_columns}
\title{Constrain table width by wrapping text}
\usage{
fit_columns(x, max_width, no_wrap = NULL, unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{max_width}{maximum total table width (in inches by default)}

\item{no_wrap}{column names or indices that must keep their
optimal width (no compression, no text wrapping).}

\item{unit}{unit for \code{max_width}, one of "in", "cm", "mm".}
}
\description{
Shrink column widths so that the total table width
does not exceed \code{max_width}. Font sizes are unchanged; text wraps
inside the narrower cells.

Columns that cannot shrink below their longest word (plus padding)
are clamped at that floor; remaining space is iteratively
redistributed among unclamped columns.

Columns listed in \code{no_wrap} keep their optimal width (as computed
by \code{\link[=dim_pretty]{dim_pretty()}}) and are not compressed.

To constrain width by reducing font size instead, see
\code{\link[=fit_to_width]{fit_to_width()}}. To size columns to their content without a
width constraint, see \code{\link[=autofit]{autofit()}}.
}
\examples{
ft <- qflextable(head(mtcars))
ft <- fit_columns(ft, max_width = 5)
flextable_dim(ft)$widths
ft
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{fit_to_width}
\alias{fit_to_width}
\title{Constrain table width by shrinking font size}
\usage{
fit_to_width(x, max_width, inc = 1L, max_iter = 20, unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{max_width}{maximum width to fit in inches}

\item{inc}{the font size decrease for each step}

\item{max_iter}{maximum iterations}

\item{unit}{unit for max_width, one of "in", "cm", "mm".}
}
\description{
Decrease font size incrementally until the table fits
within \code{max_width}.

To constrain width by wrapping text instead, see \code{\link[=fit_columns]{fit_columns()}}.
To size columns to their content without a width constraint,
see \code{\link[=autofit]{autofit()}}.
}
\examples{
ft_1 <- qflextable(head(mtcars))
ft_1 <- width(ft_1, width = 1)
ft_1

ft_2 <- fit_to_width(ft_1, max_width = 4)
ft_2
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{colformat_lgl}
\alias{colformat_lgl}
\title{format logical cells}
\usage{
colformat_lgl(
  x,
  i = NULL,
  j = NULL,
  true = "true",
  false = "false",
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
)
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{columns selection.}

\item{false, true}{string to be used for logical}

\item{na_str, nan_str}{string to be used for NA and NaN values}

\item{prefix, suffix}{string to be used as prefix or suffix}
}
\description{
Format logical cells in a flextable.
}
\examples{
dat <- data.frame(a = c(TRUE, FALSE), b = c(FALSE, TRUE))

z <- flextable(dat)
z <- colformat_lgl(x = z, j = c("a", "b"))
autofit(z)
}
\seealso{
Other cells formatters: 
\code{\link{colformat_char}()},
\code{\link{colformat_datetime}()},
\code{\link{colformat_date}()},
\code{\link{colformat_double}()},
\code{\link{colformat_image}()},
\code{\link{colformat_int}()},
\code{\link{colformat_num}()},
\code{\link{set_formatter}()}
}
\concept{cells formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{as_flextable.summarizor}
\alias{as_flextable.summarizor}
\title{Transform a 'summarizor' object into a flextable}
\usage{
\method{as_flextable}{summarizor}(x, ...)
}
\arguments{
\item{x}{result from \code{\link[=summarizor]{summarizor()}}}

\item{...}{arguments for \code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}}}
}
\description{
\code{summarizor} object should be transformed into a flextable
with method \code{\link[=as_flextable]{as_flextable()}}.
}
\examples{
\dontshow{
data.table::setDTthreads(1)
}
z <- summarizor(CO2[-c(1, 4)],
  by = "Treatment",
  overall_label = "Overall"
)
ft_1 <- as_flextable(z, spread_first_col = TRUE)
ft_1 <- prepend_chunks(ft_1,
  i = ~ is.na(variable), j = 1,
  as_chunk("\t")
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_pct}
\alias{fmt_pct}
\title{Format numbers as percentages}
\usage{
fmt_pct(x)
}
\arguments{
\item{x}{numeric values}
}
\description{
Format numeric values as percentages (e.g. \code{"45.0\%"}).
}
\examples{
library(flextable)

df <- data.frame(zz = .45)

ft_1 <- flextable(df)
ft_1 <- mk_par(
  x = ft_1, j = 1, part = "body",
  value = as_paragraph(as_chunk(zz, formatter = fmt_pct))
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other value formatters:
\code{\link[=fmt_2stats]{fmt_2stats()}},
\code{\link[=fmt_avg_dev]{fmt_avg_dev()}},
\code{\link[=fmt_dbl]{fmt_dbl()}},
\code{\link[=fmt_header_n]{fmt_header_n()}},
\code{\link[=fmt_int]{fmt_int()}},
\code{\link[=fmt_n_percent]{fmt_n_percent()}},
\code{\link[=fmt_signif_after_zeros]{fmt_signif_after_zeros()}}
}
\concept{value formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_flextable.R
\name{merge_v}
\alias{merge_v}
\title{Merge flextable cells vertically}
\usage{
merge_v(x, j = NULL, target = NULL, part = "body", combine = FALSE)
}
\arguments{
\item{x}{\code{flextable} object}

\item{j}{column to used to find consecutive values to be merged. Columns
from orignal dataset can also be used.}

\item{target}{columns names where cells have to be merged.}

\item{part}{partname of the table where merge has to be done.}

\item{combine}{If the value is TRUE, the columns defined by \code{j} will
be combined into a single column/value and the consecutive values of
this result will be used. Otherwise, the columns are inspected one
by one to perform cell merges.}
}
\description{
Merge flextable cells vertically when consecutive cells have
identical values. Text of formatted values are used to compare
values if available.

Two options are available, either a column-by-column algorithm or an
algorithm where the combinations of these columns are used once for all
target columns.
}
\section{Illustrations}{


\if{html}{\figure{fig_merge_v_1.png}{options: width="600"}}

\if{html}{\figure{fig_merge_v_2.png}{options: width="600"}}
}

\examples{
ft_merge <- flextable(mtcars)
ft_merge <- merge_v(ft_merge, j = c("gear", "carb"))
ft_merge

data_ex <- structure(list(srdr_id = c(
  "175124", "175124", "172525", "172525",
  "172545", "172545", "172609", "172609", "172609"
), substances = c(
  "alcohol",
  "alcohol", "alcohol", "alcohol", "cannabis",
  "cannabis", "alcohol\n cannabis\n other drugs",
  "alcohol\n cannabis\n other drugs",
  "alcohol\n cannabis\n other drugs"
), full_name = c(
  "TAU", "MI", "TAU", "MI (parent)", "TAU", "MI",
  "TAU", "MI", "MI"
), article_arm_name = c(
  "Control", "WISEteens",
  "Treatment as usual", "Brief MI (b-MI)", "Assessed control",
  "Intervention", "Control", "Computer BI", "Therapist BI"
)), row.names = c(
  NA,
  -9L
), class = c("tbl_df", "tbl", "data.frame"))


ft_1 <- flextable(data_ex)
ft_1 <- theme_box(ft_1)
ft_2 <- merge_v(ft_1, j = "srdr_id",
  target = c("srdr_id", "substances"))
  ft_2
}
\seealso{
Other flextable merging function: 
\code{\link{merge_at}()},
\code{\link{merge_h_range}()},
\code{\link{merge_h}()},
\code{\link{merge_none}()}
}
\concept{flextable merging function}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{summarizor}
\alias{summarizor}
\title{Prepare descriptive statistics for flextable}
\usage{
summarizor(
  x,
  by = character(),
  overall_label = NULL,
  num_stats = c("mean_sd", "median_iqr", "range"),
  hide_null_na = TRUE,
  use_labels = TRUE
)
}
\arguments{
\item{x}{dataset}

\item{by}{columns names to be used as grouping columns}

\item{overall_label}{label to use as overall label}

\item{num_stats}{available statistics for numerical columns
to show, available options are "mean_sd", "median_iqr" and "range".}

\item{hide_null_na}{if TRUE (default), NA counts will not be shown when 0.}

\item{use_labels}{Logical; if TRUE, any column labels or value labels
present in the dataset will be used for display purposes. Defaults
to TRUE.}
}
\description{
It performs a univariate statistical analysis of a dataset
by group and formats the results so that they can be used with
the \code{\link[=tabulator]{tabulator()}} function or directly with \link[=as_flextable.summarizor]{as_flextable}.
\if{html}{\out{
<img src="https://www.ardata.fr/img/flextable-imgs/flextable-018-square.png" alt="summarizor illustration" style="width:100\%;">
}}
}
\note{
This is very first version of the function; be aware it
can evolve or change.
}
\examples{
\dontshow{
data.table::setDTthreads(1)
}
z <- summarizor(CO2[-c(1, 4)],
  by = "Treatment",
  overall_label = "Overall"
)
ft_1 <- as_flextable(z)
ft_1

ft_2 <- as_flextable(z, sep_w = 0, spread_first_col = TRUE)
ft_2

z <- summarizor(CO2[-c(1, 4)])
ft_3 <- as_flextable(z, sep_w = 0, spread_first_col = TRUE)
ft_3
}
\seealso{
\code{\link[=fmt_summarizor]{fmt_summarizor()}}, \code{\link[=labelizor]{labelizor()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{gg_chunk}
\alias{gg_chunk}
\title{ggplot chunk}
\usage{
gg_chunk(value, width = 1, height = 0.2, unit = "in", res = 300, alt = "")
}
\arguments{
\item{value}{gg objects, stored in a list column; or a list
of 'ggplot' objects.}

\item{width, height}{size of the resulting png file.}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{res}{resolution of the png image in ppi}

\item{alt}{alternative text for the image (used for accessibility)}
}
\description{
This function is used to insert mini gg plots into
flextable with functions:
\itemize{
\item \code{\link[=compose]{compose()}} and \code{\link[=as_paragraph]{as_paragraph()}},
\item \code{\link[=append_chunks]{append_chunks()}},
\item \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format. With Quarto (\code{format: docx}) or
\code{rmarkdown::word_document()}, the resulting file must be repaired
with \code{\link[=repair_docx]{repair_docx()}}.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\examples{
library(data.table)
library(flextable)
if (require("ggplot2")) {
  my_cor_plot <- function(x) {
    cols <- colnames(x)[sapply(x, is.numeric)]
    x <- x[, .SD, .SDcols = cols]
    cormat <- as.data.table(cor(x))
    cormat$var1 <- colnames(cormat)
    cormat <- melt(cormat,
      id.vars = "var1", measure.vars = cormat$var1,
      variable.name = "var2", value.name = "correlation"
    )
    ggplot(data = cormat, aes(x = var1, y = var2, fill = correlation)) +
      geom_tile() +
      coord_equal() +
      scale_fill_gradient2(
        low = "blue",
        mid = "white", high = "red", limits = c(-1, 1),
        guide = "none"
      ) +
      theme_void()
  }
  z <- as.data.table(iris)
  z <- z[, list(gg = list(my_cor_plot(.SD))), by = "Species"]
  ft <- flextable(z)
  ft <- mk_par(ft,
    j = "gg",
    value = as_paragraph(
      gg_chunk(value = gg, width = 1, height = 1)
    )
  )
  ft
}
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{get_flextable_defaults}
\alias{get_flextable_defaults}
\title{Get flextable defaults formatting properties}
\usage{
get_flextable_defaults()
}
\value{
a list containing default values.
}
\description{
The current formatting properties
are automatically applied to every flextable you produce.
These default values are returned by this function.
}
\examples{
get_flextable_defaults()
}
\seealso{
Other themes and defaults:
\code{\link[=set_flextable_defaults]{set_flextable_defaults()}},
\code{\link[=theme_alafoli]{theme_alafoli()}},
\code{\link[=theme_apa]{theme_apa()}},
\code{\link[=theme_booktabs]{theme_booktabs()}},
\code{\link[=theme_borderless]{theme_borderless()}},
\code{\link[=theme_box]{theme_box()}},
\code{\link[=theme_tron]{theme_tron()}},
\code{\link[=theme_tron_legacy]{theme_tron_legacy()}},
\code{\link[=theme_vader]{theme_vader()}},
\code{\link[=theme_vanilla]{theme_vanilla()}},
\code{\link[=theme_zebra]{theme_zebra()}}
}
\concept{themes_and_defaults}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_word_field}
\alias{as_word_field}
\title{Word dynamic field chunk}
\usage{
as_word_field(x, props = NULL, width = 0.1, height = 0.15, unit = "in")
}
\arguments{
\item{x}{computed field strings}

\item{props}{text properties (see \code{\link[=fp_text_default]{fp_text_default()}} or \code{\link[officer:fp_text]{officer::fp_text()}})
object to be used to format the text. If not specified, it will use
the default text properties of the cell(s).}

\item{width, height}{size computed field}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}
}
\description{
\code{as_word_field()} inserts a Word field code (e.g. page
numbers, dates, cross-references) as a chunk inside a
flextable cell. Field codes are Word's mechanism for
auto-computed values; see
\href{https://support.microsoft.com/en-US/Word/list-of-field-codes-in-word}{Microsoft's field-code reference}
for the available codes.

The chunk is used with \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}. It only has an effect in Word
(docx) output; other formats ignore it. To apply it
conditionally, use the post-processing step (see
\code{set_flextable_defaults(post_process_docx = ...)}).

\strong{Important}: fields are inserted but not computed.
After opening the document in Word, select all text
and press \code{F9} (on macOS: \code{Fn + F9}) to refresh the
field values.
}
\examples{
library(flextable)

# define some default values ----
set_flextable_defaults(font.size = 22, border.color = "gray")

# an example with append_chunks ----
pp_docx <- function(x) {
  x <- add_header_lines(x, "Page ")
  x <- append_chunks(
    x = x, i = 1, part = "header", j = 1,
    as_word_field(x = "Page")
  )
  align(x, part = "header", align = "left")
}
ft_1 <- flextable(cars)
ft_1 <- autofit(ft_1)
ft_1 <- pp_docx(ft_1)

## or:
# set_flextable_defaults(post_process_docx = pp_docx)
## to prevent this line addition when output is not docx

# print(ft_1, preview = "docx")

# an example with compose ----

library(officer)
ft_2 <- flextable(head(cars))
ft_2 <- add_footer_lines(ft_2, "temp text")
ft_2 <- compose(
  x = ft_2, part = "footer", i = 1, j = 1,
  as_paragraph(
    "p. ",
    as_word_field(x = "Page", width = .05),
    " on ", as_word_field(x = "NumPages", width = .05)
  )
)
ft_2 <- autofit(ft_2, part = c("header", "body"))

doc <- read_docx()
doc <- body_add_flextable(doc, ft_2)
doc <- body_add_break(doc)
doc <- body_add_flextable(doc, ft_2)
outfile <- print(doc, target = tempfile(fileext = ".docx"))

# reset default values ----
init_flextable_defaults()
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_2stats}
\alias{fmt_2stats}
\alias{fmt_summarizor}
\title{Format summarizor statistics as text}
\usage{
fmt_2stats(stat, num1, num2, cts, pcts, ...)

fmt_summarizor(stat, num1, num2, cts, pcts, ...)
}
\arguments{
\item{stat}{a character column containing the name of statictics}

\item{num1}{a numeric statistic to display such as a mean or a median}

\item{num2}{a numeric statistic to display such as a standard
deviation or a median absolute deviation.}

\item{cts}{a count to display}

\item{pcts}{a percentage to display}

\item{...}{unused arguments}
}
\description{
Format the output of \code{\link[=summarizor]{summarizor()}} into display strings:
quantitative variables are shown as mean (sd), median (IQR), or range;
qualitative variables are shown as count (percentage).
}
\examples{
library(flextable)
z <- summarizor(iris, by = "Species")

tab_1 <- tabulator(
  x = z,
  rows = c("variable", "stat"),
  columns = "Species",
  blah = as_paragraph(
    as_chunk(
      fmt_summarizor(
        stat = stat,
        num1 = value1, num2 = value2,
        cts = cts, pcts = percent
      )
    )
  )
)

ft_1 <- as_flextable(x = tab_1, separate_with = "variable")
ft_1 <- labelizor(
  x = ft_1, j = "stat",
  labels = c(
    mean_sd = "Moyenne (ecart-type)",
    median_iqr = "Mediane (IQR)",
    range = "Etendue",
    missing = "Valeurs manquantes"
  )
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=summarizor]{summarizor()}}, \code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other value formatters:
\code{\link[=fmt_avg_dev]{fmt_avg_dev()}},
\code{\link[=fmt_dbl]{fmt_dbl()}},
\code{\link[=fmt_header_n]{fmt_header_n()}},
\code{\link[=fmt_int]{fmt_int()}},
\code{\link[=fmt_n_percent]{fmt_n_percent()}},
\code{\link[=fmt_pct]{fmt_pct()}},
\code{\link[=fmt_signif_after_zeros]{fmt_signif_after_zeros()}}
}
\concept{value formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable.R
\name{regulartable}
\alias{regulartable}
\title{flextable old functions}
\usage{
regulartable(data, col_keys = names(data), cwidth = 0.75, cheight = 0.25)
}
\arguments{
\item{data}{dataset}

\item{col_keys}{columns names/keys to display. If some column names are not in
the dataset, they will be added as blank columns by default.}

\item{cwidth, cheight}{initial width and height to use for cell sizes in inches.}
}
\description{
The function is maintained for compatibility with old codes
mades by users but be aware it produces the same exact object than \code{\link[=flextable]{flextable()}}.
This function should be deprecated then removed in the next versions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paginate.R
\name{paginate}
\alias{paginate}
\title{Prevent page breaks inside a flextable}
\usage{
paginate(
  x,
  init = NULL,
  hdr_ftr = TRUE,
  group = character(),
  group_def = c("rle", "nonempty", "starts")
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{init}{init value for keep_with_next property, it default
value is \code{get_flextable_defaults()$keep_with_next}.}

\item{hdr_ftr}{if TRUE (default), prevent breaks between table body
and header and between table body and footer.}

\item{group}{name of a column to use for finding groups (when
\code{group_def} is \code{"rle"} or \code{"nonempty"}) or an integer vector
of row indices that start new groups (when \code{group_def} is \code{"starts"}).}

\item{group_def}{algorithm to be used to identify groups
that should not be split into two pages, one of 'rle', 'nonempty',
'starts':
\itemize{
\item 'rle': runs of equal values are used to define the groups,
to be used with \code{\link[=tabulator]{tabulator()}}.
\item 'nonempty': non empty value start a new group,
to be used with \code{\link[=as_flextable.tabular]{as_flextable.tabular()}}.
\item 'starts': \code{group} is a numeric vector of body row indices
where new groups begin. Page breaks are allowed before these rows.
}}
}
\value{
updated flextable object
}
\description{
Prevents breaks between tables rows you want to stay together.
This feature only applies to Word and RTF output.
}
\details{
The pagination of tables allows you to control their position
in relation to page breaks.

For small tables, a simple setting is usually used that indicates
that all rows should be displayed together:

\if{html}{\out{<div class="sourceCode">}}\preformatted{paginate(x, init = TRUE, hdr_ftr = TRUE)
}\if{html}{\out{</div>}}

For large tables, it is recommended to use a setting that
indicates that all rows of the header should be bound to
the first row of the table to avoid the case where the header
is displayed alone at the bottom of the page and then repeated
on the next one:

\if{html}{\out{<div class="sourceCode">}}\preformatted{paginate(x, init = FALSE, hdr_ftr = TRUE)
}\if{html}{\out{</div>}}

For tables that present groups that you don't want to be presented
on two pages, you must use a parameterization involving the notion
of group and an algorithm for determining the groups.

\if{html}{\out{<div class="sourceCode">}}\preformatted{paginate(x, group = "grp", group_def = "rle")
}\if{html}{\out{</div>}}
}
\examples{
\dontshow{
data.table::setDTthreads(1)
}
library(data.table)
library(flextable)

init_flextable_defaults()

multi_fun <- function(x) {
  list(mean = mean(x), sd = sd(x))
}

dat <- as.data.table(ggplot2::diamonds)
dat <- dat[clarity \%in\% c("I1", "SI1", "VS2")]

dat <- dat[, unlist(lapply(.SD, multi_fun),
  recursive = FALSE
),
.SDcols = c("z", "y"),
by = c("cut", "color", "clarity")
]

tab <- tabulator(
  x = dat, rows = c("cut", "color"),
  columns = "clarity",
  `z stats` = as_paragraph(as_chunk(fmt_avg_dev(z.mean, z.sd, digit2 = 2))),
  `y stats` = as_paragraph(as_chunk(fmt_avg_dev(y.mean, y.sd, digit2 = 2)))
)

ft_1 <- as_flextable(tab)
ft_1 <- autofit(x = ft_1, add_w = .05)
ft_1 <- paginate(ft_1, group = "cut", group_def = "rle")

save_as_docx(ft_1, path = tempfile(fileext = ".docx"))
save_as_rtf(ft_1, path = tempfile(fileext = ".rtf"))
}
\seealso{
\code{\link[=knit_print.flextable]{knit_print.flextable()}}

Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitters.R
\name{split_rows}
\alias{split_rows}
\title{Split a flextable into pages by rows}
\usage{
split_rows(x, max_height, group = integer(0), unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{max_height}{Maximum height for each page, including
header and footer (in inches by default).}

\item{group}{Integer vector of body row indices that start a new group.
Rows belonging to the same group are kept together on a single page.
Default is \code{integer(0)} (no grouping, every row is independent).}

\item{unit}{Unit for \code{max_height}, one of \code{"in"}, \code{"cm"}, \code{"mm"}.}
}
\value{
A list of flextable objects. If no splitting is needed, a
single-element list is returned.
}
\description{
Split a flextable into a list of flextables whose body
rows fit within a given height (in inches). Header and footer are
repeated on every page. An optional \code{group} argument keeps row groups
together (no page break inside a group).
}
\note{
Footnotes are currently repeated on every page, even when
they reference rows that only appear on a specific page. This
limitation will be resolved in a future version when footnotes
are restructured to track their association with body rows.
}
\examples{
ft <- flextable(iris)
ft_pages <- split_rows(ft, max_height = 3)
length(ft_pages)
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{border_remove}
\alias{border_remove}
\title{Remove borders}
\usage{
border_remove(x)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}
}
\description{
The function is deleting all borders of the flextable object.
}
\examples{
dat <- iris[c(1:5, 51:55, 101:105), ]
ft_1 <- flextable(dat)
ft_1 <- theme_box(ft_1)
ft_1

# remove all borders
ft_2 <- border_remove(x = ft_1)
ft_2
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_h]{border_inner_h()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=hline]{hline()}},
\code{\link[=hline_bottom]{hline_bottom()}},
\code{\link[=hline_top]{hline_top()}},
\code{\link[=surround]{surround()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}},
\code{\link[=vline_right]{vline_right()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_sub}
\alias{as_sub}
\title{subscript chunk}
\usage{
as_sub(x)
}
\arguments{
\item{x}{value, if a chunk, the chunk will be updated}
}
\description{
The function is producing a chunk with
subscript vertical alignment.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
\section{Illustrations}{


\if{html}{\figure{fig_as_sub_1.png}{options: width="300"}}
}

\examples{
ft <- flextable( head(iris), col_keys = c("dummy") )

ft <- compose(ft, i = 1, j = "dummy", part = "header",
    value = as_paragraph(
      as_sub("Sepal.Length"),
      " anything "
    ) )

ft <- autofit(ft)
ft
}
\seealso{
Other chunk elements for paragraph: 
\code{\link{as_bracket}()},
\code{\link{as_b}()},
\code{\link{as_chunk}()},
\code{\link{as_equation}()},
\code{\link{as_highlight}()},
\code{\link{as_image}()},
\code{\link{as_i}()},
\code{\link{as_sup}()},
\code{\link{as_word_field}()},
\code{\link{colorize}()},
\code{\link{gg_chunk}()},
\code{\link{grid_chunk}()},
\code{\link{hyperlink_text}()},
\code{\link{linerange}()},
\code{\link{lollipop}()},
\code{\link{minibar}()},
\code{\link{plot_chunk}()}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{colformat_double}
\alias{colformat_double}
\title{Format double cells}
\usage{
colformat_double(
  x,
  i = NULL,
  j = NULL,
  big.mark = get_flextable_defaults()$big.mark,
  decimal.mark = get_flextable_defaults()$decimal.mark,
  digits = get_flextable_defaults()$digits,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{big.mark, digits, decimal.mark}{see \code{\link[=formatC]{formatC()}}}

\item{na_str, nan_str}{string to be used for NA and NaN values}

\item{prefix, suffix}{string to be used as prefix or suffix}
}
\description{
Format numeric cells in a flextable using
\code{\link[=formatC]{formatC()}} with explicit control over digits and
decimal mark.
}
\examples{
dat <- mtcars
ft <- flextable(head(dat))
ft <- colformat_double(
  x = ft,
  big.mark = ",", digits = 2, na_str = "N/A"
)
autofit(ft)
}
\seealso{
Other cells formatters:
\code{\link[=colformat_char]{colformat_char()}},
\code{\link[=colformat_date]{colformat_date()}},
\code{\link[=colformat_datetime]{colformat_datetime()}},
\code{\link[=colformat_image]{colformat_image()}},
\code{\link[=colformat_int]{colformat_int()}},
\code{\link[=colformat_lgl]{colformat_lgl()}},
\code{\link[=colformat_num]{colformat_num()}},
\code{\link[=set_formatter]{set_formatter()}}
}
\concept{cells formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{delete_rows}
\alias{delete_rows}
\title{Delete flextable rows}
\usage{
delete_rows(x, i = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
The function removes one or more rows
from a 'flextable'.
}
\details{
Deleting one or more rows will result in the deletion
of any span parameters that may have been set previously.
They will have to be redone after this operation or
performed only after this deletion.
}
\examples{
ft <- flextable(head(iris))
ft <- delete_rows(ft, i = 1:5, part = "body")
ft
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{align}
\alias{align}
\alias{align_text_col}
\alias{align_nottext_col}
\title{Set text alignment}
\usage{
align(
  x,
  i = NULL,
  j = NULL,
  align = "left",
  part = c("body", "header", "footer", "all")
)

align_text_col(x, align = "left", header = TRUE, footer = TRUE)

align_nottext_col(x, align = "right", header = TRUE, footer = TRUE)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{align}{text alignment - a single character value, or a vector of
character values equal in length to the number of columns selected by \code{j}.
Expected values must be from the set ('left', 'right', 'center', or 'justify').

If the number of columns is a multiple of the length of the \code{align} parameter,
then the values in \code{align} will be recycled across the remaining columns.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}

\item{header}{should the header be aligned with the body}

\item{footer}{should the footer be aligned with the body}
}
\description{
Change the text alignment of selected rows and columns of a flextable.
}
\examples{
# Table of 6 columns
ft_car <- flextable(head(mtcars)[, 2:7])

# All 6 columns right aligned
align(ft_car, align = "right", part = "all")

# Manually specify alignment of each column
align(
  ft_car,
  align = c("left", "right", "left", "center", "center", "right"),
  part = "all"
)

# Center-align column 2 and left-align column 5
align(ft_car, j = c(2, 5), align = c("center", "left"), part = "all")

# Alternate left and center alignment across columns 1-4 for header only
align(ft_car, j = 1:4, align = c("left", "center"), part = "header")
ftab <- flextable(mtcars)
ftab <- align_text_col(ftab, align = "left")
ftab <- align_nottext_col(ftab, align = "right")
ftab
}
\seealso{
Other formatting shortcuts:
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{set_header_labels}
\alias{set_header_labels}
\title{Rename column labels in the header}
\usage{
set_header_labels(x, ..., values = NULL)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{named arguments (names are data colnames), each element is a single character
value specifying label to use.}

\item{values}{a named list (names are data colnames), each element is a single character
value specifying label to use. If provided, argument \code{...} will be ignored.
It can also be a unamed character vector, in that case, it must have the same
length than the number of columns of the flextable.}
}
\description{
Change the display labels in the bottom row of the header.
Unlike \code{\link[=set_header_df]{set_header_df()}} which replaces the entire header structure,
this function only modifies column labels in the last header row.
}
\examples{
ft <- flextable(head(iris))
ft <- set_header_labels(ft,
  Sepal.Length = "Sepal length",
  Sepal.Width = "Sepal width", Petal.Length = "Petal length",
  Petal.Width = "Petal width"
)

ft <- flextable(head(iris))
ft <- set_header_labels(ft,
  values = list(
    Sepal.Length = "Sepal length",
    Sepal.Width = "Sepal width",
    Petal.Length = "Petal length",
    Petal.Width = "Petal width"
  )
)
ft

ft <- flextable(head(iris))
ft <- set_header_labels(
  x = ft,
  values = c(
    "Sepal length",
    "Sepal width", "Petal length",
    "Petal width", "Species")
)
ft
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{note_docx_images}
\alias{note_docx_images}
\title{internal utils for roxygen tags reuse}
\description{
internal utils for roxygen tags reuse
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format. With Quarto (\code{format: docx}) or
\code{rmarkdown::word_document()}, the resulting file must be repaired
with \code{\link[=repair_docx]{repair_docx()}}.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_grobs.R
\name{gen_grob}
\alias{gen_grob}
\title{Render a flextable as a graphic object}
\usage{
gen_grob(
  x,
  ...,
  fit = c("auto", "width", "fixed"),
  scaling = c("min", "full", "fixed"),
  wrapping = TRUE,
  autowidths = TRUE,
  just = NULL
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{Reserved for extra arguments}

\item{fit}{Determines the fitting/scaling of the grob on its parent viewport.
One of \code{auto}, \code{width}, \code{fixed}, \code{TRUE}, \code{FALSE}:
\itemize{
\item \code{auto} or \code{TRUE} (default): The grob is resized to fit in the parent viewport.
The table row heights and column widths are resized proportionally.
\item \code{width}: The grob is resized horizontally to fit the width
of the parent viewport. The column widths are resized proportionally.
The row heights are unaffected and the table height may be smaller or larger
than the height of the parent viewport.
\item \code{fixed} or \code{FALSE}: The grob will have fixed dimensions,
as determined by the column widths and the row heights.
}}

\item{scaling}{Determines the scaling of the table contents.
One of \code{min}, \code{full}, \code{fixed}, \code{TRUE}, \code{FALSE}:
\itemize{
\item \code{min} or \code{TRUE} (default):
When the parent viewport is smaller than the necessary,
the various content sizes (text font size, line width and image dimensions)
will decrease accordingly so that the content can still fit.
When the parent viewport is larger than the necessary,
the content sizes will remain the same, they will not increase.
\item \code{full}: Same as \code{min}, except that the content sizes are scaled fully,
they will increase or decrease, according to the size of the drawing surface.
\item \code{fixed} or \code{FALSE}: The content sizes will not be scaled.
}}

\item{wrapping}{Determines the soft wrapping (line breaking) method
for the table cell contents. One of \code{TRUE}, \code{FALSE}:
\itemize{
\item \code{TRUE}: Text content may wrap into separate lines at normal word break points
(such as a space or tab character between two words) or at newline characters
anywhere in the text content. If a word does not fit in the available cell width,
then the text content may wrap at any character.
Non-text content (such as images) is also wrapped into new lines,
according to the available cell width.
\item \code{FALSE}: Text content may wrap only with a newline character.
Non-text content is not wrapped.
}

Superscript and subscript chunks do not wrap.
Newline and tab characters are removed from these chunk types.}

\item{autowidths}{If \code{TRUE} (default) the column widths are adjusted
in order to fit the contents of the cells (taking into account the \code{wrapping} setting).}

\item{just}{Justification of viewport layout,
same as \code{just} argument in \code{\link[grid:grid.layout]{grid::grid.layout()}}.
When set to \code{NULL} (default), it is determined according to the \code{fit} argument.}
}
\value{
a grob (gTree) object made with package \code{grid}
}
\description{
\code{gen_grob()} converts a flextable into a Grid Graphics
object (\code{grob}) that can be drawn on any R graphic device.
This is the function behind \code{\link[=save_as_image]{save_as_image()}} and the
patchwork integration (\code{\link[=wrap_flextable]{wrap_flextable()}}).

Typical uses:
\itemize{
\item embed a flextable in a \code{ggplot2} plot (via
\code{\link[=wrap_flextable]{wrap_flextable()}} or cowplot)
\item export a flextable as a PNG or SVG image (via
\code{\link[=save_as_image]{save_as_image()}})
}

Text wrapping and scaling are supported. The \code{fit}
argument controls how the table adapts to the available
space (fixed size, auto-fit width, or fill the device).

Not recommended for very large tables because the
grid calculations can be slow.

Limitations: equations (\code{\link[=as_equation]{as_equation()}}) and hyperlinks
(\code{\link[officer:hyperlink_ftext]{officer::hyperlink_ftext()}}) are not rendered.

Use a 'ragg', 'svglite' or 'ggiraph' device for correct
rendering.
}
\section{size}{


The size of the flextable can be known by using the method
\link[=dim.flextableGrob]{dim} on the grob.
}

\section{caption}{

It's important to note that captions are not part of the table itself.
This means when exporting a table to PNG or SVG formats (image formats),
the caption won't be included. Captions are intended for document outputs
like Word, HTML, or PDF, where tables are embedded within the document
itself.
}

\examples{
\dontshow{
data.table::setDTthreads(1)
}
library(ragg)
library(gdtools)
register_liberationsans()

set_flextable_defaults(font.family = "Liberation Sans")

ft <- flextable(head(mtcars))
gr <- gen_grob(ft)
\dontshow{
cap <- ragg::agg_capture(width = 7, height = 4, units = "in", res = 150)
grDevices::dev.control("enable")
}
plot(gr)
\dontshow{
raster <- cap()
dev.off()
plot(as.raster(raster))
}
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{italic}
\alias{italic}
\title{Set italic font}
\usage{
italic(x, i = NULL, j = NULL, italic = TRUE, part = "body")
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{columns selection}

\item{italic}{boolean value}

\item{part}{partname of the table (one of 'all', 'body', 'header', 'footer')}
}
\description{
change font decoration of selected rows and columns of a flextable.
}
\section{Illustrations}{


\if{html}{\figure{fig_italic_1.png}{options: width="400"}}
}

\examples{
ft <- flextable(head(mtcars))
ft <- italic(ft, italic = TRUE, part = "header")
}
\seealso{
Other sugar functions for table style: 
\code{\link{align}()},
\code{\link{bg}()},
\code{\link{bold}()},
\code{\link{color}()},
\code{\link{empty_blanks}()},
\code{\link{fontsize}()},
\code{\link{font}()},
\code{\link{highlight}()},
\code{\link{line_spacing}()},
\code{\link{padding}()},
\code{\link{rotate}()},
\code{\link{valign}()}
}
\concept{sugar functions for table style}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_avg_dev}
\alias{fmt_avg_dev}
\title{Format mean and standard deviation as text}
\usage{
fmt_avg_dev(avg, dev, digit1 = 1, digit2 = 1)
}
\arguments{
\item{avg, dev}{mean and sd values}

\item{digit1, digit2}{number of digits to show when printing 'mean' and 'sd'.}
}
\description{
Format means and standard deviations as \code{mean (sd)}.
}
\examples{
library(flextable)

df <- data.frame(avg = 1:3 * 3, sd = 1:3)

ft_1 <- flextable(df, col_keys = "avg")
ft_1 <- mk_par(
  x = ft_1, j = 1, part = "body",
  value = as_paragraph(fmt_avg_dev(avg = avg, dev = sd))
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other value formatters:
\code{\link[=fmt_2stats]{fmt_2stats()}},
\code{\link[=fmt_dbl]{fmt_dbl()}},
\code{\link[=fmt_header_n]{fmt_header_n()}},
\code{\link[=fmt_int]{fmt_int()}},
\code{\link[=fmt_n_percent]{fmt_n_percent()}},
\code{\link[=fmt_pct]{fmt_pct()}},
\code{\link[=fmt_signif_after_zeros]{fmt_signif_after_zeros()}}
}
\concept{value formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{save_as_image}
\alias{save_as_image}
\title{Save a flextable in a 'png' or 'svg' file}
\usage{
save_as_image(x, path, expand = 10, res = 200, ...)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{path}{image file to be created. It should end with '.png'
or '.svg'.}

\item{expand}{space in pixels to add around the table.}

\item{res}{The resolution of the device}

\item{...}{unused arguments}
}
\value{
a string containing the full name of the generated file
}
\description{
Save a flextable as a png or svg image.
This function uses R graphic system to create an image from the flextable,
allowing for high-quality image output. See \code{\link[=gen_grob]{gen_grob()}} for more options.
}
\section{caption}{

It's important to note that captions are not part of the table itself.
This means when exporting a table to PNG or SVG formats (image formats),
the caption won't be included. Captions are intended for document outputs
like Word, HTML, or PDF, where tables are embedded within the document
itself.
}

\examples{
library(gdtools)
register_liberationsans()
set_flextable_defaults(font.family = "Liberation Sans")

ft <- flextable(head(mtcars))
ft <- autofit(ft)
tf <- tempfile(fileext = ".png")
save_as_image(x = ft, path = tf)

init_flextable_defaults()
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/themes.R
\name{theme_box}
\alias{theme_box}
\title{Apply box theme}
\usage{
theme_box(x)
}
\arguments{
\item{x}{a flextable object}
}
\description{
Apply theme box to a flextable
}
\section{Illustrations}{


\if{html}{\figure{fig_theme_box_1.png}{options: width="400"}}
}

\section{behavior}{

Theme functions are not like 'ggplot2' themes. They are applied to the existing
table \strong{immediately}. If you add a row in the footer, the new row is not formatted
with the theme. The theme function applies the theme only to existing elements
when the function is called.

That is why theme functions should be applied after all elements of the table
have been added (mainly additionnal header or footer rows).

If you want to automatically apply a theme function to each flextable,
you can use the \code{theme_fun} argument of \code{\link[=set_flextable_defaults]{set_flextable_defaults()}}; be
aware that this theme function is applied as the last instruction when
calling \code{flextable()} - so if you add headers or footers to the array,
they will not be formatted with the theme.

You can also use the \code{post_process_html} argument
of \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} (or \code{post_process_pdf},
\code{post_process_docx}, \code{post_process_pptx}) to specify a theme
to be applied systematically before the \code{flextable()} is printed;
in this case, don't forget to take care that the theme doesn't
override any formatting done before the print statement.
}

\examples{
ft <- flextable(head(airquality))
ft <- theme_box(ft)
ft
}
\seealso{
Other functions related to themes: 
\code{\link{get_flextable_defaults}()},
\code{\link{set_flextable_defaults}()},
\code{\link{theme_alafoli}()},
\code{\link{theme_apa}()},
\code{\link{theme_booktabs}()},
\code{\link{theme_tron_legacy}()},
\code{\link{theme_tron}()},
\code{\link{theme_vader}()},
\code{\link{theme_vanilla}()},
\code{\link{theme_zebra}()}
}
\concept{functions related to themes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable.R
\name{set_caption}
\alias{set_caption}
\title{Set flextable caption}
\usage{
set_caption(
  x,
  caption = NULL,
  autonum = NULL,
  word_stylename = "Table Caption",
  style = word_stylename,
  fp_p = fp_par(padding = 3),
  align_with_table = TRUE,
  html_classes = NULL,
  html_escape = TRUE
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{caption}{caption value. The caption can be either a string either
a call to \code{\link[=as_paragraph]{as_paragraph()}}. In the latter case, users are free to format
the caption with colors, italic fonts, also mixed with images or
equations. Note that Quarto does not allow the use of this feature.

Caption as a string does not support 'Markdown' syntax. If you want to
add a bold text in the caption, use \code{as_paragraph('a ', as_b('bold'), ' text')}
when providing caption.}

\item{autonum}{an autonum representation. See \code{\link[officer:run_autonum]{officer::run_autonum()}}.
This has an effect only when the output is "Word" (in which case the object
is used to define the Word auto-numbering), "html" and "pdf" (in which case only
the bookmark identifier will be used). If used, the caption is preceded
by an auto-number sequence.}

\item{word_stylename, style}{'Word' style name to associate with caption paragraph. These names are available with
function \code{\link[officer:styles_info]{officer::styles_info()}} when output is Word. Argument \code{style}
is deprecated in favor of \code{word_stylename}. If the caption is defined with
\code{as_paragraph()}, some of the formattings of the paragraph style will be
replaced by the formattings associated with the chunks (such as the font).}

\item{fp_p}{paragraph formatting properties associated with the caption, see \code{\link[officer:fp_par]{officer::fp_par()}}.
It applies when possible, i.e. in HTML and 'Word' but not with bookdown.}

\item{align_with_table}{if TRUE, caption is aligned as the flextable, if FALSE,
\code{fp_p} will not be updated and alignement is as defined with \code{fp_p}.
It applies when possible, i.e. in HTML and 'Word' but not with bookdown.}

\item{html_classes}{css class(es) to apply to associate with caption paragraph
when output is 'Word'.}

\item{html_escape}{should HTML entities be escaped so that it can be safely
included as text or an attribute value within an HTML document.}
}
\description{
Set caption value in a flextable. The function
can also be used to define formattings that will be applied
if possible to Word and HTML outputs.
\itemize{
\item The caption will be associated with a paragraph style when
the output is Word. It can also be numbered as a auto-numbered
Word computed value.
\item The PowerPoint format ignores captions. PowerPoint documents are not
structured and do not behave as HTML documents and paginated documents
(word, pdf), and it's not possible to know where we should create
a shape to contain the caption (technically it can't be in the
PowerPoint shape containing the table).
}

When working with 'R Markdown' or 'Quarto', the caption settings
defined with \code{set_caption()} will be prioritized over knitr chunk options.

Caption value can be a single string or the result to a call to
\code{\link[=as_paragraph]{as_paragraph()}}. With the latter, the caption is made of
formatted chunks whereas with the former, caption will not be
associated with any formatting.
}
\details{
The behavior of captions in the 'flextable' package varies depending on the formats
and technologies used.

The values set by the \code{set_caption()} function will be prioritized whenever
possible, including the caption ID and associated paragraph style.
However, it's important to note that the behavior may differ across different tools.
Here's what we have observed and attempted to respect, but please inform us
if you believe our observations are incorrect:
\itemize{
\item In Word and HTML documents created with 'rmarkdown' \code{rmarkdown::word_document()}
and \code{rmarkdown::html_document()}, numbered and cross-referenced captions are not
typically expected.
\item In PDF documents created with 'rmarkdown' \code{rmarkdown::pdf_document()}, numbers
are automatically added before the caption.
\item In Word and HTML documents created with 'bookdown', numbered and cross-referenced
captions are expected. 'bookdown' handles this functionality, but due to
technical reasons, the caption should not be defined within an HTML or XML block.
Therefore, when using 'flextable', the ability to format the caption content is
lost (this limitation does not apply to PDF documents).
\item HTML and PDF documents created with Quarto handle captions and
cross-references differently. Quarto replaces captions with 'tbl-cap' and 'label'
values.
\item Word documents created with Quarto present another specific case.
Currently, Quarto does not inject captions using the 'tbl-cap' and label values.
However, this is a temporary situation that is expected to change in the future.
The 'flextable' package will adapt accordingly as Quarto evolves.
\item When using the \code{body_add_flextable()} function, all the options
specified with \code{set_caption()} will be enabled.
}

Using \code{\link[=body_add_flextable]{body_add_flextable()}} enable all options specified with \code{set_caption()}.
}
\section{R Markdown}{


flextable captions can be defined from R Markdown documents by using
\code{knitr::opts_chunk$set()}. User don't always have to call \code{set_caption()}
to set a caption, he can use knitr chunk options instead. A typical call
would be:

\if{html}{\out{<div class="sourceCode">}}\preformatted{```\{r\}
#| tab.id: bookmark_id
#| tab.cap: caption text
flextable(head(cars))
```
}\if{html}{\out{</div>}}

\code{tab.id} is the caption id or bookmark, \code{tab.cap} is the caption
text. There are many options that can replace \code{set_caption()}
features. The following knitr chunk options are available:\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   Word stylename to use for table captions. \tab tab.cap.style \tab NULL \cr
   caption id/bookmark \tab tab.id \tab NULL \cr
   caption \tab tab.cap \tab NULL \cr
   display table caption on top of the table or not \tab tab.topcaption \tab TRUE \cr
   caption table sequence identifier. \tab tab.lp \tab "tab:" \cr
   prefix for numbering chunk (default to   "Table "). \tab tab.cap.pre \tab Table \cr
   suffix for numbering chunk (default to   ": "). \tab tab.cap.sep \tab " :" \cr
   title number depth \tab tab.cap.tnd \tab 0 \cr
   separator to use between title number and table number. \tab tab.cap.tns \tab "-" \cr
   caption prefix formatting properties \tab tab.cap.fp_text \tab fp_text_lite(bold = TRUE) \cr
}


See \link{knit_print.flextable} for more details.
}

\section{Formatting the caption}{


To create captions in R Markdown using the 'flextable' package and 'officer'
package, you can utilize the \code{as_paragraph()} function. This approach is
recommended when your captions require complex content, such as a combination of
different text styles or the inclusion of images and equations.

The caption is constructed as a paragraph consisting of multiple chunks. Each
chunk represents a specific portion of the caption with its desired formatting,
such as red bold text or Arial italic text.

By default, if no specific formatting is specified (using either "a string" or
\code{as_chunk("a string")}), the \code{fp_text_default()} function sets the font settings
for the caption, including the font family, boldness, italics, color, etc. The
default values can be modified using the \code{set_flextable_defaults()} function.
However, it is recommended to explicitly use \code{as_chunk()} to define the desired
formatting.

It's important to note that the style properties of the caption will not
override the formatting of the individual elements within it. Therefore, you
need to explicitly specify the font to be used for the caption.

Here's an example of how to set a caption for a flextable in R Markdown using
the 'officer' package:

\if{html}{\out{<div class="sourceCode">}}\preformatted{library(flextable)
library(officer)

ftab <- flextable(head(cars)) \%>\%
  set_caption(
    as_paragraph(
      as_chunk("caption", props = fp_text_default(font.family = "Cambria"))
    ), word_stylename = "Table Caption"
  )

print(ftab, preview = "docx")
}\if{html}{\out{</div>}}

In this example, the \code{set_caption()} function sets the caption for the
flextable. The caption is created using \code{as_paragraph()} with a single chunk
created using \code{as_chunk("caption", props = fp_text_default(font.family = "Cambria"))}. The \code{word_stylename} parameter is used to specify the table
caption style in the resulting Word document. Finally, the \code{print()} function
generates the flextable with the caption, and \code{preview = "docx"} displays a
preview of the resulting Word document.
}

\section{Using 'Quarto'}{


In 'Quarto', captions and cross-references are handled differently
compared to 'R Markdown', where flextable takes care of the job.
In Quarto, the responsibility for managing captions lies with the Quarto
framework itself. Consequently, the \code{set_caption()} function in 'flextable'
is not as useful in a 'Quarto' document. The formatting and numbering of
captions are determined by Quarto rather than flextable. Please refer to
the Quarto documentation for more information on how to work with captions
in Quarto.
}

\examples{
ftab <- flextable(head(iris))
ftab <- set_caption(ftab, "my caption")
ftab

library(officer)
autonum <- run_autonum(seq_id = "tab", bkm = "mtcars")
ftab <- flextable(head(mtcars))
ftab <- set_caption(ftab, caption = "mtcars data", autonum = autonum)
ftab
}
\seealso{
\code{\link[=flextable]{flextable()}}, \code{\link[=knit_print.flextable]{knit_print.flextable()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{plot.flextable}
\alias{plot.flextable}
\title{Plot a flextable}
\usage{
\method{plot}{flextable}(x, ...)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{additional arguments passed to \code{\link[=gen_grob]{gen_grob()}}.}
}
\description{
plots a flextable as a grid grob object
and display the result in a new graphics window.
'ragg' or 'svglite' or 'ggiraph' graphical device drivers
should be used to ensure a correct rendering.
}
\section{caption}{

It's important to note that captions are not part of the table itself.
This means when exporting a table to PNG or SVG formats (image formats),
the caption won't be included. Captions are intended for document outputs
like Word, HTML, or PDF, where tables are embedded within the document
itself.
}

\examples{
library(gdtools)
library(ragg)
register_liberationsans()
set_flextable_defaults(font.family = "Liberation Sans")
ftab <- as_flextable(cars)

\dontshow{
cap <- ragg::agg_capture(width = 7, height = 6, units = "in", res = 150)
grDevices::dev.control("enable")
}
plot(ftab)
\dontshow{
raster <- cap()
dev.off()
plot(as.raster(raster))
init_flextable_defaults()
}
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{delete_part}
\alias{delete_part}
\title{Delete flextable part}
\usage{
delete_part(x, part = "header")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
indicate to not print a part of
the flextable, i.e. an header, footer or the body.
}
\examples{
ft <- flextable(head(iris))
ft <- delete_part(x = ft, part = "header")
ft
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz-defunct.R
\name{lollipop}
\alias{lollipop}
\title{Mini lollipop chart chunk wrapper}
\usage{
lollipop(...)
}
\description{
Function is now defunct, use \code{\link[=gg_chunk]{gg_chunk()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{hrule}
\alias{hrule}
\title{Set how row heights are determined}
\usage{
hrule(x, i = NULL, rule = "auto", part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{rule}{specify the meaning of the height. Possible values
are "atleast" (height should be at least the value specified), "exact"
(height should be exactly the value specified), or the default value "auto"
(height is determined based on the height of the contents, so the value is ignored).}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{hrule()} controls whether row heights are automatic,
minimum or fixed. This only affects Word and PowerPoint
outputs; it has no effect on HTML or PDF.
\itemize{
\item \code{"auto"} (default): the row height adjusts to fit the
content; any value set by \code{\link[=height]{height()}} is ignored.
\item \code{"atleast"}: the row is at least as tall as the value
set by \code{\link[=height]{height()}}, but can grow if the content is taller.
\item \code{"exact"}: the row is exactly the height set by
\code{\link[=height]{height()}}; content that overflows is clipped.
}

For PDF see the \code{ft.arraystretch} chunk option.
}
\examples{

ft_1 <- flextable(head(iris))
ft_1 <- width(ft_1, width = 1.5)
ft_1 <- height(ft_1, height = 0.75, part = "header")
ft_1 <- hrule(ft_1, rule = "exact", part = "header")
ft_1

ft_2 <- hrule(ft_1, rule = "auto", part = "header")
ft_2
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{add_footer}
\alias{add_footer}
\title{Add footer rows with one value per column}
\usage{
add_footer(x, top = TRUE, ..., values = NULL)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{top}{should the rows be inserted at the top or the bottom.}

\item{...}{named arguments (names are data colnames) of values
to add. It is important to insert data of the same type as the
original data, otherwise it will be transformed (probably
into strings if you add a \code{character} where a \code{double} is expected).
This makes possible to still format cell contents with the \verb{colformat_*}
functions, for example \code{\link[=colformat_num]{colformat_num()}}.}

\item{values}{a list of name-value pairs of labels or values,
names should be existing col_key values. This argument can be used
instead of \code{...} for programming purpose (If \code{values} is
supplied argument \code{...} is ignored).}
}
\description{
Add new rows to the footer where each value maps to a named column.
Unlike \code{\link[=add_footer_row]{add_footer_row()}} where labels can span multiple columns,
here each value fills exactly one column.

If some columns are not provided, they will be replaced by
\code{NA} and displayed as empty.
}
\examples{
new_row <- as.list(colMeans(iris[, -5]))
new_row$Species <- "Means"

formatter <- function(x) sprintf("\%.1f", x)

ft <- flextable(data = head(iris))
ft <- add_footer(ft, values = new_row)

# cosmetics
ft <- compose(
  x = ft, j = 1:4,
  value = as_paragraph(
    as_chunk(., formatter = formatter)
  ),
  part = "footer", use_dot = TRUE
)
ft <- align(ft, part = "footer", align = "right", j = 1:4)
ft
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df_printer.R
\name{use_df_printer}
\alias{use_df_printer}
\title{Set data.frame automatic printing as a flextable}
\usage{
use_df_printer()
}
\description{
Define \code{\link[=df_printer]{df_printer()}} as data.frame
print method in an R Markdown document.

In a setup run chunk:

\if{html}{\out{<div class="sourceCode">}}\preformatted{flextable::use_df_printer()
}\if{html}{\out{</div>}}
}
\seealso{
\code{\link[=df_printer]{df_printer()}}, \code{\link[=flextable]{flextable()}}

Other flextable configuration:
\code{\link[=add_latex_dep]{add_latex_dep()}},
\code{\link[=use_flextable_qmd]{use_flextable_qmd()}},
\code{\link[=use_model_printer]{use_model_printer()}}
}
\concept{flextable_configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{flextable_to_rmd}
\alias{flextable_to_rmd}
\title{Print a flextable inside knitr loops and conditionals}
\usage{
flextable_to_rmd(x, ...)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{unused argument}
}
\description{
Print flextable in R Markdown or Quarto documents
within \code{for} loop or \code{if} statement.

The function is particularly useful when you want
to generate flextable in a loop from a R Markdown document.

Inside R Markdown document, chunk option \code{results} must be
set to 'asis'.

See \link{knit_print.flextable} for more details.
}
\examples{
\dontrun{
library(rmarkdown)
if (pandoc_available() &&
  pandoc_version() > numeric_version("2")) {
  demo_loop <- system.file(
    package = "flextable",
    "examples/rmd",
    "loop_with_flextable.Rmd"
  )
  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy(demo_loop, to = rmd_file, overwrite = TRUE)
  render(
    input = rmd_file, output_format = "html_document",
    output_file = "loop_with_flextable.html"
  )
}
}
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.lm}
\alias{as_flextable.lm}
\title{Transform a 'lm' object into a flextable}
\usage{
\method{as_flextable}{lm}(x, ...)
}
\arguments{
\item{x}{lm model}

\item{...}{unused argument}
}
\description{
produce a flextable describing a
linear model produced by function \code{lm}.

You can remove significance stars by setting options
\code{options(show.signif.stars = FALSE)}.
}
\examples{
if (require("broom")) {
  lmod <- lm(rating ~ complaints + privileges +
    learning + raises + critical, data = attitude)
  ft <- as_flextable(lmod)
  ft
}
}
\seealso{
Other as_flextable methods: 
\code{\link{as_flextable}()},
\code{\link{as_flextable.compact_summary}()},
\code{\link{as_flextable.data.frame}()},
\code{\link{as_flextable.gam}()},
\code{\link{as_flextable.glm}()},
\code{\link{as_flextable.grouped_data}()},
\code{\link{as_flextable.htest}()},
\code{\link{as_flextable.kmeans}()},
\code{\link{as_flextable.merMod}()},
\code{\link{as_flextable.pam}()},
\code{\link{as_flextable.summarizor}()},
\code{\link{as_flextable.table}()},
\code{\link{as_flextable.tabular}()},
\code{\link{as_flextable.tabulator}()},
\code{\link{as_flextable.xtable}()},
\code{\link{compact_summary}()}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{chunk_dataframe}
\alias{chunk_dataframe}
\title{create a chunk representation suitable for flextable}
\usage{
chunk_dataframe(...)
}
\arguments{
\item{...}{values to set.}
}
\value{
a data.frame with an additional class "chunk" that makes it
suitable for beeing used in \code{\link[=as_paragraph]{as_paragraph()}}
}
\description{
This function is to be used by external packages
that want to provide an object that can be inserted as a chunk
object in paragraphs of a flextable object.
}
\section{text pattern with default values}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{chunk_dataframe(txt = c("any text", "other text"))
}\if{html}{\out{</div>}}
}

\section{text pattern with bold set to TRUE}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{chunk_dataframe(
  txt = c("any text", "other text"),
  bold = c(TRUE, TRUE))
}\if{html}{\out{</div>}}
}

\section{text pattern with control over all formatting properties}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{chunk_dataframe(
  txt = c("any text", "other text"),
  font.size = c(12, 10),
  italic = c(FALSE, TRUE),
  bold = c(FALSE, TRUE),
  underlined = c(FALSE, TRUE),
  color = c("black", "red"),
  shading.color = c("transparent", "yellow"),
  font.family = c("Arial", "Arial"),
  hansi.family = c("Arial", "Arial"),
  eastasia.family = c("Arial", "Arial"),
  cs.family = c("Arial", "Arial"),
  vertical.align = c("top", "bottom") )
}\if{html}{\out{</div>}}
}

\section{text with url pattern}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{chunk_dataframe(
  txt = c("any text", "other text"),
  url = rep("https://www.google.fr", 2),
  font.size = c(12, 10),
  italic = c(FALSE, TRUE),
  bold = c(FALSE, TRUE),
  underlined = c(FALSE, TRUE),
  color = c("black", "red"),
  shading.color = c("transparent", "yellow"),
  font.family = c("Arial", "Arial"),
  hansi.family = c("Arial", "Arial"),
  eastasia.family = c("Arial", "Arial"),
  cs.family = c("Arial", "Arial"),
  vertical.align = c("top", "bottom") )
}\if{html}{\out{</div>}}
}

\section{images pattern}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{chunk_dataframe(width = width, height = height, img_data = files )
}\if{html}{\out{</div>}}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nrowcol.R
\name{ncol_keys}
\alias{ncol_keys}
\title{Number of columns}
\usage{
ncol_keys(x)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}
}
\description{
returns the number of columns displayed
}
\examples{
library(flextable)
ft <- qflextable(head(cars))
ncol_keys(ft)
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable_tabulator.R
\name{as_flextable.tabulator}
\alias{as_flextable.tabulator}
\title{tabulator to flextable}
\usage{
\method{as_flextable}{tabulator}(
  x,
  separate_with = character(0),
  big_border = fp_border_default(width = 1.5),
  small_border = fp_border_default(width = 0.75),
  rows_alignment = "left",
  columns_alignment = "center",
  label_rows = x$rows,
  spread_first_col = FALSE,
  expand_single = FALSE,
  sep_w = 0.05,
  unit = "in",
  ...
)
}
\arguments{
\item{x}{result from \code{\link[=tabulator]{tabulator()}}}

\item{separate_with}{columns used to sepatate the groups
with an horizontal line.}

\item{big_border, small_border}{big and small border properties defined
by a call to \code{\link[=fp_border_default]{fp_border_default()}} or \code{\link[=fp_border]{fp_border()}}.}

\item{rows_alignment, columns_alignment}{alignments to apply to
columns corresponding to \code{rows} and \code{columns}; see arguments
\code{rows} and \code{columns} in \code{\link[=tabulator]{tabulator()}}.}

\item{label_rows}{labels to use for the first column names, i.e.
the \emph{row} column names. It must be a named vector, the values will
be matched based on the names.}

\item{spread_first_col}{if TRUE, first row is spread as a new line separator
instead of being a column. This helps to reduce the width and allows for
clear divisions.}

\item{expand_single}{if FALSE (the default), groups with only one
row will not be expanded with a title row. If TRUE,
single row groups and multi-row groups are all
restructured.}

\item{sep_w}{blank column separators'width to be used. If 0,
blank column separators will not be used.}

\item{unit}{unit of argument \code{sep_w}, one of "in", "cm", "mm".}

\item{...}{unused argument}
}
\description{
\code{tabulator} object can be transformed as a flextable
with method \code{\link[=as_flextable]{as_flextable()}}.
}
\examples{
library(flextable)

set_flextable_defaults(digits = 2, border.color = "gray")

if(require("stats")){
  dat <- aggregate(breaks ~ wool + tension,
    data = warpbreaks, mean)

  cft_1 <- tabulator(x = dat,
                     rows = "wool",
    columns = "tension",
    `mean` = as_paragraph(as_chunk(breaks)),
    `(N)` = as_paragraph(
      as_chunk(length(breaks) ))
  )

  ft_1 <- as_flextable(cft_1, sep_w = .1)
  ft_1

  set_flextable_defaults(padding = 1, font.size = 9, border.color = "orange")
  ft_2 <- as_flextable(cft_1, sep_w = 0)
  ft_2

  set_flextable_defaults(padding = 6, font.size = 11,
                         border.color = "white", font.color = "white",
                         background.color = "#333333")
  ft_3 <- as_flextable(
    x = cft_1, sep_w = 0,
    rows_alignment = "center",
    columns_alignment = "right")
  ft_3
}

init_flextable_defaults()
}
\seealso{
\code{\link[=summarizor]{summarizor()}}, \code{\link[=as_grouped_data]{as_grouped_data()}}

Other as_flextable methods: 
\code{\link{as_flextable.data.frame}()},
\code{\link{as_flextable.glm}()},
\code{\link{as_flextable.grouped_data}()},
\code{\link{as_flextable.htest}()},
\code{\link{as_flextable.kmeans}()},
\code{\link{as_flextable.lm}()},
\code{\link{as_flextable.merMod}()},
\code{\link{as_flextable.pam}()},
\code{\link{as_flextable.summarizor}()},
\code{\link{as_flextable.xtable}()},
\code{\link{as_flextable}()}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.data.frame}
\alias{as_flextable.data.frame}
\title{Printing data.frame summary as a flextable}
\usage{
\method{as_flextable}{data.frame}(
  x,
  max_row = 10,
  split_colnames = FALSE,
  short_strings = FALSE,
  short_size = 35,
  short_suffix = "...",
  do_autofit = TRUE,
  show_coltype = TRUE,
  color_coltype = "#999999",
  ...
)
}
\arguments{
\item{x}{a data.frame}

\item{max_row}{The number of rows to print. Default to 10.}

\item{split_colnames}{Should the column names be split
(with non alpha-numeric characters). Default to FALSE.}

\item{short_strings}{Should the character column be shorten.
Default to FALSE.}

\item{short_size}{Maximum length of character column if
\code{short_strings} is TRUE. Default to 35.}

\item{short_suffix}{Suffix to add when character values are shorten.
Default to "...".}

\item{do_autofit}{Use \code{\link[=autofit]{autofit()}} before rendering the table.
Default to TRUE.}

\item{show_coltype}{Show column types.
Default to TRUE.}

\item{color_coltype}{Color to use for column types.
Default to "#999999".}

\item{...}{unused arguments}
}
\description{
It displays the first rows and shows the column types.
If there is only one row, a simplified vertical table is produced.
}
\examples{
as_flextable(mtcars)
}
\seealso{
Other as_flextable methods: 
\code{\link{as_flextable.glm}()},
\code{\link{as_flextable.grouped_data}()},
\code{\link{as_flextable.htest}()},
\code{\link{as_flextable.kmeans}()},
\code{\link{as_flextable.lm}()},
\code{\link{as_flextable.merMod}()},
\code{\link{as_flextable.pam}()},
\code{\link{as_flextable.summarizor}()},
\code{\link{as_flextable.tabulator}()},
\code{\link{as_flextable.xtable}()},
\code{\link{as_flextable}()}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_structure.R
\name{information_data_cell}
\alias{information_data_cell}
\title{Get cell-level information from a flextable}
\usage{
information_data_cell(x)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}
}
\value{
a data.frame containing information about cells:
\itemize{
\item formatting properties,
\item part (\code{.part}), row (\code{.row_id}) and column (\code{.col_id}).
}
}
\description{
This function takes a flextable object and returns a data.frame containing
information about each cell within the flextable. The data.frame includes
details about formatting properties and position within the
row and column.
}
\section{don't use this}{


These data structures should not be used, as they
represent an interpretation of the underlying data
structures, which may evolve over time.

\strong{They are exported to enable two packages that exploit
these structures to make a transition, and should not
remain available for long.}
}

\examples{
ft <- as_flextable(iris)
x <- information_data_cell(ft)
head(x)
}
\seealso{
Other table data extraction:
\code{\link[=information_data_chunk]{information_data_chunk()}},
\code{\link[=information_data_paragraph]{information_data_paragraph()}}
}
\concept{table data extraction}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{font}
\alias{font}
\title{Set font}
\usage{
font(
  x,
  i = NULL,
  j = NULL,
  fontname,
  part = "body",
  cs.family = fontname,
  hansi.family = fontname,
  eastasia.family = fontname
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{fontname}{single character value, the font family name.
With Word and PowerPoint output, this value specifies the font to
be used for formatting characters in the Unicode range (U+0000-U+007F).}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}

\item{cs.family}{Optional font to be used for formatting
characters in a complex script Unicode range. For example, Arabic
text might be displayed using the "Arial Unicode MS" font.
Used only with Word and PowerPoint outputs. The default value is the value
of \code{fontname}.}

\item{hansi.family}{Optional font to be used for formatting
characters in a Unicode range that does not fall into one of the
other categories.
Used only with Word and PowerPoint outputs. The default value is the value
of \code{fontname}.}

\item{eastasia.family}{Optional font to be used for
formatting characters in an East Asian Unicode range. For example,
Japanese text might be displayed using the "MS Mincho" font.
Used only with Word and PowerPoint outputs. The default value is the value
of \code{fontname}.}
}
\description{
Change the font of selected rows and columns of a flextable.

Fonts impact the readability and aesthetics of the table. Font families
refer to a set of typefaces that share common design features, such as 'Arial'
and 'Open Sans'.

'Google Fonts' is a popular library of free web fonts that can be
easily integrated into flextable with the \code{\link[gdtools:register_gfont]{gdtools::register_gfont()}} function.
When the output is HTML, the font will be automatically added to the HTML
document.
}
\examples{
library(gdtools)
fontname <- "Brush Script MT"

if (font_family_exists(fontname)) {
  ft_1 <- flextable(head(iris))
  ft_2 <- font(ft_1, fontname = fontname, part = "header")
  ft_2 <- font(ft_2, fontname = fontname, j = 5)
  ft_2
}
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/body_add_flextable.R
\name{body_add_flextable}
\alias{body_add_flextable}
\title{Add flextable into a Word document}
\usage{
body_add_flextable(
  x,
  value,
  align = NULL,
  pos = "after",
  split = NULL,
  topcaption = TRUE,
  keepnext = NULL
)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{\code{flextable} object}

\item{align}{left, center (default) or right.
The \code{align} parameter is still supported for the time being, but
we recommend using \code{set_flextable_defaults(table_align = "center")} instead
that will set this default alignment for all flextables during
the R session, or to define alignement for each table with
\code{set_table_properties(align = "center")}.}

\item{pos}{where to add the flextable relative to the cursor,
one of "after", "before", "on" (end of line).}

\item{split}{set to TRUE if you want to activate Word
option 'Allow row to break across pages'.
This argument is still supported for the time being, but
we recommend using \code{set_flextable_defaults(split = TRUE)} instead
that will set this as default setting for all flextables during
the R session, or to define alignement for each table with
\code{\link[=set_table_properties]{set_table_properties()}} with argument \code{opts_word=list(split = TRUE)}
instead.}

\item{topcaption}{if TRUE caption is added before the table, if FALSE,
caption is added after the table.}

\item{keepnext}{Defunct in favor of \code{\link[=paginate]{paginate()}}. The default value
used for \code{keep_with_next} is set with
\code{set_flextable_defaults(keep_with_next = TRUE)}.}
}
\description{
Add a flextable into a Word document
created with 'officer'.
}
\details{
Use the \code{\link[=paginate]{paginate()}} function to define whether the table should
be displayed on one or more pages, and whether the header should be
displayed with the first lines of the table body on the same page.

Use the \code{\link[=set_caption]{set_caption()}} function to define formatted captions
(with \code{\link[=as_paragraph]{as_paragraph()}}) or simple captions (with a string).
\code{topcaption} can be used to insert the caption before the table
(default) or after the table (use \code{FALSE}).
}
\examples{
\dontshow{
init_flextable_defaults()
}
library(officer)

# define global settings
set_flextable_defaults(
  split = TRUE,
  table_align = "center",
  table.layout = "autofit"
)

# produce 3 flextable
ft_1 <- flextable(head(airquality, n = 20))
ft_1 <- color(ft_1, i = ~ Temp > 70, color = "red", j = "Temp")
ft_1 <- highlight(ft_1, i = ~ Wind < 8, color = "yellow", j = "Wind")
ft_1 <- set_caption(
  x = ft_1,
  autonum = run_autonum(seq_id = "tab"),
  caption = "Daily air quality measurements"
)
ft_1 <- paginate(ft_1, init = TRUE, hdr_ftr = TRUE)

ft_2 <- proc_freq(mtcars, "vs", "gear")
ft_2 <- set_caption(
  x = ft_2,
  autonum = run_autonum(seq_id = "tab", bkm = "mtcars"),
  caption = as_paragraph(
    as_b("mtcars"), " ",
    colorize("table", color = "orange")
  ),
  fp_p = fp_par(keep_with_next = TRUE)
)
ft_2 <- paginate(ft_2, init = TRUE, hdr_ftr = TRUE)

ft_3 <- summarizor(iris, by = "Species")
ft_3 <- as_flextable(ft_3, spread_first_col = TRUE)
ft_3 <- set_caption(
  x = ft_3,
  autonum = run_autonum(seq_id = "tab"),
  caption = "iris summary"
)
ft_3 <- paginate(ft_3, init = TRUE, hdr_ftr = TRUE)

# add the 3 flextable in a new Word document
doc <- read_docx()
doc <- body_add_flextable(doc, value = ft_1)
doc <- body_add_par(doc, value = "")
doc <- body_add_flextable(doc, value = ft_2)
doc <- body_add_par(doc, value = "")
doc <- body_add_flextable(doc, value = ft_3)

fileout <- tempfile(fileext = ".docx")
print(doc, target = fileout)
\dontshow{
init_flextable_defaults()
}
}
\seealso{
\code{\link[=knit_print.flextable]{knit_print.flextable()}}, \code{\link[=save_as_docx]{save_as_docx()}}

Other functions for officer integration:
\code{\link[=body_replace_flextable_at_bkm]{body_replace_flextable_at_bkm()}},
\code{\link[=ph_with.flextable]{ph_with.flextable()}},
\code{\link[=rtf_add.flextable]{rtf_add.flextable()}}
}
\concept{officer_integration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{colorize}
\alias{colorize}
\title{colorize chunk}
\usage{
colorize(x, color)
}
\arguments{
\item{x}{value, if a chunk, the chunk will be updated}

\item{color}{color to use as text highlighting color as character vector.}
}
\description{
The function is producing a chunk with
a font in color.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
\examples{
ft <- flextable( head(iris),
  col_keys = c("Sepal.Length", "dummy") )

ft <- compose(ft, j = "dummy",
  value = as_paragraph(colorize(Sepal.Length, color = "red")) )

ft
}
\seealso{
Other chunk elements for paragraph: 
\code{\link{as_bracket}()},
\code{\link{as_b}()},
\code{\link{as_chunk}()},
\code{\link{as_equation}()},
\code{\link{as_highlight}()},
\code{\link{as_image}()},
\code{\link{as_i}()},
\code{\link{as_sub}()},
\code{\link{as_sup}()},
\code{\link{as_word_field}()},
\code{\link{gg_chunk}()},
\code{\link{grid_chunk}()},
\code{\link{hyperlink_text}()},
\code{\link{linerange}()},
\code{\link{lollipop}()},
\code{\link{minibar}()},
\code{\link{plot_chunk}()}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_with.R
\name{ph_with.flextable}
\alias{ph_with.flextable}
\title{Add a flextable into a PowerPoint slide}
\usage{
\method{ph_with}{flextable}(x, value, location, ...)
}
\arguments{
\item{x}{a pptx device}

\item{value}{flextable object}

\item{location}{a location for a placeholder. See \code{\link[officer:ph_location_type]{officer::ph_location_type()}}
for example.}

\item{...}{unused arguments.}
}
\description{
Add a flextable in a PowerPoint document object produced
by \code{\link[officer:read_pptx]{officer::read_pptx()}}.

This function will create a native PowerPoint table from the flextable
and the result can be eventually edited.
}
\note{
The width and height of the table can not be set with \code{location}. Use
functions \code{\link[=width]{width()}}, \code{\link[=height]{height()}}, \code{\link[=autofit]{autofit()}}
and \code{\link[=dim_pretty]{dim_pretty()}} instead. The overall size is resulting from
cells, paragraphs and text properties (i.e. padding, font size, border widths).
}
\section{caption}{

Captions are not printed in PowerPoint slides.

While captions are useful for document formats like Word, RTF, HTML, or PDF,
they aren't directly supported in PowerPoint slides. Unlike documents with
a defined layout, PowerPoint slides lack a structured document flow. They
don't function like HTML documents or paginated formats (RTF, Word, PDF).
This makes it technically challenging to determine the ideal placement for
a caption within a slide. Additionally, including a caption within the table
itself isn't feasible.
}

\examples{
library(officer)

ft <- flextable(head(iris))

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content", "Office Theme")
doc <- ph_with(doc, ft, location = ph_location_left())

fileout <- tempfile(fileext = ".pptx")
print(doc, target = fileout)
}
\seealso{
Other functions for officer integration:
\code{\link[=body_add_flextable]{body_add_flextable()}},
\code{\link[=body_replace_flextable_at_bkm]{body_replace_flextable_at_bkm()}},
\code{\link[=rtf_add.flextable]{rtf_add.flextable()}}
}
\concept{officer_integration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{add_header_lines}
\alias{add_header_lines}
\title{Add full-width rows to the header}
\usage{
add_header_lines(x, values = character(0), top = TRUE)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{values}{a character vector or a call to \code{\link[=as_paragraph]{as_paragraph()}}
to get formated content, each element will
be added as a new row.}

\item{top}{should the row be inserted at the top
or the bottom. Default to TRUE.}
}
\description{
Add one or more rows to the header where each label
spans all columns (all cells merged into one). Useful for adding
titles or subtitles above the column headers.
}
\examples{
# ex 1----
ft_1 <- flextable(head(iris))
ft_1 <- add_header_lines(ft_1, values = "blah blah")
ft_1 <- add_header_lines(ft_1, values = c("blah 1", "blah 2"))
ft_1 <- autofit(ft_1)
ft_1

# ex 2----
ft01 <- fp_text_default(color = "red")
ft02 <- fp_text_default(color = "orange")
ref <- c("(1)", "(2)")
pars <- as_paragraph(
  as_chunk(ref, props = ft02), " ",
  as_chunk(rep("My tailor is rich", length(ref)), props = ft01)
)

ft_2 <- flextable(head(mtcars))
ft_2 <- add_header_lines(ft_2, values = pars, top = FALSE)
ft_2 <- add_header_lines(ft_2, values = ref, top = TRUE)
ft_2 <- add_footer_lines(ft_2, values = "blah", top = TRUE)
ft_2 <- add_footer_lines(ft_2, values = pars, top = TRUE)
ft_2 <- add_footer_lines(ft_2, values = ref, top = FALSE)
ft_2 <- autofit(ft_2)
ft_2
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{before}
\alias{before}
\title{Detect rows before a given value}
\usage{
before(x, entries)
}
\arguments{
\item{x}{an atomic vector of values to be tested}

\item{entries}{a sequence of items to be searched in \code{x}.}
}
\description{
Returns a logical vector indicating which elements
of \code{x} appear before the first occurrence of any of the
\code{entries} values. Useful as a row selector in \code{\link[=hline]{hline()}} to
insert a border above a summary row such as \code{"Total"}.
}
\examples{
library(flextable)
library(officer)

dat <- data.frame(
  stringsAsFactors = FALSE,
  check.names = FALSE,
  Level = c("setosa", "versicolor", "virginica", "<NA>", "Total"),
  Freq = as.integer(c(50, 50, 50, 0, 150)),
  `\% Valid` = c(
    100 / 3,
    100 / 3, 100 / 3, NA, 100
  ),
  `\% Valid Cum.` = c(100 / 3, 100 * 2 / 3, 100, NA, 100),
  `\% Total` = c(
    100 / 3,
    100 / 3, 100 / 3, 0, 100
  ),
  `\% Total Cum.` = c(
    100 / 3,
    100 * 2 / 3, 100, 100, 100
  )
)

ft <- flextable(dat)
ft <- hline(ft,
  i = ~ before(Level, "Total"),
  border = fp_border_default(width = 2)
)
ft
}
\seealso{
\code{\link[=hline]{hline()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{autofit}
\alias{autofit}
\title{Adjust columns to their content size}
\usage{
autofit(
  x,
  add_w = 0.1,
  add_h = 0.1,
  part = c("body", "header"),
  unit = "in",
  hspans = "none"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{add_w}{extra width to add in inches}

\item{add_h}{extra height to add in inches}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}

\item{unit}{unit for add_h and add_w, one of "in", "cm", "mm".}

\item{hspans}{specifies how cells that are horizontally are included in the calculation.
It must be one of the following values "none", "divided" or "included". If
"none", widths of horizontally spanned cells is set to 0 (then do not affect the
widths); if "divided", widths of horizontally spanned cells is divided by
the number of spanned cells; if "included", all widths (included horizontally
spanned cells) will be used in the calculation.}
}
\description{
Compute and apply the minimum widths and heights needed
to display each cell's content on a single line, with an optional
extra margin (\code{add_w}, \code{add_h}).

This function sizes columns to fit their content. It does not
constrain the table to a given total width. To enforce a maximum
width, use \code{\link[=fit_columns]{fit_columns()}} (wraps text) or \code{\link[=fit_to_width]{fit_to_width()}}
(shrinks font size).

Note that this function is not related to 'Microsoft Word'
\emph{Autofit} feature.

There is an alternative to fixed-width layouts that works
well with HTML and Word output that can be set
with \code{set_table_properties(layout = "autofit")}, see
\code{\link[=set_table_properties]{set_table_properties()}}.
}
\examples{
ft_1 <- flextable(head(mtcars))
ft_1
ft_2 <- autofit(ft_1)
ft_2
}
\seealso{
Other functions for flextable size management:
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{dim_pretty}
\alias{dim_pretty}
\title{Calculate optimal column widths and row heights}
\usage{
dim_pretty(x, part = "all", unit = "in", hspans = "none")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}

\item{unit}{unit for returned values, one of "in", "cm", "mm".}

\item{hspans}{specifies how cells that are horizontally are included in the calculation.
It must be one of the following values "none", "divided" or "included". If
"none", widths of horizontally spanned cells is set to 0 (then do not affect the
widths); if "divided", widths of horizontally spanned cells is divided by
the number of spanned cells; if "included", all widths (included horizontally
spanned cells) will be used in the calculation.}
}
\description{
return minimum estimated widths and heights for
each table columns and rows in inches.
}
\examples{
ftab <- flextable(head(mtcars))
dim_pretty(ftab)
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doc-selectors.R
\name{flextable_selectors}
\alias{flextable_selectors}
\title{Selectors in flextable}
\description{
Selectors are a core feature of flextable that allow you to specify which
parts (\code{part}), rows (\code{i}) and columns (\code{j}) should be affected by formatting,
styling, or content operations.

Many flextable functions support these selectors, including \code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}}, \code{\link[=color]{color()}}, \code{\link[=padding]{padding()}}, \code{\link[=fontsize]{fontsize()}}, \code{\link[=italic]{italic()}}, \code{\link[=align]{align()}},
\code{\link[=mk_par]{mk_par()}}, \code{\link[=hline]{hline()}}, \code{\link[=vline]{vline()}}, and many others.

Selectors make conditional formatting easy and enable seamless piping of
multiple operations using \verb{\%>\%} or \verb{|>}.
}
\section{Row selection with the \code{i} parameter}{


The \code{i} parameter is used to select specific rows in a flextable for
formatting, styling, or content operations.

When \code{i = NULL} (the default), operations apply to all rows.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{i = ~ condition           # Formula (body only)
i = 1:5                   # Integer vector
i = c(TRUE, FALSE, ...)   # Logical vector
i = NULL                  # All rows (default)
}\if{html}{\out{</div>}}

Best practices:
\itemize{
\item Use formulas (\code{i = ~ condition}) for conditional selection in body
\item Use integers for positional selection in any part.
\item Use \code{nrow_part()} for dynamic row selection
}
}

\section{Column selection with the \code{j} parameter}{

The \code{j} parameter is used to select specific columns in a flextable for
formatting, styling, or content operations.

When \code{j = NULL} (the default), operations apply to all columns.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{j = ~ col1 + col2         # Formula (select multiple)
j = ~ . - col1            # Formula (exclude columns)
j = c("col1", "col2")     # Character vector (recommended)
j = 1:5                   # Integer vector
j = c(TRUE, FALSE, ...)   # Logical vector
j = NULL                  # All columns (default)
}\if{html}{\out{</div>}}

Best practices:
\itemize{
\item Use character vectors (\code{j = c("col1", "col2")}) for clarity and
maintainability.
\item Use formulas (\code{j = ~ col1 + col2}) for excluding columns.
\item Avoid integer positions when possible (less maintainable).
\item Column selectors work with all parts (header, body, footer, all).
}
}

\section{Part selection with the \code{part} parameter}{


The \code{part} parameter specifies which section of the flextable should be
affected by formatting, styling, or content operations.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{part = "body"    # Data rows (default)
part = "header"  # Header rows
part = "footer"  # Footer rows
part = "all"     # All parts
}\if{html}{\out{</div>}}
\itemize{
\item When \code{part = "body"} (the default), operations apply only to the data rows.
\item When \code{part = "all"}, the operation is applied to each part independently.
}

Formula row selectors (\code{i = ~ condition}) CANNOT be used with:
\itemize{
\item \code{part = "header"} - Headers contain only character values
\item \code{part = "footer"} - Footers contain only character values
\item \code{part = "all"} - Header/footer are character-only while body has original data types
}

This restriction exists because formula selectors evaluate conditions using
the actual data types from your dataset (numeric, logical, etc.). Headers and
footers store only character representations of values, so conditional
expressions like \code{i = ~ price < 330} cannot be evaluated on them.
}

\section{Advanced Programming with Selectors}{


Function \code{\link[=nrow_part]{nrow_part()}} returns the number of lines in a part.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Format the last row differently
ft <- flextable(iris[48:52,])
ft <- bold(ft, i = nrow_part(ft, part = "body"))

# Add footnote to last row
ft <- footnote(
  ft,
  i = nrow_part(ft, part = "body"), j = 1:4,
  value = as_paragraph("Calculated mean")
)
ft
}\if{html}{\out{</div>}}

Function \code{\link[=ncol_keys]{ncol_keys()}} returns the number of columns.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Get column count
ncol_keys(ft)

# Useful for programmatic selection
ncols <- ncol_keys(ft)
ft <- bg(ft, j = rep(c(TRUE, FALSE), length.out = ncols), bg = "#f0f0f0")
ft
}\if{html}{\out{</div>}}
}

\concept{selectors in flextable}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/border_fix.R
\name{fix_border_issues}
\alias{fix_border_issues}
\title{Fix border issues when cells are merged}
\usage{
fix_border_issues(x, part = "all")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
When cells are merged, the rendered borders will be those
of the first cell. If a column is made of three merged cells, the bottom
border that will be seen will be the bottom border of the first cell in the
column. From a user point of view, this is wrong, the bottom should be the one
defined for cell 3. This function modify the border values to avoid that effect.

Note since version \verb{0.9.7} that the function is called automatically
before rendering, user should not have to call this function anymore.
}
\examples{
library(officer)
dat <- data.frame(a = 1:5, b = 6:10)
ft <- flextable(dat)
ft <- theme_box(ft)
ft <- merge_at(ft, i = 4:5, j = 1, part = "body")
ft <- hline(ft,
  i = 5, part = "body",
  border = fp_border(color = "red", width = 5)
)
print(ft)
ft <- fix_border_issues(ft)
print(ft)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{use_model_printer}
\alias{use_model_printer}
\title{set model automatic printing as a flextable}
\usage{
use_model_printer()
}
\description{
Define \code{\link[=as_flextable]{as_flextable()}} as
print method in an R Markdown document for models
of class:
\itemize{
\item lm
\item glm
\item models from package 'lme' and 'lme4'
\item htest (t.test, chisq.test, ...)
\item gam
\item kmeans and pam
}

In a setup run chunk:

\if{html}{\out{<div class="sourceCode">}}\preformatted{flextable::use_model_printer()
}\if{html}{\out{</div>}}
}
\seealso{
\code{\link[=use_df_printer]{use_df_printer()}}, \code{\link[=flextable]{flextable()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compose.R
\name{compose}
\alias{compose}
\alias{mk_par}
\title{Define displayed values and mixed content}
\usage{
compose(x, i = NULL, j = NULL, value, part = "body", use_dot = FALSE)

mk_par(x, i = NULL, j = NULL, value, part = "body", use_dot = FALSE)
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{column selection}

\item{value}{a call to function \code{\link[=as_paragraph]{as_paragraph()}}.}

\item{part}{partname of the table (one of 'all', 'body', 'header', 'footer')}

\item{use_dot}{by default \code{use_dot=FALSE}; if \code{use_dot=TRUE},
\code{value} is evaluated within a data.frame augmented of a column named \code{.}
containing the \code{j}th column.}
}
\description{
Modify flextable displayed values with eventually
mixed content paragraphs.

Function is handling complex formatting as image insertion with
\code{\link[=as_image]{as_image()}}, superscript with \code{\link[=as_sup]{as_sup()}}, formated
text with \code{\link[=as_chunk]{as_chunk()}} and several other \emph{chunk} functions.

Function \code{mk_par} is another name for \code{compose} as
there is an unwanted \strong{conflict with package 'purrr'}.

If you only need to add some content at the end
or the beginning of paragraphs and keep existing
content as it is, functions \code{\link[=append_chunks]{append_chunks()}} and
\code{\link[=prepend_chunks]{prepend_chunks()}} should be prefered.
}
\section{Illustrations}{


\if{html}{\figure{fig_compose_1.png}{options: width="117"}}

\if{html}{\figure{fig_compose_2.png}{options: width="400"}}
}

\examples{
ft_1 <- flextable(head(cars, n = 5), col_keys = c("speed", "dist", "comment"))
ft_1 <- mk_par(
  x = ft_1, j = "comment",
  i = ~ dist > 9,
  value = as_paragraph(
    colorize(as_i("speed: "), color = "gray"),
    as_sup(sprintf("\%.0f", speed))
  )
)
ft_1 <- set_table_properties(ft_1, layout = "autofit")
ft_1

# using `use_dot = TRUE` ----
set.seed(8)
dat <- iris[sample.int(n = 150, size = 10),]
dat <- dat[order(dat$Species),]


ft_2 <- flextable(dat)
ft_2 <- mk_par(ft_2, j = ~ . -Species,
  value = as_paragraph(
    minibar(., barcol = "white",
            height = .1)
  ), use_dot = TRUE
  )
ft_2 <- theme_vader(ft_2)
ft_2 <- autofit(ft_2)
ft_2
}
\seealso{
\code{\link[=fp_text_default]{fp_text_default()}}, \code{\link[=as_chunk]{as_chunk()}}, \code{\link[=as_b]{as_b()}}, \code{\link[=as_word_field]{as_word_field()}}, \code{\link[=labelizor]{labelizor()}}

Other functions for mixed content paragraphs: 
\code{\link{append_chunks}()},
\code{\link{as_paragraph}()},
\code{\link{prepend_chunks}()}
}
\concept{functions for mixed content paragraphs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{delete_columns}
\alias{delete_columns}
\title{Delete flextable columns}
\usage{
delete_columns(x, j = NULL)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}
}
\description{
The function removes one or more columns
from a 'flextable'.
}
\details{
Deleting one or more columns will result in the deletion
of any span parameters that may have been set previously.
They will have to be redone after this operation or
performed only after this deletion.
}
\examples{
ft <- flextable(head(iris))
ft <- delete_columns(ft, j = "Species")
ft
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.grouped_data}
\alias{as_flextable.grouped_data}
\title{Transform a 'grouped_data' object into a flextable}
\usage{
\method{as_flextable}{grouped_data}(x, col_keys = NULL, hide_grouplabel = FALSE, ...)
}
\arguments{
\item{x}{'grouped_data' object to be transformed into a "flextable"}

\item{col_keys}{columns names/keys to display. If some column names are not in
the dataset, they will be added as blank columns by default.}

\item{hide_grouplabel}{if TRUE, group label will not be rendered, only
level/value will be rendered.}

\item{...}{unused argument}
}
\description{
Produce a flextable from a table
produced by function \code{\link[=as_grouped_data]{as_grouped_data()}}.
}
\examples{
library(data.table)
CO2 <- CO2
setDT(CO2)
CO2$conc <- as.integer(CO2$conc)

data_co2 <- dcast(CO2, Treatment + conc ~ Type,
  value.var = "uptake", fun.aggregate = mean
)
data_co2 <- as_grouped_data(x = data_co2, groups = c("Treatment"))

ft <- as_flextable(data_co2)
ft <- add_footer_lines(ft, "dataset CO2 has been used for this flextable")
ft <- add_header_lines(ft, "mean of carbon dioxide uptake in grass plants")
ft <- set_header_labels(ft, conc = "Concentration")
ft <- autofit(ft)
ft <- width(ft, width = c(1, 1, 1))
ft
}
\seealso{
\code{\link[=as_grouped_data]{as_grouped_data()}}

Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{bold}
\alias{bold}
\title{Set bold font}
\usage{
bold(x, i = NULL, j = NULL, bold = TRUE, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{bold}{boolean value}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Change the font weight of selected rows and columns of a flextable.
}
\examples{
ft <- flextable(head(iris))
ft <- bold(ft, bold = TRUE, part = "header")
}
\seealso{
Other sugar functions for table style: 
\code{\link{align}()},
\code{\link{bg}()},
\code{\link{color}()},
\code{\link{empty_blanks}()},
\code{\link{font}()},
\code{\link{fontsize}()},
\code{\link{highlight}()},
\code{\link{italic}()},
\code{\link{keep_with_next}()},
\code{\link{line_spacing}()},
\code{\link{padding}()},
\code{\link{rotate}()},
\code{\link{style}()},
\code{\link{tab_settings}()},
\code{\link{valign}()}
}
\concept{sugar functions for table style}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{tab_settings}
\alias{tab_settings}
\title{Set tabulation marks configuration}
\usage{
tab_settings(x, i = NULL, j = NULL, value = TRUE, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{value}{an object generated by \code{\link[officer:fp_tabs]{officer::fp_tabs()}}.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Define tabulation marks configuration.
Specifying the positions and types of tabulation marks in table
paragraphs helps organize content, especially in clinical tables,
by aligning numbers properly.
}
\examples{
library(officer)
library(flextable)

z <- data.frame(
  Statistic = c("Median (Q1 ; Q3)", "Min ; Max"),
  Value = c(
    "\t999.99\t(99.9 ; 99.9)",
    "\t9.99\t(9999.9 ; 99.9)"
  )
)

ts <- fp_tabs(
  fp_tab(pos = 0.4, style = "decimal"),
  fp_tab(pos = 1.4, style = "decimal")
)

zz <- flextable(z)
zz <- tab_settings(zz, j = 2, value = ts)
zz <- width(zz, width = c(1.5, 2))


save_as_docx(zz, path = tempfile(fileext = ".docx"))
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{set_header_footer_df}
\alias{set_header_footer_df}
\alias{set_header_df}
\alias{set_footer_df}
\title{Set flextable's header or footer rows}
\usage{
set_header_df(x, mapping = NULL, key = "col_keys")

set_footer_df(x, mapping = NULL, key = "col_keys")
}
\arguments{
\item{x}{a \code{flextable} object}

\item{mapping}{a \code{data.frame} specyfing for each colname
content of the column.}

\item{key}{column to use as key when joigning data_mapping.}
}
\description{
Use a data.frame to specify flextable's header or footer rows.

The data.frame must contain a column whose values match flextable
\code{col_keys} argument, this column will be used as join key. The
other columns will be displayed as header or footer rows. The leftmost column
is used as the top header/footer row and the rightmost column
is used as the bottom header/footer row.
}
\section{Illustrations}{


\if{html}{\figure{fig_set_header_footer_df_1.png}{options: width="400"}}

\if{html}{\figure{fig_set_header_footer_df_2.png}{options: width="400"}}
}

\examples{
typology <- data.frame(
  col_keys = c(
    "Sepal.Length", "Sepal.Width", "Petal.Length",
    "Petal.Width", "Species"
  ),
  what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
  measure = c("Length", "Width", "Length", "Width", "Species"),
  stringsAsFactors = FALSE
)

ft_1 <- flextable(head(iris))
ft_1 <- set_header_df(ft_1, mapping = typology, key = "col_keys")
ft_1 <- merge_h(ft_1, part = "header")
ft_1 <- merge_v(ft_1, j = "Species", part = "header")
ft_1 <- theme_vanilla(ft_1)
ft_1 <- fix_border_issues(ft_1)
ft_1


typology <- data.frame(
  col_keys = c(
    "Sepal.Length", "Sepal.Width", "Petal.Length",
    "Petal.Width", "Species"
  ),
  unit = c("(cm)", "(cm)", "(cm)", "(cm)", ""),
  stringsAsFactors = FALSE
)
ft_2 <- set_footer_df(ft_1, mapping = typology, key = "col_keys")
ft_2 <- italic(ft_2, italic = TRUE, part = "footer")
ft_2 <- theme_booktabs(ft_2)
ft_2 <- fix_border_issues(ft_2)
ft_2
}
\seealso{
Other functions to add rows in a flextable: 
\code{\link{add_body_row}()},
\code{\link{add_body}()},
\code{\link{add_footer_lines}()},
\code{\link{add_footer_row}()},
\code{\link{add_footer}()},
\code{\link{add_header_row}()},
\code{\link{add_header}()},
\code{\link{separate_header}()},
\code{\link{set_header_labels}()}
}
\concept{functions to add rows in a flextable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz-defunct.R
\name{footers_flextable_at_bkm}
\alias{footers_flextable_at_bkm}
\title{Add flextable at a bookmark location in document's footer}
\usage{
footers_flextable_at_bkm(...)
}
\arguments{
\item{...}{unused}
}
\description{
Function is now defunct, use \code{\link[officer:prop_section]{officer::prop_section()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df_printer.R
\name{df_printer}
\alias{df_printer}
\title{data.frame automatic printing as a flextable}
\usage{
df_printer(dat, ...)
}
\arguments{
\item{dat}{the data.frame}

\item{...}{unused argument}
}
\description{
Create a summary from a data.frame as a flextable. This function
is to be used in an R Markdown document.

To use that function, you must declare it in the part \code{df_print} of the 'YAML'
header of your R Markdown document:

\if{html}{\out{<div class="sourceCode">}}\preformatted{---
df_print: !expr function(x) flextable::df_printer(x)
---
}\if{html}{\out{</div>}}

We notice an unexpected behavior with bookdown. When using bookdown it
is necessary to use \code{\link[=use_df_printer]{use_df_printer()}} instead in a setup run chunk:

\if{html}{\out{<div class="sourceCode">}}\preformatted{use_df_printer()
}\if{html}{\out{</div>}}
}
\details{
'knitr' chunk options are available to customize the output:
\itemize{
\item \code{ft_max_row}: The number of rows to print. Default to 10.
\item \code{ft_split_colnames}: Should the column names be split
(with non alpha-numeric characters). Default to FALSE.
\item \code{ft_short_strings}: Should the character column be shorten.
Default to FALSE.
\item \code{ft_short_size}: Maximum length of character column if
\code{ft_short_strings} is TRUE. Default to 35.
\item \code{ft_short_suffix}: Suffix to add when character values are shorten.
Default to "...".
\item \code{ft_do_autofit}: Use autofit() before rendering the table.
Default to TRUE.
\item \code{ft_show_coltype}: Show column types.
Default to TRUE.
\item \code{ft_color_coltype}: Color to use for column types.
Default to "#999999".
}
}
\examples{
df_printer(head(mtcars))
}
\seealso{
Other functions for flextable output and export:
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proc_freq.R
\name{proc_freq}
\alias{proc_freq}
\title{Frequency table}
\usage{
proc_freq(
  x,
  row = character(),
  col = character(),
  include.row_percent = TRUE,
  include.column_percent = TRUE,
  include.table_percent = TRUE,
  include.table_count = TRUE,
  weight = character(),
  count_format_fun = fmt_int,
  ...
)
}
\arguments{
\item{x}{a \code{data.frame} object containing variable(s) to use for counts.}

\item{row}{\code{characer} column names for row}

\item{col}{\code{characer} column names for column}

\item{include.row_percent}{\code{boolean} whether to include the row percents; defaults to \code{TRUE}}

\item{include.column_percent}{\code{boolean} whether to include the column percents; defaults to \code{TRUE}}

\item{include.table_percent}{\code{boolean} whether to include the table percents; defaults to \code{TRUE}}

\item{include.table_count}{\code{boolean} whether to include the table counts; defaults to \code{TRUE}}

\item{weight}{\code{character} column name for weight}

\item{count_format_fun}{a function to format the count values,
defaults to \link{fmt_int}.}

\item{...}{unused arguments}
}
\description{
This function computes a one or two way
contingency table and creates a flextable from the result.

The function is largely inspired by "PROC FREQ" from "SAS"
and was written with the intent to make it
as compact as possible.
}
\examples{
proc_freq(mtcars, "vs", "gear")
proc_freq(mtcars, "gear", "vs", weight = "wt")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_dbl}
\alias{fmt_dbl}
\title{Format numbers as doubles}
\usage{
fmt_dbl(x)
}
\arguments{
\item{x}{numeric values}
}
\description{
Format numeric values with decimal digits, using
the flextable default settings for separators and precision.
}
\examples{
library(flextable)

df <- data.frame(zz = .45)

ft_1 <- flextable(df)
ft_1 <- mk_par(
  x = ft_1, j = 1, part = "body",
  value = as_paragraph(as_chunk(zz, formatter = fmt_dbl))
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other text formatter functions: 
\code{\link{fmt_2stats}()},
\code{\link{fmt_avg_dev}()},
\code{\link{fmt_header_n}()},
\code{\link{fmt_int}()},
\code{\link{fmt_n_percent}()},
\code{\link{fmt_pct}()},
\code{\link{fmt_signif_after_zeros}()}
}
\concept{text formatter functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable.R
\name{flextable}
\alias{flextable}
\alias{qflextable}
\title{Create a flextable from a data frame}
\usage{
flextable(
  data,
  col_keys = names(data),
  cwidth = 0.75,
  cheight = 0.25,
  defaults = list(),
  theme_fun = theme_booktabs,
  use_labels = TRUE
)

qflextable(data)
}
\arguments{
\item{data}{dataset}

\item{col_keys}{columns names/keys to display. If some column names are not in
the dataset, they will be added as blank columns by default.}

\item{cwidth, cheight}{initial width and height to use for cell sizes in inches.}

\item{defaults, theme_fun}{deprecated, use \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} instead.}

\item{use_labels}{Logical; if TRUE, any column labels or value labels
present in the dataset will be used for display purposes. Defaults
to TRUE.}
}
\description{
Create a flextable object with function \code{flextable}.

\code{flextable} are designed to make tabular reporting easier for
R users. Functions are available to let you format text, paragraphs and cells;
table cells can be merge vertically or horizontally, row headers can easily
be defined, rows heights and columns widths can be manually set or automatically
computed.

If working with 'R Markdown' documents, you should read about knitr
chunk options in \code{\link[=knit_print.flextable]{knit_print.flextable()}} and about setting
default values with \code{\link[=set_flextable_defaults]{set_flextable_defaults()}}.
}
\section{Reuse frequently used parameters}{


Some default formatting properties are automatically
applied to every flextable you produce.

It is highly recommended to use this function because
its use will minimize the code. For example, instead of
calling the \code{fontsize()} function over and over again for
each new flextable, set the font size default value by
calling (before creating the flextables)
\code{set_flextable_defaults(font.size = 11)}. This is also
a simple way to have homogeneous arrays and make the
documents containing them easier to read.

You can change these default values with function
\code{\link[=set_flextable_defaults]{set_flextable_defaults()}}. You can reset them
with function \code{\link[=init_flextable_defaults]{init_flextable_defaults()}}. You
can access these values by calling \code{\link[=get_flextable_defaults]{get_flextable_defaults()}}.
}

\section{new lines and tabulations}{


The 'flextable' package will translate for you
the new lines expressed in the form \verb{\\n} and
the tabs expressed in the form \verb{\\t}.

The new lines will be transformed into "soft-return",
that is to say a simple carriage return and not a
new paragraph.

Tabs are different depending on the output format:
\itemize{
\item HTML is using entity \emph{em space}
\item Word - a Word 'tab' element
\item PowerPoint - a PowerPoint 'tab' element
\item latex - tag "\\quad "
\item Typst - horizontal space \verb{#h(1em)}
}
}

\section{flextable parts}{


A \code{flextable} is made of 3 parts: header, body and footer.

Most functions have an argument named \code{part} that will be used
to specify what part of of the table should be modified.
}

\section{qflextable}{

\code{qflextable} is a convenient tool to produce quickly
a flextable for reporting where layout is fixed (see
\code{\link[=set_table_properties]{set_table_properties()}}) and columns
widths are adjusted with \code{\link[=autofit]{autofit()}}.
}

\examples{
ft <- flextable(head(mtcars))
ft
}
\seealso{
\code{\link[=style]{style()}}, \code{\link[=autofit]{autofit()}}, \code{\link[=theme_booktabs]{theme_booktabs()}}, \code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=compose]{compose()}}, \code{\link[=footnote]{footnote()}}, \code{\link[=set_caption]{set_caption()}}

Other flextable constructors:
\code{\link[=as_flextable]{as_flextable()}}
}
\concept{flextable_constructors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_grouped_data}
\alias{as_grouped_data}
\title{grouped data transformation}
\usage{
as_grouped_data(x, groups, columns = NULL, expand_single = TRUE)
}
\arguments{
\item{x}{dataset}

\item{groups}{columns names to be used as row separators.}

\item{columns}{columns names to keep}

\item{expand_single}{if FALSE, groups with only one
row will not be expanded with a title row. If TRUE (the
default), single row groups and multi-row groups are all
restructured.}
}
\description{
Repeated consecutive values of group columns will
be used to define the title of the groups and will
be added as a row title.
}
\examples{
# as_grouped_data -----
library(data.table)
CO2 <- CO2
setDT(CO2)
CO2$conc <- as.integer(CO2$conc)

data_co2 <- dcast(CO2, Treatment + conc ~ Type,
  value.var = "uptake", fun.aggregate = mean)
data_co2
data_co2 <- as_grouped_data(x = data_co2, groups = c("Treatment"))
data_co2
}
\seealso{
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{flextable_dim}
\alias{flextable_dim}
\title{Get overall width and height of a flextable}
\usage{
flextable_dim(x, unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{unit}{unit for returned values, one of "in", "cm", "mm".}
}
\description{
Returns the width, height and
aspect ratio of a flextable in a named list.
The aspect ratio is the ratio corresponding to \code{height/width}.

Names of the list are \code{widths}, \code{heights} and \code{aspect_ratio}.
}
\examples{
ftab <- flextable(head(iris))
flextable_dim(ftab)
ftab <- autofit(ftab)
flextable_dim(ftab)
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.glm}
\alias{as_flextable.glm}
\title{Transform a 'glm' object into a flextable}
\usage{
\method{as_flextable}{glm}(x, ...)
}
\arguments{
\item{x}{glm model}

\item{...}{unused argument}
}
\description{
produce a flextable describing a
generalized linear model produced by function \code{glm}.

You can remove significance stars by setting options
\code{options(show.signif.stars = FALSE)}.
}
\examples{
if (require("broom")) {
  dat <- attitude
  dat$high.rating <- (dat$rating > 70)
  probit.model <- glm(high.rating ~ learning + critical +
    advance, data = dat, family = binomial(link = "probit"))
  ft <- as_flextable(probit.model)
  ft
}
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{vline_left}
\alias{vline_left}
\title{set flextable left vertical borders}
\usage{
vline_left(x, i = NULL, border = NULL, part = "all")
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{border}{border properties defined by a call to \code{\link[=fp_border]{fp_border()}}}

\item{part}{partname of the table (one of 'all', 'body', 'header', 'footer')}
}
\description{
The function is applying vertical borders to the
left side of one or all parts of a flextable. The line is the
left border of selected cells of the first column.
}
\section{Illustrations}{


\if{html}{\figure{fig_vline_left_1.png}{options: width="400"}}
}

\examples{
library(officer)
std_border = fp_border(color="orange")

ft <- flextable(head(iris))
ft <- border_remove(x = ft)

# add vertical border on the left side of the table
ft <- vline_left(ft, border = std_border )
ft
}
\seealso{
Other borders management: 
\code{\link{border_inner_h}()},
\code{\link{border_inner_v}()},
\code{\link{border_inner}()},
\code{\link{border_outer}()},
\code{\link{border_remove}()},
\code{\link{hline_bottom}()},
\code{\link{hline_top}()},
\code{\link{hline}()},
\code{\link{surround}()},
\code{\link{vline_right}()},
\code{\link{vline}()}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{vline_right}
\alias{vline_right}
\title{Set the right border of the table}
\usage{
vline_right(x, i = NULL, border = NULL, part = "all")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{border}{border properties defined by a call to
\code{\link[officer:fp_border]{officer::fp_border()}}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{vline_right()} draws a vertical line along the \strong{right edge}
of the table by setting the right border of the last column.
It does not accept a column selector \code{j} because it always
targets the last column.

An optional row selector \code{i} lets you restrict the line to
specific rows.

Unlike \code{\link[=vline]{vline()}}, which adds inner lines to the right of
arbitrary columns, \code{vline_right()} is meant for the outer
right edge of the table.
}
\examples{
library(officer)
std_border <- fp_border(color = "orange")

ft <- flextable(head(iris))
ft <- border_remove(x = ft)

# add a border on the right edge of the table
ft <- vline_right(ft, border = std_border)
ft
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_h]{border_inner_h()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=border_remove]{border_remove()}},
\code{\link[=hline]{hline()}},
\code{\link[=hline_bottom]{hline_bottom()}},
\code{\link[=hline_top]{hline_top()}},
\code{\link[=surround]{surround()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compact_summary.R
\name{compact_summary}
\alias{compact_summary}
\title{Compact Summary of a Dataset}
\usage{
compact_summary(x, show_type = FALSE, show_na = FALSE, max_levels = 10L)
}
\arguments{
\item{x}{A data.frame.}

\item{show_type}{If \code{TRUE}, a \emph{Type} column is added when the
object is rendered as a flextable.}

\item{show_na}{If \code{TRUE}, a \emph{NA} column showing the count of
missing values per column is added when rendered as a
flextable.}

\item{max_levels}{Maximum number of levels or unique values
displayed for factor and character columns. Additional
values are replaced by \code{", ..."}.}
}
\value{
A data.frame with additional class \code{"compact_summary"}.
}
\description{
Produces a compact summary of a data.frame. Each row of the
result describes one column of the input with a type-specific
synopsis.

Supported types and their summaries:
\itemize{
\item \strong{numeric / integer}: count of non-\code{NA} values, min and max.
\item \strong{character}: number of unique values, first values listed.
\item \strong{factor}: number of levels, levels listed.
\item \strong{logical}: count of non-\code{NA} values, counts of \code{TRUE} and
\code{FALSE}.
\item \strong{Date}: count of non-\code{NA} values, date range.
\item \strong{POSIXct / POSIXlt}: count of non-\code{NA} values, datetime
range.
\item \strong{hms / difftime}: count of non-\code{NA} values, time range.
}

Character and factor columns share the same summary layout
but report a different type label.

The result has class \code{"compact_summary"} and can be converted
into a flextable with \code{\link[=as_flextable]{as_flextable()}}.
}
\examples{
z <- compact_summary(iris)
as_flextable(z)

z <- compact_summary(iris, show_type = TRUE, show_na = TRUE)
as_flextable(z)
}
\seealso{
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}}

Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable_rtables.R
\name{as_flextable.TableTree}
\alias{as_flextable.TableTree}
\alias{as_flextable.ElementaryTable}
\title{Transform an rtables object into a flextable}
\usage{
\method{as_flextable}{TableTree}(x, indent_padding = 4, ...)

\method{as_flextable}{ElementaryTable}(x, indent_padding = 4, ...)
}
\arguments{
\item{x}{a \code{TableTree} or \code{ElementaryTable} object produced
by the 'rtables' package.}

\item{indent_padding}{base left padding in points per indentation
level. Default is \code{4}.}

\item{...}{unused arguments}
}
\value{
a flextable object.
}
\description{
produce a flextable from a \code{TableTree} or
\code{ElementaryTable} object produced with the
'rtables' package.

The conversion uses \code{\link[formatters:matrix_form]{formatters::matrix_form()}} to extract the
formatted content, column spans, alignments, indentation
and footnotes, then maps them to flextable features.

Indentation of row labels is rendered with left padding.

Label rows (\code{LabelRow}) are displayed in bold in the first column.
Content rows (\code{ContentRow}) are displayed entirely in bold.

When \code{LabelRow} groups exist, \code{\link[=paginate]{paginate()}} is applied so
that all rows belonging to the same group are kept together
on the same page in Word and RTF output.

To paginate the resulting flextable into multiple pages, use
\code{\link[=split_to_pages]{split_to_pages()}}, \code{\link[=split_rows]{split_rows()}}, or \code{\link[=split_columns]{split_columns()}} after
calling \code{as_flextable()}.
}
\examples{
if (require("rtables", character.only = TRUE, quietly = TRUE)) {
  library(rtables)

  lyt <- basic_table(title = "Demographic Summary") \%>\%
    split_cols_by("ARM") \%>\%
    split_rows_by("SEX") \%>\%
    analyze("AGE", afun = mean, format = "xx.x")

  tbl <- build_table(lyt, DM)
  as_flextable(tbl)
}
}
\seealso{
\code{\link[=split_to_pages]{split_to_pages()}}, \code{\link[=split_rows]{split_rows()}}, \code{\link[=split_columns]{split_columns()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{style}
\alias{style}
\title{Set formatting properties on a flextable selection}
\usage{
style(
  x,
  i = NULL,
  j = NULL,
  pr_t = NULL,
  pr_p = NULL,
  pr_c = NULL,
  part = "body"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{pr_t}{an \code{\link[officer:fp_text]{officer::fp_text()}} or \code{\link[officer:fp_text_lite]{officer::fp_text_lite()}} object
defining text formatting (font, size, color, bold, ...).}

\item{pr_p}{an \code{\link[officer:fp_par]{officer::fp_par()}} or \code{\link[officer:fp_par_lite]{officer::fp_par_lite()}} object
defining paragraph formatting (alignment, padding, line spacing, ...).}

\item{pr_c}{an \code{\link[officer:fp_cell]{officer::fp_cell()}} object defining cell formatting
(background, borders, vertical alignment, ...).}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{style()} applies text, paragraph and cell formatting properties
to a selection of rows and columns in one call. It is a lower-level
function that bundles what the convenience functions do individually:
\itemize{
\item text (\code{pr_t}): font family, size, color, bold, italic, etc.,
same properties as \code{\link[=bold]{bold()}}, \code{\link[=italic]{italic()}}, \code{\link[=color]{color()}}, \code{\link[=fontsize]{fontsize()}},
\code{\link[=font]{font()}}, \code{\link[=highlight]{highlight()}}.
\item paragraph (\code{pr_p}): alignment, padding, line spacing,
same properties as \code{\link[=align]{align()}}, \code{\link[=padding]{padding()}}, \code{\link[=line_spacing]{line_spacing()}}.
\item cell (\code{pr_c}): background colour, borders, vertical alignment,
same properties as \code{\link[=bg]{bg()}}, \code{\link[=border]{border()}}, \code{\link[=valign]{valign()}}.
}

Use \code{style()} when you need to set several property types at once
on the same selection; use the convenience functions when you
only need to change one aspect.

Note that \code{style()} modifies an existing flextable.
To change the initial formatting applied to every new
flextable, use \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} instead.
}
\examples{
library(officer)
def_cell <- fp_cell(border = fp_border(color = "wheat"))

def_par <- fp_par(text.align = "center")

ft <- flextable(head(mtcars))

ft <- style(ft, pr_c = def_cell, pr_p = def_par, part = "all")
ft <- style(ft, ~ drat > 3.5, ~ vs + am + gear + carb,
  pr_t = fp_text(color = "red", italic = TRUE)
)

ft
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{border}
\alias{border}
\title{Set cell borders}
\usage{
border(
  x,
  i = NULL,
  j = NULL,
  border = NULL,
  border.top = NULL,
  border.bottom = NULL,
  border.left = NULL,
  border.right = NULL,
  part = "body"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{border}{border (shortcut for top, bottom, left and right)}

\item{border.top}{border top}

\item{border.bottom}{border bottom}

\item{border.left}{border left}

\item{border.right}{border right}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
change borders of selected rows and columns of a flextable.
This function is \strong{not to be used by end user} (it requires careful
settings to avoid overlapping borders) but only for programming purposes.

If you need to add borders, use instead other functions:

These set borders for the whole table : \code{\link[=border_outer]{border_outer()}},
\code{\link[=border_inner_h]{border_inner_h()}} and \code{\link[=border_inner_v]{border_inner_v()}}.

To add horizontal or vertical lines in the table at specific location,
use:
\itemize{
\item \code{\link[=hline]{hline()}}: set bottom borders (inner horizontal)
\item \code{\link[=vline]{vline()}}: set right borders (inner vertical)
}

To add following horizontal or vertical lines at
beginning or end of the table, use:
\itemize{
\item \code{\link[=hline_top]{hline_top()}}: set the top border (outer horizontal)
\item \code{\link[=hline_bottom]{hline_bottom()}}: set the bottom border (outer horizontal)
\item \code{\link[=vline_left]{vline_left()}}: set the left border (outer vertical)
\item \code{\link[=vline_right]{vline_right()}}: set the right border (outer vertical)
}

If you want to highlight specific cells with some borders, use
\code{\link[=surround]{surround()}}.
}
\note{
pdf and pptx outputs do not support \code{border()} usage.
}
\examples{
library(officer)
ftab <- flextable(head(mtcars))
ftab <- border(ftab, border.top = fp_border(color = "orange"))
ftab
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{as_image}
\alias{as_image}
\title{image chunk wrapper}
\usage{
as_image(src, width = NULL, height = NULL, unit = "in", guess_size = TRUE, ...)
}
\arguments{
\item{src}{image filename}

\item{width, height}{size of the image file. It can be ignored
if parameter \code{guess_size=TRUE}, see parameter \code{guess_size}.}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{guess_size}{If package 'magick' is installed, this option
can be used (set it to \code{TRUE} and don't provide values for paramters
\code{width} and \code{height}). When the flextable will be printed,
the images will be read and width and height will be guessed. This
should be avoid if possible as it can be an extensive task when
several images.}

\item{...}{unused argument}
}
\description{
The function lets add images within flextable
objects with functions:
\itemize{
\item \code{\link[=compose]{compose()}} and \code{\link[=as_paragraph]{as_paragraph()}},
\item \code{\link[=append_chunks]{append_chunks()}},
\item \code{\link[=prepend_chunks]{prepend_chunks()}}
\if{html}{\out{
<img src="https://www.ardata.fr/img/flextable-imgs/flextable-006.png" alt="as_image illustration" style="width:100\%;">
}}
}
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\examples{
img.file <- file.path( R.home("doc"),
  "html", "logo.jpg" )
if (require("magick")) {
  myft <- flextable( head(iris))
  myft <- compose( myft, i = 1:3, j = 1,
   value = as_paragraph(
     as_image(src = img.file),
     " ",
     as_chunk(Sepal.Length,
       props = fp_text_default(color = "red"))
   ),
   part = "body")
  ft <- autofit(myft)
  ft
}
}
\seealso{
\code{\link[=compose]{compose()}}, \code{\link[=as_paragraph]{as_paragraph()}}

Other chunk elements for paragraph: 
\code{\link{as_bracket}()},
\code{\link{as_b}()},
\code{\link{as_chunk}()},
\code{\link{as_equation}()},
\code{\link{as_highlight}()},
\code{\link{as_i}()},
\code{\link{as_sub}()},
\code{\link{as_sup}()},
\code{\link{as_word_field}()},
\code{\link{colorize}()},
\code{\link{gg_chunk}()},
\code{\link{grid_chunk}()},
\code{\link{hyperlink_text}()},
\code{\link{linerange}()},
\code{\link{lollipop}()},
\code{\link{minibar}()},
\code{\link{plot_chunk}()}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{padding}
\alias{padding}
\title{Set paragraph paddings}
\usage{
padding(
  x,
  i = NULL,
  j = NULL,
  padding = NULL,
  padding.top = NULL,
  padding.bottom = NULL,
  padding.left = NULL,
  padding.right = NULL,
  part = "body"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{padding}{padding (shortcut for top, bottom, left and right), unit is pts (points).}

\item{padding.top}{padding top, unit is pts (points).}

\item{padding.bottom}{padding bottom, unit is pts (points).}

\item{padding.left}{padding left, unit is pts (points).}

\item{padding.right}{padding right, unit is pts (points).}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Change the padding of selected rows and columns of a flextable.
}
\note{
In PDF output, only \code{padding.left} and \code{padding.right} are supported.
\code{padding.top} and \code{padding.bottom} are ignored due to LaTeX limitations.
Global horizontal spacing can also be set with
\code{set_table_properties(opts_pdf = list(tabcolsep = 1))}.
}
\examples{
ft_1 <- flextable(head(iris))
ft_1 <- theme_vader(ft_1)
ft_1 <- padding(ft_1, padding.top = 4, part = "all")
ft_1 <- padding(ft_1, j = 1, padding.right = 40)
ft_1 <- padding(ft_1, i = 3, padding.top = 40)
ft_1 <- padding(ft_1, padding.top = 10, part = "header")
ft_1 <- padding(ft_1, padding.bottom = 10, part = "header")
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{colformat_image}
\alias{colformat_image}
\title{format cells as images}
\usage{
colformat_image(
  x,
  i = NULL,
  j = NULL,
  width,
  height,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
)
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{columns selection.}

\item{width, height}{size of the png file in inches}

\item{na_str, nan_str}{string to be used for NA and NaN values}

\item{prefix, suffix}{string to be used as prefix or suffix}
}
\description{
Format image paths as images in a flextable.
}
\section{Illustrations}{


\if{html}{\figure{fig_colformat_image_1.png}{options: width="400"}}
}

\examples{
img.file <- file.path( R.home("doc"), "html", "logo.jpg" )

dat <- head(iris)
dat$Species <- as.character(dat$Species)
dat[c(1, 3, 5), "Species"] <- img.file

myft <- flextable( dat)
myft <- colformat_image(
  myft, i = c(1, 3, 5),
  j = "Species", width = .20, height = .15)
ft <- autofit(myft)
ft
}
\seealso{
Other cells formatters: 
\code{\link{colformat_char}()},
\code{\link{colformat_datetime}()},
\code{\link{colformat_date}()},
\code{\link{colformat_double}()},
\code{\link{colformat_int}()},
\code{\link{colformat_lgl}()},
\code{\link{colformat_num}()},
\code{\link{set_formatter}()}
}
\concept{cells formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{hline_top}
\alias{hline_top}
\title{Set the top border of a table part}
\usage{
hline_top(x, j = NULL, border = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{border}{border properties defined by a call to
\code{\link[officer:fp_border]{officer::fp_border()}}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{hline_top()} draws a horizontal line at the \strong{very top} of a
table part. It does not accept a row selector \code{i} because it
always targets the first row.

When the part above exists (e.g. header above body), the line
is stored as the bottom border of that adjacent part so that
it renders seamlessly.

Unlike \code{\link[=hline]{hline()}}, which adds inner lines below arbitrary rows,
\code{hline_top()} is meant for the outer top edge of a part.
}
\examples{
library(officer)
big_border <- fp_border(color = "orange", width = 3)

ft <- flextable(head(iris))
ft <- border_remove(x = ft)

# add a thick line on top of each part
ft <- hline_top(ft, part = "all", border = big_border)
ft
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_h]{border_inner_h()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=border_remove]{border_remove()}},
\code{\link[=hline]{hline()}},
\code{\link[=hline_bottom]{hline_bottom()}},
\code{\link[=surround]{surround()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}},
\code{\link[=vline_right]{vline_right()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_str.R
\name{flextable_html_dependency}
\alias{flextable_html_dependency}
\title{htmlDependency for flextable objects}
\usage{
flextable_html_dependency()
}
\description{
When using loops in an R Markdown for HTML document, the
htmlDependency object for flextable must also be added at least once.
}
\examples{
if(require("htmltools"))
  div(flextable_html_dependency())
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{fp_border_default}
\alias{fp_border_default}
\title{Create border formatting with flextable defaults}
\usage{
fp_border_default(
  color = flextable_global$defaults$border.color,
  style = "solid",
  width = flextable_global$defaults$border.width
)
}
\arguments{
\item{color}{border color - single character value (e.g. "#000000" or "black")}

\item{style}{border style - single character value : See Details for supported
border styles.}

\item{width}{border width - an integer value : 0>= value}
}
\description{
Create a \code{\link[officer:fp_border]{officer::fp_border()}} object whose unspecified
arguments inherit from \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} (border color, width).

Use \code{fp_border_default()} instead of \code{officer::fp_border()} so that
borders automatically match the table's default style.
}
\examples{
library(flextable)

set_flextable_defaults(
  border.color = "orange"
)

z <- flextable(head(cars))
z <- theme_vanilla(z)
z <- vline(
  z,
  j = 1, part = "all",
  border = officer::fp_border()
)
z <- vline(
  z,
  j = 2, part = "all",
  border = fp_border_default()
)
z

init_flextable_defaults()

}
\seealso{
\code{\link[=hline]{hline()}}, \code{\link[=vline]{vline()}}

Other default formatting properties:
\code{\link[=fp_text_default]{fp_text_default()}}
}
\concept{default_formatting_properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz-defunct.R
\name{as_raster}
\alias{as_raster}
\title{Transform a flextable into a raster}
\usage{
as_raster(...)
}
\arguments{
\item{...}{unused}
}
\description{
Function is now defunct, use \code{\link[=gen_grob]{gen_grob()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{bg}
\alias{bg}
\title{Set background color}
\usage{
bg(x, i = NULL, j = NULL, bg, part = "body", source = j)
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{columns selection}

\item{bg}{color to use as background color. If a function, function need to return
a character vector of colors.}

\item{part}{partname of the table (one of 'all', 'body', 'header', 'footer')}

\item{source}{if bg is a function, source is specifying the dataset column to be used
as argument to \code{bg}. This is only useful if j is colored with values contained in
other columns.}
}
\description{
Change background color of selected rows and
columns of a flextable. A function can be used instead of
fixed colors.

When \code{bg} is a function, it is possible to color cells based on values
located in other columns, using hidden columns (those not used by
argument \code{colkeys}) is a common use case. The argument \code{source}
has to be used to define what are the columns to be used for the color
definition and the argument \code{j} has to be used to define where to apply
the colors and only accept values from \code{colkeys}.
}
\note{
Word does not allow you to apply transparency to table cells or paragraph shading.
}
\section{Illustrations}{


\if{html}{\figure{fig_bg_1.png}{options: width="400"}}

\if{html}{\figure{fig_bg_2.png}{options: width="300"}}
}

\examples{
ft_1 <- flextable(head(mtcars))
ft_1 <- bg(ft_1, bg = "wheat", part = "header")
ft_1 <- bg(ft_1, i = ~ qsec < 18, bg = "#EFEFEF", part = "body")
ft_1 <- bg(ft_1, j = "drat", bg = "#606060", part = "all")
ft_1 <- color(ft_1, j = "drat", color = "white", part = "all")
ft_1

if (require("scales")) {
  ft_2 <- flextable(head(iris))
  colourer <- col_numeric(
    palette = c("wheat", "red"),
    domain = c(0, 7)
  )
  ft_2 <- bg(ft_2,
    j = c(
      "Sepal.Length", "Sepal.Width",
      "Petal.Length", "Petal.Width"
    ),
    bg = colourer, part = "body"
  )
  ft_2
}
}
\seealso{
Other sugar functions for table style: 
\code{\link{align}()},
\code{\link{bold}()},
\code{\link{color}()},
\code{\link{empty_blanks}()},
\code{\link{fontsize}()},
\code{\link{font}()},
\code{\link{highlight}()},
\code{\link{italic}()},
\code{\link{line_spacing}()},
\code{\link{padding}()},
\code{\link{rotate}()},
\code{\link{valign}()}
}
\concept{sugar functions for table style}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{border_inner_h}
\alias{border_inner_h}
\title{Set inner horizontal borders}
\usage{
border_inner_h(x, border = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{border}{border properties defined by a call to \code{\link[officer:fp_border]{officer::fp_border()}}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
The function is applying a border to inner content of one
or all parts of a flextable.
}
\examples{
library(officer)
std_border <- fp_border(color = "orange", width = 1)

dat <- iris[c(1:5, 51:55, 101:105), ]
ft <- flextable(dat)
ft <- border_remove(x = ft)

# add inner horizontal borders
ft <- border_inner_h(ft, border = std_border)
ft
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=border_remove]{border_remove()}},
\code{\link[=hline]{hline()}},
\code{\link[=hline_bottom]{hline_bottom()}},
\code{\link[=hline_top]{hline_top()}},
\code{\link[=surround]{surround()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}},
\code{\link[=vline_right]{vline_right()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clintables.R
\name{shift_table}
\alias{shift_table}
\title{Create a shift table}
\usage{
shift_table(
  x,
  cn_visit = "VISIT",
  cn_visit_num = "VISITNUM",
  cn_grade = "LBNRIND",
  cn_usubjid = "USUBJID",
  cn_lab_cat = NA_character_,
  cn_is_baseline = "LBBLFL",
  baseline_identifier = "Y",
  cn_treatment = NA_character_,
  grade_levels = c("LOW", "NORMAL", "HIGH"),
  grade_labels = c("Low", "Normal", "High")
)
}
\arguments{
\item{x}{Laboratory Tests Results data frame.}

\item{cn_visit}{column name containing visit names, default to "VISIT".}

\item{cn_visit_num}{column name containing visit numbers, default to "VISITNUM".}

\item{cn_grade}{column name containing reference range indicators, default to "LBNRIND".}

\item{cn_usubjid}{column name containing unique subject inditifiers, default to "USUBJID".}

\item{cn_lab_cat}{column name containing lab tests or examination names, default to "LBTEST".}

\item{cn_is_baseline}{column name containing baseline flags, default to "LBBLFL".}

\item{baseline_identifier}{baseline flag value to use for baseline
identification. Its default is "Y".}

\item{cn_treatment}{column name containing treatment names, default to \code{NA}.}

\item{grade_levels}{levels to use for reference range indicators}

\item{grade_labels}{labels to use for reference range indicators}
}
\value{
the shift table as a data.frame. Additionnal elements are provided
in attributes:
\itemize{
\item "VISIT_N": count of unique subject id per visits, labs and eventually
treatments. This element is supposed to be used as value for argument
\code{hidden_data} of function \code{tabulator()}.
\item "FUN_VISIT": a utility function to easily turn \emph{visit} column as a factor
column. It should be applied after the shift table creation.
\item "FUN_GRADE": a utility function to easily turn \emph{grade} column as a factor
column. It adds "MISSING/Missing" and "SUM/Sum" at the end of the set of values specified
in arguments \code{grade_levels} and \code{grade_labels}. It should be applied after the shift
table creation.
}
}
\description{
Create a shift table ready to be used with \code{tabulator()}.

The function is transforming a dataset representing some
'Laboratory Tests Results' structured as \emph{CDISC clinical trial
data sets} format to a dataset representing the shift table.

Shift tables are tables used in clinical trial analysis.
They show the progression of change from the baseline, with the progression
often being along time; the number of subjects is displayed in different
range (e.g. low, normal, or high) at baseline and at selected time points
or intervals.
}
\examples{
\dontshow{
data.table::setDTthreads(1)
}
library(data.table)
library(flextable)

# data simulation ----
USUBJID <- sprintf("01-ABC-\%04.0f", 1:200)
VISITS <- c("SCREENING 1", "WEEK 2", "MONTH 3")
LBTEST <- c("Albumin", "Sodium")

VISITNUM <- seq_along(VISITS)
LBBLFL <- rep(NA_character_, length(VISITNUM))
LBBLFL[1] <- "Y"

VISIT <- data.frame(
  VISIT = VISITS, VISITNUM = VISITNUM,
  LBBLFL = LBBLFL, stringsAsFactors = FALSE
)
labdata <- expand.grid(
  USUBJID = USUBJID, LBTEST = LBTEST,
  VISITNUM = VISITNUM,
  stringsAsFactors = FALSE
)
setDT(labdata)

labdata <- merge(labdata, VISIT, by = "VISITNUM")

subject_elts <- unique(labdata[, .SD, .SDcols = "USUBJID"])
subject_elts <- unique(subject_elts)
subject_elts[, c("TREAT") := list(
  sample(x = c("Treatment", "Placebo"), size = .N, replace = TRUE)
)]
subject_elts[, c("TREAT") := list(
  factor(.SD$TREAT, levels = c("Treatment", "Placebo"))
)]
setDF(subject_elts)
labdata <- merge(labdata, subject_elts,
  by = "USUBJID", all.x = TRUE, all.y = FALSE
)
labdata[, c("LBNRIND") := list(
  sample(
    x = c("LOW", "NORMAL", "HIGH"), size = .N,
    replace = TRUE, prob = c(.03, .9, .07)
  )
)]

setDF(labdata)




# shift table calculation ----

SHIFT_TABLE <- shift_table(
  x = labdata, cn_visit = "VISIT",
  cn_grade = "LBNRIND",
  cn_usubjid = "USUBJID",
  cn_lab_cat = "LBTEST",
  cn_treatment = "TREAT",
  cn_is_baseline = "LBBLFL",
  baseline_identifier = "Y",
  grade_levels = c("LOW", "NORMAL", "HIGH")
)

# get attrs for post treatment ----
SHIFT_TABLE_VISIT <- attr(SHIFT_TABLE, "VISIT_N")
visit_as_factor <- attr(SHIFT_TABLE, "FUN_VISIT")
range_as_factor <- attr(SHIFT_TABLE, "FUN_GRADE")

# post treatments ----
SHIFT_TABLE$VISIT <- visit_as_factor(SHIFT_TABLE$VISIT)
SHIFT_TABLE$BASELINE <- range_as_factor(SHIFT_TABLE$BASELINE)
SHIFT_TABLE$LBNRIND <- range_as_factor(SHIFT_TABLE$LBNRIND)

SHIFT_TABLE_VISIT$VISIT <- visit_as_factor(SHIFT_TABLE_VISIT$VISIT)

# tabulator ----

my_format <- function(z) {
  formatC(z * 100,
    digits = 1, format = "f",
    flag = "0", width = 4
  )
}

tab <- tabulator(
  x = SHIFT_TABLE,
  hidden_data = SHIFT_TABLE_VISIT,
  row_compose = list(
    VISIT = as_paragraph(VISIT, "\n(N=", N_VISIT, ")")
  ),
  rows = c("LBTEST", "VISIT", "BASELINE"),
  columns = c("TREAT", "LBNRIND"),
  `n` = as_paragraph(N),
  `\%` = as_paragraph(as_chunk(PCT, formatter = my_format))
)

# as_flextable ----

ft_1 <- as_flextable(
  x = tab, separate_with = "VISIT",
  label_rows = c(
    LBTEST = "Lab Test", VISIT = "Visit",
    BASELINE = "Reference Range Indicator"
  )
)

ft_1
}
\concept{tools for clinical reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf_str.R
\name{rtf_add.flextable}
\alias{rtf_add.flextable}
\title{Add a 'flextable' into an RTF document}
\usage{
\method{rtf_add}{flextable}(x, value, ...)
}
\arguments{
\item{x}{rtf object, created by \code{\link[officer:rtf_doc]{officer::rtf_doc()}}.}

\item{value}{a flextable object}

\item{...}{unused arguments}
}
\description{
\code{\link[officer:rtf_add]{officer::rtf_add()}} method for adding
flextable objects into 'RTF' documents.
}
\examples{
library(flextable)
library(officer)

ft <- flextable(head(iris))
ft <- autofit(ft)

z <- rtf_doc()
z <- rtf_add(z, ft)

print(z, target = tempfile(fileext = ".rtf"))
}
\seealso{
Other functions for officer integration:
\code{\link[=body_add_flextable]{body_add_flextable()}},
\code{\link[=body_replace_flextable_at_bkm]{body_replace_flextable_at_bkm()}},
\code{\link[=ph_with.flextable]{ph_with.flextable()}}
}
\concept{officer_integration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{add_header_row}
\alias{add_header_row}
\title{Add a header row with spanning labels}
\usage{
add_header_row(x, top = TRUE, values = character(0), colwidths = integer(0))
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{top}{should the row be inserted at the
top or the bottom. Default to TRUE.}

\item{values}{values to add, a character vector (as header rows
contains only character values/columns), a list
or a call to \code{\link[=as_paragraph]{as_paragraph()}}.}

\item{colwidths}{the number of columns used for each label}
}
\description{
Add a single row to the header where labels can span
multiple columns (merged cells) via the \code{colwidths} argument.

Labels are associated with a number of columns
to merge that default to one if not specified.
In this case, you have to make sure that the
number of labels is equal to the number of columns displayed.

The function can add only one single row by call.

Labels can also be formatted with \code{\link[=as_paragraph]{as_paragraph()}}.
}
\examples{
library(flextable)

ft01 <- fp_text_default(color = "red")
ft02 <- fp_text_default(color = "orange")

pars <- as_paragraph(
  as_chunk(c("(1)", "(2)"), props = ft02), " ",
  as_chunk(c(
    "My tailor is rich",
    "My baker is rich"
  ), props = ft01)
)

ft_1 <- flextable(head(mtcars))
ft_1 <- add_header_row(ft_1,
  values = pars,
  colwidths = c(5, 6), top = FALSE
)
ft_1 <- add_header_row(ft_1,
  values = pars,
  colwidths = c(3, 8), top = TRUE
)
ft_1

ft_2 <- flextable(head(airquality))
ft_2 <- add_header_row(ft_2,
  values = c("Measure", "Time"),
  colwidths = c(4, 2), top = TRUE
)
ft_2 <- theme_box(ft_2)
ft_2
}
\seealso{
\code{\link[=flextable]{flextable()}}, \code{\link[=set_caption]{set_caption()}}

Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{border_inner_v}
\alias{border_inner_v}
\title{Set inner vertical borders}
\usage{
border_inner_v(x, border = NULL, part = "all")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{border}{border properties defined by a call to \code{\link[officer:fp_border]{officer::fp_border()}}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
The function is applying a vertical border to inner content of one
or all parts of a flextable.
}
\examples{
library(officer)
std_border <- fp_border(color = "orange", width = 1)

dat <- iris[c(1:5, 51:55, 101:105), ]
ft <- flextable(dat)
ft <- border_remove(x = ft)

# add inner vertical borders
ft <- border_inner_v(ft, border = std_border)
ft
}
\seealso{
Other borders management: 
\code{\link{border_inner}()},
\code{\link{border_inner_h}()},
\code{\link{border_outer}()},
\code{\link{border_remove}()},
\code{\link{hline}()},
\code{\link{hline_bottom}()},
\code{\link{hline_top}()},
\code{\link{surround}()},
\code{\link{vline}()},
\code{\link{vline_left}()},
\code{\link{vline_right}()}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_signif_after_zeros}
\alias{fmt_signif_after_zeros}
\title{Format with significant figures after zeros}
\usage{
fmt_signif_after_zeros(x, digits = 3)
}
\arguments{
\item{x}{numeric values}

\item{digits}{number of digits displayed after the leading zeros}
}
\description{
Rounds significant figures after zeros in numeric vectors.
The number of digits displayed after the leading zeros is
customizable using the \code{digits} parameter.
}
\examples{
x <- data.frame(
  x = c(0.00000004567, 2.000003456, 3, pi)
)
ft_1 <- flextable(x)
ft_1 <- align(x = ft_1, j = 1, align = "left")
mk_par(ft_1, value = as_paragraph(
  fmt_signif_after_zeros(x)))
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other value formatters:
\code{\link[=fmt_2stats]{fmt_2stats()}},
\code{\link[=fmt_avg_dev]{fmt_avg_dev()}},
\code{\link[=fmt_dbl]{fmt_dbl()}},
\code{\link[=fmt_header_n]{fmt_header_n()}},
\code{\link[=fmt_int]{fmt_int()}},
\code{\link[=fmt_n_percent]{fmt_n_percent()}},
\code{\link[=fmt_pct]{fmt_pct()}}
}
\concept{value formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table-properties.R
\name{set_table_properties}
\alias{set_table_properties}
\title{Set table layout and width properties}
\usage{
set_table_properties(
  x,
  layout = "fixed",
  width = 0,
  align = NULL,
  opts_html = list(),
  opts_word = list(),
  opts_pdf = list(),
  word_title = NULL,
  word_description = NULL
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{layout}{'autofit' or 'fixed' algorithm. Default to 'fixed'.}

\item{width}{The parameter has a different effect depending on the
output format. Users should consider it as a minimum width.
In HTML, it is the minimum width of the space that the
table should occupy. In Word, it is a preferred size and Word
may decide not to strictly stick to it. It has no effect on
PowerPoint and PDF output. Its default value is 0, as an effect, it
only use necessary width to display all content. It is not used by the
PDF output.}

\item{align}{alignment in document (only Word, HTML and PDF),
supported values are 'left', 'center' and 'right'.}

\item{opts_html}{html options as a list. Supported elements are:
\itemize{
\item 'extra_css': extra css instructions to be integrated with the HTML
code of the table.
\item 'scroll': NULL or a list if you want to add a scroll-box.
\itemize{
\item Use an empty list to add an horizontal scroll.  The with
is fixed, corresponding to the container's width.
\item If the list has a value named \code{height} it will be used as
height and the scroll will happen also vertically. The height
will be in pixel if numeric, if a string it should be a valid css
measure.
\item If the list has a value named \code{freeze_first_column} set to \code{TRUE},
the first column is set as a \emph{sticky} column.
\item If the list has a value named \code{add_css} it will be used as extra
css to add, .i.e: \verb{border:1px solid red;}.
}
\item 'extra_class': extra classes to add in the table tag
}}

\item{opts_word}{Word options as a list. Supported elements are:
\itemize{
\item 'split':  Word option 'Allow row to break across pages' can be
activated when TRUE.
\item 'keep_with_next': Word option 'keep rows together' is
activated when TRUE. It avoids page break within tables. This
is handy for small tables, i.e. less than a page height.
\item 'repeat_headers': Word option 'Repeat as header row' is
activated and associated to header rows when TRUE.
}}

\item{opts_pdf}{PDF options as a list. Supported elements are:
\itemize{
\item 'tabcolsep': space between the text and the left/right border of its containing
cell.
\item 'arraystretch': height of each row relative to its default
height, the default value is 1.5.
\item 'float': type of floating placement in the PDF document, one of:
\itemize{
\item 'none' (the default value), table is placed after the preceding
paragraph.
\item 'float', table can float to a place in the text where it fits best
\item 'wrap-r', wrap text around the table positioned to the right side of the text
\item 'wrap-l', wrap text around the table positioned to the left side of the text
\item 'wrap-i', wrap text around the table positioned inside edge-near the binding
\item 'wrap-o', wrap text around the table positioned outside edge-far from the binding
}
\item 'fonts_ignore': if TRUE, pdf-engine 'pdflatex' can be used instead of
'xelatex' or 'lualatex.' If pdflatex is used, fonts will be ignored because they are
not supported by pdflatex, whereas with the xelatex and lualatex engines they are.
\item 'caption_repeat': a boolean that indicates if the caption should be
repeated along pages. Its default value is \code{TRUE}.
\item 'footer_repeat': a boolean that indicates if the footer should be
repeated along pages. Its default value is \code{TRUE}.
\item 'default_line_color': default line color, restored globally after
the flextable is produced.
}}

\item{word_title}{alternative text for Word table (used as title of the table)}

\item{word_description}{alternative text for Word table (used as description of the table)}
}
\description{
Set table layout and table width. Default to fixed
algorithm.

If layout is fixed, column widths will be used to display the table;
\code{width} is ignored.

If layout is autofit, column widths will not be used;
table width is used (as a percentage).
}
\note{
PowerPoint output ignore 'autofit layout'.
}
\examples{
library(flextable)
ft_1 <- flextable(head(cars))
ft_1 <- autofit(ft_1)
ft_2 <- set_table_properties(ft_1, width = .5, layout = "autofit")
ft_2
ft_3 <- set_table_properties(ft_1, width = 1, layout = "autofit")

# add scroll for HTML ----
set.seed(2)
dat <- lapply(1:14, function(x) rnorm(n = 20))
dat <- setNames(dat, paste0("colname", 1:14))
dat <- as.data.frame(dat)

ft_4 <- flextable(dat)
ft_4 <- colformat_double(ft_4)
ft_4 <- bg(ft_4, j = 1, bg = "#DDDDDD", part = "all")
ft_4 <- bg(ft_4, i = 1, bg = "#DDDDDD", part = "header")
ft_4 <- autofit(ft_4)
ft_4 <- set_table_properties(
  x = ft_4,
  opts_html = list(
    scroll = list(
      height = "500px",
      freeze_first_column = TRUE
    )
  )
)
ft_4
}
\seealso{
\code{\link[=flextable]{flextable()}}, \code{\link[=as_flextable]{as_flextable()}}, \code{\link[=autofit]{autofit()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}}

Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{continuous_summary}
\alias{continuous_summary}
\title{continuous columns summary}
\usage{
continuous_summary(
  dat,
  columns = NULL,
  by = character(0),
  hide_grouplabel = TRUE,
  digits = 3
)
}
\arguments{
\item{dat}{a data.frame}

\item{columns}{continuous variables to be summarized. If NULL all
continuous variables are summarized.}

\item{by}{discrete variables to use as groups when summarizing.}

\item{hide_grouplabel}{if TRUE, group label will not be rendered, only
level/value will be rendered.}

\item{digits}{the desired number of digits after the decimal point}
}
\description{
create a data.frame summary for continuous variables
}
\examples{
ft_1 <- continuous_summary(iris, names(iris)[1:4], by = "Species",
  hide_grouplabel = FALSE)
ft_1
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{add_body}
\alias{add_body}
\title{Add column values as new lines in body}
\usage{
add_body(x, top = TRUE, ..., values = NULL)
}
\arguments{
\item{x}{a flextable object}

\item{top}{should the rows be inserted at the top or the bottom.}

\item{...}{named arguments (names are data colnames) of values
to add. It is important to insert data of the same type as the
original data, otherwise it will be transformed (probably
into strings if you add a \code{character} where a \code{double} is expected).
This makes possible to still format cell contents with the \verb{colformat_*}
functions, for example \code{\link[=colformat_num]{colformat_num()}}.}

\item{values}{a list of name-value pairs of labels or values,
names should be existing col_key values. This argument can be used
instead of \code{...} for programming purpose (If \code{values} is
supplied argument \code{...} is ignored).}
}
\description{
The function adds a list of values to be inserted as
new rows in the body. The values are inserted in
existing columns of the input data of the flextable.
Rows can be inserted at the top or the bottom of
the body.

If some columns are not provided, they will be replaced by
\code{NA} and displayed as empty.
}
\examples{
ft <- flextable(head(iris),
  col_keys = c(
    "Species", "Sepal.Length", "Petal.Length",
    "Sepal.Width", "Petal.Width"
  )
)

ft <- add_body(
  x = ft, Sepal.Length = 1:5,
  Sepal.Width = 1:5 * 2, Petal.Length = 1:5 * 3,
  Petal.Width = 1:5 + 10, Species = "Blah", top = FALSE
)

ft <- theme_booktabs(ft)
ft
}
\seealso{
\code{\link[=flextable]{flextable()}}

Other functions to add rows in a flextable: 
\code{\link{add_body_row}()},
\code{\link{add_footer_lines}()},
\code{\link{add_footer_row}()},
\code{\link{add_footer}()},
\code{\link{add_header_row}()},
\code{\link{add_header}()},
\code{\link{separate_header}()},
\code{\link{set_header_footer_df}},
\code{\link{set_header_labels}()}
}
\concept{functions to add rows in a flextable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_i}
\alias{as_i}
\title{Italic chunk}
\usage{
as_i(x)
}
\arguments{
\item{x}{value, if a chunk, the chunk will be updated}
}
\description{
The function is producing a chunk with
italic font.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
\examples{
ft <- flextable(head(iris),
  col_keys = c("Sepal.Length", "dummy")
)

ft <- compose(ft,
  j = "dummy",
  value = as_paragraph(as_i(Sepal.Length))
)

ft
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.merMod}
\alias{as_flextable.merMod}
\alias{as_flextable.lme}
\alias{as_flextable.gls}
\alias{as_flextable.nlme}
\alias{as_flextable.brmsfit}
\alias{as_flextable.glmmTMB}
\alias{as_flextable.glmmadmb}
\title{Transform a 'merMod' or 'lme' object into a flextable}
\usage{
\method{as_flextable}{merMod}(x, add.random = TRUE, ...)

\method{as_flextable}{lme}(x, add.random = TRUE, ...)

\method{as_flextable}{gls}(x, add.random = TRUE, ...)

\method{as_flextable}{nlme}(x, add.random = TRUE, ...)

\method{as_flextable}{brmsfit}(x, add.random = TRUE, ...)

\method{as_flextable}{glmmTMB}(x, add.random = TRUE, ...)

\method{as_flextable}{glmmadmb}(x, add.random = TRUE, ...)
}
\arguments{
\item{x}{a mixed model}

\item{add.random}{TRUE or FALSE, if TRUE random effects are
added to the table.}

\item{...}{unused argument}
}
\description{
produce a flextable describing a
mixed model. The function is only using package 'broom.mixed'
that provides the data presented in the resulting flextable.

You can remove significance stars by setting options
\code{options(show.signif.stars = FALSE)}.
}
\examples{
if (require("broom.mixed") && require("nlme")) {
  m1 <- lme(distance ~ age, data = Orthodont)
  ft <- as_flextable(m1)
  ft
}
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_flextable.R
\name{merge_h}
\alias{merge_h}
\title{Merge flextable cells horizontally}
\usage{
merge_h(x, i = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
Merge flextable cells horizontally when consecutive cells have
identical values. Text of formatted values are used to compare
values.
}
\examples{
library(flextable)

schedule <- data.frame(
  time = c("9h", "10h", "11h", "14h", "15h", "16h"),
  monday = c("Math", "Math", "French", "History", "Science", "French"),
  tuesday = c("English", "Math", "Art", "Math", "Math", "French"),
  wednesday = c("Science", "Math", "Science", "English", "English", "French"),
  stringsAsFactors = FALSE
)

ft <- flextable(schedule)
ft <- theme_box(ft)
ft <- merge_h(ft)
ft
}
\seealso{
Other cell merging functions:
\code{\link[=merge_at]{merge_at()}},
\code{\link[=merge_h_range]{merge_h_range()}},
\code{\link[=merge_none]{merge_none()}},
\code{\link[=merge_v]{merge_v()}}
}
\concept{cell merging functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{print.flextable}
\alias{print.flextable}
\title{Print a flextable}
\usage{
\method{print}{flextable}(x, preview = "html", align = "center", ...)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{preview}{preview type, one of c("html", "pptx", "docx", "rtf", "pdf, "log").
When \code{"log"} is used, a description of the flextable is printed.}

\item{align}{left, center (default) or right. Only for docx/html/pdf.}

\item{...}{arguments for 'pdf_document' call when preview is "pdf".}
}
\description{
print a flextable object to format \code{html}, \code{docx},
\code{pptx} or as text (not for display but for informative purpose).
This function is to be used in an interactive context.
}
\note{
When argument \code{preview} is set to \code{"docx"} or \code{"pptx"}, an
external client linked to these formats (Office is installed) is used to
edit a document. The document is saved in the temporary directory of
the R session and will be removed when R session will be ended.

When argument \code{preview} is set to \code{"html"}, an
external client linked to these HTML format is used to display the table.
If RStudio is used, the Viewer is used to display the table.

Note also that a print method is used when flextable are used within
R markdown documents. See \code{\link[=knit_print.flextable]{knit_print.flextable()}}.
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable_gam.R
\name{as_flextable.gam}
\alias{as_flextable.gam}
\title{Transform a 'gam' model into a flextable}
\usage{
\method{as_flextable}{gam}(x, ...)
}
\arguments{
\item{x}{gam model}

\item{...}{unused argument}
}
\description{
produce a flextable describing a
generalized additive model produced by function \code{mgcv::gam}.
\if{html}{\out{
<img src="https://www.ardata.fr/img/flextable-imgs/flextable-002.png" alt="as_flextable.gam illustration" style="width:100\%;">
}}
}
\examples{
if (require("mgcv")) {
  set.seed(2)

  # Simulated data
  dat <- gamSim(1, n = 400, dist = "normal", scale = 2)

  # basic GAM model
  b <- gam(y ~ s(x0) + s(x1) + s(x2) + s(x3), data = dat)

  ft <- as_flextable(b)
  ft
}
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patchwork.R
\name{wrap_flextable}
\alias{wrap_flextable}
\title{Wrap a flextable for use with patchwork}
\usage{
wrap_flextable(
  x,
  panel = c("body", "full", "rows", "cols"),
  space = c("free", "free_x", "free_y", "fixed"),
  n_row_headers = 0L,
  flex_body = FALSE,
  flex_cols = FALSE,
  expand = 0.6,
  just = c("left", "right", "center")
)
}
\arguments{
\item{x}{A flextable object.}

\item{panel}{What portion of the table should be aligned with the panel
region? \code{"body"} means that header and footer will be placed outside the
panel region. \code{"full"} means that the whole table will be placed inside the
panel region. \code{"rows"} keeps all rows inside the panel but is otherwise
equivalent to \code{"body"}. \code{"cols"} places all columns within the panel region
but keeps column headers on top.}

\item{space}{How should the dimension of the table influence the final
composition? \code{"fixed"} means that the table width and height will set the
dimensions of the area it occupies. \code{"free"} means the table dimensions will
not influence the sizing. \code{"free_x"} and \code{"free_y"} allow freeing either
direction.}

\item{n_row_headers}{Number of leading columns to treat as row headers.
These columns will be placed outside the panel region and will not
participate in alignment with the plot axes.}

\item{flex_body}{If \code{TRUE}, the table body row heights become flexible:
the adjacent ggplot determines the panel height and the body rows
stretch equally to fill it. Header and footer keep their fixed size.
This is useful to align table rows with discrete bars or categories
in a neighbouring plot. Implies \code{free_y} for \code{space}.}

\item{flex_cols}{If \code{TRUE}, the data column widths (all columns after
\code{n_row_headers}) become flexible: the adjacent ggplot determines the
panel width and the data columns stretch to fill it. Row header
columns keep their fixed width. This is useful to align table columns
with discrete x-axis categories in a neighbouring plot.
Implies \code{free_x} for \code{space}.}

\item{expand}{Expansion value matching the ggplot discrete axis expansion
(\code{ggplot2::expansion(add = expand)}). Default is \code{0.6}, which is the
ggplot2 default for discrete axes. Only used when \code{flex_cols = TRUE}.}

\item{just}{Horizontal alignment of the table within its patchwork panel.
One of \code{"left"} (default), \code{"right"}, or \code{"center"}. Useful when the
table is narrower than the available panel width.
Ignored when \code{flex_cols = TRUE} (columns fill the panel).}
}
\value{
A patchwork-compliant object that can be combined with ggplot2 plots
using \code{+}, \code{|}, or \code{/} operators.
}
\description{
This function wraps a flextable as a patchwork-compliant patch, similar
to what \code{\link[patchwork:wrap_table]{patchwork::wrap_table()}} does for gt tables. It allows flextable
objects to be combined with ggplot2 plots in a patchwork layout, with
optional alignment of table headers and body with plot panel areas.

Note this is experimental and may change in the future.
}
\examples{
\dontshow{if (requireNamespace("patchwork", quietly = TRUE) && requireNamespace("ggplot2", quietly = TRUE) && requireNamespace("ragg", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(gdtools)
font_set_liberation()
library(ggplot2)
library(patchwork)

set_flextable_defaults(
  font.family = "Liberation Sans",
  font.size = 10,
  big.mark = "",
  border.color = "grey60"
)

# Adapted from <https://r-graph-gallery.com/web-dumbell-chart.html>

dataset <- data.frame(
  team = c(
    "FC Bayern Munchen", "SV Werder Bremen", "Borussia Dortmund",
    "VfB Stuttgart", "Borussia M'gladbach", "Hamburger SV",
    "Eintracht Frankfurt", "FC Schalke 04", "1. FC Koln",
    "Bayer 04 Leverkusen"
  ),
  matches = c(2000, 1992, 1924, 1924, 1898, 1866, 1856, 1832, 1754, 1524),
  won     = c(1206,  818,  881,  782,  763,  746,  683,  700,  674,  669),
  lost    = c( 363,  676,  563,  673,  636,  625,  693,  669,  628,  447)
)
dataset$win_pct  <- dataset$won  / dataset$matches * 100
dataset$loss_pct <- dataset$lost / dataset$matches * 100
dataset$team <- factor(dataset$team, levels = rev(dataset$team))

# -- dumbbell chart --
pal <- c(lost = "#EFAC00", won = "#28A87D")
df_long <- reshape(dataset, direction = "long",
  varying = list(c("loss_pct", "win_pct")),
  v.names = "pct", timevar = "type",
  times = c("lost", "won"), idvar = "team"
)

p <- ggplot(df_long, aes(x = pct / 100, y = team)) +
  stat_summary(
    geom = "linerange", fun.min = "min", fun.max = "max",
    linewidth = .7, color = "grey60"
  ) +
  geom_point(aes(fill = type), size = 4, shape = 21,
    stroke = .8, color = "white"
  ) +
  scale_x_continuous(
    labels = scales::percent,
    expand = expansion(add = c(.02, .02))
  ) +
  scale_y_discrete(name = NULL, guide = "none") +
  scale_fill_manual(
    values = pal,
    labels = c(lost = "Lost", won = "Won")
  ) +
  labs(x = NULL, fill = NULL) +
  theme_minimal(base_family = "Liberation Sans", base_size = 10) +
  theme(
    legend.position = "top",
    legend.justification = "left",
    panel.grid.minor = element_blank(),
    panel.grid.major.y = element_blank()
  )

# -- flextable --
ft_dat <- dataset[, c("matches", "win_pct", "loss_pct", "team")]
ft_dat$team <- as.character(ft_dat$team)

ft <- flextable(ft_dat)
ft <- border_remove(ft)
ft <- bold(ft, part = "header")
ft <- colformat_double(ft, j = c("win_pct", "loss_pct"),
  digits = 1, suffix = "\%"
)
ft <- set_header_labels(ft,
  team = "Team", matches = "GP",
  win_pct = "", loss_pct = ""
)
ft <- color(ft, color = "#28A87D", j = 2)
ft <- color(ft, color = "#EFAC00", j = 3)
ft <- bold(ft, bold = TRUE, j = 2:3)
ft <- italic(ft, italic = TRUE, j = 4)
ft <- align(ft, align = "right", part = "all")
ft <- autofit(ft)

\dontshow{
cap <- ragg::agg_capture(width = 7, height = 6, units = "in", res = 150)
grDevices::dev.control("enable")
}
print(
  wrap_flextable(ft, flex_body = TRUE, just = "right") +
    p + plot_layout(widths = c(1.1, 2))
)
\dontshow{
raster <- cap()
dev.off()
plot(as.raster(raster))
init_flextable_defaults()
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{colformat_num}
\alias{colformat_num}
\title{format numeric cells}
\usage{
colformat_num(
  x,
  i = NULL,
  j = NULL,
  big.mark = get_flextable_defaults()$big.mark,
  decimal.mark = get_flextable_defaults()$decimal.mark,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = "",
  ...
)
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{columns selection.}

\item{big.mark, decimal.mark}{see \code{\link[=format]{format()}}}

\item{na_str, nan_str}{string to be used for NA and NaN values}

\item{prefix, suffix}{string to be used as prefix or suffix}

\item{...}{additional argument for function \code{\link[=format]{format()}}, \code{scientific}
and \code{digits} can not be used.}
}
\description{
Format numeric cells in a flextable.

The function is different from \code{\link[=colformat_double]{colformat_double()}} on numeric type
columns. The function uses the \code{\link[=format]{format()}} function of R on numeric
type columns. So this is normally what you see on the R console
most of the time (but scientific mode is disabled and NA are replaced).
}
\section{format call}{


Function \code{\link[=format]{format()}} is called with the following values:
\itemize{
\item \code{trim} is set to TRUE,
\item \code{scientific} is set to FALSE,
\item \code{big.mark} is set to the value of \code{big.mark} argument,
\item \code{decimal.mark} is set to the value of \code{decimal.mark} argument,
\item other arguments are passed 'as is' to the format function.
}

argument \code{digits} is ignored as it is not the same \code{digits} that users
want, this one will be used by \code{\link[=format]{format()}} and not \code{\link[=formatC]{formatC()}}.
To change the digit argument use \code{options(digits=4)} instead.

This argument will not be changed because \code{colformat_num()}
is supposed to format things roughly as what you see on the R console.

If these functions does not fit your needs, use \code{\link[=set_formatter]{set_formatter()}}
that lets you use any format function.
}

\section{Illustrations}{


\if{html}{\figure{fig_colformat_num_1.png}{options: width="400"}}
}

\examples{
dat <- mtcars
dat[2,1] <- NA
ft <- flextable(head(dat))
ft <- colformat_num(x = ft,
  big.mark=" ", decimal.mark = ",",
  na_str = "N/A")
ft <- autofit(ft)
ft
}
\seealso{
Other cells formatters: 
\code{\link{colformat_char}()},
\code{\link{colformat_datetime}()},
\code{\link{colformat_date}()},
\code{\link{colformat_double}()},
\code{\link{colformat_image}()},
\code{\link{colformat_int}()},
\code{\link{colformat_lgl}()},
\code{\link{set_formatter}()}
}
\concept{cells formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_header_n}
\alias{fmt_header_n}
\title{Format count as '(N=XX)' for column headers}
\usage{
fmt_header_n(n, newline = TRUE)
}
\arguments{
\item{n}{count values}

\item{newline}{indicates to prefix the text with a new line
(sof return).}
}
\description{
Format counts as \verb{\\n(N=XX)} for appending
sample sizes to column titles.
}
\examples{
library(flextable)

df <- data.frame(zz = 1)

ft_1 <- flextable(df)
ft_1 <- append_chunks(
  x = ft_1, j = 1, part = "header",
  value = as_chunk(fmt_header_n(200))
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other value formatters:
\code{\link[=fmt_2stats]{fmt_2stats()}},
\code{\link[=fmt_avg_dev]{fmt_avg_dev()}},
\code{\link[=fmt_dbl]{fmt_dbl()}},
\code{\link[=fmt_int]{fmt_int()}},
\code{\link[=fmt_n_percent]{fmt_n_percent()}},
\code{\link[=fmt_pct]{fmt_pct()}},
\code{\link[=fmt_signif_after_zeros]{fmt_signif_after_zeros()}}
}
\concept{value formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable}
\alias{as_flextable}
\title{Method to transform objects into flextables}
\usage{
as_flextable(x, ...)
}
\arguments{
\item{x}{object to be transformed as flextable}

\item{...}{arguments for custom methods}
}
\description{
This is a convenient function
to let users create flextable bindings
from any objects. Users should consult documentation
of corresponding method to understand the details and
see what arguments can be used.
}
\seealso{
Other flextable constructors:
\code{\link[=flextable]{flextable()}}
}
\concept{flextable_constructors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/append_chunk.R
\name{append_chunks}
\alias{append_chunks}
\title{Append chunks to flextable content}
\usage{
append_chunks(x, ..., i = NULL, j = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{chunks to be appened, see \code{\link[=as_chunk]{as_chunk()}}, \code{\link[=gg_chunk]{gg_chunk()}} and other
chunk elements for paragraph.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
append chunks (for example chunk \code{\link[=as_chunk]{as_chunk()}})
in a flextable.
}
\examples{
library(flextable)
img.file <- file.path(R.home("doc"), "html", "logo.jpg")

ft_1 <- flextable(head(cars))

ft_1 <- append_chunks(ft_1,
  # where to append
  i = c(1, 3, 5),
  j = 1,
  # what to append
  as_chunk(" "),
  as_image(src = img.file, width = .20, height = .15)
)
ft_1 <- set_table_properties(ft_1, layout = "autofit")
ft_1
}
\seealso{
\code{\link[=as_chunk]{as_chunk()}}, \code{\link[=as_sup]{as_sup()}}, \code{\link[=as_sub]{as_sub()}}, \code{\link[=colorize]{colorize()}}

Other functions to compose cell content:
\code{\link[=as_paragraph]{as_paragraph()}},
\code{\link[=compose]{compose()}},
\code{\link[=footnote]{footnote()}},
\code{\link[=labelizor]{labelizor()}},
\code{\link[=prepend_chunks]{prepend_chunks()}},
\code{\link[=void]{void()}}
}
\concept{cell_content_composition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/body_add_flextable.R
\name{body_replace_flextable_at_bkm}
\alias{body_replace_flextable_at_bkm}
\title{Add flextable at bookmark location in a Word document}
\usage{
body_replace_flextable_at_bkm(
  x,
  bookmark,
  value,
  align = "center",
  split = FALSE
)
}
\arguments{
\item{x}{an rdocx object}

\item{bookmark}{bookmark id}

\item{value}{\code{flextable} object}

\item{align}{left, center (default) or right.}

\item{split}{set to TRUE if you want to activate Word
option 'Allow row to break across pages'.}
}
\description{
Use this function if you want to replace a paragraph containing
a bookmark with a flextable. As a side effect, the bookmark will be lost.
}
\seealso{
Other functions for officer integration:
\code{\link[=body_add_flextable]{body_add_flextable()}},
\code{\link[=ph_with.flextable]{ph_with.flextable()}},
\code{\link[=rtf_add.flextable]{rtf_add.flextable()}}
}
\concept{officer_integration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_flextable.R
\name{merge_at}
\alias{merge_at}
\title{Merge flextable cells into a single one}
\usage{
merge_at(x, i = NULL, j = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
Merge flextable cells into a single one. All
rows and columns must be consecutive.
}
\examples{
ft_merge <- flextable(head(mtcars), cwidth = .5)
ft_merge <- merge_at(ft_merge, i = 1:2, j = 1:2)
ft_merge
}
\seealso{
Other cell merging functions:
\code{\link[=merge_h]{merge_h()}},
\code{\link[=merge_h_range]{merge_h_range()}},
\code{\link[=merge_none]{merge_none()}},
\code{\link[=merge_v]{merge_v()}}
}
\concept{cell merging functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{colformat_int}
\alias{colformat_int}
\title{format integer cells}
\usage{
colformat_int(
  x,
  i = NULL,
  j = NULL,
  big.mark = get_flextable_defaults()$big.mark,
  na_str = get_flextable_defaults()$na_str,
  nan_str = get_flextable_defaults()$nan_str,
  prefix = "",
  suffix = ""
)
}
\arguments{
\item{x}{a flextable object}

\item{i}{rows selection}

\item{j}{columns selection.}

\item{big.mark}{see \code{\link[=format]{format()}}}

\item{na_str, nan_str}{string to be used for NA and NaN values}

\item{prefix, suffix}{string to be used as prefix or suffix}
}
\description{
Format integer cells in a flextable.
}
\examples{
z <- flextable(head(mtcars))
j <- c("vs", "am", "gear", "carb")
z <- colformat_int(x = z, j = j, prefix = "# ")
z
}
\seealso{
Other cells formatters: 
\code{\link{colformat_char}()},
\code{\link{colformat_datetime}()},
\code{\link{colformat_date}()},
\code{\link{colformat_double}()},
\code{\link{colformat_image}()},
\code{\link{colformat_lgl}()},
\code{\link{colformat_num}()},
\code{\link{set_formatter}()}
}
\concept{cells formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{keep_with_next}
\alias{keep_with_next}
\title{Set Word 'Keep with next' instructions}
\usage{
keep_with_next(x, i = NULL, value = TRUE, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{value}{TRUE or FALSE. When applied to a group, all rows
except the last one should be flagged with the 'Keep with next' attribute.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
The 'Keep with next' functionality in 'Word', applied
to the rows of a table, ensures that rows with that attribute
stay together and do not break across multiple pages.

This function provides better control of page breaks than
the global \code{keep_with_next} parameter.
}
\examples{
library(flextable)
dat <- iris[c(1:25, 51:75, 101:125), ]
ft <- qflextable(dat)
ft <- keep_with_next(
  x = ft,
  i = c(1:24, 26:49, 51:74),
  value = TRUE
)

save_as_docx(ft, path = tempfile(fileext = ".docx"))
}
\seealso{
\code{\link[=paginate]{paginate()}}

Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitters.R
\name{split_columns}
\alias{split_columns}
\title{Split a flextable into pages by columns}
\usage{
split_columns(x, max_width, rep_cols = NULL, unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{max_width}{Maximum width for each page (in inches by default).}

\item{rep_cols}{Columns to repeat on every page. Can be a character
vector of column names or an integer vector of column positions.
\code{NULL} (default) means no repetition. Repeated columns appear at
the beginning of each page in the order specified.}

\item{unit}{Unit for \code{max_width}, one of \code{"in"}, \code{"cm"}, \code{"mm"}.}
}
\value{
A list of flextable objects. If no splitting is needed, a
single-element list is returned.
}
\description{
Split a flextable into a list of flextables whose columns
fit within a given width (in inches). This is useful for paginating
wide tables in Word or PowerPoint output.
}
\examples{
ft <- flextable(head(mtcars))
ft_pages <- split_columns(ft, max_width = 5)
length(ft_pages)
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{linerange}
\alias{linerange}
\title{mini linerange chunk wrapper}
\usage{
linerange(
  value,
  min = NULL,
  max = NULL,
  rangecol = "#CCCCCC",
  stickcol = "#FF0000",
  bg = "transparent",
  width = 1,
  height = 0.2,
  raster_width = 30,
  unit = "in"
)
}
\arguments{
\item{value}{values containing the bar size}

\item{min}{min bar size. Default min of value}

\item{max}{max bar size. Default max of value}

\item{rangecol}{bar color}

\item{stickcol}{jauge color}

\item{bg}{background color}

\item{width, height}{size of the resulting png file in inches}

\item{raster_width}{number of pixels used as width
when interpolating value.}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}
}
\description{
This function is used to insert lineranges into
flextable with functions:
\itemize{
\item \code{\link[=compose]{compose()}} and \code{\link[=as_paragraph]{as_paragraph()}},
\item \code{\link[=append_chunks]{append_chunks()}},
\item \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\examples{
myft <- flextable( head(iris, n = 10 ))

myft <- compose( myft, j = 1,
  value = as_paragraph(
    linerange(value = Sepal.Length)
  ),
  part = "body")

autofit(myft)
}
\seealso{
\code{\link[=compose]{compose()}}, \code{\link[=as_paragraph]{as_paragraph()}}

Other chunk elements for paragraph: 
\code{\link{as_bracket}()},
\code{\link{as_b}()},
\code{\link{as_chunk}()},
\code{\link{as_equation}()},
\code{\link{as_highlight}()},
\code{\link{as_image}()},
\code{\link{as_i}()},
\code{\link{as_sub}()},
\code{\link{as_sup}()},
\code{\link{as_word_field}()},
\code{\link{colorize}()},
\code{\link{gg_chunk}()},
\code{\link{grid_chunk}()},
\code{\link{hyperlink_text}()},
\code{\link{lollipop}()},
\code{\link{minibar}()},
\code{\link{plot_chunk}()}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_bracket}
\alias{as_bracket}
\title{Bracket chunk}
\usage{
as_bracket(..., sep = ", ", p = "(", s = ")")
}
\arguments{
\item{...}{text and column names}

\item{sep}{separator}

\item{p}{prefix, default to '('}

\item{s}{suffix, default to ')'}
}
\description{
The function is producing a chunk by
pasting values and add the result in brackets.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
\examples{
ft <- flextable(head(iris),
  col_keys = c("Species", "Sepal", "Petal")
)
ft <- set_header_labels(ft, Sepal = "Sepal", Petal = "Petal")
ft <- compose(ft,
  j = "Sepal",
  value = as_paragraph(as_bracket(Sepal.Length, Sepal.Width))
)
ft <- compose(ft,
  j = "Petal",
  value = as_paragraph(as_bracket(Petal.Length, Petal.Width))
)
ft
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{minibar}
\alias{minibar}
\title{Mini barplot chunk}
\usage{
minibar(
  value,
  max = NULL,
  barcol = "#CCCCCC",
  bg = "transparent",
  width = 1,
  height = 0.2,
  unit = "in",
  alt = ""
)
}
\arguments{
\item{value}{values containing the bar size}

\item{max}{max bar size}

\item{barcol}{bar color}

\item{bg}{background color}

\item{width, height}{size of the resulting png file in inches}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{alt}{alternative text for the image (used for accessibility)}
}
\description{
This function is used to insert bars into
flextable with functions:
\itemize{
\item \code{\link[=compose]{compose()}} and \code{\link[=as_paragraph]{as_paragraph()}},
\item \code{\link[=append_chunks]{append_chunks()}},
\item \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format. With Quarto (\code{format: docx}) or
\code{rmarkdown::word_document()}, the resulting file must be repaired
with \code{\link[=repair_docx]{repair_docx()}}.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\examples{
ft <- flextable(head(iris, n = 10))

ft <- compose(ft,
  j = 1,
  value = as_paragraph(
    minibar(value = Sepal.Length, max = max(Sepal.Length))
  ),
  part = "body"
)

ft <- autofit(ft)
ft
}
\seealso{
\code{\link[=compose]{compose()}}, \code{\link[=as_paragraph]{as_paragraph()}}

Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz-defunct.R
\name{set_formatter_type}
\alias{set_formatter_type}
\title{Set Formatter by Types of Columns}
\usage{
set_formatter_type(...)
}
\description{
Function is now defunct, use \code{colformat_} functions instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{rotate}
\alias{rotate}
\title{Rotate cell text}
\usage{
rotate(x, i = NULL, j = NULL, rotation, align = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{rotation}{one of "lrtb", "tbrl", "btlr".}

\item{align}{vertical alignment of paragraph within cell,
one of "center" or "top" or "bottom".}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
It can be useful to change the text direction
when table headers are large. For example, header labels can
be rendered as "tbrl" (top to bottom and right to left), corresponding
to a 90-degree rotation, or "btlr", corresponding to a 270-degree
rotation.
This function changes cell text direction. By default, it is
"lrtb", which means from left to right and top to bottom.

'Word' and 'PowerPoint' do not handle automatic height with rotated headers.
Therefore, you need to set header heights (with the \code{\link[=height]{height()}} function)
and set the rule to "exact" for row heights (with the \code{\link[=hrule]{hrule()}} function);
otherwise, Word and PowerPoint outputs will have insufficient height
to properly display the text.

flextable does not rotate text by arbitrary angles. It only
rotates by right angles (90-degree increments). This choice ensures
consistent rendering across Word, PowerPoint (limited to angles 0, 270, and 90),
HTML, and PDF.
}
\details{
When the \code{\link[=autofit]{autofit()}} function is used, rotation will be
ignored. In that case, use \link{dim_pretty} and \link{width} instead
of \code{autofit()}.
}
\examples{
library(flextable)

ft_1 <- flextable(head(iris))

ft_1 <- rotate(ft_1, j = 1:4, align = "bottom", rotation = "tbrl", part = "header")
ft_1 <- rotate(ft_1, j = 5, align = "bottom", rotation = "btlr", part = "header")

# if output is docx or pptx, think about (1) set header heights
# and (2) set rule "exact" for rows heights because Word
# and PowerPoint don't handle auto height with rotated headers
ft_1 <- height(ft_1, height = 1.2, part = "header")
ft_1 <- hrule(ft_1, i = 1, rule = "exact", part = "header")

ft_1

dat <- data.frame(
  a = c("left-top", "left-middle", "left-bottom"),
  b = c("center-top", "center-middle", "center-bottom"),
  c = c("right-top", "right-middle", "right-bottom")
)

ft_2 <- flextable(dat)
ft_2 <- theme_box(ft_2)
ft_2 <- height_all(x = ft_2, height = 1.3, part = "body")
ft_2 <- hrule(ft_2, rule = "exact")
ft_2 <- rotate(ft_2, rotation = "tbrl")
ft_2 <- width(ft_2, width = 1.3)

ft_2 <- align(ft_2, j = 1, align = "left")
ft_2 <- align(ft_2, j = 2, align = "center")
ft_2 <- align(ft_2, j = 3, align = "right")

ft_2 <- valign(ft_2, i = 1, valign = "top")
ft_2 <- valign(ft_2, i = 2, valign = "center")
ft_2 <- valign(ft_2, i = 3, valign = "bottom")

ft_2
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_structure.R
\name{information_data_chunk}
\alias{information_data_chunk}
\title{Get chunk-level content information from a flextable}
\usage{
information_data_chunk(x, expand_special_chars = TRUE)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}
}
\value{
a data.frame containing information about chunks:
\itemize{
\item text chunk (column \code{txt}) and other content (\code{url}
for the linked url, \code{eq_data} for content of type 'equation',
\code{word_field_data} for content of type 'word_field' and
\code{img_data} for content of type 'image'),
\item formatting properties,
\item part (\code{.part}), position within the paragraph (\code{.chunk_index}),
row (\code{.row_id}) and column (\code{.col_id}).
}
}
\description{
This function takes a flextable object and returns a data.frame containing
information about each text chunk within the flextable. The data.frame includes
details such as the text content, formatting properties, position within the
paragraph, paragraph row, and column.
}
\section{don't use this}{


These data structures should not be used, as they
represent an interpretation of the underlying data
structures, which may evolve over time.

\strong{They are exported to enable two packages that exploit
these structures to make a transition, and should not
remain available for long.}
}

\examples{
ft <- as_flextable(iris)
x <- information_data_chunk(ft)
head(x)
}
\seealso{
Other table data extraction:
\code{\link[=information_data_cell]{information_data_cell()}},
\code{\link[=information_data_paragraph]{information_data_paragraph()}}
}
\concept{table data extraction}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{add_footer_lines}
\alias{add_footer_lines}
\title{Add labels as new rows in the footer}
\usage{
add_footer_lines(x, values = character(0), top = FALSE)
}
\arguments{
\item{x}{a \code{flextable} object}

\item{values}{a character vector or a call to \code{\link[=as_paragraph]{as_paragraph()}}
to get formated content, each element will
be added as a new row.}

\item{top}{should the row be inserted at the top
or the bottom. Default to TRUE.}
}
\description{
Add labels as new rows in the footer,
where all columns are merged.

This is a sugar function to be used when you need to
add labels in the footer, a footnote for example.
}
\section{Illustrations}{

\if{html}{\figure{fig_add_footer_lines_1.png}{options: width="300"}}
}

\examples{
ft_1 <- flextable(head(iris))
ft_1 <- add_footer_lines(ft_1,
  values = c("blah 1", "blah 2")
)
ft_1
}
\seealso{
Other functions to add rows in a flextable: 
\code{\link{add_body_row}()},
\code{\link{add_body}()},
\code{\link{add_footer_row}()},
\code{\link{add_footer}()},
\code{\link{add_header_row}()},
\code{\link{add_header}()},
\code{\link{separate_header}()},
\code{\link{set_header_footer_df}},
\code{\link{set_header_labels}()}
}
\concept{functions to add rows in a flextable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_chunk}
\alias{as_chunk}
\title{Text chunk}
\usage{
as_chunk(x, props = NULL, formatter = format_fun, ...)
}
\arguments{
\item{x}{text or any element that can be formatted as text
with function provided in argument \code{formatter}.}

\item{props}{an \code{\link[=fp_text_default]{fp_text_default()}} or \code{\link[officer:fp_text]{officer::fp_text()}} object to be used to format the text.
If not specified, it will be the default value corresponding to the cell.}

\item{formatter}{a function that will format x as a character vector.}

\item{...}{additional arguments for \code{formatter} function.}
}
\description{
The function lets add formated text in flextable
cells.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.

It should be used inside a call to \code{\link[=as_paragraph]{as_paragraph()}}.
}
\examples{
library(officer)

ft <- flextable(head(iris))

ft <- compose(ft,
  j = "Sepal.Length",
  value = as_paragraph(
    "Sepal.Length value is ",
    as_chunk(Sepal.Length, props = fp_text(color = "red"))
  ),
  part = "body"
)
ft <- color(ft, color = "gray40", part = "all")
ft <- autofit(ft)
ft
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_grobs.R
\name{dim.flextableGrob}
\alias{dim.flextableGrob}
\title{Get optimal width and height of a flextable grob}
\usage{
\method{dim}{flextableGrob}(x)
}
\arguments{
\item{x}{a flextableGrob object}
}
\value{
a named list with two elements, \code{width} and \code{height}.
Values are expressed in inches.
}
\description{
returns the optimal width and height for the grob,
according to the grob generation parameters.
}
\examples{
ftab <- flextable(head(iris))
gr <- gen_grob(ftab)
dim(gr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{color}
\alias{color}
\title{Set font color}
\usage{
color(x, i = NULL, j = NULL, color, part = "body", source = j)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{color}{color to use as font color. If a function, the function must return
a character vector of colors.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}

\item{source}{if color is a function, source specifies the dataset column to be used
as an argument to \code{color}. This is only useful when j is colored with values contained in
other columns.}
}
\description{
Change the text color of selected rows and
columns of a flextable. A function can be used instead of
fixed colors.

When \code{color} is a function, it is possible to color cells based on values
located in other columns; using hidden columns (those not used by
argument \code{colkeys}) is a common use case. The argument \code{source}
must be used to define the columns to be used for the color
definition, and the argument \code{j} must be used to define where to apply
the colors and only accepts values from \code{colkeys}.
}
\examples{
ft <- flextable(head(mtcars))
ft <- color(ft, color = "orange", part = "header")
ft <- color(ft,
  color = "red",
  i = ~ qsec < 18 & vs < 1
)
ft

if (require("scales")) {
  scale <- scales::col_numeric(domain = c(-1, 1), palette = "RdBu")
  x <- as.data.frame(cor(iris[-5]))
  x <- cbind(
    data.frame(
      colname = colnames(x),
      stringsAsFactors = FALSE
    ),
    x
  )
}
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.kmeans}
\alias{as_flextable.kmeans}
\title{tabular summary for kmeans}
\usage{
\method{as_flextable}{kmeans}(x, digits = 4, ...)
}
\arguments{
\item{x}{a \code{\link[=kmeans]{kmeans()}} object}

\item{digits}{number of digits for the numeric columns}

\item{...}{unused argument}
}
\description{
produce a flextable describing a
kmeans object. The function is only using package 'broom'
that provides the data presented in the resulting flextable.
}
\examples{
if(require("stats")){
  cl <- kmeans(scale(mtcars[1:7]), 5)
  ft <- as_flextable(cl)
  ft
}
}
\seealso{
Other as_flextable methods: 
\code{\link{as_flextable.data.frame}()},
\code{\link{as_flextable.glm}()},
\code{\link{as_flextable.grouped_data}()},
\code{\link{as_flextable.htest}()},
\code{\link{as_flextable.lm}()},
\code{\link{as_flextable.merMod}()},
\code{\link{as_flextable.pam}()},
\code{\link{as_flextable.summarizor}()},
\code{\link{as_flextable.tabulator}()},
\code{\link{as_flextable.xtable}()},
\code{\link{as_flextable}()}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{height}
\alias{height}
\alias{height_all}
\title{Set flextable rows height}
\usage{
height(x, i = NULL, height, part = "body", unit = "in")

height_all(x, height, part = "all", unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{height}{height in inches}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}

\item{unit}{unit for height, one of "in", "cm", "mm".}
}
\description{
control rows height for a part of the flextable when the line
height adjustment is "atleast" or "exact" (see \code{\link[=hrule]{hrule()}}).
}
\note{
This function has no effect when the rule for line height is set to
"auto" (see \code{\link[=hrule]{hrule()}}), which is the default case, except with PowerPoint
which does not support this automatic line height adjustment feature.
}
\section{height_all}{

\code{height_all} is a convenient function for
setting the same height to all rows (selected
with argument \code{part}).
}

\examples{
ft_1 <- flextable(head(iris))
ft_1 <- height(ft_1, height = .5)
ft_1 <- hrule(ft_1, rule = "exact")
ft_1


ft_2 <- flextable(head(iris))
ft_2 <- height_all(ft_2, height = 1)
ft_2 <- hrule(ft_2, rule = "exact")
ft_2
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{save_as_docx}
\alias{save_as_docx}
\title{save flextable objects in an Word file}
\usage{
save_as_docx(..., values = NULL, path, pr_section = NULL, align = "center")
}
\arguments{
\item{...}{flextable objects, objects, possibly named. If named objects, names are
used as titles.}

\item{values}{a list (possibly named), each element is a flextable object. If named objects, names are
used as titles. If provided, argument \code{...} will be ignored.}

\item{path}{Word file to be created}

\item{pr_section}{a \link{prop_section} object that can be used to define page
layout such as orientation, width and height.}

\item{align}{left, center (default) or right.}
}
\description{
sugar function to save flextable objects in an Word file.
}
\examples{

tf <- tempfile(fileext = ".docx")

library(officer)
ft1 <- flextable(head(iris))
save_as_docx(ft1, path = tf)


ft2 <- flextable(head(mtcars))
sect_properties <- prop_section(
  page_size = page_size(
    orient = "landscape",
    width = 8.3, height = 11.7
  ),
  type = "continuous",
  page_margins = page_mar()
)
save_as_docx(
  `iris table` = ft1, `mtcars table` = ft2,
  path = tf, pr_section = sect_properties
)
}
\seealso{
Other flextable print function: 
\code{\link{as_raster}()},
\code{\link{df_printer}()},
\code{\link{flextable_to_rmd}()},
\code{\link{gen_grob}()},
\code{\link{htmltools_value}()},
\code{\link{knit_print.flextable}()},
\code{\link{plot.flextable}()},
\code{\link{print.flextable}()},
\code{\link{save_as_html}()},
\code{\link{save_as_image}()},
\code{\link{save_as_pptx}()}
}
\concept{flextable print function}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{repair_docx}
\alias{repair_docx}
\title{Repair flextables in a 'Word' document produced by 'pandoc'}
\usage{
repair_docx(path, target = path)
}
\arguments{
\item{path}{path of the 'docx' file to repair. The file is modified
in place unless \code{target} is provided.}

\item{target}{path of the resulting 'docx' file. Defaults to \code{path}.}
}
\value{
the path of the repaired file, invisibly.
}
\description{
Word documents produced by Quarto (\code{format: docx}) or by
\code{rmarkdown::word_document()} are written by 'pandoc', which copies the
raw OOXML of flextable objects verbatim: images and hyperlinks are
left as unresolved references (file paths and URLs instead of
relationship identifiers) and \strong{Word refuses to open the file}.

This function reads such a document and rewrites it with the package
'officer', whose writer resolves these references: images and SVG
files are embedded, hyperlinks are registered, and the other
officer-flavored contents are processed as well (sections, footnotes,
comments, custom style names, list markers, documents poured with
\code{officer::block_pour_docx()}).

It is not needed when the document is produced with
\code{officedown::rdocx_document()} or \code{\link[=save_as_docx]{save_as_docx()}}, where the file is
already written by 'officer'.
}
\section{What to do with Quarto}{


The repair must run after \verb{quarto render}. Quarto can do it for you
with a \emph{post-render script}, but this can \strong{only} be declared in a
\verb{_quarto.yml} file - the \verb{project:} key is ignored in the YAML header
of a \code{.qmd} file.
\enumerate{
\item Next to your \code{.qmd} file, create a file \verb{_quarto.yml} containing:
}

\if{html}{\out{<div class="sourceCode yaml">}}\preformatted{project:
  post-render: repair.R
}\if{html}{\out{</div>}}
\enumerate{
\item In the same directory, create the file \code{repair.R} containing:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{files <- strsplit(Sys.getenv("QUARTO_PROJECT_OUTPUT_FILES"), "\\n")[[1]]
for (f in files[grepl("\\\\.docx$", files, ignore.case = TRUE)]) \{
  flextable::repair_docx(f)
\}
}\if{html}{\out{</div>}}
\enumerate{
\item Render as usual, either the whole project (\verb{quarto render}) or a
single file (\verb{quarto render doc.qmd}); the script runs in both cases
and only processes the files that were rendered.
}

Without a \verb{_quarto.yml}, call \code{repair_docx("doc.docx")} manually
after each render.
}

\section{What to do with R Markdown}{


Prefer \code{officedown::rdocx_document()} as output format: the file is
post-processed by 'officer' and no repair is needed. With
\code{rmarkdown::word_document()}, call
\code{repair_docx()} on the output file after \code{rmarkdown::render()}.
}

\section{Requirements and limits}{

\itemize{
\item The images referenced in the document must still exist, at paths
resolvable \strong{from the directory where \code{repair_docx()} runs}: a
Quarto post-render script runs at the project root, so images
referenced with paths relative to a sub-directory document will not
be found - use absolute paths (e.g. \code{here::here()}) in that case.
\item Mini graphics generated by \code{\link[=minibar]{minibar()}}, \code{\link[=linerange]{linerange()}},
\code{\link[=gg_chunk]{gg_chunk()}}, \code{\link[=plot_chunk]{plot_chunk()}} and \code{\link[=grid_chunk]{grid_chunk()}} are written to
temporary 'PNG' files that no longer exist once rendering has
completed: \strong{these can not be repaired afterwards}. With Quarto or
\code{rmarkdown::word_document()}, use image files you manage yourself
(\code{\link[=as_image]{as_image()}}, \code{\link[=colformat_image]{colformat_image()}}); mini graphics require
\code{officedown::rdocx_document()} or \code{\link[=save_as_docx]{save_as_docx()}}.
\item When a referenced image can not be found, the rewrite fails with
an error \verb{cannot open the connection} and a warning naming the
missing file.
}
}

\examples{
\dontrun{
repair_docx("document.docx")
}
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{add_footer_row}
\alias{add_footer_row}
\title{Add footer labels}
\usage{
add_footer_row(x, top = TRUE, values = character(0), colwidths = integer(0))
}
\arguments{
\item{x}{a flextable object}

\item{top}{should the row be inserted at the top or the bottom.}

\item{values}{values to add. It can be a \code{list}, a \code{character()} vector
or a call to \code{\link[=as_paragraph]{as_paragraph()}}.

If it is a list, it can be a named list with the names of the columns of the
original data.frame or the \code{colkeys}; this is the recommended method because
it allows to keep the original data types and therefore allows to perform
conditional formatting. If a character, columns of the original data.frame
stored in the flextable object are changed to \code{character()}; this is often
not an issue with footer and header but can be inconvenient if adding
rows into body as it will change data types to character and prevent
efficient conditional formatting.}

\item{colwidths}{the number of columns to merge in the row for each label}
}
\description{
Add a row of new columns labels in footer part.
Labels can be spanned along multiple columns, as merged cells.

Labels are associated with a number of columns
to merge that default to one if not specified.
In this case, you have to make sure that the
number of labels is equal to the number of columns displayed.

The function can add only one single row by call.

Labels can be formatted with \code{\link[=as_paragraph]{as_paragraph()}}.
}
\examples{
library(flextable)

ft01 <- fp_text_default(color = "red")
ft02 <- fp_text_default(color = "orange")

pars <- as_paragraph(
  as_chunk(c("(1)", "(2)"), props = ft02), " ",
  as_chunk(
    c(
      "My tailor is rich",
      "My baker is rich"
    ),
    props = ft01
  )
)

ft_1 <- flextable(head(mtcars))
ft_1 <- add_footer_row(ft_1,
  values = pars,
  colwidths = c(5, 6), top = FALSE
)
ft_1 <- add_footer_row(ft_1,
  values = pars,
  colwidths = c(3, 8), top = TRUE
)
ft_1

ft_2 <- flextable(head(airquality))
ft_2 <- add_footer_row(ft_2,
  values = c("Measure", "Time"),
  colwidths = c(4, 2), top = TRUE
)
ft_2 <- theme_box(ft_2)
ft_2
}
\seealso{
\code{\link[=flextable]{flextable()}}, \code{\link[=set_caption]{set_caption()}}

Other functions to add rows in a flextable: 
\code{\link{add_body_row}()},
\code{\link{add_body}()},
\code{\link{add_footer_lines}()},
\code{\link{add_footer}()},
\code{\link{add_header_row}()},
\code{\link{add_header}()},
\code{\link{separate_header}()},
\code{\link{set_header_footer_df}},
\code{\link{set_header_labels}()}
}
\concept{functions to add rows in a flextable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment_rows.R
\name{separate_header}
\alias{separate_header}
\title{Split column names using a separator into multiple rows}
\usage{
separate_header(
  x,
  opts = c("span-top", "center-hspan", "bottom-vspan", "default-theme"),
  split = "[_\\\\.]",
  fixed = FALSE
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{opts}{Optional treatments to apply to the resulting header part.
This should be a character vector with support for multiple values.

Supported values include:
\itemize{
\item "span-top": This operation spans empty cells with the first non-empty
cell, applied column by column.
\item "center-hspan": Center the cells that are horizontally spanned.
\item "bottom-vspan": Aligns to the bottom the cells treated at the  when "span-top" is applied.
\item "default-theme": Applies the theme set in \code{set_flextable_defaults(theme_fun = ...)} to the new header part.
}}

\item{split}{a regular expression (unless \code{fixed = TRUE})
to use for splitting.}

\item{fixed}{logical. If TRUE match \code{split} exactly,
otherwise use regular expressions.}
}
\description{
This function is used to separate and place individual
labels in their own rows if your variable names contain multiple
delimited labels.
\if{html}{\out{
<img src="https://www.ardata.fr/img/flextable-imgs/flextable-016.png" alt="add_header illustration" style="width:100\%;">
}}
}
\examples{
library(flextable)

x <- data.frame(
  Species = as.factor(c("setosa", "versicolor", "virginica")),
  Sepal.Length_mean = c(5.006, 5.936, 6.588),
  Sepal.Length_sd = c(0.35249, 0.51617, 0.63588),
  Sepal.Width_mean = c(3.428, 2.77, 2.974),
  Sepal.Width_sd = c(0.37906, 0.3138, 0.3225),
  Petal.Length_mean = c(1.462, 4.26, 5.552),
  Petal.Length_sd = c(0.17366, 0.46991, 0.55189),
  Petal.Width_mean = c(0.246, 1.326, 2.026),
  Petal.Width_sd = c(0.10539, 0.19775, 0.27465)
)

ft_1 <- flextable(x)
ft_1 <- colformat_double(ft_1, digits = 2)
ft_1 <- theme_box(ft_1)
ft_1 <- separate_header(
  x = ft_1,
  opts = c("span-top", "bottom-vspan")
)
ft_1
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}},
\code{\link[=split_to_pages]{split_to_pages()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{save_as_html}
\alias{save_as_html}
\title{Save flextable objects in an 'HTML' file}
\usage{
save_as_html(..., values = NULL, path, lang = "en", title = "&#32;")
}
\arguments{
\item{...}{flextable objects, objects, possibly named. If named objects, names are
used as titles.}

\item{values}{a list (possibly named), each element is a flextable object. If named objects, names are
used as titles. If provided, argument \code{...} will be ignored.}

\item{path}{HTML file to be created}

\item{lang}{language of the document using IETF language tags}

\item{title}{page title}
}
\value{
a string containing the full name of the generated file
}
\description{
save a flextable in an 'HTML' file. This function
is useful to save the flextable in 'HTML' file without using
R Markdown (it is highly recommanded to use R Markdown
instead).
}
\examples{
ft1 <- flextable(head(iris))
tf1 <- tempfile(fileext = ".html")
if (rmarkdown::pandoc_available()) {
  save_as_html(ft1, path = tf1)
  # browseURL(tf1)
}

ft2 <- flextable(head(mtcars))
tf2 <- tempfile(fileext = ".html")
if (rmarkdown::pandoc_available()) {
  save_as_html(
    `iris table` = ft1,
    `mtcars table` = ft2,
    path = tf2,
    title = "rhoooo"
  )
  # browseURL(tf2)
}
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/body_add_flextable.R
\name{headers_flextable_at_bkm}
\alias{headers_flextable_at_bkm}
\title{add flextable at a bookmark location in document's header}
\usage{
headers_flextable_at_bkm(x, bookmark, value)
}
\arguments{
\item{x}{an rdocx object}

\item{bookmark}{bookmark id}

\item{value}{a flextable object}
}
\description{
replace in the header of a document  a paragraph containing a bookmark by a flextable.
A bookmark will be considered as valid if enclosing words
within a paragraph; i.e., a bookmark along two or more paragraphs is invalid,
a bookmark set on a whole paragraph is also invalid, but bookmarking few words
inside a paragraph is valid.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compose.R
\name{labelizor}
\alias{labelizor}
\title{Replace displayed text with labels}
\usage{
labelizor(x, j = NULL, labels, part = "all")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{labels}{a named vector whose names will be used to identify
values to replace and values will be used as labels.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{labelizor()} substitutes text values shown in a flextable
with human-readable labels. This is useful to turn column
values such as variable names, factor levels or coded strings
into presentation-ready wording (e.g. \code{"Sepal.Length"} to
\code{"Sepal Length"}).

\code{labels} can be either a \strong{named character vector} (names
identify values to find, values are the replacement labels)
or a \strong{function} applied to every text chunk (e.g. \code{\link[=toupper]{toupper()}}).

Only the displayed content is affected; the underlying data
of the flextable is unchanged.
}
\examples{
\dontshow{
data.table::setDTthreads(1)
}
z <- summarizor(
  x = CO2[-c(1, 4)],
  by = "Treatment",
  overall_label = "Overall"
)

ft_1 <- as_flextable(z, separate_with = "variable")

ft_1 <- labelizor(
  x = ft_1, j = c("stat"),
  labels = c(Missing = "Kouign amann")
)

ft_1 <- labelizor(
  x = ft_1, j = c("variable"),
  labels = toupper
)

ft_1
}
\seealso{
\code{\link[=mk_par]{mk_par()}}, \code{\link[=append_chunks]{append_chunks()}}, \code{\link[=prepend_chunks]{prepend_chunks()}}

Other functions to compose cell content:
\code{\link[=append_chunks]{append_chunks()}},
\code{\link[=as_paragraph]{as_paragraph()}},
\code{\link[=compose]{compose()}},
\code{\link[=footnote]{footnote()}},
\code{\link[=prepend_chunks]{prepend_chunks()}},
\code{\link[=void]{void()}}
}
\concept{cell_content_composition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{hline}
\alias{hline}
\title{Set horizontal borders below selected rows}
\usage{
hline(x, i = NULL, j = NULL, border = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{border}{border properties defined by a call to
\code{\link[officer:fp_border]{officer::fp_border()}}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{hline()} draws a horizontal line \strong{below} each selected row by
setting the bottom border of cells at row \code{i} (and the top border
of cells at row \code{i + 1} so that the line renders consistently
across output formats).

Use the \code{i} selector to target specific rows (e.g. a formula
such as \code{~ before(col, "Total")}). When \code{i} is \code{NULL} (the
default) the border is added below every row, which produces
a full grid of inner horizontal lines.

For the \strong{outer} edges of the table, use \code{\link[=hline_top]{hline_top()}} and
\code{\link[=hline_bottom]{hline_bottom()}} instead; those always target the very first
or very last row of a part.
}
\examples{
library(officer)
std_border <- fp_border(color = "gray")

ft <- flextable(head(iris))
ft <- border_remove(x = ft)

# add horizontal borders below every row
ft <- hline(ft, part = "all", border = std_border)
ft
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_h]{border_inner_h()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=border_remove]{border_remove()}},
\code{\link[=hline_bottom]{hline_bottom()}},
\code{\link[=hline_top]{hline_top()}},
\code{\link[=surround]{surround()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}},
\code{\link[=vline_right]{vline_right()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/typst_str.R
\name{save_as_typst}
\alias{save_as_typst}
\title{Save flextable objects in a Typst file}
\usage{
save_as_typst(
  ...,
  values = NULL,
  path,
  page = "#set page(width: auto, height: auto, margin: 12pt)"
)
}
\arguments{
\item{...}{flextable objects, possibly named.}

\item{values}{a list of flextable objects. If provided, \code{...} is ignored.}

\item{path}{Typst file to be created.}

\item{page}{a Typst \verb{#set page(...)} rule used as document setup. The
default produces a tightly cropped page fitting the table(s).}
}
\value{
the path to the generated file, invisibly.
}
\description{
Writes one or more flextables to a standalone Typst
(\code{.typ}) document. Images are embedded as base64 (decoded inline at
compile time), so the file is self-contained and compiles offline with
\verb{typst compile}. Equations require the \code{mitex} Typst package, which is
fetched from the Typst package registry on first compilation.
}
\examples{
ft <- flextable(head(iris))
tf <- tempfile(fileext = ".typ")
save_as_typst(ft, path = tf)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/empty.R
\name{void}
\alias{void}
\title{Clear the displayed content of selected columns}
\usage{
void(x, j = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{void()} replaces the visible text of the selected columns
with an empty string. The columns themselves (and their
headers) remain in the table, but the cell values are no
longer displayed.

This is useful when a column should stay in the layout
(e.g. to preserve its width or to keep its header label)
but its body values should be hidden, for instance
after using \code{\link[=compose]{compose()}} to build a richer display in a
neighbouring column that already incorporates those values.

The underlying dataset is not modified; only the displayed
content is affected. To remove a column entirely, use
the \code{col_keys} argument of \code{\link[=flextable]{flextable()}} instead.
}
\examples{
ftab <- flextable(head(mtcars))
ftab <- void(ftab, ~ vs + am + gear + carb)
ftab
}
\seealso{
Other functions to compose cell content:
\code{\link[=append_chunks]{append_chunks()}},
\code{\link[=as_paragraph]{as_paragraph()}},
\code{\link[=compose]{compose()}},
\code{\link[=footnote]{footnote()}},
\code{\link[=labelizor]{labelizor()}},
\code{\link[=prepend_chunks]{prepend_chunks()}}
}
\concept{cell_content_composition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{save_as_rtf}
\alias{save_as_rtf}
\title{Save flextable objects in an 'RTF' file}
\usage{
save_as_rtf(..., values = NULL, path, pr_section = NULL)
}
\arguments{
\item{...}{flextable objects, objects, possibly named. If named objects, names are
used as titles.}

\item{values}{a list (possibly named), each element is a flextable object. If named objects, names are
used as titles. If provided, argument \code{...} will be ignored.}

\item{path}{Word file to be created}

\item{pr_section}{a \link[officer:prop_section]{officer::prop_section} object that can be used to define page
layout such as orientation, width and height.}
}
\value{
a string containing the full name of the generated file
}
\description{
sugar function to save flextable objects in an 'RTF' file.
}
\examples{

tf <- tempfile(fileext = ".rtf")

library(officer)
ft1 <- flextable(head(iris))
save_as_rtf(ft1, path = tf)


ft2 <- flextable(head(mtcars))
sect_properties <- prop_section(
  page_size = page_size(
    orient = "landscape",
    width = 8.3, height = 11.7
  ),
  type = "continuous",
  page_margins = page_mar(),
  header_default = block_list(
    fpar(ftext("text for default page header")),
    qflextable(data.frame(a = 1L))
  )
)
tf <- tempfile(fileext = ".rtf")
save_as_rtf(
  `iris table` = ft1, `mtcars table` = ft2,
  path = tf, pr_section = sect_properties
)
}
\seealso{
\code{\link[=paginate]{paginate()}}

Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_qmd}
\alias{as_qmd}
\title{Quarto inline markdown chunk}
\usage{
as_qmd(x, display = x)
}
\arguments{
\item{x}{character vector of Quarto markdown content.}

\item{display}{character vector of display text used
as fallback when the Lua filter is not active.
Defaults to \code{x}.}
}
\description{
\code{as_qmd()} creates a chunk for inline Quarto markdown
content (text-level) that fits within a table
cell paragraph. This enables cross-references
(\verb{@fig-xxx}, \verb{@tbl-xxx}), links, bold/italic, math,
inline code, shortcodes and other inline Quarto markdown
features inside flextable cells.

It is not designed for block-level elements such as
headings, bullet lists or fenced code blocks.

The chunk is used with \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}. It requires the \code{flextable-qmd} Lua
filter extension (see \code{\link[=use_flextable_qmd]{use_flextable_qmd()}}) and works with
HTML, PDF, Typst and Word (docx) Quarto output formats.
}
\section{Setup}{

\enumerate{
\item Install the extension once per project:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{flextable::use_flextable_qmd()
}\if{html}{\out{</div>}}
\enumerate{
\item Add the filter to your Quarto document YAML.
For HTML and PDF, a single line is enough:
}

\if{html}{\out{<div class="sourceCode yaml">}}\preformatted{filters:
  - flextable-qmd
}\if{html}{\out{</div>}}

For Word (docx), an additional post-render filter
removes the wrapper table that Quarto adds around
labelled flextables:

\if{html}{\out{<div class="sourceCode yaml">}}\preformatted{filters:
  - flextable-qmd
  - at: post-render
    path: _extensions/flextable-qmd/unwrap-float.lua
}\if{html}{\out{</div>}}
}

\section{Supported markdown}{

\itemize{
\item Cross-references: \verb{@fig-xxx}, \verb{@tbl-xxx}
\item Bold / italic: \verb{**bold**}, \verb{*italic*}
\item Inline code: \code{`code`}
\item Links: \verb{[text](url)} (internal and external)
\item Math: \verb{$\\\\alpha + \\\\beta$}
\item Shortcodes and other Quarto markdown constructs
}
}

\section{Limitations}{


Each table cell in a flextable contains a single paragraph
built from inline chunks (see \code{\link[=as_paragraph]{as_paragraph()}}). There is no
mechanism to insert block-level structures (multiple
paragraphs, lists, headings, fenced code blocks, callouts,
etc.) inside a cell. Because \code{as_qmd()} produces one of
these inline chunks, only inline markdown is supported.
}

\examples{
library(flextable)

dat <- data.frame(
  label = c("Bold", "Link", "Code"),
  content = c(
    "This is **bold** text",
    "Visit [Quarto](https://quarto.org)",
    "Use `print()` here"
  )
)
ft <- flextable(dat)
ft <- mk_par(ft, j = "content",
  value = as_paragraph(as_qmd(content)))
ft
}
\seealso{
\code{\link[=use_flextable_qmd]{use_flextable_qmd()}} to install the Lua filter extension,
\code{\link[=knit_print.flextable]{knit_print.flextable()}} for rendering options in knitr documents.

Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/append_chunk.R
\name{prepend_chunks}
\alias{prepend_chunks}
\title{Prepend chunks to flextable content}
\usage{
prepend_chunks(x, ..., i = NULL, j = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{chunks to be prepended, see \code{\link[=as_chunk]{as_chunk()}}, \code{\link[=gg_chunk]{gg_chunk()}} and other
chunk elements for paragraph.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
prepend chunks (for example chunk \code{\link[=as_chunk]{as_chunk()}})
in a flextable.
}
\examples{
x <- flextable(head(iris))
x <- prepend_chunks(
  x,
  i = 1, j = 1,
  colorize(as_b("Hello "), color = "red"),
  colorize(as_i("World"), color = "magenta")
)
x
}
\seealso{
Other functions to compose cell content:
\code{\link[=append_chunks]{append_chunks()}},
\code{\link[=as_paragraph]{as_paragraph()}},
\code{\link[=compose]{compose()}},
\code{\link[=footnote]{footnote()}},
\code{\link[=labelizor]{labelizor()}},
\code{\link[=void]{void()}}
}
\concept{cell_content_composition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable-package.R
\docType{package}
\name{flextable-package}
\alias{flextable-package}
\title{flextable: Functions for Tabular Reporting}
\description{
The flextable package facilitates access to and manipulation of
tabular reporting elements from R.

The documentation of functions can be opened with command \code{help(package = "flextable")}.

\code{flextable()} function is producing flexible tables where each cell
can contain several chunks of text with their own set of formatting
properties (bold, font color, etc.). Function \code{\link[=mk_par]{mk_par()}} lets customise
text of cells.

Each cell holds a single paragraph composed of inline chunks
(see \code{\link[=as_paragraph]{as_paragraph()}}). This means cell content is strictly
inline: bold, italic, links, images, equations, inline code, etc.
Block-level structures (multiple paragraphs, bullet lists,
headings or fenced code blocks) cannot be placed inside a cell.
Soft line breaks (\verb{\\n}) are however supported.

The \code{\link[=as_flextable]{as_flextable()}} function is used to transform specific objects into
flextable objects. For example, you can transform a crosstab produced with
the 'tables' package into a flextable which can then be formatted,
annotated or augmented with footnotes.

In order to reduce the homogenization efforts and the number of functions
to be called, it is recommended to define formatting properties such as
font, border color, number of decimals displayed which will then be applied
by default. See \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} for more details.
}
\section{Table Structure}{


A flextable is composed of three distinct parts:
\itemize{
\item \code{header}: By default, contains one row with the column names from the data.frame
\item \code{body}: Contains the actual data from the data.frame
\item \code{footer}: Empty by default, but can contain content (commonly used for footnotes or #' summary rows)
}

\if{html}{\out{<div class="sourceCode">}}\preformatted{       HEADER         <- Column names, labels, spanning headers
--------------------
        BODY          <- Data rows
--------------------
       FOOTER         <- Summary rows, notes, footnotes (optional)
}\if{html}{\out{</div>}}

A basic flextable has:
\itemize{
\item in the part 'header': 1 row with column names
\item in the part 'body': as many rows as there are in the input data.frame
\item no footer
}

Rows and columns can be added or removed to the basic flextable:
\itemize{
\item Add new rows in header with \code{\link[=add_header]{add_header()}}, \code{\link[=add_header_row]{add_header_row()}},
\code{\link[=add_header_lines]{add_header_lines()}} and \code{\link[=set_header_labels]{set_header_labels()}}.
\item Add new rows in footer with \code{\link[=add_footer]{add_footer()}}, \code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=set_header_footer_df]{set_header_footer_df()}} and \code{\link[=add_footer_row]{add_footer_row()}}.
\item Add new rows in body with \code{\link[=add_body]{add_body()}} and \code{\link[=add_body_row]{add_body_row()}}.
\item Delete columns with \code{\link[=delete_columns]{delete_columns()}}.
\item Delete a part with \code{\link[=delete_part]{delete_part()}}.
\item Use column names to separate a simple header row into multiple nested
rows with \code{\link[=separate_header]{separate_header()}}.
}
}

\section{Selectors in flextable}{

Selectors are a core feature of flextable that allow you to specify which
parts (\code{part}), rows (\code{i}) and columns (\code{j}) should be affected by formatting,
styling, or content operations. See the corresponding manual:
<\code{\link[=flextable_selectors]{Selectors in flextable}}>.
}

\seealso{
\url{https://davidgohel.github.io/flextable/},
\url{https://ardata-fr.github.io/flextable-book/}, \code{\link[=flextable]{flextable()}}
}
\author{
\strong{Maintainer}: David Gohel \email{david.gohel@ardata.fr}

Authors:
\itemize{
  \item David Gohel \email{david.gohel@ardata.fr}
  \item Panagiotis Skintzos \email{panagiotis.skintzos@ardata.fr}
}

Other contributors:
\itemize{
  \item ArData [copyright holder]
  \item Clementine Jager [contributor]
  \item Eli Daniels [contributor]
  \item Quentin Fazilleau [contributor]
  \item Maxim Nazarov [contributor]
  \item Titouan Robert [contributor]
  \item Michael Barrowman [contributor]
  \item Atsushi Yasumoto [contributor]
  \item Paul Julian [contributor]
  \item Sean Browning [contributor]
  \item Rémi Thériault (\href{https://orcid.org/0000-0003-4315-6788}{ORCID}) [contributor]
  \item Samuel Jobert [contributor]
  \item Keith Newman [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable_tabulator.R
\name{tabulator}
\alias{tabulator}
\alias{summary.tabulator}
\title{Create pivot-style summary tables}
\usage{
tabulator(
  x,
  rows,
  columns,
  datasup_first = NULL,
  datasup_last = NULL,
  hidden_data = NULL,
  row_compose = list(),
  ...
)

\method{summary}{tabulator}(object, ...)
}
\arguments{
\item{x}{an aggregated data.frame}

\item{rows}{column names to use in rows dimensions}

\item{columns}{column names to use in columns dimensions}

\item{datasup_first}{additional data that will be merged with
table and placed after the columns presenting
the row dimensions.}

\item{datasup_last}{additional data that will be merged with
table and placed at the end of the table.}

\item{hidden_data}{additional data that will be merged with
table, the columns are not presented but can be used
with \code{\link[=compose]{compose()}} or \code{\link[=mk_par]{mk_par()}} function.}

\item{row_compose}{a list of call to \code{\link[=as_paragraph]{as_paragraph()}} - these
calls will be applied to the row dimensions (the name is
used to target the displayed column).}

\item{...}{named arguments calling function \code{\link[=as_paragraph]{as_paragraph()}}.
The names are used as labels and the values are evaluated
when the flextable is created.}

\item{object}{an object returned by function
\code{tabulator()}.}
}
\value{
an object of class \code{tabulator}.
}
\description{
It tabulates a data.frame representing an aggregation
which is then transformed as a flextable with
\link[=as_flextable.tabulator]{as_flextable}. The function
allows to define any display with the syntax of flextable in
a table whose layout is showing dimensions of the aggregation
across rows and columns.
\if{html}{\out{
<img src="https://www.ardata.fr/img/flextable-imgs/ft-0-7-2-001-square.png" alt="tabulator illustration" style="width:100\%;">
}}
}
\section{Methods (by generic)}{
\itemize{
\item \code{summary(tabulator)}: call \code{summary()} to get
a data.frame describing mappings between variables
and their names in the flextable. This data.frame contains
a column named \code{col_keys} where are stored the names that
can be used for further selections.

}}
\note{
This is very first version of the function; be aware it
can evolve or change.
}
\examples{
\dontrun{
set_flextable_defaults(digits = 2, border.color = "gray")

library(data.table)
# example 1 ----
if (require("stats")) {
  dat <- aggregate(breaks ~ wool + tension,
    data = warpbreaks, mean
  )

  cft_1 <- tabulator(
    x = dat, rows = "wool",
    columns = "tension",
    `mean` = as_paragraph(as_chunk(breaks)),
    `(N)` = as_paragraph(as_chunk(length(breaks), formatter = fmt_int))
  )

  ft_1 <- as_flextable(cft_1)
  ft_1
}

# example 2 ----
if (require("ggplot2")) {
  multi_fun <- function(x) {
    list(mean = mean(x), sd = sd(x))
  }

  dat <- as.data.table(ggplot2::diamonds)
  dat <- dat[cut \%in\% c("Fair", "Good", "Very Good")]

  dat <- dat[, unlist(lapply(.SD, multi_fun),
    recursive = FALSE
  ),
  .SDcols = c("z", "y"),
  by = c("cut", "color")
  ]

  tab_2 <- tabulator(
    x = dat, rows = "color",
    columns = "cut",
    `z stats` = as_paragraph(as_chunk(fmt_avg_dev(z.mean, z.sd, digit2 = 2))),
    `y stats` = as_paragraph(as_chunk(fmt_avg_dev(y.mean, y.sd, digit2 = 2)))
  )
  ft_2 <- as_flextable(tab_2)
  ft_2 <- autofit(x = ft_2, add_w = .05)
  ft_2
}

# example 3 ----
# data.table version
dat <- melt(as.data.table(iris),
  id.vars = "Species",
  variable.name = "name", value.name = "value"
)
dat <- dat[,
  list(
    avg = mean(value, na.rm = TRUE),
    sd = sd(value, na.rm = TRUE)
  ),
  by = c("Species", "name")
]
# dplyr version
# library(dplyr)
# dat <- iris \%>\%
#   pivot_longer(cols = -c(Species)) \%>\%
#   group_by(Species, name) \%>\%
#   summarise(avg = mean(value, na.rm = TRUE),
#   sd = sd(value, na.rm = TRUE),
#   .groups = "drop")

tab_3 <- tabulator(
  x = dat, rows = c("Species"),
  columns = "name",
  `mean (sd)` = as_paragraph(
    as_chunk(avg),
    " (", as_chunk(sd), ")"
  )
)
ft_3 <- as_flextable(tab_3)
ft_3

init_flextable_defaults()
}
}
\seealso{
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}}, \code{\link[=summarizor]{summarizor()}},
\code{\link[=as_grouped_data]{as_grouped_data()}}, \code{\link[=tabulator_colnames]{tabulator_colnames()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{plot_chunk}
\alias{plot_chunk}
\title{Mini plot chunk}
\usage{
plot_chunk(
  value,
  width = 1,
  height = 0.2,
  type = "box",
  free_scale = FALSE,
  unit = "in",
  alt = "",
  ...
)
}
\arguments{
\item{value}{a numeric vector, stored in a list column.}

\item{width, height}{size of the resulting png file in inches}

\item{type}{type of the plot: 'box', 'line', 'points' or 'density'.}

\item{free_scale}{Should scales be free (TRUE or FALSE, the default value).}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{alt}{alternative text for the image (used for accessibility)}

\item{...}{arguments sent to plot functions (see \code{\link[=par]{par()}})}
}
\description{
This function is used to insert mini plots into
flextable with functions:
\itemize{
\item \code{\link[=compose]{compose()}} and \code{\link[=as_paragraph]{as_paragraph()}},
\item \code{\link[=append_chunks]{append_chunks()}},
\item \code{\link[=prepend_chunks]{prepend_chunks()}}.
}

Available plots are 'box', 'line', 'points', 'density'.
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format. With Quarto (\code{format: docx}) or
\code{rmarkdown::word_document()}, the resulting file must be repaired
with \code{\link[=repair_docx]{repair_docx()}}.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\examples{
library(data.table)
library(flextable)

z <- as.data.table(iris)
z <- z[, list(
  Sepal.Length = mean(Sepal.Length, na.rm = TRUE),
  z = list(.SD$Sepal.Length)
), by = "Species"]

ft <- flextable(z,
  col_keys = c("Species", "Sepal.Length", "box", "density")
)
ft <- mk_par(ft, j = "box", value = as_paragraph(
  plot_chunk(
    value = z, type = "box",
    border = "red", col = "transparent"
  )
))
ft <- mk_par(ft, j = "density", value = as_paragraph(
  plot_chunk(value = z, type = "dens", col = "red")
))
ft <- set_table_properties(ft, layout = "autofit", width = .6)
ft <- set_header_labels(ft, box = "boxplot", density = "density")
theme_vanilla(ft)
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{knit_print.flextable}
\alias{knit_print.flextable}
\title{Render flextable in knitr documents}
\usage{
\method{knit_print}{flextable}(x, ...)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{unused.}
}
\description{
This function is called automatically by knitr to display
a flextable in R Markdown and Quarto documents. You do not need to call
it directly.

Supported output formats: HTML, Word (docx), PDF, PowerPoint (pptx) and
Typst (\code{format: typst}, Quarto only). For other formats (e.g.,
\code{github_document}, \code{beamer_presentation}), the table is rendered as a
PNG image.
}
\section{Getting started}{


No special setup is needed: place a flextable object in a code chunk
and it will be rendered in the output document.

Add a caption with \code{\link[=set_caption]{set_caption()}}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ft <- set_caption(ft, caption = "My table caption")
}\if{html}{\out{</div>}}

In Quarto documents, use chunk options \code{tbl-cap} and \code{label} instead:

\if{html}{\out{<div class="sourceCode">}}\preformatted{```\{r\}
#| label: tbl-mytable
#| tbl-cap: "My table caption"
ft
```
}\if{html}{\out{</div>}}
}

\section{Captions}{


\strong{Recommended method:} use \code{\link[=set_caption]{set_caption()}} to define the caption
directly on the flextable object. When \code{set_caption()} is used,
chunk options related to captions are ignored.

\strong{Alternative (R Markdown only):} use knitr chunk options \code{tab.cap}
and \code{tab.id}:\tabular{lcc}{
   \strong{Description} \tab \strong{Chunk option} \tab \strong{Default} \cr
   Caption text \tab tab.cap \tab NULL \cr
   Caption id/bookmark \tab tab.id \tab NULL \cr
   Caption on top of the table \tab tab.topcaption \tab TRUE \cr
   Caption sequence identifier \tab tab.lp \tab "tab:" \cr
   Word style for captions \tab tab.cap.style \tab NULL \cr
}


\strong{Bookdown:} cross-references use the pattern
\verb{\\@ref(tab:chunk_label)}. The usual bookdown numbering applies.

\strong{Quarto:} cross-references use \verb{@tbl-chunk_label}. To embed
cross-references or other Quarto markdown inside flextable cells,
use \code{\link[=as_qmd]{as_qmd()}} with the \code{flextable-qmd} extension
(see \code{\link[=use_flextable_qmd]{use_flextable_qmd()}}).
}

\section{Chunk options}{


Use \code{knitr::opts_chunk$set(...)} to set defaults for the whole document.

\strong{All formats:}
\itemize{
\item \code{ft.align}: table alignment, one of \code{'left'}, \code{'center'} (default)
or \code{'right'}.
}

\strong{HTML:}
\itemize{
\item \code{ft.htmlscroll}: \code{TRUE} or \code{FALSE} (default) to enable horizontal
scrolling. See \code{\link[=set_table_properties]{set_table_properties()}} for finer control.
}

\strong{Word:}
\itemize{
\item \code{ft.split}: allow rows to break across pages (\code{TRUE} by default).
}

\strong{PDF:}
\itemize{
\item \code{ft.tabcolsep}: space between text and cell borders (default 0 pt).
\item \code{ft.arraystretch}: row height multiplier (default 1.5).
\item \code{ft.latex.float}: floating placement. One of \code{'none'} (default),
\code{'float'}, \code{'wrap-r'}, \code{'wrap-l'}, \code{'wrap-i'}, \code{'wrap-o'}.
}

\strong{PowerPoint:}
\itemize{
\item \code{ft.left}, \code{ft.top}: top-left coordinates of the table placeholder
in inches (defaults: 1 and 2).
}
}

\section{Word with officedown}{


When using \code{officedown::rdocx_document()}, additional caption options
are available:\tabular{lcc}{
   \strong{Description} \tab \strong{Chunk option} \tab \strong{Default} \cr
   Numbering prefix \tab tab.cap.pre \tab "Table " \cr
   Numbering suffix \tab tab.cap.sep \tab ": " \cr
   Title number depth \tab tab.cap.tnd \tab 0 \cr
   Caption prefix formatting \tab tab.cap.fp_text \tab \code{fp_text_lite(bold=TRUE)} \cr
   Title number / table number separator \tab tab.cap.tns \tab "-" \cr
}
}

\section{Quarto}{


flextable works natively in Quarto documents for HTML, PDF, Word and
Typst.

The \code{flextable-qmd} Lua filter extension enables Quarto markdown
inside flextable cells: cross-references (\verb{@tbl-xxx}, \verb{@fig-xxx}),
bold/italic, links, math, inline code and shortcodes.
See \code{\link[=as_qmd]{as_qmd()}} and \code{\link[=use_flextable_qmd]{use_flextable_qmd()}} for setup instructions.
}

\section{PDF accessibility (PDF/UA-2)}{


Quarto's \code{pdf-standard: ua-2} injects \verb{\\DocumentMetadata\{tagging=on\}}
in the LaTeX preamble, activating 'tagpdf'.
This code patches LaTeX commands at compile time to insert
PDF structure tags. Neither Quarto nor flextable control this process.

The tagging code does not yet support \code{longtable} + \code{colortbl},
\code{booktabs} rules, and \code{cline}, which flextable relies on. Compilation
fails with \verb{\\ERRORtbl/row}. When these upstream issues are resolved,
flextable PDF output will support tagging without changes. Other formats are
not affected.
}

\section{PDF limitations}{


The following properties are not supported in PDF output:
\code{padding.top}, \code{padding.bottom}, \code{line_spacing} and row \code{height}.
Justified text is converted to left-aligned.

To use system fonts, set \code{latex_engine: xelatex} in the YAML
header (the default \code{pdflatex} engine does not support them).

See \code{\link[=add_latex_dep]{add_latex_dep()}} when caching flextable results.
}

\section{PowerPoint limitations}{


PowerPoint only supports fixed table layout. Use \code{\link[=autofit]{autofit()}} to
adjust column widths. Images inside table cells are not supported
(this is a PowerPoint limitation).
}

\section{HTML note}{


HTML output uses Shadow DOM to isolate table styles from the rest
of the page.
}

\examples{
\dontrun{
library(rmarkdown)
if (pandoc_available() &&
  pandoc_version() > numeric_version("2")) {
  demo_loop <- system.file(
    package = "flextable",
    "examples/rmd",
    "demo.Rmd"
  )
  rmd_file <- tempfile(fileext = ".Rmd")
  file.copy(demo_loop, to = rmd_file, overwrite = TRUE)
  render(
    input = rmd_file, output_format = "html_document",
    output_file = "demo.html"
  )
}
}
}
\seealso{
\code{\link[=set_caption]{set_caption()}}, \code{\link[=as_qmd]{as_qmd()}}, \code{\link[=use_flextable_qmd]{use_flextable_qmd()}},
\code{\link[=paginate]{paginate()}}

Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{to_html.flextable}
\alias{to_html.flextable}
\title{Get HTML code as a string}
\usage{
\method{to_html}{flextable}(x, type = c("table", "img"), ...)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{type}{output type. one of "table" or "img".}

\item{...}{unused}
}
\value{
If \code{type='img'}, the result will be a string
containing HTML code of an image tag, otherwise, the
result will be a string containing HTML code of
a table tag.
}
\description{
Generate HTML code of corresponding
flextable as an HTML table or an HTML image.
}
\examples{
library(officer)
library(flextable)
x <- to_html(as_flextable(cars))
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=htmltools_value]{htmltools_value()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_grobs.R
\name{plot.flextableGrob}
\alias{plot.flextableGrob}
\title{plot a flextable grob}
\usage{
\method{plot}{flextableGrob}(x, ...)
}
\arguments{
\item{x}{a flextableGrob object}

\item{...}{additional arguments passed to other functions}
}
\description{
plot a flextable grob
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{dim.flextable}
\alias{dim.flextable}
\title{Get column widths and row heights of a flextable}
\usage{
\method{dim}{flextable}(x)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}
}
\description{
returns widths and heights for each table columns and rows.
Values are expressed in inches.
}
\examples{
ftab <- flextable(head(iris))
dim(ftab)
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}},
\code{\link[=width]{width()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df_printer.R
\name{as_flextable.table}
\alias{as_flextable.table}
\title{Transform a 'table' object into a flextable}
\usage{
\method{as_flextable}{table}(x, ...)
}
\arguments{
\item{x}{table object}

\item{...}{arguments used by \code{\link[=proc_freq]{proc_freq()}}.}
}
\description{
produce a flextable describing a
count table produced by function \code{table()}.

This function uses the \code{\link[=proc_freq]{proc_freq()}} function.
}
\examples{
tab <- with(warpbreaks, table(wool, tension))
ft <- as_flextable(tab)
ft
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{fontsize}
\alias{fontsize}
\title{Set font size}
\usage{
fontsize(x, i = NULL, j = NULL, size = 11, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{size}{integer value (points)}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Change the font size of selected rows and columns of a flextable.
}
\examples{
ft <- flextable(head(iris))
ft <- fontsize(ft, size = 14, part = "header")
ft <- fontsize(ft, size = 14, j = 2)
ft <- fontsize(ft, size = 7, j = 3)
ft
}
\seealso{
Other sugar functions for table style: 
\code{\link{align}()},
\code{\link{bg}()},
\code{\link{bold}()},
\code{\link{color}()},
\code{\link{empty_blanks}()},
\code{\link{font}()},
\code{\link{highlight}()},
\code{\link{italic}()},
\code{\link{keep_with_next}()},
\code{\link{line_spacing}()},
\code{\link{padding}()},
\code{\link{rotate}()},
\code{\link{style}()},
\code{\link{tab_settings}()},
\code{\link{valign}()}
}
\concept{sugar functions for table style}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_equation}
\alias{as_equation}
\title{Equation chunk}
\usage{
as_equation(x, width = 1, height = 0.2, unit = "in", props = NULL)
}
\arguments{
\item{x}{values containing the 'MathJax' equations}

\item{width, height}{size of the resulting equation}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{props}{an \code{\link[=fp_text_default]{fp_text_default()}} or \code{\link[officer:fp_text]{officer::fp_text()}} object to be used to format the text.
If not specified, it will be the default value corresponding to the cell.}
}
\description{
This function is used to insert equations into
flextable.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.

To use this function, package 'equatags' is required;
also \code{equatags::mathjax_install()} must be executed only once
to install necessary dependencies.
}
\examples{
library(flextable)
if (require("equatags")) {
  eqs <- c(
    "(ax^2 + bx + c = 0)",
    "a \\\\ne 0",
    "x = {-b \\\\pm \\\\sqrt{b^2-4ac} \\\\over 2a}"
  )
  df <- data.frame(formula = eqs)
  df


  ft <- flextable(df)
  ft <- compose(
    x = ft, j = "formula",
    value = as_paragraph(as_equation(formula, width = 2, height = .5))
  )
  ft <- align(ft, align = "center", part = "all")
  ft <- width(ft, width = 2)
  ft
}
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/footnote.R
\name{footnote}
\alias{footnote}
\title{Add footnotes to flextable}
\usage{
footnote(
  x,
  i = NULL,
  j = NULL,
  value,
  ref_symbols = NULL,
  part = "body",
  inline = FALSE,
  sep = "; ",
  symbol_sep = ""
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{value}{a call to function \code{\link[=as_paragraph]{as_paragraph()}}.}

\item{ref_symbols}{character value, symbols to append that will be used
as references to notes.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}

\item{inline}{whether to add footnote on same line as previous footnote or not}

\item{sep}{used only when inline = TRUE, character string to use as
a separator between footnotes.}

\item{symbol_sep}{separator to insert between multiple
footnote symbols in the same cell (e.g. \code{","} to
produce \verb{1,2} instead of \code{12}). Default is \code{""}
(no separator, backward compatible).}
}
\description{
The function let add footnotes to a flextable object
by adding some symbols in the flextable and associated notes in
the footer of the flextable.

Symbols are added to the cells designated by the selection \code{i}
and \code{j}. If you use i = c(1,3) and j = c(2,5), then you will
add the symbols (or the repeated symbol) to cells \verb{[1,2]}
and \verb{[3,5]}.

\if{html}{\out{
<img src="https://www.ardata.fr/img/flextable-imgs/flextable-016.png" alt="add_footer illustration" style="width:100\%;">

See https://www.ardata.fr/en/flextable-gallery/2022-06-23-separate-headers/ for the example
shown
}}
}
\examples{
ft_1 <- flextable(head(iris))
ft_1 <- footnote(ft_1,
  i = 1, j = 1:3,
  value = as_paragraph(
    c(
      "This is footnote one",
      "This is footnote two",
      "This is footnote three"
    )
  ),
  ref_symbols = c("a", "b", "c"),
  part = "header"
)
ft_1 <- valign(ft_1, valign = "bottom", part = "header")
ft_1 <- autofit(ft_1)

ft_2 <- flextable(head(iris))
ft_2 <- autofit(ft_2)
ft_2 <- footnote(ft_2,
  i = 1, j = 1:2,
  value = as_paragraph(
    c(
      "This is footnote one",
      "This is footnote two"
    )
  ),
  ref_symbols = c("a", "b"),
  part = "header", inline = TRUE
)
ft_2 <- footnote(ft_2,
  i = 1, j = 3:4,
  value = as_paragraph(
    c(
      "This is footnote three",
      "This is footnote four"
    )
  ),
  ref_symbols = c("c", "d"),
  part = "header", inline = TRUE
)
ft_2

ft_3 <- flextable(head(iris))
ft_3 <- autofit(ft_3)
ft_3 <- footnote(
  x = ft_3, i = 1:3, j = 1:3,
  ref_symbols = "a",
  value = as_paragraph("This is footnote one")
)
ft_3
}
\seealso{
Other functions to compose cell content:
\code{\link[=append_chunks]{append_chunks()}},
\code{\link[=as_paragraph]{as_paragraph()}},
\code{\link[=compose]{compose()}},
\code{\link[=labelizor]{labelizor()}},
\code{\link[=prepend_chunks]{prepend_chunks()}},
\code{\link[=void]{void()}}
}
\concept{cell_content_composition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{hline_bottom}
\alias{hline_bottom}
\title{Set the bottom border of a table part}
\usage{
hline_bottom(x, j = NULL, border = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{border}{border properties defined by a call to
\code{\link[officer:fp_border]{officer::fp_border()}}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{hline_bottom()} draws a horizontal line at the \strong{very bottom}
of a table part. It does not accept a row selector \code{i} because
it always targets the last row.

Unlike \code{\link[=hline]{hline()}}, which adds inner lines below arbitrary rows,
\code{hline_bottom()} is meant for the outer bottom edge of a part.
}
\examples{
library(officer)
big_border <- fp_border(color = "orange", width = 3)

ft <- flextable(head(iris))
ft <- border_remove(x = ft)

# add a thick line at the bottom of the body
ft <- hline_bottom(ft, part = "body", border = big_border)
ft
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_h]{border_inner_h()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=border_remove]{border_remove()}},
\code{\link[=hline]{hline()}},
\code{\link[=hline_top]{hline_top()}},
\code{\link[=surround]{surround()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}},
\code{\link[=vline_right]{vline_right()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.tabular.R
\name{as_flextable.tabular}
\alias{as_flextable.tabular}
\title{Transform a 'tables::tabular' object into a flextable}
\usage{
\method{as_flextable}{tabular}(
  x,
  spread_first_col = FALSE,
  fp_p = fp_par(text.align = "center", padding.top = 4),
  row_title = as_paragraph(as_chunk(.row_title)),
  add_tab = FALSE,
  ...
)
}
\arguments{
\item{x}{object produced by \code{\link[tables:tabular]{tables::tabular()}}.}

\item{spread_first_col}{if TRUE, first row is spread as a new line separator
instead of being a column. This helps to reduce the width and allows for
clear divisions.}

\item{fp_p}{paragraph formatting properties associated with row titles,
see \code{\link[officer:fp_par]{officer::fp_par()}}.}

\item{row_title}{a call to \code{\link[=as_paragraph]{as_paragraph()}} - it
will be applied to the row titles if any
when \code{spread_first_col=TRUE}.}

\item{add_tab}{adds a tab in front of "list_data"
label lines (located in column \code{.type}).}

\item{...}{unused argument}
}
\description{
Produce a flextable from a 'tabular' object
produced with function \code{\link[tables:tabular]{tables::tabular()}}.

When \code{as_flextable.tabular=TRUE}, the first column is
used as row separator acting as a row title. It can
be formated with arguments \code{fp_p} (the formatting
properties of the paragraph) and \code{row_title} that
specifies the content and eventually formattings
of the content.

Two hidden columns can be used for conditional formatting
after the creation of the flextable (use only when
\code{spread_first_col=TRUE}):
\itemize{
\item The column \code{.row_title} that contains the title label
\item The column \code{.type} that can contain the following values:
\itemize{
\item "one_row": Indicates that there is only one row for this group. In this case, the row is not expanded with a title above.
\item "list_title": Indicates a row that serves as a title for the data that are displayed after it.
\item "list_data": Indicates rows that follow a title and contain data to be displayed.
}
}

The result is paginated (see \code{\link[=paginate]{paginate()}}).
}
\examples{
if (require("tables")) {
  set.seed(42)
  genders <- c("Male", "Female")
  status <- c("low", "medium", "high")
  Sex <- factor(sample(genders, 100, rep = TRUE))
  Status <- factor(sample(status, 100, rep = TRUE))
  z <- rnorm(100) + 5
  fmt <- function(x) {
    s <- format(x, digits = 2)
    even <- ((1:length(s)) \%\% 2) == 0
    s[even] <- sprintf("(\%s)", s[even])
    s
  }
  tab <- tabular(
    Justify(c) * Heading() * z *
      Sex * Heading(Statistic) *
      Format(fmt()) *
      (mean + sd) ~ Status
  )
  as_flextable(tab)
}

if (require("tables")) {
  tab <- tabular(
    (Species + 1) ~ (n = 1) + Format(digits = 2) *
      (Sepal.Length + Sepal.Width) * (mean + sd),
    data = iris
  )
  as_flextable(tab)
}

if (require("tables")) {
  x <- tabular((Factor(gear, "Gears") + 1)
  * ((n = 1) + Percent()
      + (RowPct = Percent("row"))
      + (ColPct = Percent("col")))
  ~ (Factor(carb, "Carburetors") + 1)
    * Format(digits = 1), data = mtcars)

  ft <- as_flextable(
    x,
    spread_first_col = TRUE,
    row_title = as_paragraph(
      colorize("Gears: ", color = "#666666"),
      colorize(as_b(.row_title), color = "red")
    )
  )
  ft
}

if (require("tables")) {
  tab <- tabular(
    (mean + mean) * (Sepal.Length + Sepal.Width) ~ 1,
    data = iris
  )
  as_flextable(tab)
}
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.htest]{as_flextable.htest()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/printers.R
\name{htmltools_value}
\alias{htmltools_value}
\title{Convert a flextable to an HTML object}
\usage{
htmltools_value(
  x,
  ft.align = NULL,
  ft.shadow = NULL,
  extra_dependencies = NULL
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{ft.align}{flextable alignment, supported values are 'left', 'center' and 'right'.}

\item{ft.shadow}{deprecated.}

\item{extra_dependencies}{a list of HTML dependencies to
add in the HTML output.}
}
\value{
an object marked as \link[htmltools:HTML]{htmltools::HTML} ready to be used within
a call to \code{shiny::renderUI} for example.
}
\description{
get a \code{\link[htmltools:div]{htmltools::div()}} from a flextable object.
This can be used in a shiny application. For an output within
"R Markdown" document, use \link{knit_print.flextable}.
}
\examples{
htmltools_value(flextable(iris[1:5, ]))
}
\seealso{
Other functions for flextable output and export:
\code{\link[=df_printer]{df_printer()}},
\code{\link[=flextable_to_rmd]{flextable_to_rmd()}},
\code{\link[=gen_grob]{gen_grob()}},
\code{\link[=knit_print.flextable]{knit_print.flextable()}},
\code{\link[=plot.flextable]{plot.flextable()}},
\code{\link[=print.flextable]{print.flextable()}},
\code{\link[=repair_docx]{repair_docx()}},
\code{\link[=save_as_docx]{save_as_docx()}},
\code{\link[=save_as_html]{save_as_html()}},
\code{\link[=save_as_image]{save_as_image()}},
\code{\link[=save_as_pptx]{save_as_pptx()}},
\code{\link[=save_as_rtf]{save_as_rtf()}},
\code{\link[=to_html.flextable]{to_html.flextable()}},
\code{\link[=wrap_flextable]{wrap_flextable()}}
}
\concept{flextable_output_export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doc-selectors.R
\name{args_x_i_part_no_all}
\alias{args_x_i_part_no_all}
\title{internal utils for roxygen tags reuse}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
internal utils for roxygen tags reuse
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{hyperlink_text}
\alias{hyperlink_text}
\title{Hyperlink chunk}
\usage{
hyperlink_text(x, props = NULL, formatter = format_fun, url, ...)
}
\arguments{
\item{x}{text or any element that can be formatted as text
with function provided in argument \code{formatter}.}

\item{props}{an \code{\link[=fp_text_default]{fp_text_default()}} or \code{\link[officer:fp_text]{officer::fp_text()}} object to be used to format the text.
If not specified, it will be the default value corresponding to the cell.}

\item{formatter}{a function that will format x as a character vector.}

\item{url}{url to be used}

\item{...}{additional arguments for \code{formatter} function.}
}
\description{
The function lets add hyperlinks within flextable
objects.

It is used to add it to the content of a cell of the
flextable with the functions \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}
or \code{\link[=prepend_chunks]{prepend_chunks()}}.

URL are not encoded, they are preserved 'as is'.
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format. With Quarto (\code{format: docx}) or
\code{rmarkdown::word_document()}, the resulting file must be repaired
with \code{\link[=repair_docx]{repair_docx()}}.
}
\examples{
dat <- data.frame(
  col = "Google it",
  href = "https://www.google.fr/search?source=hp&q=flextable+R+package",
  stringsAsFactors = FALSE
)

ftab <- flextable(dat)
ftab <- compose(
  x = ftab, j = "col",
  value = as_paragraph(
    "This is a link: ",
    hyperlink_text(x = col, url = href)
  )
)
ftab
}
\seealso{
\code{\link[=compose]{compose()}}

Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=grid_chunk]{grid_chunk()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_flextable.R
\name{merge_h_range}
\alias{merge_h_range}
\title{Rowwise merge of a range of columns}
\usage{
merge_h_range(x, i = NULL, j1 = NULL, j2 = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j1, j2}{selected columns that will define the range of columns to merge.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
Merge flextable columns into a single one for each selected rows. All
columns must be consecutive.
}
\examples{
ft <- flextable(head(mtcars), cwidth = .5)
ft <- theme_box(ft)
ft <- merge_h_range(ft, i = ~ cyl == 6, j1 = "am", j2 = "carb")
ft <- flextable::align(ft, i = ~ cyl == 6, align = "center")
ft
}
\seealso{
Other cell merging functions:
\code{\link[=merge_at]{merge_at()}},
\code{\link[=merge_h]{merge_h()}},
\code{\link[=merge_none]{merge_none()}},
\code{\link[=merge_v]{merge_v()}}
}
\concept{cell merging functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizor.R
\name{fmt_int}
\alias{fmt_int}
\title{Format numbers as integers}
\usage{
fmt_int(x)
}
\arguments{
\item{x}{numeric values}
}
\description{
Format numeric values as integers (no decimals).
}
\examples{
library(flextable)

df <- data.frame(zz = 1.23)

ft_1 <- flextable(df)
ft_1 <- mk_par(
  x = ft_1, j = 1, part = "body",
  value = as_paragraph(as_chunk(zz, formatter = fmt_int))
)
ft_1 <- autofit(ft_1)
ft_1
}
\seealso{
\code{\link[=tabulator]{tabulator()}}, \code{\link[=mk_par]{mk_par()}}

Other text formatter functions: 
\code{\link{fmt_2stats}()},
\code{\link{fmt_avg_dev}()},
\code{\link{fmt_dbl}()},
\code{\link{fmt_header_n}()},
\code{\link{fmt_n_percent}()},
\code{\link{fmt_pct}()},
\code{\link{fmt_signif_after_zeros}()}
}
\concept{text formatter functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{use_flextable_qmd}
\alias{use_flextable_qmd}
\title{Install the flextable-qmd Quarto extension}
\usage{
use_flextable_qmd(path = ".", quiet = FALSE)
}
\arguments{
\item{path}{Path to the Quarto project root. Defaults
to the current working directory.}

\item{quiet}{If \code{TRUE}, suppress informational messages.}
}
\value{
The path to the installed extension (invisibly).
}
\description{
Copies the \code{flextable-qmd} Quarto extension (bundled with
flextable) into the \verb{_extensions/} directory of a
Quarto project. The extension provides Lua filters
that resolve Quarto markdown content produced by
\code{\link[=as_qmd]{as_qmd()}} inside flextable cells for HTML, PDF,
Typst and Word (docx) output formats.

After installation, add the filter to your document
or project YAML:

\if{html}{\out{<div class="sourceCode yaml">}}\preformatted{filters:
  - flextable-qmd
}\if{html}{\out{</div>}}

This single filter is all that is required for HTML,
PDF and Typst (\code{format: typst}) output.

For Word (docx) output with labelled flextable chunks
(e.g. \verb{#| label: tbl-xxx}), add the post-render filter
to remove the wrapper table Quarto creates around the
flextable (this extra filter is not needed for Typst,
where the table is embedded directly in the figure):

\if{html}{\out{<div class="sourceCode yaml">}}\preformatted{filters:
  - flextable-qmd
  - at: post-render
    path: _extensions/flextable-qmd/unwrap-float.lua
}\if{html}{\out{</div>}}
}
\examples{
\dontrun{
use_flextable_qmd()
}
}
\seealso{
\code{\link[=as_qmd]{as_qmd()}} for creating Quarto markdown chunks,
\code{\link[=knit_print.flextable]{knit_print.flextable()}} for rendering options in knitr documents.

Other flextable configuration:
\code{\link[=add_latex_dep]{add_latex_dep()}},
\code{\link[=use_df_printer]{use_df_printer()}},
\code{\link[=use_model_printer]{use_model_printer()}}
}
\concept{flextable_configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_images.R
\name{grid_chunk}
\alias{grid_chunk}
\title{Grid Graphics chunk}
\usage{
grid_chunk(value, width = 1, height = 0.2, unit = "in", res = 300, alt = "")
}
\arguments{
\item{value}{grid objects, stored in a list column; or a list
of grid objects.}

\item{width, height}{size of the resulting png file}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{res}{resolution of the png image in ppi}

\item{alt}{alternative text for the image (used for accessibility)}
}
\description{
This function is used to insert grid objects into
flextable with functions:
\itemize{
\item \code{\link[=compose]{compose()}} and \code{\link[=as_paragraph]{as_paragraph()}},
\item \code{\link[=append_chunks]{append_chunks()}},
\item \code{\link[=prepend_chunks]{prepend_chunks()}}.
}
}
\note{
This chunk option requires package officedown in a R Markdown
context with Word output format. With Quarto (\code{format: docx}) or
\code{rmarkdown::word_document()}, the resulting file must be repaired
with \code{\link[=repair_docx]{repair_docx()}}.

PowerPoint cannot mix images and text in a paragraph, images
are removed when outputing to PowerPoint format.
}
\examples{
library(flextable)
ft_1 <- flextable(head(cars))
if (require("grid")) {
  ft_1 <- prepend_chunks(
    x = ft_1, i = 2, j = 2,
    grid_chunk(
      list(
        circleGrob(gp = gpar(
          fill = "#ec11c2",
          col = "transparent"
        ))
      ),
      width = .15, height = .15
    )
  )
}
ft_1
}
\seealso{
Other chunk elements for paragraph:
\code{\link[=as_b]{as_b()}},
\code{\link[=as_bracket]{as_bracket()}},
\code{\link[=as_chunk]{as_chunk()}},
\code{\link[=as_equation]{as_equation()}},
\code{\link[=as_highlight]{as_highlight()}},
\code{\link[=as_i]{as_i()}},
\code{\link[=as_image]{as_image()}},
\code{\link[=as_qmd]{as_qmd()}},
\code{\link[=as_strike]{as_strike()}},
\code{\link[=as_sub]{as_sub()}},
\code{\link[=as_sup]{as_sup()}},
\code{\link[=as_word_field]{as_word_field()}},
\code{\link[=colorize]{colorize()}},
\code{\link[=gg_chunk]{gg_chunk()}},
\code{\link[=hyperlink_text]{hyperlink_text()}},
\code{\link[=linerange]{linerange()}},
\code{\link[=minibar]{minibar()}},
\code{\link[=plot_chunk]{plot_chunk()}}
}
\concept{chunk elements for paragraph}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitters.R
\name{split_to_pages}
\alias{split_to_pages}
\title{Split a flextable into pages by rows and columns}
\usage{
split_to_pages(
  x,
  max_width = NULL,
  max_height = NULL,
  rep_cols = NULL,
  group = integer(0),
  unit = "in"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{max_width}{Maximum width per page (in inches by default).
\code{NULL} means no column splitting.}

\item{max_height}{Maximum height per page, including header and
footer (in inches by default). \code{NULL} means no row splitting.}

\item{rep_cols}{Columns to repeat on every horizontal page. Can be
a character vector of column names, an integer vector of column
positions, or \code{NULL} (default, no repetition).}

\item{group}{Integer vector of body row indices that start a new
group. Default is \code{integer(0)}.}

\item{unit}{Unit for \code{max_width} and \code{max_height}, one of
\code{"in"}, \code{"cm"}, \code{"mm"}.}
}
\value{
A list of flextable objects.
}
\description{
Split a flextable into a list of flextables that fit
within the given height and width constraints. Rows are split
first, then columns are split within each row page.

This is a convenience wrapper around \code{\link[=split_rows]{split_rows()}} and
\code{\link[=split_columns]{split_columns()}}.
}
\examples{
ft <- flextable(iris)
ft_pages <- split_to_pages(ft, max_width = 4, max_height = 5)
length(ft_pages)
}
\seealso{
Other row and column operations:
\code{\link[=add_body]{add_body()}},
\code{\link[=add_body_row]{add_body_row()}},
\code{\link[=add_footer]{add_footer()}},
\code{\link[=add_footer_lines]{add_footer_lines()}},
\code{\link[=add_footer_row]{add_footer_row()}},
\code{\link[=add_header]{add_header()}},
\code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_header_row]{add_header_row()}},
\code{\link[=delete_columns]{delete_columns()}},
\code{\link[=delete_part]{delete_part()}},
\code{\link[=delete_rows]{delete_rows()}},
\code{\link[=paginate]{paginate()}},
\code{\link[=separate_header]{separate_header()}},
\code{\link{set_header_footer_df}},
\code{\link[=set_header_labels]{set_header_labels()}},
\code{\link[=split_columns]{split_columns()}},
\code{\link[=split_rows]{split_rows()}}
}
\concept{table_structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{indentation}
\alias{indentation}
\title{Set paragraph indents}
\usage{
indentation(
  x,
  i = NULL,
  j = NULL,
  first_line = NULL,
  hanging = NULL,
  part = "body"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{first_line}{first-line indent in pts (points). Use NA to unset.}

\item{hanging}{hanging indent in pts (points). Use NA to unset. If both
\code{first_line} and \code{hanging} are provided, \code{hanging} wins.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Change the first-line or hanging indent of paragraphs in
selected rows and columns of a flextable.

A \emph{first-line} indent moves only the first line of the paragraph to the
right. A \emph{hanging} indent moves all lines \strong{but} the first one to the
right; the indented lines start at the paragraph left padding plus the
hanging value, so \code{hanging} should not exceed the left padding
(see \code{\link[=padding]{padding()}}).
}
\examples{
ft_1 <- flextable(head(iris))
ft_1 <- mk_par(
  x = ft_1, i = 1, j = 1, part = "header",
  value = as_paragraph("A rather long header that will wrap on two lines")
)
ft_1 <- padding(ft_1, i = 1, j = 1, padding.left = 30, part = "header")
ft_1 <- indentation(ft_1, i = 1, j = 1, hanging = 20, part = "header")
ft_1 <- width(ft_1, j = 1, width = 1.2)
ft_1
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{highlight}
\alias{highlight}
\title{Set text highlight color}
\usage{
highlight(x, i = NULL, j = NULL, color = "yellow", part = "body", source = j)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{color}{color to use as text highlighting color.
If a function, the function must return a character vector of colors.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}

\item{source}{if color is a function, source specifies the dataset column to be used
as an argument to \code{color}. This is only useful when j is colored with values contained in
other columns.}
}
\description{
Change the text highlight color of selected rows and
columns of a flextable. A function can be used instead of
fixed colors.

When \code{color} is a function, it is possible to color cells based on values
located in other columns; using hidden columns (those not used by
argument \code{colkeys}) is a common use case. The argument \code{source}
must be used to define the columns to be used for the color
definition, and the argument \code{j} must be used to define where to apply
the colors and only accepts values from \code{colkeys}.
}
\examples{
my_color_fun <- function(x) {
  out <- rep("yellow", length(x))
  out[x < quantile(x, .75)] <- "pink"
  out[x < quantile(x, .50)] <- "wheat"
  out[x < quantile(x, .25)] <- "gray90"
  out
}
ft <- flextable(head(mtcars, n = 10))
ft <- highlight(ft, j = "disp", i = ~ disp > 200, color = "yellow")
ft <- highlight(ft, j = ~ drat + wt + qsec, color = my_color_fun)
ft
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=empty_blanks]{empty_blanks()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/borders.R
\name{surround}
\alias{surround}
\title{Surround cells with borders}
\usage{
surround(
  x,
  i = NULL,
  j = NULL,
  border = NULL,
  border.top = NULL,
  border.bottom = NULL,
  border.left = NULL,
  border.right = NULL,
  part = "body"
)
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{i}{row selector, see section \emph{Row selection with the \code{i} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{border}{border (shortcut for top, bottom, left and right)}

\item{border.top}{border top}

\item{border.bottom}{border bottom}

\item{border.left}{border left}

\item{border.right}{border right}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
\code{surround()} draws borders around specific cells, highlighting
them individually.

To set borders for the whole table, use \code{\link[=border_outer]{border_outer()}},
\code{\link[=border_inner_h]{border_inner_h()}} and \code{\link[=border_inner_v]{border_inner_v()}}.

All the following functions also support the
row and column selector \code{i} and \code{j}:
\itemize{
\item \code{\link[=hline]{hline()}}: set bottom borders (inner horizontal)
\item \code{\link[=vline]{vline()}}: set right borders (inner vertical)
\item \code{\link[=hline_top]{hline_top()}}: set the top border (outer horizontal)
\item \code{\link[=vline_left]{vline_left()}}: set the left border (outer vertical)
}
}
\examples{
library(officer)
library(flextable)

# cell to highlight
vary_i <- 1:3
vary_j <- 1:3

std_border <- fp_border(color = "orange")

ft <- flextable(head(iris))
ft <- border_remove(x = ft)
ft <- border_outer(x = ft, border = std_border)

for (id in seq_along(vary_i)) {
  ft <- bg(
    x = ft,
    i = vary_i[id],
    j = vary_j[id], bg = "yellow"
  )
  ft <- surround(
    x = ft,
    i = vary_i[id],
    j = vary_j[id],
    border.left = std_border,
    border.right = std_border,
    part = "body"
  )
}

ft <- autofit(ft)
ft
# # render
# print(ft, preview = "pptx")
# print(ft, preview = "docx")
# print(ft, preview = "pdf")
# print(ft, preview = "html")
}
\seealso{
Other borders management:
\code{\link[=border_inner]{border_inner()}},
\code{\link[=border_inner_h]{border_inner_h()}},
\code{\link[=border_inner_v]{border_inner_v()}},
\code{\link[=border_outer]{border_outer()}},
\code{\link[=border_remove]{border_remove()}},
\code{\link[=hline]{hline()}},
\code{\link[=hline_bottom]{hline_bottom()}},
\code{\link[=hline_top]{hline_top()}},
\code{\link[=vline]{vline()}},
\code{\link[=vline_left]{vline_left()}},
\code{\link[=vline_right]{vline_right()}}
}
\concept{borders management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{fp_text_default}
\alias{fp_text_default}
\title{Create text formatting with flextable defaults}
\usage{
fp_text_default(
  color = flextable_global$defaults$font.color,
  font.size = flextable_global$defaults$font.size,
  bold = FALSE,
  italic = FALSE,
  underlined = FALSE,
  strike = FALSE,
  font.family = flextable_global$defaults$font.family,
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  vertical.align = "baseline",
  shading.color = "transparent"
)
}
\arguments{
\item{color}{font color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{font.size}{font size (in point) - 0 or positive integer value.}

\item{bold}{is bold}

\item{italic}{is italic}

\item{underlined}{is underlined}

\item{strike}{is strikethrough}

\item{font.family}{single character value. Specifies the font to
be used to format characters in the Unicode range (U+0000-U+007F).}

\item{cs.family}{optional font to be used to format
characters in a complex script Unicode range. For example, Arabic
text might be displayed using the "Arial Unicode MS" font.}

\item{eastasia.family}{optional font to be used to
format characters in an East Asian Unicode range. For example,
Japanese text might be displayed using the "MS Mincho" font.}

\item{hansi.family}{optional. Specifies the font to be used to format
characters in a Unicode range which does not fall into one of the
other categories.}

\item{vertical.align}{single character value specifying font vertical alignments.
Expected value is one of the following : default \code{'baseline'}
or \code{'subscript'} or \code{'superscript'}}

\item{shading.color}{shading color - a single character value specifying
a valid color (e.g. "#000000" or "black").}
}
\description{
Create a \code{\link[officer:fp_text]{officer::fp_text()}} object whose unspecified
arguments inherit from \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} (font family, size, color).

Use \code{fp_text_default()} instead of \code{officer::fp_text()} when building
chunks with \code{\link[=as_chunk]{as_chunk()}}: only override the properties you need,
and the rest will match the table's default style.

See also \code{\link[=set_flextable_defaults]{set_flextable_defaults()}} to modify the inherited values.
}
\examples{
library(flextable)

set_flextable_defaults(
  font.size = 11, font.color = "#303030",
  padding = 3, table.layout = "autofit"
)
z <- flextable(head(cars))

z <- compose(
  x = z,
  i = ~ speed < 6,
  j = "speed",
  value = as_paragraph(
    as_chunk("slow... ", props = fp_text_default(color = "red")),
    as_chunk(speed, props = fp_text_default(italic = TRUE))
  )
)
z

init_flextable_defaults()

}
\seealso{
\code{\link[=as_chunk]{as_chunk()}}, \code{\link[=compose]{compose()}}, \code{\link[=append_chunks]{append_chunks()}}, \code{\link[=prepend_chunks]{prepend_chunks()}}

Other default formatting properties:
\code{\link[=fp_border_default]{fp_border_default()}}
}
\concept{default_formatting_properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_flextable.R
\name{as_flextable.htest}
\alias{as_flextable.htest}
\title{Transform a 'htest' object into a flextable}
\usage{
\method{as_flextable}{htest}(x, ...)
}
\arguments{
\item{x}{htest object}

\item{...}{unused argument}
}
\description{
produce a flextable describing an
object oof class \code{htest}.
}
\examples{
if (require("stats")) {
  M <- as.table(rbind(c(762, 327, 468), c(484, 239, 477)))
  dimnames(M) <- list(
    gender = c("F", "M"),
    party = c("Democrat", "Independent", "Republican")
  )
  ft_1 <- as_flextable(chisq.test(M))
  ft_1
}
}
\seealso{
Other as_flextable methods:
\code{\link[=as_flextable.compact_summary]{as_flextable.compact_summary()}},
\code{\link[=as_flextable.data.frame]{as_flextable.data.frame()}},
\code{\link[=as_flextable.gam]{as_flextable.gam()}},
\code{\link[=as_flextable.glm]{as_flextable.glm()}},
\code{\link[=as_flextable.grouped_data]{as_flextable.grouped_data()}},
\code{\link[=as_flextable.kmeans]{as_flextable.kmeans()}},
\code{\link[=as_flextable.lm]{as_flextable.lm()}},
\code{\link[=as_flextable.merMod]{as_flextable.merMod()}},
\code{\link[=as_flextable.pam]{as_flextable.pam()}},
\code{\link[=as_flextable.summarizor]{as_flextable.summarizor()}},
\code{\link[=as_flextable.table]{as_flextable.table()}},
\code{\link[=as_flextable.tabular]{as_flextable.tabular()}},
\code{\link[=as_flextable.tabulator]{as_flextable.tabulator()}},
\code{\link[=as_flextable.xtable]{as_flextable.xtable()}},
\code{\link[=compact_summary]{compact_summary()}}
}
\concept{as_flextable methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_content.R
\name{as_paragraph}
\alias{as_paragraph}
\title{Build a paragraph from chunks}
\usage{
as_paragraph(..., list_values = NULL)
}
\arguments{
\item{...}{chunk elements that are defining the paragraph content.
If a character is used, it is transformed to a chunk object with
function \code{\link[=as_chunk]{as_chunk()}}.}

\item{list_values}{a list of chunk elements that are defining
the paragraph content. If specified argument \code{...} is unused.}
}
\description{
\code{as_paragraph()} assembles one or more chunks into a single paragraph
that defines the content of a flextable cell. Each cell in a flextable
contains exactly one paragraph; a paragraph is an ordered sequence of
chunks.

Chunks are the smallest content units and can be created with
\code{\link[=as_chunk]{as_chunk()}} (formatted text), \code{\link[=as_b]{as_b()}} / \code{\link[=as_i]{as_i()}} (bold / italic
shortcuts), \code{\link[=minibar]{minibar()}} (inline bar), \code{\link[=as_image]{as_image()}} (image),
\code{\link[=gg_chunk]{gg_chunk()}} (ggplot), \code{\link[=as_equation]{as_equation()}} (equation) or
\code{\link[=hyperlink_text]{hyperlink_text()}} (link). Plain character strings passed to
\code{as_paragraph()} are automatically converted to chunks via
\code{\link[=as_chunk]{as_chunk()}}.

The resulting paragraph is passed to the \code{value} argument of
\code{\link[=compose]{compose()}}, \code{\link[=mk_par]{mk_par()}}, \code{\link[=add_header_lines]{add_header_lines()}},
\code{\link[=add_footer_lines]{add_footer_lines()}} or \code{\link[=footnote]{footnote()}} to set cell content.
}
\examples{
library(flextable)
ft <- flextable(airquality[sample.int(150, size = 10), ])
ft <- compose(ft,
  j = "Wind",
  value = as_paragraph(
    as_chunk(Wind, props = fp_text_default(color = "orange")),
    " ",
    minibar(value = Wind, max = max(airquality$Wind), barcol = "orange", bg = "black", height = .15)
  ),
  part = "body"
)
ft <- autofit(ft)
ft
}
\seealso{
\code{\link[=as_chunk]{as_chunk()}}, \code{\link[=minibar]{minibar()}},
\code{\link[=as_image]{as_image()}}, \code{\link[=hyperlink_text]{hyperlink_text()}}

Other functions to compose cell content:
\code{\link[=append_chunks]{append_chunks()}},
\code{\link[=compose]{compose()}},
\code{\link[=footnote]{footnote()}},
\code{\link[=labelizor]{labelizor()}},
\code{\link[=prepend_chunks]{prepend_chunks()}},
\code{\link[=void]{void()}}
}
\concept{cell_content_composition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_flextable.R
\name{merge_none}
\alias{merge_none}
\title{Delete flextable merging information}
\usage{
merge_none(x, part = "all")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Delete all merging information from a flextable.
}
\examples{
typology <- data.frame(
  col_keys = c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Species"),
  what = c("Sepal", "Sepal", "Petal", "Petal", "Species"),
  measure = c("Length", "Width", "Length", "Width", "Species"),
  stringsAsFactors = FALSE
)

ft <- flextable(head(iris))
ft <- set_header_df(ft, mapping = typology, key = "col_keys")
ft <- merge_v(ft, j = c("Species"))

ft <- theme_tron_legacy(merge_none(ft))
ft
}
\seealso{
Other cell merging functions:
\code{\link[=merge_at]{merge_at()}},
\code{\link[=merge_h]{merge_h()}},
\code{\link[=merge_h_range]{merge_h_range()}},
\code{\link[=merge_v]{merge_v()}}
}
\concept{cell merging functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flextable_sizes.R
\name{width}
\alias{width}
\title{Set columns width}
\usage{
width(x, j = NULL, width, unit = "in")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{j}{column selector, see section \emph{Column selection with the \code{j} parameter}
in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.}

\item{width}{width in inches}

\item{unit}{unit for width, one of "in", "cm", "mm".}
}
\description{
Defines the widths of one or more columns in the
table. This function will have no effect if you have
used \code{set_table_properties(layout = "autofit")}.

\code{\link[=set_table_properties]{set_table_properties()}} can provide an alternative to fixed-width layouts
that is supported with HTML and Word output that can be set
with \code{set_table_properties(layout = "autofit")}.
}
\details{
Heights are not used when flextable is been rendered into HTML.
}
\examples{

ft <- flextable(head(iris))
ft <- width(ft, width = 1.5)
ft
}
\seealso{
Other functions for flextable size management:
\code{\link[=autofit]{autofit()}},
\code{\link[=dim.flextable]{dim.flextable()}},
\code{\link[=dim_pretty]{dim_pretty()}},
\code{\link[=fit_columns]{fit_columns()}},
\code{\link[=fit_to_width]{fit_to_width()}},
\code{\link[=flextable_dim]{flextable_dim()}},
\code{\link[=height]{height()}},
\code{\link[=hrule]{hrule()}},
\code{\link[=ncol_keys]{ncol_keys()}},
\code{\link[=nrow_part]{nrow_part()}},
\code{\link[=set_table_properties]{set_table_properties()}}
}
\concept{functions for flextable size management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/styles.R
\name{empty_blanks}
\alias{empty_blanks}
\title{Make blank columns transparent}
\usage{
empty_blanks(x, width = 0.05, unit = "in", part = "all")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{width}{width of blank columns (.1 inch by default).}

\item{unit}{unit for width, one of "in", "cm", "mm".}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' can be used.}
}
\description{
Blank columns are set as transparent. This is a shortcut function
that deletes top and bottom borders, changes the background color to
transparent, displays empty content, and sets blank column widths.
}
\examples{
typology <- data.frame(
  col_keys = c(
    "Sepal.Length", "Sepal.Width", "Petal.Length",
    "Petal.Width", "Species"
  ),
  what = c("Sepal", "Sepal", "Petal", "Petal", " "),
  measure = c("Length", "Width", "Length", "Width", "Species"),
  stringsAsFactors = FALSE
)
typology

ftab <- flextable(head(iris), col_keys = c(
  "Species",
  "break1", "Sepal.Length", "Sepal.Width",
  "break2", "Petal.Length", "Petal.Width"
))
ftab <- set_header_df(ftab, mapping = typology, key = "col_keys")
ftab <- merge_h(ftab, part = "header")
ftab <- theme_vanilla(ftab)
ftab <- empty_blanks(ftab)
ftab <- width(ftab, j = c(2, 5), width = .1)
ftab
}
\seealso{
Other formatting shortcuts:
\code{\link[=align]{align()}},
\code{\link[=bg]{bg()}},
\code{\link[=bold]{bold()}},
\code{\link[=color]{color()}},
\code{\link[=font]{font()}},
\code{\link[=fontsize]{fontsize()}},
\code{\link[=highlight]{highlight()}},
\code{\link[=indentation]{indentation()}},
\code{\link[=italic]{italic()}},
\code{\link[=keep_with_next]{keep_with_next()}},
\code{\link[=line_spacing]{line_spacing()}},
\code{\link[=padding]{padding()}},
\code{\link[=rotate]{rotate()}},
\code{\link[=style]{style()}},
\code{\link[=tab_settings]{tab_settings()}},
\code{\link[=valign]{valign()}}
}
\concept{formatting_shortcuts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{set_flextable_defaults}
\alias{set_flextable_defaults}
\alias{init_flextable_defaults}
\title{Modify flextable defaults formatting properties}
\usage{
set_flextable_defaults(
  font.family = NULL,
  font.size = NULL,
  font.color = NULL,
  text.align = NULL,
  padding = NULL,
  padding.bottom = NULL,
  padding.top = NULL,
  padding.left = NULL,
  padding.right = NULL,
  border.color = NULL,
  border.width = NULL,
  background.color = NULL,
  line_spacing = NULL,
  table.layout = NULL,
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  decimal.mark = NULL,
  big.mark = NULL,
  digits = NULL,
  pct_digits = NULL,
  na_str = NULL,
  nan_str = NULL,
  fmt_date = NULL,
  fmt_datetime = NULL,
  extra_css = NULL,
  scroll = NULL,
  table_align = "center",
  split = NULL,
  keep_with_next = NULL,
  tabcolsep = NULL,
  arraystretch = NULL,
  float = NULL,
  fonts_ignore = NULL,
  theme_fun = NULL,
  post_process_all = NULL,
  post_process_pdf = NULL,
  post_process_docx = NULL,
  post_process_html = NULL,
  post_process_pptx = NULL,
  ...
)

init_flextable_defaults()
}
\arguments{
\item{font.family}{single character value. When format is Word, it specifies the font to
be used to format characters in the Unicode range (U+0000-U+007F). If you
want to use non ascii characters in Word, you should also set \code{hansi.family}
to the same family name.}

\item{font.size}{font size (in point) - 0 or positive integer value.}

\item{font.color}{font color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{text.align}{text alignment - a single character value, expected value
is one of 'left', 'right', 'center', 'justify'.}

\item{padding}{padding (shortcut for top, bottom, left and right padding)}

\item{padding.bottom, padding.top, padding.left, padding.right}{paragraph paddings - 0 or
positive integer value.}

\item{border.color}{border color - single character value
(e.g. "#000000" or "black").}

\item{border.width}{border width in points.}

\item{background.color}{cell background color - a single character value specifying a
valid color (e.g. "#000000" or "black").}

\item{line_spacing}{space between lines of text, 1 is single line spacing, 2 is double line spacing.}

\item{table.layout}{'autofit' or 'fixed' algorithm. Default to 'autofit'.}

\item{cs.family}{optional and only for Word. Font to be used to format
characters in a complex script Unicode range. For example, Arabic
text might be displayed using the "Arial Unicode MS" font.}

\item{eastasia.family}{optional and only for Word. Font to be used to
format characters in an East Asian Unicode range. For example,
Japanese text might be displayed using the "MS Mincho" font.}

\item{hansi.family}{optional and only for Word. Font to be used to format
characters in a Unicode range which does not fall into one of the
other categories.}

\item{decimal.mark, big.mark, na_str, nan_str}{\link{formatC} arguments used by \code{\link[=colformat_num]{colformat_num()}},
\code{\link[=colformat_double]{colformat_double()}}, and \code{\link[=colformat_int]{colformat_int()}}.}

\item{digits}{\link{formatC} argument used by \code{\link[=colformat_double]{colformat_double()}}.}

\item{pct_digits}{number of digits for percentages.}

\item{fmt_date, fmt_datetime}{formats for date and datetime columns as
documented in \code{\link[=strptime]{strptime()}}. Default to '\%Y-\%m-\%d' and '\%Y-\%m-\%d \%H:\%M:\%S'.}

\item{extra_css}{css instructions to be integrated with the table.}

\item{scroll}{NULL or a list if you want to add a scroll-box.
See \strong{scroll} element of argument \code{opts_html} in function \code{\link[=set_table_properties]{set_table_properties()}}.}

\item{table_align}{default flextable alignment, supported values are 'left', 'center'
and 'right'.}

\item{split}{Word option 'Allow row to break across pages' can be
activated when TRUE.}

\item{keep_with_next}{default initialization value used by the \code{\link[=paginate]{paginate()}}
function corresponding to the Word option "keep rows together" that will
be defined in the array.}

\item{tabcolsep}{space between the text and the left/right border of its containing
cell.}

\item{arraystretch}{height of each row relative to its default
height, the default value is 1.5.}

\item{float}{type of floating placement in the PDF document, one of:
\itemize{
\item 'none' (the default value), table is placed after the preceding
paragraph.
\item 'float', table can float to a place in the text where it fits best
\item 'wrap-r', wrap text around the table positioned to the right side of the text
\item 'wrap-l', wrap text around the table positioned to the left side of the text
\item 'wrap-i', wrap text around the table positioned inside edge-near the binding
\item 'wrap-o', wrap text around the table positioned outside edge-far from the binding
}}

\item{fonts_ignore}{if TRUE, pdf-engine pdflatex can be used instead of
xelatex or lualatex. If pdflatex is used, fonts will be ignored because they are
not supported by pdflatex, whereas with the xelatex and lualatex engines they are.}

\item{theme_fun}{a single character value (the name of the theme function
to be applied) or a theme function (input is a flextable, output is a flextable).}

\item{post_process_all}{Post-processing function
that will allow you to customize the the table. It will be executed before
call to post_process_pdf(), post_process_docx(),
post_process_html(), post_process_pptx().}

\item{post_process_pdf, post_process_docx, post_process_html, post_process_pptx}{Post-processing functions
that will allow you to customize the display by output type (pdf, html, docx, pptx).
They are executed just before printing the table.}

\item{...}{unused or deprecated arguments}
}
\value{
a list containing previous default values.
}
\description{
The current formatting properties (see \code{\link[=get_flextable_defaults]{get_flextable_defaults()}})
are automatically applied to every flextable you produce.
Use \code{set_flextable_defaults()} to override them. Use \code{init_flextable_defaults()}
to re-init all values with the package defaults.
}
\examples{
ft_1 <- qflextable(head(airquality))
ft_1

old <- set_flextable_defaults(
  font.color = "#AA8855",
  border.color = "#8855AA"
)
ft_2 <- qflextable(head(airquality))
ft_2

do.call(set_flextable_defaults, old)
}
\seealso{
Other themes and defaults:
\code{\link[=get_flextable_defaults]{get_flextable_defaults()}},
\code{\link[=theme_alafoli]{theme_alafoli()}},
\code{\link[=theme_apa]{theme_apa()}},
\code{\link[=theme_booktabs]{theme_booktabs()}},
\code{\link[=theme_borderless]{theme_borderless()}},
\code{\link[=theme_box]{theme_box()}},
\code{\link[=theme_tron]{theme_tron()}},
\code{\link[=theme_tron_legacy]{theme_tron_legacy()}},
\code{\link[=theme_vader]{theme_vader()}},
\code{\link[=theme_vanilla]{theme_vanilla()}},
\code{\link[=theme_zebra]{theme_zebra()}}
}
\concept{themes_and_defaults}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{set_formatter}
\alias{set_formatter}
\title{Set column formatter functions}
\usage{
set_formatter(x, ..., values = NULL, part = "body")
}
\arguments{
\item{x}{a 'flextable' object, see \link{flextable-package} to learn how to create
'flextable' object.}

\item{...}{Name-value pairs of functions, names should be existing col_key values}

\item{values}{format functions, If values is supplied argument \code{...} is ignored.
\itemize{
\item It can be a list of name-value pairs of functions, names
should be existing col_key values.
\item If \code{values} is a single function, it will be applied to each column.
}}

\item{part}{part selector, see section \emph{Part selection with the \code{part}
parameter} in <\code{\link[=flextable_selectors]{Selectors in flextable}}>.
Value 'all' is not allowed by the function.}
}
\description{
Apply formatter functions to column keys.

Functions should have a single argument (the vector) and should
return the formatted values as a character vector.
}
\examples{
ft <- flextable(head(iris))
ft <- set_formatter(
  x = ft,
  Sepal.Length = function(x) sprintf("\%.02f", x),
  Sepal.Width = function(x) sprintf("\%.04f", x)
)
ft <- theme_vanilla(ft)
ft
}
\seealso{
Other cells formatters:
\code{\link[=colformat_char]{colformat_char()}},
\code{\link[=colformat_date]{colformat_date()}},
\code{\link[=colformat_datetime]{colformat_datetime()}},
\code{\link[=colformat_double]{colformat_double()}},
\code{\link[=colformat_image]{colformat_image()}},
\code{\link[=colformat_int]{colformat_int()}},
\code{\link[=colformat_lgl]{colformat_lgl()}},
\code{\link[=colformat_num]{colformat_num()}}
}
\concept{cells formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/latex_str.R
\name{add_latex_dep}
\alias{add_latex_dep}
\title{Add latex dependencies}
\usage{
add_latex_dep(float = FALSE, wrapfig = FALSE)
}
\arguments{
\item{float}{load package 'float'}

\item{wrapfig}{load package 'wrapfig'}
}
\description{
Manually add flextable latex dependencies to
the knitr session via \code{\link[knitr:knit_meta_add]{knitr::knit_meta_add()}}.

When enabling caching in 'R Markdown' documents for PDF output,
the flextable cached result is used directly. Call \code{add_latex_dep()} in a
non cached chunk so that flextable latex dependencies are added
to knitr metadata.
}
\examples{
add_latex_dep()
}
\seealso{
\code{\link[=knit_print.flextable]{knit_print.flextable()}}

Other flextable configuration:
\code{\link[=use_df_printer]{use_df_printer()}},
\code{\link[=use_flextable_qmd]{use_flextable_qmd()}},
\code{\link[=use_model_printer]{use_model_printer()}}
}
\concept{flextable_configuration}
\keyword{internal}
